#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#define LANGUAGE_VERSION 14
#define STATE_COUNT 2654
#define LARGE_STATE_COUNT 584
#define SYMBOL_COUNT 425
#define ALIAS_COUNT 1
#define TOKEN_COUNT 193
#define EXTERNAL_TOKEN_COUNT 12
#define FIELD_COUNT 31
#define MAX_ALIAS_SEQUENCE_LENGTH 12
#define PRODUCTION_ID_COUNT 217

enum ts_symbol_identifiers {
  sym_name = 1,
  anon_sym_QMARK_GT = 2,
  sym_php_tag = 3,
  anon_sym_SEMI = 4,
  anon_sym_AMP = 5,
  aux_sym_function_static_declaration_token1 = 6,
  anon_sym_COMMA = 7,
  anon_sym_EQ = 8,
  aux_sym_global_declaration_token1 = 9,
  aux_sym_namespace_definition_token1 = 10,
  aux_sym_namespace_use_declaration_token1 = 11,
  aux_sym_namespace_use_clause_token1 = 12,
  aux_sym__namespace_use_type_token1 = 13,
  aux_sym__namespace_use_type_token2 = 14,
  anon_sym_BSLASH = 15,
  anon_sym_LBRACE = 16,
  anon_sym_RBRACE = 17,
  aux_sym_trait_declaration_token1 = 18,
  aux_sym_interface_declaration_token1 = 19,
  aux_sym_base_clause_token1 = 20,
  aux_sym_enum_declaration_token1 = 21,
  anon_sym_COLON = 22,
  anon_sym_string = 23,
  anon_sym_int = 24,
  aux_sym_enum_case_token1 = 25,
  aux_sym_class_declaration_token1 = 26,
  aux_sym_final_modifier_token1 = 27,
  aux_sym_abstract_modifier_token1 = 28,
  aux_sym_readonly_modifier_token1 = 29,
  aux_sym_class_interface_clause_token1 = 30,
  anon_sym_EQ_GT = 31,
  sym_var_modifier = 32,
  aux_sym_use_instead_of_clause_token1 = 33,
  aux_sym_visibility_modifier_token1 = 34,
  aux_sym_visibility_modifier_token2 = 35,
  aux_sym_visibility_modifier_token3 = 36,
  anon_sym_LPAREN = 37,
  anon_sym_RPAREN = 38,
  anon_sym_LPAREN2 = 39,
  anon_sym_RPAREN2 = 40,
  aux_sym__arrow_function_header_token1 = 41,
  anon_sym_DOT_DOT_DOT = 42,
  anon_sym_QMARK = 43,
  sym_bottom_type = 44,
  anon_sym_PIPE = 45,
  anon_sym_array = 46,
  aux_sym_primitive_type_token1 = 47,
  anon_sym_iterable = 48,
  anon_sym_bool = 49,
  anon_sym_float = 50,
  anon_sym_void = 51,
  anon_sym_mixed = 52,
  anon_sym_false = 53,
  anon_sym_null = 54,
  anon_sym_true = 55,
  aux_sym_cast_type_token1 = 56,
  aux_sym_cast_type_token2 = 57,
  aux_sym_cast_type_token3 = 58,
  aux_sym_cast_type_token4 = 59,
  aux_sym_cast_type_token5 = 60,
  aux_sym_cast_type_token6 = 61,
  aux_sym_cast_type_token7 = 62,
  aux_sym_cast_type_token8 = 63,
  aux_sym_cast_type_token9 = 64,
  aux_sym_cast_type_token10 = 65,
  aux_sym_cast_type_token11 = 66,
  aux_sym_cast_type_token12 = 67,
  aux_sym_echo_statement_token1 = 68,
  aux_sym_exit_statement_token1 = 69,
  anon_sym_unset = 70,
  aux_sym_declare_statement_token1 = 71,
  aux_sym_declare_statement_token2 = 72,
  anon_sym_ticks = 73,
  anon_sym_encoding = 74,
  anon_sym_strict_types = 75,
  sym_float = 76,
  aux_sym_try_statement_token1 = 77,
  aux_sym_catch_clause_token1 = 78,
  aux_sym_finally_clause_token1 = 79,
  aux_sym_goto_statement_token1 = 80,
  aux_sym_continue_statement_token1 = 81,
  aux_sym_break_statement_token1 = 82,
  sym_integer = 83,
  aux_sym_return_statement_token1 = 84,
  aux_sym_throw_expression_token1 = 85,
  aux_sym_while_statement_token1 = 86,
  aux_sym_while_statement_token2 = 87,
  aux_sym_do_statement_token1 = 88,
  aux_sym_for_statement_token1 = 89,
  aux_sym_for_statement_token2 = 90,
  aux_sym_foreach_statement_token1 = 91,
  aux_sym_foreach_statement_token2 = 92,
  aux_sym_if_statement_token1 = 93,
  aux_sym_if_statement_token2 = 94,
  aux_sym_else_if_clause_token1 = 95,
  aux_sym_else_clause_token1 = 96,
  aux_sym_match_expression_token1 = 97,
  aux_sym_match_default_expression_token1 = 98,
  aux_sym_switch_statement_token1 = 99,
  aux_sym_switch_block_token1 = 100,
  anon_sym_PLUS = 101,
  anon_sym_DASH = 102,
  anon_sym_TILDE = 103,
  anon_sym_BANG = 104,
  anon_sym_AT = 105,
  aux_sym_clone_expression_token1 = 106,
  anon_sym_COLON_COLON = 107,
  aux_sym_print_intrinsic_token1 = 108,
  aux_sym__new_non_dereferencable_expression_token1 = 109,
  anon_sym_DASH_DASH = 110,
  anon_sym_PLUS_PLUS = 111,
  anon_sym_STAR_STAR_EQ = 112,
  anon_sym_STAR_EQ = 113,
  anon_sym_SLASH_EQ = 114,
  anon_sym_PERCENT_EQ = 115,
  anon_sym_PLUS_EQ = 116,
  anon_sym_DASH_EQ = 117,
  anon_sym_DOT_EQ = 118,
  anon_sym_LT_LT_EQ = 119,
  anon_sym_GT_GT_EQ = 120,
  anon_sym_AMP_EQ = 121,
  anon_sym_CARET_EQ = 122,
  anon_sym_PIPE_EQ = 123,
  anon_sym_QMARK_QMARK_EQ = 124,
  anon_sym_DASH_GT = 125,
  anon_sym_QMARK_DASH_GT = 126,
  aux_sym__list_destructing_token1 = 127,
  anon_sym_LBRACK = 128,
  anon_sym_RBRACK = 129,
  anon_sym_self = 130,
  anon_sym_parent = 131,
  aux_sym__argument_name_token1 = 132,
  aux_sym__argument_name_token2 = 133,
  anon_sym_POUND_LBRACK = 134,
  sym_escape_sequence = 135,
  anon_sym_BSLASHu = 136,
  aux_sym_encapsed_string_token1 = 137,
  anon_sym_DQUOTE = 138,
  aux_sym_string_token1 = 139,
  anon_sym_SQUOTE = 140,
  aux_sym_string_token2 = 141,
  aux_sym_string_content_token1 = 142,
  anon_sym_LT_LT_LT = 143,
  anon_sym_DQUOTE2 = 144,
  sym__new_line = 145,
  anon_sym_SQUOTE2 = 146,
  anon_sym_BQUOTE = 147,
  anon_sym_DOLLAR = 148,
  aux_sym_yield_expression_token1 = 149,
  aux_sym_yield_expression_token2 = 150,
  aux_sym_binary_expression_token1 = 151,
  anon_sym_QMARK_QMARK = 152,
  anon_sym_STAR_STAR = 153,
  aux_sym_binary_expression_token2 = 154,
  aux_sym_binary_expression_token3 = 155,
  aux_sym_binary_expression_token4 = 156,
  anon_sym_PIPE_PIPE = 157,
  anon_sym_AMP_AMP = 158,
  anon_sym_CARET = 159,
  anon_sym_EQ_EQ = 160,
  anon_sym_BANG_EQ = 161,
  anon_sym_LT_GT = 162,
  anon_sym_EQ_EQ_EQ = 163,
  anon_sym_BANG_EQ_EQ = 164,
  anon_sym_LT = 165,
  anon_sym_GT = 166,
  anon_sym_LT_EQ = 167,
  anon_sym_GT_EQ = 168,
  anon_sym_LT_EQ_GT = 169,
  anon_sym_LT_LT = 170,
  anon_sym_GT_GT = 171,
  anon_sym_DOT = 172,
  anon_sym_STAR = 173,
  anon_sym_SLASH = 174,
  anon_sym_PERCENT = 175,
  aux_sym_include_expression_token1 = 176,
  aux_sym_include_once_expression_token1 = 177,
  aux_sym_require_expression_token1 = 178,
  aux_sym_require_once_expression_token1 = 179,
  sym_comment = 180,
  sym__automatic_semicolon = 181,
  sym_encapsed_string_chars = 182,
  sym_encapsed_string_chars_after_variable = 183,
  sym_execution_string_chars = 184,
  sym_execution_string_chars_after_variable = 185,
  sym_encapsed_string_chars_heredoc = 186,
  sym_encapsed_string_chars_after_variable_heredoc = 187,
  sym__eof = 188,
  sym_heredoc_start = 189,
  sym_heredoc_end = 190,
  sym_nowdoc_string = 191,
  sym_sentinel_error = 192,
  sym_program = 193,
  sym_statement = 194,
  sym_empty_statement = 195,
  sym_reference_modifier = 196,
  sym_function_static_declaration = 197,
  sym_static_variable_declaration = 198,
  sym_global_declaration = 199,
  sym_namespace_definition = 200,
  sym_namespace_use_declaration = 201,
  sym_namespace_use_clause = 202,
  sym_qualified_name = 203,
  sym__name = 204,
  sym_namespace_name = 205,
  sym__namespace_use_group = 206,
  sym_namespace_use_group = 207,
  sym_trait_declaration = 208,
  sym_interface_declaration = 209,
  sym_base_clause = 210,
  sym_enum_declaration = 211,
  sym_enum_declaration_list = 212,
  sym__enum_member_declaration = 213,
  sym_enum_case = 214,
  sym_class_declaration = 215,
  sym_declaration_list = 216,
  sym_final_modifier = 217,
  sym_abstract_modifier = 218,
  sym_readonly_modifier = 219,
  sym_class_interface_clause = 220,
  sym__member_declaration = 221,
  sym_const_declaration = 222,
  sym__class_const_declaration = 223,
  sym__const_declaration = 224,
  sym_property_declaration = 225,
  sym__modifier = 226,
  sym_property_element = 227,
  sym_property_hook_list = 228,
  sym_property_hook = 229,
  sym__property_hook_body = 230,
  sym_method_declaration = 231,
  sym_static_modifier = 232,
  sym_use_declaration = 233,
  sym_use_list = 234,
  sym_use_instead_of_clause = 235,
  sym_use_as_clause = 236,
  sym_visibility_modifier = 237,
  sym_function_definition = 238,
  sym__function_definition_header = 239,
  sym_anonymous_function = 240,
  sym_anonymous_function_use_clause = 241,
  sym__anonymous_function_header = 242,
  sym__arrow_function_header = 243,
  sym_arrow_function = 244,
  sym_formal_parameters = 245,
  sym_property_promotion_parameter = 246,
  sym_simple_parameter = 247,
  sym_variadic_parameter = 248,
  sym_type = 249,
  sym__types = 250,
  sym_named_type = 251,
  sym_optional_type = 252,
  sym_union_type = 253,
  sym_intersection_type = 254,
  sym_disjunctive_normal_form_type = 255,
  sym_primitive_type = 256,
  sym_cast_type = 257,
  sym__return_type = 258,
  sym_const_element = 259,
  sym_echo_statement = 260,
  sym_exit_statement = 261,
  sym_unset_statement = 262,
  sym_declare_statement = 263,
  sym_declare_directive = 264,
  sym_literal = 265,
  sym_try_statement = 266,
  sym_catch_clause = 267,
  sym_type_list = 268,
  sym_finally_clause = 269,
  sym_goto_statement = 270,
  sym_continue_statement = 271,
  sym_break_statement = 272,
  sym_return_statement = 273,
  sym_throw_expression = 274,
  sym_while_statement = 275,
  sym_do_statement = 276,
  sym_for_statement = 277,
  sym__expressions = 278,
  sym_sequence_expression = 279,
  sym_foreach_statement = 280,
  sym_foreach_pair = 281,
  sym__foreach_value = 282,
  sym_if_statement = 283,
  sym_colon_block = 284,
  sym_else_if_clause = 285,
  sym_else_clause = 286,
  sym_else_if_clause_2 = 287,
  sym_else_clause_2 = 288,
  sym_match_expression = 289,
  sym_match_block = 290,
  sym_match_condition_list = 291,
  sym_match_conditional_expression = 292,
  sym_match_default_expression = 293,
  sym_switch_statement = 294,
  sym_switch_block = 295,
  sym_case_statement = 296,
  sym_default_statement = 297,
  sym_compound_statement = 298,
  sym_named_label_statement = 299,
  sym_expression_statement = 300,
  sym_expression = 301,
  sym__unary_expression = 302,
  sym_unary_op_expression = 303,
  sym_error_suppression_expression = 304,
  sym_clone_expression = 305,
  sym_primary_expression = 306,
  sym_parenthesized_expression = 307,
  sym_class_constant_access_expression = 308,
  sym_print_intrinsic = 309,
  sym_object_creation_expression = 310,
  sym__new_non_dereferencable_expression = 311,
  sym__new_dereferencable_expression = 312,
  sym__class_name_reference = 313,
  sym_anonymous_class = 314,
  sym_update_expression = 315,
  sym_cast_expression = 316,
  sym_cast_variable = 317,
  sym_assignment_expression = 318,
  sym_reference_assignment_expression = 319,
  sym_conditional_expression = 320,
  sym_augmented_assignment_expression = 321,
  sym__variable_member_access_expression = 322,
  sym_member_access_expression = 323,
  sym__variable_nullsafe_member_access_expression = 324,
  sym_nullsafe_member_access_expression = 325,
  sym__variable_scoped_property_access_expression = 326,
  sym_scoped_property_access_expression = 327,
  sym_list_literal = 328,
  sym__list_destructing = 329,
  sym__array_destructing = 330,
  sym__array_destructing_element = 331,
  sym_function_call_expression = 332,
  sym__callable_expression = 333,
  sym_scoped_call_expression = 334,
  sym__scope_resolution_qualifier = 335,
  sym_relative_scope = 336,
  sym_variadic_placeholder = 337,
  sym_arguments = 338,
  sym_argument = 339,
  sym__argument_name = 340,
  sym_member_call_expression = 341,
  sym_nullsafe_member_call_expression = 342,
  sym_variadic_unpacking = 343,
  sym__member_name = 344,
  sym__variable_subscript_expression = 345,
  sym__dereferencable_subscript_expression = 346,
  sym__dereferencable_expression = 347,
  sym__dereferencable_scalar = 348,
  sym_array_creation_expression = 349,
  sym_attribute_group = 350,
  sym_attribute_list = 351,
  sym_attribute = 352,
  sym__complex_string_part = 353,
  sym__simple_string_member_access_expression = 354,
  sym__simple_string_subscript_unary_expression = 355,
  sym__simple_string_array_access_argument = 356,
  sym__simple_string_subscript_expression = 357,
  sym__simple_string_part = 358,
  aux_sym__interpolated_string_body = 359,
  aux_sym__interpolated_string_body_heredoc = 360,
  sym_encapsed_string = 361,
  sym_string = 362,
  sym_string_content = 363,
  sym_heredoc_body = 364,
  sym_heredoc = 365,
  sym_nowdoc_body = 366,
  sym_nowdoc = 367,
  aux_sym__interpolated_execution_operator_body = 368,
  sym_shell_command_expression = 369,
  sym_boolean = 370,
  sym_null = 371,
  sym__string = 372,
  sym_dynamic_variable_name = 373,
  sym__simple_variable = 374,
  sym__new_variable = 375,
  sym__callable_variable = 376,
  sym_variable_name = 377,
  sym_by_ref = 378,
  sym_yield_expression = 379,
  sym_array_element_initializer = 380,
  sym_binary_expression = 381,
  sym_include_expression = 382,
  sym_include_once_expression = 383,
  sym_require_expression = 384,
  sym_require_once_expression = 385,
  sym__reserved_identifier = 386,
  sym__identifier = 387,
  sym__semicolon = 388,
  aux_sym_program_repeat1 = 389,
  aux_sym_function_static_declaration_repeat1 = 390,
  aux_sym_global_declaration_repeat1 = 391,
  aux_sym_namespace_use_declaration_repeat1 = 392,
  aux_sym_namespace_name_repeat1 = 393,
  aux_sym_base_clause_repeat1 = 394,
  aux_sym_enum_declaration_list_repeat1 = 395,
  aux_sym_class_declaration_repeat1 = 396,
  aux_sym_declaration_list_repeat1 = 397,
  aux_sym__const_declaration_repeat1 = 398,
  aux_sym_property_declaration_repeat1 = 399,
  aux_sym_property_hook_list_repeat1 = 400,
  aux_sym_use_list_repeat1 = 401,
  aux_sym_anonymous_function_use_clause_repeat1 = 402,
  aux_sym_formal_parameters_repeat1 = 403,
  aux_sym_union_type_repeat1 = 404,
  aux_sym_intersection_type_repeat1 = 405,
  aux_sym_disjunctive_normal_form_type_repeat1 = 406,
  aux_sym_unset_statement_repeat1 = 407,
  aux_sym_try_statement_repeat1 = 408,
  aux_sym_type_list_repeat1 = 409,
  aux_sym_if_statement_repeat1 = 410,
  aux_sym_if_statement_repeat2 = 411,
  aux_sym_match_block_repeat1 = 412,
  aux_sym_match_condition_list_repeat1 = 413,
  aux_sym_switch_block_repeat1 = 414,
  aux_sym__list_destructing_repeat1 = 415,
  aux_sym__array_destructing_repeat1 = 416,
  aux_sym_arguments_repeat1 = 417,
  aux_sym_array_creation_expression_repeat1 = 418,
  aux_sym_attribute_group_repeat1 = 419,
  aux_sym_attribute_list_repeat1 = 420,
  aux_sym_string_repeat1 = 421,
  aux_sym_string_content_repeat1 = 422,
  aux_sym_heredoc_body_repeat1 = 423,
  aux_sym_nowdoc_body_repeat1 = 424,
  alias_sym_operation = 425,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_name] = "name",
  [anon_sym_QMARK_GT] = "\?>",
  [sym_php_tag] = "php_tag",
  [anon_sym_SEMI] = ";",
  [anon_sym_AMP] = "&",
  [aux_sym_function_static_declaration_token1] = "static",
  [anon_sym_COMMA] = ",",
  [anon_sym_EQ] = "=",
  [aux_sym_global_declaration_token1] = "global",
  [aux_sym_namespace_definition_token1] = "namespace",
  [aux_sym_namespace_use_declaration_token1] = "use",
  [aux_sym_namespace_use_clause_token1] = "as",
  [aux_sym__namespace_use_type_token1] = "function",
  [aux_sym__namespace_use_type_token2] = "const",
  [anon_sym_BSLASH] = "\\",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [aux_sym_trait_declaration_token1] = "trait",
  [aux_sym_interface_declaration_token1] = "interface",
  [aux_sym_base_clause_token1] = "extends",
  [aux_sym_enum_declaration_token1] = "enum",
  [anon_sym_COLON] = ":",
  [anon_sym_string] = "string",
  [anon_sym_int] = "int",
  [aux_sym_enum_case_token1] = "case",
  [aux_sym_class_declaration_token1] = "class",
  [aux_sym_final_modifier_token1] = "final",
  [aux_sym_abstract_modifier_token1] = "abstract",
  [aux_sym_readonly_modifier_token1] = "readonly",
  [aux_sym_class_interface_clause_token1] = "implements",
  [anon_sym_EQ_GT] = "=>",
  [sym_var_modifier] = "var_modifier",
  [aux_sym_use_instead_of_clause_token1] = "insteadof",
  [aux_sym_visibility_modifier_token1] = "public",
  [aux_sym_visibility_modifier_token2] = "protected",
  [aux_sym_visibility_modifier_token3] = "private",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_LPAREN2] = "(",
  [anon_sym_RPAREN2] = ")",
  [aux_sym__arrow_function_header_token1] = "fn",
  [anon_sym_DOT_DOT_DOT] = "...",
  [anon_sym_QMARK] = "\?",
  [sym_bottom_type] = "bottom_type",
  [anon_sym_PIPE] = "|",
  [anon_sym_array] = "array",
  [aux_sym_primitive_type_token1] = "callable",
  [anon_sym_iterable] = "iterable",
  [anon_sym_bool] = "bool",
  [anon_sym_float] = "float",
  [anon_sym_void] = "void",
  [anon_sym_mixed] = "mixed",
  [anon_sym_false] = "false",
  [anon_sym_null] = "null",
  [anon_sym_true] = "true",
  [aux_sym_cast_type_token1] = "cast_type_token1",
  [aux_sym_cast_type_token2] = "cast_type_token2",
  [aux_sym_cast_type_token3] = "cast_type_token3",
  [aux_sym_cast_type_token4] = "cast_type_token4",
  [aux_sym_cast_type_token5] = "cast_type_token5",
  [aux_sym_cast_type_token6] = "cast_type_token6",
  [aux_sym_cast_type_token7] = "cast_type_token7",
  [aux_sym_cast_type_token8] = "cast_type_token8",
  [aux_sym_cast_type_token9] = "cast_type_token9",
  [aux_sym_cast_type_token10] = "cast_type_token10",
  [aux_sym_cast_type_token11] = "cast_type_token11",
  [aux_sym_cast_type_token12] = "cast_type_token12",
  [aux_sym_echo_statement_token1] = "echo",
  [aux_sym_exit_statement_token1] = "exit",
  [anon_sym_unset] = "unset",
  [aux_sym_declare_statement_token1] = "declare",
  [aux_sym_declare_statement_token2] = "enddeclare",
  [anon_sym_ticks] = "ticks",
  [anon_sym_encoding] = "encoding",
  [anon_sym_strict_types] = "strict_types",
  [sym_float] = "float",
  [aux_sym_try_statement_token1] = "try",
  [aux_sym_catch_clause_token1] = "catch",
  [aux_sym_finally_clause_token1] = "finally",
  [aux_sym_goto_statement_token1] = "goto",
  [aux_sym_continue_statement_token1] = "continue",
  [aux_sym_break_statement_token1] = "break",
  [sym_integer] = "integer",
  [aux_sym_return_statement_token1] = "return",
  [aux_sym_throw_expression_token1] = "throw",
  [aux_sym_while_statement_token1] = "while",
  [aux_sym_while_statement_token2] = "endwhile",
  [aux_sym_do_statement_token1] = "do",
  [aux_sym_for_statement_token1] = "for",
  [aux_sym_for_statement_token2] = "endfor",
  [aux_sym_foreach_statement_token1] = "foreach",
  [aux_sym_foreach_statement_token2] = "endforeach",
  [aux_sym_if_statement_token1] = "if",
  [aux_sym_if_statement_token2] = "endif",
  [aux_sym_else_if_clause_token1] = "elseif",
  [aux_sym_else_clause_token1] = "else",
  [aux_sym_match_expression_token1] = "match",
  [aux_sym_match_default_expression_token1] = "default",
  [aux_sym_switch_statement_token1] = "switch",
  [aux_sym_switch_block_token1] = "endswitch",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_TILDE] = "~",
  [anon_sym_BANG] = "!",
  [anon_sym_AT] = "@",
  [aux_sym_clone_expression_token1] = "clone",
  [anon_sym_COLON_COLON] = "::",
  [aux_sym_print_intrinsic_token1] = "print",
  [aux_sym__new_non_dereferencable_expression_token1] = "new",
  [anon_sym_DASH_DASH] = "--",
  [anon_sym_PLUS_PLUS] = "++",
  [anon_sym_STAR_STAR_EQ] = "**=",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_PERCENT_EQ] = "%=",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_DASH_EQ] = "-=",
  [anon_sym_DOT_EQ] = ".=",
  [anon_sym_LT_LT_EQ] = "<<=",
  [anon_sym_GT_GT_EQ] = ">>=",
  [anon_sym_AMP_EQ] = "&=",
  [anon_sym_CARET_EQ] = "^=",
  [anon_sym_PIPE_EQ] = "|=",
  [anon_sym_QMARK_QMARK_EQ] = "\?\?=",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_QMARK_DASH_GT] = "\?->",
  [aux_sym__list_destructing_token1] = "list",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_self] = "self",
  [anon_sym_parent] = "parent",
  [aux_sym__argument_name_token1] = "_argument_name_token1",
  [aux_sym__argument_name_token2] = "_argument_name_token2",
  [anon_sym_POUND_LBRACK] = "#[",
  [sym_escape_sequence] = "escape_sequence",
  [anon_sym_BSLASHu] = "string_content",
  [aux_sym_encapsed_string_token1] = "encapsed_string_token1",
  [anon_sym_DQUOTE] = "\"",
  [aux_sym_string_token1] = "string_token1",
  [anon_sym_SQUOTE] = "'",
  [aux_sym_string_token2] = "escape_sequence",
  [aux_sym_string_content_token1] = "string_content_token1",
  [anon_sym_LT_LT_LT] = "<<<",
  [anon_sym_DQUOTE2] = "\"",
  [sym__new_line] = "_new_line",
  [anon_sym_SQUOTE2] = "'",
  [anon_sym_BQUOTE] = "`",
  [anon_sym_DOLLAR] = "$",
  [aux_sym_yield_expression_token1] = "yield",
  [aux_sym_yield_expression_token2] = "from",
  [aux_sym_binary_expression_token1] = "instanceof",
  [anon_sym_QMARK_QMARK] = "\?\?",
  [anon_sym_STAR_STAR] = "**",
  [aux_sym_binary_expression_token2] = "and",
  [aux_sym_binary_expression_token3] = "or",
  [aux_sym_binary_expression_token4] = "xor",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_CARET] = "^",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_LT_GT] = "<>",
  [anon_sym_EQ_EQ_EQ] = "===",
  [anon_sym_BANG_EQ_EQ] = "!==",
  [anon_sym_LT] = "<",
  [anon_sym_GT] = ">",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_LT_EQ_GT] = "<=>",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_DOT] = ".",
  [anon_sym_STAR] = "*",
  [anon_sym_SLASH] = "/",
  [anon_sym_PERCENT] = "%",
  [aux_sym_include_expression_token1] = "include",
  [aux_sym_include_once_expression_token1] = "include_once",
  [aux_sym_require_expression_token1] = "require",
  [aux_sym_require_once_expression_token1] = "require_once",
  [sym_comment] = "comment",
  [sym__automatic_semicolon] = "_automatic_semicolon",
  [sym_encapsed_string_chars] = "string_content",
  [sym_encapsed_string_chars_after_variable] = "string_content",
  [sym_execution_string_chars] = "string_content",
  [sym_execution_string_chars_after_variable] = "string_content",
  [sym_encapsed_string_chars_heredoc] = "string_content",
  [sym_encapsed_string_chars_after_variable_heredoc] = "string_content",
  [sym__eof] = "_eof",
  [sym_heredoc_start] = "heredoc_start",
  [sym_heredoc_end] = "heredoc_end",
  [sym_nowdoc_string] = "nowdoc_string",
  [sym_sentinel_error] = "sentinel_error",
  [sym_program] = "program",
  [sym_statement] = "statement",
  [sym_empty_statement] = "empty_statement",
  [sym_reference_modifier] = "reference_modifier",
  [sym_function_static_declaration] = "function_static_declaration",
  [sym_static_variable_declaration] = "static_variable_declaration",
  [sym_global_declaration] = "global_declaration",
  [sym_namespace_definition] = "namespace_definition",
  [sym_namespace_use_declaration] = "namespace_use_declaration",
  [sym_namespace_use_clause] = "namespace_use_clause",
  [sym_qualified_name] = "qualified_name",
  [sym__name] = "_name",
  [sym_namespace_name] = "namespace_name",
  [sym__namespace_use_group] = "_namespace_use_group",
  [sym_namespace_use_group] = "namespace_use_group",
  [sym_trait_declaration] = "trait_declaration",
  [sym_interface_declaration] = "interface_declaration",
  [sym_base_clause] = "base_clause",
  [sym_enum_declaration] = "enum_declaration",
  [sym_enum_declaration_list] = "enum_declaration_list",
  [sym__enum_member_declaration] = "_enum_member_declaration",
  [sym_enum_case] = "enum_case",
  [sym_class_declaration] = "class_declaration",
  [sym_declaration_list] = "declaration_list",
  [sym_final_modifier] = "final_modifier",
  [sym_abstract_modifier] = "abstract_modifier",
  [sym_readonly_modifier] = "readonly_modifier",
  [sym_class_interface_clause] = "class_interface_clause",
  [sym__member_declaration] = "_member_declaration",
  [sym_const_declaration] = "const_declaration",
  [sym__class_const_declaration] = "const_declaration",
  [sym__const_declaration] = "_const_declaration",
  [sym_property_declaration] = "property_declaration",
  [sym__modifier] = "_modifier",
  [sym_property_element] = "property_element",
  [sym_property_hook_list] = "property_hook_list",
  [sym_property_hook] = "property_hook",
  [sym__property_hook_body] = "_property_hook_body",
  [sym_method_declaration] = "method_declaration",
  [sym_static_modifier] = "static_modifier",
  [sym_use_declaration] = "use_declaration",
  [sym_use_list] = "use_list",
  [sym_use_instead_of_clause] = "use_instead_of_clause",
  [sym_use_as_clause] = "use_as_clause",
  [sym_visibility_modifier] = "visibility_modifier",
  [sym_function_definition] = "function_definition",
  [sym__function_definition_header] = "_function_definition_header",
  [sym_anonymous_function] = "anonymous_function",
  [sym_anonymous_function_use_clause] = "anonymous_function_use_clause",
  [sym__anonymous_function_header] = "_anonymous_function_header",
  [sym__arrow_function_header] = "_arrow_function_header",
  [sym_arrow_function] = "arrow_function",
  [sym_formal_parameters] = "formal_parameters",
  [sym_property_promotion_parameter] = "property_promotion_parameter",
  [sym_simple_parameter] = "simple_parameter",
  [sym_variadic_parameter] = "variadic_parameter",
  [sym_type] = "type",
  [sym__types] = "_types",
  [sym_named_type] = "named_type",
  [sym_optional_type] = "optional_type",
  [sym_union_type] = "union_type",
  [sym_intersection_type] = "intersection_type",
  [sym_disjunctive_normal_form_type] = "disjunctive_normal_form_type",
  [sym_primitive_type] = "primitive_type",
  [sym_cast_type] = "cast_type",
  [sym__return_type] = "_return_type",
  [sym_const_element] = "const_element",
  [sym_echo_statement] = "echo_statement",
  [sym_exit_statement] = "exit_statement",
  [sym_unset_statement] = "unset_statement",
  [sym_declare_statement] = "declare_statement",
  [sym_declare_directive] = "declare_directive",
  [sym_literal] = "literal",
  [sym_try_statement] = "try_statement",
  [sym_catch_clause] = "catch_clause",
  [sym_type_list] = "type_list",
  [sym_finally_clause] = "finally_clause",
  [sym_goto_statement] = "goto_statement",
  [sym_continue_statement] = "continue_statement",
  [sym_break_statement] = "break_statement",
  [sym_return_statement] = "return_statement",
  [sym_throw_expression] = "throw_expression",
  [sym_while_statement] = "while_statement",
  [sym_do_statement] = "do_statement",
  [sym_for_statement] = "for_statement",
  [sym__expressions] = "_expressions",
  [sym_sequence_expression] = "sequence_expression",
  [sym_foreach_statement] = "foreach_statement",
  [sym_foreach_pair] = "pair",
  [sym__foreach_value] = "_foreach_value",
  [sym_if_statement] = "if_statement",
  [sym_colon_block] = "colon_block",
  [sym_else_if_clause] = "else_if_clause",
  [sym_else_clause] = "else_clause",
  [sym_else_if_clause_2] = "else_if_clause",
  [sym_else_clause_2] = "else_clause",
  [sym_match_expression] = "match_expression",
  [sym_match_block] = "match_block",
  [sym_match_condition_list] = "match_condition_list",
  [sym_match_conditional_expression] = "match_conditional_expression",
  [sym_match_default_expression] = "match_default_expression",
  [sym_switch_statement] = "switch_statement",
  [sym_switch_block] = "switch_block",
  [sym_case_statement] = "case_statement",
  [sym_default_statement] = "default_statement",
  [sym_compound_statement] = "compound_statement",
  [sym_named_label_statement] = "named_label_statement",
  [sym_expression_statement] = "expression_statement",
  [sym_expression] = "expression",
  [sym__unary_expression] = "_unary_expression",
  [sym_unary_op_expression] = "unary_op_expression",
  [sym_error_suppression_expression] = "error_suppression_expression",
  [sym_clone_expression] = "clone_expression",
  [sym_primary_expression] = "primary_expression",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym_class_constant_access_expression] = "class_constant_access_expression",
  [sym_print_intrinsic] = "print_intrinsic",
  [sym_object_creation_expression] = "object_creation_expression",
  [sym__new_non_dereferencable_expression] = "_new_non_dereferencable_expression",
  [sym__new_dereferencable_expression] = "_new_dereferencable_expression",
  [sym__class_name_reference] = "_class_name_reference",
  [sym_anonymous_class] = "anonymous_class",
  [sym_update_expression] = "update_expression",
  [sym_cast_expression] = "cast_expression",
  [sym_cast_variable] = "cast_expression",
  [sym_assignment_expression] = "assignment_expression",
  [sym_reference_assignment_expression] = "reference_assignment_expression",
  [sym_conditional_expression] = "conditional_expression",
  [sym_augmented_assignment_expression] = "augmented_assignment_expression",
  [sym__variable_member_access_expression] = "member_access_expression",
  [sym_member_access_expression] = "member_access_expression",
  [sym__variable_nullsafe_member_access_expression] = "nullsafe_member_access_expression",
  [sym_nullsafe_member_access_expression] = "nullsafe_member_access_expression",
  [sym__variable_scoped_property_access_expression] = "scoped_property_access_expression",
  [sym_scoped_property_access_expression] = "scoped_property_access_expression",
  [sym_list_literal] = "list_literal",
  [sym__list_destructing] = "_list_destructing",
  [sym__array_destructing] = "_array_destructing",
  [sym__array_destructing_element] = "_array_destructing_element",
  [sym_function_call_expression] = "function_call_expression",
  [sym__callable_expression] = "_callable_expression",
  [sym_scoped_call_expression] = "scoped_call_expression",
  [sym__scope_resolution_qualifier] = "_scope_resolution_qualifier",
  [sym_relative_scope] = "relative_scope",
  [sym_variadic_placeholder] = "variadic_placeholder",
  [sym_arguments] = "arguments",
  [sym_argument] = "argument",
  [sym__argument_name] = "_argument_name",
  [sym_member_call_expression] = "member_call_expression",
  [sym_nullsafe_member_call_expression] = "nullsafe_member_call_expression",
  [sym_variadic_unpacking] = "variadic_unpacking",
  [sym__member_name] = "_member_name",
  [sym__variable_subscript_expression] = "subscript_expression",
  [sym__dereferencable_subscript_expression] = "subscript_expression",
  [sym__dereferencable_expression] = "_dereferencable_expression",
  [sym__dereferencable_scalar] = "_dereferencable_scalar",
  [sym_array_creation_expression] = "array_creation_expression",
  [sym_attribute_group] = "attribute_group",
  [sym_attribute_list] = "attribute_list",
  [sym_attribute] = "attribute",
  [sym__complex_string_part] = "_complex_string_part",
  [sym__simple_string_member_access_expression] = "member_access_expression",
  [sym__simple_string_subscript_unary_expression] = "unary_op_expression",
  [sym__simple_string_array_access_argument] = "_simple_string_array_access_argument",
  [sym__simple_string_subscript_expression] = "subscript_expression",
  [sym__simple_string_part] = "_simple_string_part",
  [aux_sym__interpolated_string_body] = "_interpolated_string_body",
  [aux_sym__interpolated_string_body_heredoc] = "_interpolated_string_body_heredoc",
  [sym_encapsed_string] = "encapsed_string",
  [sym_string] = "string",
  [sym_string_content] = "string_content",
  [sym_heredoc_body] = "heredoc_body",
  [sym_heredoc] = "heredoc",
  [sym_nowdoc_body] = "nowdoc_body",
  [sym_nowdoc] = "nowdoc",
  [aux_sym__interpolated_execution_operator_body] = "_interpolated_execution_operator_body",
  [sym_shell_command_expression] = "shell_command_expression",
  [sym_boolean] = "boolean",
  [sym_null] = "null",
  [sym__string] = "_string",
  [sym_dynamic_variable_name] = "dynamic_variable_name",
  [sym__simple_variable] = "_simple_variable",
  [sym__new_variable] = "_new_variable",
  [sym__callable_variable] = "_callable_variable",
  [sym_variable_name] = "variable_name",
  [sym_by_ref] = "by_ref",
  [sym_yield_expression] = "yield_expression",
  [sym_array_element_initializer] = "array_element_initializer",
  [sym_binary_expression] = "binary_expression",
  [sym_include_expression] = "include_expression",
  [sym_include_once_expression] = "include_once_expression",
  [sym_require_expression] = "require_expression",
  [sym_require_once_expression] = "require_once_expression",
  [sym__reserved_identifier] = "name",
  [sym__identifier] = "_identifier",
  [sym__semicolon] = "_semicolon",
  [aux_sym_program_repeat1] = "program_repeat1",
  [aux_sym_function_static_declaration_repeat1] = "function_static_declaration_repeat1",
  [aux_sym_global_declaration_repeat1] = "global_declaration_repeat1",
  [aux_sym_namespace_use_declaration_repeat1] = "namespace_use_declaration_repeat1",
  [aux_sym_namespace_name_repeat1] = "namespace_name_repeat1",
  [aux_sym_base_clause_repeat1] = "base_clause_repeat1",
  [aux_sym_enum_declaration_list_repeat1] = "enum_declaration_list_repeat1",
  [aux_sym_class_declaration_repeat1] = "class_declaration_repeat1",
  [aux_sym_declaration_list_repeat1] = "declaration_list_repeat1",
  [aux_sym__const_declaration_repeat1] = "_const_declaration_repeat1",
  [aux_sym_property_declaration_repeat1] = "property_declaration_repeat1",
  [aux_sym_property_hook_list_repeat1] = "property_hook_list_repeat1",
  [aux_sym_use_list_repeat1] = "use_list_repeat1",
  [aux_sym_anonymous_function_use_clause_repeat1] = "anonymous_function_use_clause_repeat1",
  [aux_sym_formal_parameters_repeat1] = "formal_parameters_repeat1",
  [aux_sym_union_type_repeat1] = "union_type_repeat1",
  [aux_sym_intersection_type_repeat1] = "intersection_type_repeat1",
  [aux_sym_disjunctive_normal_form_type_repeat1] = "disjunctive_normal_form_type_repeat1",
  [aux_sym_unset_statement_repeat1] = "unset_statement_repeat1",
  [aux_sym_try_statement_repeat1] = "try_statement_repeat1",
  [aux_sym_type_list_repeat1] = "type_list_repeat1",
  [aux_sym_if_statement_repeat1] = "if_statement_repeat1",
  [aux_sym_if_statement_repeat2] = "if_statement_repeat2",
  [aux_sym_match_block_repeat1] = "match_block_repeat1",
  [aux_sym_match_condition_list_repeat1] = "match_condition_list_repeat1",
  [aux_sym_switch_block_repeat1] = "switch_block_repeat1",
  [aux_sym__list_destructing_repeat1] = "_list_destructing_repeat1",
  [aux_sym__array_destructing_repeat1] = "_array_destructing_repeat1",
  [aux_sym_arguments_repeat1] = "arguments_repeat1",
  [aux_sym_array_creation_expression_repeat1] = "array_creation_expression_repeat1",
  [aux_sym_attribute_group_repeat1] = "attribute_group_repeat1",
  [aux_sym_attribute_list_repeat1] = "attribute_list_repeat1",
  [aux_sym_string_repeat1] = "string_repeat1",
  [aux_sym_string_content_repeat1] = "string_content_repeat1",
  [aux_sym_heredoc_body_repeat1] = "heredoc_body_repeat1",
  [aux_sym_nowdoc_body_repeat1] = "nowdoc_body_repeat1",
  [alias_sym_operation] = "operation",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_name] = sym_name,
  [anon_sym_QMARK_GT] = anon_sym_QMARK_GT,
  [sym_php_tag] = sym_php_tag,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_AMP] = anon_sym_AMP,
  [aux_sym_function_static_declaration_token1] = aux_sym_function_static_declaration_token1,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_EQ] = anon_sym_EQ,
  [aux_sym_global_declaration_token1] = aux_sym_global_declaration_token1,
  [aux_sym_namespace_definition_token1] = aux_sym_namespace_definition_token1,
  [aux_sym_namespace_use_declaration_token1] = aux_sym_namespace_use_declaration_token1,
  [aux_sym_namespace_use_clause_token1] = aux_sym_namespace_use_clause_token1,
  [aux_sym__namespace_use_type_token1] = aux_sym__namespace_use_type_token1,
  [aux_sym__namespace_use_type_token2] = aux_sym__namespace_use_type_token2,
  [anon_sym_BSLASH] = anon_sym_BSLASH,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [aux_sym_trait_declaration_token1] = aux_sym_trait_declaration_token1,
  [aux_sym_interface_declaration_token1] = aux_sym_interface_declaration_token1,
  [aux_sym_base_clause_token1] = aux_sym_base_clause_token1,
  [aux_sym_enum_declaration_token1] = aux_sym_enum_declaration_token1,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_string] = anon_sym_string,
  [anon_sym_int] = anon_sym_int,
  [aux_sym_enum_case_token1] = aux_sym_enum_case_token1,
  [aux_sym_class_declaration_token1] = aux_sym_class_declaration_token1,
  [aux_sym_final_modifier_token1] = aux_sym_final_modifier_token1,
  [aux_sym_abstract_modifier_token1] = aux_sym_abstract_modifier_token1,
  [aux_sym_readonly_modifier_token1] = aux_sym_readonly_modifier_token1,
  [aux_sym_class_interface_clause_token1] = aux_sym_class_interface_clause_token1,
  [anon_sym_EQ_GT] = anon_sym_EQ_GT,
  [sym_var_modifier] = sym_var_modifier,
  [aux_sym_use_instead_of_clause_token1] = aux_sym_use_instead_of_clause_token1,
  [aux_sym_visibility_modifier_token1] = aux_sym_visibility_modifier_token1,
  [aux_sym_visibility_modifier_token2] = aux_sym_visibility_modifier_token2,
  [aux_sym_visibility_modifier_token3] = aux_sym_visibility_modifier_token3,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_LPAREN2] = anon_sym_LPAREN,
  [anon_sym_RPAREN2] = anon_sym_RPAREN,
  [aux_sym__arrow_function_header_token1] = aux_sym__arrow_function_header_token1,
  [anon_sym_DOT_DOT_DOT] = anon_sym_DOT_DOT_DOT,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [sym_bottom_type] = sym_bottom_type,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_array] = anon_sym_array,
  [aux_sym_primitive_type_token1] = aux_sym_primitive_type_token1,
  [anon_sym_iterable] = anon_sym_iterable,
  [anon_sym_bool] = anon_sym_bool,
  [anon_sym_float] = anon_sym_float,
  [anon_sym_void] = anon_sym_void,
  [anon_sym_mixed] = anon_sym_mixed,
  [anon_sym_false] = anon_sym_false,
  [anon_sym_null] = anon_sym_null,
  [anon_sym_true] = anon_sym_true,
  [aux_sym_cast_type_token1] = aux_sym_cast_type_token1,
  [aux_sym_cast_type_token2] = aux_sym_cast_type_token2,
  [aux_sym_cast_type_token3] = aux_sym_cast_type_token3,
  [aux_sym_cast_type_token4] = aux_sym_cast_type_token4,
  [aux_sym_cast_type_token5] = aux_sym_cast_type_token5,
  [aux_sym_cast_type_token6] = aux_sym_cast_type_token6,
  [aux_sym_cast_type_token7] = aux_sym_cast_type_token7,
  [aux_sym_cast_type_token8] = aux_sym_cast_type_token8,
  [aux_sym_cast_type_token9] = aux_sym_cast_type_token9,
  [aux_sym_cast_type_token10] = aux_sym_cast_type_token10,
  [aux_sym_cast_type_token11] = aux_sym_cast_type_token11,
  [aux_sym_cast_type_token12] = aux_sym_cast_type_token12,
  [aux_sym_echo_statement_token1] = aux_sym_echo_statement_token1,
  [aux_sym_exit_statement_token1] = aux_sym_exit_statement_token1,
  [anon_sym_unset] = anon_sym_unset,
  [aux_sym_declare_statement_token1] = aux_sym_declare_statement_token1,
  [aux_sym_declare_statement_token2] = aux_sym_declare_statement_token2,
  [anon_sym_ticks] = anon_sym_ticks,
  [anon_sym_encoding] = anon_sym_encoding,
  [anon_sym_strict_types] = anon_sym_strict_types,
  [sym_float] = sym_float,
  [aux_sym_try_statement_token1] = aux_sym_try_statement_token1,
  [aux_sym_catch_clause_token1] = aux_sym_catch_clause_token1,
  [aux_sym_finally_clause_token1] = aux_sym_finally_clause_token1,
  [aux_sym_goto_statement_token1] = aux_sym_goto_statement_token1,
  [aux_sym_continue_statement_token1] = aux_sym_continue_statement_token1,
  [aux_sym_break_statement_token1] = aux_sym_break_statement_token1,
  [sym_integer] = sym_integer,
  [aux_sym_return_statement_token1] = aux_sym_return_statement_token1,
  [aux_sym_throw_expression_token1] = aux_sym_throw_expression_token1,
  [aux_sym_while_statement_token1] = aux_sym_while_statement_token1,
  [aux_sym_while_statement_token2] = aux_sym_while_statement_token2,
  [aux_sym_do_statement_token1] = aux_sym_do_statement_token1,
  [aux_sym_for_statement_token1] = aux_sym_for_statement_token1,
  [aux_sym_for_statement_token2] = aux_sym_for_statement_token2,
  [aux_sym_foreach_statement_token1] = aux_sym_foreach_statement_token1,
  [aux_sym_foreach_statement_token2] = aux_sym_foreach_statement_token2,
  [aux_sym_if_statement_token1] = aux_sym_if_statement_token1,
  [aux_sym_if_statement_token2] = aux_sym_if_statement_token2,
  [aux_sym_else_if_clause_token1] = aux_sym_else_if_clause_token1,
  [aux_sym_else_clause_token1] = aux_sym_else_clause_token1,
  [aux_sym_match_expression_token1] = aux_sym_match_expression_token1,
  [aux_sym_match_default_expression_token1] = aux_sym_match_default_expression_token1,
  [aux_sym_switch_statement_token1] = aux_sym_switch_statement_token1,
  [aux_sym_switch_block_token1] = aux_sym_switch_block_token1,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_AT] = anon_sym_AT,
  [aux_sym_clone_expression_token1] = aux_sym_clone_expression_token1,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [aux_sym_print_intrinsic_token1] = aux_sym_print_intrinsic_token1,
  [aux_sym__new_non_dereferencable_expression_token1] = aux_sym__new_non_dereferencable_expression_token1,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [anon_sym_PLUS_PLUS] = anon_sym_PLUS_PLUS,
  [anon_sym_STAR_STAR_EQ] = anon_sym_STAR_STAR_EQ,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_PERCENT_EQ] = anon_sym_PERCENT_EQ,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_DASH_EQ] = anon_sym_DASH_EQ,
  [anon_sym_DOT_EQ] = anon_sym_DOT_EQ,
  [anon_sym_LT_LT_EQ] = anon_sym_LT_LT_EQ,
  [anon_sym_GT_GT_EQ] = anon_sym_GT_GT_EQ,
  [anon_sym_AMP_EQ] = anon_sym_AMP_EQ,
  [anon_sym_CARET_EQ] = anon_sym_CARET_EQ,
  [anon_sym_PIPE_EQ] = anon_sym_PIPE_EQ,
  [anon_sym_QMARK_QMARK_EQ] = anon_sym_QMARK_QMARK_EQ,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_QMARK_DASH_GT] = anon_sym_QMARK_DASH_GT,
  [aux_sym__list_destructing_token1] = aux_sym__list_destructing_token1,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_self] = anon_sym_self,
  [anon_sym_parent] = anon_sym_parent,
  [aux_sym__argument_name_token1] = aux_sym__argument_name_token1,
  [aux_sym__argument_name_token2] = aux_sym__argument_name_token2,
  [anon_sym_POUND_LBRACK] = anon_sym_POUND_LBRACK,
  [sym_escape_sequence] = sym_escape_sequence,
  [anon_sym_BSLASHu] = sym_string_content,
  [aux_sym_encapsed_string_token1] = aux_sym_encapsed_string_token1,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [aux_sym_string_token1] = aux_sym_string_token1,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [aux_sym_string_token2] = sym_escape_sequence,
  [aux_sym_string_content_token1] = aux_sym_string_content_token1,
  [anon_sym_LT_LT_LT] = anon_sym_LT_LT_LT,
  [anon_sym_DQUOTE2] = anon_sym_DQUOTE,
  [sym__new_line] = sym__new_line,
  [anon_sym_SQUOTE2] = anon_sym_SQUOTE,
  [anon_sym_BQUOTE] = anon_sym_BQUOTE,
  [anon_sym_DOLLAR] = anon_sym_DOLLAR,
  [aux_sym_yield_expression_token1] = aux_sym_yield_expression_token1,
  [aux_sym_yield_expression_token2] = aux_sym_yield_expression_token2,
  [aux_sym_binary_expression_token1] = aux_sym_binary_expression_token1,
  [anon_sym_QMARK_QMARK] = anon_sym_QMARK_QMARK,
  [anon_sym_STAR_STAR] = anon_sym_STAR_STAR,
  [aux_sym_binary_expression_token2] = aux_sym_binary_expression_token2,
  [aux_sym_binary_expression_token3] = aux_sym_binary_expression_token3,
  [aux_sym_binary_expression_token4] = aux_sym_binary_expression_token4,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_LT_GT] = anon_sym_LT_GT,
  [anon_sym_EQ_EQ_EQ] = anon_sym_EQ_EQ_EQ,
  [anon_sym_BANG_EQ_EQ] = anon_sym_BANG_EQ_EQ,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_LT_EQ_GT] = anon_sym_LT_EQ_GT,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [aux_sym_include_expression_token1] = aux_sym_include_expression_token1,
  [aux_sym_include_once_expression_token1] = aux_sym_include_once_expression_token1,
  [aux_sym_require_expression_token1] = aux_sym_require_expression_token1,
  [aux_sym_require_once_expression_token1] = aux_sym_require_once_expression_token1,
  [sym_comment] = sym_comment,
  [sym__automatic_semicolon] = sym__automatic_semicolon,
  [sym_encapsed_string_chars] = sym_string_content,
  [sym_encapsed_string_chars_after_variable] = sym_string_content,
  [sym_execution_string_chars] = sym_string_content,
  [sym_execution_string_chars_after_variable] = sym_string_content,
  [sym_encapsed_string_chars_heredoc] = sym_string_content,
  [sym_encapsed_string_chars_after_variable_heredoc] = sym_string_content,
  [sym__eof] = sym__eof,
  [sym_heredoc_start] = sym_heredoc_start,
  [sym_heredoc_end] = sym_heredoc_end,
  [sym_nowdoc_string] = sym_nowdoc_string,
  [sym_sentinel_error] = sym_sentinel_error,
  [sym_program] = sym_program,
  [sym_statement] = sym_statement,
  [sym_empty_statement] = sym_empty_statement,
  [sym_reference_modifier] = sym_reference_modifier,
  [sym_function_static_declaration] = sym_function_static_declaration,
  [sym_static_variable_declaration] = sym_static_variable_declaration,
  [sym_global_declaration] = sym_global_declaration,
  [sym_namespace_definition] = sym_namespace_definition,
  [sym_namespace_use_declaration] = sym_namespace_use_declaration,
  [sym_namespace_use_clause] = sym_namespace_use_clause,
  [sym_qualified_name] = sym_qualified_name,
  [sym__name] = sym__name,
  [sym_namespace_name] = sym_namespace_name,
  [sym__namespace_use_group] = sym__namespace_use_group,
  [sym_namespace_use_group] = sym_namespace_use_group,
  [sym_trait_declaration] = sym_trait_declaration,
  [sym_interface_declaration] = sym_interface_declaration,
  [sym_base_clause] = sym_base_clause,
  [sym_enum_declaration] = sym_enum_declaration,
  [sym_enum_declaration_list] = sym_enum_declaration_list,
  [sym__enum_member_declaration] = sym__enum_member_declaration,
  [sym_enum_case] = sym_enum_case,
  [sym_class_declaration] = sym_class_declaration,
  [sym_declaration_list] = sym_declaration_list,
  [sym_final_modifier] = sym_final_modifier,
  [sym_abstract_modifier] = sym_abstract_modifier,
  [sym_readonly_modifier] = sym_readonly_modifier,
  [sym_class_interface_clause] = sym_class_interface_clause,
  [sym__member_declaration] = sym__member_declaration,
  [sym_const_declaration] = sym_const_declaration,
  [sym__class_const_declaration] = sym_const_declaration,
  [sym__const_declaration] = sym__const_declaration,
  [sym_property_declaration] = sym_property_declaration,
  [sym__modifier] = sym__modifier,
  [sym_property_element] = sym_property_element,
  [sym_property_hook_list] = sym_property_hook_list,
  [sym_property_hook] = sym_property_hook,
  [sym__property_hook_body] = sym__property_hook_body,
  [sym_method_declaration] = sym_method_declaration,
  [sym_static_modifier] = sym_static_modifier,
  [sym_use_declaration] = sym_use_declaration,
  [sym_use_list] = sym_use_list,
  [sym_use_instead_of_clause] = sym_use_instead_of_clause,
  [sym_use_as_clause] = sym_use_as_clause,
  [sym_visibility_modifier] = sym_visibility_modifier,
  [sym_function_definition] = sym_function_definition,
  [sym__function_definition_header] = sym__function_definition_header,
  [sym_anonymous_function] = sym_anonymous_function,
  [sym_anonymous_function_use_clause] = sym_anonymous_function_use_clause,
  [sym__anonymous_function_header] = sym__anonymous_function_header,
  [sym__arrow_function_header] = sym__arrow_function_header,
  [sym_arrow_function] = sym_arrow_function,
  [sym_formal_parameters] = sym_formal_parameters,
  [sym_property_promotion_parameter] = sym_property_promotion_parameter,
  [sym_simple_parameter] = sym_simple_parameter,
  [sym_variadic_parameter] = sym_variadic_parameter,
  [sym_type] = sym_type,
  [sym__types] = sym__types,
  [sym_named_type] = sym_named_type,
  [sym_optional_type] = sym_optional_type,
  [sym_union_type] = sym_union_type,
  [sym_intersection_type] = sym_intersection_type,
  [sym_disjunctive_normal_form_type] = sym_disjunctive_normal_form_type,
  [sym_primitive_type] = sym_primitive_type,
  [sym_cast_type] = sym_cast_type,
  [sym__return_type] = sym__return_type,
  [sym_const_element] = sym_const_element,
  [sym_echo_statement] = sym_echo_statement,
  [sym_exit_statement] = sym_exit_statement,
  [sym_unset_statement] = sym_unset_statement,
  [sym_declare_statement] = sym_declare_statement,
  [sym_declare_directive] = sym_declare_directive,
  [sym_literal] = sym_literal,
  [sym_try_statement] = sym_try_statement,
  [sym_catch_clause] = sym_catch_clause,
  [sym_type_list] = sym_type_list,
  [sym_finally_clause] = sym_finally_clause,
  [sym_goto_statement] = sym_goto_statement,
  [sym_continue_statement] = sym_continue_statement,
  [sym_break_statement] = sym_break_statement,
  [sym_return_statement] = sym_return_statement,
  [sym_throw_expression] = sym_throw_expression,
  [sym_while_statement] = sym_while_statement,
  [sym_do_statement] = sym_do_statement,
  [sym_for_statement] = sym_for_statement,
  [sym__expressions] = sym__expressions,
  [sym_sequence_expression] = sym_sequence_expression,
  [sym_foreach_statement] = sym_foreach_statement,
  [sym_foreach_pair] = sym_foreach_pair,
  [sym__foreach_value] = sym__foreach_value,
  [sym_if_statement] = sym_if_statement,
  [sym_colon_block] = sym_colon_block,
  [sym_else_if_clause] = sym_else_if_clause,
  [sym_else_clause] = sym_else_clause,
  [sym_else_if_clause_2] = sym_else_if_clause,
  [sym_else_clause_2] = sym_else_clause,
  [sym_match_expression] = sym_match_expression,
  [sym_match_block] = sym_match_block,
  [sym_match_condition_list] = sym_match_condition_list,
  [sym_match_conditional_expression] = sym_match_conditional_expression,
  [sym_match_default_expression] = sym_match_default_expression,
  [sym_switch_statement] = sym_switch_statement,
  [sym_switch_block] = sym_switch_block,
  [sym_case_statement] = sym_case_statement,
  [sym_default_statement] = sym_default_statement,
  [sym_compound_statement] = sym_compound_statement,
  [sym_named_label_statement] = sym_named_label_statement,
  [sym_expression_statement] = sym_expression_statement,
  [sym_expression] = sym_expression,
  [sym__unary_expression] = sym__unary_expression,
  [sym_unary_op_expression] = sym_unary_op_expression,
  [sym_error_suppression_expression] = sym_error_suppression_expression,
  [sym_clone_expression] = sym_clone_expression,
  [sym_primary_expression] = sym_primary_expression,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym_class_constant_access_expression] = sym_class_constant_access_expression,
  [sym_print_intrinsic] = sym_print_intrinsic,
  [sym_object_creation_expression] = sym_object_creation_expression,
  [sym__new_non_dereferencable_expression] = sym__new_non_dereferencable_expression,
  [sym__new_dereferencable_expression] = sym__new_dereferencable_expression,
  [sym__class_name_reference] = sym__class_name_reference,
  [sym_anonymous_class] = sym_anonymous_class,
  [sym_update_expression] = sym_update_expression,
  [sym_cast_expression] = sym_cast_expression,
  [sym_cast_variable] = sym_cast_expression,
  [sym_assignment_expression] = sym_assignment_expression,
  [sym_reference_assignment_expression] = sym_reference_assignment_expression,
  [sym_conditional_expression] = sym_conditional_expression,
  [sym_augmented_assignment_expression] = sym_augmented_assignment_expression,
  [sym__variable_member_access_expression] = sym_member_access_expression,
  [sym_member_access_expression] = sym_member_access_expression,
  [sym__variable_nullsafe_member_access_expression] = sym_nullsafe_member_access_expression,
  [sym_nullsafe_member_access_expression] = sym_nullsafe_member_access_expression,
  [sym__variable_scoped_property_access_expression] = sym_scoped_property_access_expression,
  [sym_scoped_property_access_expression] = sym_scoped_property_access_expression,
  [sym_list_literal] = sym_list_literal,
  [sym__list_destructing] = sym__list_destructing,
  [sym__array_destructing] = sym__array_destructing,
  [sym__array_destructing_element] = sym__array_destructing_element,
  [sym_function_call_expression] = sym_function_call_expression,
  [sym__callable_expression] = sym__callable_expression,
  [sym_scoped_call_expression] = sym_scoped_call_expression,
  [sym__scope_resolution_qualifier] = sym__scope_resolution_qualifier,
  [sym_relative_scope] = sym_relative_scope,
  [sym_variadic_placeholder] = sym_variadic_placeholder,
  [sym_arguments] = sym_arguments,
  [sym_argument] = sym_argument,
  [sym__argument_name] = sym__argument_name,
  [sym_member_call_expression] = sym_member_call_expression,
  [sym_nullsafe_member_call_expression] = sym_nullsafe_member_call_expression,
  [sym_variadic_unpacking] = sym_variadic_unpacking,
  [sym__member_name] = sym__member_name,
  [sym__variable_subscript_expression] = sym__variable_subscript_expression,
  [sym__dereferencable_subscript_expression] = sym__variable_subscript_expression,
  [sym__dereferencable_expression] = sym__dereferencable_expression,
  [sym__dereferencable_scalar] = sym__dereferencable_scalar,
  [sym_array_creation_expression] = sym_array_creation_expression,
  [sym_attribute_group] = sym_attribute_group,
  [sym_attribute_list] = sym_attribute_list,
  [sym_attribute] = sym_attribute,
  [sym__complex_string_part] = sym__complex_string_part,
  [sym__simple_string_member_access_expression] = sym_member_access_expression,
  [sym__simple_string_subscript_unary_expression] = sym_unary_op_expression,
  [sym__simple_string_array_access_argument] = sym__simple_string_array_access_argument,
  [sym__simple_string_subscript_expression] = sym__variable_subscript_expression,
  [sym__simple_string_part] = sym__simple_string_part,
  [aux_sym__interpolated_string_body] = aux_sym__interpolated_string_body,
  [aux_sym__interpolated_string_body_heredoc] = aux_sym__interpolated_string_body_heredoc,
  [sym_encapsed_string] = sym_encapsed_string,
  [sym_string] = sym_string,
  [sym_string_content] = sym_string_content,
  [sym_heredoc_body] = sym_heredoc_body,
  [sym_heredoc] = sym_heredoc,
  [sym_nowdoc_body] = sym_nowdoc_body,
  [sym_nowdoc] = sym_nowdoc,
  [aux_sym__interpolated_execution_operator_body] = aux_sym__interpolated_execution_operator_body,
  [sym_shell_command_expression] = sym_shell_command_expression,
  [sym_boolean] = sym_boolean,
  [sym_null] = sym_null,
  [sym__string] = sym__string,
  [sym_dynamic_variable_name] = sym_dynamic_variable_name,
  [sym__simple_variable] = sym__simple_variable,
  [sym__new_variable] = sym__new_variable,
  [sym__callable_variable] = sym__callable_variable,
  [sym_variable_name] = sym_variable_name,
  [sym_by_ref] = sym_by_ref,
  [sym_yield_expression] = sym_yield_expression,
  [sym_array_element_initializer] = sym_array_element_initializer,
  [sym_binary_expression] = sym_binary_expression,
  [sym_include_expression] = sym_include_expression,
  [sym_include_once_expression] = sym_include_once_expression,
  [sym_require_expression] = sym_require_expression,
  [sym_require_once_expression] = sym_require_once_expression,
  [sym__reserved_identifier] = sym_name,
  [sym__identifier] = sym__identifier,
  [sym__semicolon] = sym__semicolon,
  [aux_sym_program_repeat1] = aux_sym_program_repeat1,
  [aux_sym_function_static_declaration_repeat1] = aux_sym_function_static_declaration_repeat1,
  [aux_sym_global_declaration_repeat1] = aux_sym_global_declaration_repeat1,
  [aux_sym_namespace_use_declaration_repeat1] = aux_sym_namespace_use_declaration_repeat1,
  [aux_sym_namespace_name_repeat1] = aux_sym_namespace_name_repeat1,
  [aux_sym_base_clause_repeat1] = aux_sym_base_clause_repeat1,
  [aux_sym_enum_declaration_list_repeat1] = aux_sym_enum_declaration_list_repeat1,
  [aux_sym_class_declaration_repeat1] = aux_sym_class_declaration_repeat1,
  [aux_sym_declaration_list_repeat1] = aux_sym_declaration_list_repeat1,
  [aux_sym__const_declaration_repeat1] = aux_sym__const_declaration_repeat1,
  [aux_sym_property_declaration_repeat1] = aux_sym_property_declaration_repeat1,
  [aux_sym_property_hook_list_repeat1] = aux_sym_property_hook_list_repeat1,
  [aux_sym_use_list_repeat1] = aux_sym_use_list_repeat1,
  [aux_sym_anonymous_function_use_clause_repeat1] = aux_sym_anonymous_function_use_clause_repeat1,
  [aux_sym_formal_parameters_repeat1] = aux_sym_formal_parameters_repeat1,
  [aux_sym_union_type_repeat1] = aux_sym_union_type_repeat1,
  [aux_sym_intersection_type_repeat1] = aux_sym_intersection_type_repeat1,
  [aux_sym_disjunctive_normal_form_type_repeat1] = aux_sym_disjunctive_normal_form_type_repeat1,
  [aux_sym_unset_statement_repeat1] = aux_sym_unset_statement_repeat1,
  [aux_sym_try_statement_repeat1] = aux_sym_try_statement_repeat1,
  [aux_sym_type_list_repeat1] = aux_sym_type_list_repeat1,
  [aux_sym_if_statement_repeat1] = aux_sym_if_statement_repeat1,
  [aux_sym_if_statement_repeat2] = aux_sym_if_statement_repeat2,
  [aux_sym_match_block_repeat1] = aux_sym_match_block_repeat1,
  [aux_sym_match_condition_list_repeat1] = aux_sym_match_condition_list_repeat1,
  [aux_sym_switch_block_repeat1] = aux_sym_switch_block_repeat1,
  [aux_sym__list_destructing_repeat1] = aux_sym__list_destructing_repeat1,
  [aux_sym__array_destructing_repeat1] = aux_sym__array_destructing_repeat1,
  [aux_sym_arguments_repeat1] = aux_sym_arguments_repeat1,
  [aux_sym_array_creation_expression_repeat1] = aux_sym_array_creation_expression_repeat1,
  [aux_sym_attribute_group_repeat1] = aux_sym_attribute_group_repeat1,
  [aux_sym_attribute_list_repeat1] = aux_sym_attribute_list_repeat1,
  [aux_sym_string_repeat1] = aux_sym_string_repeat1,
  [aux_sym_string_content_repeat1] = aux_sym_string_content_repeat1,
  [aux_sym_heredoc_body_repeat1] = aux_sym_heredoc_body_repeat1,
  [aux_sym_nowdoc_body_repeat1] = aux_sym_nowdoc_body_repeat1,
  [alias_sym_operation] = alias_sym_operation,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_name] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_QMARK_GT] = {
    .visible = true,
    .named = false,
  },
  [sym_php_tag] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_function_static_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_global_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_namespace_definition_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_namespace_use_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_namespace_use_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__namespace_use_type_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__namespace_use_type_token2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BSLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_trait_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_interface_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_base_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_enum_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_string] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_int] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_enum_case_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_class_declaration_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_final_modifier_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_abstract_modifier_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_readonly_modifier_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_class_interface_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_GT] = {
    .visible = true,
    .named = false,
  },
  [sym_var_modifier] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_use_instead_of_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_visibility_modifier_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_visibility_modifier_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_visibility_modifier_token3] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__arrow_function_header_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [sym_bottom_type] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_array] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_primitive_type_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_iterable] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_bool] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_float] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_void] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_mixed] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_false] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_null] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_true] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_cast_type_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token4] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token5] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token6] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token7] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token8] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token9] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token10] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token11] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_cast_type_token12] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_echo_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_exit_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unset] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_declare_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_declare_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ticks] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_encoding] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_strict_types] = {
    .visible = true,
    .named = false,
  },
  [sym_float] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_try_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_catch_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_finally_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_goto_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_continue_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_break_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [sym_integer] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_return_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_throw_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_while_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_while_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_do_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_for_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_for_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_foreach_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_foreach_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_if_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_if_statement_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_else_if_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_else_clause_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_match_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_match_default_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_switch_statement_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_switch_block_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_clone_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_print_intrinsic_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__new_non_dereferencable_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_QMARK_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__list_destructing_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_self] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_parent] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__argument_name_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__argument_name_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_POUND_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_BSLASHu] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_encapsed_string_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_string_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_string_token2] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_string_content_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_LT_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [sym__new_line] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_SQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_yield_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_yield_expression_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_binary_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_binary_expression_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_binary_expression_token3] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_binary_expression_token4] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_include_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_include_once_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_require_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_require_once_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym__automatic_semicolon] = {
    .visible = false,
    .named = true,
  },
  [sym_encapsed_string_chars] = {
    .visible = true,
    .named = true,
  },
  [sym_encapsed_string_chars_after_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_execution_string_chars] = {
    .visible = true,
    .named = true,
  },
  [sym_execution_string_chars_after_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_encapsed_string_chars_heredoc] = {
    .visible = true,
    .named = true,
  },
  [sym_encapsed_string_chars_after_variable_heredoc] = {
    .visible = true,
    .named = true,
  },
  [sym__eof] = {
    .visible = false,
    .named = true,
  },
  [sym_heredoc_start] = {
    .visible = true,
    .named = true,
  },
  [sym_heredoc_end] = {
    .visible = true,
    .named = true,
  },
  [sym_nowdoc_string] = {
    .visible = true,
    .named = true,
  },
  [sym_sentinel_error] = {
    .visible = true,
    .named = true,
  },
  [sym_program] = {
    .visible = true,
    .named = true,
  },
  [sym_statement] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_empty_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_reference_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_function_static_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_static_variable_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_global_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_namespace_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_namespace_use_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_namespace_use_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_qualified_name] = {
    .visible = true,
    .named = true,
  },
  [sym__name] = {
    .visible = false,
    .named = true,
  },
  [sym_namespace_name] = {
    .visible = true,
    .named = true,
  },
  [sym__namespace_use_group] = {
    .visible = false,
    .named = true,
  },
  [sym_namespace_use_group] = {
    .visible = true,
    .named = true,
  },
  [sym_trait_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_base_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym__enum_member_declaration] = {
    .visible = false,
    .named = true,
  },
  [sym_enum_case] = {
    .visible = true,
    .named = true,
  },
  [sym_class_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_final_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_abstract_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_readonly_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_class_interface_clause] = {
    .visible = true,
    .named = true,
  },
  [sym__member_declaration] = {
    .visible = false,
    .named = true,
  },
  [sym_const_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__class_const_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__const_declaration] = {
    .visible = false,
    .named = true,
  },
  [sym_property_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__modifier] = {
    .visible = false,
    .named = true,
  },
  [sym_property_element] = {
    .visible = true,
    .named = true,
  },
  [sym_property_hook_list] = {
    .visible = true,
    .named = true,
  },
  [sym_property_hook] = {
    .visible = true,
    .named = true,
  },
  [sym__property_hook_body] = {
    .visible = false,
    .named = true,
  },
  [sym_method_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_static_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_use_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_use_list] = {
    .visible = true,
    .named = true,
  },
  [sym_use_instead_of_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_use_as_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_visibility_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_function_definition] = {
    .visible = true,
    .named = true,
  },
  [sym__function_definition_header] = {
    .visible = false,
    .named = true,
  },
  [sym_anonymous_function] = {
    .visible = true,
    .named = true,
  },
  [sym_anonymous_function_use_clause] = {
    .visible = true,
    .named = true,
  },
  [sym__anonymous_function_header] = {
    .visible = false,
    .named = true,
  },
  [sym__arrow_function_header] = {
    .visible = false,
    .named = true,
  },
  [sym_arrow_function] = {
    .visible = true,
    .named = true,
  },
  [sym_formal_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_property_promotion_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_simple_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_variadic_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__types] = {
    .visible = false,
    .named = true,
  },
  [sym_named_type] = {
    .visible = true,
    .named = true,
  },
  [sym_optional_type] = {
    .visible = true,
    .named = true,
  },
  [sym_union_type] = {
    .visible = true,
    .named = true,
  },
  [sym_intersection_type] = {
    .visible = true,
    .named = true,
  },
  [sym_disjunctive_normal_form_type] = {
    .visible = true,
    .named = true,
  },
  [sym_primitive_type] = {
    .visible = true,
    .named = true,
  },
  [sym_cast_type] = {
    .visible = true,
    .named = true,
  },
  [sym__return_type] = {
    .visible = false,
    .named = true,
  },
  [sym_const_element] = {
    .visible = true,
    .named = true,
  },
  [sym_echo_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_exit_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_unset_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_declare_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_declare_directive] = {
    .visible = true,
    .named = true,
  },
  [sym_literal] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_try_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_catch_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_type_list] = {
    .visible = true,
    .named = true,
  },
  [sym_finally_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_goto_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_continue_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_break_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_return_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_throw_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_while_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_do_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_for_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__expressions] = {
    .visible = false,
    .named = true,
  },
  [sym_sequence_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_foreach_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_foreach_pair] = {
    .visible = true,
    .named = true,
  },
  [sym__foreach_value] = {
    .visible = false,
    .named = true,
  },
  [sym_if_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_colon_block] = {
    .visible = true,
    .named = true,
  },
  [sym_else_if_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_else_if_clause_2] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause_2] = {
    .visible = true,
    .named = true,
  },
  [sym_match_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_match_block] = {
    .visible = true,
    .named = true,
  },
  [sym_match_condition_list] = {
    .visible = true,
    .named = true,
  },
  [sym_match_conditional_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_match_default_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_block] = {
    .visible = true,
    .named = true,
  },
  [sym_case_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_default_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_compound_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_named_label_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_expression_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__unary_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_unary_op_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_error_suppression_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_clone_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_primary_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_class_constant_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_print_intrinsic] = {
    .visible = true,
    .named = true,
  },
  [sym_object_creation_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__new_non_dereferencable_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__new_dereferencable_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__class_name_reference] = {
    .visible = false,
    .named = true,
  },
  [sym_anonymous_class] = {
    .visible = true,
    .named = true,
  },
  [sym_update_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_cast_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_cast_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_reference_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_conditional_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_augmented_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__variable_member_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_member_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__variable_nullsafe_member_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_nullsafe_member_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__variable_scoped_property_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_scoped_property_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_list_literal] = {
    .visible = true,
    .named = true,
  },
  [sym__list_destructing] = {
    .visible = false,
    .named = true,
  },
  [sym__array_destructing] = {
    .visible = false,
    .named = true,
  },
  [sym__array_destructing_element] = {
    .visible = false,
    .named = true,
  },
  [sym_function_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__callable_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_scoped_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__scope_resolution_qualifier] = {
    .visible = false,
    .named = true,
  },
  [sym_relative_scope] = {
    .visible = true,
    .named = true,
  },
  [sym_variadic_placeholder] = {
    .visible = true,
    .named = true,
  },
  [sym_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_argument] = {
    .visible = true,
    .named = true,
  },
  [sym__argument_name] = {
    .visible = false,
    .named = true,
  },
  [sym_member_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_nullsafe_member_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_variadic_unpacking] = {
    .visible = true,
    .named = true,
  },
  [sym__member_name] = {
    .visible = false,
    .named = true,
  },
  [sym__variable_subscript_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__dereferencable_subscript_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__dereferencable_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__dereferencable_scalar] = {
    .visible = false,
    .named = true,
  },
  [sym_array_creation_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_group] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_list] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym__complex_string_part] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_string_member_access_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_string_subscript_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_string_array_access_argument] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_string_subscript_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_string_part] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__interpolated_string_body] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__interpolated_string_body_heredoc] = {
    .visible = false,
    .named = false,
  },
  [sym_encapsed_string] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_string_content] = {
    .visible = true,
    .named = true,
  },
  [sym_heredoc_body] = {
    .visible = true,
    .named = true,
  },
  [sym_heredoc] = {
    .visible = true,
    .named = true,
  },
  [sym_nowdoc_body] = {
    .visible = true,
    .named = true,
  },
  [sym_nowdoc] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__interpolated_execution_operator_body] = {
    .visible = false,
    .named = false,
  },
  [sym_shell_command_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean] = {
    .visible = true,
    .named = true,
  },
  [sym_null] = {
    .visible = true,
    .named = true,
  },
  [sym__string] = {
    .visible = false,
    .named = true,
  },
  [sym_dynamic_variable_name] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_variable] = {
    .visible = false,
    .named = true,
  },
  [sym__new_variable] = {
    .visible = false,
    .named = true,
  },
  [sym__callable_variable] = {
    .visible = false,
    .named = true,
  },
  [sym_variable_name] = {
    .visible = true,
    .named = true,
  },
  [sym_by_ref] = {
    .visible = true,
    .named = true,
  },
  [sym_yield_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_array_element_initializer] = {
    .visible = true,
    .named = true,
  },
  [sym_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_include_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_include_once_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_require_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_require_once_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__reserved_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym__identifier] = {
    .visible = false,
    .named = true,
  },
  [sym__semicolon] = {
    .visible = false,
    .named = true,
  },
  [aux_sym_program_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_function_static_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_global_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_namespace_use_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_namespace_name_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_base_clause_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_enum_declaration_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_declaration_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__const_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_property_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_property_hook_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_use_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_anonymous_function_use_clause_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_formal_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_union_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_intersection_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_disjunctive_normal_form_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_unset_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_try_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_if_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_if_statement_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_match_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_match_condition_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_switch_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__list_destructing_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__array_destructing_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_array_creation_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_group_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_heredoc_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_nowdoc_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_operation] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_alias = 1,
  field_alternative = 2,
  field_argument = 3,
  field_arguments = 4,
  field_attributes = 5,
  field_body = 6,
  field_condition = 7,
  field_conditional_expressions = 8,
  field_default_value = 9,
  field_end_tag = 10,
  field_final = 11,
  field_function = 12,
  field_identifier = 13,
  field_initialize = 14,
  field_left = 15,
  field_name = 16,
  field_object = 17,
  field_operator = 18,
  field_parameters = 19,
  field_prefix = 20,
  field_readonly = 21,
  field_reference_modifier = 22,
  field_return_expression = 23,
  field_return_type = 24,
  field_right = 25,
  field_scope = 26,
  field_static_modifier = 27,
  field_type = 28,
  field_update = 29,
  field_value = 30,
  field_visibility = 31,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alias] = "alias",
  [field_alternative] = "alternative",
  [field_argument] = "argument",
  [field_arguments] = "arguments",
  [field_attributes] = "attributes",
  [field_body] = "body",
  [field_condition] = "condition",
  [field_conditional_expressions] = "conditional_expressions",
  [field_default_value] = "default_value",
  [field_end_tag] = "end_tag",
  [field_final] = "final",
  [field_function] = "function",
  [field_identifier] = "identifier",
  [field_initialize] = "initialize",
  [field_left] = "left",
  [field_name] = "name",
  [field_object] = "object",
  [field_operator] = "operator",
  [field_parameters] = "parameters",
  [field_prefix] = "prefix",
  [field_readonly] = "readonly",
  [field_reference_modifier] = "reference_modifier",
  [field_return_expression] = "return_expression",
  [field_return_type] = "return_type",
  [field_right] = "right",
  [field_scope] = "scope",
  [field_static_modifier] = "static_modifier",
  [field_type] = "type",
  [field_update] = "update",
  [field_value] = "value",
  [field_visibility] = "visibility",
};

static const TSFieldMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [1] = {.index = 0, .length = 1},
  [3] = {.index = 1, .length = 2},
  [4] = {.index = 3, .length = 2},
  [5] = {.index = 5, .length = 1},
  [6] = {.index = 6, .length = 1},
  [7] = {.index = 7, .length = 1},
  [8] = {.index = 8, .length = 1},
  [9] = {.index = 9, .length = 2},
  [11] = {.index = 11, .length = 2},
  [12] = {.index = 13, .length = 4},
  [13] = {.index = 17, .length = 6},
  [14] = {.index = 23, .length = 2},
  [15] = {.index = 25, .length = 2},
  [16] = {.index = 27, .length = 1},
  [17] = {.index = 28, .length = 1},
  [18] = {.index = 29, .length = 2},
  [19] = {.index = 31, .length = 2},
  [20] = {.index = 33, .length = 2},
  [21] = {.index = 35, .length = 2},
  [23] = {.index = 37, .length = 2},
  [24] = {.index = 39, .length = 2},
  [25] = {.index = 41, .length = 2},
  [26] = {.index = 43, .length = 2},
  [27] = {.index = 45, .length = 2},
  [28] = {.index = 47, .length = 6},
  [29] = {.index = 53, .length = 3},
  [30] = {.index = 56, .length = 2},
  [31] = {.index = 58, .length = 2},
  [32] = {.index = 60, .length = 2},
  [33] = {.index = 62, .length = 5},
  [34] = {.index = 67, .length = 2},
  [35] = {.index = 69, .length = 1},
  [36] = {.index = 70, .length = 1},
  [37] = {.index = 71, .length = 2},
  [38] = {.index = 73, .length = 2},
  [39] = {.index = 75, .length = 2},
  [40] = {.index = 77, .length = 2},
  [41] = {.index = 79, .length = 3},
  [42] = {.index = 82, .length = 2},
  [43] = {.index = 84, .length = 1},
  [44] = {.index = 85, .length = 2},
  [45] = {.index = 87, .length = 3},
  [46] = {.index = 90, .length = 1},
  [47] = {.index = 91, .length = 2},
  [48] = {.index = 93, .length = 2},
  [50] = {.index = 95, .length = 2},
  [51] = {.index = 97, .length = 1},
  [52] = {.index = 98, .length = 3},
  [53] = {.index = 101, .length = 3},
  [54] = {.index = 104, .length = 2},
  [57] = {.index = 106, .length = 2},
  [58] = {.index = 108, .length = 2},
  [59] = {.index = 110, .length = 3},
  [60] = {.index = 113, .length = 2},
  [61] = {.index = 5, .length = 1},
  [62] = {.index = 115, .length = 1},
  [63] = {.index = 116, .length = 1},
  [64] = {.index = 117, .length = 3},
  [65] = {.index = 120, .length = 3},
  [66] = {.index = 123, .length = 2},
  [67] = {.index = 125, .length = 2},
  [68] = {.index = 127, .length = 3},
  [69] = {.index = 130, .length = 3},
  [70] = {.index = 133, .length = 3},
  [71] = {.index = 136, .length = 3},
  [72] = {.index = 139, .length = 3},
  [73] = {.index = 142, .length = 3},
  [74] = {.index = 145, .length = 2},
  [75] = {.index = 147, .length = 2},
  [76] = {.index = 149, .length = 2},
  [77] = {.index = 151, .length = 2},
  [78] = {.index = 153, .length = 3},
  [79] = {.index = 156, .length = 3},
  [80] = {.index = 159, .length = 2},
  [81] = {.index = 161, .length = 3},
  [82] = {.index = 164, .length = 2},
  [83] = {.index = 166, .length = 3},
  [84] = {.index = 169, .length = 3},
  [85] = {.index = 172, .length = 3},
  [86] = {.index = 175, .length = 2},
  [87] = {.index = 177, .length = 3},
  [88] = {.index = 180, .length = 3},
  [89] = {.index = 183, .length = 1},
  [90] = {.index = 184, .length = 3},
  [91] = {.index = 187, .length = 2},
  [92] = {.index = 189, .length = 2},
  [93] = {.index = 189, .length = 2},
  [94] = {.index = 191, .length = 2},
  [95] = {.index = 193, .length = 4},
  [96] = {.index = 197, .length = 2},
  [97] = {.index = 199, .length = 1},
  [98] = {.index = 200, .length = 2},
  [99] = {.index = 202, .length = 3},
  [100] = {.index = 205, .length = 3},
  [101] = {.index = 208, .length = 2},
  [102] = {.index = 210, .length = 3},
  [103] = {.index = 213, .length = 2},
  [104] = {.index = 215, .length = 4},
  [105] = {.index = 219, .length = 3},
  [106] = {.index = 222, .length = 3},
  [108] = {.index = 225, .length = 4},
  [109] = {.index = 229, .length = 3},
  [110] = {.index = 232, .length = 3},
  [111] = {.index = 235, .length = 4},
  [112] = {.index = 239, .length = 4},
  [113] = {.index = 243, .length = 3},
  [114] = {.index = 246, .length = 1},
  [115] = {.index = 247, .length = 2},
  [116] = {.index = 249, .length = 3},
  [117] = {.index = 252, .length = 4},
  [118] = {.index = 256, .length = 3},
  [119] = {.index = 259, .length = 3},
  [120] = {.index = 262, .length = 3},
  [121] = {.index = 265, .length = 3},
  [122] = {.index = 268, .length = 4},
  [123] = {.index = 272, .length = 4},
  [124] = {.index = 276, .length = 3},
  [125] = {.index = 279, .length = 4},
  [126] = {.index = 283, .length = 3},
  [127] = {.index = 286, .length = 2},
  [128] = {.index = 288, .length = 4},
  [129] = {.index = 292, .length = 4},
  [130] = {.index = 296, .length = 3},
  [131] = {.index = 299, .length = 2},
  [132] = {.index = 301, .length = 1},
  [133] = {.index = 302, .length = 1},
  [134] = {.index = 303, .length = 2},
  [135] = {.index = 305, .length = 1},
  [136] = {.index = 306, .length = 1},
  [137] = {.index = 307, .length = 2},
  [140] = {.index = 309, .length = 3},
  [141] = {.index = 312, .length = 3},
  [142] = {.index = 315, .length = 3},
  [143] = {.index = 318, .length = 4},
  [144] = {.index = 322, .length = 4},
  [145] = {.index = 326, .length = 3},
  [146] = {.index = 326, .length = 3},
  [147] = {.index = 329, .length = 5},
  [148] = {.index = 334, .length = 4},
  [149] = {.index = 338, .length = 3},
  [150] = {.index = 341, .length = 2},
  [151] = {.index = 343, .length = 4},
  [152] = {.index = 347, .length = 4},
  [153] = {.index = 351, .length = 1},
  [154] = {.index = 352, .length = 1},
  [155] = {.index = 353, .length = 4},
  [156] = {.index = 357, .length = 4},
  [157] = {.index = 361, .length = 5},
  [158] = {.index = 366, .length = 4},
  [159] = {.index = 370, .length = 4},
  [160] = {.index = 374, .length = 4},
  [161] = {.index = 378, .length = 1},
  [162] = {.index = 379, .length = 5},
  [163] = {.index = 384, .length = 4},
  [164] = {.index = 388, .length = 2},
  [165] = {.index = 390, .length = 2},
  [166] = {.index = 392, .length = 1},
  [167] = {.index = 393, .length = 2},
  [168] = {.index = 395, .length = 1},
  [169] = {.index = 396, .length = 2},
  [170] = {.index = 398, .length = 1},
  [171] = {.index = 399, .length = 1},
  [172] = {.index = 400, .length = 2},
  [173] = {.index = 402, .length = 3},
  [174] = {.index = 405, .length = 3},
  [175] = {.index = 408, .length = 5},
  [176] = {.index = 413, .length = 3},
  [177] = {.index = 416, .length = 5},
  [178] = {.index = 421, .length = 2},
  [179] = {.index = 423, .length = 2},
  [180] = {.index = 425, .length = 2},
  [181] = {.index = 427, .length = 2},
  [182] = {.index = 429, .length = 5},
  [183] = {.index = 434, .length = 5},
  [184] = {.index = 439, .length = 5},
  [185] = {.index = 444, .length = 2},
  [186] = {.index = 446, .length = 2},
  [187] = {.index = 448, .length = 3},
  [188] = {.index = 451, .length = 3},
  [189] = {.index = 454, .length = 2},
  [190] = {.index = 456, .length = 3},
  [191] = {.index = 459, .length = 2},
  [192] = {.index = 461, .length = 3},
  [193] = {.index = 464, .length = 2},
  [194] = {.index = 466, .length = 2},
  [195] = {.index = 468, .length = 3},
  [196] = {.index = 471, .length = 3},
  [197] = {.index = 474, .length = 3},
  [198] = {.index = 477, .length = 3},
  [199] = {.index = 480, .length = 3},
  [200] = {.index = 483, .length = 3},
  [201] = {.index = 486, .length = 6},
  [202] = {.index = 492, .length = 3},
  [203] = {.index = 495, .length = 4},
  [204] = {.index = 499, .length = 3},
  [205] = {.index = 502, .length = 4},
  [206] = {.index = 506, .length = 4},
  [207] = {.index = 510, .length = 4},
  [208] = {.index = 514, .length = 4},
  [209] = {.index = 518, .length = 2},
  [210] = {.index = 520, .length = 2},
  [211] = {.index = 522, .length = 2},
  [212] = {.index = 524, .length = 5},
  [213] = {.index = 529, .length = 3},
  [214] = {.index = 532, .length = 3},
  [215] = {.index = 535, .length = 3},
  [216] = {.index = 538, .length = 4},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_type, 0, .inherited = true},
  [1] =
    {field_name, 0, .inherited = true},
    {field_object, 0, .inherited = true},
  [3] =
    {field_name, 0, .inherited = true},
    {field_scope, 0, .inherited = true},
  [5] =
    {field_name, 0},
  [6] =
    {field_body, 1},
  [7] =
    {field_parameters, 1},
  [8] =
    {field_prefix, 0},
  [9] =
    {field_argument, 1},
    {field_operator, 0},
  [11] =
    {field_arguments, 1},
    {field_function, 0},
  [13] =
    {field_body, 1},
    {field_name, 0, .inherited = true},
    {field_parameters, 0, .inherited = true},
    {field_return_type, 0, .inherited = true},
  [17] =
    {field_attributes, 0, .inherited = true},
    {field_body, 1},
    {field_parameters, 0, .inherited = true},
    {field_reference_modifier, 0, .inherited = true},
    {field_return_type, 0, .inherited = true},
    {field_static_modifier, 0, .inherited = true},
  [23] =
    {field_argument, 0},
    {field_operator, 1},
  [25] =
    {field_body, 2},
    {field_name, 1},
  [27] =
    {field_name, 1},
  [28] =
    {field_type, 0},
  [29] =
    {field_body, 1, .inherited = true},
    {field_type, 1, .inherited = true},
  [31] =
    {field_parameters, 2},
    {field_reference_modifier, 1},
  [33] =
    {field_parameters, 1},
    {field_return_type, 2, .inherited = true},
  [35] =
    {field_name, 1},
    {field_parameters, 2},
  [37] =
    {field_body, 2},
    {field_condition, 1},
  [39] =
    {field_end_tag, 2},
    {field_identifier, 1},
  [41] =
    {field_name, 2},
    {field_scope, 0},
  [43] =
    {field_prefix, 0},
    {field_prefix, 1},
  [45] =
    {field_parameters, 2},
    {field_static_modifier, 0},
  [47] =
    {field_attributes, 0, .inherited = true},
    {field_body, 2},
    {field_parameters, 0, .inherited = true},
    {field_reference_modifier, 0, .inherited = true},
    {field_return_type, 0, .inherited = true},
    {field_static_modifier, 0, .inherited = true},
  [53] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [56] =
    {field_left, 0},
    {field_right, 2},
  [58] =
    {field_name, 2, .inherited = true},
    {field_object, 0},
  [60] =
    {field_attributes, 0},
    {field_parameters, 2},
  [62] =
    {field_attributes, 0},
    {field_body, 2},
    {field_name, 1, .inherited = true},
    {field_parameters, 1, .inherited = true},
    {field_return_type, 1, .inherited = true},
  [67] =
    {field_name, 0},
    {field_value, 2},
  [69] =
    {field_alias, 2},
  [70] =
    {field_body, 2},
  [71] =
    {field_name, 1},
    {field_reference_modifier, 0},
  [73] =
    {field_name, 1},
    {field_visibility, 0},
  [75] =
    {field_name, 1},
    {field_type, 0},
  [77] =
    {field_attributes, 0},
    {field_name, 1},
  [79] =
    {field_parameters, 2},
    {field_reference_modifier, 1},
    {field_return_type, 3, .inherited = true},
  [82] =
    {field_name, 2},
    {field_parameters, 3},
  [84] =
    {field_return_type, 1},
  [85] =
    {field_parameters, 1},
    {field_return_type, 3, .inherited = true},
  [87] =
    {field_name, 1},
    {field_parameters, 2},
    {field_return_type, 3, .inherited = true},
  [90] =
    {field_type, 1},
  [91] =
    {field_attributes, 0, .inherited = true},
    {field_type, 0, .inherited = true},
  [93] =
    {field_body, 3},
    {field_name, 1},
  [95] =
    {field_type, 1},
    {field_value, 3},
  [97] =
    {field_alternative, 0},
  [98] =
    {field_alternative, 3},
    {field_body, 2},
    {field_condition, 1},
  [101] =
    {field_alternative, 3, .inherited = true},
    {field_body, 2},
    {field_condition, 1},
  [104] =
    {field_attributes, 0},
    {field_body, 2},
  [106] =
    {field_name, 2},
    {field_object, 0},
  [108] =
    {field_end_tag, 3},
    {field_identifier, 1},
  [110] =
    {field_end_tag, 3},
    {field_identifier, 1},
    {field_value, 2},
  [113] =
    {field_end_tag, 3},
    {field_identifier, 2},
  [115] =
    {field_reference_modifier, 0},
  [116] =
    {field_name, 0, .inherited = true},
  [117] =
    {field_parameters, 3},
    {field_reference_modifier, 2},
    {field_static_modifier, 0},
  [120] =
    {field_parameters, 2},
    {field_return_type, 3, .inherited = true},
    {field_static_modifier, 0},
  [123] =
    {field_alternative, 3},
    {field_condition, 0},
  [125] =
    {field_left, 0},
    {field_right, 3},
  [127] =
    {field_arguments, 3},
    {field_name, 2, .inherited = true},
    {field_scope, 0},
  [130] =
    {field_arguments, 3},
    {field_name, 2, .inherited = true},
    {field_object, 0},
  [133] =
    {field_attributes, 0},
    {field_parameters, 3},
    {field_reference_modifier, 2},
  [136] =
    {field_attributes, 0},
    {field_parameters, 2},
    {field_return_type, 3, .inherited = true},
  [139] =
    {field_attributes, 0},
    {field_body, 3},
    {field_name, 2},
  [142] =
    {field_attributes, 0},
    {field_parameters, 3},
    {field_static_modifier, 1},
  [145] =
    {field_body, 3},
    {field_name, 2},
  [147] =
    {field_alias, 3},
    {field_type, 0},
  [149] =
    {field_body, 3},
    {field_type, 0},
  [151] =
    {field_name, 2},
    {field_reference_modifier, 0},
  [153] =
    {field_name, 2},
    {field_readonly, 1},
    {field_visibility, 0},
  [156] =
    {field_name, 2},
    {field_type, 1},
    {field_visibility, 0},
  [159] =
    {field_name, 2},
    {field_type, 0},
  [161] =
    {field_name, 2},
    {field_reference_modifier, 1},
    {field_type, 0},
  [164] =
    {field_attributes, 0},
    {field_name, 2},
  [166] =
    {field_attributes, 0},
    {field_name, 2},
    {field_reference_modifier, 1},
  [169] =
    {field_attributes, 0},
    {field_name, 2},
    {field_visibility, 1},
  [172] =
    {field_attributes, 0},
    {field_name, 2},
    {field_type, 1},
  [175] =
    {field_default_value, 2},
    {field_name, 0},
  [177] =
    {field_parameters, 2},
    {field_reference_modifier, 1},
    {field_return_type, 4, .inherited = true},
  [180] =
    {field_name, 2},
    {field_parameters, 3},
    {field_return_type, 4, .inherited = true},
  [183] =
    {field_type, 1, .inherited = true},
  [184] =
    {field_name, 0, .inherited = true},
    {field_parameters, 0, .inherited = true},
    {field_return_type, 0, .inherited = true},
  [187] =
    {field_attributes, 0},
    {field_type, 1, .inherited = true},
  [189] =
    {field_body, 4},
    {field_name, 1},
  [191] =
    {field_body, 1},
    {field_condition, 3},
  [193] =
    {field_alternative, 3, .inherited = true},
    {field_alternative, 4},
    {field_body, 2},
    {field_condition, 1},
  [197] =
    {field_alternative, 0, .inherited = true},
    {field_alternative, 1, .inherited = true},
  [199] =
    {field_body, 3},
  [200] =
    {field_attributes, 0},
    {field_body, 3},
  [202] =
    {field_end_tag, 4},
    {field_identifier, 1},
    {field_value, 3},
  [205] =
    {field_end_tag, 4},
    {field_identifier, 1},
    {field_value, 2},
  [208] =
    {field_end_tag, 4},
    {field_identifier, 2},
  [210] =
    {field_end_tag, 4},
    {field_identifier, 2},
    {field_value, 3},
  [213] =
    {field_name, 0, .inherited = true},
    {field_reference_modifier, 1},
  [215] =
    {field_parameters, 3},
    {field_reference_modifier, 2},
    {field_return_type, 4, .inherited = true},
    {field_static_modifier, 0},
  [219] =
    {field_parameters, 2},
    {field_return_type, 4, .inherited = true},
    {field_static_modifier, 0},
  [222] =
    {field_alternative, 4},
    {field_body, 2},
    {field_condition, 0},
  [225] =
    {field_attributes, 0},
    {field_parameters, 3},
    {field_reference_modifier, 2},
    {field_return_type, 4, .inherited = true},
  [229] =
    {field_attributes, 0},
    {field_parameters, 2},
    {field_return_type, 4, .inherited = true},
  [232] =
    {field_attributes, 0},
    {field_body, 4},
    {field_name, 2},
  [235] =
    {field_attributes, 0},
    {field_parameters, 4},
    {field_reference_modifier, 3},
    {field_static_modifier, 1},
  [239] =
    {field_attributes, 0},
    {field_parameters, 3},
    {field_return_type, 4, .inherited = true},
    {field_static_modifier, 1},
  [243] =
    {field_attributes, 0},
    {field_body, 4},
    {field_name, 3},
  [246] =
    {field_type, 2},
  [247] =
    {field_body, 4},
    {field_name, 2},
  [249] =
    {field_default_value, 3},
    {field_name, 1},
    {field_reference_modifier, 0},
  [252] =
    {field_name, 3},
    {field_readonly, 1},
    {field_type, 2},
    {field_visibility, 0},
  [256] =
    {field_default_value, 3},
    {field_name, 1},
    {field_visibility, 0},
  [259] =
    {field_name, 3},
    {field_reference_modifier, 1},
    {field_type, 0},
  [262] =
    {field_default_value, 3},
    {field_name, 1},
    {field_type, 0},
  [265] =
    {field_attributes, 0},
    {field_name, 3},
    {field_reference_modifier, 1},
  [268] =
    {field_attributes, 0},
    {field_name, 3},
    {field_readonly, 2},
    {field_visibility, 1},
  [272] =
    {field_attributes, 0},
    {field_name, 3},
    {field_type, 2},
    {field_visibility, 1},
  [276] =
    {field_attributes, 0},
    {field_name, 3},
    {field_type, 1},
  [279] =
    {field_attributes, 0},
    {field_name, 3},
    {field_reference_modifier, 2},
    {field_type, 1},
  [283] =
    {field_attributes, 0},
    {field_default_value, 3},
    {field_name, 1},
  [286] =
    {field_attributes, 0},
    {field_type, 2, .inherited = true},
  [288] =
    {field_attributes, 0},
    {field_name, 1, .inherited = true},
    {field_parameters, 1, .inherited = true},
    {field_return_type, 1, .inherited = true},
  [292] =
    {field_body, 2},
    {field_name, 1, .inherited = true},
    {field_parameters, 1, .inherited = true},
    {field_return_type, 1, .inherited = true},
  [296] =
    {field_name, 1, .inherited = true},
    {field_parameters, 1, .inherited = true},
    {field_return_type, 1, .inherited = true},
  [299] =
    {field_body, 5},
    {field_name, 1},
  [301] =
    {field_body, 5},
  [302] =
    {field_return_expression, 2},
  [303] =
    {field_conditional_expressions, 0},
    {field_return_expression, 2},
  [305] =
    {field_value, 1},
  [306] =
    {field_body, 4},
  [307] =
    {field_attributes, 0},
    {field_body, 4},
  [309] =
    {field_end_tag, 5},
    {field_identifier, 1},
    {field_value, 3},
  [312] =
    {field_end_tag, 5},
    {field_identifier, 2},
    {field_value, 4},
  [315] =
    {field_end_tag, 5},
    {field_identifier, 2},
    {field_value, 3},
  [318] =
    {field_parameters, 3},
    {field_reference_modifier, 2},
    {field_return_type, 5, .inherited = true},
    {field_static_modifier, 0},
  [322] =
    {field_attributes, 0},
    {field_parameters, 3},
    {field_reference_modifier, 2},
    {field_return_type, 5, .inherited = true},
  [326] =
    {field_attributes, 0},
    {field_body, 5},
    {field_name, 2},
  [329] =
    {field_attributes, 0},
    {field_parameters, 4},
    {field_reference_modifier, 3},
    {field_return_type, 5, .inherited = true},
    {field_static_modifier, 1},
  [334] =
    {field_attributes, 0},
    {field_parameters, 3},
    {field_return_type, 5, .inherited = true},
    {field_static_modifier, 1},
  [338] =
    {field_attributes, 0},
    {field_body, 5},
    {field_name, 3},
  [341] =
    {field_body, 5},
    {field_name, 2},
  [343] =
    {field_default_value, 4},
    {field_name, 2},
    {field_readonly, 1},
    {field_visibility, 0},
  [347] =
    {field_default_value, 4},
    {field_name, 2},
    {field_type, 1},
    {field_visibility, 0},
  [351] =
    {field_body, 1, .inherited = true},
  [352] =
    {field_body, 0},
  [353] =
    {field_default_value, 4},
    {field_name, 2},
    {field_reference_modifier, 1},
    {field_type, 0},
  [357] =
    {field_attributes, 0},
    {field_default_value, 4},
    {field_name, 2},
    {field_reference_modifier, 1},
  [361] =
    {field_attributes, 0},
    {field_name, 4},
    {field_readonly, 2},
    {field_type, 3},
    {field_visibility, 1},
  [366] =
    {field_attributes, 0},
    {field_default_value, 4},
    {field_name, 2},
    {field_visibility, 1},
  [370] =
    {field_attributes, 0},
    {field_name, 4},
    {field_reference_modifier, 2},
    {field_type, 1},
  [374] =
    {field_attributes, 0},
    {field_default_value, 4},
    {field_name, 2},
    {field_type, 1},
  [378] =
    {field_attributes, 0},
  [379] =
    {field_attributes, 0},
    {field_body, 3},
    {field_name, 2, .inherited = true},
    {field_parameters, 2, .inherited = true},
    {field_return_type, 2, .inherited = true},
  [384] =
    {field_attributes, 0},
    {field_name, 2, .inherited = true},
    {field_parameters, 2, .inherited = true},
    {field_return_type, 2, .inherited = true},
  [388] =
    {field_body, 4},
    {field_type, 2},
  [390] =
    {field_body, 6},
    {field_update, 4},
  [392] =
    {field_update, 4},
  [393] =
    {field_body, 6},
    {field_condition, 3},
  [395] =
    {field_condition, 3},
  [396] =
    {field_body, 6},
    {field_initialize, 2},
  [398] =
    {field_initialize, 2},
  [399] =
    {field_body, 6},
  [400] =
    {field_attributes, 0},
    {field_body, 5},
  [402] =
    {field_end_tag, 6},
    {field_identifier, 2},
    {field_value, 4},
  [405] =
    {field_attributes, 0},
    {field_body, 6},
    {field_name, 2},
  [408] =
    {field_attributes, 0},
    {field_parameters, 4},
    {field_reference_modifier, 3},
    {field_return_type, 6, .inherited = true},
    {field_static_modifier, 1},
  [413] =
    {field_attributes, 0},
    {field_body, 6},
    {field_name, 3},
  [416] =
    {field_default_value, 5},
    {field_name, 3},
    {field_readonly, 1},
    {field_type, 2},
    {field_visibility, 0},
  [421] =
    {field_body, 2, .inherited = true},
    {field_parameters, 1},
  [423] =
    {field_body, 2, .inherited = true},
    {field_reference_modifier, 0},
  [425] =
    {field_body, 2, .inherited = true},
    {field_final, 0},
  [427] =
    {field_attributes, 0},
    {field_body, 2, .inherited = true},
  [429] =
    {field_attributes, 0},
    {field_default_value, 5},
    {field_name, 3},
    {field_readonly, 2},
    {field_visibility, 1},
  [434] =
    {field_attributes, 0},
    {field_default_value, 5},
    {field_name, 3},
    {field_type, 2},
    {field_visibility, 1},
  [439] =
    {field_attributes, 0},
    {field_default_value, 5},
    {field_name, 3},
    {field_reference_modifier, 2},
    {field_type, 1},
  [444] =
    {field_attributes, 0},
    {field_type, 2},
  [446] =
    {field_name, 1},
    {field_value, 3},
  [448] =
    {field_body, 5},
    {field_name, 3},
    {field_type, 2},
  [451] =
    {field_body, 7},
    {field_condition, 3},
    {field_update, 5},
  [454] =
    {field_condition, 3},
    {field_update, 5},
  [456] =
    {field_body, 7},
    {field_initialize, 2},
    {field_update, 5},
  [459] =
    {field_initialize, 2},
    {field_update, 5},
  [461] =
    {field_body, 7},
    {field_condition, 4},
    {field_initialize, 2},
  [464] =
    {field_condition, 4},
    {field_initialize, 2},
  [466] =
    {field_attributes, 0},
    {field_body, 6},
  [468] =
    {field_body, 3, .inherited = true},
    {field_parameters, 2},
    {field_reference_modifier, 0},
  [471] =
    {field_body, 3, .inherited = true},
    {field_final, 0},
    {field_parameters, 2},
  [474] =
    {field_body, 3, .inherited = true},
    {field_final, 0},
    {field_reference_modifier, 1},
  [477] =
    {field_attributes, 0},
    {field_body, 3, .inherited = true},
    {field_parameters, 2},
  [480] =
    {field_attributes, 0},
    {field_body, 3, .inherited = true},
    {field_reference_modifier, 1},
  [483] =
    {field_attributes, 0},
    {field_body, 3, .inherited = true},
    {field_final, 1},
  [486] =
    {field_attributes, 0},
    {field_default_value, 6},
    {field_name, 4},
    {field_readonly, 2},
    {field_type, 3},
    {field_visibility, 1},
  [492] =
    {field_attributes, 0},
    {field_name, 2},
    {field_value, 4},
  [495] =
    {field_body, 8},
    {field_condition, 4},
    {field_initialize, 2},
    {field_update, 6},
  [499] =
    {field_condition, 4},
    {field_initialize, 2},
    {field_update, 6},
  [502] =
    {field_body, 4, .inherited = true},
    {field_final, 0},
    {field_parameters, 3},
    {field_reference_modifier, 1},
  [506] =
    {field_attributes, 0},
    {field_body, 4, .inherited = true},
    {field_parameters, 3},
    {field_reference_modifier, 1},
  [510] =
    {field_attributes, 0},
    {field_body, 4, .inherited = true},
    {field_final, 1},
    {field_parameters, 3},
  [514] =
    {field_attributes, 0},
    {field_body, 4, .inherited = true},
    {field_final, 1},
    {field_reference_modifier, 2},
  [518] =
    {field_body, 7},
    {field_update, 4},
  [520] =
    {field_body, 7},
    {field_condition, 3},
  [522] =
    {field_body, 7},
    {field_initialize, 2},
  [524] =
    {field_attributes, 0},
    {field_body, 5, .inherited = true},
    {field_final, 1},
    {field_parameters, 4},
    {field_reference_modifier, 2},
  [529] =
    {field_body, 8},
    {field_condition, 3},
    {field_update, 5},
  [532] =
    {field_body, 8},
    {field_initialize, 2},
    {field_update, 5},
  [535] =
    {field_body, 8},
    {field_condition, 4},
    {field_initialize, 2},
  [538] =
    {field_body, 9},
    {field_condition, 4},
    {field_initialize, 2},
    {field_update, 6},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [2] = {
    [0] = sym_object_creation_expression,
  },
  [10] = {
    [0] = sym_list_literal,
  },
  [22] = {
    [0] = anon_sym_array,
  },
  [49] = {
    [2] = alias_sym_operation,
  },
  [55] = {
    [1] = sym_list_literal,
  },
  [56] = {
    [2] = sym_list_literal,
  },
  [61] = {
    [0] = sym_name,
  },
  [92] = {
    [3] = sym_primitive_type,
  },
  [107] = {
    [3] = sym_name,
  },
  [131] = {
    [3] = sym_primitive_type,
  },
  [138] = {
    [3] = sym_list_literal,
  },
  [139] = {
    [4] = sym_list_literal,
  },
  [145] = {
    [4] = sym_primitive_type,
  },
  [174] = {
    [4] = sym_primitive_type,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym_expression, 2,
    sym_expression,
    sym_name,
  sym__new_dereferencable_expression, 2,
    sym__new_dereferencable_expression,
    sym_object_creation_expression,
  sym__list_destructing, 2,
    sym__list_destructing,
    sym_list_literal,
  sym__array_destructing, 2,
    sym__array_destructing,
    sym_list_literal,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 5,
  [6] = 6,
  [7] = 7,
  [8] = 8,
  [9] = 9,
  [10] = 9,
  [11] = 2,
  [12] = 7,
  [13] = 8,
  [14] = 9,
  [15] = 9,
  [16] = 16,
  [17] = 17,
  [18] = 18,
  [19] = 19,
  [20] = 20,
  [21] = 21,
  [22] = 19,
  [23] = 20,
  [24] = 24,
  [25] = 25,
  [26] = 26,
  [27] = 16,
  [28] = 28,
  [29] = 17,
  [30] = 18,
  [31] = 20,
  [32] = 28,
  [33] = 25,
  [34] = 21,
  [35] = 35,
  [36] = 19,
  [37] = 16,
  [38] = 24,
  [39] = 25,
  [40] = 21,
  [41] = 28,
  [42] = 17,
  [43] = 18,
  [44] = 20,
  [45] = 21,
  [46] = 19,
  [47] = 16,
  [48] = 24,
  [49] = 25,
  [50] = 50,
  [51] = 28,
  [52] = 17,
  [53] = 18,
  [54] = 24,
  [55] = 55,
  [56] = 56,
  [57] = 57,
  [58] = 58,
  [59] = 59,
  [60] = 60,
  [61] = 61,
  [62] = 62,
  [63] = 63,
  [64] = 64,
  [65] = 65,
  [66] = 66,
  [67] = 67,
  [68] = 68,
  [69] = 69,
  [70] = 70,
  [71] = 71,
  [72] = 72,
  [73] = 73,
  [74] = 74,
  [75] = 55,
  [76] = 62,
  [77] = 77,
  [78] = 78,
  [79] = 62,
  [80] = 56,
  [81] = 57,
  [82] = 58,
  [83] = 59,
  [84] = 60,
  [85] = 61,
  [86] = 63,
  [87] = 64,
  [88] = 65,
  [89] = 66,
  [90] = 67,
  [91] = 68,
  [92] = 69,
  [93] = 70,
  [94] = 71,
  [95] = 72,
  [96] = 73,
  [97] = 74,
  [98] = 62,
  [99] = 77,
  [100] = 78,
  [101] = 77,
  [102] = 77,
  [103] = 78,
  [104] = 77,
  [105] = 78,
  [106] = 55,
  [107] = 55,
  [108] = 78,
  [109] = 109,
  [110] = 110,
  [111] = 111,
  [112] = 111,
  [113] = 111,
  [114] = 111,
  [115] = 109,
  [116] = 110,
  [117] = 117,
  [118] = 117,
  [119] = 117,
  [120] = 117,
  [121] = 121,
  [122] = 122,
  [123] = 121,
  [124] = 122,
  [125] = 122,
  [126] = 121,
  [127] = 121,
  [128] = 121,
  [129] = 122,
  [130] = 122,
  [131] = 122,
  [132] = 132,
  [133] = 132,
  [134] = 132,
  [135] = 135,
  [136] = 135,
  [137] = 132,
  [138] = 132,
  [139] = 132,
  [140] = 135,
  [141] = 135,
  [142] = 135,
  [143] = 143,
  [144] = 143,
  [145] = 145,
  [146] = 143,
  [147] = 145,
  [148] = 143,
  [149] = 145,
  [150] = 145,
  [151] = 143,
  [152] = 143,
  [153] = 145,
  [154] = 145,
  [155] = 155,
  [156] = 156,
  [157] = 157,
  [158] = 156,
  [159] = 155,
  [160] = 160,
  [161] = 161,
  [162] = 160,
  [163] = 163,
  [164] = 164,
  [165] = 165,
  [166] = 161,
  [167] = 165,
  [168] = 164,
  [169] = 163,
  [170] = 170,
  [171] = 171,
  [172] = 172,
  [173] = 172,
  [174] = 174,
  [175] = 175,
  [176] = 175,
  [177] = 177,
  [178] = 171,
  [179] = 177,
  [180] = 180,
  [181] = 181,
  [182] = 182,
  [183] = 180,
  [184] = 180,
  [185] = 185,
  [186] = 180,
  [187] = 187,
  [188] = 187,
  [189] = 189,
  [190] = 190,
  [191] = 191,
  [192] = 187,
  [193] = 193,
  [194] = 194,
  [195] = 190,
  [196] = 196,
  [197] = 187,
  [198] = 194,
  [199] = 199,
  [200] = 200,
  [201] = 201,
  [202] = 199,
  [203] = 189,
  [204] = 201,
  [205] = 196,
  [206] = 189,
  [207] = 201,
  [208] = 200,
  [209] = 209,
  [210] = 191,
  [211] = 191,
  [212] = 196,
  [213] = 193,
  [214] = 196,
  [215] = 189,
  [216] = 201,
  [217] = 200,
  [218] = 193,
  [219] = 193,
  [220] = 191,
  [221] = 200,
  [222] = 222,
  [223] = 223,
  [224] = 224,
  [225] = 225,
  [226] = 226,
  [227] = 226,
  [228] = 226,
  [229] = 226,
  [230] = 226,
  [231] = 225,
  [232] = 226,
  [233] = 233,
  [234] = 234,
  [235] = 235,
  [236] = 233,
  [237] = 237,
  [238] = 238,
  [239] = 239,
  [240] = 240,
  [241] = 234,
  [242] = 242,
  [243] = 242,
  [244] = 239,
  [245] = 234,
  [246] = 242,
  [247] = 247,
  [248] = 234,
  [249] = 233,
  [250] = 240,
  [251] = 239,
  [252] = 239,
  [253] = 239,
  [254] = 238,
  [255] = 239,
  [256] = 238,
  [257] = 257,
  [258] = 242,
  [259] = 259,
  [260] = 260,
  [261] = 261,
  [262] = 262,
  [263] = 263,
  [264] = 264,
  [265] = 265,
  [266] = 266,
  [267] = 267,
  [268] = 268,
  [269] = 269,
  [270] = 270,
  [271] = 271,
  [272] = 272,
  [273] = 273,
  [274] = 274,
  [275] = 275,
  [276] = 276,
  [277] = 277,
  [278] = 278,
  [279] = 279,
  [280] = 280,
  [281] = 281,
  [282] = 282,
  [283] = 283,
  [284] = 284,
  [285] = 285,
  [286] = 286,
  [287] = 287,
  [288] = 288,
  [289] = 289,
  [290] = 290,
  [291] = 291,
  [292] = 292,
  [293] = 293,
  [294] = 294,
  [295] = 295,
  [296] = 296,
  [297] = 260,
  [298] = 261,
  [299] = 262,
  [300] = 263,
  [301] = 264,
  [302] = 265,
  [303] = 266,
  [304] = 267,
  [305] = 274,
  [306] = 275,
  [307] = 276,
  [308] = 277,
  [309] = 259,
  [310] = 278,
  [311] = 279,
  [312] = 280,
  [313] = 282,
  [314] = 284,
  [315] = 315,
  [316] = 316,
  [317] = 317,
  [318] = 288,
  [319] = 289,
  [320] = 320,
  [321] = 286,
  [322] = 315,
  [323] = 316,
  [324] = 324,
  [325] = 325,
  [326] = 326,
  [327] = 320,
  [328] = 287,
  [329] = 290,
  [330] = 330,
  [331] = 291,
  [332] = 292,
  [333] = 333,
  [334] = 293,
  [335] = 317,
  [336] = 320,
  [337] = 295,
  [338] = 338,
  [339] = 339,
  [340] = 324,
  [341] = 325,
  [342] = 342,
  [343] = 343,
  [344] = 344,
  [345] = 345,
  [346] = 286,
  [347] = 287,
  [348] = 288,
  [349] = 289,
  [350] = 290,
  [351] = 291,
  [352] = 292,
  [353] = 330,
  [354] = 296,
  [355] = 260,
  [356] = 261,
  [357] = 262,
  [358] = 263,
  [359] = 264,
  [360] = 265,
  [361] = 266,
  [362] = 267,
  [363] = 274,
  [364] = 275,
  [365] = 276,
  [366] = 277,
  [367] = 259,
  [368] = 278,
  [369] = 279,
  [370] = 280,
  [371] = 282,
  [372] = 284,
  [373] = 315,
  [374] = 316,
  [375] = 317,
  [376] = 295,
  [377] = 377,
  [378] = 378,
  [379] = 338,
  [380] = 339,
  [381] = 320,
  [382] = 295,
  [383] = 338,
  [384] = 339,
  [385] = 296,
  [386] = 325,
  [387] = 286,
  [388] = 287,
  [389] = 288,
  [390] = 289,
  [391] = 290,
  [392] = 291,
  [393] = 292,
  [394] = 293,
  [395] = 296,
  [396] = 260,
  [397] = 261,
  [398] = 262,
  [399] = 263,
  [400] = 264,
  [401] = 265,
  [402] = 266,
  [403] = 267,
  [404] = 274,
  [405] = 275,
  [406] = 276,
  [407] = 277,
  [408] = 259,
  [409] = 278,
  [410] = 279,
  [411] = 280,
  [412] = 282,
  [413] = 284,
  [414] = 315,
  [415] = 316,
  [416] = 317,
  [417] = 320,
  [418] = 339,
  [419] = 325,
  [420] = 339,
  [421] = 325,
  [422] = 339,
  [423] = 325,
  [424] = 339,
  [425] = 339,
  [426] = 339,
  [427] = 427,
  [428] = 428,
  [429] = 429,
  [430] = 430,
  [431] = 431,
  [432] = 432,
  [433] = 433,
  [434] = 434,
  [435] = 330,
  [436] = 330,
  [437] = 293,
  [438] = 438,
  [439] = 438,
  [440] = 438,
  [441] = 438,
  [442] = 438,
  [443] = 443,
  [444] = 444,
  [445] = 445,
  [446] = 446,
  [447] = 445,
  [448] = 446,
  [449] = 449,
  [450] = 450,
  [451] = 451,
  [452] = 452,
  [453] = 453,
  [454] = 454,
  [455] = 455,
  [456] = 456,
  [457] = 457,
  [458] = 458,
  [459] = 459,
  [460] = 460,
  [461] = 451,
  [462] = 462,
  [463] = 463,
  [464] = 464,
  [465] = 465,
  [466] = 466,
  [467] = 467,
  [468] = 468,
  [469] = 469,
  [470] = 470,
  [471] = 471,
  [472] = 472,
  [473] = 473,
  [474] = 474,
  [475] = 475,
  [476] = 476,
  [477] = 477,
  [478] = 478,
  [479] = 479,
  [480] = 480,
  [481] = 481,
  [482] = 482,
  [483] = 483,
  [484] = 484,
  [485] = 485,
  [486] = 486,
  [487] = 487,
  [488] = 488,
  [489] = 489,
  [490] = 490,
  [491] = 491,
  [492] = 492,
  [493] = 493,
  [494] = 494,
  [495] = 495,
  [496] = 496,
  [497] = 497,
  [498] = 498,
  [499] = 499,
  [500] = 500,
  [501] = 501,
  [502] = 502,
  [503] = 503,
  [504] = 504,
  [505] = 505,
  [506] = 506,
  [507] = 507,
  [508] = 508,
  [509] = 509,
  [510] = 510,
  [511] = 511,
  [512] = 512,
  [513] = 513,
  [514] = 514,
  [515] = 515,
  [516] = 516,
  [517] = 517,
  [518] = 518,
  [519] = 519,
  [520] = 520,
  [521] = 521,
  [522] = 522,
  [523] = 523,
  [524] = 524,
  [525] = 525,
  [526] = 526,
  [527] = 527,
  [528] = 453,
  [529] = 529,
  [530] = 530,
  [531] = 531,
  [532] = 532,
  [533] = 533,
  [534] = 534,
  [535] = 535,
  [536] = 536,
  [537] = 537,
  [538] = 538,
  [539] = 539,
  [540] = 540,
  [541] = 541,
  [542] = 542,
  [543] = 543,
  [544] = 544,
  [545] = 545,
  [546] = 546,
  [547] = 547,
  [548] = 548,
  [549] = 549,
  [550] = 550,
  [551] = 551,
  [552] = 552,
  [553] = 553,
  [554] = 554,
  [555] = 555,
  [556] = 556,
  [557] = 557,
  [558] = 558,
  [559] = 559,
  [560] = 560,
  [561] = 561,
  [562] = 562,
  [563] = 563,
  [564] = 564,
  [565] = 565,
  [566] = 566,
  [567] = 567,
  [568] = 568,
  [569] = 569,
  [570] = 570,
  [571] = 571,
  [572] = 572,
  [573] = 573,
  [574] = 574,
  [575] = 575,
  [576] = 576,
  [577] = 577,
  [578] = 578,
  [579] = 579,
  [580] = 579,
  [581] = 579,
  [582] = 579,
  [583] = 583,
  [584] = 584,
  [585] = 585,
  [586] = 586,
  [587] = 587,
  [588] = 588,
  [589] = 589,
  [590] = 590,
  [591] = 591,
  [592] = 592,
  [593] = 593,
  [594] = 594,
  [595] = 595,
  [596] = 596,
  [597] = 597,
  [598] = 598,
  [599] = 599,
  [600] = 600,
  [601] = 601,
  [602] = 602,
  [603] = 603,
  [604] = 604,
  [605] = 605,
  [606] = 606,
  [607] = 607,
  [608] = 608,
  [609] = 609,
  [610] = 610,
  [611] = 611,
  [612] = 612,
  [613] = 613,
  [614] = 614,
  [615] = 588,
  [616] = 616,
  [617] = 589,
  [618] = 597,
  [619] = 606,
  [620] = 607,
  [621] = 621,
  [622] = 608,
  [623] = 584,
  [624] = 600,
  [625] = 625,
  [626] = 603,
  [627] = 627,
  [628] = 602,
  [629] = 587,
  [630] = 630,
  [631] = 611,
  [632] = 601,
  [633] = 609,
  [634] = 586,
  [635] = 585,
  [636] = 593,
  [637] = 596,
  [638] = 616,
  [639] = 639,
  [640] = 594,
  [641] = 598,
  [642] = 642,
  [643] = 643,
  [644] = 616,
  [645] = 604,
  [646] = 643,
  [647] = 647,
  [648] = 591,
  [649] = 592,
  [650] = 612,
  [651] = 610,
  [652] = 595,
  [653] = 605,
  [654] = 614,
  [655] = 613,
  [656] = 656,
  [657] = 590,
  [658] = 603,
  [659] = 621,
  [660] = 616,
  [661] = 599,
  [662] = 621,
  [663] = 643,
  [664] = 664,
  [665] = 665,
  [666] = 621,
  [667] = 643,
  [668] = 665,
  [669] = 614,
  [670] = 670,
  [671] = 671,
  [672] = 613,
  [673] = 673,
  [674] = 674,
  [675] = 674,
  [676] = 676,
  [677] = 677,
  [678] = 673,
  [679] = 676,
  [680] = 680,
  [681] = 681,
  [682] = 682,
  [683] = 681,
  [684] = 684,
  [685] = 685,
  [686] = 686,
  [687] = 687,
  [688] = 687,
  [689] = 686,
  [690] = 686,
  [691] = 685,
  [692] = 692,
  [693] = 693,
  [694] = 687,
  [695] = 695,
  [696] = 686,
  [697] = 697,
  [698] = 686,
  [699] = 692,
  [700] = 700,
  [701] = 701,
  [702] = 702,
  [703] = 703,
  [704] = 704,
  [705] = 517,
  [706] = 588,
  [707] = 589,
  [708] = 478,
  [709] = 593,
  [710] = 607,
  [711] = 608,
  [712] = 712,
  [713] = 701,
  [714] = 594,
  [715] = 609,
  [716] = 596,
  [717] = 604,
  [718] = 598,
  [719] = 584,
  [720] = 601,
  [721] = 587,
  [722] = 611,
  [723] = 585,
  [724] = 600,
  [725] = 725,
  [726] = 606,
  [727] = 593,
  [728] = 728,
  [729] = 586,
  [730] = 730,
  [731] = 597,
  [732] = 602,
  [733] = 733,
  [734] = 595,
  [735] = 590,
  [736] = 588,
  [737] = 737,
  [738] = 599,
  [739] = 612,
  [740] = 589,
  [741] = 592,
  [742] = 591,
  [743] = 743,
  [744] = 610,
  [745] = 605,
  [746] = 746,
  [747] = 747,
  [748] = 748,
  [749] = 749,
  [750] = 750,
  [751] = 751,
  [752] = 752,
  [753] = 753,
  [754] = 703,
  [755] = 755,
  [756] = 756,
  [757] = 605,
  [758] = 612,
  [759] = 592,
  [760] = 594,
  [761] = 601,
  [762] = 586,
  [763] = 595,
  [764] = 596,
  [765] = 598,
  [766] = 600,
  [767] = 590,
  [768] = 599,
  [769] = 591,
  [770] = 584,
  [771] = 606,
  [772] = 607,
  [773] = 608,
  [774] = 609,
  [775] = 610,
  [776] = 587,
  [777] = 611,
  [778] = 585,
  [779] = 597,
  [780] = 602,
  [781] = 604,
  [782] = 782,
  [783] = 783,
  [784] = 784,
  [785] = 785,
  [786] = 786,
  [787] = 787,
  [788] = 788,
  [789] = 789,
  [790] = 790,
  [791] = 791,
  [792] = 792,
  [793] = 793,
  [794] = 794,
  [795] = 795,
  [796] = 782,
  [797] = 453,
  [798] = 583,
  [799] = 799,
  [800] = 800,
  [801] = 801,
  [802] = 802,
  [803] = 803,
  [804] = 804,
  [805] = 805,
  [806] = 806,
  [807] = 807,
  [808] = 808,
  [809] = 809,
  [810] = 810,
  [811] = 811,
  [812] = 812,
  [813] = 813,
  [814] = 814,
  [815] = 815,
  [816] = 816,
  [817] = 817,
  [818] = 818,
  [819] = 819,
  [820] = 820,
  [821] = 821,
  [822] = 822,
  [823] = 823,
  [824] = 824,
  [825] = 825,
  [826] = 826,
  [827] = 451,
  [828] = 828,
  [829] = 728,
  [830] = 830,
  [831] = 588,
  [832] = 589,
  [833] = 833,
  [834] = 834,
  [835] = 833,
  [836] = 836,
  [837] = 593,
  [838] = 838,
  [839] = 834,
  [840] = 684,
  [841] = 725,
  [842] = 684,
  [843] = 598,
  [844] = 600,
  [845] = 604,
  [846] = 584,
  [847] = 606,
  [848] = 607,
  [849] = 608,
  [850] = 609,
  [851] = 851,
  [852] = 851,
  [853] = 594,
  [854] = 854,
  [855] = 793,
  [856] = 597,
  [857] = 587,
  [858] = 703,
  [859] = 611,
  [860] = 585,
  [861] = 602,
  [862] = 712,
  [863] = 730,
  [864] = 601,
  [865] = 865,
  [866] = 866,
  [867] = 586,
  [868] = 596,
  [869] = 593,
  [870] = 746,
  [871] = 700,
  [872] = 599,
  [873] = 873,
  [874] = 700,
  [875] = 733,
  [876] = 876,
  [877] = 877,
  [878] = 591,
  [879] = 610,
  [880] = 725,
  [881] = 728,
  [882] = 595,
  [883] = 588,
  [884] = 589,
  [885] = 793,
  [886] = 592,
  [887] = 612,
  [888] = 605,
  [889] = 590,
  [890] = 751,
  [891] = 809,
  [892] = 786,
  [893] = 593,
  [894] = 787,
  [895] = 811,
  [896] = 815,
  [897] = 783,
  [898] = 804,
  [899] = 816,
  [900] = 817,
  [901] = 805,
  [902] = 806,
  [903] = 605,
  [904] = 904,
  [905] = 818,
  [906] = 819,
  [907] = 820,
  [908] = 821,
  [909] = 822,
  [910] = 812,
  [911] = 813,
  [912] = 612,
  [913] = 592,
  [914] = 594,
  [915] = 753,
  [916] = 800,
  [917] = 743,
  [918] = 793,
  [919] = 737,
  [920] = 597,
  [921] = 602,
  [922] = 789,
  [923] = 790,
  [924] = 924,
  [925] = 749,
  [926] = 792,
  [927] = 814,
  [928] = 830,
  [929] = 601,
  [930] = 750,
  [931] = 801,
  [932] = 823,
  [933] = 586,
  [934] = 810,
  [935] = 596,
  [936] = 803,
  [937] = 747,
  [938] = 598,
  [939] = 600,
  [940] = 590,
  [941] = 784,
  [942] = 807,
  [943] = 478,
  [944] = 599,
  [945] = 794,
  [946] = 591,
  [947] = 604,
  [948] = 584,
  [949] = 606,
  [950] = 607,
  [951] = 608,
  [952] = 795,
  [953] = 583,
  [954] = 799,
  [955] = 824,
  [956] = 825,
  [957] = 826,
  [958] = 788,
  [959] = 791,
  [960] = 517,
  [961] = 609,
  [962] = 748,
  [963] = 583,
  [964] = 703,
  [965] = 610,
  [966] = 752,
  [967] = 587,
  [968] = 611,
  [969] = 585,
  [970] = 802,
  [971] = 785,
  [972] = 904,
  [973] = 808,
  [974] = 595,
  [975] = 725,
  [976] = 976,
  [977] = 977,
  [978] = 978,
  [979] = 836,
  [980] = 980,
  [981] = 981,
  [982] = 982,
  [983] = 983,
  [984] = 828,
  [985] = 985,
  [986] = 684,
  [987] = 838,
  [988] = 988,
  [989] = 989,
  [990] = 793,
  [991] = 991,
  [992] = 992,
  [993] = 993,
  [994] = 994,
  [995] = 755,
  [996] = 996,
  [997] = 756,
  [998] = 728,
  [999] = 999,
  [1000] = 1000,
  [1001] = 1001,
  [1002] = 1002,
  [1003] = 1003,
  [1004] = 1004,
  [1005] = 1005,
  [1006] = 1006,
  [1007] = 1007,
  [1008] = 593,
  [1009] = 876,
  [1010] = 700,
  [1011] = 877,
  [1012] = 1012,
  [1013] = 1013,
  [1014] = 1014,
  [1015] = 1015,
  [1016] = 583,
  [1017] = 1017,
  [1018] = 1018,
  [1019] = 1019,
  [1020] = 1020,
  [1021] = 1021,
  [1022] = 1022,
  [1023] = 1023,
  [1024] = 1024,
  [1025] = 1025,
  [1026] = 1026,
  [1027] = 1027,
  [1028] = 1028,
  [1029] = 1029,
  [1030] = 1030,
  [1031] = 1031,
  [1032] = 1032,
  [1033] = 1033,
  [1034] = 1034,
  [1035] = 1035,
  [1036] = 1036,
  [1037] = 1037,
  [1038] = 1038,
  [1039] = 1039,
  [1040] = 1040,
  [1041] = 1041,
  [1042] = 1042,
  [1043] = 1043,
  [1044] = 1044,
  [1045] = 1045,
  [1046] = 1022,
  [1047] = 1023,
  [1048] = 996,
  [1049] = 1001,
  [1050] = 1025,
  [1051] = 1032,
  [1052] = 976,
  [1053] = 1028,
  [1054] = 1037,
  [1055] = 1031,
  [1056] = 1036,
  [1057] = 453,
  [1058] = 1012,
  [1059] = 1041,
  [1060] = 1044,
  [1061] = 1013,
  [1062] = 1035,
  [1063] = 1040,
  [1064] = 1043,
  [1065] = 1015,
  [1066] = 1017,
  [1067] = 1018,
  [1068] = 1019,
  [1069] = 1021,
  [1070] = 1022,
  [1071] = 1023,
  [1072] = 1024,
  [1073] = 1025,
  [1074] = 1032,
  [1075] = 1029,
  [1076] = 1037,
  [1077] = 1039,
  [1078] = 1042,
  [1079] = 1014,
  [1080] = 1014,
  [1081] = 1038,
  [1082] = 1029,
  [1083] = 988,
  [1084] = 1084,
  [1085] = 1042,
  [1086] = 1027,
  [1087] = 1039,
  [1088] = 451,
  [1089] = 985,
  [1090] = 980,
  [1091] = 1033,
  [1092] = 1002,
  [1093] = 1036,
  [1094] = 982,
  [1095] = 1003,
  [1096] = 1006,
  [1097] = 991,
  [1098] = 983,
  [1099] = 1000,
  [1100] = 977,
  [1101] = 1005,
  [1102] = 1030,
  [1103] = 993,
  [1104] = 1031,
  [1105] = 1004,
  [1106] = 978,
  [1107] = 1038,
  [1108] = 981,
  [1109] = 1034,
  [1110] = 1007,
  [1111] = 1012,
  [1112] = 992,
  [1113] = 1041,
  [1114] = 1044,
  [1115] = 1013,
  [1116] = 1035,
  [1117] = 1040,
  [1118] = 1043,
  [1119] = 989,
  [1120] = 1015,
  [1121] = 1017,
  [1122] = 1028,
  [1123] = 1030,
  [1124] = 1020,
  [1125] = 1026,
  [1126] = 1027,
  [1127] = 1033,
  [1128] = 1034,
  [1129] = 999,
  [1130] = 1018,
  [1131] = 1020,
  [1132] = 994,
  [1133] = 1026,
  [1134] = 1019,
  [1135] = 1021,
  [1136] = 1024,
  [1137] = 1043,
  [1138] = 1015,
  [1139] = 1017,
  [1140] = 1018,
  [1141] = 1019,
  [1142] = 1021,
  [1143] = 1022,
  [1144] = 1023,
  [1145] = 1024,
  [1146] = 1025,
  [1147] = 1032,
  [1148] = 1037,
  [1149] = 1039,
  [1150] = 1042,
  [1151] = 1036,
  [1152] = 1014,
  [1153] = 1029,
  [1154] = 1031,
  [1155] = 1038,
  [1156] = 1012,
  [1157] = 1028,
  [1158] = 1030,
  [1159] = 1020,
  [1160] = 1026,
  [1161] = 1027,
  [1162] = 1033,
  [1163] = 1034,
  [1164] = 1041,
  [1165] = 1044,
  [1166] = 1013,
  [1167] = 1035,
  [1168] = 1040,
  [1169] = 1169,
  [1170] = 1170,
  [1171] = 1171,
  [1172] = 1172,
  [1173] = 1173,
  [1174] = 1174,
  [1175] = 1175,
  [1176] = 1176,
  [1177] = 1177,
  [1178] = 1178,
  [1179] = 1179,
  [1180] = 1180,
  [1181] = 1181,
  [1182] = 1182,
  [1183] = 1183,
  [1184] = 1184,
  [1185] = 1020,
  [1186] = 1186,
  [1187] = 1184,
  [1188] = 1180,
  [1189] = 1189,
  [1190] = 1182,
  [1191] = 1191,
  [1192] = 1192,
  [1193] = 1193,
  [1194] = 1194,
  [1195] = 1186,
  [1196] = 1193,
  [1197] = 1194,
  [1198] = 1198,
  [1199] = 1199,
  [1200] = 1200,
  [1201] = 1201,
  [1202] = 1202,
  [1203] = 1203,
  [1204] = 1020,
  [1205] = 1205,
  [1206] = 1206,
  [1207] = 1207,
  [1208] = 1208,
  [1209] = 1209,
  [1210] = 1210,
  [1211] = 1211,
  [1212] = 1212,
  [1213] = 1213,
  [1214] = 1214,
  [1215] = 1215,
  [1216] = 1211,
  [1217] = 1217,
  [1218] = 1218,
  [1219] = 1189,
  [1220] = 1183,
  [1221] = 1221,
  [1222] = 1222,
  [1223] = 1211,
  [1224] = 1224,
  [1225] = 1189,
  [1226] = 1226,
  [1227] = 1183,
  [1228] = 1228,
  [1229] = 1229,
  [1230] = 1230,
  [1231] = 1228,
  [1232] = 1228,
  [1233] = 1233,
  [1234] = 1234,
  [1235] = 1235,
  [1236] = 1236,
  [1237] = 1237,
  [1238] = 1238,
  [1239] = 1239,
  [1240] = 1240,
  [1241] = 1241,
  [1242] = 1242,
  [1243] = 1237,
  [1244] = 1242,
  [1245] = 1245,
  [1246] = 1246,
  [1247] = 1247,
  [1248] = 1234,
  [1249] = 1235,
  [1250] = 1239,
  [1251] = 1240,
  [1252] = 1252,
  [1253] = 1252,
  [1254] = 1245,
  [1255] = 1246,
  [1256] = 1234,
  [1257] = 1235,
  [1258] = 1240,
  [1259] = 1237,
  [1260] = 1246,
  [1261] = 1234,
  [1262] = 1235,
  [1263] = 1240,
  [1264] = 1237,
  [1265] = 1246,
  [1266] = 1234,
  [1267] = 1235,
  [1268] = 1240,
  [1269] = 1246,
  [1270] = 1234,
  [1271] = 1235,
  [1272] = 1240,
  [1273] = 1246,
  [1274] = 1246,
  [1275] = 1246,
  [1276] = 1276,
  [1277] = 1277,
  [1278] = 1245,
  [1279] = 1247,
  [1280] = 1246,
  [1281] = 1281,
  [1282] = 1247,
  [1283] = 1247,
  [1284] = 1252,
  [1285] = 1252,
  [1286] = 1237,
  [1287] = 1287,
  [1288] = 1288,
  [1289] = 1289,
  [1290] = 1288,
  [1291] = 1291,
  [1292] = 1292,
  [1293] = 1293,
  [1294] = 1294,
  [1295] = 1293,
  [1296] = 1296,
  [1297] = 1297,
  [1298] = 1297,
  [1299] = 1299,
  [1300] = 1300,
  [1301] = 1301,
  [1302] = 1302,
  [1303] = 1303,
  [1304] = 1304,
  [1305] = 1305,
  [1306] = 1306,
  [1307] = 1307,
  [1308] = 1307,
  [1309] = 1309,
  [1310] = 1309,
  [1311] = 1311,
  [1312] = 1305,
  [1313] = 1309,
  [1314] = 1304,
  [1315] = 1306,
  [1316] = 1311,
  [1317] = 1306,
  [1318] = 1305,
  [1319] = 1319,
  [1320] = 1320,
  [1321] = 1320,
  [1322] = 1299,
  [1323] = 1319,
  [1324] = 1320,
  [1325] = 1300,
  [1326] = 1319,
  [1327] = 1327,
  [1328] = 1327,
  [1329] = 1302,
  [1330] = 1301,
  [1331] = 1331,
  [1332] = 1303,
  [1333] = 1333,
  [1334] = 1334,
  [1335] = 1333,
  [1336] = 1333,
  [1337] = 1337,
  [1338] = 1338,
  [1339] = 1339,
  [1340] = 1340,
  [1341] = 1340,
  [1342] = 1342,
  [1343] = 1343,
  [1344] = 1344,
  [1345] = 1344,
  [1346] = 1346,
  [1347] = 1347,
  [1348] = 1348,
  [1349] = 1342,
  [1350] = 1347,
  [1351] = 1351,
  [1352] = 1352,
  [1353] = 1348,
  [1354] = 1354,
  [1355] = 1355,
  [1356] = 1356,
  [1357] = 1357,
  [1358] = 1358,
  [1359] = 1359,
  [1360] = 1360,
  [1361] = 1361,
  [1362] = 1358,
  [1363] = 1359,
  [1364] = 1364,
  [1365] = 1360,
  [1366] = 1366,
  [1367] = 1364,
  [1368] = 1368,
  [1369] = 1369,
  [1370] = 1370,
  [1371] = 1371,
  [1372] = 1372,
  [1373] = 1373,
  [1374] = 1374,
  [1375] = 1375,
  [1376] = 1376,
  [1377] = 1377,
  [1378] = 1361,
  [1379] = 1379,
  [1380] = 1380,
  [1381] = 1381,
  [1382] = 1382,
  [1383] = 1383,
  [1384] = 1384,
  [1385] = 1385,
  [1386] = 1386,
  [1387] = 830,
  [1388] = 1388,
  [1389] = 1385,
  [1390] = 1390,
  [1391] = 1391,
  [1392] = 1392,
  [1393] = 1393,
  [1394] = 1394,
  [1395] = 1395,
  [1396] = 1396,
  [1397] = 1397,
  [1398] = 1398,
  [1399] = 1399,
  [1400] = 1400,
  [1401] = 1401,
  [1402] = 1402,
  [1403] = 1403,
  [1404] = 1404,
  [1405] = 1405,
  [1406] = 1406,
  [1407] = 1407,
  [1408] = 1408,
  [1409] = 1409,
  [1410] = 1410,
  [1411] = 1411,
  [1412] = 1409,
  [1413] = 1407,
  [1414] = 1407,
  [1415] = 1408,
  [1416] = 1416,
  [1417] = 1417,
  [1418] = 1408,
  [1419] = 1409,
  [1420] = 1408,
  [1421] = 1409,
  [1422] = 1422,
  [1423] = 1407,
  [1424] = 1424,
  [1425] = 1409,
  [1426] = 1407,
  [1427] = 1408,
  [1428] = 1416,
  [1429] = 1422,
  [1430] = 1430,
  [1431] = 1417,
  [1432] = 1424,
  [1433] = 1433,
  [1434] = 1434,
  [1435] = 1430,
  [1436] = 1411,
  [1437] = 1408,
  [1438] = 1409,
  [1439] = 1407,
  [1440] = 1408,
  [1441] = 1416,
  [1442] = 1417,
  [1443] = 1407,
  [1444] = 1409,
  [1445] = 1445,
  [1446] = 1446,
  [1447] = 1447,
  [1448] = 1448,
  [1449] = 1445,
  [1450] = 1450,
  [1451] = 1451,
  [1452] = 1452,
  [1453] = 1453,
  [1454] = 1454,
  [1455] = 1455,
  [1456] = 1456,
  [1457] = 1457,
  [1458] = 1458,
  [1459] = 1459,
  [1460] = 1460,
  [1461] = 1461,
  [1462] = 1462,
  [1463] = 1463,
  [1464] = 1464,
  [1465] = 1465,
  [1466] = 1466,
  [1467] = 1467,
  [1468] = 1468,
  [1469] = 1469,
  [1470] = 1470,
  [1471] = 597,
  [1472] = 1472,
  [1473] = 1473,
  [1474] = 1474,
  [1475] = 1474,
  [1476] = 1476,
  [1477] = 1459,
  [1478] = 1478,
  [1479] = 597,
  [1480] = 1480,
  [1481] = 1481,
  [1482] = 1482,
  [1483] = 597,
  [1484] = 1484,
  [1485] = 1485,
  [1486] = 1486,
  [1487] = 1487,
  [1488] = 1488,
  [1489] = 1489,
  [1490] = 1490,
  [1491] = 1491,
  [1492] = 1492,
  [1493] = 1493,
  [1494] = 1494,
  [1495] = 1410,
  [1496] = 1496,
  [1497] = 1497,
  [1498] = 1496,
  [1499] = 1499,
  [1500] = 1500,
  [1501] = 1501,
  [1502] = 444,
  [1503] = 1503,
  [1504] = 1504,
  [1505] = 1503,
  [1506] = 1493,
  [1507] = 1500,
  [1508] = 443,
  [1509] = 1509,
  [1510] = 1499,
  [1511] = 1511,
  [1512] = 1512,
  [1513] = 1392,
  [1514] = 1514,
  [1515] = 1514,
  [1516] = 1516,
  [1517] = 1517,
  [1518] = 1518,
  [1519] = 1519,
  [1520] = 1520,
  [1521] = 1521,
  [1522] = 1522,
  [1523] = 1517,
  [1524] = 1524,
  [1525] = 1525,
  [1526] = 602,
  [1527] = 604,
  [1528] = 1528,
  [1529] = 1529,
  [1530] = 1410,
  [1531] = 1531,
  [1532] = 1532,
  [1533] = 1533,
  [1534] = 1534,
  [1535] = 1529,
  [1536] = 1536,
  [1537] = 1537,
  [1538] = 597,
  [1539] = 1539,
  [1540] = 1540,
  [1541] = 1541,
  [1542] = 1542,
  [1543] = 1543,
  [1544] = 1544,
  [1545] = 1545,
  [1546] = 1546,
  [1547] = 1547,
  [1548] = 1548,
  [1549] = 1549,
  [1550] = 1550,
  [1551] = 1551,
  [1552] = 1552,
  [1553] = 1553,
  [1554] = 1554,
  [1555] = 1555,
  [1556] = 873,
  [1557] = 1557,
  [1558] = 1517,
  [1559] = 1559,
  [1560] = 1560,
  [1561] = 1549,
  [1562] = 1552,
  [1563] = 1563,
  [1564] = 1564,
  [1565] = 445,
  [1566] = 1542,
  [1567] = 1542,
  [1568] = 1543,
  [1569] = 1543,
  [1570] = 1554,
  [1571] = 1564,
  [1572] = 1572,
  [1573] = 684,
  [1574] = 446,
  [1575] = 445,
  [1576] = 1576,
  [1577] = 1577,
  [1578] = 602,
  [1579] = 1579,
  [1580] = 604,
  [1581] = 602,
  [1582] = 1582,
  [1583] = 604,
  [1584] = 1584,
  [1585] = 1585,
  [1586] = 597,
  [1587] = 597,
  [1588] = 1588,
  [1589] = 1582,
  [1590] = 1590,
  [1591] = 1579,
  [1592] = 1554,
  [1593] = 1593,
  [1594] = 1590,
  [1595] = 1532,
  [1596] = 1596,
  [1597] = 1597,
  [1598] = 1593,
  [1599] = 1599,
  [1600] = 1600,
  [1601] = 1601,
  [1602] = 1602,
  [1603] = 1596,
  [1604] = 1604,
  [1605] = 1605,
  [1606] = 1606,
  [1607] = 1607,
  [1608] = 1608,
  [1609] = 1609,
  [1610] = 1584,
  [1611] = 1599,
  [1612] = 1600,
  [1613] = 1608,
  [1614] = 1601,
  [1615] = 1590,
  [1616] = 1579,
  [1617] = 1590,
  [1618] = 1579,
  [1619] = 1590,
  [1620] = 1579,
  [1621] = 1602,
  [1622] = 1590,
  [1623] = 1579,
  [1624] = 1590,
  [1625] = 1590,
  [1626] = 1590,
  [1627] = 1590,
  [1628] = 1590,
  [1629] = 1590,
  [1630] = 1604,
  [1631] = 446,
  [1632] = 1605,
  [1633] = 1410,
  [1634] = 1606,
  [1635] = 1635,
  [1636] = 1636,
  [1637] = 1607,
  [1638] = 1638,
  [1639] = 1639,
  [1640] = 1640,
  [1641] = 1609,
  [1642] = 1642,
  [1643] = 1643,
  [1644] = 1532,
  [1645] = 1597,
  [1646] = 1640,
  [1647] = 1647,
  [1648] = 1648,
  [1649] = 1649,
  [1650] = 1650,
  [1651] = 1651,
  [1652] = 1652,
  [1653] = 1653,
  [1654] = 1654,
  [1655] = 1655,
  [1656] = 1656,
  [1657] = 1657,
  [1658] = 1658,
  [1659] = 1659,
  [1660] = 1648,
  [1661] = 1661,
  [1662] = 1649,
  [1663] = 1663,
  [1664] = 1664,
  [1665] = 1665,
  [1666] = 1666,
  [1667] = 1667,
  [1668] = 454,
  [1669] = 1663,
  [1670] = 449,
  [1671] = 1671,
  [1672] = 1672,
  [1673] = 1673,
  [1674] = 1674,
  [1675] = 1675,
  [1676] = 1655,
  [1677] = 1466,
  [1678] = 1678,
  [1679] = 1679,
  [1680] = 1680,
  [1681] = 1681,
  [1682] = 1682,
  [1683] = 1683,
  [1684] = 1684,
  [1685] = 1685,
  [1686] = 1650,
  [1687] = 1687,
  [1688] = 1688,
  [1689] = 1661,
  [1690] = 1448,
  [1691] = 1691,
  [1692] = 1692,
  [1693] = 1450,
  [1694] = 1694,
  [1695] = 1695,
  [1696] = 1696,
  [1697] = 1697,
  [1698] = 1692,
  [1699] = 1699,
  [1700] = 1700,
  [1701] = 1701,
  [1702] = 1672,
  [1703] = 1703,
  [1704] = 1704,
  [1705] = 1705,
  [1706] = 1446,
  [1707] = 1680,
  [1708] = 1708,
  [1709] = 1709,
  [1710] = 1464,
  [1711] = 1711,
  [1712] = 1473,
  [1713] = 1713,
  [1714] = 1714,
  [1715] = 1715,
  [1716] = 450,
  [1717] = 1714,
  [1718] = 1666,
  [1719] = 1719,
  [1720] = 1465,
  [1721] = 451,
  [1722] = 1722,
  [1723] = 1723,
  [1724] = 1456,
  [1725] = 1461,
  [1726] = 1465,
  [1727] = 1727,
  [1728] = 1728,
  [1729] = 790,
  [1730] = 1730,
  [1731] = 1731,
  [1732] = 453,
  [1733] = 452,
  [1734] = 1734,
  [1735] = 1735,
  [1736] = 1736,
  [1737] = 1737,
  [1738] = 1738,
  [1739] = 1652,
  [1740] = 1740,
  [1741] = 1695,
  [1742] = 1742,
  [1743] = 1743,
  [1744] = 1697,
  [1745] = 1745,
  [1746] = 1746,
  [1747] = 1709,
  [1748] = 1748,
  [1749] = 1749,
  [1750] = 1750,
  [1751] = 1658,
  [1752] = 1752,
  [1753] = 1753,
  [1754] = 1754,
  [1755] = 1659,
  [1756] = 1464,
  [1757] = 1473,
  [1758] = 1470,
  [1759] = 1470,
  [1760] = 1760,
  [1761] = 1761,
  [1762] = 1762,
  [1763] = 1763,
  [1764] = 1734,
  [1765] = 1765,
  [1766] = 1766,
  [1767] = 1456,
  [1768] = 1458,
  [1769] = 1769,
  [1770] = 1461,
  [1771] = 1771,
  [1772] = 1673,
  [1773] = 1773,
  [1774] = 1774,
  [1775] = 1775,
  [1776] = 1776,
  [1777] = 1777,
  [1778] = 1778,
  [1779] = 1779,
  [1780] = 1780,
  [1781] = 1781,
  [1782] = 1774,
  [1783] = 1467,
  [1784] = 1784,
  [1785] = 1458,
  [1786] = 1786,
  [1787] = 1654,
  [1788] = 1788,
  [1789] = 1789,
  [1790] = 1790,
  [1791] = 1791,
  [1792] = 1777,
  [1793] = 1793,
  [1794] = 1794,
  [1795] = 1795,
  [1796] = 1777,
  [1797] = 1797,
  [1798] = 1798,
  [1799] = 1455,
  [1800] = 1800,
  [1801] = 1801,
  [1802] = 1802,
  [1803] = 1780,
  [1804] = 1468,
  [1805] = 1805,
  [1806] = 1806,
  [1807] = 1807,
  [1808] = 1777,
  [1809] = 1809,
  [1810] = 1775,
  [1811] = 1811,
  [1812] = 1812,
  [1813] = 1813,
  [1814] = 1814,
  [1815] = 1815,
  [1816] = 1816,
  [1817] = 1817,
  [1818] = 1233,
  [1819] = 1753,
  [1820] = 1777,
  [1821] = 1779,
  [1822] = 1814,
  [1823] = 1651,
  [1824] = 1824,
  [1825] = 1825,
  [1826] = 1701,
  [1827] = 1827,
  [1828] = 1828,
  [1829] = 1829,
  [1830] = 1797,
  [1831] = 1831,
  [1832] = 1825,
  [1833] = 1833,
  [1834] = 1834,
  [1835] = 1835,
  [1836] = 1836,
  [1837] = 1778,
  [1838] = 1838,
  [1839] = 1839,
  [1840] = 1840,
  [1841] = 1841,
  [1842] = 1842,
  [1843] = 1777,
  [1844] = 1844,
  [1845] = 1777,
  [1846] = 1846,
  [1847] = 1847,
  [1848] = 1848,
  [1849] = 1809,
  [1850] = 1813,
  [1851] = 700,
  [1852] = 1719,
  [1853] = 1853,
  [1854] = 1854,
  [1855] = 1833,
  [1856] = 1841,
  [1857] = 1848,
  [1858] = 1466,
  [1859] = 1859,
  [1860] = 1836,
  [1861] = 1476,
  [1862] = 1779,
  [1863] = 1453,
  [1864] = 1833,
  [1865] = 1829,
  [1866] = 1866,
  [1867] = 1853,
  [1868] = 1854,
  [1869] = 1869,
  [1870] = 1870,
  [1871] = 1871,
  [1872] = 1872,
  [1873] = 1805,
  [1874] = 1779,
  [1875] = 1828,
  [1876] = 1790,
  [1877] = 1877,
  [1878] = 1878,
  [1879] = 1879,
  [1880] = 1880,
  [1881] = 1881,
  [1882] = 542,
  [1883] = 1883,
  [1884] = 544,
  [1885] = 455,
  [1886] = 519,
  [1887] = 497,
  [1888] = 1888,
  [1889] = 1889,
  [1890] = 550,
  [1891] = 551,
  [1892] = 553,
  [1893] = 1893,
  [1894] = 1894,
  [1895] = 1895,
  [1896] = 508,
  [1897] = 1897,
  [1898] = 556,
  [1899] = 1899,
  [1900] = 1900,
  [1901] = 1901,
  [1902] = 1902,
  [1903] = 1903,
  [1904] = 563,
  [1905] = 569,
  [1906] = 1906,
  [1907] = 575,
  [1908] = 1908,
  [1909] = 456,
  [1910] = 1910,
  [1911] = 1910,
  [1912] = 1912,
  [1913] = 1913,
  [1914] = 458,
  [1915] = 1915,
  [1916] = 462,
  [1917] = 1917,
  [1918] = 463,
  [1919] = 467,
  [1920] = 1920,
  [1921] = 468,
  [1922] = 1922,
  [1923] = 1923,
  [1924] = 1924,
  [1925] = 1925,
  [1926] = 1926,
  [1927] = 1927,
  [1928] = 470,
  [1929] = 1929,
  [1930] = 1930,
  [1931] = 1931,
  [1932] = 1932,
  [1933] = 1933,
  [1934] = 1934,
  [1935] = 487,
  [1936] = 488,
  [1937] = 489,
  [1938] = 490,
  [1939] = 491,
  [1940] = 1913,
  [1941] = 492,
  [1942] = 1942,
  [1943] = 493,
  [1944] = 1944,
  [1945] = 1945,
  [1946] = 496,
  [1947] = 1878,
  [1948] = 498,
  [1949] = 1949,
  [1950] = 1950,
  [1951] = 1951,
  [1952] = 501,
  [1953] = 502,
  [1954] = 560,
  [1955] = 503,
  [1956] = 504,
  [1957] = 1957,
  [1958] = 505,
  [1959] = 1959,
  [1960] = 1960,
  [1961] = 1942,
  [1962] = 1962,
  [1963] = 516,
  [1964] = 1902,
  [1965] = 1705,
  [1966] = 1966,
  [1967] = 457,
  [1968] = 1931,
  [1969] = 1925,
  [1970] = 1924,
  [1971] = 469,
  [1972] = 1972,
  [1973] = 1973,
  [1974] = 1974,
  [1975] = 500,
  [1976] = 1976,
  [1977] = 475,
  [1978] = 485,
  [1979] = 1816,
  [1980] = 1950,
  [1981] = 518,
  [1982] = 486,
  [1983] = 1902,
  [1984] = 578,
  [1985] = 499,
  [1986] = 520,
  [1987] = 1987,
  [1988] = 1925,
  [1989] = 525,
  [1990] = 1951,
  [1991] = 1991,
  [1992] = 1992,
  [1993] = 1993,
  [1994] = 1994,
  [1995] = 1995,
  [1996] = 1996,
  [1997] = 526,
  [1998] = 1998,
  [1999] = 1902,
  [2000] = 2000,
  [2001] = 533,
  [2002] = 536,
  [2003] = 2003,
  [2004] = 1925,
  [2005] = 2005,
  [2006] = 459,
  [2007] = 2007,
  [2008] = 1776,
  [2009] = 2009,
  [2010] = 2010,
  [2011] = 515,
  [2012] = 537,
  [2013] = 2013,
  [2014] = 1902,
  [2015] = 2015,
  [2016] = 555,
  [2017] = 557,
  [2018] = 1925,
  [2019] = 559,
  [2020] = 561,
  [2021] = 540,
  [2022] = 2022,
  [2023] = 541,
  [2024] = 1487,
  [2025] = 1754,
  [2026] = 2026,
  [2027] = 543,
  [2028] = 567,
  [2029] = 2029,
  [2030] = 2030,
  [2031] = 2031,
  [2032] = 570,
  [2033] = 571,
  [2034] = 2034,
  [2035] = 548,
  [2036] = 572,
  [2037] = 573,
  [2038] = 574,
  [2039] = 2039,
  [2040] = 2040,
  [2041] = 464,
  [2042] = 2042,
  [2043] = 460,
  [2044] = 2044,
  [2045] = 484,
  [2046] = 2046,
  [2047] = 2047,
  [2048] = 2048,
  [2049] = 1957,
  [2050] = 2050,
  [2051] = 1846,
  [2052] = 2015,
  [2053] = 1925,
  [2054] = 532,
  [2055] = 2022,
  [2056] = 465,
  [2057] = 2057,
  [2058] = 2058,
  [2059] = 2047,
  [2060] = 2060,
  [2061] = 1991,
  [2062] = 2062,
  [2063] = 2063,
  [2064] = 2064,
  [2065] = 2065,
  [2066] = 2066,
  [2067] = 466,
  [2068] = 2068,
  [2069] = 2069,
  [2070] = 1915,
  [2071] = 2071,
  [2072] = 2072,
  [2073] = 2073,
  [2074] = 2074,
  [2075] = 483,
  [2076] = 2076,
  [2077] = 2077,
  [2078] = 554,
  [2079] = 558,
  [2080] = 2080,
  [2081] = 2081,
  [2082] = 1895,
  [2083] = 2083,
  [2084] = 2084,
  [2085] = 2060,
  [2086] = 1959,
  [2087] = 1897,
  [2088] = 562,
  [2089] = 1466,
  [2090] = 564,
  [2091] = 565,
  [2092] = 494,
  [2093] = 2093,
  [2094] = 2094,
  [2095] = 2095,
  [2096] = 2096,
  [2097] = 545,
  [2098] = 2029,
  [2099] = 2031,
  [2100] = 566,
  [2101] = 568,
  [2102] = 2102,
  [2103] = 2103,
  [2104] = 1906,
  [2105] = 1899,
  [2106] = 2106,
  [2107] = 1934,
  [2108] = 2108,
  [2109] = 2109,
  [2110] = 1883,
  [2111] = 495,
  [2112] = 2040,
  [2113] = 2030,
  [2114] = 2114,
  [2115] = 1917,
  [2116] = 1927,
  [2117] = 1920,
  [2118] = 1923,
  [2119] = 576,
  [2120] = 1880,
  [2121] = 577,
  [2122] = 2122,
  [2123] = 1922,
  [2124] = 1926,
  [2125] = 2125,
  [2126] = 2126,
  [2127] = 2127,
  [2128] = 2128,
  [2129] = 2129,
  [2130] = 2130,
  [2131] = 2131,
  [2132] = 2132,
  [2133] = 2133,
  [2134] = 471,
  [2135] = 2135,
  [2136] = 2063,
  [2137] = 2137,
  [2138] = 2138,
  [2139] = 2139,
  [2140] = 2064,
  [2141] = 472,
  [2142] = 523,
  [2143] = 473,
  [2144] = 2144,
  [2145] = 552,
  [2146] = 2146,
  [2147] = 2147,
  [2148] = 506,
  [2149] = 2149,
  [2150] = 2150,
  [2151] = 2151,
  [2152] = 2152,
  [2153] = 507,
  [2154] = 2154,
  [2155] = 1902,
  [2156] = 2156,
  [2157] = 474,
  [2158] = 510,
  [2159] = 1791,
  [2160] = 2160,
  [2161] = 521,
  [2162] = 2162,
  [2163] = 547,
  [2164] = 2034,
  [2165] = 1929,
  [2166] = 2166,
  [2167] = 2167,
  [2168] = 1930,
  [2169] = 549,
  [2170] = 2076,
  [2171] = 2171,
  [2172] = 514,
  [2173] = 2173,
  [2174] = 511,
  [2175] = 2175,
  [2176] = 2048,
  [2177] = 2177,
  [2178] = 527,
  [2179] = 531,
  [2180] = 538,
  [2181] = 539,
  [2182] = 1996,
  [2183] = 1933,
  [2184] = 2065,
  [2185] = 476,
  [2186] = 546,
  [2187] = 2187,
  [2188] = 509,
  [2189] = 513,
  [2190] = 2146,
  [2191] = 512,
  [2192] = 2192,
  [2193] = 2150,
  [2194] = 522,
  [2195] = 524,
  [2196] = 1888,
  [2197] = 2154,
  [2198] = 2160,
  [2199] = 529,
  [2200] = 477,
  [2201] = 2201,
  [2202] = 479,
  [2203] = 480,
  [2204] = 481,
  [2205] = 482,
  [2206] = 1289,
  [2207] = 530,
  [2208] = 1908,
  [2209] = 534,
  [2210] = 1900,
  [2211] = 1901,
  [2212] = 1903,
  [2213] = 2213,
  [2214] = 2192,
  [2215] = 535,
  [2216] = 2216,
  [2217] = 2217,
  [2218] = 2218,
  [2219] = 2219,
  [2220] = 2220,
  [2221] = 2221,
  [2222] = 2222,
  [2223] = 2223,
  [2224] = 2224,
  [2225] = 2220,
  [2226] = 2226,
  [2227] = 2227,
  [2228] = 2228,
  [2229] = 2229,
  [2230] = 2230,
  [2231] = 2231,
  [2232] = 2232,
  [2233] = 2233,
  [2234] = 2234,
  [2235] = 2235,
  [2236] = 2236,
  [2237] = 2219,
  [2238] = 2238,
  [2239] = 2239,
  [2240] = 2240,
  [2241] = 2241,
  [2242] = 2242,
  [2243] = 2243,
  [2244] = 2244,
  [2245] = 2227,
  [2246] = 2246,
  [2247] = 2247,
  [2248] = 2248,
  [2249] = 2249,
  [2250] = 2250,
  [2251] = 2219,
  [2252] = 2252,
  [2253] = 2230,
  [2254] = 2247,
  [2255] = 2227,
  [2256] = 2256,
  [2257] = 2223,
  [2258] = 2246,
  [2259] = 2259,
  [2260] = 2260,
  [2261] = 2260,
  [2262] = 2262,
  [2263] = 2263,
  [2264] = 2264,
  [2265] = 2231,
  [2266] = 2266,
  [2267] = 2267,
  [2268] = 2268,
  [2269] = 2269,
  [2270] = 2270,
  [2271] = 1974,
  [2272] = 2272,
  [2273] = 2233,
  [2274] = 2274,
  [2275] = 2275,
  [2276] = 2276,
  [2277] = 2277,
  [2278] = 2227,
  [2279] = 2250,
  [2280] = 2241,
  [2281] = 2281,
  [2282] = 2282,
  [2283] = 2283,
  [2284] = 2284,
  [2285] = 2227,
  [2286] = 2286,
  [2287] = 2287,
  [2288] = 2288,
  [2289] = 2289,
  [2290] = 2290,
  [2291] = 1674,
  [2292] = 2292,
  [2293] = 2293,
  [2294] = 2294,
  [2295] = 2295,
  [2296] = 2296,
  [2297] = 2297,
  [2298] = 2219,
  [2299] = 2299,
  [2300] = 2263,
  [2301] = 876,
  [2302] = 2302,
  [2303] = 2303,
  [2304] = 2304,
  [2305] = 2275,
  [2306] = 2306,
  [2307] = 2307,
  [2308] = 2308,
  [2309] = 2309,
  [2310] = 2310,
  [2311] = 2311,
  [2312] = 2312,
  [2313] = 2313,
  [2314] = 2314,
  [2315] = 2315,
  [2316] = 2316,
  [2317] = 2317,
  [2318] = 2302,
  [2319] = 2319,
  [2320] = 2320,
  [2321] = 2321,
  [2322] = 2322,
  [2323] = 2323,
  [2324] = 2312,
  [2325] = 2325,
  [2326] = 2326,
  [2327] = 2327,
  [2328] = 2173,
  [2329] = 2329,
  [2330] = 2330,
  [2331] = 2224,
  [2332] = 2332,
  [2333] = 2333,
  [2334] = 2334,
  [2335] = 2283,
  [2336] = 2243,
  [2337] = 2337,
  [2338] = 2338,
  [2339] = 2294,
  [2340] = 2340,
  [2341] = 2315,
  [2342] = 2342,
  [2343] = 1394,
  [2344] = 2266,
  [2345] = 2345,
  [2346] = 2322,
  [2347] = 2347,
  [2348] = 2348,
  [2349] = 2349,
  [2350] = 2350,
  [2351] = 2351,
  [2352] = 2352,
  [2353] = 2287,
  [2354] = 2306,
  [2355] = 2310,
  [2356] = 2340,
  [2357] = 2267,
  [2358] = 1727,
  [2359] = 2236,
  [2360] = 2252,
  [2361] = 2361,
  [2362] = 2362,
  [2363] = 2363,
  [2364] = 2364,
  [2365] = 2295,
  [2366] = 2362,
  [2367] = 2367,
  [2368] = 2314,
  [2369] = 2369,
  [2370] = 2317,
  [2371] = 2371,
  [2372] = 2026,
  [2373] = 2326,
  [2374] = 2374,
  [2375] = 2330,
  [2376] = 2369,
  [2377] = 2294,
  [2378] = 2378,
  [2379] = 2268,
  [2380] = 2380,
  [2381] = 2381,
  [2382] = 2382,
  [2383] = 2106,
  [2384] = 2382,
  [2385] = 2363,
  [2386] = 2386,
  [2387] = 2268,
  [2388] = 2269,
  [2389] = 2364,
  [2390] = 2390,
  [2391] = 2325,
  [2392] = 2294,
  [2393] = 2393,
  [2394] = 2394,
  [2395] = 2289,
  [2396] = 2342,
  [2397] = 2397,
  [2398] = 2290,
  [2399] = 2232,
  [2400] = 2400,
  [2401] = 2270,
  [2402] = 2268,
  [2403] = 2403,
  [2404] = 2404,
  [2405] = 2219,
  [2406] = 2406,
  [2407] = 1382,
  [2408] = 2408,
  [2409] = 2409,
  [2410] = 2219,
  [2411] = 1715,
  [2412] = 2412,
  [2413] = 2227,
  [2414] = 2414,
  [2415] = 2415,
  [2416] = 2416,
  [2417] = 2412,
  [2418] = 2276,
  [2419] = 2419,
  [2420] = 2420,
  [2421] = 2421,
  [2422] = 2422,
  [2423] = 2423,
  [2424] = 2424,
  [2425] = 2425,
  [2426] = 2426,
  [2427] = 2427,
  [2428] = 2428,
  [2429] = 2429,
  [2430] = 2430,
  [2431] = 2431,
  [2432] = 2432,
  [2433] = 2433,
  [2434] = 2434,
  [2435] = 2435,
  [2436] = 2436,
  [2437] = 2437,
  [2438] = 2438,
  [2439] = 2439,
  [2440] = 2440,
  [2441] = 2441,
  [2442] = 2431,
  [2443] = 2443,
  [2444] = 2444,
  [2445] = 2445,
  [2446] = 2446,
  [2447] = 2447,
  [2448] = 2448,
  [2449] = 2449,
  [2450] = 2419,
  [2451] = 2451,
  [2452] = 2452,
  [2453] = 2419,
  [2454] = 2454,
  [2455] = 2455,
  [2456] = 2456,
  [2457] = 2457,
  [2458] = 2458,
  [2459] = 2432,
  [2460] = 2449,
  [2461] = 2461,
  [2462] = 2462,
  [2463] = 2463,
  [2464] = 2464,
  [2465] = 2465,
  [2466] = 2466,
  [2467] = 2467,
  [2468] = 2468,
  [2469] = 2469,
  [2470] = 2470,
  [2471] = 2471,
  [2472] = 2472,
  [2473] = 2473,
  [2474] = 2474,
  [2475] = 2475,
  [2476] = 2476,
  [2477] = 2435,
  [2478] = 2478,
  [2479] = 2479,
  [2480] = 2480,
  [2481] = 2481,
  [2482] = 2468,
  [2483] = 2483,
  [2484] = 2484,
  [2485] = 2471,
  [2486] = 2486,
  [2487] = 2441,
  [2488] = 2471,
  [2489] = 2489,
  [2490] = 2447,
  [2491] = 2486,
  [2492] = 2420,
  [2493] = 2493,
  [2494] = 2419,
  [2495] = 2456,
  [2496] = 1788,
  [2497] = 2497,
  [2498] = 2498,
  [2499] = 2499,
  [2500] = 2500,
  [2501] = 2501,
  [2502] = 2447,
  [2503] = 2503,
  [2504] = 2504,
  [2505] = 2505,
  [2506] = 2428,
  [2507] = 1893,
  [2508] = 2508,
  [2509] = 2509,
  [2510] = 2510,
  [2511] = 2511,
  [2512] = 2489,
  [2513] = 2513,
  [2514] = 2463,
  [2515] = 2505,
  [2516] = 2516,
  [2517] = 2517,
  [2518] = 2518,
  [2519] = 2478,
  [2520] = 2462,
  [2521] = 2472,
  [2522] = 2473,
  [2523] = 2523,
  [2524] = 2524,
  [2525] = 2505,
  [2526] = 2526,
  [2527] = 2527,
  [2528] = 2420,
  [2529] = 2529,
  [2530] = 2430,
  [2531] = 2434,
  [2532] = 2532,
  [2533] = 2533,
  [2534] = 2534,
  [2535] = 2535,
  [2536] = 2536,
  [2537] = 2503,
  [2538] = 2447,
  [2539] = 2461,
  [2540] = 2458,
  [2541] = 2455,
  [2542] = 2475,
  [2543] = 2543,
  [2544] = 2544,
  [2545] = 2545,
  [2546] = 2546,
  [2547] = 2504,
  [2548] = 2475,
  [2549] = 2532,
  [2550] = 2544,
  [2551] = 2489,
  [2552] = 2449,
  [2553] = 2461,
  [2554] = 2462,
  [2555] = 2447,
  [2556] = 2493,
  [2557] = 2465,
  [2558] = 2478,
  [2559] = 2559,
  [2560] = 2534,
  [2561] = 2561,
  [2562] = 2486,
  [2563] = 2563,
  [2564] = 2472,
  [2565] = 2478,
  [2566] = 2446,
  [2567] = 2473,
  [2568] = 2486,
  [2569] = 2419,
  [2570] = 2462,
  [2571] = 2571,
  [2572] = 2572,
  [2573] = 2509,
  [2574] = 2574,
  [2575] = 2498,
  [2576] = 2439,
  [2577] = 2427,
  [2578] = 2465,
  [2579] = 2470,
  [2580] = 2580,
  [2581] = 2464,
  [2582] = 2582,
  [2583] = 2457,
  [2584] = 2465,
  [2585] = 2510,
  [2586] = 2508,
  [2587] = 2587,
  [2588] = 2588,
  [2589] = 2436,
  [2590] = 2504,
  [2591] = 2497,
  [2592] = 2451,
  [2593] = 2468,
  [2594] = 2527,
  [2595] = 2449,
  [2596] = 2471,
  [2597] = 2472,
  [2598] = 2473,
  [2599] = 2501,
  [2600] = 2475,
  [2601] = 2467,
  [2602] = 2419,
  [2603] = 2461,
  [2604] = 2486,
  [2605] = 2462,
  [2606] = 2606,
  [2607] = 2499,
  [2608] = 2608,
  [2609] = 2468,
  [2610] = 2452,
  [2611] = 2611,
  [2612] = 2580,
  [2613] = 2439,
  [2614] = 2614,
  [2615] = 2510,
  [2616] = 2454,
  [2617] = 2505,
  [2618] = 2505,
  [2619] = 2479,
  [2620] = 2620,
  [2621] = 2621,
  [2622] = 2622,
  [2623] = 2587,
  [2624] = 2535,
  [2625] = 2435,
  [2626] = 2425,
  [2627] = 2447,
  [2628] = 2628,
  [2629] = 2629,
  [2630] = 2630,
  [2631] = 2484,
  [2632] = 2504,
  [2633] = 2516,
  [2634] = 2509,
  [2635] = 2635,
  [2636] = 2504,
  [2637] = 2489,
  [2638] = 2425,
  [2639] = 2608,
  [2640] = 2420,
  [2641] = 2484,
  [2642] = 2175,
  [2643] = 2439,
  [2644] = 2559,
  [2645] = 2635,
  [2646] = 2478,
  [2647] = 2510,
  [2648] = 2419,
  [2649] = 2559,
  [2650] = 2484,
  [2651] = 2423,
  [2652] = 2559,
  [2653] = 2653,
};

static TSCharacterRange sym_escape_sequence_character_set_1[] = {
  {'"', '"'}, {'$', '$'}, {'0', '7'}, {'\\', '\\'}, {'`', '`'}, {'e', 'f'}, {'n', 'n'}, {'r', 'r'},
  {'t', 'v'}, {'x', 'x'},
};

static TSCharacterRange sym_name_character_set_1[] = {
  {'0', '9'}, {'A', 'Z'}, {'_', '_'}, {'a', 'z'}, {0x80, 0x9f}, {0xa1, 0x200a}, {0x200c, 0x205f}, {0x2061, 0xfefe},
  {0xff00, 0xffff},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(66);
      ADVANCE_MAP(
        '!', 124,
        '"', 172,
        '#', 253,
        '$', 176,
        '%', 213,
        '&', 73,
        '\'', 174,
        '(', 83,
        ')', 84,
        '*', 208,
        '+', 116,
        ',', 74,
        '-', 119,
        '.', 206,
        '/', 211,
        '0', 106,
        ':', 81,
        ';', 70,
        '<', 190,
        '=', 75,
        '>', 194,
        '?', 89,
        '@', 125,
        '[', 144,
        '\\', 77,
        ']', 145,
        '^', 184,
        '_', 248,
        '`', 175,
        '{', 78,
        '|', 94,
        '}', 79,
        '~', 122,
        'A', 238,
        'a', 238,
        'B', 214,
        'b', 214,
        'E', 217,
        'e', 217,
        'F', 218,
        'f', 218,
        'I', 232,
        'i', 232,
        'N', 246,
        'n', 246,
        'S', 244,
        's', 244,
        'T', 237,
        't', 237,
        'U', 230,
        'u', 230,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(108);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(63);
      if (('C' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 1:
      if (lookahead == '\n') ADVANCE(251);
      END_STATE();
    case 2:
      if (lookahead == '\n') ADVANCE(251);
      if (lookahead == '\r') ADVANCE(1);
      if (lookahead != 0 &&
          lookahead != '>') ADVANCE(252);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '\n', 173,
        '\r', 173,
        '"', 172,
        '#', 254,
        '$', 176,
        '-', 32,
        '/', 21,
        '[', 144,
        '\\', 36,
        '{', 78,
      );
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(4);
      END_STATE();
    case 4:
      ADVANCE_MAP(
        '\n', 173,
        '\r', 173,
        '#', 254,
        '$', 176,
        '-', 32,
        '/', 21,
        '[', 144,
        '\\', 37,
        '{', 78,
      );
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(4);
      END_STATE();
    case 5:
      ADVANCE_MAP(
        '!', 124,
        '"', 158,
        '#', 253,
        '$', 176,
        '%', 212,
        '&', 72,
        '\'', 160,
        '(', 85,
        ')', 86,
        '*', 209,
        '+', 115,
        ',', 74,
        '-', 118,
        '.', 206,
        '/', 210,
        '0', 106,
        ':', 80,
        ';', 70,
        '<', 191,
        '=', 31,
        '>', 195,
        '?', 92,
        '@', 125,
        '[', 144,
        '\\', 76,
        ']', 145,
        '^', 183,
        '_', 248,
        '`', 175,
        '{', 78,
        '|', 95,
        '}', 79,
        '~', 122,
        'A', 238,
        'a', 238,
        'B', 215,
        'b', 215,
        'E', 217,
        'e', 217,
        'F', 219,
        'f', 219,
        'N', 246,
        'n', 246,
        'T', 237,
        't', 237,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(108);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(5);
      if (('C' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 6:
      ADVANCE_MAP(
        '!', 123,
        '"', 158,
        '#', 253,
        '$', 176,
        '&', 71,
        '\'', 160,
        '(', 85,
        ')', 86,
        '+', 115,
        ',', 74,
        '-', 118,
        '.', 102,
        '/', 21,
        '0', 106,
        '<', 28,
        '@', 125,
        '[', 144,
        '\\', 76,
        ']', 145,
        '_', 248,
        '`', 175,
        '~', 122,
        'A', 238,
        'a', 238,
        'B', 215,
        'b', 215,
        'E', 217,
        'e', 217,
        'F', 219,
        'f', 219,
        'N', 246,
        'n', 246,
        'T', 237,
        't', 237,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(108);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(6);
      if (('C' <= lookahead && lookahead <= 'Z') ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 7:
      ADVANCE_MAP(
        '!', 123,
        '"', 158,
        '#', 253,
        '$', 176,
        '\'', 160,
        '(', 85,
        '+', 115,
        '-', 118,
        '.', 103,
        '/', 21,
        '0', 106,
        '<', 28,
        '@', 125,
        '[', 144,
        '\\', 76,
        '_', 248,
        '`', 175,
        '~', 122,
        'A', 238,
        'a', 238,
        'B', 214,
        'b', 214,
        'E', 217,
        'e', 217,
        'F', 218,
        'f', 218,
        'I', 232,
        'i', 232,
        'N', 246,
        'n', 246,
        'S', 244,
        's', 244,
        'T', 237,
        't', 237,
        'U', 230,
        'u', 230,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(108);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(7);
      if (('C' <= lookahead && lookahead <= 'Z') ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 8:
      ADVANCE_MAP(
        '!', 30,
        '"', 158,
        '#', 254,
        '$', 176,
        '%', 212,
        '&', 72,
        '(', 85,
        ')', 86,
        '*', 209,
        '+', 115,
        ',', 74,
        '-', 120,
        '.', 205,
        '/', 210,
        ':', 81,
        ';', 70,
        '<', 193,
        '=', 75,
        '>', 195,
        '?', 91,
        '[', 144,
        '\\', 36,
        ']', 145,
        '^', 183,
        '`', 175,
        '{', 78,
        '|', 95,
        '}', 79,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(9);
      if (('A' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 9:
      ADVANCE_MAP(
        '!', 30,
        '"', 158,
        '#', 254,
        '$', 176,
        '%', 212,
        '&', 72,
        '(', 85,
        ')', 86,
        '*', 209,
        '+', 115,
        ',', 74,
        '-', 120,
        '.', 205,
        '/', 210,
        ':', 81,
        ';', 70,
        '<', 193,
        '=', 75,
        '>', 195,
        '?', 91,
        '[', 144,
        '\\', 37,
        ']', 145,
        '^', 183,
        '`', 175,
        '{', 78,
        '|', 95,
        '}', 79,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(9);
      if (('A' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 10:
      ADVANCE_MAP(
        '!', 30,
        '#', 254,
        '$', 176,
        '%', 212,
        '&', 72,
        '\'', 174,
        '(', 85,
        ')', 86,
        '*', 209,
        '+', 114,
        ',', 74,
        '-', 121,
        '.', 205,
        '/', 210,
        '0', 111,
        ':', 81,
        ';', 70,
        '<', 193,
        '=', 75,
        '>', 195,
        '?', 91,
        '[', 144,
        '\\', 76,
        ']', 145,
        '^', 183,
        '{', 78,
        '|', 95,
        '}', 79,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(113);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(11);
      if (('A' <= lookahead && lookahead <= '_') ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 11:
      ADVANCE_MAP(
        '!', 30,
        '#', 254,
        '$', 176,
        '%', 212,
        '&', 72,
        '(', 85,
        ')', 86,
        '*', 209,
        '+', 114,
        ',', 74,
        '-', 121,
        '.', 205,
        '/', 210,
        '0', 111,
        ':', 81,
        ';', 70,
        '<', 193,
        '=', 75,
        '>', 195,
        '?', 91,
        '[', 144,
        '\\', 76,
        ']', 145,
        '^', 183,
        '{', 78,
        '|', 95,
        '}', 79,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(113);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(11);
      if (('A' <= lookahead && lookahead <= '_') ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 12:
      ADVANCE_MAP(
        '!', 30,
        '#', 254,
        '$', 176,
        '%', 212,
        '&', 72,
        '(', 85,
        ')', 86,
        '*', 209,
        '+', 114,
        ',', 74,
        '-', 117,
        '.', 205,
        '/', 210,
        ':', 81,
        ';', 70,
        '<', 193,
        '=', 75,
        '>', 195,
        '?', 92,
        '\\', 76,
        ']', 145,
        '^', 183,
        '{', 78,
        '|', 95,
        '}', 79,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(12);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 13:
      if (lookahead == '"') ADVANCE(157);
      if (lookahead == '\'') ADVANCE(159);
      END_STATE();
    case 14:
      ADVANCE_MAP(
        '"', 158,
        '#', 254,
        '$', 176,
        '&', 71,
        '\'', 160,
        '(', 85,
        ')', 86,
        '.', 24,
        '/', 21,
        '<', 28,
        '[', 144,
        '\\', 76,
        '|', 93,
        '}', 79,
        'A', 238,
        'a', 238,
        'B', 215,
        'b', 215,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(14);
      if (('C' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 15:
      ADVANCE_MAP(
        '"', 158,
        '#', 254,
        '\'', 160,
        '.', 103,
        '/', 21,
        '0', 106,
        '<', 28,
        '_', 57,
        'B', 13,
        'b', 13,
        'E', 40,
        'e', 40,
        'F', 41,
        'f', 41,
        'N', 51,
        'n', 51,
        'T', 48,
        't', 48,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(108);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(15);
      END_STATE();
    case 16:
      ADVANCE_MAP(
        '#', 253,
        '$', 176,
        '&', 71,
        '(', 85,
        ')', 86,
        ',', 74,
        '.', 24,
        '/', 21,
        '?', 88,
        '\\', 76,
        '}', 79,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(16);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 17:
      ADVANCE_MAP(
        '#', 254,
        '$', 176,
        '&', 71,
        '(', 83,
        ')', 84,
        '/', 21,
        ';', 70,
        '?', 88,
        '\\', 76,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(19);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 18:
      ADVANCE_MAP(
        '#', 254,
        '$', 176,
        '&', 71,
        '(', 85,
        ')', 86,
        '.', 24,
        '/', 21,
        '=', 33,
        '?', 88,
        '\\', 76,
        '{', 78,
        '|', 93,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(18);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 19:
      ADVANCE_MAP(
        '#', 254,
        '$', 176,
        '&', 71,
        '(', 85,
        '/', 21,
        ';', 70,
        '?', 88,
        '\\', 76,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(19);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 20:
      if (lookahead == '#') ADVANCE(169);
      if (lookahead == '\'') ADVANCE(160);
      if (lookahead == '/') ADVANCE(165);
      if (lookahead == '\\') ADVANCE(39);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(164);
      if (lookahead != 0) ADVANCE(170);
      END_STATE();
    case 21:
      if (lookahead == '*') ADVANCE(23);
      if (lookahead == '/') ADVANCE(252);
      END_STATE();
    case 22:
      if (lookahead == '*') ADVANCE(22);
      if (lookahead == '/') ADVANCE(251);
      if (lookahead != 0) ADVANCE(23);
      END_STATE();
    case 23:
      if (lookahead == '*') ADVANCE(22);
      if (lookahead != 0) ADVANCE(23);
      END_STATE();
    case 24:
      if (lookahead == '.') ADVANCE(26);
      END_STATE();
    case 25:
      if (lookahead == '.') ADVANCE(103);
      if (lookahead == '_') ADVANCE(57);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(40);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(25);
      END_STATE();
    case 26:
      if (lookahead == '.') ADVANCE(87);
      END_STATE();
    case 27:
      if (lookahead == '<') ADVANCE(171);
      END_STATE();
    case 28:
      if (lookahead == '<') ADVANCE(27);
      END_STATE();
    case 29:
      if (lookahead == '<') ADVANCE(27);
      if (lookahead == '?') ADVANCE(69);
      END_STATE();
    case 30:
      if (lookahead == '=') ADVANCE(186);
      END_STATE();
    case 31:
      if (lookahead == '=') ADVANCE(185);
      if (lookahead == '>') ADVANCE(82);
      END_STATE();
    case 32:
      if (lookahead == '>') ADVANCE(142);
      END_STATE();
    case 33:
      if (lookahead == '>') ADVANCE(82);
      END_STATE();
    case 34:
      if (lookahead == '>') ADVANCE(67);
      END_STATE();
    case 35:
      if (lookahead == '>') ADVANCE(143);
      END_STATE();
    case 36:
      if (lookahead == 'u') ADVANCE(156);
      if (lookahead == 'x') ADVANCE(61);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(153);
      if (set_contains(sym_escape_sequence_character_set_1, 10, lookahead)) ADVANCE(151);
      END_STATE();
    case 37:
      if (lookahead == 'u') ADVANCE(155);
      END_STATE();
    case 38:
      if (lookahead == '}') ADVANCE(151);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(38);
      END_STATE();
    case 39:
      if (lookahead == '\'' ||
          lookahead == '\\') ADVANCE(161);
      if (lookahead != 0) ADVANCE(170);
      END_STATE();
    case 40:
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(58);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(105);
      END_STATE();
    case 41:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(44);
      END_STATE();
    case 42:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(148);
      END_STATE();
    case 43:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(47);
      END_STATE();
    case 44:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(49);
      END_STATE();
    case 45:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(146);
      END_STATE();
    case 46:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(45);
      END_STATE();
    case 47:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(68);
      END_STATE();
    case 48:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(50);
      END_STATE();
    case 49:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(42);
      END_STATE();
    case 50:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(42);
      END_STATE();
    case 51:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(46);
      END_STATE();
    case 52:
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(109);
      END_STATE();
    case 53:
      if (lookahead == '8' ||
          lookahead == '9') ADVANCE(25);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(107);
      END_STATE();
    case 54:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(112);
      END_STATE();
    case 55:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(108);
      END_STATE();
    case 56:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(103);
      END_STATE();
    case 57:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(25);
      END_STATE();
    case 58:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(105);
      END_STATE();
    case 59:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(113);
      END_STATE();
    case 60:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(110);
      END_STATE();
    case 61:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(154);
      END_STATE();
    case 62:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(38);
      END_STATE();
    case 63:
      if (eof) ADVANCE(66);
      ADVANCE_MAP(
        '!', 124,
        '"', 158,
        '#', 253,
        '$', 176,
        '%', 213,
        '&', 73,
        '\'', 160,
        '(', 85,
        ')', 86,
        '*', 208,
        '+', 116,
        ',', 74,
        '-', 119,
        '.', 206,
        '/', 211,
        '0', 106,
        ':', 81,
        ';', 70,
        '<', 190,
        '=', 75,
        '>', 194,
        '?', 89,
        '@', 125,
        '[', 144,
        '\\', 77,
        ']', 145,
        '^', 184,
        '_', 248,
        '`', 175,
        '{', 78,
        '|', 94,
        '}', 79,
        '~', 122,
        'A', 238,
        'a', 238,
        'B', 214,
        'b', 214,
        'E', 217,
        'e', 217,
        'F', 218,
        'f', 218,
        'I', 232,
        'i', 232,
        'N', 246,
        'n', 246,
        'S', 244,
        's', 244,
        'T', 237,
        't', 237,
        'U', 230,
        'u', 230,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(108);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(63);
      if (('C' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 64:
      if (eof) ADVANCE(66);
      ADVANCE_MAP(
        '!', 123,
        '"', 158,
        '#', 253,
        '$', 176,
        '&', 71,
        '\'', 160,
        '(', 85,
        ')', 86,
        '+', 115,
        ',', 74,
        '-', 118,
        '.', 103,
        '/', 21,
        '0', 106,
        ':', 80,
        ';', 70,
        '<', 29,
        '=', 33,
        '?', 34,
        '@', 125,
        '[', 144,
        '\\', 76,
        ']', 145,
        '_', 248,
        '`', 175,
        '{', 78,
        '}', 79,
        '~', 122,
        'A', 238,
        'a', 238,
        'B', 215,
        'b', 215,
        'E', 217,
        'e', 217,
        'F', 219,
        'f', 219,
        'N', 246,
        'n', 246,
        'T', 237,
        't', 237,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(108);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(64);
      if (('C' <= lookahead && lookahead <= 'Z') ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 65:
      if (eof) ADVANCE(66);
      ADVANCE_MAP(
        '!', 30,
        '"', 158,
        '#', 254,
        '$', 176,
        '%', 213,
        '&', 73,
        '\'', 160,
        '(', 85,
        ')', 86,
        '*', 208,
        '+', 116,
        ',', 74,
        '-', 119,
        '.', 207,
        '/', 211,
        '0', 111,
        ':', 81,
        ';', 70,
        '<', 192,
        '=', 75,
        '>', 194,
        '?', 90,
        '[', 144,
        '\\', 76,
        ']', 145,
        '^', 184,
        '{', 78,
        '|', 94,
        '}', 79,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(113);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(65);
      if (('A' <= lookahead && lookahead <= '_') ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 66:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 67:
      ACCEPT_TOKEN(anon_sym_QMARK_GT);
      END_STATE();
    case 68:
      ACCEPT_TOKEN(sym_php_tag);
      END_STATE();
    case 69:
      ACCEPT_TOKEN(sym_php_tag);
      if (lookahead == '=') ADVANCE(68);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(43);
      END_STATE();
    case 70:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 71:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 72:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(182);
      END_STATE();
    case 73:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(182);
      if (lookahead == '=') ADVANCE(138);
      END_STATE();
    case 74:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 75:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(185);
      if (lookahead == '>') ADVANCE(82);
      END_STATE();
    case 76:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      END_STATE();
    case 77:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      if (lookahead == '\'' ||
          lookahead == '\\') ADVANCE(161);
      END_STATE();
    case 78:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 79:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 80:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(126);
      END_STATE();
    case 82:
      ACCEPT_TOKEN(anon_sym_EQ_GT);
      END_STATE();
    case 83:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 84:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 85:
      ACCEPT_TOKEN(anon_sym_LPAREN2);
      END_STATE();
    case 86:
      ACCEPT_TOKEN(anon_sym_RPAREN2);
      END_STATE();
    case 87:
      ACCEPT_TOKEN(anon_sym_DOT_DOT_DOT);
      END_STATE();
    case 88:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 89:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '-') ADVANCE(35);
      if (lookahead == '>') ADVANCE(67);
      if (lookahead == '?') ADVANCE(178);
      END_STATE();
    case 90:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '-') ADVANCE(35);
      if (lookahead == '?') ADVANCE(178);
      END_STATE();
    case 91:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '-') ADVANCE(35);
      if (lookahead == '?') ADVANCE(177);
      END_STATE();
    case 92:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '?') ADVANCE(177);
      END_STATE();
    case 93:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 94:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '=') ADVANCE(140);
      if (lookahead == '|') ADVANCE(181);
      END_STATE();
    case 95:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '|') ADVANCE(181);
      END_STATE();
    case 96:
      ACCEPT_TOKEN(aux_sym_cast_type_token1);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 97:
      ACCEPT_TOKEN(aux_sym_cast_type_token3);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 98:
      ACCEPT_TOKEN(aux_sym_cast_type_token6);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 99:
      ACCEPT_TOKEN(aux_sym_cast_type_token8);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 100:
      ACCEPT_TOKEN(aux_sym_cast_type_token11);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 101:
      ACCEPT_TOKEN(aux_sym_cast_type_token12);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 102:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '.') ADVANCE(26);
      if (lookahead == '_') ADVANCE(56);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(40);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(103);
      END_STATE();
    case 103:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '_') ADVANCE(56);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(40);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(103);
      END_STATE();
    case 104:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '_') ADVANCE(249);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(104);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0x9f) ||
          (0xa1 <= lookahead && lookahead <= 0x200a) ||
          (0x200c <= lookahead && lookahead <= 0x205f) ||
          (0x2061 <= lookahead && lookahead <= 0xfefe) ||
          (0xff00 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 105:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == '_') ADVANCE(58);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(105);
      END_STATE();
    case 106:
      ACCEPT_TOKEN(sym_integer);
      ADVANCE_MAP(
        '.', 103,
        '_', 53,
        'B', 52,
        'b', 52,
        'E', 40,
        'e', 40,
        'O', 112,
        'o', 112,
        'X', 60,
        'x', 60,
        '8', 25,
        '9', 25,
      );
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(107);
      END_STATE();
    case 107:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '.') ADVANCE(103);
      if (lookahead == '_') ADVANCE(53);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(40);
      if (lookahead == '8' ||
          lookahead == '9') ADVANCE(25);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(107);
      END_STATE();
    case 108:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '.') ADVANCE(103);
      if (lookahead == '_') ADVANCE(55);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(40);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(108);
      END_STATE();
    case 109:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(52);
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(109);
      END_STATE();
    case 110:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(60);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(110);
      END_STATE();
    case 111:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(54);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(52);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(112);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(60);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(112);
      END_STATE();
    case 112:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(54);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(112);
      END_STATE();
    case 113:
      ACCEPT_TOKEN(sym_integer);
      if (lookahead == '_') ADVANCE(59);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(113);
      END_STATE();
    case 114:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 115:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(128);
      END_STATE();
    case 116:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(128);
      if (lookahead == '=') ADVANCE(133);
      END_STATE();
    case 117:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 118:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(127);
      END_STATE();
    case 119:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(127);
      if (lookahead == '=') ADVANCE(134);
      if (lookahead == '>') ADVANCE(142);
      END_STATE();
    case 120:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(127);
      if (lookahead == '>') ADVANCE(142);
      END_STATE();
    case 121:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '>') ADVANCE(142);
      END_STATE();
    case 122:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 123:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 124:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(186);
      END_STATE();
    case 125:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 126:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 127:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      END_STATE();
    case 128:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      END_STATE();
    case 129:
      ACCEPT_TOKEN(anon_sym_STAR_STAR_EQ);
      END_STATE();
    case 130:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 131:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 132:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 134:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      END_STATE();
    case 135:
      ACCEPT_TOKEN(anon_sym_DOT_EQ);
      END_STATE();
    case 136:
      ACCEPT_TOKEN(anon_sym_LT_LT_EQ);
      END_STATE();
    case 137:
      ACCEPT_TOKEN(anon_sym_GT_GT_EQ);
      END_STATE();
    case 138:
      ACCEPT_TOKEN(anon_sym_AMP_EQ);
      END_STATE();
    case 139:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(anon_sym_PIPE_EQ);
      END_STATE();
    case 141:
      ACCEPT_TOKEN(anon_sym_QMARK_QMARK_EQ);
      END_STATE();
    case 142:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 143:
      ACCEPT_TOKEN(anon_sym_QMARK_DASH_GT);
      END_STATE();
    case 144:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(aux_sym__argument_name_token1);
      END_STATE();
    case 147:
      ACCEPT_TOKEN(aux_sym__argument_name_token1);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 148:
      ACCEPT_TOKEN(aux_sym__argument_name_token2);
      END_STATE();
    case 149:
      ACCEPT_TOKEN(aux_sym__argument_name_token2);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 150:
      ACCEPT_TOKEN(anon_sym_POUND_LBRACK);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(sym_escape_sequence);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(151);
      END_STATE();
    case 153:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(152);
      END_STATE();
    case 154:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(151);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(anon_sym_BSLASHu);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(anon_sym_BSLASHu);
      if (lookahead == '{') ADVANCE(62);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(aux_sym_encapsed_string_token1);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(aux_sym_string_token1);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(aux_sym_string_token2);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '\n') ADVANCE(170);
      if (lookahead == '\r') ADVANCE(163);
      if (lookahead == '>') ADVANCE(170);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(168);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '\n') ADVANCE(170);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(170);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '#') ADVANCE(169);
      if (lookahead == '/') ADVANCE(165);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) ADVANCE(164);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(170);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '*') ADVANCE(167);
      if (lookahead == '/') ADVANCE(168);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(170);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '*') ADVANCE(166);
      if (lookahead == '/') ADVANCE(170);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(167);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '*') ADVANCE(166);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(167);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '?') ADVANCE(162);
      if (lookahead == '\n' ||
          lookahead == '\r') ADVANCE(170);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(168);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == '?' ||
          lookahead == '[') ADVANCE(170);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '[' &&
          lookahead != '\\') ADVANCE(168);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(170);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(anon_sym_LT_LT_LT);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_DQUOTE2);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(sym__new_line);
      if (lookahead == '\n') ADVANCE(173);
      if (lookahead == '\r') ADVANCE(173);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(anon_sym_SQUOTE2);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_BQUOTE);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(anon_sym_QMARK_QMARK);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_QMARK_QMARK);
      if (lookahead == '=') ADVANCE(141);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      if (lookahead == '=') ADVANCE(129);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(139);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      if (lookahead == '=') ADVANCE(188);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      if (lookahead == '=') ADVANCE(189);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_LT_GT);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_EQ_EQ_EQ);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_BANG_EQ_EQ);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(201);
      if (lookahead == '=') ADVANCE(196);
      if (lookahead == '>') ADVANCE(187);
      if (lookahead == '?') ADVANCE(69);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(200);
      if (lookahead == '=') ADVANCE(196);
      if (lookahead == '>') ADVANCE(187);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(202);
      if (lookahead == '=') ADVANCE(196);
      if (lookahead == '>') ADVANCE(187);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(199);
      if (lookahead == '=') ADVANCE(196);
      if (lookahead == '>') ADVANCE(187);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(197);
      if (lookahead == '>') ADVANCE(204);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(197);
      if (lookahead == '>') ADVANCE(203);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      if (lookahead == '>') ADVANCE(198);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_LT_EQ_GT);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '<') ADVANCE(171);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '<') ADVANCE(171);
      if (lookahead == '=') ADVANCE(136);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '=') ADVANCE(136);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '=') ADVANCE(137);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(26);
      if (lookahead == '_') ADVANCE(56);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(40);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(103);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '=') ADVANCE(135);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(180);
      if (lookahead == '=') ADVANCE(130);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(179);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(23);
      if (lookahead == '/') ADVANCE(252);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(23);
      if (lookahead == '/') ADVANCE(252);
      if (lookahead == '=') ADVANCE(131);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(132);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == '"') ADVANCE(157);
      if (lookahead == '\'') ADVANCE(159);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(234);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == '"') ADVANCE(157);
      if (lookahead == '\'') ADVANCE(159);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == '.') ADVANCE(103);
      if (lookahead == '_') ADVANCE(248);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(217);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(216);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0x80 <= lookahead && lookahead <= 0x9f) ||
          (0xa1 <= lookahead && lookahead <= 0x200a) ||
          (0x200c <= lookahead && lookahead <= 0x205f) ||
          (0x2061 <= lookahead && lookahead <= 0xfefe) ||
          (0xff00 <= lookahead && lookahead <= 0xffff)) ADVANCE(250);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(58);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(104);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(229);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(233);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(229);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(247);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(242);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(149);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(243);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(100);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(231);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(97);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(147);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(227);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(239);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(240);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(224);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(241);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(221);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(226);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(220);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(225);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(245);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(235);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(222);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(223);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(98);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(99);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(101);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(236);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(222);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(228);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(sym_name);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(96);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(sym_name);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(216);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(sym_name);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(104);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(sym_name);
      if (set_contains(sym_name_character_set_1, 9, lookahead)) ADVANCE(250);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(sym_comment);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '?') ADVANCE(2);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(252);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '[') ADVANCE(150);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '?') ADVANCE(252);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '?' &&
          lookahead != '[') ADVANCE(252);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      ADVANCE_MAP(
        'A', 1,
        'B', 2,
        'E', 3,
        'F', 4,
        'I', 5,
        'M', 6,
        'N', 7,
        'P', 8,
        'S', 9,
        'T', 10,
        'U', 11,
        'V', 12,
        'a', 13,
        'b', 14,
        'e', 15,
        'f', 16,
        'i', 17,
        'm', 18,
        'n', 19,
        'p', 20,
        's', 21,
        't', 22,
        'u', 23,
        'v', 24,
        'C', 25,
        'c', 25,
        'D', 26,
        'd', 26,
        'G', 27,
        'g', 27,
        'L', 28,
        'l', 28,
        'O', 29,
        'o', 29,
        'R', 30,
        'r', 30,
        'W', 31,
        'w', 31,
        'X', 32,
        'x', 32,
        'Y', 33,
        'y', 33,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0x2060 ||
          lookahead == 0xfeff) SKIP(0);
      END_STATE();
    case 1:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(34);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(35);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(36);
      END_STATE();
    case 2:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(37);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(38);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(39);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        'C', 40,
        'c', 40,
        'L', 41,
        'l', 41,
        'N', 42,
        'n', 42,
        'X', 43,
        'x', 43,
      );
      END_STATE();
    case 4:
      ADVANCE_MAP(
        'I', 44,
        'i', 44,
        'N', 45,
        'n', 45,
        'O', 46,
        'o', 46,
        'R', 47,
        'r', 47,
        'U', 48,
        'u', 48,
      );
      END_STATE();
    case 5:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(49);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(50);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(51);
      END_STATE();
    case 6:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(52);
      END_STATE();
    case 7:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(53);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(54);
      END_STATE();
    case 8:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(55);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(56);
      END_STATE();
    case 9:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(57);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(58);
      END_STATE();
    case 10:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(59);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(60);
      END_STATE();
    case 11:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(61);
      END_STATE();
    case 12:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(62);
      END_STATE();
    case 13:
      if (lookahead == 'r') ADVANCE(63);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(34);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(35);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(36);
      END_STATE();
    case 14:
      if (lookahead == 'O') ADVANCE(38);
      if (lookahead == 'o') ADVANCE(64);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(37);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(39);
      END_STATE();
    case 15:
      ADVANCE_MAP(
        'N', 42,
        'n', 65,
        'C', 40,
        'c', 40,
        'L', 41,
        'l', 41,
        'X', 43,
        'x', 43,
      );
      END_STATE();
    case 16:
      ADVANCE_MAP(
        'a', 66,
        'l', 67,
        'I', 44,
        'i', 44,
        'N', 45,
        'n', 45,
        'O', 46,
        'o', 46,
        'R', 47,
        'r', 47,
        'U', 48,
        'u', 48,
      );
      END_STATE();
    case 17:
      if (lookahead == 'N') ADVANCE(51);
      if (lookahead == 'n') ADVANCE(68);
      if (lookahead == 't') ADVANCE(69);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(49);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(50);
      END_STATE();
    case 18:
      if (lookahead == 'i') ADVANCE(70);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(52);
      END_STATE();
    case 19:
      if (lookahead == 'E') ADVANCE(54);
      if (lookahead == 'e') ADVANCE(71);
      if (lookahead == 'u') ADVANCE(72);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(53);
      END_STATE();
    case 20:
      if (lookahead == 'a') ADVANCE(73);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(55);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(56);
      END_STATE();
    case 21:
      if (lookahead == 'T') ADVANCE(57);
      if (lookahead == 'e') ADVANCE(74);
      if (lookahead == 't') ADVANCE(75);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(58);
      END_STATE();
    case 22:
      if (lookahead == 'R') ADVANCE(60);
      if (lookahead == 'i') ADVANCE(76);
      if (lookahead == 'r') ADVANCE(77);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(59);
      END_STATE();
    case 23:
      if (lookahead == 'n') ADVANCE(78);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(61);
      END_STATE();
    case 24:
      if (lookahead == 'o') ADVANCE(79);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(62);
      END_STATE();
    case 25:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(80);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(81);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(82);
      END_STATE();
    case 26:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(83);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(84);
      END_STATE();
    case 27:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(85);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(86);
      END_STATE();
    case 28:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(87);
      END_STATE();
    case 29:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(88);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(89);
      END_STATE();
    case 30:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(90);
      END_STATE();
    case 31:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(91);
      END_STATE();
    case 32:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(92);
      END_STATE();
    case 33:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(93);
      END_STATE();
    case 34:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(94);
      END_STATE();
    case 35:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(95);
      END_STATE();
    case 36:
      ACCEPT_TOKEN(aux_sym_namespace_use_clause_token1);
      END_STATE();
    case 37:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(96);
      END_STATE();
    case 38:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(97);
      END_STATE();
    case 39:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(98);
      END_STATE();
    case 40:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(99);
      END_STATE();
    case 41:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(100);
      END_STATE();
    case 42:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(101);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(102);
      END_STATE();
    case 43:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(103);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(104);
      END_STATE();
    case 44:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(105);
      END_STATE();
    case 45:
      ACCEPT_TOKEN(aux_sym__arrow_function_header_token1);
      END_STATE();
    case 46:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(106);
      END_STATE();
    case 47:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(107);
      END_STATE();
    case 48:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(108);
      END_STATE();
    case 49:
      ACCEPT_TOKEN(aux_sym_if_statement_token1);
      END_STATE();
    case 50:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(109);
      END_STATE();
    case 51:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(110);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(111);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(112);
      END_STATE();
    case 52:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(113);
      END_STATE();
    case 53:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(114);
      END_STATE();
    case 54:
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(115);
      END_STATE();
    case 55:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(116);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(117);
      END_STATE();
    case 56:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(118);
      END_STATE();
    case 57:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(119);
      END_STATE();
    case 58:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(120);
      END_STATE();
    case 59:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(121);
      END_STATE();
    case 60:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(122);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(123);
      END_STATE();
    case 61:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(124);
      END_STATE();
    case 62:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(125);
      END_STATE();
    case 63:
      if (lookahead == 'r') ADVANCE(126);
      END_STATE();
    case 64:
      if (lookahead == 'O') ADVANCE(97);
      if (lookahead == 'o') ADVANCE(127);
      END_STATE();
    case 65:
      if (lookahead == 'c') ADVANCE(128);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(101);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(102);
      END_STATE();
    case 66:
      if (lookahead == 'l') ADVANCE(129);
      END_STATE();
    case 67:
      if (lookahead == 'o') ADVANCE(130);
      END_STATE();
    case 68:
      if (lookahead == 'T') ADVANCE(112);
      if (lookahead == 't') ADVANCE(131);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(110);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(111);
      END_STATE();
    case 69:
      if (lookahead == 'e') ADVANCE(132);
      END_STATE();
    case 70:
      if (lookahead == 'x') ADVANCE(133);
      END_STATE();
    case 71:
      if (lookahead == 'v') ADVANCE(134);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(115);
      END_STATE();
    case 72:
      if (lookahead == 'l') ADVANCE(135);
      END_STATE();
    case 73:
      if (lookahead == 'r') ADVANCE(136);
      END_STATE();
    case 74:
      if (lookahead == 'l') ADVANCE(137);
      END_STATE();
    case 75:
      if (lookahead == 'r') ADVANCE(138);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(119);
      END_STATE();
    case 76:
      if (lookahead == 'c') ADVANCE(139);
      END_STATE();
    case 77:
      if (lookahead == 'u') ADVANCE(140);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(122);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(123);
      END_STATE();
    case 78:
      if (lookahead == 's') ADVANCE(141);
      END_STATE();
    case 79:
      if (lookahead == 'i') ADVANCE(142);
      END_STATE();
    case 80:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(143);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(144);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(145);
      END_STATE();
    case 81:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(146);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(147);
      END_STATE();
    case 82:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(148);
      END_STATE();
    case 83:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(149);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(150);
      END_STATE();
    case 84:
      ACCEPT_TOKEN(aux_sym_do_statement_token1);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(151);
      END_STATE();
    case 85:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(152);
      END_STATE();
    case 86:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(153);
      END_STATE();
    case 87:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(154);
      END_STATE();
    case 88:
      if (lookahead == 'J' ||
          lookahead == 'j') ADVANCE(155);
      END_STATE();
    case 89:
      ACCEPT_TOKEN(aux_sym_binary_expression_token3);
      END_STATE();
    case 90:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(156);
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(157);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(158);
      END_STATE();
    case 91:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(159);
      END_STATE();
    case 92:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(160);
      END_STATE();
    case 93:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(161);
      END_STATE();
    case 94:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(162);
      END_STATE();
    case 95:
      ACCEPT_TOKEN(aux_sym_binary_expression_token2);
      END_STATE();
    case 96:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(163);
      END_STATE();
    case 97:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(164);
      END_STATE();
    case 98:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(165);
      END_STATE();
    case 99:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(166);
      END_STATE();
    case 100:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(167);
      END_STATE();
    case 101:
      ADVANCE_MAP(
        'D', 168,
        'd', 168,
        'F', 169,
        'f', 169,
        'I', 170,
        'i', 170,
        'S', 171,
        's', 171,
        'W', 172,
        'w', 172,
      );
      END_STATE();
    case 102:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(173);
      END_STATE();
    case 103:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(174);
      END_STATE();
    case 104:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(175);
      END_STATE();
    case 105:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(176);
      END_STATE();
    case 106:
      ACCEPT_TOKEN(aux_sym_for_statement_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(177);
      END_STATE();
    case 107:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(178);
      END_STATE();
    case 108:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(179);
      END_STATE();
    case 109:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(180);
      END_STATE();
    case 110:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(181);
      END_STATE();
    case 111:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(182);
      END_STATE();
    case 112:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(183);
      END_STATE();
    case 113:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(184);
      END_STATE();
    case 114:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(185);
      END_STATE();
    case 115:
      ACCEPT_TOKEN(aux_sym__new_non_dereferencable_expression_token1);
      END_STATE();
    case 116:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(186);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(187);
      END_STATE();
    case 117:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(188);
      END_STATE();
    case 118:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(189);
      END_STATE();
    case 119:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(190);
      END_STATE();
    case 120:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(191);
      END_STATE();
    case 121:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(192);
      END_STATE();
    case 122:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(193);
      END_STATE();
    case 123:
      ACCEPT_TOKEN(aux_sym_try_statement_token1);
      END_STATE();
    case 124:
      ACCEPT_TOKEN(aux_sym_namespace_use_declaration_token1);
      END_STATE();
    case 125:
      ACCEPT_TOKEN(sym_var_modifier);
      END_STATE();
    case 126:
      if (lookahead == 'a') ADVANCE(194);
      END_STATE();
    case 127:
      if (lookahead == 'L') ADVANCE(164);
      if (lookahead == 'l') ADVANCE(195);
      END_STATE();
    case 128:
      if (lookahead == 'o') ADVANCE(196);
      END_STATE();
    case 129:
      if (lookahead == 's') ADVANCE(197);
      END_STATE();
    case 130:
      if (lookahead == 'a') ADVANCE(198);
      END_STATE();
    case 131:
      ACCEPT_TOKEN(anon_sym_int);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(183);
      END_STATE();
    case 132:
      if (lookahead == 'r') ADVANCE(199);
      END_STATE();
    case 133:
      if (lookahead == 'e') ADVANCE(200);
      END_STATE();
    case 134:
      if (lookahead == 'e') ADVANCE(201);
      END_STATE();
    case 135:
      if (lookahead == 'l') ADVANCE(202);
      END_STATE();
    case 136:
      if (lookahead == 'e') ADVANCE(203);
      END_STATE();
    case 137:
      if (lookahead == 'f') ADVANCE(204);
      END_STATE();
    case 138:
      if (lookahead == 'i') ADVANCE(205);
      END_STATE();
    case 139:
      if (lookahead == 'k') ADVANCE(206);
      END_STATE();
    case 140:
      if (lookahead == 'e') ADVANCE(207);
      END_STATE();
    case 141:
      if (lookahead == 'e') ADVANCE(208);
      END_STATE();
    case 142:
      if (lookahead == 'd') ADVANCE(209);
      END_STATE();
    case 143:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(210);
      END_STATE();
    case 144:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(211);
      END_STATE();
    case 145:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(212);
      END_STATE();
    case 146:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(213);
      END_STATE();
    case 147:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(214);
      END_STATE();
    case 148:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(215);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(216);
      END_STATE();
    case 149:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(217);
      END_STATE();
    case 150:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(218);
      END_STATE();
    case 151:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(219);
      END_STATE();
    case 152:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(220);
      END_STATE();
    case 153:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(221);
      END_STATE();
    case 154:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(222);
      END_STATE();
    case 155:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(223);
      END_STATE();
    case 156:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(224);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(225);
      END_STATE();
    case 157:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(226);
      END_STATE();
    case 158:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(227);
      END_STATE();
    case 159:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(228);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(aux_sym_binary_expression_token4);
      END_STATE();
    case 161:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(229);
      END_STATE();
    case 162:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(230);
      END_STATE();
    case 163:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(231);
      END_STATE();
    case 164:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(232);
      END_STATE();
    case 165:
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(233);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(aux_sym_echo_statement_token1);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(aux_sym_else_clause_token1);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(234);
      END_STATE();
    case 168:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(235);
      END_STATE();
    case 169:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(236);
      END_STATE();
    case 170:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(237);
      END_STATE();
    case 171:
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(238);
      END_STATE();
    case 172:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(239);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(aux_sym_enum_declaration_token1);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(aux_sym_exit_statement_token1);
      END_STATE();
    case 175:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(240);
      END_STATE();
    case 176:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(241);
      END_STATE();
    case 177:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(242);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(aux_sym_yield_expression_token2);
      END_STATE();
    case 179:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(243);
      END_STATE();
    case 180:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(244);
      END_STATE();
    case 181:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(245);
      END_STATE();
    case 182:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(246);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(247);
      END_STATE();
    case 183:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(248);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(249);
      END_STATE();
    case 184:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(250);
      END_STATE();
    case 185:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(251);
      END_STATE();
    case 186:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(252);
      END_STATE();
    case 187:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(253);
      END_STATE();
    case 188:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(254);
      END_STATE();
    case 189:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(255);
      END_STATE();
    case 190:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(256);
      END_STATE();
    case 191:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(257);
      END_STATE();
    case 192:
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(258);
      END_STATE();
    case 193:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(259);
      END_STATE();
    case 194:
      if (lookahead == 'y') ADVANCE(260);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_bool);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(232);
      END_STATE();
    case 196:
      if (lookahead == 'd') ADVANCE(261);
      END_STATE();
    case 197:
      if (lookahead == 'e') ADVANCE(262);
      END_STATE();
    case 198:
      if (lookahead == 't') ADVANCE(263);
      END_STATE();
    case 199:
      if (lookahead == 'a') ADVANCE(264);
      END_STATE();
    case 200:
      if (lookahead == 'd') ADVANCE(265);
      END_STATE();
    case 201:
      if (lookahead == 'r') ADVANCE(266);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_null);
      END_STATE();
    case 203:
      if (lookahead == 'n') ADVANCE(267);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_self);
      END_STATE();
    case 205:
      if (lookahead == 'c') ADVANCE(268);
      if (lookahead == 'n') ADVANCE(269);
      END_STATE();
    case 206:
      if (lookahead == 's') ADVANCE(270);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_true);
      END_STATE();
    case 208:
      if (lookahead == 't') ADVANCE(271);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_void);
      END_STATE();
    case 210:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(272);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(aux_sym_enum_case_token1);
      END_STATE();
    case 212:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(273);
      END_STATE();
    case 213:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(274);
      END_STATE();
    case 214:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(275);
      END_STATE();
    case 215:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(276);
      END_STATE();
    case 216:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(277);
      END_STATE();
    case 217:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(278);
      END_STATE();
    case 218:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(279);
      END_STATE();
    case 219:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(280);
      END_STATE();
    case 220:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(281);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(aux_sym_goto_statement_token1);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(aux_sym__list_destructing_token1);
      END_STATE();
    case 223:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(282);
      END_STATE();
    case 224:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(283);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(aux_sym_cast_type_token10);
      END_STATE();
    case 226:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(284);
      END_STATE();
    case 227:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(285);
      END_STATE();
    case 228:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(286);
      END_STATE();
    case 229:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(287);
      END_STATE();
    case 230:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(288);
      END_STATE();
    case 231:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(289);
      END_STATE();
    case 232:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(290);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(aux_sym_break_statement_token1);
      END_STATE();
    case 234:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(291);
      END_STATE();
    case 235:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(292);
      END_STATE();
    case 236:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(293);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(aux_sym_if_statement_token2);
      END_STATE();
    case 238:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(294);
      END_STATE();
    case 239:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(295);
      END_STATE();
    case 240:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(296);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(aux_sym_final_modifier_token1);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(297);
      END_STATE();
    case 242:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(298);
      END_STATE();
    case 243:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(299);
      END_STATE();
    case 244:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(300);
      END_STATE();
    case 245:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(301);
      END_STATE();
    case 246:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(302);
      END_STATE();
    case 247:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(303);
      END_STATE();
    case 248:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(304);
      END_STATE();
    case 249:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(305);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(aux_sym_match_expression_token1);
      END_STATE();
    case 251:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(306);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(aux_sym_print_intrinsic_token1);
      END_STATE();
    case 253:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(307);
      END_STATE();
    case 254:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(308);
      END_STATE();
    case 255:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(309);
      END_STATE();
    case 256:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(310);
      END_STATE();
    case 257:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(311);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(aux_sym_throw_expression_token1);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(aux_sym_trait_declaration_token1);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(anon_sym_array);
      END_STATE();
    case 261:
      if (lookahead == 'i') ADVANCE(312);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(anon_sym_false);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(anon_sym_float);
      END_STATE();
    case 264:
      if (lookahead == 'b') ADVANCE(313);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(anon_sym_mixed);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(sym_bottom_type);
      END_STATE();
    case 267:
      if (lookahead == 't') ADVANCE(314);
      END_STATE();
    case 268:
      if (lookahead == 't') ADVANCE(315);
      END_STATE();
    case 269:
      if (lookahead == 'g') ADVANCE(316);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(anon_sym_ticks);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(anon_sym_unset);
      END_STATE();
    case 272:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(317);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(aux_sym_catch_clause_token1);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(aux_sym_class_declaration_token1);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(aux_sym_clone_expression_token1);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(aux_sym__namespace_use_type_token2);
      END_STATE();
    case 277:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(318);
      END_STATE();
    case 278:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(319);
      END_STATE();
    case 279:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(320);
      END_STATE();
    case 280:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(321);
      END_STATE();
    case 281:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(322);
      END_STATE();
    case 282:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(323);
      END_STATE();
    case 283:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(324);
      END_STATE();
    case 284:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(325);
      END_STATE();
    case 285:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(326);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(aux_sym_while_statement_token1);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(aux_sym_yield_expression_token1);
      END_STATE();
    case 288:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(327);
      END_STATE();
    case 289:
      ACCEPT_TOKEN(aux_sym_cast_type_token2);
      END_STATE();
    case 290:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(328);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(aux_sym_else_if_clause_token1);
      END_STATE();
    case 292:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(329);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(aux_sym_for_statement_token2);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(330);
      END_STATE();
    case 294:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(331);
      END_STATE();
    case 295:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(332);
      END_STATE();
    case 296:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(333);
      END_STATE();
    case 297:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(334);
      END_STATE();
    case 298:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(335);
      END_STATE();
    case 299:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(336);
      END_STATE();
    case 300:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(337);
      END_STATE();
    case 301:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(338);
      END_STATE();
    case 302:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(339);
      END_STATE();
    case 303:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(340);
      END_STATE();
    case 304:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(341);
      END_STATE();
    case 305:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(342);
      END_STATE();
    case 306:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(343);
      END_STATE();
    case 307:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(344);
      END_STATE();
    case 308:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(345);
      END_STATE();
    case 309:
      ACCEPT_TOKEN(aux_sym_visibility_modifier_token1);
      END_STATE();
    case 310:
      ACCEPT_TOKEN(aux_sym_function_static_declaration_token1);
      END_STATE();
    case 311:
      ACCEPT_TOKEN(aux_sym_switch_statement_token1);
      END_STATE();
    case 312:
      if (lookahead == 'n') ADVANCE(346);
      END_STATE();
    case 313:
      if (lookahead == 'l') ADVANCE(347);
      END_STATE();
    case 314:
      ACCEPT_TOKEN(anon_sym_parent);
      END_STATE();
    case 315:
      if (lookahead == '_') ADVANCE(348);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(anon_sym_string);
      END_STATE();
    case 317:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(349);
      END_STATE();
    case 318:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(350);
      END_STATE();
    case 319:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(351);
      END_STATE();
    case 320:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(352);
      END_STATE();
    case 321:
      ACCEPT_TOKEN(aux_sym_cast_type_token5);
      END_STATE();
    case 322:
      ACCEPT_TOKEN(aux_sym_global_declaration_token1);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(aux_sym_cast_type_token9);
      END_STATE();
    case 324:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(353);
      END_STATE();
    case 325:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(354);
      END_STATE();
    case 326:
      ACCEPT_TOKEN(aux_sym_return_statement_token1);
      END_STATE();
    case 327:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(355);
      END_STATE();
    case 328:
      ACCEPT_TOKEN(aux_sym_cast_type_token4);
      END_STATE();
    case 329:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(356);
      END_STATE();
    case 330:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(357);
      END_STATE();
    case 331:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(358);
      END_STATE();
    case 332:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(359);
      END_STATE();
    case 333:
      ACCEPT_TOKEN(aux_sym_base_clause_token1);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(aux_sym_finally_clause_token1);
      END_STATE();
    case 335:
      ACCEPT_TOKEN(aux_sym_foreach_statement_token1);
      END_STATE();
    case 336:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(360);
      END_STATE();
    case 337:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(361);
      END_STATE();
    case 338:
      ACCEPT_TOKEN(aux_sym_include_expression_token1);
      if (lookahead == '_') ADVANCE(362);
      END_STATE();
    case 339:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(363);
      END_STATE();
    case 340:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(364);
      END_STATE();
    case 341:
      ACCEPT_TOKEN(aux_sym_cast_type_token7);
      END_STATE();
    case 342:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(365);
      END_STATE();
    case 343:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(366);
      END_STATE();
    case 344:
      ACCEPT_TOKEN(aux_sym_visibility_modifier_token3);
      END_STATE();
    case 345:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(367);
      END_STATE();
    case 346:
      if (lookahead == 'g') ADVANCE(368);
      END_STATE();
    case 347:
      if (lookahead == 'e') ADVANCE(369);
      END_STATE();
    case 348:
      if (lookahead == 't') ADVANCE(370);
      END_STATE();
    case 349:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(371);
      END_STATE();
    case 350:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(372);
      END_STATE();
    case 351:
      ACCEPT_TOKEN(aux_sym_declare_statement_token1);
      END_STATE();
    case 352:
      ACCEPT_TOKEN(aux_sym_match_default_expression_token1);
      END_STATE();
    case 353:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(373);
      END_STATE();
    case 354:
      ACCEPT_TOKEN(aux_sym_require_expression_token1);
      if (lookahead == '_') ADVANCE(374);
      END_STATE();
    case 355:
      ACCEPT_TOKEN(aux_sym_abstract_modifier_token1);
      END_STATE();
    case 356:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(375);
      END_STATE();
    case 357:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(376);
      END_STATE();
    case 358:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(377);
      END_STATE();
    case 359:
      ACCEPT_TOKEN(aux_sym_while_statement_token2);
      END_STATE();
    case 360:
      ACCEPT_TOKEN(aux_sym__namespace_use_type_token1);
      END_STATE();
    case 361:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(378);
      END_STATE();
    case 362:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(379);
      END_STATE();
    case 363:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(380);
      END_STATE();
    case 364:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(381);
      END_STATE();
    case 365:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(382);
      END_STATE();
    case 366:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(383);
      END_STATE();
    case 367:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(384);
      END_STATE();
    case 368:
      ACCEPT_TOKEN(anon_sym_encoding);
      END_STATE();
    case 369:
      ACCEPT_TOKEN(anon_sym_iterable);
      END_STATE();
    case 370:
      if (lookahead == 'y') ADVANCE(385);
      END_STATE();
    case 371:
      ACCEPT_TOKEN(aux_sym_primitive_type_token1);
      END_STATE();
    case 372:
      ACCEPT_TOKEN(aux_sym_continue_statement_token1);
      END_STATE();
    case 373:
      ACCEPT_TOKEN(aux_sym_readonly_modifier_token1);
      END_STATE();
    case 374:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(386);
      END_STATE();
    case 375:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(387);
      END_STATE();
    case 376:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(388);
      END_STATE();
    case 377:
      ACCEPT_TOKEN(aux_sym_switch_block_token1);
      END_STATE();
    case 378:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(389);
      END_STATE();
    case 379:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(390);
      END_STATE();
    case 380:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(391);
      END_STATE();
    case 381:
      ACCEPT_TOKEN(aux_sym_use_instead_of_clause_token1);
      END_STATE();
    case 382:
      ACCEPT_TOKEN(aux_sym_interface_declaration_token1);
      END_STATE();
    case 383:
      ACCEPT_TOKEN(aux_sym_namespace_definition_token1);
      END_STATE();
    case 384:
      ACCEPT_TOKEN(aux_sym_visibility_modifier_token2);
      END_STATE();
    case 385:
      if (lookahead == 'p') ADVANCE(392);
      END_STATE();
    case 386:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(393);
      END_STATE();
    case 387:
      ACCEPT_TOKEN(aux_sym_declare_statement_token2);
      END_STATE();
    case 388:
      ACCEPT_TOKEN(aux_sym_foreach_statement_token2);
      END_STATE();
    case 389:
      ACCEPT_TOKEN(aux_sym_class_interface_clause_token1);
      END_STATE();
    case 390:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(394);
      END_STATE();
    case 391:
      ACCEPT_TOKEN(aux_sym_binary_expression_token1);
      END_STATE();
    case 392:
      if (lookahead == 'e') ADVANCE(395);
      END_STATE();
    case 393:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(396);
      END_STATE();
    case 394:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(397);
      END_STATE();
    case 395:
      if (lookahead == 's') ADVANCE(398);
      END_STATE();
    case 396:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(399);
      END_STATE();
    case 397:
      ACCEPT_TOKEN(aux_sym_include_once_expression_token1);
      END_STATE();
    case 398:
      ACCEPT_TOKEN(anon_sym_strict_types);
      END_STATE();
    case 399:
      ACCEPT_TOKEN(aux_sym_require_once_expression_token1);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 64},
  [2] = {.lex_state = 64},
  [3] = {.lex_state = 64},
  [4] = {.lex_state = 64},
  [5] = {.lex_state = 64},
  [6] = {.lex_state = 64},
  [7] = {.lex_state = 64},
  [8] = {.lex_state = 64},
  [9] = {.lex_state = 64, .external_lex_state = 2},
  [10] = {.lex_state = 64, .external_lex_state = 2},
  [11] = {.lex_state = 64},
  [12] = {.lex_state = 64},
  [13] = {.lex_state = 64},
  [14] = {.lex_state = 64, .external_lex_state = 2},
  [15] = {.lex_state = 64, .external_lex_state = 2},
  [16] = {.lex_state = 64, .external_lex_state = 2},
  [17] = {.lex_state = 64, .external_lex_state = 2},
  [18] = {.lex_state = 64, .external_lex_state = 2},
  [19] = {.lex_state = 64, .external_lex_state = 2},
  [20] = {.lex_state = 64, .external_lex_state = 2},
  [21] = {.lex_state = 64, .external_lex_state = 2},
  [22] = {.lex_state = 64, .external_lex_state = 2},
  [23] = {.lex_state = 64, .external_lex_state = 2},
  [24] = {.lex_state = 64, .external_lex_state = 2},
  [25] = {.lex_state = 64, .external_lex_state = 2},
  [26] = {.lex_state = 64},
  [27] = {.lex_state = 64, .external_lex_state = 2},
  [28] = {.lex_state = 64, .external_lex_state = 2},
  [29] = {.lex_state = 64, .external_lex_state = 2},
  [30] = {.lex_state = 64, .external_lex_state = 2},
  [31] = {.lex_state = 64, .external_lex_state = 2},
  [32] = {.lex_state = 64, .external_lex_state = 2},
  [33] = {.lex_state = 64, .external_lex_state = 2},
  [34] = {.lex_state = 64, .external_lex_state = 2},
  [35] = {.lex_state = 64},
  [36] = {.lex_state = 64, .external_lex_state = 2},
  [37] = {.lex_state = 64, .external_lex_state = 2},
  [38] = {.lex_state = 64, .external_lex_state = 2},
  [39] = {.lex_state = 64, .external_lex_state = 2},
  [40] = {.lex_state = 64, .external_lex_state = 2},
  [41] = {.lex_state = 64, .external_lex_state = 2},
  [42] = {.lex_state = 64, .external_lex_state = 2},
  [43] = {.lex_state = 64, .external_lex_state = 2},
  [44] = {.lex_state = 64, .external_lex_state = 2},
  [45] = {.lex_state = 64, .external_lex_state = 2},
  [46] = {.lex_state = 64, .external_lex_state = 2},
  [47] = {.lex_state = 64, .external_lex_state = 2},
  [48] = {.lex_state = 64, .external_lex_state = 2},
  [49] = {.lex_state = 64, .external_lex_state = 2},
  [50] = {.lex_state = 64},
  [51] = {.lex_state = 64, .external_lex_state = 2},
  [52] = {.lex_state = 64, .external_lex_state = 2},
  [53] = {.lex_state = 64, .external_lex_state = 2},
  [54] = {.lex_state = 64, .external_lex_state = 2},
  [55] = {.lex_state = 64},
  [56] = {.lex_state = 64},
  [57] = {.lex_state = 64},
  [58] = {.lex_state = 64},
  [59] = {.lex_state = 64},
  [60] = {.lex_state = 64},
  [61] = {.lex_state = 64},
  [62] = {.lex_state = 64},
  [63] = {.lex_state = 64},
  [64] = {.lex_state = 64},
  [65] = {.lex_state = 64},
  [66] = {.lex_state = 64},
  [67] = {.lex_state = 64},
  [68] = {.lex_state = 64},
  [69] = {.lex_state = 64},
  [70] = {.lex_state = 64},
  [71] = {.lex_state = 64},
  [72] = {.lex_state = 64},
  [73] = {.lex_state = 64},
  [74] = {.lex_state = 64},
  [75] = {.lex_state = 64},
  [76] = {.lex_state = 64},
  [77] = {.lex_state = 64},
  [78] = {.lex_state = 64},
  [79] = {.lex_state = 64},
  [80] = {.lex_state = 64},
  [81] = {.lex_state = 64},
  [82] = {.lex_state = 64},
  [83] = {.lex_state = 64},
  [84] = {.lex_state = 64},
  [85] = {.lex_state = 64},
  [86] = {.lex_state = 64},
  [87] = {.lex_state = 64},
  [88] = {.lex_state = 64},
  [89] = {.lex_state = 64},
  [90] = {.lex_state = 64},
  [91] = {.lex_state = 64},
  [92] = {.lex_state = 64},
  [93] = {.lex_state = 64},
  [94] = {.lex_state = 64},
  [95] = {.lex_state = 64},
  [96] = {.lex_state = 64},
  [97] = {.lex_state = 64},
  [98] = {.lex_state = 64},
  [99] = {.lex_state = 64},
  [100] = {.lex_state = 64},
  [101] = {.lex_state = 64},
  [102] = {.lex_state = 64},
  [103] = {.lex_state = 64},
  [104] = {.lex_state = 64},
  [105] = {.lex_state = 64},
  [106] = {.lex_state = 64},
  [107] = {.lex_state = 64},
  [108] = {.lex_state = 64},
  [109] = {.lex_state = 64},
  [110] = {.lex_state = 64},
  [111] = {.lex_state = 64},
  [112] = {.lex_state = 64},
  [113] = {.lex_state = 64},
  [114] = {.lex_state = 64},
  [115] = {.lex_state = 64},
  [116] = {.lex_state = 64},
  [117] = {.lex_state = 5},
  [118] = {.lex_state = 5},
  [119] = {.lex_state = 5, .external_lex_state = 2},
  [120] = {.lex_state = 5},
  [121] = {.lex_state = 7},
  [122] = {.lex_state = 7},
  [123] = {.lex_state = 7},
  [124] = {.lex_state = 7},
  [125] = {.lex_state = 7},
  [126] = {.lex_state = 7},
  [127] = {.lex_state = 7},
  [128] = {.lex_state = 7},
  [129] = {.lex_state = 7},
  [130] = {.lex_state = 7},
  [131] = {.lex_state = 7},
  [132] = {.lex_state = 6},
  [133] = {.lex_state = 6},
  [134] = {.lex_state = 6},
  [135] = {.lex_state = 6},
  [136] = {.lex_state = 6},
  [137] = {.lex_state = 6},
  [138] = {.lex_state = 6},
  [139] = {.lex_state = 6},
  [140] = {.lex_state = 6},
  [141] = {.lex_state = 6},
  [142] = {.lex_state = 6},
  [143] = {.lex_state = 6},
  [144] = {.lex_state = 6},
  [145] = {.lex_state = 6},
  [146] = {.lex_state = 6},
  [147] = {.lex_state = 6},
  [148] = {.lex_state = 6},
  [149] = {.lex_state = 6},
  [150] = {.lex_state = 6},
  [151] = {.lex_state = 6},
  [152] = {.lex_state = 6},
  [153] = {.lex_state = 6},
  [154] = {.lex_state = 6},
  [155] = {.lex_state = 6},
  [156] = {.lex_state = 6},
  [157] = {.lex_state = 6},
  [158] = {.lex_state = 6},
  [159] = {.lex_state = 6},
  [160] = {.lex_state = 6},
  [161] = {.lex_state = 6},
  [162] = {.lex_state = 6},
  [163] = {.lex_state = 64},
  [164] = {.lex_state = 6},
  [165] = {.lex_state = 6},
  [166] = {.lex_state = 6},
  [167] = {.lex_state = 6},
  [168] = {.lex_state = 6},
  [169] = {.lex_state = 64},
  [170] = {.lex_state = 64},
  [171] = {.lex_state = 64},
  [172] = {.lex_state = 6},
  [173] = {.lex_state = 6},
  [174] = {.lex_state = 64},
  [175] = {.lex_state = 64},
  [176] = {.lex_state = 64},
  [177] = {.lex_state = 64},
  [178] = {.lex_state = 64},
  [179] = {.lex_state = 64},
  [180] = {.lex_state = 64},
  [181] = {.lex_state = 6},
  [182] = {.lex_state = 64},
  [183] = {.lex_state = 64},
  [184] = {.lex_state = 64},
  [185] = {.lex_state = 64},
  [186] = {.lex_state = 64},
  [187] = {.lex_state = 64},
  [188] = {.lex_state = 64},
  [189] = {.lex_state = 64},
  [190] = {.lex_state = 64, .external_lex_state = 2},
  [191] = {.lex_state = 64},
  [192] = {.lex_state = 64},
  [193] = {.lex_state = 64},
  [194] = {.lex_state = 64, .external_lex_state = 2},
  [195] = {.lex_state = 64, .external_lex_state = 2},
  [196] = {.lex_state = 64},
  [197] = {.lex_state = 64},
  [198] = {.lex_state = 64, .external_lex_state = 2},
  [199] = {.lex_state = 64, .external_lex_state = 2},
  [200] = {.lex_state = 64},
  [201] = {.lex_state = 64},
  [202] = {.lex_state = 64, .external_lex_state = 2},
  [203] = {.lex_state = 64},
  [204] = {.lex_state = 64},
  [205] = {.lex_state = 64},
  [206] = {.lex_state = 64},
  [207] = {.lex_state = 64},
  [208] = {.lex_state = 64},
  [209] = {.lex_state = 64},
  [210] = {.lex_state = 64},
  [211] = {.lex_state = 64},
  [212] = {.lex_state = 64},
  [213] = {.lex_state = 64},
  [214] = {.lex_state = 64},
  [215] = {.lex_state = 64},
  [216] = {.lex_state = 64},
  [217] = {.lex_state = 64},
  [218] = {.lex_state = 64},
  [219] = {.lex_state = 64},
  [220] = {.lex_state = 64},
  [221] = {.lex_state = 64},
  [222] = {.lex_state = 6},
  [223] = {.lex_state = 6},
  [224] = {.lex_state = 64},
  [225] = {.lex_state = 64},
  [226] = {.lex_state = 64},
  [227] = {.lex_state = 64},
  [228] = {.lex_state = 64},
  [229] = {.lex_state = 64},
  [230] = {.lex_state = 64},
  [231] = {.lex_state = 64},
  [232] = {.lex_state = 64},
  [233] = {.lex_state = 64},
  [234] = {.lex_state = 64},
  [235] = {.lex_state = 64},
  [236] = {.lex_state = 64},
  [237] = {.lex_state = 64},
  [238] = {.lex_state = 64},
  [239] = {.lex_state = 64},
  [240] = {.lex_state = 64},
  [241] = {.lex_state = 64},
  [242] = {.lex_state = 64},
  [243] = {.lex_state = 64},
  [244] = {.lex_state = 64},
  [245] = {.lex_state = 64},
  [246] = {.lex_state = 64},
  [247] = {.lex_state = 64},
  [248] = {.lex_state = 64},
  [249] = {.lex_state = 64},
  [250] = {.lex_state = 64},
  [251] = {.lex_state = 64},
  [252] = {.lex_state = 64},
  [253] = {.lex_state = 64},
  [254] = {.lex_state = 64},
  [255] = {.lex_state = 64},
  [256] = {.lex_state = 64},
  [257] = {.lex_state = 64},
  [258] = {.lex_state = 64},
  [259] = {.lex_state = 64},
  [260] = {.lex_state = 64},
  [261] = {.lex_state = 64},
  [262] = {.lex_state = 64},
  [263] = {.lex_state = 64},
  [264] = {.lex_state = 64},
  [265] = {.lex_state = 64},
  [266] = {.lex_state = 64},
  [267] = {.lex_state = 64},
  [268] = {.lex_state = 64},
  [269] = {.lex_state = 64},
  [270] = {.lex_state = 64},
  [271] = {.lex_state = 64},
  [272] = {.lex_state = 64},
  [273] = {.lex_state = 64},
  [274] = {.lex_state = 64},
  [275] = {.lex_state = 64},
  [276] = {.lex_state = 64},
  [277] = {.lex_state = 64},
  [278] = {.lex_state = 64},
  [279] = {.lex_state = 64},
  [280] = {.lex_state = 64},
  [281] = {.lex_state = 64},
  [282] = {.lex_state = 64},
  [283] = {.lex_state = 64},
  [284] = {.lex_state = 64},
  [285] = {.lex_state = 64},
  [286] = {.lex_state = 64},
  [287] = {.lex_state = 64},
  [288] = {.lex_state = 64},
  [289] = {.lex_state = 64},
  [290] = {.lex_state = 64},
  [291] = {.lex_state = 64},
  [292] = {.lex_state = 64},
  [293] = {.lex_state = 64},
  [294] = {.lex_state = 64},
  [295] = {.lex_state = 64},
  [296] = {.lex_state = 64},
  [297] = {.lex_state = 64},
  [298] = {.lex_state = 64},
  [299] = {.lex_state = 64},
  [300] = {.lex_state = 64},
  [301] = {.lex_state = 64},
  [302] = {.lex_state = 64},
  [303] = {.lex_state = 64},
  [304] = {.lex_state = 64},
  [305] = {.lex_state = 64},
  [306] = {.lex_state = 64},
  [307] = {.lex_state = 64},
  [308] = {.lex_state = 64},
  [309] = {.lex_state = 64},
  [310] = {.lex_state = 64},
  [311] = {.lex_state = 64},
  [312] = {.lex_state = 64},
  [313] = {.lex_state = 64},
  [314] = {.lex_state = 64},
  [315] = {.lex_state = 64},
  [316] = {.lex_state = 64},
  [317] = {.lex_state = 64},
  [318] = {.lex_state = 64},
  [319] = {.lex_state = 64},
  [320] = {.lex_state = 64},
  [321] = {.lex_state = 64},
  [322] = {.lex_state = 64},
  [323] = {.lex_state = 64},
  [324] = {.lex_state = 64},
  [325] = {.lex_state = 64},
  [326] = {.lex_state = 64},
  [327] = {.lex_state = 64},
  [328] = {.lex_state = 64},
  [329] = {.lex_state = 64},
  [330] = {.lex_state = 64},
  [331] = {.lex_state = 64},
  [332] = {.lex_state = 64},
  [333] = {.lex_state = 64},
  [334] = {.lex_state = 64},
  [335] = {.lex_state = 64},
  [336] = {.lex_state = 64},
  [337] = {.lex_state = 64},
  [338] = {.lex_state = 64},
  [339] = {.lex_state = 64},
  [340] = {.lex_state = 64},
  [341] = {.lex_state = 64},
  [342] = {.lex_state = 64},
  [343] = {.lex_state = 64},
  [344] = {.lex_state = 64},
  [345] = {.lex_state = 64},
  [346] = {.lex_state = 64},
  [347] = {.lex_state = 64},
  [348] = {.lex_state = 64},
  [349] = {.lex_state = 64},
  [350] = {.lex_state = 64},
  [351] = {.lex_state = 64},
  [352] = {.lex_state = 64},
  [353] = {.lex_state = 64},
  [354] = {.lex_state = 64},
  [355] = {.lex_state = 64},
  [356] = {.lex_state = 64},
  [357] = {.lex_state = 64},
  [358] = {.lex_state = 64},
  [359] = {.lex_state = 64},
  [360] = {.lex_state = 64},
  [361] = {.lex_state = 64},
  [362] = {.lex_state = 64},
  [363] = {.lex_state = 64},
  [364] = {.lex_state = 64},
  [365] = {.lex_state = 64},
  [366] = {.lex_state = 64},
  [367] = {.lex_state = 64},
  [368] = {.lex_state = 64},
  [369] = {.lex_state = 64},
  [370] = {.lex_state = 64},
  [371] = {.lex_state = 64},
  [372] = {.lex_state = 64},
  [373] = {.lex_state = 64},
  [374] = {.lex_state = 64},
  [375] = {.lex_state = 64},
  [376] = {.lex_state = 64},
  [377] = {.lex_state = 64},
  [378] = {.lex_state = 64},
  [379] = {.lex_state = 64},
  [380] = {.lex_state = 64},
  [381] = {.lex_state = 64},
  [382] = {.lex_state = 64},
  [383] = {.lex_state = 64},
  [384] = {.lex_state = 64},
  [385] = {.lex_state = 64},
  [386] = {.lex_state = 64},
  [387] = {.lex_state = 64},
  [388] = {.lex_state = 64},
  [389] = {.lex_state = 64},
  [390] = {.lex_state = 64},
  [391] = {.lex_state = 64},
  [392] = {.lex_state = 64},
  [393] = {.lex_state = 64},
  [394] = {.lex_state = 64},
  [395] = {.lex_state = 64},
  [396] = {.lex_state = 64},
  [397] = {.lex_state = 64},
  [398] = {.lex_state = 64},
  [399] = {.lex_state = 64},
  [400] = {.lex_state = 64},
  [401] = {.lex_state = 64},
  [402] = {.lex_state = 64},
  [403] = {.lex_state = 64},
  [404] = {.lex_state = 64},
  [405] = {.lex_state = 64},
  [406] = {.lex_state = 64},
  [407] = {.lex_state = 64},
  [408] = {.lex_state = 64},
  [409] = {.lex_state = 64},
  [410] = {.lex_state = 64},
  [411] = {.lex_state = 64},
  [412] = {.lex_state = 64},
  [413] = {.lex_state = 64},
  [414] = {.lex_state = 64},
  [415] = {.lex_state = 64},
  [416] = {.lex_state = 64},
  [417] = {.lex_state = 64},
  [418] = {.lex_state = 64},
  [419] = {.lex_state = 64},
  [420] = {.lex_state = 64},
  [421] = {.lex_state = 64},
  [422] = {.lex_state = 64},
  [423] = {.lex_state = 64},
  [424] = {.lex_state = 64},
  [425] = {.lex_state = 64},
  [426] = {.lex_state = 64},
  [427] = {.lex_state = 64},
  [428] = {.lex_state = 64},
  [429] = {.lex_state = 64},
  [430] = {.lex_state = 64},
  [431] = {.lex_state = 64},
  [432] = {.lex_state = 64},
  [433] = {.lex_state = 64},
  [434] = {.lex_state = 64},
  [435] = {.lex_state = 64},
  [436] = {.lex_state = 64},
  [437] = {.lex_state = 64},
  [438] = {.lex_state = 64},
  [439] = {.lex_state = 64},
  [440] = {.lex_state = 64},
  [441] = {.lex_state = 64},
  [442] = {.lex_state = 64},
  [443] = {.lex_state = 64},
  [444] = {.lex_state = 64},
  [445] = {.lex_state = 64},
  [446] = {.lex_state = 64},
  [447] = {.lex_state = 64},
  [448] = {.lex_state = 64},
  [449] = {.lex_state = 64},
  [450] = {.lex_state = 64},
  [451] = {.lex_state = 64},
  [452] = {.lex_state = 64},
  [453] = {.lex_state = 64},
  [454] = {.lex_state = 64},
  [455] = {.lex_state = 64},
  [456] = {.lex_state = 64},
  [457] = {.lex_state = 64},
  [458] = {.lex_state = 64},
  [459] = {.lex_state = 64},
  [460] = {.lex_state = 64},
  [461] = {.lex_state = 64},
  [462] = {.lex_state = 64},
  [463] = {.lex_state = 64},
  [464] = {.lex_state = 64},
  [465] = {.lex_state = 64},
  [466] = {.lex_state = 64},
  [467] = {.lex_state = 64},
  [468] = {.lex_state = 64},
  [469] = {.lex_state = 64},
  [470] = {.lex_state = 64},
  [471] = {.lex_state = 64},
  [472] = {.lex_state = 64},
  [473] = {.lex_state = 64},
  [474] = {.lex_state = 64},
  [475] = {.lex_state = 64},
  [476] = {.lex_state = 64},
  [477] = {.lex_state = 64},
  [478] = {.lex_state = 64},
  [479] = {.lex_state = 64},
  [480] = {.lex_state = 64},
  [481] = {.lex_state = 64},
  [482] = {.lex_state = 64},
  [483] = {.lex_state = 64},
  [484] = {.lex_state = 64},
  [485] = {.lex_state = 64},
  [486] = {.lex_state = 64},
  [487] = {.lex_state = 64},
  [488] = {.lex_state = 64},
  [489] = {.lex_state = 64},
  [490] = {.lex_state = 64},
  [491] = {.lex_state = 64},
  [492] = {.lex_state = 64},
  [493] = {.lex_state = 64},
  [494] = {.lex_state = 64},
  [495] = {.lex_state = 64},
  [496] = {.lex_state = 64},
  [497] = {.lex_state = 64},
  [498] = {.lex_state = 64},
  [499] = {.lex_state = 64},
  [500] = {.lex_state = 64},
  [501] = {.lex_state = 64},
  [502] = {.lex_state = 64},
  [503] = {.lex_state = 64},
  [504] = {.lex_state = 64},
  [505] = {.lex_state = 64},
  [506] = {.lex_state = 64},
  [507] = {.lex_state = 64},
  [508] = {.lex_state = 64},
  [509] = {.lex_state = 64},
  [510] = {.lex_state = 64},
  [511] = {.lex_state = 64},
  [512] = {.lex_state = 64},
  [513] = {.lex_state = 64},
  [514] = {.lex_state = 64},
  [515] = {.lex_state = 64},
  [516] = {.lex_state = 64},
  [517] = {.lex_state = 64},
  [518] = {.lex_state = 64},
  [519] = {.lex_state = 64},
  [520] = {.lex_state = 64},
  [521] = {.lex_state = 64},
  [522] = {.lex_state = 64},
  [523] = {.lex_state = 64},
  [524] = {.lex_state = 64},
  [525] = {.lex_state = 64},
  [526] = {.lex_state = 64},
  [527] = {.lex_state = 64},
  [528] = {.lex_state = 64},
  [529] = {.lex_state = 64},
  [530] = {.lex_state = 64},
  [531] = {.lex_state = 64},
  [532] = {.lex_state = 64},
  [533] = {.lex_state = 64},
  [534] = {.lex_state = 64},
  [535] = {.lex_state = 64},
  [536] = {.lex_state = 64},
  [537] = {.lex_state = 64},
  [538] = {.lex_state = 64},
  [539] = {.lex_state = 64},
  [540] = {.lex_state = 64},
  [541] = {.lex_state = 64},
  [542] = {.lex_state = 64},
  [543] = {.lex_state = 64},
  [544] = {.lex_state = 64},
  [545] = {.lex_state = 64},
  [546] = {.lex_state = 64},
  [547] = {.lex_state = 64},
  [548] = {.lex_state = 64},
  [549] = {.lex_state = 64},
  [550] = {.lex_state = 64},
  [551] = {.lex_state = 64},
  [552] = {.lex_state = 64},
  [553] = {.lex_state = 64},
  [554] = {.lex_state = 64},
  [555] = {.lex_state = 64},
  [556] = {.lex_state = 64},
  [557] = {.lex_state = 64},
  [558] = {.lex_state = 64},
  [559] = {.lex_state = 64},
  [560] = {.lex_state = 64},
  [561] = {.lex_state = 64},
  [562] = {.lex_state = 64},
  [563] = {.lex_state = 64},
  [564] = {.lex_state = 64},
  [565] = {.lex_state = 64},
  [566] = {.lex_state = 64},
  [567] = {.lex_state = 64},
  [568] = {.lex_state = 64},
  [569] = {.lex_state = 64},
  [570] = {.lex_state = 64},
  [571] = {.lex_state = 64},
  [572] = {.lex_state = 64},
  [573] = {.lex_state = 64},
  [574] = {.lex_state = 64},
  [575] = {.lex_state = 64},
  [576] = {.lex_state = 64},
  [577] = {.lex_state = 64},
  [578] = {.lex_state = 64},
  [579] = {.lex_state = 64},
  [580] = {.lex_state = 64},
  [581] = {.lex_state = 64},
  [582] = {.lex_state = 64},
  [583] = {.lex_state = 64},
  [584] = {.lex_state = 65},
  [585] = {.lex_state = 65},
  [586] = {.lex_state = 65},
  [587] = {.lex_state = 65},
  [588] = {.lex_state = 65},
  [589] = {.lex_state = 65},
  [590] = {.lex_state = 65},
  [591] = {.lex_state = 65},
  [592] = {.lex_state = 65},
  [593] = {.lex_state = 65},
  [594] = {.lex_state = 65},
  [595] = {.lex_state = 65},
  [596] = {.lex_state = 65},
  [597] = {.lex_state = 65},
  [598] = {.lex_state = 65},
  [599] = {.lex_state = 65},
  [600] = {.lex_state = 65},
  [601] = {.lex_state = 65},
  [602] = {.lex_state = 65},
  [603] = {.lex_state = 65},
  [604] = {.lex_state = 65},
  [605] = {.lex_state = 65},
  [606] = {.lex_state = 65},
  [607] = {.lex_state = 65},
  [608] = {.lex_state = 65},
  [609] = {.lex_state = 65},
  [610] = {.lex_state = 65},
  [611] = {.lex_state = 65},
  [612] = {.lex_state = 65},
  [613] = {.lex_state = 65},
  [614] = {.lex_state = 65},
  [615] = {.lex_state = 65, .external_lex_state = 2},
  [616] = {.lex_state = 65},
  [617] = {.lex_state = 65, .external_lex_state = 2},
  [618] = {.lex_state = 65, .external_lex_state = 2},
  [619] = {.lex_state = 65, .external_lex_state = 2},
  [620] = {.lex_state = 65, .external_lex_state = 2},
  [621] = {.lex_state = 65},
  [622] = {.lex_state = 65, .external_lex_state = 2},
  [623] = {.lex_state = 65, .external_lex_state = 2},
  [624] = {.lex_state = 65, .external_lex_state = 2},
  [625] = {.lex_state = 14},
  [626] = {.lex_state = 65, .external_lex_state = 2},
  [627] = {.lex_state = 14},
  [628] = {.lex_state = 65, .external_lex_state = 2},
  [629] = {.lex_state = 65, .external_lex_state = 2},
  [630] = {.lex_state = 14},
  [631] = {.lex_state = 65, .external_lex_state = 2},
  [632] = {.lex_state = 65, .external_lex_state = 2},
  [633] = {.lex_state = 65, .external_lex_state = 2},
  [634] = {.lex_state = 65, .external_lex_state = 2},
  [635] = {.lex_state = 65, .external_lex_state = 2},
  [636] = {.lex_state = 65, .external_lex_state = 2},
  [637] = {.lex_state = 65, .external_lex_state = 2},
  [638] = {.lex_state = 65, .external_lex_state = 2},
  [639] = {.lex_state = 14},
  [640] = {.lex_state = 65, .external_lex_state = 2},
  [641] = {.lex_state = 65, .external_lex_state = 2},
  [642] = {.lex_state = 14},
  [643] = {.lex_state = 65},
  [644] = {.lex_state = 65},
  [645] = {.lex_state = 65, .external_lex_state = 2},
  [646] = {.lex_state = 65},
  [647] = {.lex_state = 65},
  [648] = {.lex_state = 65, .external_lex_state = 2},
  [649] = {.lex_state = 65, .external_lex_state = 2},
  [650] = {.lex_state = 65, .external_lex_state = 2},
  [651] = {.lex_state = 65, .external_lex_state = 2},
  [652] = {.lex_state = 65, .external_lex_state = 2},
  [653] = {.lex_state = 65, .external_lex_state = 2},
  [654] = {.lex_state = 65, .external_lex_state = 2},
  [655] = {.lex_state = 65, .external_lex_state = 2},
  [656] = {.lex_state = 14},
  [657] = {.lex_state = 65, .external_lex_state = 2},
  [658] = {.lex_state = 65},
  [659] = {.lex_state = 65, .external_lex_state = 2},
  [660] = {.lex_state = 65},
  [661] = {.lex_state = 65, .external_lex_state = 2},
  [662] = {.lex_state = 65},
  [663] = {.lex_state = 65, .external_lex_state = 2},
  [664] = {.lex_state = 65},
  [665] = {.lex_state = 65},
  [666] = {.lex_state = 65},
  [667] = {.lex_state = 65},
  [668] = {.lex_state = 65},
  [669] = {.lex_state = 65},
  [670] = {.lex_state = 65},
  [671] = {.lex_state = 65},
  [672] = {.lex_state = 65},
  [673] = {.lex_state = 65},
  [674] = {.lex_state = 14},
  [675] = {.lex_state = 14},
  [676] = {.lex_state = 65},
  [677] = {.lex_state = 65},
  [678] = {.lex_state = 65},
  [679] = {.lex_state = 65},
  [680] = {.lex_state = 65},
  [681] = {.lex_state = 14},
  [682] = {.lex_state = 65},
  [683] = {.lex_state = 14},
  [684] = {.lex_state = 10},
  [685] = {.lex_state = 14},
  [686] = {.lex_state = 14},
  [687] = {.lex_state = 14},
  [688] = {.lex_state = 14},
  [689] = {.lex_state = 14},
  [690] = {.lex_state = 14},
  [691] = {.lex_state = 14},
  [692] = {.lex_state = 14},
  [693] = {.lex_state = 65},
  [694] = {.lex_state = 14},
  [695] = {.lex_state = 65},
  [696] = {.lex_state = 14},
  [697] = {.lex_state = 14},
  [698] = {.lex_state = 14},
  [699] = {.lex_state = 14},
  [700] = {.lex_state = 10},
  [701] = {.lex_state = 12},
  [702] = {.lex_state = 12},
  [703] = {.lex_state = 8},
  [704] = {.lex_state = 12},
  [705] = {.lex_state = 10},
  [706] = {.lex_state = 8},
  [707] = {.lex_state = 8},
  [708] = {.lex_state = 10},
  [709] = {.lex_state = 10},
  [710] = {.lex_state = 8},
  [711] = {.lex_state = 8},
  [712] = {.lex_state = 8},
  [713] = {.lex_state = 12, .external_lex_state = 2},
  [714] = {.lex_state = 8},
  [715] = {.lex_state = 8},
  [716] = {.lex_state = 8},
  [717] = {.lex_state = 8},
  [718] = {.lex_state = 8},
  [719] = {.lex_state = 8},
  [720] = {.lex_state = 8},
  [721] = {.lex_state = 8},
  [722] = {.lex_state = 8},
  [723] = {.lex_state = 8},
  [724] = {.lex_state = 8},
  [725] = {.lex_state = 8},
  [726] = {.lex_state = 8},
  [727] = {.lex_state = 8},
  [728] = {.lex_state = 8},
  [729] = {.lex_state = 8},
  [730] = {.lex_state = 10},
  [731] = {.lex_state = 8},
  [732] = {.lex_state = 8},
  [733] = {.lex_state = 8},
  [734] = {.lex_state = 8},
  [735] = {.lex_state = 8},
  [736] = {.lex_state = 10},
  [737] = {.lex_state = 10},
  [738] = {.lex_state = 8},
  [739] = {.lex_state = 8},
  [740] = {.lex_state = 10},
  [741] = {.lex_state = 8},
  [742] = {.lex_state = 8},
  [743] = {.lex_state = 10},
  [744] = {.lex_state = 8},
  [745] = {.lex_state = 8},
  [746] = {.lex_state = 8},
  [747] = {.lex_state = 10},
  [748] = {.lex_state = 10},
  [749] = {.lex_state = 10},
  [750] = {.lex_state = 10},
  [751] = {.lex_state = 10},
  [752] = {.lex_state = 10},
  [753] = {.lex_state = 10},
  [754] = {.lex_state = 10},
  [755] = {.lex_state = 10},
  [756] = {.lex_state = 10},
  [757] = {.lex_state = 10},
  [758] = {.lex_state = 10},
  [759] = {.lex_state = 10},
  [760] = {.lex_state = 10},
  [761] = {.lex_state = 10},
  [762] = {.lex_state = 10},
  [763] = {.lex_state = 10},
  [764] = {.lex_state = 10},
  [765] = {.lex_state = 10},
  [766] = {.lex_state = 10},
  [767] = {.lex_state = 10},
  [768] = {.lex_state = 10},
  [769] = {.lex_state = 10},
  [770] = {.lex_state = 10},
  [771] = {.lex_state = 10},
  [772] = {.lex_state = 10},
  [773] = {.lex_state = 10},
  [774] = {.lex_state = 10},
  [775] = {.lex_state = 10},
  [776] = {.lex_state = 10},
  [777] = {.lex_state = 10},
  [778] = {.lex_state = 10},
  [779] = {.lex_state = 10},
  [780] = {.lex_state = 10},
  [781] = {.lex_state = 10},
  [782] = {.lex_state = 16},
  [783] = {.lex_state = 10},
  [784] = {.lex_state = 10},
  [785] = {.lex_state = 10},
  [786] = {.lex_state = 10},
  [787] = {.lex_state = 10},
  [788] = {.lex_state = 10},
  [789] = {.lex_state = 10},
  [790] = {.lex_state = 10},
  [791] = {.lex_state = 10},
  [792] = {.lex_state = 10},
  [793] = {.lex_state = 10},
  [794] = {.lex_state = 10},
  [795] = {.lex_state = 10},
  [796] = {.lex_state = 16},
  [797] = {.lex_state = 12},
  [798] = {.lex_state = 10},
  [799] = {.lex_state = 10},
  [800] = {.lex_state = 10},
  [801] = {.lex_state = 10},
  [802] = {.lex_state = 10},
  [803] = {.lex_state = 10},
  [804] = {.lex_state = 10},
  [805] = {.lex_state = 10},
  [806] = {.lex_state = 10},
  [807] = {.lex_state = 10},
  [808] = {.lex_state = 10},
  [809] = {.lex_state = 10},
  [810] = {.lex_state = 10},
  [811] = {.lex_state = 10},
  [812] = {.lex_state = 10},
  [813] = {.lex_state = 10},
  [814] = {.lex_state = 10},
  [815] = {.lex_state = 10},
  [816] = {.lex_state = 10},
  [817] = {.lex_state = 10},
  [818] = {.lex_state = 10},
  [819] = {.lex_state = 10},
  [820] = {.lex_state = 10},
  [821] = {.lex_state = 10},
  [822] = {.lex_state = 10},
  [823] = {.lex_state = 10},
  [824] = {.lex_state = 10},
  [825] = {.lex_state = 10},
  [826] = {.lex_state = 10},
  [827] = {.lex_state = 12},
  [828] = {.lex_state = 10},
  [829] = {.lex_state = 10},
  [830] = {.lex_state = 12},
  [831] = {.lex_state = 8, .external_lex_state = 2},
  [832] = {.lex_state = 8, .external_lex_state = 2},
  [833] = {.lex_state = 16},
  [834] = {.lex_state = 16},
  [835] = {.lex_state = 16},
  [836] = {.lex_state = 10},
  [837] = {.lex_state = 12},
  [838] = {.lex_state = 10},
  [839] = {.lex_state = 16},
  [840] = {.lex_state = 10, .external_lex_state = 2},
  [841] = {.lex_state = 10},
  [842] = {.lex_state = 12},
  [843] = {.lex_state = 8, .external_lex_state = 2},
  [844] = {.lex_state = 8, .external_lex_state = 2},
  [845] = {.lex_state = 8, .external_lex_state = 2},
  [846] = {.lex_state = 8, .external_lex_state = 2},
  [847] = {.lex_state = 8, .external_lex_state = 2},
  [848] = {.lex_state = 8, .external_lex_state = 2},
  [849] = {.lex_state = 8, .external_lex_state = 2},
  [850] = {.lex_state = 8, .external_lex_state = 2},
  [851] = {.lex_state = 16},
  [852] = {.lex_state = 16},
  [853] = {.lex_state = 8, .external_lex_state = 2},
  [854] = {.lex_state = 6},
  [855] = {.lex_state = 10},
  [856] = {.lex_state = 8, .external_lex_state = 2},
  [857] = {.lex_state = 8, .external_lex_state = 2},
  [858] = {.lex_state = 8, .external_lex_state = 2},
  [859] = {.lex_state = 8, .external_lex_state = 2},
  [860] = {.lex_state = 8, .external_lex_state = 2},
  [861] = {.lex_state = 8, .external_lex_state = 2},
  [862] = {.lex_state = 8, .external_lex_state = 2},
  [863] = {.lex_state = 10, .external_lex_state = 2},
  [864] = {.lex_state = 8, .external_lex_state = 2},
  [865] = {.lex_state = 16},
  [866] = {.lex_state = 16},
  [867] = {.lex_state = 8, .external_lex_state = 2},
  [868] = {.lex_state = 8, .external_lex_state = 2},
  [869] = {.lex_state = 8, .external_lex_state = 2},
  [870] = {.lex_state = 8, .external_lex_state = 2},
  [871] = {.lex_state = 10, .external_lex_state = 2},
  [872] = {.lex_state = 8, .external_lex_state = 2},
  [873] = {.lex_state = 6},
  [874] = {.lex_state = 12},
  [875] = {.lex_state = 8, .external_lex_state = 2},
  [876] = {.lex_state = 12},
  [877] = {.lex_state = 12},
  [878] = {.lex_state = 8, .external_lex_state = 2},
  [879] = {.lex_state = 8, .external_lex_state = 2},
  [880] = {.lex_state = 8, .external_lex_state = 2},
  [881] = {.lex_state = 8, .external_lex_state = 2},
  [882] = {.lex_state = 8, .external_lex_state = 2},
  [883] = {.lex_state = 10, .external_lex_state = 2},
  [884] = {.lex_state = 10, .external_lex_state = 2},
  [885] = {.lex_state = 10, .external_lex_state = 2},
  [886] = {.lex_state = 8, .external_lex_state = 2},
  [887] = {.lex_state = 8, .external_lex_state = 2},
  [888] = {.lex_state = 8, .external_lex_state = 2},
  [889] = {.lex_state = 8, .external_lex_state = 2},
  [890] = {.lex_state = 10, .external_lex_state = 2},
  [891] = {.lex_state = 10, .external_lex_state = 2},
  [892] = {.lex_state = 10, .external_lex_state = 2},
  [893] = {.lex_state = 10, .external_lex_state = 2},
  [894] = {.lex_state = 10, .external_lex_state = 2},
  [895] = {.lex_state = 10, .external_lex_state = 2},
  [896] = {.lex_state = 10, .external_lex_state = 2},
  [897] = {.lex_state = 10, .external_lex_state = 2},
  [898] = {.lex_state = 10, .external_lex_state = 2},
  [899] = {.lex_state = 10, .external_lex_state = 2},
  [900] = {.lex_state = 10, .external_lex_state = 2},
  [901] = {.lex_state = 10, .external_lex_state = 2},
  [902] = {.lex_state = 10, .external_lex_state = 2},
  [903] = {.lex_state = 10, .external_lex_state = 2},
  [904] = {.lex_state = 10, .external_lex_state = 2},
  [905] = {.lex_state = 10, .external_lex_state = 2},
  [906] = {.lex_state = 10, .external_lex_state = 2},
  [907] = {.lex_state = 10, .external_lex_state = 2},
  [908] = {.lex_state = 10, .external_lex_state = 2},
  [909] = {.lex_state = 10, .external_lex_state = 2},
  [910] = {.lex_state = 10, .external_lex_state = 2},
  [911] = {.lex_state = 10, .external_lex_state = 2},
  [912] = {.lex_state = 10, .external_lex_state = 2},
  [913] = {.lex_state = 10, .external_lex_state = 2},
  [914] = {.lex_state = 10, .external_lex_state = 2},
  [915] = {.lex_state = 10, .external_lex_state = 2},
  [916] = {.lex_state = 10, .external_lex_state = 2},
  [917] = {.lex_state = 10, .external_lex_state = 2},
  [918] = {.lex_state = 10},
  [919] = {.lex_state = 10, .external_lex_state = 2},
  [920] = {.lex_state = 10, .external_lex_state = 2},
  [921] = {.lex_state = 10, .external_lex_state = 2},
  [922] = {.lex_state = 10, .external_lex_state = 2},
  [923] = {.lex_state = 10, .external_lex_state = 2},
  [924] = {.lex_state = 10},
  [925] = {.lex_state = 10, .external_lex_state = 2},
  [926] = {.lex_state = 10, .external_lex_state = 2},
  [927] = {.lex_state = 10, .external_lex_state = 2},
  [928] = {.lex_state = 12, .external_lex_state = 2},
  [929] = {.lex_state = 10, .external_lex_state = 2},
  [930] = {.lex_state = 10, .external_lex_state = 2},
  [931] = {.lex_state = 10, .external_lex_state = 2},
  [932] = {.lex_state = 10, .external_lex_state = 2},
  [933] = {.lex_state = 10, .external_lex_state = 2},
  [934] = {.lex_state = 10, .external_lex_state = 2},
  [935] = {.lex_state = 10, .external_lex_state = 2},
  [936] = {.lex_state = 10, .external_lex_state = 2},
  [937] = {.lex_state = 10, .external_lex_state = 2},
  [938] = {.lex_state = 10, .external_lex_state = 2},
  [939] = {.lex_state = 10, .external_lex_state = 2},
  [940] = {.lex_state = 10, .external_lex_state = 2},
  [941] = {.lex_state = 10, .external_lex_state = 2},
  [942] = {.lex_state = 10, .external_lex_state = 2},
  [943] = {.lex_state = 10, .external_lex_state = 2},
  [944] = {.lex_state = 10, .external_lex_state = 2},
  [945] = {.lex_state = 10, .external_lex_state = 2},
  [946] = {.lex_state = 10, .external_lex_state = 2},
  [947] = {.lex_state = 10, .external_lex_state = 2},
  [948] = {.lex_state = 10, .external_lex_state = 2},
  [949] = {.lex_state = 10, .external_lex_state = 2},
  [950] = {.lex_state = 10, .external_lex_state = 2},
  [951] = {.lex_state = 10, .external_lex_state = 2},
  [952] = {.lex_state = 10, .external_lex_state = 2},
  [953] = {.lex_state = 10, .external_lex_state = 2},
  [954] = {.lex_state = 10, .external_lex_state = 2},
  [955] = {.lex_state = 10, .external_lex_state = 2},
  [956] = {.lex_state = 10, .external_lex_state = 2},
  [957] = {.lex_state = 10, .external_lex_state = 2},
  [958] = {.lex_state = 10, .external_lex_state = 2},
  [959] = {.lex_state = 10, .external_lex_state = 2},
  [960] = {.lex_state = 10, .external_lex_state = 2},
  [961] = {.lex_state = 10, .external_lex_state = 2},
  [962] = {.lex_state = 10, .external_lex_state = 2},
  [963] = {.lex_state = 12},
  [964] = {.lex_state = 10, .external_lex_state = 2},
  [965] = {.lex_state = 10, .external_lex_state = 2},
  [966] = {.lex_state = 10, .external_lex_state = 2},
  [967] = {.lex_state = 10, .external_lex_state = 2},
  [968] = {.lex_state = 10, .external_lex_state = 2},
  [969] = {.lex_state = 10, .external_lex_state = 2},
  [970] = {.lex_state = 10, .external_lex_state = 2},
  [971] = {.lex_state = 10, .external_lex_state = 2},
  [972] = {.lex_state = 10, .external_lex_state = 2},
  [973] = {.lex_state = 10, .external_lex_state = 2},
  [974] = {.lex_state = 10, .external_lex_state = 2},
  [975] = {.lex_state = 10, .external_lex_state = 2},
  [976] = {.lex_state = 12},
  [977] = {.lex_state = 12},
  [978] = {.lex_state = 12},
  [979] = {.lex_state = 10, .external_lex_state = 2},
  [980] = {.lex_state = 12},
  [981] = {.lex_state = 12},
  [982] = {.lex_state = 12},
  [983] = {.lex_state = 12},
  [984] = {.lex_state = 10, .external_lex_state = 2},
  [985] = {.lex_state = 12},
  [986] = {.lex_state = 12, .external_lex_state = 2},
  [987] = {.lex_state = 10, .external_lex_state = 2},
  [988] = {.lex_state = 12},
  [989] = {.lex_state = 12},
  [990] = {.lex_state = 10},
  [991] = {.lex_state = 12},
  [992] = {.lex_state = 12},
  [993] = {.lex_state = 12},
  [994] = {.lex_state = 12},
  [995] = {.lex_state = 10, .external_lex_state = 2},
  [996] = {.lex_state = 12},
  [997] = {.lex_state = 10, .external_lex_state = 2},
  [998] = {.lex_state = 10, .external_lex_state = 2},
  [999] = {.lex_state = 12},
  [1000] = {.lex_state = 12},
  [1001] = {.lex_state = 12},
  [1002] = {.lex_state = 12},
  [1003] = {.lex_state = 12},
  [1004] = {.lex_state = 12},
  [1005] = {.lex_state = 12},
  [1006] = {.lex_state = 12},
  [1007] = {.lex_state = 12},
  [1008] = {.lex_state = 12, .external_lex_state = 2},
  [1009] = {.lex_state = 12, .external_lex_state = 2},
  [1010] = {.lex_state = 12, .external_lex_state = 2},
  [1011] = {.lex_state = 12, .external_lex_state = 2},
  [1012] = {.lex_state = 12},
  [1013] = {.lex_state = 12},
  [1014] = {.lex_state = 12},
  [1015] = {.lex_state = 12},
  [1016] = {.lex_state = 12, .external_lex_state = 2},
  [1017] = {.lex_state = 12},
  [1018] = {.lex_state = 12},
  [1019] = {.lex_state = 12},
  [1020] = {.lex_state = 12},
  [1021] = {.lex_state = 12},
  [1022] = {.lex_state = 12},
  [1023] = {.lex_state = 12},
  [1024] = {.lex_state = 12},
  [1025] = {.lex_state = 12},
  [1026] = {.lex_state = 12},
  [1027] = {.lex_state = 12},
  [1028] = {.lex_state = 12},
  [1029] = {.lex_state = 12},
  [1030] = {.lex_state = 12},
  [1031] = {.lex_state = 12},
  [1032] = {.lex_state = 12},
  [1033] = {.lex_state = 12},
  [1034] = {.lex_state = 12},
  [1035] = {.lex_state = 12},
  [1036] = {.lex_state = 12},
  [1037] = {.lex_state = 12},
  [1038] = {.lex_state = 12},
  [1039] = {.lex_state = 12},
  [1040] = {.lex_state = 12},
  [1041] = {.lex_state = 12},
  [1042] = {.lex_state = 12},
  [1043] = {.lex_state = 12},
  [1044] = {.lex_state = 12},
  [1045] = {.lex_state = 12},
  [1046] = {.lex_state = 12, .external_lex_state = 2},
  [1047] = {.lex_state = 12, .external_lex_state = 2},
  [1048] = {.lex_state = 12, .external_lex_state = 2},
  [1049] = {.lex_state = 12, .external_lex_state = 2},
  [1050] = {.lex_state = 12, .external_lex_state = 2},
  [1051] = {.lex_state = 12, .external_lex_state = 2},
  [1052] = {.lex_state = 12, .external_lex_state = 2},
  [1053] = {.lex_state = 12, .external_lex_state = 2},
  [1054] = {.lex_state = 12, .external_lex_state = 2},
  [1055] = {.lex_state = 12},
  [1056] = {.lex_state = 12},
  [1057] = {.lex_state = 12, .external_lex_state = 2},
  [1058] = {.lex_state = 12},
  [1059] = {.lex_state = 12},
  [1060] = {.lex_state = 12},
  [1061] = {.lex_state = 12},
  [1062] = {.lex_state = 12},
  [1063] = {.lex_state = 12},
  [1064] = {.lex_state = 12},
  [1065] = {.lex_state = 12},
  [1066] = {.lex_state = 12},
  [1067] = {.lex_state = 12},
  [1068] = {.lex_state = 12},
  [1069] = {.lex_state = 12},
  [1070] = {.lex_state = 12},
  [1071] = {.lex_state = 12},
  [1072] = {.lex_state = 12},
  [1073] = {.lex_state = 12},
  [1074] = {.lex_state = 12},
  [1075] = {.lex_state = 12, .external_lex_state = 2},
  [1076] = {.lex_state = 12},
  [1077] = {.lex_state = 12},
  [1078] = {.lex_state = 12},
  [1079] = {.lex_state = 12, .external_lex_state = 2},
  [1080] = {.lex_state = 12},
  [1081] = {.lex_state = 12},
  [1082] = {.lex_state = 12},
  [1083] = {.lex_state = 12, .external_lex_state = 2},
  [1084] = {.lex_state = 18},
  [1085] = {.lex_state = 12, .external_lex_state = 2},
  [1086] = {.lex_state = 12, .external_lex_state = 2},
  [1087] = {.lex_state = 12, .external_lex_state = 2},
  [1088] = {.lex_state = 12, .external_lex_state = 2},
  [1089] = {.lex_state = 12, .external_lex_state = 2},
  [1090] = {.lex_state = 12, .external_lex_state = 2},
  [1091] = {.lex_state = 12, .external_lex_state = 2},
  [1092] = {.lex_state = 12, .external_lex_state = 2},
  [1093] = {.lex_state = 12, .external_lex_state = 2},
  [1094] = {.lex_state = 12, .external_lex_state = 2},
  [1095] = {.lex_state = 12, .external_lex_state = 2},
  [1096] = {.lex_state = 12, .external_lex_state = 2},
  [1097] = {.lex_state = 12, .external_lex_state = 2},
  [1098] = {.lex_state = 12, .external_lex_state = 2},
  [1099] = {.lex_state = 12, .external_lex_state = 2},
  [1100] = {.lex_state = 12, .external_lex_state = 2},
  [1101] = {.lex_state = 12, .external_lex_state = 2},
  [1102] = {.lex_state = 12, .external_lex_state = 2},
  [1103] = {.lex_state = 12, .external_lex_state = 2},
  [1104] = {.lex_state = 12, .external_lex_state = 2},
  [1105] = {.lex_state = 12, .external_lex_state = 2},
  [1106] = {.lex_state = 12, .external_lex_state = 2},
  [1107] = {.lex_state = 12, .external_lex_state = 2},
  [1108] = {.lex_state = 12, .external_lex_state = 2},
  [1109] = {.lex_state = 12, .external_lex_state = 2},
  [1110] = {.lex_state = 12, .external_lex_state = 2},
  [1111] = {.lex_state = 12, .external_lex_state = 2},
  [1112] = {.lex_state = 12, .external_lex_state = 2},
  [1113] = {.lex_state = 12, .external_lex_state = 2},
  [1114] = {.lex_state = 12, .external_lex_state = 2},
  [1115] = {.lex_state = 12, .external_lex_state = 2},
  [1116] = {.lex_state = 12, .external_lex_state = 2},
  [1117] = {.lex_state = 12, .external_lex_state = 2},
  [1118] = {.lex_state = 12, .external_lex_state = 2},
  [1119] = {.lex_state = 12, .external_lex_state = 2},
  [1120] = {.lex_state = 12, .external_lex_state = 2},
  [1121] = {.lex_state = 12, .external_lex_state = 2},
  [1122] = {.lex_state = 12},
  [1123] = {.lex_state = 12},
  [1124] = {.lex_state = 12},
  [1125] = {.lex_state = 12},
  [1126] = {.lex_state = 12},
  [1127] = {.lex_state = 12},
  [1128] = {.lex_state = 12},
  [1129] = {.lex_state = 12, .external_lex_state = 2},
  [1130] = {.lex_state = 12, .external_lex_state = 2},
  [1131] = {.lex_state = 12, .external_lex_state = 2},
  [1132] = {.lex_state = 12, .external_lex_state = 2},
  [1133] = {.lex_state = 12, .external_lex_state = 2},
  [1134] = {.lex_state = 12, .external_lex_state = 2},
  [1135] = {.lex_state = 12, .external_lex_state = 2},
  [1136] = {.lex_state = 12, .external_lex_state = 2},
  [1137] = {.lex_state = 12},
  [1138] = {.lex_state = 12},
  [1139] = {.lex_state = 12},
  [1140] = {.lex_state = 12},
  [1141] = {.lex_state = 12},
  [1142] = {.lex_state = 12},
  [1143] = {.lex_state = 12},
  [1144] = {.lex_state = 12},
  [1145] = {.lex_state = 12},
  [1146] = {.lex_state = 12},
  [1147] = {.lex_state = 12},
  [1148] = {.lex_state = 12},
  [1149] = {.lex_state = 12},
  [1150] = {.lex_state = 12},
  [1151] = {.lex_state = 12},
  [1152] = {.lex_state = 12},
  [1153] = {.lex_state = 12},
  [1154] = {.lex_state = 12},
  [1155] = {.lex_state = 12},
  [1156] = {.lex_state = 12},
  [1157] = {.lex_state = 12},
  [1158] = {.lex_state = 12},
  [1159] = {.lex_state = 12},
  [1160] = {.lex_state = 12},
  [1161] = {.lex_state = 12},
  [1162] = {.lex_state = 12},
  [1163] = {.lex_state = 12},
  [1164] = {.lex_state = 12},
  [1165] = {.lex_state = 12},
  [1166] = {.lex_state = 12},
  [1167] = {.lex_state = 12},
  [1168] = {.lex_state = 12},
  [1169] = {.lex_state = 12},
  [1170] = {.lex_state = 12},
  [1171] = {.lex_state = 12, .external_lex_state = 2},
  [1172] = {.lex_state = 12},
  [1173] = {.lex_state = 12},
  [1174] = {.lex_state = 12},
  [1175] = {.lex_state = 12},
  [1176] = {.lex_state = 12},
  [1177] = {.lex_state = 12},
  [1178] = {.lex_state = 12},
  [1179] = {.lex_state = 12},
  [1180] = {.lex_state = 12, .external_lex_state = 2},
  [1181] = {.lex_state = 12},
  [1182] = {.lex_state = 12},
  [1183] = {.lex_state = 12, .external_lex_state = 2},
  [1184] = {.lex_state = 12},
  [1185] = {.lex_state = 12},
  [1186] = {.lex_state = 12, .external_lex_state = 2},
  [1187] = {.lex_state = 12},
  [1188] = {.lex_state = 12, .external_lex_state = 2},
  [1189] = {.lex_state = 12, .external_lex_state = 2},
  [1190] = {.lex_state = 12},
  [1191] = {.lex_state = 12, .external_lex_state = 2},
  [1192] = {.lex_state = 12},
  [1193] = {.lex_state = 12, .external_lex_state = 2},
  [1194] = {.lex_state = 12, .external_lex_state = 2},
  [1195] = {.lex_state = 12, .external_lex_state = 2},
  [1196] = {.lex_state = 12, .external_lex_state = 2},
  [1197] = {.lex_state = 12, .external_lex_state = 2},
  [1198] = {.lex_state = 12, .external_lex_state = 2},
  [1199] = {.lex_state = 12},
  [1200] = {.lex_state = 12, .external_lex_state = 2},
  [1201] = {.lex_state = 12},
  [1202] = {.lex_state = 12, .external_lex_state = 2},
  [1203] = {.lex_state = 12, .external_lex_state = 2},
  [1204] = {.lex_state = 12},
  [1205] = {.lex_state = 12},
  [1206] = {.lex_state = 12},
  [1207] = {.lex_state = 12},
  [1208] = {.lex_state = 16},
  [1209] = {.lex_state = 12},
  [1210] = {.lex_state = 12},
  [1211] = {.lex_state = 16},
  [1212] = {.lex_state = 12},
  [1213] = {.lex_state = 12},
  [1214] = {.lex_state = 12},
  [1215] = {.lex_state = 12},
  [1216] = {.lex_state = 16},
  [1217] = {.lex_state = 12},
  [1218] = {.lex_state = 12},
  [1219] = {.lex_state = 12},
  [1220] = {.lex_state = 12},
  [1221] = {.lex_state = 12},
  [1222] = {.lex_state = 12},
  [1223] = {.lex_state = 16},
  [1224] = {.lex_state = 12},
  [1225] = {.lex_state = 12},
  [1226] = {.lex_state = 12},
  [1227] = {.lex_state = 12},
  [1228] = {.lex_state = 16},
  [1229] = {.lex_state = 12},
  [1230] = {.lex_state = 12},
  [1231] = {.lex_state = 16},
  [1232] = {.lex_state = 16},
  [1233] = {.lex_state = 17},
  [1234] = {.lex_state = 12},
  [1235] = {.lex_state = 12},
  [1236] = {.lex_state = 12},
  [1237] = {.lex_state = 12},
  [1238] = {.lex_state = 12},
  [1239] = {.lex_state = 12},
  [1240] = {.lex_state = 12},
  [1241] = {.lex_state = 12},
  [1242] = {.lex_state = 12},
  [1243] = {.lex_state = 12},
  [1244] = {.lex_state = 12},
  [1245] = {.lex_state = 12},
  [1246] = {.lex_state = 12},
  [1247] = {.lex_state = 12},
  [1248] = {.lex_state = 12},
  [1249] = {.lex_state = 12},
  [1250] = {.lex_state = 12},
  [1251] = {.lex_state = 12},
  [1252] = {.lex_state = 12},
  [1253] = {.lex_state = 12},
  [1254] = {.lex_state = 12},
  [1255] = {.lex_state = 12},
  [1256] = {.lex_state = 12},
  [1257] = {.lex_state = 12},
  [1258] = {.lex_state = 12},
  [1259] = {.lex_state = 12},
  [1260] = {.lex_state = 12},
  [1261] = {.lex_state = 12},
  [1262] = {.lex_state = 12},
  [1263] = {.lex_state = 12},
  [1264] = {.lex_state = 12},
  [1265] = {.lex_state = 12},
  [1266] = {.lex_state = 12},
  [1267] = {.lex_state = 12},
  [1268] = {.lex_state = 12},
  [1269] = {.lex_state = 12},
  [1270] = {.lex_state = 12},
  [1271] = {.lex_state = 12},
  [1272] = {.lex_state = 12},
  [1273] = {.lex_state = 12},
  [1274] = {.lex_state = 12},
  [1275] = {.lex_state = 12},
  [1276] = {.lex_state = 12},
  [1277] = {.lex_state = 12},
  [1278] = {.lex_state = 12},
  [1279] = {.lex_state = 12},
  [1280] = {.lex_state = 12},
  [1281] = {.lex_state = 12},
  [1282] = {.lex_state = 12},
  [1283] = {.lex_state = 12},
  [1284] = {.lex_state = 12},
  [1285] = {.lex_state = 12},
  [1286] = {.lex_state = 12},
  [1287] = {.lex_state = 12},
  [1288] = {.lex_state = 16},
  [1289] = {.lex_state = 12},
  [1290] = {.lex_state = 16},
  [1291] = {.lex_state = 12},
  [1292] = {.lex_state = 12},
  [1293] = {.lex_state = 16},
  [1294] = {.lex_state = 16},
  [1295] = {.lex_state = 16},
  [1296] = {.lex_state = 12},
  [1297] = {.lex_state = 12},
  [1298] = {.lex_state = 12},
  [1299] = {.lex_state = 16},
  [1300] = {.lex_state = 16},
  [1301] = {.lex_state = 16},
  [1302] = {.lex_state = 16},
  [1303] = {.lex_state = 16},
  [1304] = {.lex_state = 12},
  [1305] = {.lex_state = 12},
  [1306] = {.lex_state = 12},
  [1307] = {.lex_state = 65},
  [1308] = {.lex_state = 65},
  [1309] = {.lex_state = 12},
  [1310] = {.lex_state = 12},
  [1311] = {.lex_state = 65},
  [1312] = {.lex_state = 12},
  [1313] = {.lex_state = 12},
  [1314] = {.lex_state = 12},
  [1315] = {.lex_state = 12},
  [1316] = {.lex_state = 65},
  [1317] = {.lex_state = 12},
  [1318] = {.lex_state = 12},
  [1319] = {.lex_state = 12},
  [1320] = {.lex_state = 12},
  [1321] = {.lex_state = 12},
  [1322] = {.lex_state = 16},
  [1323] = {.lex_state = 12},
  [1324] = {.lex_state = 12},
  [1325] = {.lex_state = 16},
  [1326] = {.lex_state = 12},
  [1327] = {.lex_state = 65},
  [1328] = {.lex_state = 65},
  [1329] = {.lex_state = 16},
  [1330] = {.lex_state = 16},
  [1331] = {.lex_state = 65},
  [1332] = {.lex_state = 16},
  [1333] = {.lex_state = 65},
  [1334] = {.lex_state = 65},
  [1335] = {.lex_state = 65},
  [1336] = {.lex_state = 65},
  [1337] = {.lex_state = 65},
  [1338] = {.lex_state = 15},
  [1339] = {.lex_state = 65},
  [1340] = {.lex_state = 65},
  [1341] = {.lex_state = 65},
  [1342] = {.lex_state = 65},
  [1343] = {.lex_state = 65},
  [1344] = {.lex_state = 65},
  [1345] = {.lex_state = 65},
  [1346] = {.lex_state = 3, .external_lex_state = 3},
  [1347] = {.lex_state = 65},
  [1348] = {.lex_state = 65},
  [1349] = {.lex_state = 65},
  [1350] = {.lex_state = 65},
  [1351] = {.lex_state = 14},
  [1352] = {.lex_state = 3, .external_lex_state = 3},
  [1353] = {.lex_state = 65},
  [1354] = {.lex_state = 3, .external_lex_state = 4},
  [1355] = {.lex_state = 3, .external_lex_state = 3},
  [1356] = {.lex_state = 3, .external_lex_state = 3},
  [1357] = {.lex_state = 3, .external_lex_state = 3},
  [1358] = {.lex_state = 8, .external_lex_state = 5},
  [1359] = {.lex_state = 8, .external_lex_state = 6},
  [1360] = {.lex_state = 8, .external_lex_state = 5},
  [1361] = {.lex_state = 8, .external_lex_state = 6},
  [1362] = {.lex_state = 8, .external_lex_state = 5},
  [1363] = {.lex_state = 8, .external_lex_state = 6},
  [1364] = {.lex_state = 65},
  [1365] = {.lex_state = 8, .external_lex_state = 5},
  [1366] = {.lex_state = 16},
  [1367] = {.lex_state = 65},
  [1368] = {.lex_state = 16},
  [1369] = {.lex_state = 16},
  [1370] = {.lex_state = 16},
  [1371] = {.lex_state = 16},
  [1372] = {.lex_state = 8, .external_lex_state = 5},
  [1373] = {.lex_state = 8, .external_lex_state = 6},
  [1374] = {.lex_state = 16},
  [1375] = {.lex_state = 16},
  [1376] = {.lex_state = 16},
  [1377] = {.lex_state = 16},
  [1378] = {.lex_state = 8, .external_lex_state = 6},
  [1379] = {.lex_state = 16},
  [1380] = {.lex_state = 16},
  [1381] = {.lex_state = 16},
  [1382] = {.lex_state = 16},
  [1383] = {.lex_state = 16},
  [1384] = {.lex_state = 65},
  [1385] = {.lex_state = 65},
  [1386] = {.lex_state = 16},
  [1387] = {.lex_state = 65},
  [1388] = {.lex_state = 16},
  [1389] = {.lex_state = 65},
  [1390] = {.lex_state = 8, .external_lex_state = 4},
  [1391] = {.lex_state = 16},
  [1392] = {.lex_state = 18},
  [1393] = {.lex_state = 16},
  [1394] = {.lex_state = 16},
  [1395] = {.lex_state = 16},
  [1396] = {.lex_state = 16},
  [1397] = {.lex_state = 16},
  [1398] = {.lex_state = 16},
  [1399] = {.lex_state = 16},
  [1400] = {.lex_state = 16},
  [1401] = {.lex_state = 16},
  [1402] = {.lex_state = 16},
  [1403] = {.lex_state = 16},
  [1404] = {.lex_state = 16},
  [1405] = {.lex_state = 16},
  [1406] = {.lex_state = 16},
  [1407] = {.lex_state = 65},
  [1408] = {.lex_state = 65},
  [1409] = {.lex_state = 65},
  [1410] = {.lex_state = 65},
  [1411] = {.lex_state = 16},
  [1412] = {.lex_state = 65},
  [1413] = {.lex_state = 65},
  [1414] = {.lex_state = 65},
  [1415] = {.lex_state = 65},
  [1416] = {.lex_state = 65},
  [1417] = {.lex_state = 65},
  [1418] = {.lex_state = 65},
  [1419] = {.lex_state = 65},
  [1420] = {.lex_state = 65},
  [1421] = {.lex_state = 65},
  [1422] = {.lex_state = 65},
  [1423] = {.lex_state = 65},
  [1424] = {.lex_state = 65},
  [1425] = {.lex_state = 65},
  [1426] = {.lex_state = 65},
  [1427] = {.lex_state = 65},
  [1428] = {.lex_state = 65},
  [1429] = {.lex_state = 65},
  [1430] = {.lex_state = 16},
  [1431] = {.lex_state = 65},
  [1432] = {.lex_state = 65},
  [1433] = {.lex_state = 16},
  [1434] = {.lex_state = 65},
  [1435] = {.lex_state = 16},
  [1436] = {.lex_state = 16},
  [1437] = {.lex_state = 65},
  [1438] = {.lex_state = 65},
  [1439] = {.lex_state = 65},
  [1440] = {.lex_state = 65},
  [1441] = {.lex_state = 65},
  [1442] = {.lex_state = 65},
  [1443] = {.lex_state = 65},
  [1444] = {.lex_state = 65},
  [1445] = {.lex_state = 65},
  [1446] = {.lex_state = 18},
  [1447] = {.lex_state = 65},
  [1448] = {.lex_state = 18},
  [1449] = {.lex_state = 65},
  [1450] = {.lex_state = 18},
  [1451] = {.lex_state = 65},
  [1452] = {.lex_state = 65},
  [1453] = {.lex_state = 18},
  [1454] = {.lex_state = 65},
  [1455] = {.lex_state = 18},
  [1456] = {.lex_state = 65},
  [1457] = {.lex_state = 3, .external_lex_state = 7},
  [1458] = {.lex_state = 65},
  [1459] = {.lex_state = 65},
  [1460] = {.lex_state = 65},
  [1461] = {.lex_state = 65},
  [1462] = {.lex_state = 65},
  [1463] = {.lex_state = 65},
  [1464] = {.lex_state = 65},
  [1465] = {.lex_state = 65},
  [1466] = {.lex_state = 65},
  [1467] = {.lex_state = 18},
  [1468] = {.lex_state = 18},
  [1469] = {.lex_state = 65},
  [1470] = {.lex_state = 65},
  [1471] = {.lex_state = 3, .external_lex_state = 7},
  [1472] = {.lex_state = 65},
  [1473] = {.lex_state = 65},
  [1474] = {.lex_state = 65},
  [1475] = {.lex_state = 65},
  [1476] = {.lex_state = 18},
  [1477] = {.lex_state = 65},
  [1478] = {.lex_state = 65, .external_lex_state = 2},
  [1479] = {.lex_state = 8, .external_lex_state = 8},
  [1480] = {.lex_state = 65, .external_lex_state = 2},
  [1481] = {.lex_state = 65, .external_lex_state = 2},
  [1482] = {.lex_state = 65, .external_lex_state = 2},
  [1483] = {.lex_state = 8, .external_lex_state = 9},
  [1484] = {.lex_state = 8, .external_lex_state = 8},
  [1485] = {.lex_state = 65, .external_lex_state = 2},
  [1486] = {.lex_state = 65, .external_lex_state = 2},
  [1487] = {.lex_state = 18},
  [1488] = {.lex_state = 65, .external_lex_state = 2},
  [1489] = {.lex_state = 65, .external_lex_state = 2},
  [1490] = {.lex_state = 65},
  [1491] = {.lex_state = 8, .external_lex_state = 9},
  [1492] = {.lex_state = 65},
  [1493] = {.lex_state = 65},
  [1494] = {.lex_state = 65},
  [1495] = {.lex_state = 65, .external_lex_state = 2},
  [1496] = {.lex_state = 65},
  [1497] = {.lex_state = 65},
  [1498] = {.lex_state = 65},
  [1499] = {.lex_state = 65},
  [1500] = {.lex_state = 65},
  [1501] = {.lex_state = 65},
  [1502] = {.lex_state = 65},
  [1503] = {.lex_state = 65},
  [1504] = {.lex_state = 65},
  [1505] = {.lex_state = 65},
  [1506] = {.lex_state = 65},
  [1507] = {.lex_state = 65},
  [1508] = {.lex_state = 65},
  [1509] = {.lex_state = 65},
  [1510] = {.lex_state = 65},
  [1511] = {.lex_state = 65},
  [1512] = {.lex_state = 65, .external_lex_state = 2},
  [1513] = {.lex_state = 65, .external_lex_state = 2},
  [1514] = {.lex_state = 65},
  [1515] = {.lex_state = 65},
  [1516] = {.lex_state = 65},
  [1517] = {.lex_state = 10},
  [1518] = {.lex_state = 65, .external_lex_state = 2},
  [1519] = {.lex_state = 65},
  [1520] = {.lex_state = 65, .external_lex_state = 2},
  [1521] = {.lex_state = 65},
  [1522] = {.lex_state = 65},
  [1523] = {.lex_state = 10},
  [1524] = {.lex_state = 65, .external_lex_state = 2},
  [1525] = {.lex_state = 65, .external_lex_state = 2},
  [1526] = {.lex_state = 3, .external_lex_state = 3},
  [1527] = {.lex_state = 3, .external_lex_state = 3},
  [1528] = {.lex_state = 65, .external_lex_state = 2},
  [1529] = {.lex_state = 65},
  [1530] = {.lex_state = 14},
  [1531] = {.lex_state = 65, .external_lex_state = 2},
  [1532] = {.lex_state = 3, .external_lex_state = 3},
  [1533] = {.lex_state = 65},
  [1534] = {.lex_state = 65},
  [1535] = {.lex_state = 65},
  [1536] = {.lex_state = 65, .external_lex_state = 2},
  [1537] = {.lex_state = 65, .external_lex_state = 2},
  [1538] = {.lex_state = 3, .external_lex_state = 3},
  [1539] = {.lex_state = 65},
  [1540] = {.lex_state = 3, .external_lex_state = 3},
  [1541] = {.lex_state = 65},
  [1542] = {.lex_state = 3, .external_lex_state = 3},
  [1543] = {.lex_state = 3, .external_lex_state = 3},
  [1544] = {.lex_state = 65, .external_lex_state = 2},
  [1545] = {.lex_state = 65, .external_lex_state = 2},
  [1546] = {.lex_state = 65, .external_lex_state = 2},
  [1547] = {.lex_state = 65, .external_lex_state = 2},
  [1548] = {.lex_state = 65, .external_lex_state = 2},
  [1549] = {.lex_state = 65},
  [1550] = {.lex_state = 65},
  [1551] = {.lex_state = 65},
  [1552] = {.lex_state = 65},
  [1553] = {.lex_state = 65},
  [1554] = {.lex_state = 3, .external_lex_state = 3},
  [1555] = {.lex_state = 65, .external_lex_state = 2},
  [1556] = {.lex_state = 18},
  [1557] = {.lex_state = 65, .external_lex_state = 2},
  [1558] = {.lex_state = 10},
  [1559] = {.lex_state = 65, .external_lex_state = 2},
  [1560] = {.lex_state = 65, .external_lex_state = 2},
  [1561] = {.lex_state = 65},
  [1562] = {.lex_state = 65},
  [1563] = {.lex_state = 65},
  [1564] = {.lex_state = 65},
  [1565] = {.lex_state = 65},
  [1566] = {.lex_state = 8, .external_lex_state = 5},
  [1567] = {.lex_state = 8, .external_lex_state = 6},
  [1568] = {.lex_state = 8, .external_lex_state = 6},
  [1569] = {.lex_state = 8, .external_lex_state = 5},
  [1570] = {.lex_state = 8, .external_lex_state = 6},
  [1571] = {.lex_state = 65},
  [1572] = {.lex_state = 65},
  [1573] = {.lex_state = 14},
  [1574] = {.lex_state = 65},
  [1575] = {.lex_state = 65},
  [1576] = {.lex_state = 65},
  [1577] = {.lex_state = 65},
  [1578] = {.lex_state = 8, .external_lex_state = 5},
  [1579] = {.lex_state = 65},
  [1580] = {.lex_state = 8, .external_lex_state = 5},
  [1581] = {.lex_state = 8, .external_lex_state = 6},
  [1582] = {.lex_state = 20},
  [1583] = {.lex_state = 8, .external_lex_state = 6},
  [1584] = {.lex_state = 65},
  [1585] = {.lex_state = 65},
  [1586] = {.lex_state = 8, .external_lex_state = 5},
  [1587] = {.lex_state = 8, .external_lex_state = 6},
  [1588] = {.lex_state = 65},
  [1589] = {.lex_state = 20},
  [1590] = {.lex_state = 65},
  [1591] = {.lex_state = 65},
  [1592] = {.lex_state = 8, .external_lex_state = 5},
  [1593] = {.lex_state = 65},
  [1594] = {.lex_state = 65},
  [1595] = {.lex_state = 8, .external_lex_state = 6},
  [1596] = {.lex_state = 65},
  [1597] = {.lex_state = 20},
  [1598] = {.lex_state = 65},
  [1599] = {.lex_state = 65},
  [1600] = {.lex_state = 65},
  [1601] = {.lex_state = 65},
  [1602] = {.lex_state = 65},
  [1603] = {.lex_state = 65},
  [1604] = {.lex_state = 65},
  [1605] = {.lex_state = 65},
  [1606] = {.lex_state = 65},
  [1607] = {.lex_state = 65},
  [1608] = {.lex_state = 65},
  [1609] = {.lex_state = 65},
  [1610] = {.lex_state = 65},
  [1611] = {.lex_state = 65},
  [1612] = {.lex_state = 65},
  [1613] = {.lex_state = 65},
  [1614] = {.lex_state = 65},
  [1615] = {.lex_state = 65},
  [1616] = {.lex_state = 65},
  [1617] = {.lex_state = 65},
  [1618] = {.lex_state = 65},
  [1619] = {.lex_state = 65},
  [1620] = {.lex_state = 65},
  [1621] = {.lex_state = 65},
  [1622] = {.lex_state = 65},
  [1623] = {.lex_state = 65},
  [1624] = {.lex_state = 65},
  [1625] = {.lex_state = 65},
  [1626] = {.lex_state = 65},
  [1627] = {.lex_state = 65},
  [1628] = {.lex_state = 65},
  [1629] = {.lex_state = 65},
  [1630] = {.lex_state = 65},
  [1631] = {.lex_state = 65},
  [1632] = {.lex_state = 65},
  [1633] = {.lex_state = 65},
  [1634] = {.lex_state = 65},
  [1635] = {.lex_state = 65},
  [1636] = {.lex_state = 65},
  [1637] = {.lex_state = 65},
  [1638] = {.lex_state = 8, .external_lex_state = 5},
  [1639] = {.lex_state = 20},
  [1640] = {.lex_state = 65},
  [1641] = {.lex_state = 65},
  [1642] = {.lex_state = 8, .external_lex_state = 6},
  [1643] = {.lex_state = 65},
  [1644] = {.lex_state = 8, .external_lex_state = 5},
  [1645] = {.lex_state = 20},
  [1646] = {.lex_state = 65},
  [1647] = {.lex_state = 65},
  [1648] = {.lex_state = 65, .external_lex_state = 2},
  [1649] = {.lex_state = 65, .external_lex_state = 2},
  [1650] = {.lex_state = 65, .external_lex_state = 2},
  [1651] = {.lex_state = 65, .external_lex_state = 2},
  [1652] = {.lex_state = 65},
  [1653] = {.lex_state = 65, .external_lex_state = 2},
  [1654] = {.lex_state = 65, .external_lex_state = 2},
  [1655] = {.lex_state = 65},
  [1656] = {.lex_state = 14},
  [1657] = {.lex_state = 65},
  [1658] = {.lex_state = 65, .external_lex_state = 2},
  [1659] = {.lex_state = 65, .external_lex_state = 2},
  [1660] = {.lex_state = 65, .external_lex_state = 2},
  [1661] = {.lex_state = 65, .external_lex_state = 2},
  [1662] = {.lex_state = 65, .external_lex_state = 2},
  [1663] = {.lex_state = 65},
  [1664] = {.lex_state = 65},
  [1665] = {.lex_state = 65},
  [1666] = {.lex_state = 65},
  [1667] = {.lex_state = 16},
  [1668] = {.lex_state = 65},
  [1669] = {.lex_state = 65},
  [1670] = {.lex_state = 65},
  [1671] = {.lex_state = 65},
  [1672] = {.lex_state = 65, .external_lex_state = 2},
  [1673] = {.lex_state = 65, .external_lex_state = 2},
  [1674] = {.lex_state = 65, .external_lex_state = 2},
  [1675] = {.lex_state = 16},
  [1676] = {.lex_state = 65},
  [1677] = {.lex_state = 65, .external_lex_state = 2},
  [1678] = {.lex_state = 65},
  [1679] = {.lex_state = 65},
  [1680] = {.lex_state = 65, .external_lex_state = 2},
  [1681] = {.lex_state = 16},
  [1682] = {.lex_state = 16},
  [1683] = {.lex_state = 16},
  [1684] = {.lex_state = 14},
  [1685] = {.lex_state = 65},
  [1686] = {.lex_state = 65, .external_lex_state = 2},
  [1687] = {.lex_state = 65},
  [1688] = {.lex_state = 65},
  [1689] = {.lex_state = 65, .external_lex_state = 2},
  [1690] = {.lex_state = 65, .external_lex_state = 2},
  [1691] = {.lex_state = 16},
  [1692] = {.lex_state = 65, .external_lex_state = 2},
  [1693] = {.lex_state = 65, .external_lex_state = 2},
  [1694] = {.lex_state = 65},
  [1695] = {.lex_state = 65, .external_lex_state = 2},
  [1696] = {.lex_state = 65, .external_lex_state = 2},
  [1697] = {.lex_state = 65, .external_lex_state = 2},
  [1698] = {.lex_state = 65, .external_lex_state = 2},
  [1699] = {.lex_state = 65},
  [1700] = {.lex_state = 65},
  [1701] = {.lex_state = 65, .external_lex_state = 2},
  [1702] = {.lex_state = 65, .external_lex_state = 2},
  [1703] = {.lex_state = 65, .external_lex_state = 2},
  [1704] = {.lex_state = 16},
  [1705] = {.lex_state = 65, .external_lex_state = 2},
  [1706] = {.lex_state = 65, .external_lex_state = 2},
  [1707] = {.lex_state = 65, .external_lex_state = 2},
  [1708] = {.lex_state = 16},
  [1709] = {.lex_state = 65, .external_lex_state = 2},
  [1710] = {.lex_state = 65, .external_lex_state = 2},
  [1711] = {.lex_state = 65},
  [1712] = {.lex_state = 65, .external_lex_state = 2},
  [1713] = {.lex_state = 65, .external_lex_state = 2},
  [1714] = {.lex_state = 65, .external_lex_state = 2},
  [1715] = {.lex_state = 65, .external_lex_state = 2},
  [1716] = {.lex_state = 65},
  [1717] = {.lex_state = 65, .external_lex_state = 2},
  [1718] = {.lex_state = 65},
  [1719] = {.lex_state = 65, .external_lex_state = 2},
  [1720] = {.lex_state = 65, .external_lex_state = 2},
  [1721] = {.lex_state = 16},
  [1722] = {.lex_state = 65, .external_lex_state = 2},
  [1723] = {.lex_state = 16},
  [1724] = {.lex_state = 14},
  [1725] = {.lex_state = 14},
  [1726] = {.lex_state = 14},
  [1727] = {.lex_state = 65, .external_lex_state = 2},
  [1728] = {.lex_state = 65},
  [1729] = {.lex_state = 65},
  [1730] = {.lex_state = 65},
  [1731] = {.lex_state = 65},
  [1732] = {.lex_state = 16},
  [1733] = {.lex_state = 65},
  [1734] = {.lex_state = 65, .external_lex_state = 2},
  [1735] = {.lex_state = 65},
  [1736] = {.lex_state = 16},
  [1737] = {.lex_state = 16},
  [1738] = {.lex_state = 16},
  [1739] = {.lex_state = 65},
  [1740] = {.lex_state = 65},
  [1741] = {.lex_state = 65, .external_lex_state = 2},
  [1742] = {.lex_state = 65},
  [1743] = {.lex_state = 16},
  [1744] = {.lex_state = 65, .external_lex_state = 2},
  [1745] = {.lex_state = 65, .external_lex_state = 2},
  [1746] = {.lex_state = 65, .external_lex_state = 2},
  [1747] = {.lex_state = 65, .external_lex_state = 2},
  [1748] = {.lex_state = 65},
  [1749] = {.lex_state = 65},
  [1750] = {.lex_state = 65},
  [1751] = {.lex_state = 65, .external_lex_state = 2},
  [1752] = {.lex_state = 65},
  [1753] = {.lex_state = 65, .external_lex_state = 2},
  [1754] = {.lex_state = 65, .external_lex_state = 2},
  [1755] = {.lex_state = 65, .external_lex_state = 2},
  [1756] = {.lex_state = 14},
  [1757] = {.lex_state = 14},
  [1758] = {.lex_state = 65, .external_lex_state = 2},
  [1759] = {.lex_state = 14},
  [1760] = {.lex_state = 65},
  [1761] = {.lex_state = 16},
  [1762] = {.lex_state = 16},
  [1763] = {.lex_state = 16},
  [1764] = {.lex_state = 65, .external_lex_state = 2},
  [1765] = {.lex_state = 16},
  [1766] = {.lex_state = 16},
  [1767] = {.lex_state = 65, .external_lex_state = 2},
  [1768] = {.lex_state = 65, .external_lex_state = 2},
  [1769] = {.lex_state = 65},
  [1770] = {.lex_state = 65, .external_lex_state = 2},
  [1771] = {.lex_state = 65, .external_lex_state = 2},
  [1772] = {.lex_state = 65, .external_lex_state = 2},
  [1773] = {.lex_state = 65},
  [1774] = {.lex_state = 65},
  [1775] = {.lex_state = 65},
  [1776] = {.lex_state = 65, .external_lex_state = 2},
  [1777] = {.lex_state = 65},
  [1778] = {.lex_state = 65},
  [1779] = {.lex_state = 65},
  [1780] = {.lex_state = 65},
  [1781] = {.lex_state = 65},
  [1782] = {.lex_state = 65},
  [1783] = {.lex_state = 65, .external_lex_state = 2},
  [1784] = {.lex_state = 65},
  [1785] = {.lex_state = 14},
  [1786] = {.lex_state = 65},
  [1787] = {.lex_state = 65},
  [1788] = {.lex_state = 65, .external_lex_state = 2},
  [1789] = {.lex_state = 65, .external_lex_state = 2},
  [1790] = {.lex_state = 65, .external_lex_state = 2},
  [1791] = {.lex_state = 65, .external_lex_state = 2},
  [1792] = {.lex_state = 65},
  [1793] = {.lex_state = 65},
  [1794] = {.lex_state = 65},
  [1795] = {.lex_state = 65},
  [1796] = {.lex_state = 65},
  [1797] = {.lex_state = 65},
  [1798] = {.lex_state = 65},
  [1799] = {.lex_state = 65, .external_lex_state = 2},
  [1800] = {.lex_state = 65, .external_lex_state = 2},
  [1801] = {.lex_state = 65},
  [1802] = {.lex_state = 65},
  [1803] = {.lex_state = 65},
  [1804] = {.lex_state = 65, .external_lex_state = 2},
  [1805] = {.lex_state = 65},
  [1806] = {.lex_state = 65},
  [1807] = {.lex_state = 3, .external_lex_state = 10},
  [1808] = {.lex_state = 65},
  [1809] = {.lex_state = 65},
  [1810] = {.lex_state = 65},
  [1811] = {.lex_state = 65},
  [1812] = {.lex_state = 20},
  [1813] = {.lex_state = 65},
  [1814] = {.lex_state = 3, .external_lex_state = 11},
  [1815] = {.lex_state = 65},
  [1816] = {.lex_state = 65, .external_lex_state = 2},
  [1817] = {.lex_state = 65},
  [1818] = {.lex_state = 17, .external_lex_state = 2},
  [1819] = {.lex_state = 65},
  [1820] = {.lex_state = 65},
  [1821] = {.lex_state = 65},
  [1822] = {.lex_state = 3, .external_lex_state = 11},
  [1823] = {.lex_state = 65},
  [1824] = {.lex_state = 65},
  [1825] = {.lex_state = 65},
  [1826] = {.lex_state = 65},
  [1827] = {.lex_state = 65},
  [1828] = {.lex_state = 65},
  [1829] = {.lex_state = 65},
  [1830] = {.lex_state = 65},
  [1831] = {.lex_state = 65, .external_lex_state = 2},
  [1832] = {.lex_state = 65},
  [1833] = {.lex_state = 65},
  [1834] = {.lex_state = 65},
  [1835] = {.lex_state = 65, .external_lex_state = 2},
  [1836] = {.lex_state = 65},
  [1837] = {.lex_state = 65},
  [1838] = {.lex_state = 65},
  [1839] = {.lex_state = 65},
  [1840] = {.lex_state = 65},
  [1841] = {.lex_state = 65},
  [1842] = {.lex_state = 65},
  [1843] = {.lex_state = 65},
  [1844] = {.lex_state = 65, .external_lex_state = 2},
  [1845] = {.lex_state = 65},
  [1846] = {.lex_state = 65, .external_lex_state = 2},
  [1847] = {.lex_state = 3, .external_lex_state = 10},
  [1848] = {.lex_state = 65},
  [1849] = {.lex_state = 65},
  [1850] = {.lex_state = 65},
  [1851] = {.lex_state = 14},
  [1852] = {.lex_state = 65},
  [1853] = {.lex_state = 3, .external_lex_state = 11},
  [1854] = {.lex_state = 65},
  [1855] = {.lex_state = 65},
  [1856] = {.lex_state = 65},
  [1857] = {.lex_state = 65},
  [1858] = {.lex_state = 14},
  [1859] = {.lex_state = 65},
  [1860] = {.lex_state = 65},
  [1861] = {.lex_state = 65, .external_lex_state = 2},
  [1862] = {.lex_state = 65},
  [1863] = {.lex_state = 65, .external_lex_state = 2},
  [1864] = {.lex_state = 65},
  [1865] = {.lex_state = 65},
  [1866] = {.lex_state = 20},
  [1867] = {.lex_state = 3, .external_lex_state = 11},
  [1868] = {.lex_state = 65},
  [1869] = {.lex_state = 65},
  [1870] = {.lex_state = 65},
  [1871] = {.lex_state = 65, .external_lex_state = 2},
  [1872] = {.lex_state = 65, .external_lex_state = 2},
  [1873] = {.lex_state = 65},
  [1874] = {.lex_state = 65},
  [1875] = {.lex_state = 65},
  [1876] = {.lex_state = 65, .external_lex_state = 2},
  [1877] = {.lex_state = 65},
  [1878] = {.lex_state = 65, .external_lex_state = 2},
  [1879] = {.lex_state = 65},
  [1880] = {.lex_state = 65},
  [1881] = {.lex_state = 65},
  [1882] = {.lex_state = 65},
  [1883] = {.lex_state = 65},
  [1884] = {.lex_state = 65},
  [1885] = {.lex_state = 65},
  [1886] = {.lex_state = 65},
  [1887] = {.lex_state = 65},
  [1888] = {.lex_state = 65},
  [1889] = {.lex_state = 65},
  [1890] = {.lex_state = 65},
  [1891] = {.lex_state = 65},
  [1892] = {.lex_state = 65},
  [1893] = {.lex_state = 65, .external_lex_state = 2},
  [1894] = {.lex_state = 65},
  [1895] = {.lex_state = 65},
  [1896] = {.lex_state = 65},
  [1897] = {.lex_state = 65},
  [1898] = {.lex_state = 65},
  [1899] = {.lex_state = 3, .external_lex_state = 11},
  [1900] = {.lex_state = 65, .external_lex_state = 2},
  [1901] = {.lex_state = 65, .external_lex_state = 2},
  [1902] = {.lex_state = 65},
  [1903] = {.lex_state = 65, .external_lex_state = 2},
  [1904] = {.lex_state = 65},
  [1905] = {.lex_state = 65},
  [1906] = {.lex_state = 65, .external_lex_state = 2},
  [1907] = {.lex_state = 65},
  [1908] = {.lex_state = 65},
  [1909] = {.lex_state = 65},
  [1910] = {.lex_state = 65},
  [1911] = {.lex_state = 65},
  [1912] = {.lex_state = 65},
  [1913] = {.lex_state = 65},
  [1914] = {.lex_state = 65},
  [1915] = {.lex_state = 65},
  [1916] = {.lex_state = 65},
  [1917] = {.lex_state = 65, .external_lex_state = 2},
  [1918] = {.lex_state = 65},
  [1919] = {.lex_state = 65},
  [1920] = {.lex_state = 65, .external_lex_state = 2},
  [1921] = {.lex_state = 65},
  [1922] = {.lex_state = 65, .external_lex_state = 2},
  [1923] = {.lex_state = 3, .external_lex_state = 11},
  [1924] = {.lex_state = 3, .external_lex_state = 11},
  [1925] = {.lex_state = 65},
  [1926] = {.lex_state = 65, .external_lex_state = 2},
  [1927] = {.lex_state = 65, .external_lex_state = 2},
  [1928] = {.lex_state = 65},
  [1929] = {.lex_state = 65, .external_lex_state = 2},
  [1930] = {.lex_state = 65, .external_lex_state = 2},
  [1931] = {.lex_state = 65, .external_lex_state = 2},
  [1932] = {.lex_state = 65},
  [1933] = {.lex_state = 65, .external_lex_state = 2},
  [1934] = {.lex_state = 65},
  [1935] = {.lex_state = 65},
  [1936] = {.lex_state = 65},
  [1937] = {.lex_state = 65},
  [1938] = {.lex_state = 65},
  [1939] = {.lex_state = 65},
  [1940] = {.lex_state = 65},
  [1941] = {.lex_state = 65},
  [1942] = {.lex_state = 65, .external_lex_state = 2},
  [1943] = {.lex_state = 65},
  [1944] = {.lex_state = 65},
  [1945] = {.lex_state = 65},
  [1946] = {.lex_state = 65},
  [1947] = {.lex_state = 65, .external_lex_state = 2},
  [1948] = {.lex_state = 65},
  [1949] = {.lex_state = 65},
  [1950] = {.lex_state = 65, .external_lex_state = 2},
  [1951] = {.lex_state = 65, .external_lex_state = 2},
  [1952] = {.lex_state = 65},
  [1953] = {.lex_state = 65},
  [1954] = {.lex_state = 65},
  [1955] = {.lex_state = 65},
  [1956] = {.lex_state = 65},
  [1957] = {.lex_state = 65, .external_lex_state = 2},
  [1958] = {.lex_state = 65},
  [1959] = {.lex_state = 65, .external_lex_state = 2},
  [1960] = {.lex_state = 65},
  [1961] = {.lex_state = 65, .external_lex_state = 2},
  [1962] = {.lex_state = 65},
  [1963] = {.lex_state = 65},
  [1964] = {.lex_state = 65},
  [1965] = {.lex_state = 65},
  [1966] = {.lex_state = 65},
  [1967] = {.lex_state = 65},
  [1968] = {.lex_state = 65, .external_lex_state = 2},
  [1969] = {.lex_state = 65},
  [1970] = {.lex_state = 3, .external_lex_state = 11},
  [1971] = {.lex_state = 65},
  [1972] = {.lex_state = 65},
  [1973] = {.lex_state = 14},
  [1974] = {.lex_state = 65, .external_lex_state = 2},
  [1975] = {.lex_state = 65},
  [1976] = {.lex_state = 65},
  [1977] = {.lex_state = 65},
  [1978] = {.lex_state = 65},
  [1979] = {.lex_state = 65},
  [1980] = {.lex_state = 65, .external_lex_state = 2},
  [1981] = {.lex_state = 65},
  [1982] = {.lex_state = 65},
  [1983] = {.lex_state = 65},
  [1984] = {.lex_state = 65},
  [1985] = {.lex_state = 65},
  [1986] = {.lex_state = 65},
  [1987] = {.lex_state = 65},
  [1988] = {.lex_state = 65},
  [1989] = {.lex_state = 65},
  [1990] = {.lex_state = 65, .external_lex_state = 2},
  [1991] = {.lex_state = 65, .external_lex_state = 2},
  [1992] = {.lex_state = 65},
  [1993] = {.lex_state = 65},
  [1994] = {.lex_state = 65},
  [1995] = {.lex_state = 14},
  [1996] = {.lex_state = 65, .external_lex_state = 2},
  [1997] = {.lex_state = 65},
  [1998] = {.lex_state = 65},
  [1999] = {.lex_state = 65},
  [2000] = {.lex_state = 65},
  [2001] = {.lex_state = 65},
  [2002] = {.lex_state = 65},
  [2003] = {.lex_state = 14},
  [2004] = {.lex_state = 65},
  [2005] = {.lex_state = 65},
  [2006] = {.lex_state = 65},
  [2007] = {.lex_state = 65},
  [2008] = {.lex_state = 65},
  [2009] = {.lex_state = 65},
  [2010] = {.lex_state = 65},
  [2011] = {.lex_state = 65},
  [2012] = {.lex_state = 65},
  [2013] = {.lex_state = 65},
  [2014] = {.lex_state = 65},
  [2015] = {.lex_state = 65, .external_lex_state = 2},
  [2016] = {.lex_state = 65},
  [2017] = {.lex_state = 65},
  [2018] = {.lex_state = 65},
  [2019] = {.lex_state = 65},
  [2020] = {.lex_state = 65},
  [2021] = {.lex_state = 65},
  [2022] = {.lex_state = 65, .external_lex_state = 2},
  [2023] = {.lex_state = 65},
  [2024] = {.lex_state = 65, .external_lex_state = 2},
  [2025] = {.lex_state = 65},
  [2026] = {.lex_state = 65, .external_lex_state = 2},
  [2027] = {.lex_state = 65},
  [2028] = {.lex_state = 65},
  [2029] = {.lex_state = 65, .external_lex_state = 2},
  [2030] = {.lex_state = 65, .external_lex_state = 2},
  [2031] = {.lex_state = 65, .external_lex_state = 2},
  [2032] = {.lex_state = 65},
  [2033] = {.lex_state = 65},
  [2034] = {.lex_state = 65, .external_lex_state = 12},
  [2035] = {.lex_state = 65},
  [2036] = {.lex_state = 65},
  [2037] = {.lex_state = 65},
  [2038] = {.lex_state = 65},
  [2039] = {.lex_state = 65},
  [2040] = {.lex_state = 65, .external_lex_state = 2},
  [2041] = {.lex_state = 65},
  [2042] = {.lex_state = 65},
  [2043] = {.lex_state = 65},
  [2044] = {.lex_state = 65},
  [2045] = {.lex_state = 65},
  [2046] = {.lex_state = 65, .external_lex_state = 2},
  [2047] = {.lex_state = 65, .external_lex_state = 2},
  [2048] = {.lex_state = 65, .external_lex_state = 2},
  [2049] = {.lex_state = 65, .external_lex_state = 2},
  [2050] = {.lex_state = 65},
  [2051] = {.lex_state = 65},
  [2052] = {.lex_state = 65, .external_lex_state = 2},
  [2053] = {.lex_state = 65},
  [2054] = {.lex_state = 65},
  [2055] = {.lex_state = 65, .external_lex_state = 2},
  [2056] = {.lex_state = 65},
  [2057] = {.lex_state = 14},
  [2058] = {.lex_state = 65},
  [2059] = {.lex_state = 65, .external_lex_state = 2},
  [2060] = {.lex_state = 65},
  [2061] = {.lex_state = 65, .external_lex_state = 2},
  [2062] = {.lex_state = 65},
  [2063] = {.lex_state = 65, .external_lex_state = 2},
  [2064] = {.lex_state = 65, .external_lex_state = 2},
  [2065] = {.lex_state = 65},
  [2066] = {.lex_state = 65},
  [2067] = {.lex_state = 65},
  [2068] = {.lex_state = 65},
  [2069] = {.lex_state = 65},
  [2070] = {.lex_state = 65},
  [2071] = {.lex_state = 65},
  [2072] = {.lex_state = 65},
  [2073] = {.lex_state = 65, .external_lex_state = 2},
  [2074] = {.lex_state = 65},
  [2075] = {.lex_state = 65},
  [2076] = {.lex_state = 65, .external_lex_state = 2},
  [2077] = {.lex_state = 65},
  [2078] = {.lex_state = 65},
  [2079] = {.lex_state = 65},
  [2080] = {.lex_state = 65},
  [2081] = {.lex_state = 65},
  [2082] = {.lex_state = 65},
  [2083] = {.lex_state = 65},
  [2084] = {.lex_state = 65},
  [2085] = {.lex_state = 65},
  [2086] = {.lex_state = 65, .external_lex_state = 2},
  [2087] = {.lex_state = 65},
  [2088] = {.lex_state = 65},
  [2089] = {.lex_state = 65},
  [2090] = {.lex_state = 65},
  [2091] = {.lex_state = 65},
  [2092] = {.lex_state = 65},
  [2093] = {.lex_state = 65},
  [2094] = {.lex_state = 65},
  [2095] = {.lex_state = 65},
  [2096] = {.lex_state = 65},
  [2097] = {.lex_state = 65},
  [2098] = {.lex_state = 65, .external_lex_state = 2},
  [2099] = {.lex_state = 65, .external_lex_state = 2},
  [2100] = {.lex_state = 65},
  [2101] = {.lex_state = 65},
  [2102] = {.lex_state = 65},
  [2103] = {.lex_state = 65},
  [2104] = {.lex_state = 65, .external_lex_state = 2},
  [2105] = {.lex_state = 3, .external_lex_state = 11},
  [2106] = {.lex_state = 65, .external_lex_state = 2},
  [2107] = {.lex_state = 65},
  [2108] = {.lex_state = 65},
  [2109] = {.lex_state = 65},
  [2110] = {.lex_state = 65},
  [2111] = {.lex_state = 65},
  [2112] = {.lex_state = 65, .external_lex_state = 2},
  [2113] = {.lex_state = 65, .external_lex_state = 2},
  [2114] = {.lex_state = 65},
  [2115] = {.lex_state = 65, .external_lex_state = 2},
  [2116] = {.lex_state = 65, .external_lex_state = 2},
  [2117] = {.lex_state = 65, .external_lex_state = 2},
  [2118] = {.lex_state = 3, .external_lex_state = 11},
  [2119] = {.lex_state = 65},
  [2120] = {.lex_state = 65},
  [2121] = {.lex_state = 65},
  [2122] = {.lex_state = 65},
  [2123] = {.lex_state = 65, .external_lex_state = 2},
  [2124] = {.lex_state = 65, .external_lex_state = 2},
  [2125] = {.lex_state = 65},
  [2126] = {.lex_state = 65},
  [2127] = {.lex_state = 65},
  [2128] = {.lex_state = 65, .external_lex_state = 2},
  [2129] = {.lex_state = 65, .external_lex_state = 2},
  [2130] = {.lex_state = 65},
  [2131] = {.lex_state = 65},
  [2132] = {.lex_state = 65},
  [2133] = {.lex_state = 65},
  [2134] = {.lex_state = 65},
  [2135] = {.lex_state = 65},
  [2136] = {.lex_state = 65, .external_lex_state = 2},
  [2137] = {.lex_state = 65},
  [2138] = {.lex_state = 65},
  [2139] = {.lex_state = 65},
  [2140] = {.lex_state = 65, .external_lex_state = 2},
  [2141] = {.lex_state = 65},
  [2142] = {.lex_state = 65},
  [2143] = {.lex_state = 65},
  [2144] = {.lex_state = 65, .external_lex_state = 2},
  [2145] = {.lex_state = 65},
  [2146] = {.lex_state = 65, .external_lex_state = 2},
  [2147] = {.lex_state = 65},
  [2148] = {.lex_state = 65},
  [2149] = {.lex_state = 65},
  [2150] = {.lex_state = 65, .external_lex_state = 2},
  [2151] = {.lex_state = 65},
  [2152] = {.lex_state = 65},
  [2153] = {.lex_state = 65},
  [2154] = {.lex_state = 65},
  [2155] = {.lex_state = 65},
  [2156] = {.lex_state = 65},
  [2157] = {.lex_state = 65},
  [2158] = {.lex_state = 65},
  [2159] = {.lex_state = 65},
  [2160] = {.lex_state = 65},
  [2161] = {.lex_state = 65},
  [2162] = {.lex_state = 65},
  [2163] = {.lex_state = 65},
  [2164] = {.lex_state = 65, .external_lex_state = 12},
  [2165] = {.lex_state = 65, .external_lex_state = 2},
  [2166] = {.lex_state = 65},
  [2167] = {.lex_state = 65},
  [2168] = {.lex_state = 65, .external_lex_state = 2},
  [2169] = {.lex_state = 65},
  [2170] = {.lex_state = 65, .external_lex_state = 2},
  [2171] = {.lex_state = 65},
  [2172] = {.lex_state = 65},
  [2173] = {.lex_state = 65, .external_lex_state = 2},
  [2174] = {.lex_state = 65},
  [2175] = {.lex_state = 65, .external_lex_state = 2},
  [2176] = {.lex_state = 65, .external_lex_state = 2},
  [2177] = {.lex_state = 65},
  [2178] = {.lex_state = 65},
  [2179] = {.lex_state = 65},
  [2180] = {.lex_state = 65},
  [2181] = {.lex_state = 65},
  [2182] = {.lex_state = 65, .external_lex_state = 2},
  [2183] = {.lex_state = 65, .external_lex_state = 2},
  [2184] = {.lex_state = 65},
  [2185] = {.lex_state = 65},
  [2186] = {.lex_state = 65},
  [2187] = {.lex_state = 65},
  [2188] = {.lex_state = 65},
  [2189] = {.lex_state = 65},
  [2190] = {.lex_state = 65, .external_lex_state = 2},
  [2191] = {.lex_state = 65},
  [2192] = {.lex_state = 65, .external_lex_state = 2},
  [2193] = {.lex_state = 65, .external_lex_state = 2},
  [2194] = {.lex_state = 65},
  [2195] = {.lex_state = 65},
  [2196] = {.lex_state = 65},
  [2197] = {.lex_state = 65},
  [2198] = {.lex_state = 65},
  [2199] = {.lex_state = 65},
  [2200] = {.lex_state = 65},
  [2201] = {.lex_state = 65},
  [2202] = {.lex_state = 65},
  [2203] = {.lex_state = 65},
  [2204] = {.lex_state = 65},
  [2205] = {.lex_state = 65},
  [2206] = {.lex_state = 65, .external_lex_state = 2},
  [2207] = {.lex_state = 65},
  [2208] = {.lex_state = 65},
  [2209] = {.lex_state = 65},
  [2210] = {.lex_state = 65, .external_lex_state = 2},
  [2211] = {.lex_state = 65, .external_lex_state = 2},
  [2212] = {.lex_state = 65, .external_lex_state = 2},
  [2213] = {.lex_state = 65},
  [2214] = {.lex_state = 65, .external_lex_state = 2},
  [2215] = {.lex_state = 65},
  [2216] = {.lex_state = 65},
  [2217] = {.lex_state = 65},
  [2218] = {.lex_state = 65},
  [2219] = {.lex_state = 65},
  [2220] = {.lex_state = 65},
  [2221] = {.lex_state = 65},
  [2222] = {.lex_state = 65},
  [2223] = {.lex_state = 65},
  [2224] = {.lex_state = 65},
  [2225] = {.lex_state = 65},
  [2226] = {.lex_state = 65},
  [2227] = {.lex_state = 65},
  [2228] = {.lex_state = 65},
  [2229] = {.lex_state = 65, .external_lex_state = 2},
  [2230] = {.lex_state = 65},
  [2231] = {.lex_state = 65},
  [2232] = {.lex_state = 65},
  [2233] = {.lex_state = 65},
  [2234] = {.lex_state = 65},
  [2235] = {.lex_state = 65},
  [2236] = {.lex_state = 3, .external_lex_state = 11},
  [2237] = {.lex_state = 65},
  [2238] = {.lex_state = 65},
  [2239] = {.lex_state = 65},
  [2240] = {.lex_state = 65},
  [2241] = {.lex_state = 65},
  [2242] = {.lex_state = 65, .external_lex_state = 2},
  [2243] = {.lex_state = 65},
  [2244] = {.lex_state = 65},
  [2245] = {.lex_state = 65},
  [2246] = {.lex_state = 65},
  [2247] = {.lex_state = 65},
  [2248] = {.lex_state = 65},
  [2249] = {.lex_state = 65, .external_lex_state = 13},
  [2250] = {.lex_state = 65},
  [2251] = {.lex_state = 65},
  [2252] = {.lex_state = 3, .external_lex_state = 11},
  [2253] = {.lex_state = 65},
  [2254] = {.lex_state = 65},
  [2255] = {.lex_state = 65},
  [2256] = {.lex_state = 65},
  [2257] = {.lex_state = 65},
  [2258] = {.lex_state = 65},
  [2259] = {.lex_state = 65},
  [2260] = {.lex_state = 65},
  [2261] = {.lex_state = 65},
  [2262] = {.lex_state = 65},
  [2263] = {.lex_state = 65},
  [2264] = {.lex_state = 65},
  [2265] = {.lex_state = 65},
  [2266] = {.lex_state = 65},
  [2267] = {.lex_state = 65},
  [2268] = {.lex_state = 65},
  [2269] = {.lex_state = 65},
  [2270] = {.lex_state = 65},
  [2271] = {.lex_state = 65},
  [2272] = {.lex_state = 65},
  [2273] = {.lex_state = 65},
  [2274] = {.lex_state = 65},
  [2275] = {.lex_state = 65},
  [2276] = {.lex_state = 65},
  [2277] = {.lex_state = 65},
  [2278] = {.lex_state = 65},
  [2279] = {.lex_state = 65},
  [2280] = {.lex_state = 65},
  [2281] = {.lex_state = 65},
  [2282] = {.lex_state = 65},
  [2283] = {.lex_state = 3, .external_lex_state = 11},
  [2284] = {.lex_state = 65},
  [2285] = {.lex_state = 65},
  [2286] = {.lex_state = 65},
  [2287] = {.lex_state = 65},
  [2288] = {.lex_state = 65},
  [2289] = {.lex_state = 65},
  [2290] = {.lex_state = 65},
  [2291] = {.lex_state = 65},
  [2292] = {.lex_state = 65},
  [2293] = {.lex_state = 65},
  [2294] = {.lex_state = 65},
  [2295] = {.lex_state = 65},
  [2296] = {.lex_state = 65},
  [2297] = {.lex_state = 65},
  [2298] = {.lex_state = 65},
  [2299] = {.lex_state = 65},
  [2300] = {.lex_state = 65},
  [2301] = {.lex_state = 65},
  [2302] = {.lex_state = 65},
  [2303] = {.lex_state = 65},
  [2304] = {.lex_state = 65},
  [2305] = {.lex_state = 65},
  [2306] = {.lex_state = 65},
  [2307] = {.lex_state = 65},
  [2308] = {.lex_state = 65},
  [2309] = {.lex_state = 65},
  [2310] = {.lex_state = 65},
  [2311] = {.lex_state = 65},
  [2312] = {.lex_state = 65},
  [2313] = {.lex_state = 65},
  [2314] = {.lex_state = 65},
  [2315] = {.lex_state = 65},
  [2316] = {.lex_state = 65, .external_lex_state = 2},
  [2317] = {.lex_state = 65},
  [2318] = {.lex_state = 65, .external_lex_state = 2},
  [2319] = {.lex_state = 65},
  [2320] = {.lex_state = 65},
  [2321] = {.lex_state = 65},
  [2322] = {.lex_state = 3, .external_lex_state = 11},
  [2323] = {.lex_state = 65},
  [2324] = {.lex_state = 65},
  [2325] = {.lex_state = 3, .external_lex_state = 11},
  [2326] = {.lex_state = 65},
  [2327] = {.lex_state = 65},
  [2328] = {.lex_state = 65},
  [2329] = {.lex_state = 65},
  [2330] = {.lex_state = 65},
  [2331] = {.lex_state = 65},
  [2332] = {.lex_state = 65, .external_lex_state = 2},
  [2333] = {.lex_state = 65},
  [2334] = {.lex_state = 65},
  [2335] = {.lex_state = 3, .external_lex_state = 11},
  [2336] = {.lex_state = 65},
  [2337] = {.lex_state = 65},
  [2338] = {.lex_state = 65},
  [2339] = {.lex_state = 65},
  [2340] = {.lex_state = 65},
  [2341] = {.lex_state = 65},
  [2342] = {.lex_state = 65},
  [2343] = {.lex_state = 65},
  [2344] = {.lex_state = 65},
  [2345] = {.lex_state = 65},
  [2346] = {.lex_state = 3, .external_lex_state = 11},
  [2347] = {.lex_state = 65},
  [2348] = {.lex_state = 65},
  [2349] = {.lex_state = 65},
  [2350] = {.lex_state = 65},
  [2351] = {.lex_state = 65},
  [2352] = {.lex_state = 65},
  [2353] = {.lex_state = 65},
  [2354] = {.lex_state = 65},
  [2355] = {.lex_state = 65},
  [2356] = {.lex_state = 65},
  [2357] = {.lex_state = 65},
  [2358] = {.lex_state = 65},
  [2359] = {.lex_state = 3, .external_lex_state = 11},
  [2360] = {.lex_state = 3, .external_lex_state = 11},
  [2361] = {.lex_state = 65, .external_lex_state = 2},
  [2362] = {.lex_state = 65},
  [2363] = {.lex_state = 65},
  [2364] = {.lex_state = 65},
  [2365] = {.lex_state = 65},
  [2366] = {.lex_state = 65},
  [2367] = {.lex_state = 65},
  [2368] = {.lex_state = 65},
  [2369] = {.lex_state = 65},
  [2370] = {.lex_state = 65},
  [2371] = {.lex_state = 65},
  [2372] = {.lex_state = 65},
  [2373] = {.lex_state = 65},
  [2374] = {.lex_state = 65},
  [2375] = {.lex_state = 65},
  [2376] = {.lex_state = 65},
  [2377] = {.lex_state = 65},
  [2378] = {.lex_state = 65},
  [2379] = {.lex_state = 65},
  [2380] = {.lex_state = 65},
  [2381] = {.lex_state = 65},
  [2382] = {.lex_state = 65},
  [2383] = {.lex_state = 65},
  [2384] = {.lex_state = 65},
  [2385] = {.lex_state = 65},
  [2386] = {.lex_state = 65},
  [2387] = {.lex_state = 65},
  [2388] = {.lex_state = 65},
  [2389] = {.lex_state = 65},
  [2390] = {.lex_state = 65, .external_lex_state = 2},
  [2391] = {.lex_state = 3, .external_lex_state = 11},
  [2392] = {.lex_state = 65},
  [2393] = {.lex_state = 65},
  [2394] = {.lex_state = 65},
  [2395] = {.lex_state = 65},
  [2396] = {.lex_state = 65},
  [2397] = {.lex_state = 65, .external_lex_state = 2},
  [2398] = {.lex_state = 65},
  [2399] = {.lex_state = 65},
  [2400] = {.lex_state = 65},
  [2401] = {.lex_state = 65},
  [2402] = {.lex_state = 65},
  [2403] = {.lex_state = 65},
  [2404] = {.lex_state = 65},
  [2405] = {.lex_state = 65},
  [2406] = {.lex_state = 65},
  [2407] = {.lex_state = 65},
  [2408] = {.lex_state = 65},
  [2409] = {.lex_state = 65},
  [2410] = {.lex_state = 65},
  [2411] = {.lex_state = 65},
  [2412] = {.lex_state = 65},
  [2413] = {.lex_state = 65},
  [2414] = {.lex_state = 65},
  [2415] = {.lex_state = 65},
  [2416] = {.lex_state = 65},
  [2417] = {.lex_state = 65},
  [2418] = {.lex_state = 65},
  [2419] = {.lex_state = 65},
  [2420] = {.lex_state = 65},
  [2421] = {.lex_state = 65},
  [2422] = {.lex_state = 65},
  [2423] = {.lex_state = 65},
  [2424] = {.lex_state = 65},
  [2425] = {.lex_state = 65},
  [2426] = {.lex_state = 65},
  [2427] = {.lex_state = 65, .external_lex_state = 11},
  [2428] = {.lex_state = 65, .external_lex_state = 11},
  [2429] = {.lex_state = 65},
  [2430] = {.lex_state = 65, .external_lex_state = 11},
  [2431] = {.lex_state = 65},
  [2432] = {.lex_state = 65},
  [2433] = {.lex_state = 65},
  [2434] = {.lex_state = 65, .external_lex_state = 11},
  [2435] = {.lex_state = 65},
  [2436] = {.lex_state = 65},
  [2437] = {.lex_state = 65},
  [2438] = {.lex_state = 65},
  [2439] = {.lex_state = 65},
  [2440] = {.lex_state = 65},
  [2441] = {.lex_state = 65},
  [2442] = {.lex_state = 65},
  [2443] = {.lex_state = 65},
  [2444] = {.lex_state = 65},
  [2445] = {.lex_state = 65},
  [2446] = {.lex_state = 65},
  [2447] = {.lex_state = 65},
  [2448] = {.lex_state = 65},
  [2449] = {.lex_state = 65},
  [2450] = {.lex_state = 65},
  [2451] = {.lex_state = 65},
  [2452] = {.lex_state = 65, .external_lex_state = 12},
  [2453] = {.lex_state = 65},
  [2454] = {.lex_state = 65},
  [2455] = {.lex_state = 65},
  [2456] = {.lex_state = 10},
  [2457] = {.lex_state = 65},
  [2458] = {.lex_state = 65},
  [2459] = {.lex_state = 65},
  [2460] = {.lex_state = 65},
  [2461] = {.lex_state = 65},
  [2462] = {.lex_state = 65},
  [2463] = {.lex_state = 65},
  [2464] = {.lex_state = 65},
  [2465] = {.lex_state = 65},
  [2466] = {.lex_state = 65},
  [2467] = {.lex_state = 65},
  [2468] = {.lex_state = 65},
  [2469] = {.lex_state = 65},
  [2470] = {.lex_state = 65},
  [2471] = {.lex_state = 65},
  [2472] = {.lex_state = 65},
  [2473] = {.lex_state = 65},
  [2474] = {.lex_state = 65},
  [2475] = {.lex_state = 65},
  [2476] = {.lex_state = 65},
  [2477] = {.lex_state = 65},
  [2478] = {.lex_state = 65},
  [2479] = {.lex_state = 65},
  [2480] = {.lex_state = 65},
  [2481] = {.lex_state = 65},
  [2482] = {.lex_state = 65},
  [2483] = {.lex_state = 65},
  [2484] = {.lex_state = 65},
  [2485] = {.lex_state = 65},
  [2486] = {.lex_state = 65},
  [2487] = {.lex_state = 65},
  [2488] = {.lex_state = 65},
  [2489] = {.lex_state = 65},
  [2490] = {.lex_state = 65},
  [2491] = {.lex_state = 65},
  [2492] = {.lex_state = 65},
  [2493] = {.lex_state = 65},
  [2494] = {.lex_state = 65},
  [2495] = {.lex_state = 10},
  [2496] = {.lex_state = 65},
  [2497] = {.lex_state = 65},
  [2498] = {.lex_state = 65},
  [2499] = {.lex_state = 65},
  [2500] = {.lex_state = 65},
  [2501] = {.lex_state = 65},
  [2502] = {.lex_state = 65},
  [2503] = {.lex_state = 65},
  [2504] = {.lex_state = 65},
  [2505] = {.lex_state = 65},
  [2506] = {.lex_state = 65, .external_lex_state = 11},
  [2507] = {.lex_state = 65},
  [2508] = {.lex_state = 65, .external_lex_state = 11},
  [2509] = {.lex_state = 65},
  [2510] = {.lex_state = 65},
  [2511] = {.lex_state = 65},
  [2512] = {.lex_state = 65},
  [2513] = {.lex_state = 65},
  [2514] = {.lex_state = 65},
  [2515] = {.lex_state = 65},
  [2516] = {.lex_state = 65},
  [2517] = {.lex_state = 65},
  [2518] = {.lex_state = 65},
  [2519] = {.lex_state = 65},
  [2520] = {.lex_state = 65},
  [2521] = {.lex_state = 65},
  [2522] = {.lex_state = 65},
  [2523] = {.lex_state = 65},
  [2524] = {.lex_state = 65},
  [2525] = {.lex_state = 65},
  [2526] = {.lex_state = 65},
  [2527] = {.lex_state = 65},
  [2528] = {.lex_state = 65},
  [2529] = {.lex_state = 65},
  [2530] = {.lex_state = 65, .external_lex_state = 11},
  [2531] = {.lex_state = 65, .external_lex_state = 11},
  [2532] = {.lex_state = 65},
  [2533] = {.lex_state = 65},
  [2534] = {.lex_state = 17},
  [2535] = {.lex_state = 65},
  [2536] = {.lex_state = 65},
  [2537] = {.lex_state = 65},
  [2538] = {.lex_state = 65},
  [2539] = {.lex_state = 65},
  [2540] = {.lex_state = 65},
  [2541] = {.lex_state = 65},
  [2542] = {.lex_state = 65},
  [2543] = {.lex_state = 65},
  [2544] = {.lex_state = 65},
  [2545] = {.lex_state = 65},
  [2546] = {.lex_state = 65},
  [2547] = {.lex_state = 65},
  [2548] = {.lex_state = 65},
  [2549] = {.lex_state = 65},
  [2550] = {.lex_state = 65},
  [2551] = {.lex_state = 65},
  [2552] = {.lex_state = 65},
  [2553] = {.lex_state = 65},
  [2554] = {.lex_state = 65},
  [2555] = {.lex_state = 65},
  [2556] = {.lex_state = 65},
  [2557] = {.lex_state = 65},
  [2558] = {.lex_state = 65},
  [2559] = {.lex_state = 65},
  [2560] = {.lex_state = 17},
  [2561] = {.lex_state = 65},
  [2562] = {.lex_state = 65},
  [2563] = {.lex_state = 65},
  [2564] = {.lex_state = 65},
  [2565] = {.lex_state = 65},
  [2566] = {.lex_state = 65},
  [2567] = {.lex_state = 65},
  [2568] = {.lex_state = 65},
  [2569] = {.lex_state = 65},
  [2570] = {.lex_state = 65},
  [2571] = {.lex_state = 65},
  [2572] = {.lex_state = 65},
  [2573] = {.lex_state = 65},
  [2574] = {.lex_state = 65},
  [2575] = {.lex_state = 65},
  [2576] = {.lex_state = 65},
  [2577] = {.lex_state = 65, .external_lex_state = 11},
  [2578] = {.lex_state = 65},
  [2579] = {.lex_state = 65},
  [2580] = {.lex_state = 65, .external_lex_state = 12},
  [2581] = {.lex_state = 65},
  [2582] = {.lex_state = 65},
  [2583] = {.lex_state = 65},
  [2584] = {.lex_state = 65},
  [2585] = {.lex_state = 65},
  [2586] = {.lex_state = 65, .external_lex_state = 11},
  [2587] = {.lex_state = 65},
  [2588] = {.lex_state = 65},
  [2589] = {.lex_state = 65},
  [2590] = {.lex_state = 65},
  [2591] = {.lex_state = 65},
  [2592] = {.lex_state = 65},
  [2593] = {.lex_state = 65},
  [2594] = {.lex_state = 65},
  [2595] = {.lex_state = 65},
  [2596] = {.lex_state = 65},
  [2597] = {.lex_state = 65},
  [2598] = {.lex_state = 65},
  [2599] = {.lex_state = 65},
  [2600] = {.lex_state = 65},
  [2601] = {.lex_state = 65},
  [2602] = {.lex_state = 65},
  [2603] = {.lex_state = 65},
  [2604] = {.lex_state = 65},
  [2605] = {.lex_state = 65},
  [2606] = {.lex_state = 65},
  [2607] = {.lex_state = 65},
  [2608] = {.lex_state = 65},
  [2609] = {.lex_state = 65},
  [2610] = {.lex_state = 65, .external_lex_state = 12},
  [2611] = {.lex_state = 65},
  [2612] = {.lex_state = 65, .external_lex_state = 12},
  [2613] = {.lex_state = 65},
  [2614] = {.lex_state = 65},
  [2615] = {.lex_state = 65},
  [2616] = {.lex_state = 65},
  [2617] = {.lex_state = 65},
  [2618] = {.lex_state = 65},
  [2619] = {.lex_state = 65},
  [2620] = {.lex_state = 65},
  [2621] = {.lex_state = 65},
  [2622] = {.lex_state = 65},
  [2623] = {.lex_state = 65},
  [2624] = {.lex_state = 65},
  [2625] = {.lex_state = 65},
  [2626] = {.lex_state = 65},
  [2627] = {.lex_state = 65},
  [2628] = {.lex_state = 65},
  [2629] = {.lex_state = 65},
  [2630] = {.lex_state = 65},
  [2631] = {.lex_state = 65},
  [2632] = {.lex_state = 65},
  [2633] = {.lex_state = 65},
  [2634] = {.lex_state = 65},
  [2635] = {.lex_state = 65},
  [2636] = {.lex_state = 65},
  [2637] = {.lex_state = 65},
  [2638] = {.lex_state = 65},
  [2639] = {.lex_state = 65},
  [2640] = {.lex_state = 65},
  [2641] = {.lex_state = 65},
  [2642] = {.lex_state = 65},
  [2643] = {.lex_state = 65},
  [2644] = {.lex_state = 65},
  [2645] = {.lex_state = 65},
  [2646] = {.lex_state = 65},
  [2647] = {.lex_state = 65},
  [2648] = {.lex_state = 65},
  [2649] = {.lex_state = 65},
  [2650] = {.lex_state = 65},
  [2651] = {.lex_state = 65},
  [2652] = {.lex_state = 65},
  [2653] = {.lex_state = 65},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [0] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_name] = ACTIONS(1),
    [anon_sym_QMARK_GT] = ACTIONS(1),
    [sym_php_tag] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [aux_sym_global_declaration_token1] = ACTIONS(1),
    [aux_sym_namespace_definition_token1] = ACTIONS(1),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1),
    [aux_sym_namespace_use_clause_token1] = ACTIONS(1),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1),
    [anon_sym_BSLASH] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [aux_sym_trait_declaration_token1] = ACTIONS(1),
    [aux_sym_interface_declaration_token1] = ACTIONS(1),
    [aux_sym_base_clause_token1] = ACTIONS(1),
    [aux_sym_enum_declaration_token1] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_string] = ACTIONS(1),
    [anon_sym_int] = ACTIONS(1),
    [aux_sym_enum_case_token1] = ACTIONS(1),
    [aux_sym_class_declaration_token1] = ACTIONS(1),
    [aux_sym_final_modifier_token1] = ACTIONS(1),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1),
    [aux_sym_class_interface_clause_token1] = ACTIONS(1),
    [anon_sym_EQ_GT] = ACTIONS(1),
    [sym_var_modifier] = ACTIONS(1),
    [aux_sym_use_instead_of_clause_token1] = ACTIONS(1),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_LPAREN2] = ACTIONS(1),
    [anon_sym_RPAREN2] = ACTIONS(1),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [sym_bottom_type] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_array] = ACTIONS(1),
    [aux_sym_primitive_type_token1] = ACTIONS(1),
    [anon_sym_iterable] = ACTIONS(1),
    [anon_sym_bool] = ACTIONS(1),
    [anon_sym_float] = ACTIONS(1),
    [anon_sym_void] = ACTIONS(1),
    [anon_sym_mixed] = ACTIONS(1),
    [anon_sym_false] = ACTIONS(1),
    [anon_sym_null] = ACTIONS(1),
    [anon_sym_true] = ACTIONS(1),
    [aux_sym_cast_type_token1] = ACTIONS(1),
    [aux_sym_cast_type_token2] = ACTIONS(1),
    [aux_sym_cast_type_token3] = ACTIONS(1),
    [aux_sym_cast_type_token4] = ACTIONS(1),
    [aux_sym_cast_type_token5] = ACTIONS(1),
    [aux_sym_cast_type_token6] = ACTIONS(1),
    [aux_sym_cast_type_token7] = ACTIONS(1),
    [aux_sym_cast_type_token8] = ACTIONS(1),
    [aux_sym_cast_type_token9] = ACTIONS(1),
    [aux_sym_cast_type_token10] = ACTIONS(1),
    [aux_sym_cast_type_token11] = ACTIONS(1),
    [aux_sym_cast_type_token12] = ACTIONS(1),
    [aux_sym_echo_statement_token1] = ACTIONS(1),
    [aux_sym_exit_statement_token1] = ACTIONS(1),
    [anon_sym_unset] = ACTIONS(1),
    [aux_sym_declare_statement_token1] = ACTIONS(1),
    [aux_sym_declare_statement_token2] = ACTIONS(1),
    [anon_sym_ticks] = ACTIONS(1),
    [anon_sym_encoding] = ACTIONS(1),
    [anon_sym_strict_types] = ACTIONS(1),
    [sym_float] = ACTIONS(1),
    [aux_sym_try_statement_token1] = ACTIONS(1),
    [aux_sym_catch_clause_token1] = ACTIONS(1),
    [aux_sym_finally_clause_token1] = ACTIONS(1),
    [aux_sym_goto_statement_token1] = ACTIONS(1),
    [aux_sym_continue_statement_token1] = ACTIONS(1),
    [aux_sym_break_statement_token1] = ACTIONS(1),
    [sym_integer] = ACTIONS(1),
    [aux_sym_return_statement_token1] = ACTIONS(1),
    [aux_sym_throw_expression_token1] = ACTIONS(1),
    [aux_sym_while_statement_token1] = ACTIONS(1),
    [aux_sym_while_statement_token2] = ACTIONS(1),
    [aux_sym_do_statement_token1] = ACTIONS(1),
    [aux_sym_for_statement_token1] = ACTIONS(1),
    [aux_sym_for_statement_token2] = ACTIONS(1),
    [aux_sym_foreach_statement_token1] = ACTIONS(1),
    [aux_sym_foreach_statement_token2] = ACTIONS(1),
    [aux_sym_if_statement_token1] = ACTIONS(1),
    [aux_sym_if_statement_token2] = ACTIONS(1),
    [aux_sym_else_if_clause_token1] = ACTIONS(1),
    [aux_sym_else_clause_token1] = ACTIONS(1),
    [aux_sym_match_expression_token1] = ACTIONS(1),
    [aux_sym_match_default_expression_token1] = ACTIONS(1),
    [aux_sym_switch_statement_token1] = ACTIONS(1),
    [aux_sym_switch_block_token1] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [aux_sym_clone_expression_token1] = ACTIONS(1),
    [anon_sym_COLON_COLON] = ACTIONS(1),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1),
    [anon_sym_DASH_DASH] = ACTIONS(1),
    [anon_sym_PLUS_PLUS] = ACTIONS(1),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1),
    [anon_sym_STAR_EQ] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_PERCENT_EQ] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_DASH_EQ] = ACTIONS(1),
    [anon_sym_LT_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_AMP_EQ] = ACTIONS(1),
    [anon_sym_CARET_EQ] = ACTIONS(1),
    [anon_sym_PIPE_EQ] = ACTIONS(1),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [anon_sym_QMARK_DASH_GT] = ACTIONS(1),
    [aux_sym__list_destructing_token1] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_self] = ACTIONS(1),
    [anon_sym_parent] = ACTIONS(1),
    [aux_sym__argument_name_token1] = ACTIONS(1),
    [aux_sym__argument_name_token2] = ACTIONS(1),
    [anon_sym_POUND_LBRACK] = ACTIONS(1),
    [aux_sym_encapsed_string_token1] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [aux_sym_string_token1] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [aux_sym_string_token2] = ACTIONS(1),
    [anon_sym_LT_LT_LT] = ACTIONS(1),
    [anon_sym_DQUOTE2] = ACTIONS(1),
    [anon_sym_SQUOTE2] = ACTIONS(1),
    [anon_sym_BQUOTE] = ACTIONS(1),
    [anon_sym_DOLLAR] = ACTIONS(1),
    [aux_sym_yield_expression_token1] = ACTIONS(1),
    [aux_sym_yield_expression_token2] = ACTIONS(1),
    [aux_sym_binary_expression_token1] = ACTIONS(1),
    [anon_sym_QMARK_QMARK] = ACTIONS(1),
    [anon_sym_STAR_STAR] = ACTIONS(1),
    [aux_sym_binary_expression_token2] = ACTIONS(1),
    [aux_sym_binary_expression_token3] = ACTIONS(1),
    [aux_sym_binary_expression_token4] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_LT_GT] = ACTIONS(1),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_LT_EQ_GT] = ACTIONS(1),
    [anon_sym_LT_LT] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [aux_sym_include_expression_token1] = ACTIONS(1),
    [aux_sym_include_once_expression_token1] = ACTIONS(1),
    [aux_sym_require_expression_token1] = ACTIONS(1),
    [aux_sym_require_once_expression_token1] = ACTIONS(1),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(1),
    [sym_encapsed_string_chars] = ACTIONS(1),
    [sym_encapsed_string_chars_after_variable] = ACTIONS(1),
    [sym_execution_string_chars] = ACTIONS(1),
    [sym_execution_string_chars_after_variable] = ACTIONS(1),
    [sym_encapsed_string_chars_heredoc] = ACTIONS(1),
    [sym_encapsed_string_chars_after_variable_heredoc] = ACTIONS(1),
    [sym__eof] = ACTIONS(1),
    [sym_heredoc_start] = ACTIONS(1),
    [sym_heredoc_end] = ACTIONS(1),
    [sym_nowdoc_string] = ACTIONS(1),
    [sym_sentinel_error] = ACTIONS(1),
  },
  [1] = {
    [sym_program] = STATE(2630),
    [sym_statement] = STATE(26),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(26),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [ts_builtin_sym_end] = ACTIONS(5),
    [sym_name] = ACTIONS(7),
    [anon_sym_QMARK_GT] = ACTIONS(9),
    [sym_php_tag] = ACTIONS(11),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [2] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [ts_builtin_sym_end] = ACTIONS(137),
    [sym_name] = ACTIONS(139),
    [anon_sym_QMARK_GT] = ACTIONS(137),
    [anon_sym_SEMI] = ACTIONS(142),
    [aux_sym_function_static_declaration_token1] = ACTIONS(145),
    [aux_sym_global_declaration_token1] = ACTIONS(148),
    [aux_sym_namespace_definition_token1] = ACTIONS(151),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(154),
    [aux_sym__namespace_use_type_token1] = ACTIONS(157),
    [aux_sym__namespace_use_type_token2] = ACTIONS(160),
    [anon_sym_BSLASH] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(166),
    [anon_sym_RBRACE] = ACTIONS(137),
    [aux_sym_trait_declaration_token1] = ACTIONS(169),
    [aux_sym_interface_declaration_token1] = ACTIONS(172),
    [aux_sym_enum_declaration_token1] = ACTIONS(175),
    [aux_sym_enum_case_token1] = ACTIONS(178),
    [aux_sym_class_declaration_token1] = ACTIONS(180),
    [aux_sym_final_modifier_token1] = ACTIONS(183),
    [aux_sym_abstract_modifier_token1] = ACTIONS(186),
    [aux_sym_readonly_modifier_token1] = ACTIONS(189),
    [sym_var_modifier] = ACTIONS(192),
    [aux_sym_visibility_modifier_token1] = ACTIONS(195),
    [aux_sym_visibility_modifier_token2] = ACTIONS(195),
    [aux_sym_visibility_modifier_token3] = ACTIONS(195),
    [anon_sym_LPAREN2] = ACTIONS(198),
    [aux_sym__arrow_function_header_token1] = ACTIONS(201),
    [aux_sym_cast_type_token1] = ACTIONS(204),
    [aux_sym_echo_statement_token1] = ACTIONS(207),
    [aux_sym_exit_statement_token1] = ACTIONS(210),
    [anon_sym_unset] = ACTIONS(213),
    [aux_sym_declare_statement_token1] = ACTIONS(216),
    [aux_sym_declare_statement_token2] = ACTIONS(178),
    [sym_float] = ACTIONS(219),
    [aux_sym_try_statement_token1] = ACTIONS(222),
    [aux_sym_goto_statement_token1] = ACTIONS(225),
    [aux_sym_continue_statement_token1] = ACTIONS(228),
    [aux_sym_break_statement_token1] = ACTIONS(231),
    [sym_integer] = ACTIONS(219),
    [aux_sym_return_statement_token1] = ACTIONS(234),
    [aux_sym_throw_expression_token1] = ACTIONS(237),
    [aux_sym_while_statement_token1] = ACTIONS(240),
    [aux_sym_while_statement_token2] = ACTIONS(178),
    [aux_sym_do_statement_token1] = ACTIONS(243),
    [aux_sym_for_statement_token1] = ACTIONS(246),
    [aux_sym_for_statement_token2] = ACTIONS(178),
    [aux_sym_foreach_statement_token1] = ACTIONS(249),
    [aux_sym_foreach_statement_token2] = ACTIONS(178),
    [aux_sym_if_statement_token1] = ACTIONS(252),
    [aux_sym_if_statement_token2] = ACTIONS(178),
    [aux_sym_match_expression_token1] = ACTIONS(255),
    [aux_sym_match_default_expression_token1] = ACTIONS(178),
    [aux_sym_switch_statement_token1] = ACTIONS(258),
    [aux_sym_switch_block_token1] = ACTIONS(178),
    [anon_sym_PLUS] = ACTIONS(261),
    [anon_sym_DASH] = ACTIONS(261),
    [anon_sym_TILDE] = ACTIONS(264),
    [anon_sym_BANG] = ACTIONS(264),
    [anon_sym_AT] = ACTIONS(267),
    [aux_sym_clone_expression_token1] = ACTIONS(270),
    [aux_sym_print_intrinsic_token1] = ACTIONS(273),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(276),
    [anon_sym_DASH_DASH] = ACTIONS(279),
    [anon_sym_PLUS_PLUS] = ACTIONS(279),
    [aux_sym__list_destructing_token1] = ACTIONS(282),
    [anon_sym_LBRACK] = ACTIONS(285),
    [anon_sym_self] = ACTIONS(288),
    [anon_sym_parent] = ACTIONS(288),
    [aux_sym__argument_name_token1] = ACTIONS(291),
    [aux_sym__argument_name_token2] = ACTIONS(294),
    [anon_sym_POUND_LBRACK] = ACTIONS(297),
    [aux_sym_encapsed_string_token1] = ACTIONS(300),
    [anon_sym_DQUOTE] = ACTIONS(300),
    [aux_sym_string_token1] = ACTIONS(303),
    [anon_sym_SQUOTE] = ACTIONS(303),
    [anon_sym_LT_LT_LT] = ACTIONS(306),
    [anon_sym_BQUOTE] = ACTIONS(309),
    [anon_sym_DOLLAR] = ACTIONS(312),
    [aux_sym_yield_expression_token1] = ACTIONS(315),
    [aux_sym_include_expression_token1] = ACTIONS(318),
    [aux_sym_include_once_expression_token1] = ACTIONS(321),
    [aux_sym_require_expression_token1] = ACTIONS(324),
    [aux_sym_require_once_expression_token1] = ACTIONS(327),
    [sym_comment] = ACTIONS(3),
  },
  [3] = {
    [sym_statement] = STATE(5),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(5),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(330),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_enum_case_token1] = ACTIONS(332),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_match_default_expression_token1] = ACTIONS(332),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [aux_sym_switch_block_token1] = ACTIONS(332),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [4] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(334),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_enum_case_token1] = ACTIONS(336),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_match_default_expression_token1] = ACTIONS(336),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [aux_sym_switch_block_token1] = ACTIONS(336),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [5] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(338),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_enum_case_token1] = ACTIONS(340),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_match_default_expression_token1] = ACTIONS(340),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [aux_sym_switch_block_token1] = ACTIONS(340),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [6] = {
    [sym_statement] = STATE(4),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(4),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(342),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_enum_case_token1] = ACTIONS(344),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_match_default_expression_token1] = ACTIONS(344),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [aux_sym_switch_block_token1] = ACTIONS(344),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [7] = {
    [sym_statement] = STATE(8),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(8),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_while_statement_token2] = ACTIONS(346),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_foreach_statement_token2] = ACTIONS(346),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_if_statement_token2] = ACTIONS(346),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [8] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_while_statement_token2] = ACTIONS(348),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_foreach_statement_token2] = ACTIONS(348),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_if_statement_token2] = ACTIONS(348),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [9] = {
    [sym_statement] = STATE(542),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_colon_block] = STATE(2544),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(544),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(350),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(352),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(350),
  },
  [10] = {
    [sym_statement] = STATE(1882),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_colon_block] = STATE(2550),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(1884),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(352),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(356),
  },
  [11] = {
    [sym_statement] = STATE(11),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(11),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(139),
    [anon_sym_SEMI] = ACTIONS(142),
    [aux_sym_function_static_declaration_token1] = ACTIONS(145),
    [aux_sym_global_declaration_token1] = ACTIONS(148),
    [aux_sym_namespace_definition_token1] = ACTIONS(151),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(154),
    [aux_sym__namespace_use_type_token1] = ACTIONS(157),
    [aux_sym__namespace_use_type_token2] = ACTIONS(160),
    [anon_sym_BSLASH] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(166),
    [aux_sym_trait_declaration_token1] = ACTIONS(169),
    [aux_sym_interface_declaration_token1] = ACTIONS(172),
    [aux_sym_enum_declaration_token1] = ACTIONS(175),
    [aux_sym_class_declaration_token1] = ACTIONS(180),
    [aux_sym_final_modifier_token1] = ACTIONS(183),
    [aux_sym_abstract_modifier_token1] = ACTIONS(186),
    [aux_sym_readonly_modifier_token1] = ACTIONS(189),
    [sym_var_modifier] = ACTIONS(192),
    [aux_sym_visibility_modifier_token1] = ACTIONS(195),
    [aux_sym_visibility_modifier_token2] = ACTIONS(195),
    [aux_sym_visibility_modifier_token3] = ACTIONS(195),
    [anon_sym_LPAREN2] = ACTIONS(198),
    [aux_sym__arrow_function_header_token1] = ACTIONS(201),
    [aux_sym_cast_type_token1] = ACTIONS(204),
    [aux_sym_echo_statement_token1] = ACTIONS(207),
    [aux_sym_exit_statement_token1] = ACTIONS(210),
    [anon_sym_unset] = ACTIONS(213),
    [aux_sym_declare_statement_token1] = ACTIONS(410),
    [sym_float] = ACTIONS(219),
    [aux_sym_try_statement_token1] = ACTIONS(222),
    [aux_sym_goto_statement_token1] = ACTIONS(225),
    [aux_sym_continue_statement_token1] = ACTIONS(228),
    [aux_sym_break_statement_token1] = ACTIONS(231),
    [sym_integer] = ACTIONS(219),
    [aux_sym_return_statement_token1] = ACTIONS(234),
    [aux_sym_throw_expression_token1] = ACTIONS(237),
    [aux_sym_while_statement_token1] = ACTIONS(413),
    [aux_sym_do_statement_token1] = ACTIONS(243),
    [aux_sym_for_statement_token1] = ACTIONS(416),
    [aux_sym_foreach_statement_token1] = ACTIONS(419),
    [aux_sym_if_statement_token1] = ACTIONS(422),
    [aux_sym_if_statement_token2] = ACTIONS(178),
    [aux_sym_else_if_clause_token1] = ACTIONS(178),
    [aux_sym_else_clause_token1] = ACTIONS(178),
    [aux_sym_match_expression_token1] = ACTIONS(255),
    [aux_sym_switch_statement_token1] = ACTIONS(258),
    [anon_sym_PLUS] = ACTIONS(261),
    [anon_sym_DASH] = ACTIONS(261),
    [anon_sym_TILDE] = ACTIONS(264),
    [anon_sym_BANG] = ACTIONS(264),
    [anon_sym_AT] = ACTIONS(267),
    [aux_sym_clone_expression_token1] = ACTIONS(270),
    [aux_sym_print_intrinsic_token1] = ACTIONS(273),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(276),
    [anon_sym_DASH_DASH] = ACTIONS(279),
    [anon_sym_PLUS_PLUS] = ACTIONS(279),
    [aux_sym__list_destructing_token1] = ACTIONS(282),
    [anon_sym_LBRACK] = ACTIONS(285),
    [anon_sym_self] = ACTIONS(288),
    [anon_sym_parent] = ACTIONS(288),
    [aux_sym__argument_name_token1] = ACTIONS(291),
    [aux_sym__argument_name_token2] = ACTIONS(294),
    [anon_sym_POUND_LBRACK] = ACTIONS(297),
    [aux_sym_encapsed_string_token1] = ACTIONS(300),
    [anon_sym_DQUOTE] = ACTIONS(300),
    [aux_sym_string_token1] = ACTIONS(303),
    [anon_sym_SQUOTE] = ACTIONS(303),
    [anon_sym_LT_LT_LT] = ACTIONS(306),
    [anon_sym_BQUOTE] = ACTIONS(309),
    [anon_sym_DOLLAR] = ACTIONS(312),
    [aux_sym_yield_expression_token1] = ACTIONS(315),
    [aux_sym_include_expression_token1] = ACTIONS(318),
    [aux_sym_include_once_expression_token1] = ACTIONS(321),
    [aux_sym_require_expression_token1] = ACTIONS(324),
    [aux_sym_require_once_expression_token1] = ACTIONS(327),
    [sym_comment] = ACTIONS(3),
  },
  [12] = {
    [sym_statement] = STATE(13),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(13),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_if_statement_token2] = ACTIONS(346),
    [aux_sym_else_if_clause_token1] = ACTIONS(346),
    [aux_sym_else_clause_token1] = ACTIONS(346),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [13] = {
    [sym_statement] = STATE(11),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(11),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_if_statement_token2] = ACTIONS(348),
    [aux_sym_else_if_clause_token1] = ACTIONS(348),
    [aux_sym_else_clause_token1] = ACTIONS(348),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [14] = {
    [sym_statement] = STATE(542),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_colon_block] = STATE(2544),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(544),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(350),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(352),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(350),
  },
  [15] = {
    [sym_statement] = STATE(1882),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_colon_block] = STATE(2550),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(1884),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(356),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(352),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(356),
  },
  [16] = {
    [sym_statement] = STATE(2194),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(2195),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(445),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(447),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(445),
  },
  [17] = {
    [sym_statement] = STATE(575),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(456),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(449),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(451),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(449),
  },
  [18] = {
    [sym_statement] = STATE(458),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(462),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(453),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(455),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(453),
  },
  [19] = {
    [sym_statement] = STATE(523),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(552),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(457),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(459),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(457),
  },
  [20] = {
    [sym_statement] = STATE(487),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(488),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(461),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(463),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(461),
  },
  [21] = {
    [sym_statement] = STATE(2001),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(2001),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(465),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(467),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(465),
  },
  [22] = {
    [sym_statement] = STATE(2142),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(2145),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(469),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(471),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(469),
  },
  [23] = {
    [sym_statement] = STATE(1935),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(1936),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(473),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(475),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(473),
  },
  [24] = {
    [sym_statement] = STATE(2199),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(2207),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(477),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(479),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(477),
  },
  [25] = {
    [sym_statement] = STATE(2209),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(2215),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(481),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(483),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(481),
  },
  [26] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [ts_builtin_sym_end] = ACTIONS(485),
    [sym_name] = ACTIONS(7),
    [anon_sym_QMARK_GT] = ACTIONS(487),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [27] = {
    [sym_statement] = STATE(522),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(524),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(489),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(491),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(489),
  },
  [28] = {
    [sym_statement] = STATE(1904),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(1905),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(493),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(495),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(493),
  },
  [29] = {
    [sym_statement] = STATE(1907),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(1909),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(497),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(499),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(497),
  },
  [30] = {
    [sym_statement] = STATE(1914),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(1916),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(501),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(503),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(501),
  },
  [31] = {
    [sym_statement] = STATE(1935),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(1936),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(473),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(475),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(473),
  },
  [32] = {
    [sym_statement] = STATE(563),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(569),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(505),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(507),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(505),
  },
  [33] = {
    [sym_statement] = STATE(534),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(535),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(509),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(511),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(509),
  },
  [34] = {
    [sym_statement] = STATE(533),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(533),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(513),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(515),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(513),
  },
  [35] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [ts_builtin_sym_end] = ACTIONS(517),
    [sym_name] = ACTIONS(7),
    [anon_sym_QMARK_GT] = ACTIONS(519),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [36] = {
    [sym_statement] = STATE(523),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(552),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(457),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(459),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(457),
  },
  [37] = {
    [sym_statement] = STATE(522),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(524),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(489),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(491),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(489),
  },
  [38] = {
    [sym_statement] = STATE(529),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(530),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(521),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(523),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(521),
  },
  [39] = {
    [sym_statement] = STATE(534),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(535),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(509),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(511),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(509),
  },
  [40] = {
    [sym_statement] = STATE(533),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(533),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(513),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(515),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(513),
  },
  [41] = {
    [sym_statement] = STATE(563),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(569),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(505),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(507),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(505),
  },
  [42] = {
    [sym_statement] = STATE(575),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(456),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(449),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(451),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(449),
  },
  [43] = {
    [sym_statement] = STATE(458),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(462),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(453),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(455),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(453),
  },
  [44] = {
    [sym_statement] = STATE(487),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(488),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(461),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(463),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(461),
  },
  [45] = {
    [sym_statement] = STATE(2001),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(2001),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(465),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(467),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(465),
  },
  [46] = {
    [sym_statement] = STATE(2142),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(2145),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(469),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(471),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(469),
  },
  [47] = {
    [sym_statement] = STATE(2194),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(2195),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(445),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(447),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(445),
  },
  [48] = {
    [sym_statement] = STATE(2199),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(2207),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(477),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(479),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(477),
  },
  [49] = {
    [sym_statement] = STATE(2209),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(2215),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(481),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(483),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(481),
  },
  [50] = {
    [sym_statement] = STATE(35),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(35),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [ts_builtin_sym_end] = ACTIONS(485),
    [sym_name] = ACTIONS(7),
    [anon_sym_QMARK_GT] = ACTIONS(487),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [51] = {
    [sym_statement] = STATE(1904),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(1905),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(493),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(495),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(493),
  },
  [52] = {
    [sym_statement] = STATE(1907),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(1909),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(497),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(499),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(497),
  },
  [53] = {
    [sym_statement] = STATE(1914),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(1916),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(501),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(503),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(501),
  },
  [54] = {
    [sym_statement] = STATE(529),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(530),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(521),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(523),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(521),
  },
  [55] = {
    [sym_statement] = STATE(574),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_colon_block] = STATE(2467),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(352),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [56] = {
    [sym_statement] = STATE(57),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(57),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [aux_sym_declare_statement_token2] = ACTIONS(525),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [57] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [aux_sym_declare_statement_token2] = ACTIONS(527),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [58] = {
    [sym_statement] = STATE(59),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(59),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(529),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [59] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(531),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [60] = {
    [sym_statement] = STATE(64),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(64),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(533),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [61] = {
    [sym_statement] = STATE(65),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(65),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(535),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [62] = {
    [sym_statement] = STATE(448),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_colon_block] = STATE(1611),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(537),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [63] = {
    [sym_statement] = STATE(67),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(67),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(539),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [64] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(541),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [65] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(543),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [66] = {
    [sym_statement] = STATE(70),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(70),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(545),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [67] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(547),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [68] = {
    [sym_statement] = STATE(71),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(71),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(549),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [69] = {
    [sym_statement] = STATE(72),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(72),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(551),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [70] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(553),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [71] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(555),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [72] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(557),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [73] = {
    [sym_statement] = STATE(74),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(74),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(559),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [74] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(561),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [75] = {
    [sym_statement] = STATE(2038),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_colon_block] = STATE(2601),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(563),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(352),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [76] = {
    [sym_statement] = STATE(1631),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_colon_block] = STATE(1599),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(563),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(537),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [77] = {
    [sym_statement] = STATE(78),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(78),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(565),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [78] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(567),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [79] = {
    [sym_statement] = STATE(446),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_colon_block] = STATE(1611),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(537),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [80] = {
    [sym_statement] = STATE(81),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(81),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [aux_sym_declare_statement_token2] = ACTIONS(569),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [81] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [aux_sym_declare_statement_token2] = ACTIONS(571),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [82] = {
    [sym_statement] = STATE(83),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(83),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(573),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [83] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(575),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [84] = {
    [sym_statement] = STATE(87),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(87),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(577),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [85] = {
    [sym_statement] = STATE(88),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(88),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(579),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [86] = {
    [sym_statement] = STATE(90),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(90),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(581),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [87] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(583),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [88] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(585),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [89] = {
    [sym_statement] = STATE(93),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(93),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(587),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [90] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(589),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [91] = {
    [sym_statement] = STATE(94),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(94),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(591),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [92] = {
    [sym_statement] = STATE(95),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(95),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(593),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [93] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(595),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [94] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(597),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [95] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(599),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [96] = {
    [sym_statement] = STATE(97),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(97),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(601),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [97] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_for_statement_token2] = ACTIONS(603),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [98] = {
    [sym_statement] = STATE(1574),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_colon_block] = STATE(1599),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(563),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(537),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [99] = {
    [sym_statement] = STATE(100),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(100),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(605),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [100] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(607),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [101] = {
    [sym_statement] = STATE(108),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(108),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(609),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [102] = {
    [sym_statement] = STATE(103),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(103),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(611),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [103] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(613),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [104] = {
    [sym_statement] = STATE(105),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(105),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(615),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [105] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(617),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [106] = {
    [sym_statement] = STATE(574),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_colon_block] = STATE(2467),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(352),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [107] = {
    [sym_statement] = STATE(2038),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_colon_block] = STATE(2601),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(563),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [anon_sym_COLON] = ACTIONS(352),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [108] = {
    [sym_statement] = STATE(2),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_program_repeat1] = STATE(2),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_RBRACE] = ACTIONS(619),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [109] = {
    [sym_statement] = STATE(2161),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(563),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [110] = {
    [sym_statement] = STATE(2566),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(563),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [111] = {
    [sym_statement] = STATE(2021),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(563),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [112] = {
    [sym_statement] = STATE(540),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [113] = {
    [sym_statement] = STATE(2021),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(563),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(435),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(437),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(439),
    [aux_sym_foreach_statement_token1] = ACTIONS(441),
    [aux_sym_if_statement_token1] = ACTIONS(443),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [114] = {
    [sym_statement] = STATE(540),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(61),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(77),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(81),
    [aux_sym_foreach_statement_token1] = ACTIONS(83),
    [aux_sym_if_statement_token1] = ACTIONS(85),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [115] = {
    [sym_statement] = STATE(521),
    [sym_empty_statement] = STATE(500),
    [sym_function_static_declaration] = STATE(500),
    [sym_global_declaration] = STATE(500),
    [sym_namespace_definition] = STATE(500),
    [sym_namespace_use_declaration] = STATE(500),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(500),
    [sym_interface_declaration] = STATE(500),
    [sym_enum_declaration] = STATE(500),
    [sym_class_declaration] = STATE(500),
    [sym_final_modifier] = STATE(1341),
    [sym_abstract_modifier] = STATE(1341),
    [sym_readonly_modifier] = STATE(1341),
    [sym_const_declaration] = STATE(500),
    [sym__const_declaration] = STATE(532),
    [sym__modifier] = STATE(1341),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1341),
    [sym_function_definition] = STATE(500),
    [sym__function_definition_header] = STATE(2290),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(500),
    [sym_exit_statement] = STATE(500),
    [sym_unset_statement] = STATE(500),
    [sym_declare_statement] = STATE(500),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(500),
    [sym_goto_statement] = STATE(500),
    [sym_continue_statement] = STATE(500),
    [sym_break_statement] = STATE(500),
    [sym_return_statement] = STATE(500),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(500),
    [sym_do_statement] = STATE(500),
    [sym_for_statement] = STATE(500),
    [sym_foreach_statement] = STATE(500),
    [sym_if_statement] = STATE(500),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(500),
    [sym_compound_statement] = STATE(500),
    [sym_named_label_statement] = STATE(500),
    [sym_expression_statement] = STATE(500),
    [sym_expression] = STATE(1186),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1316),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1341),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(13),
    [aux_sym_function_static_declaration_token1] = ACTIONS(15),
    [aux_sym_global_declaration_token1] = ACTIONS(17),
    [aux_sym_namespace_definition_token1] = ACTIONS(19),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(21),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(25),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [aux_sym_trait_declaration_token1] = ACTIONS(31),
    [aux_sym_interface_declaration_token1] = ACTIONS(33),
    [aux_sym_enum_declaration_token1] = ACTIONS(35),
    [aux_sym_class_declaration_token1] = ACTIONS(37),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(45),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(55),
    [aux_sym_exit_statement_token1] = ACTIONS(57),
    [anon_sym_unset] = ACTIONS(59),
    [aux_sym_declare_statement_token1] = ACTIONS(425),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(65),
    [aux_sym_goto_statement_token1] = ACTIONS(67),
    [aux_sym_continue_statement_token1] = ACTIONS(69),
    [aux_sym_break_statement_token1] = ACTIONS(71),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(73),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(427),
    [aux_sym_do_statement_token1] = ACTIONS(79),
    [aux_sym_for_statement_token1] = ACTIONS(429),
    [aux_sym_foreach_statement_token1] = ACTIONS(431),
    [aux_sym_if_statement_token1] = ACTIONS(433),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(89),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [116] = {
    [sym_statement] = STATE(2446),
    [sym_empty_statement] = STATE(1975),
    [sym_function_static_declaration] = STATE(1975),
    [sym_global_declaration] = STATE(1975),
    [sym_namespace_definition] = STATE(1975),
    [sym_namespace_use_declaration] = STATE(1975),
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_trait_declaration] = STATE(1975),
    [sym_interface_declaration] = STATE(1975),
    [sym_enum_declaration] = STATE(1975),
    [sym_class_declaration] = STATE(1975),
    [sym_final_modifier] = STATE(1340),
    [sym_abstract_modifier] = STATE(1340),
    [sym_readonly_modifier] = STATE(1340),
    [sym_const_declaration] = STATE(1975),
    [sym__const_declaration] = STATE(2054),
    [sym__modifier] = STATE(1340),
    [sym_static_modifier] = STATE(1434),
    [sym_visibility_modifier] = STATE(1340),
    [sym_function_definition] = STATE(1975),
    [sym__function_definition_header] = STATE(2398),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_echo_statement] = STATE(1975),
    [sym_exit_statement] = STATE(1975),
    [sym_unset_statement] = STATE(1975),
    [sym_declare_statement] = STATE(1975),
    [sym_literal] = STATE(1096),
    [sym_try_statement] = STATE(1975),
    [sym_goto_statement] = STATE(1975),
    [sym_continue_statement] = STATE(1975),
    [sym_break_statement] = STATE(1975),
    [sym_return_statement] = STATE(1975),
    [sym_throw_expression] = STATE(1096),
    [sym_while_statement] = STATE(1975),
    [sym_do_statement] = STATE(1975),
    [sym_for_statement] = STATE(1975),
    [sym_foreach_statement] = STATE(1975),
    [sym_if_statement] = STATE(1975),
    [sym_match_expression] = STATE(1105),
    [sym_switch_statement] = STATE(1975),
    [sym_compound_statement] = STATE(1975),
    [sym_named_label_statement] = STATE(1975),
    [sym_expression_statement] = STATE(1975),
    [sym_expression] = STATE(1195),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1311),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_class_declaration_repeat1] = STATE(1340),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(354),
    [anon_sym_SEMI] = ACTIONS(563),
    [aux_sym_function_static_declaration_token1] = ACTIONS(358),
    [aux_sym_global_declaration_token1] = ACTIONS(360),
    [aux_sym_namespace_definition_token1] = ACTIONS(362),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(23),
    [aux_sym__namespace_use_type_token2] = ACTIONS(366),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(368),
    [aux_sym_trait_declaration_token1] = ACTIONS(370),
    [aux_sym_interface_declaration_token1] = ACTIONS(372),
    [aux_sym_enum_declaration_token1] = ACTIONS(374),
    [aux_sym_class_declaration_token1] = ACTIONS(376),
    [aux_sym_final_modifier_token1] = ACTIONS(39),
    [aux_sym_abstract_modifier_token1] = ACTIONS(41),
    [aux_sym_readonly_modifier_token1] = ACTIONS(43),
    [sym_var_modifier] = ACTIONS(378),
    [aux_sym_visibility_modifier_token1] = ACTIONS(47),
    [aux_sym_visibility_modifier_token2] = ACTIONS(47),
    [aux_sym_visibility_modifier_token3] = ACTIONS(47),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [aux_sym_echo_statement_token1] = ACTIONS(380),
    [aux_sym_exit_statement_token1] = ACTIONS(382),
    [anon_sym_unset] = ACTIONS(384),
    [aux_sym_declare_statement_token1] = ACTIONS(386),
    [sym_float] = ACTIONS(63),
    [aux_sym_try_statement_token1] = ACTIONS(388),
    [aux_sym_goto_statement_token1] = ACTIONS(390),
    [aux_sym_continue_statement_token1] = ACTIONS(392),
    [aux_sym_break_statement_token1] = ACTIONS(394),
    [sym_integer] = ACTIONS(63),
    [aux_sym_return_statement_token1] = ACTIONS(396),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_while_statement_token1] = ACTIONS(398),
    [aux_sym_do_statement_token1] = ACTIONS(400),
    [aux_sym_for_statement_token1] = ACTIONS(402),
    [aux_sym_foreach_statement_token1] = ACTIONS(404),
    [aux_sym_if_statement_token1] = ACTIONS(406),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [aux_sym_switch_statement_token1] = ACTIONS(408),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [117] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1020),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(994),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(623),
    [anon_sym_AMP] = ACTIONS(625),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym_namespace_use_clause_token1] = ACTIONS(629),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_RBRACE] = ACTIONS(623),
    [anon_sym_COLON] = ACTIONS(623),
    [anon_sym_EQ_GT] = ACTIONS(623),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [anon_sym_RPAREN2] = ACTIONS(623),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(637),
    [anon_sym_QMARK] = ACTIONS(629),
    [anon_sym_PIPE] = ACTIONS(629),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(647),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_yield_expression_token2] = ACTIONS(679),
    [aux_sym_binary_expression_token1] = ACTIONS(629),
    [anon_sym_QMARK_QMARK] = ACTIONS(623),
    [anon_sym_STAR_STAR] = ACTIONS(623),
    [aux_sym_binary_expression_token2] = ACTIONS(629),
    [aux_sym_binary_expression_token3] = ACTIONS(629),
    [aux_sym_binary_expression_token4] = ACTIONS(629),
    [anon_sym_PIPE_PIPE] = ACTIONS(623),
    [anon_sym_AMP_AMP] = ACTIONS(623),
    [anon_sym_CARET] = ACTIONS(623),
    [anon_sym_EQ_EQ] = ACTIONS(629),
    [anon_sym_BANG_EQ] = ACTIONS(629),
    [anon_sym_LT_GT] = ACTIONS(623),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(623),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(623),
    [anon_sym_LT] = ACTIONS(629),
    [anon_sym_GT] = ACTIONS(629),
    [anon_sym_LT_EQ] = ACTIONS(629),
    [anon_sym_GT_EQ] = ACTIONS(623),
    [anon_sym_LT_EQ_GT] = ACTIONS(623),
    [anon_sym_LT_LT] = ACTIONS(629),
    [anon_sym_GT_GT] = ACTIONS(623),
    [anon_sym_DOT] = ACTIONS(629),
    [anon_sym_STAR] = ACTIONS(629),
    [anon_sym_SLASH] = ACTIONS(629),
    [anon_sym_PERCENT] = ACTIONS(623),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [118] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1124),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(994),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(623),
    [anon_sym_AMP] = ACTIONS(625),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(623),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_RBRACE] = ACTIONS(623),
    [anon_sym_EQ_GT] = ACTIONS(623),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(691),
    [anon_sym_QMARK] = ACTIONS(629),
    [anon_sym_PIPE] = ACTIONS(629),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(695),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(623),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_yield_expression_token2] = ACTIONS(709),
    [aux_sym_binary_expression_token1] = ACTIONS(629),
    [anon_sym_QMARK_QMARK] = ACTIONS(623),
    [anon_sym_STAR_STAR] = ACTIONS(623),
    [aux_sym_binary_expression_token2] = ACTIONS(629),
    [aux_sym_binary_expression_token3] = ACTIONS(629),
    [aux_sym_binary_expression_token4] = ACTIONS(629),
    [anon_sym_PIPE_PIPE] = ACTIONS(623),
    [anon_sym_AMP_AMP] = ACTIONS(623),
    [anon_sym_CARET] = ACTIONS(623),
    [anon_sym_EQ_EQ] = ACTIONS(629),
    [anon_sym_BANG_EQ] = ACTIONS(629),
    [anon_sym_LT_GT] = ACTIONS(623),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(623),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(623),
    [anon_sym_LT] = ACTIONS(629),
    [anon_sym_GT] = ACTIONS(629),
    [anon_sym_LT_EQ] = ACTIONS(629),
    [anon_sym_GT_EQ] = ACTIONS(623),
    [anon_sym_LT_EQ_GT] = ACTIONS(623),
    [anon_sym_LT_LT] = ACTIONS(629),
    [anon_sym_GT_GT] = ACTIONS(623),
    [anon_sym_DOT] = ACTIONS(629),
    [anon_sym_STAR] = ACTIONS(629),
    [anon_sym_SLASH] = ACTIONS(629),
    [anon_sym_PERCENT] = ACTIONS(623),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [119] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1131),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym_variadic_unpacking] = STATE(1106),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_by_ref] = STATE(1106),
    [sym_yield_expression] = STATE(1105),
    [sym_array_element_initializer] = STATE(1132),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [anon_sym_SEMI] = ACTIONS(623),
    [anon_sym_AMP] = ACTIONS(721),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(623),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(623),
    [anon_sym_EQ_GT] = ACTIONS(623),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(723),
    [anon_sym_QMARK] = ACTIONS(629),
    [anon_sym_PIPE] = ACTIONS(629),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_yield_expression_token2] = ACTIONS(725),
    [aux_sym_binary_expression_token1] = ACTIONS(629),
    [anon_sym_QMARK_QMARK] = ACTIONS(623),
    [anon_sym_STAR_STAR] = ACTIONS(623),
    [aux_sym_binary_expression_token2] = ACTIONS(629),
    [aux_sym_binary_expression_token3] = ACTIONS(629),
    [aux_sym_binary_expression_token4] = ACTIONS(629),
    [anon_sym_PIPE_PIPE] = ACTIONS(623),
    [anon_sym_AMP_AMP] = ACTIONS(623),
    [anon_sym_CARET] = ACTIONS(623),
    [anon_sym_EQ_EQ] = ACTIONS(629),
    [anon_sym_BANG_EQ] = ACTIONS(629),
    [anon_sym_LT_GT] = ACTIONS(623),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(623),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(623),
    [anon_sym_LT] = ACTIONS(629),
    [anon_sym_GT] = ACTIONS(629),
    [anon_sym_LT_EQ] = ACTIONS(629),
    [anon_sym_GT_EQ] = ACTIONS(623),
    [anon_sym_LT_EQ_GT] = ACTIONS(623),
    [anon_sym_LT_LT] = ACTIONS(629),
    [anon_sym_GT_GT] = ACTIONS(623),
    [anon_sym_DOT] = ACTIONS(629),
    [anon_sym_STAR] = ACTIONS(629),
    [anon_sym_SLASH] = ACTIONS(629),
    [anon_sym_PERCENT] = ACTIONS(623),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(623),
  },
  [120] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1159),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(994),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(625),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(623),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LBRACE] = ACTIONS(623),
    [anon_sym_EQ_GT] = ACTIONS(623),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(623),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [anon_sym_QMARK] = ACTIONS(629),
    [anon_sym_PIPE] = ACTIONS(629),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(733),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_yield_expression_token2] = ACTIONS(747),
    [aux_sym_binary_expression_token1] = ACTIONS(629),
    [anon_sym_QMARK_QMARK] = ACTIONS(623),
    [anon_sym_STAR_STAR] = ACTIONS(623),
    [aux_sym_binary_expression_token2] = ACTIONS(629),
    [aux_sym_binary_expression_token3] = ACTIONS(629),
    [aux_sym_binary_expression_token4] = ACTIONS(629),
    [anon_sym_PIPE_PIPE] = ACTIONS(623),
    [anon_sym_AMP_AMP] = ACTIONS(623),
    [anon_sym_CARET] = ACTIONS(623),
    [anon_sym_EQ_EQ] = ACTIONS(629),
    [anon_sym_BANG_EQ] = ACTIONS(629),
    [anon_sym_LT_GT] = ACTIONS(623),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(623),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(623),
    [anon_sym_LT] = ACTIONS(629),
    [anon_sym_GT] = ACTIONS(629),
    [anon_sym_LT_EQ] = ACTIONS(629),
    [anon_sym_GT_EQ] = ACTIONS(623),
    [anon_sym_LT_EQ_GT] = ACTIONS(623),
    [anon_sym_LT_LT] = ACTIONS(629),
    [anon_sym_GT_GT] = ACTIONS(623),
    [anon_sym_DOT] = ACTIONS(629),
    [anon_sym_STAR] = ACTIONS(629),
    [anon_sym_SLASH] = ACTIONS(629),
    [anon_sym_PERCENT] = ACTIONS(623),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [121] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_cast_type] = STATE(2570),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1237),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(757),
    [aux_sym_cast_type_token2] = ACTIONS(759),
    [aux_sym_cast_type_token3] = ACTIONS(759),
    [aux_sym_cast_type_token4] = ACTIONS(759),
    [aux_sym_cast_type_token5] = ACTIONS(759),
    [aux_sym_cast_type_token6] = ACTIONS(759),
    [aux_sym_cast_type_token7] = ACTIONS(759),
    [aux_sym_cast_type_token8] = ACTIONS(759),
    [aux_sym_cast_type_token9] = ACTIONS(759),
    [aux_sym_cast_type_token10] = ACTIONS(759),
    [aux_sym_cast_type_token11] = ACTIONS(759),
    [aux_sym_cast_type_token12] = ACTIONS(759),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [122] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_cast_type] = STATE(2491),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(757),
    [aux_sym_cast_type_token2] = ACTIONS(759),
    [aux_sym_cast_type_token3] = ACTIONS(759),
    [aux_sym_cast_type_token4] = ACTIONS(759),
    [aux_sym_cast_type_token5] = ACTIONS(759),
    [aux_sym_cast_type_token6] = ACTIONS(759),
    [aux_sym_cast_type_token7] = ACTIONS(759),
    [aux_sym_cast_type_token8] = ACTIONS(759),
    [aux_sym_cast_type_token9] = ACTIONS(759),
    [aux_sym_cast_type_token10] = ACTIONS(759),
    [aux_sym_cast_type_token11] = ACTIONS(759),
    [aux_sym_cast_type_token12] = ACTIONS(759),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [123] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_cast_type] = STATE(2520),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(757),
    [aux_sym_cast_type_token2] = ACTIONS(759),
    [aux_sym_cast_type_token3] = ACTIONS(759),
    [aux_sym_cast_type_token4] = ACTIONS(759),
    [aux_sym_cast_type_token5] = ACTIONS(759),
    [aux_sym_cast_type_token6] = ACTIONS(759),
    [aux_sym_cast_type_token7] = ACTIONS(759),
    [aux_sym_cast_type_token8] = ACTIONS(759),
    [aux_sym_cast_type_token9] = ACTIONS(759),
    [aux_sym_cast_type_token10] = ACTIONS(759),
    [aux_sym_cast_type_token11] = ACTIONS(759),
    [aux_sym_cast_type_token12] = ACTIONS(759),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [124] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_cast_type] = STATE(2604),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1237),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(757),
    [aux_sym_cast_type_token2] = ACTIONS(759),
    [aux_sym_cast_type_token3] = ACTIONS(759),
    [aux_sym_cast_type_token4] = ACTIONS(759),
    [aux_sym_cast_type_token5] = ACTIONS(759),
    [aux_sym_cast_type_token6] = ACTIONS(759),
    [aux_sym_cast_type_token7] = ACTIONS(759),
    [aux_sym_cast_type_token8] = ACTIONS(759),
    [aux_sym_cast_type_token9] = ACTIONS(759),
    [aux_sym_cast_type_token10] = ACTIONS(759),
    [aux_sym_cast_type_token11] = ACTIONS(759),
    [aux_sym_cast_type_token12] = ACTIONS(759),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [125] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_cast_type] = STATE(2604),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(757),
    [aux_sym_cast_type_token2] = ACTIONS(759),
    [aux_sym_cast_type_token3] = ACTIONS(759),
    [aux_sym_cast_type_token4] = ACTIONS(759),
    [aux_sym_cast_type_token5] = ACTIONS(759),
    [aux_sym_cast_type_token6] = ACTIONS(759),
    [aux_sym_cast_type_token7] = ACTIONS(759),
    [aux_sym_cast_type_token8] = ACTIONS(759),
    [aux_sym_cast_type_token9] = ACTIONS(759),
    [aux_sym_cast_type_token10] = ACTIONS(759),
    [aux_sym_cast_type_token11] = ACTIONS(759),
    [aux_sym_cast_type_token12] = ACTIONS(759),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [126] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_cast_type] = STATE(2554),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(757),
    [aux_sym_cast_type_token2] = ACTIONS(759),
    [aux_sym_cast_type_token3] = ACTIONS(759),
    [aux_sym_cast_type_token4] = ACTIONS(759),
    [aux_sym_cast_type_token5] = ACTIONS(759),
    [aux_sym_cast_type_token6] = ACTIONS(759),
    [aux_sym_cast_type_token7] = ACTIONS(759),
    [aux_sym_cast_type_token8] = ACTIONS(759),
    [aux_sym_cast_type_token9] = ACTIONS(759),
    [aux_sym_cast_type_token10] = ACTIONS(759),
    [aux_sym_cast_type_token11] = ACTIONS(759),
    [aux_sym_cast_type_token12] = ACTIONS(759),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [127] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_cast_type] = STATE(2605),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(757),
    [aux_sym_cast_type_token2] = ACTIONS(759),
    [aux_sym_cast_type_token3] = ACTIONS(759),
    [aux_sym_cast_type_token4] = ACTIONS(759),
    [aux_sym_cast_type_token5] = ACTIONS(759),
    [aux_sym_cast_type_token6] = ACTIONS(759),
    [aux_sym_cast_type_token7] = ACTIONS(759),
    [aux_sym_cast_type_token8] = ACTIONS(759),
    [aux_sym_cast_type_token9] = ACTIONS(759),
    [aux_sym_cast_type_token10] = ACTIONS(759),
    [aux_sym_cast_type_token11] = ACTIONS(759),
    [aux_sym_cast_type_token12] = ACTIONS(759),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [128] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_cast_type] = STATE(2462),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(757),
    [aux_sym_cast_type_token2] = ACTIONS(759),
    [aux_sym_cast_type_token3] = ACTIONS(759),
    [aux_sym_cast_type_token4] = ACTIONS(759),
    [aux_sym_cast_type_token5] = ACTIONS(759),
    [aux_sym_cast_type_token6] = ACTIONS(759),
    [aux_sym_cast_type_token7] = ACTIONS(759),
    [aux_sym_cast_type_token8] = ACTIONS(759),
    [aux_sym_cast_type_token9] = ACTIONS(759),
    [aux_sym_cast_type_token10] = ACTIONS(759),
    [aux_sym_cast_type_token11] = ACTIONS(759),
    [aux_sym_cast_type_token12] = ACTIONS(759),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [129] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_cast_type] = STATE(2486),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(757),
    [aux_sym_cast_type_token2] = ACTIONS(759),
    [aux_sym_cast_type_token3] = ACTIONS(759),
    [aux_sym_cast_type_token4] = ACTIONS(759),
    [aux_sym_cast_type_token5] = ACTIONS(759),
    [aux_sym_cast_type_token6] = ACTIONS(759),
    [aux_sym_cast_type_token7] = ACTIONS(759),
    [aux_sym_cast_type_token8] = ACTIONS(759),
    [aux_sym_cast_type_token9] = ACTIONS(759),
    [aux_sym_cast_type_token10] = ACTIONS(759),
    [aux_sym_cast_type_token11] = ACTIONS(759),
    [aux_sym_cast_type_token12] = ACTIONS(759),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [130] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_cast_type] = STATE(2562),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(757),
    [aux_sym_cast_type_token2] = ACTIONS(759),
    [aux_sym_cast_type_token3] = ACTIONS(759),
    [aux_sym_cast_type_token4] = ACTIONS(759),
    [aux_sym_cast_type_token5] = ACTIONS(759),
    [aux_sym_cast_type_token6] = ACTIONS(759),
    [aux_sym_cast_type_token7] = ACTIONS(759),
    [aux_sym_cast_type_token8] = ACTIONS(759),
    [aux_sym_cast_type_token9] = ACTIONS(759),
    [aux_sym_cast_type_token10] = ACTIONS(759),
    [aux_sym_cast_type_token11] = ACTIONS(759),
    [aux_sym_cast_type_token12] = ACTIONS(759),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [131] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_cast_type] = STATE(2568),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(757),
    [aux_sym_cast_type_token2] = ACTIONS(759),
    [aux_sym_cast_type_token3] = ACTIONS(759),
    [aux_sym_cast_type_token4] = ACTIONS(759),
    [aux_sym_cast_type_token5] = ACTIONS(759),
    [aux_sym_cast_type_token6] = ACTIONS(759),
    [aux_sym_cast_type_token7] = ACTIONS(759),
    [aux_sym_cast_type_token8] = ACTIONS(759),
    [aux_sym_cast_type_token9] = ACTIONS(759),
    [aux_sym_cast_type_token10] = ACTIONS(759),
    [aux_sym_cast_type_token11] = ACTIONS(759),
    [aux_sym_cast_type_token12] = ACTIONS(759),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [132] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_variadic_placeholder] = STATE(2627),
    [sym_argument] = STATE(1964),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(771),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(775),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [133] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_variadic_placeholder] = STATE(2502),
    [sym_argument] = STATE(1902),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(789),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(775),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [134] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_variadic_placeholder] = STATE(2538),
    [sym_argument] = STATE(2155),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(791),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(775),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [135] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1199),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(673),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(673),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(673),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(673),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2149),
    [sym__array_destructing_element] = STATE(2151),
    [sym_function_call_expression] = STATE(665),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(665),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(665),
    [sym_nullsafe_member_call_expression] = STATE(665),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(676),
    [sym__dereferencable_subscript_expression] = STATE(665),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(665),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2374),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2197),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym__array_destructing_repeat1] = STATE(2156),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(795),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(691),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_RBRACK] = ACTIONS(797),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [136] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1199),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(673),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(673),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(673),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(673),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2149),
    [sym__array_destructing_element] = STATE(2151),
    [sym_function_call_expression] = STATE(665),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(665),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(665),
    [sym_nullsafe_member_call_expression] = STATE(665),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(676),
    [sym__dereferencable_subscript_expression] = STATE(665),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(665),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2374),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2197),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym__array_destructing_repeat1] = STATE(2156),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(795),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(691),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_RBRACK] = ACTIONS(799),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [137] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_variadic_placeholder] = STATE(2447),
    [sym_argument] = STATE(1983),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(801),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(775),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [138] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_variadic_placeholder] = STATE(2555),
    [sym_argument] = STATE(1999),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(803),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(775),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [139] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_variadic_placeholder] = STATE(2490),
    [sym_argument] = STATE(2014),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(805),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(775),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [140] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1199),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(673),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(673),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(673),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(673),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2149),
    [sym__array_destructing_element] = STATE(2151),
    [sym_function_call_expression] = STATE(665),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(665),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(665),
    [sym_nullsafe_member_call_expression] = STATE(665),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(676),
    [sym__dereferencable_subscript_expression] = STATE(665),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(665),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2374),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2154),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym__array_destructing_repeat1] = STATE(2156),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(807),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(691),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_RBRACK] = ACTIONS(809),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [141] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1199),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(673),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(673),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(673),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(673),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2149),
    [sym__array_destructing_element] = STATE(2151),
    [sym_function_call_expression] = STATE(665),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(665),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(665),
    [sym_nullsafe_member_call_expression] = STATE(665),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(676),
    [sym__dereferencable_subscript_expression] = STATE(665),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(665),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2374),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2197),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym__array_destructing_repeat1] = STATE(2156),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(795),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(691),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_RBRACK] = ACTIONS(811),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [142] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1199),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(673),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(673),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(673),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(673),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2149),
    [sym__array_destructing_element] = STATE(2151),
    [sym_function_call_expression] = STATE(665),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(665),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(665),
    [sym_nullsafe_member_call_expression] = STATE(665),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(676),
    [sym__dereferencable_subscript_expression] = STATE(665),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(676),
    [sym__callable_variable] = STATE(665),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2374),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2197),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym__array_destructing_repeat1] = STATE(2156),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(795),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(691),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_RBRACK] = ACTIONS(813),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [143] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_argument] = STATE(2256),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(815),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [144] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_argument] = STATE(2256),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(817),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [145] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_argument] = STATE(2256),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(819),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [146] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_argument] = STATE(2256),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(821),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [147] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_argument] = STATE(2256),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(823),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [148] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_argument] = STATE(2256),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(825),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [149] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_argument] = STATE(2256),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(827),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [150] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_argument] = STATE(2256),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(829),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [151] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_argument] = STATE(2256),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(831),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [152] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_argument] = STATE(2256),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(833),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [153] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_argument] = STATE(2256),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(835),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [154] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_argument] = STATE(2256),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(837),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [155] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1185),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2154),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(839),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(691),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(841),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [156] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1204),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2110),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(843),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(845),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [157] = {
    [sym_reference_modifier] = STATE(223),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1215),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_argument] = STATE(2256),
    [sym__argument_name] = STATE(181),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2378),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1563),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(761),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(765),
    [aux_sym_namespace_definition_token1] = ACTIONS(767),
    [aux_sym__namespace_use_type_token1] = ACTIONS(769),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(773),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(777),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(779),
    [aux_sym_match_expression_token1] = ACTIONS(781),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(783),
    [anon_sym_parent] = ACTIONS(783),
    [aux_sym__argument_name_token1] = ACTIONS(785),
    [aux_sym__argument_name_token2] = ACTIONS(787),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [158] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1204),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(1883),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(847),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(849),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [159] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1185),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2197),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(851),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(691),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(853),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [160] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1204),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2066),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(855),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [161] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1204),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2066),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(857),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [162] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1204),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2066),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(859),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [163] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_match_condition_list] = STATE(2536),
    [sym_match_conditional_expression] = STATE(2070),
    [sym_match_default_expression] = STATE(2070),
    [sym_expression] = STATE(1201),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(861),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_RBRACE] = ACTIONS(863),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [aux_sym_match_default_expression_token1] = ACTIONS(865),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [164] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1185),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2066),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(691),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(867),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [165] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1185),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2066),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(691),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(869),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [166] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1204),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2066),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(871),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [167] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1185),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2066),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(691),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(873),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [168] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1185),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2066),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(691),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(875),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [169] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_match_condition_list] = STATE(2536),
    [sym_match_conditional_expression] = STATE(1915),
    [sym_match_default_expression] = STATE(1915),
    [sym_expression] = STATE(1201),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(877),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_RBRACE] = ACTIONS(879),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [aux_sym_match_default_expression_token1] = ACTIONS(865),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [170] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1277),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(678),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(678),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(678),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(678),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2149),
    [sym__array_destructing_element] = STATE(2303),
    [sym_function_call_expression] = STATE(668),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(668),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(668),
    [sym_nullsafe_member_call_expression] = STATE(668),
    [sym__variable_subscript_expression] = STATE(679),
    [sym__dereferencable_subscript_expression] = STATE(668),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(679),
    [sym__callable_variable] = STATE(668),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2303),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(881),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(883),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(885),
    [anon_sym_RBRACK] = ACTIONS(881),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [171] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_match_condition_list] = STATE(2536),
    [sym_match_conditional_expression] = STATE(2307),
    [sym_match_default_expression] = STATE(2307),
    [sym_expression] = STATE(1201),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_RBRACE] = ACTIONS(887),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [aux_sym_match_default_expression_token1] = ACTIONS(865),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [172] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1185),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2066),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(691),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [173] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1204),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(978),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(978),
    [sym_yield_expression] = STATE(1004),
    [sym_array_element_initializer] = STATE(2066),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [174] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1276),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(671),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(671),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(671),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(671),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(1794),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(647),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(647),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(647),
    [sym_nullsafe_member_call_expression] = STATE(647),
    [sym__variable_subscript_expression] = STATE(664),
    [sym__dereferencable_subscript_expression] = STATE(647),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(664),
    [sym__callable_variable] = STATE(647),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2068),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym__list_destructing_repeat1] = STATE(2069),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(889),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(883),
    [anon_sym_RPAREN2] = ACTIONS(891),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [175] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1277),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(678),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(678),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(678),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(678),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2149),
    [sym__array_destructing_element] = STATE(2303),
    [sym_function_call_expression] = STATE(668),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(668),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(668),
    [sym_nullsafe_member_call_expression] = STATE(668),
    [sym__variable_subscript_expression] = STATE(679),
    [sym__dereferencable_subscript_expression] = STATE(668),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(679),
    [sym__callable_variable] = STATE(668),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2303),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(881),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(883),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(885),
    [anon_sym_RBRACK] = ACTIONS(893),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [176] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1277),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(678),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(678),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(678),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(678),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2149),
    [sym__array_destructing_element] = STATE(2303),
    [sym_function_call_expression] = STATE(668),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(668),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(668),
    [sym_nullsafe_member_call_expression] = STATE(668),
    [sym__variable_subscript_expression] = STATE(679),
    [sym__dereferencable_subscript_expression] = STATE(668),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(679),
    [sym__callable_variable] = STATE(668),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2303),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(881),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(883),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(885),
    [anon_sym_RBRACK] = ACTIONS(896),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [177] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_match_condition_list] = STATE(2536),
    [sym_match_conditional_expression] = STATE(2307),
    [sym_match_default_expression] = STATE(2307),
    [sym_expression] = STATE(1201),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_RBRACE] = ACTIONS(899),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [aux_sym_match_default_expression_token1] = ACTIONS(865),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [178] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_match_condition_list] = STATE(2536),
    [sym_match_conditional_expression] = STATE(2307),
    [sym_match_default_expression] = STATE(2307),
    [sym_expression] = STATE(1201),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_RBRACE] = ACTIONS(901),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [aux_sym_match_default_expression_token1] = ACTIONS(865),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [179] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_match_condition_list] = STATE(2536),
    [sym_match_conditional_expression] = STATE(2307),
    [sym_match_default_expression] = STATE(2307),
    [sym_expression] = STATE(1201),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_RBRACE] = ACTIONS(903),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [aux_sym_match_default_expression_token1] = ACTIONS(865),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [180] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_foreach_pair] = STATE(2485),
    [sym__foreach_value] = STATE(2485),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1205),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2259),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2485),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [181] = {
    [sym_reference_modifier] = STATE(222),
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1226),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2320),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1541),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(763),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [182] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1238),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(677),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(677),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(677),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(677),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2122),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(670),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(670),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(670),
    [sym_nullsafe_member_call_expression] = STATE(670),
    [sym__variable_subscript_expression] = STATE(682),
    [sym__dereferencable_subscript_expression] = STATE(670),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(682),
    [sym__callable_variable] = STATE(670),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2350),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [anon_sym_COMMA] = ACTIONS(905),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(883),
    [anon_sym_RPAREN2] = ACTIONS(905),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [183] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_foreach_pair] = STATE(2488),
    [sym__foreach_value] = STATE(2488),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1205),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2259),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2488),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [184] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_foreach_pair] = STATE(2471),
    [sym__foreach_value] = STATE(2471),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1205),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2259),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2471),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [185] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_match_condition_list] = STATE(2536),
    [sym_match_conditional_expression] = STATE(2307),
    [sym_match_default_expression] = STATE(2307),
    [sym_expression] = STATE(1201),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [aux_sym_match_default_expression_token1] = ACTIONS(865),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [186] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_foreach_pair] = STATE(2596),
    [sym__foreach_value] = STATE(2596),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1205),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2259),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2596),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [187] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2492),
    [sym_sequence_expression] = STATE(2492),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1219),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(907),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [188] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2640),
    [sym_sequence_expression] = STATE(2640),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1219),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(909),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [189] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2521),
    [sym_sequence_expression] = STATE(2521),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(911),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [190] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1193),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(465),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [anon_sym_SEMI] = ACTIONS(913),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(913),
  },
  [191] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2512),
    [sym_sequence_expression] = STATE(2512),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1219),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(915),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [192] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2528),
    [sym_sequence_expression] = STATE(2528),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1219),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(917),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [193] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2647),
    [sym_sequence_expression] = STATE(2647),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1219),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(919),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [194] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1194),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(466),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [anon_sym_SEMI] = ACTIONS(921),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(921),
  },
  [195] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1196),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(2056),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [anon_sym_SEMI] = ACTIONS(923),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(923),
  },
  [196] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2609),
    [sym_sequence_expression] = STATE(2609),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(925),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [197] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2420),
    [sym_sequence_expression] = STATE(2420),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1219),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(927),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [198] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1197),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(2067),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [anon_sym_SEMI] = ACTIONS(929),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(929),
  },
  [199] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1180),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(2075),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [anon_sym_SEMI] = ACTIONS(931),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(931),
  },
  [200] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2548),
    [sym_sequence_expression] = STATE(2548),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(933),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [201] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2522),
    [sym_sequence_expression] = STATE(2522),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(935),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [202] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1188),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [sym__semicolon] = STATE(483),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [anon_sym_SEMI] = ACTIONS(937),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
    [sym__automatic_semicolon] = ACTIONS(937),
  },
  [203] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2564),
    [sym_sequence_expression] = STATE(2564),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(939),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [204] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2567),
    [sym_sequence_expression] = STATE(2567),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(941),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [205] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2468),
    [sym_sequence_expression] = STATE(2468),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(943),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [206] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2472),
    [sym_sequence_expression] = STATE(2472),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(945),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [207] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2473),
    [sym_sequence_expression] = STATE(2473),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(947),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [208] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2475),
    [sym_sequence_expression] = STATE(2475),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(949),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [209] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__foreach_value] = STATE(2546),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1281),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2259),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2546),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [210] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2489),
    [sym_sequence_expression] = STATE(2489),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1219),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(951),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [211] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2551),
    [sym_sequence_expression] = STATE(2551),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1219),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(953),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [212] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2482),
    [sym_sequence_expression] = STATE(2482),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(955),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [213] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2585),
    [sym_sequence_expression] = STATE(2585),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1219),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(957),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [214] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2593),
    [sym_sequence_expression] = STATE(2593),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(959),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [215] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2597),
    [sym_sequence_expression] = STATE(2597),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(961),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [216] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2598),
    [sym_sequence_expression] = STATE(2598),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(963),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [217] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2600),
    [sym_sequence_expression] = STATE(2600),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(965),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [218] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2615),
    [sym_sequence_expression] = STATE(2615),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1219),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(967),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [219] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2510),
    [sym_sequence_expression] = STATE(2510),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1219),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(969),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [220] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2637),
    [sym_sequence_expression] = STATE(2637),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1219),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_SEMI] = ACTIONS(971),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [221] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__expressions] = STATE(2542),
    [sym_sequence_expression] = STATE(2542),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1225),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(973),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [222] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1206),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2299),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1533),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [223] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1230),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym_variadic_unpacking] = STATE(2262),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1534),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(729),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [224] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1055),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(693),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(693),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(693),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(693),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2201),
    [sym_function_call_expression] = STATE(680),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(680),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(680),
    [sym_nullsafe_member_call_expression] = STATE(680),
    [sym__variable_subscript_expression] = STATE(695),
    [sym__dereferencable_subscript_expression] = STATE(680),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(695),
    [sym__callable_variable] = STATE(680),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(2329),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [225] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym__expressions] = STATE(2214),
    [sym_sequence_expression] = STATE(2214),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1189),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [226] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1055),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(1007),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [227] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1154),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(1007),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(793),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [228] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1031),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(1007),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(975),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [229] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1154),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(1007),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(975),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [230] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1104),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_by_ref] = STATE(1110),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [anon_sym_AMP] = ACTIONS(977),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [231] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym__expressions] = STATE(2192),
    [sym_sequence_expression] = STATE(2192),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1189),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [232] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1055),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_by_ref] = STATE(1007),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(975),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [233] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_sequence_expression] = STATE(2302),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1227),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [234] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1054),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [anon_sym_AMP] = ACTIONS(979),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [235] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1210),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_EQ_GT] = ACTIONS(981),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [236] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_sequence_expression] = STATE(2318),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1183),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [237] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1157),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_COLON] = ACTIONS(983),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [238] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1271),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(985),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [239] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1234),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(987),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [240] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1242),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [anon_sym_RPAREN2] = ACTIONS(989),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [241] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1148),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(991),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [242] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1279),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_COLON] = ACTIONS(993),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [243] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1282),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_COLON] = ACTIONS(995),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [244] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1266),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(997),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [245] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1037),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(999),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [246] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1247),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_COLON] = ACTIONS(1001),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [247] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1153),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [anon_sym_RPAREN2] = ACTIONS(1003),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [248] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1076),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [anon_sym_AMP] = ACTIONS(1005),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [249] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_sequence_expression] = STATE(2302),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1220),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [250] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1244),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [anon_sym_RPAREN2] = ACTIONS(1007),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [251] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1256),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(1009),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [252] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1248),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(1011),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [253] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1261),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(1013),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [254] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1249),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(1015),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [255] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1270),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(1017),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [256] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1262),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_RBRACK] = ACTIONS(1019),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [257] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1210),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_EQ_GT] = ACTIONS(1021),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [258] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1283),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_COLON] = ACTIONS(1023),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [259] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1046),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [260] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1111),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [261] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1113),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [262] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1114),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [263] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1115),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [264] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1116),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [265] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1117),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [266] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1118),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [267] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1120),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [268] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1173),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [269] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1191),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [270] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1178),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [271] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1174),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [272] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1207),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [273] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1200),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [274] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1121),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [275] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1130),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [276] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1134),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [277] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1135),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [278] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1047),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [279] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1136),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [280] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1050),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [281] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1210),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [282] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1051),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [283] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1175),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [284] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1087),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [285] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1203),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [286] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1028),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [287] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(991),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [288] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(977),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [289] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1030),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [290] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1026),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [291] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1027),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [292] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1033),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [293] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1034),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [294] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1202),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [295] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1075),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [296] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1036),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [297] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1012),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [298] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1041),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [299] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1044),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [300] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1013),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [301] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1035),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [302] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1040),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [303] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1043),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [304] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1015),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [305] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1017),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [306] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1018),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [307] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1019),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [308] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1021),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [309] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1022),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [310] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1023),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [311] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1024),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [312] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1025),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [313] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1032),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [314] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1039),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [315] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1042),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [316] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1014),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [317] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1038),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [318] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1100),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [319] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1102),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [320] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1237),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [321] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1053),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [322] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1085),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [323] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1079),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [324] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1239),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [325] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1240),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [326] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1212),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [327] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1286),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [328] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1097),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [329] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1133),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [330] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1252),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [331] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1086),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [332] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1091),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [333] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1217),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [334] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1109),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [335] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1107),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [336] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1243),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [337] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1029),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [338] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1245),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [339] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1246),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [340] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1250),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [341] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1251),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [342] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1222),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [343] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1170),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [344] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1229),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [345] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1221),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [346] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1122),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [347] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(991),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [348] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(977),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [349] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1123),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [350] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1125),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [351] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1126),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [352] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1127),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [353] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1253),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [354] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1056),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [355] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1058),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [356] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1059),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [357] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1060),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [358] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1061),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [359] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1062),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [360] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1063),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [361] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1064),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [362] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1065),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [363] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1066),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [364] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1067),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [365] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1068),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [366] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1069),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [367] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1070),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [368] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1071),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [369] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1072),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [370] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1073),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [371] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1074),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [372] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1077),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [373] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1078),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [374] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1080),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [375] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1081),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [376] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1082),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [377] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1224),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [378] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1218),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [379] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1278),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [380] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1280),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [381] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1259),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [382] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1153),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [383] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1254),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [384] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1255),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [385] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1093),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [386] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1258),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [387] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1157),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [388] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(991),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [389] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(977),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [390] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1158),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [391] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1160),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [392] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1161),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [393] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1162),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [394] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1163),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [395] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1151),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [396] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1156),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [397] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1164),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [398] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1165),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [399] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1166),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [400] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1167),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [401] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1168),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [402] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1137),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [403] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1138),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [404] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1139),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [405] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1140),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [406] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1141),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [407] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1142),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [408] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1143),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [409] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1144),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [410] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1145),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [411] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1146),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [412] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1147),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [413] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1149),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [414] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1150),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [415] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1152),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [416] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1155),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [417] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1264),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [418] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1260),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [419] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1263),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [420] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1265),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [421] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1268),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [422] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1269),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [423] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1272),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [424] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1273),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [425] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1274),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [426] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1275),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [427] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1177),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [428] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1169),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [429] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1209),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [430] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1214),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [431] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1179),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [432] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1213),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(667),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(667),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(667),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(667),
    [sym_list_literal] = STATE(2553),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(660),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(660),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(660),
    [sym_nullsafe_member_call_expression] = STATE(660),
    [sym__variable_subscript_expression] = STATE(666),
    [sym__dereferencable_subscript_expression] = STATE(660),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(666),
    [sym__callable_variable] = STATE(660),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(743),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(745),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [aux_sym_require_expression_token1] = ACTIONS(753),
    [aux_sym_require_once_expression_token1] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
  },
  [433] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1171),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [434] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_match_expression] = STATE(1105),
    [sym_expression] = STATE(1198),
    [sym__unary_expression] = STATE(1083),
    [sym_unary_op_expression] = STATE(1083),
    [sym_error_suppression_expression] = STATE(1105),
    [sym_clone_expression] = STATE(1083),
    [sym_primary_expression] = STATE(1083),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1083),
    [sym_cast_variable] = STATE(663),
    [sym_assignment_expression] = STATE(1105),
    [sym_reference_assignment_expression] = STATE(1105),
    [sym_conditional_expression] = STATE(1105),
    [sym_augmented_assignment_expression] = STATE(1105),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(663),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(663),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(663),
    [sym_list_literal] = STATE(2539),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(638),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(638),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(638),
    [sym_nullsafe_member_call_expression] = STATE(638),
    [sym__variable_subscript_expression] = STATE(659),
    [sym__dereferencable_subscript_expression] = STATE(638),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(659),
    [sym__callable_variable] = STATE(638),
    [sym_variable_name] = STATE(640),
    [sym_yield_expression] = STATE(1105),
    [sym_binary_expression] = STATE(1105),
    [sym_include_expression] = STATE(1105),
    [sym_include_once_expression] = STATE(1105),
    [sym_require_expression] = STATE(1105),
    [sym_require_once_expression] = STATE(1105),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_match_expression_token1] = ACTIONS(87),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(107),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_yield_expression_token1] = ACTIONS(127),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [aux_sym_require_expression_token1] = ACTIONS(133),
    [aux_sym_require_once_expression_token1] = ACTIONS(135),
    [sym_comment] = ACTIONS(3),
  },
  [435] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1284),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [436] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1285),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(643),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(643),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(643),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(643),
    [sym_list_literal] = STATE(2603),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(616),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(616),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(616),
    [sym_nullsafe_member_call_expression] = STATE(616),
    [sym__variable_subscript_expression] = STATE(621),
    [sym__dereferencable_subscript_expression] = STATE(616),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(621),
    [sym__callable_variable] = STATE(616),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(661),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(677),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [aux_sym_require_expression_token1] = ACTIONS(685),
    [aux_sym_require_once_expression_token1] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
  },
  [437] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_match_expression] = STATE(1004),
    [sym_expression] = STATE(1128),
    [sym__unary_expression] = STATE(988),
    [sym_unary_op_expression] = STATE(988),
    [sym_error_suppression_expression] = STATE(1004),
    [sym_clone_expression] = STATE(988),
    [sym_primary_expression] = STATE(988),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(988),
    [sym_cast_variable] = STATE(646),
    [sym_assignment_expression] = STATE(1004),
    [sym_reference_assignment_expression] = STATE(1004),
    [sym_conditional_expression] = STATE(1004),
    [sym_augmented_assignment_expression] = STATE(1004),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(646),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(646),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(646),
    [sym_list_literal] = STATE(2461),
    [sym__list_destructing] = STATE(2297),
    [sym__array_destructing] = STATE(2297),
    [sym_function_call_expression] = STATE(644),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(644),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(644),
    [sym_nullsafe_member_call_expression] = STATE(644),
    [sym__variable_subscript_expression] = STATE(662),
    [sym__dereferencable_subscript_expression] = STATE(644),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(662),
    [sym__callable_variable] = STATE(644),
    [sym_variable_name] = STATE(594),
    [sym_yield_expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym_include_expression] = STATE(1004),
    [sym_include_once_expression] = STATE(1004),
    [sym_require_expression] = STATE(1004),
    [sym_require_once_expression] = STATE(1004),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_match_expression_token1] = ACTIONS(645),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [aux_sym__list_destructing_token1] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(705),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_yield_expression_token1] = ACTIONS(707),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [aux_sym_require_expression_token1] = ACTIONS(715),
    [aux_sym_require_once_expression_token1] = ACTIONS(717),
    [sym_comment] = ACTIONS(3),
  },
  [438] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__unary_expression] = STATE(982),
    [sym_unary_op_expression] = STATE(982),
    [sym_error_suppression_expression] = STATE(982),
    [sym_clone_expression] = STATE(982),
    [sym_primary_expression] = STATE(982),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(982),
    [sym_cast_variable] = STATE(614),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(614),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(614),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(614),
    [sym_function_call_expression] = STATE(603),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(603),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(603),
    [sym_nullsafe_member_call_expression] = STATE(603),
    [sym__variable_subscript_expression] = STATE(613),
    [sym__dereferencable_subscript_expression] = STATE(603),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(613),
    [sym__callable_variable] = STATE(603),
    [sym_variable_name] = STATE(594),
    [sym_include_expression] = STATE(982),
    [sym_include_once_expression] = STATE(982),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(727),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [anon_sym_PLUS] = ACTIONS(733),
    [anon_sym_DASH] = ACTIONS(733),
    [anon_sym_TILDE] = ACTIONS(735),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_AT] = ACTIONS(737),
    [aux_sym_clone_expression_token1] = ACTIONS(739),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [anon_sym_LBRACK] = ACTIONS(1025),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_include_expression_token1] = ACTIONS(749),
    [aux_sym_include_once_expression_token1] = ACTIONS(751),
    [sym_comment] = ACTIONS(3),
  },
  [439] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__unary_expression] = STATE(982),
    [sym_unary_op_expression] = STATE(982),
    [sym_error_suppression_expression] = STATE(982),
    [sym_clone_expression] = STATE(982),
    [sym_primary_expression] = STATE(982),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(982),
    [sym_cast_variable] = STATE(669),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(669),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(669),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(669),
    [sym_function_call_expression] = STATE(658),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(658),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(658),
    [sym_nullsafe_member_call_expression] = STATE(658),
    [sym__variable_subscript_expression] = STATE(672),
    [sym__dereferencable_subscript_expression] = STATE(658),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(672),
    [sym__callable_variable] = STATE(658),
    [sym_variable_name] = STATE(594),
    [sym_include_expression] = STATE(982),
    [sym_include_once_expression] = STATE(982),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(883),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [anon_sym_LBRACK] = ACTIONS(1025),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [sym_comment] = ACTIONS(3),
  },
  [440] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1579),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym__unary_expression] = STATE(1094),
    [sym_unary_op_expression] = STATE(1094),
    [sym_error_suppression_expression] = STATE(1094),
    [sym_clone_expression] = STATE(1094),
    [sym_primary_expression] = STATE(1094),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_expression] = STATE(1094),
    [sym_cast_variable] = STATE(654),
    [sym__variable_member_access_expression] = STATE(653),
    [sym_member_access_expression] = STATE(654),
    [sym__variable_nullsafe_member_access_expression] = STATE(650),
    [sym_nullsafe_member_access_expression] = STATE(654),
    [sym__variable_scoped_property_access_expression] = STATE(649),
    [sym_scoped_property_access_expression] = STATE(654),
    [sym_function_call_expression] = STATE(626),
    [sym__callable_expression] = STATE(2298),
    [sym_scoped_call_expression] = STATE(626),
    [sym__scope_resolution_qualifier] = STATE(2569),
    [sym_relative_scope] = STATE(2569),
    [sym_member_call_expression] = STATE(626),
    [sym_nullsafe_member_call_expression] = STATE(626),
    [sym__variable_subscript_expression] = STATE(655),
    [sym__dereferencable_subscript_expression] = STATE(626),
    [sym__dereferencable_expression] = STATE(1796),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(640),
    [sym__simple_variable] = STATE(640),
    [sym__new_variable] = STATE(655),
    [sym__callable_variable] = STATE(626),
    [sym_variable_name] = STATE(640),
    [sym_include_expression] = STATE(1094),
    [sym_include_once_expression] = STATE(1094),
    [sym__reserved_identifier] = STATE(1579),
    [sym__identifier] = STATE(1579),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(49),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [anon_sym_PLUS] = ACTIONS(91),
    [anon_sym_DASH] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_AT] = ACTIONS(95),
    [aux_sym_clone_expression_token1] = ACTIONS(97),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [anon_sym_LBRACK] = ACTIONS(1027),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(125),
    [aux_sym_include_expression_token1] = ACTIONS(129),
    [aux_sym_include_once_expression_token1] = ACTIONS(131),
    [sym_comment] = ACTIONS(3),
  },
  [441] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__unary_expression] = STATE(982),
    [sym_unary_op_expression] = STATE(982),
    [sym_error_suppression_expression] = STATE(982),
    [sym_clone_expression] = STATE(982),
    [sym_primary_expression] = STATE(982),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(982),
    [sym_cast_variable] = STATE(614),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(614),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(614),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(614),
    [sym_function_call_expression] = STATE(603),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(603),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(603),
    [sym_nullsafe_member_call_expression] = STATE(603),
    [sym__variable_subscript_expression] = STATE(613),
    [sym__dereferencable_subscript_expression] = STATE(603),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(613),
    [sym__callable_variable] = STATE(603),
    [sym_variable_name] = STATE(594),
    [sym_include_expression] = STATE(982),
    [sym_include_once_expression] = STATE(982),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(635),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [anon_sym_PLUS] = ACTIONS(647),
    [anon_sym_DASH] = ACTIONS(647),
    [anon_sym_TILDE] = ACTIONS(649),
    [anon_sym_BANG] = ACTIONS(649),
    [anon_sym_AT] = ACTIONS(651),
    [aux_sym_clone_expression_token1] = ACTIONS(653),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [anon_sym_LBRACK] = ACTIONS(1025),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_include_expression_token1] = ACTIONS(681),
    [aux_sym_include_once_expression_token1] = ACTIONS(683),
    [sym_comment] = ACTIONS(3),
  },
  [442] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1591),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym__unary_expression] = STATE(982),
    [sym_unary_op_expression] = STATE(982),
    [sym_error_suppression_expression] = STATE(982),
    [sym_clone_expression] = STATE(982),
    [sym_primary_expression] = STATE(982),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_expression] = STATE(982),
    [sym_cast_variable] = STATE(614),
    [sym__variable_member_access_expression] = STATE(605),
    [sym_member_access_expression] = STATE(614),
    [sym__variable_nullsafe_member_access_expression] = STATE(612),
    [sym_nullsafe_member_access_expression] = STATE(614),
    [sym__variable_scoped_property_access_expression] = STATE(592),
    [sym_scoped_property_access_expression] = STATE(614),
    [sym_function_call_expression] = STATE(603),
    [sym__callable_expression] = STATE(2410),
    [sym_scoped_call_expression] = STATE(603),
    [sym__scope_resolution_qualifier] = STATE(2419),
    [sym_relative_scope] = STATE(2419),
    [sym_member_call_expression] = STATE(603),
    [sym_nullsafe_member_call_expression] = STATE(603),
    [sym__variable_subscript_expression] = STATE(613),
    [sym__dereferencable_subscript_expression] = STATE(603),
    [sym__dereferencable_expression] = STATE(1777),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(594),
    [sym__simple_variable] = STATE(594),
    [sym__new_variable] = STATE(613),
    [sym__callable_variable] = STATE(603),
    [sym_variable_name] = STATE(594),
    [sym_include_expression] = STATE(982),
    [sym_include_once_expression] = STATE(982),
    [sym__reserved_identifier] = STATE(1591),
    [sym__identifier] = STATE(1591),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(689),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(697),
    [anon_sym_BANG] = ACTIONS(697),
    [anon_sym_AT] = ACTIONS(699),
    [aux_sym_clone_expression_token1] = ACTIONS(701),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [anon_sym_LBRACK] = ACTIONS(1025),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(675),
    [aux_sym_include_expression_token1] = ACTIONS(711),
    [aux_sym_include_once_expression_token1] = ACTIONS(713),
    [sym_comment] = ACTIONS(3),
  },
  [443] = {
    [sym_catch_clause] = STATE(444),
    [sym_finally_clause] = STATE(444),
    [aux_sym_try_statement_repeat1] = STATE(444),
    [ts_builtin_sym_end] = ACTIONS(1029),
    [sym_name] = ACTIONS(1031),
    [anon_sym_QMARK_GT] = ACTIONS(1029),
    [anon_sym_SEMI] = ACTIONS(1029),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1031),
    [aux_sym_global_declaration_token1] = ACTIONS(1031),
    [aux_sym_namespace_definition_token1] = ACTIONS(1031),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1031),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1031),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1031),
    [anon_sym_BSLASH] = ACTIONS(1029),
    [anon_sym_LBRACE] = ACTIONS(1029),
    [anon_sym_RBRACE] = ACTIONS(1029),
    [aux_sym_trait_declaration_token1] = ACTIONS(1031),
    [aux_sym_interface_declaration_token1] = ACTIONS(1031),
    [aux_sym_enum_declaration_token1] = ACTIONS(1031),
    [aux_sym_enum_case_token1] = ACTIONS(1031),
    [aux_sym_class_declaration_token1] = ACTIONS(1031),
    [aux_sym_final_modifier_token1] = ACTIONS(1031),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1031),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1031),
    [sym_var_modifier] = ACTIONS(1031),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1031),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1031),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1031),
    [anon_sym_LPAREN2] = ACTIONS(1029),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1031),
    [aux_sym_cast_type_token1] = ACTIONS(1031),
    [aux_sym_echo_statement_token1] = ACTIONS(1031),
    [aux_sym_exit_statement_token1] = ACTIONS(1031),
    [anon_sym_unset] = ACTIONS(1031),
    [aux_sym_declare_statement_token1] = ACTIONS(1031),
    [aux_sym_declare_statement_token2] = ACTIONS(1031),
    [sym_float] = ACTIONS(1031),
    [aux_sym_try_statement_token1] = ACTIONS(1031),
    [aux_sym_catch_clause_token1] = ACTIONS(1033),
    [aux_sym_finally_clause_token1] = ACTIONS(1035),
    [aux_sym_goto_statement_token1] = ACTIONS(1031),
    [aux_sym_continue_statement_token1] = ACTIONS(1031),
    [aux_sym_break_statement_token1] = ACTIONS(1031),
    [sym_integer] = ACTIONS(1031),
    [aux_sym_return_statement_token1] = ACTIONS(1031),
    [aux_sym_throw_expression_token1] = ACTIONS(1031),
    [aux_sym_while_statement_token1] = ACTIONS(1031),
    [aux_sym_while_statement_token2] = ACTIONS(1031),
    [aux_sym_do_statement_token1] = ACTIONS(1031),
    [aux_sym_for_statement_token1] = ACTIONS(1031),
    [aux_sym_for_statement_token2] = ACTIONS(1031),
    [aux_sym_foreach_statement_token1] = ACTIONS(1031),
    [aux_sym_foreach_statement_token2] = ACTIONS(1031),
    [aux_sym_if_statement_token1] = ACTIONS(1031),
    [aux_sym_if_statement_token2] = ACTIONS(1031),
    [aux_sym_else_if_clause_token1] = ACTIONS(1031),
    [aux_sym_else_clause_token1] = ACTIONS(1031),
    [aux_sym_match_expression_token1] = ACTIONS(1031),
    [aux_sym_match_default_expression_token1] = ACTIONS(1031),
    [aux_sym_switch_statement_token1] = ACTIONS(1031),
    [aux_sym_switch_block_token1] = ACTIONS(1031),
    [anon_sym_PLUS] = ACTIONS(1031),
    [anon_sym_DASH] = ACTIONS(1031),
    [anon_sym_TILDE] = ACTIONS(1029),
    [anon_sym_BANG] = ACTIONS(1029),
    [anon_sym_AT] = ACTIONS(1029),
    [aux_sym_clone_expression_token1] = ACTIONS(1031),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1031),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1031),
    [anon_sym_DASH_DASH] = ACTIONS(1029),
    [anon_sym_PLUS_PLUS] = ACTIONS(1029),
    [aux_sym__list_destructing_token1] = ACTIONS(1031),
    [anon_sym_LBRACK] = ACTIONS(1029),
    [anon_sym_self] = ACTIONS(1031),
    [anon_sym_parent] = ACTIONS(1031),
    [aux_sym__argument_name_token1] = ACTIONS(1031),
    [aux_sym__argument_name_token2] = ACTIONS(1031),
    [anon_sym_POUND_LBRACK] = ACTIONS(1029),
    [aux_sym_encapsed_string_token1] = ACTIONS(1029),
    [anon_sym_DQUOTE] = ACTIONS(1029),
    [aux_sym_string_token1] = ACTIONS(1029),
    [anon_sym_SQUOTE] = ACTIONS(1029),
    [anon_sym_LT_LT_LT] = ACTIONS(1029),
    [anon_sym_BQUOTE] = ACTIONS(1029),
    [anon_sym_DOLLAR] = ACTIONS(1029),
    [aux_sym_yield_expression_token1] = ACTIONS(1031),
    [aux_sym_include_expression_token1] = ACTIONS(1031),
    [aux_sym_include_once_expression_token1] = ACTIONS(1031),
    [aux_sym_require_expression_token1] = ACTIONS(1031),
    [aux_sym_require_once_expression_token1] = ACTIONS(1031),
    [sym_comment] = ACTIONS(3),
  },
  [444] = {
    [sym_catch_clause] = STATE(444),
    [sym_finally_clause] = STATE(444),
    [aux_sym_try_statement_repeat1] = STATE(444),
    [ts_builtin_sym_end] = ACTIONS(1037),
    [sym_name] = ACTIONS(1039),
    [anon_sym_QMARK_GT] = ACTIONS(1037),
    [anon_sym_SEMI] = ACTIONS(1037),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1039),
    [aux_sym_global_declaration_token1] = ACTIONS(1039),
    [aux_sym_namespace_definition_token1] = ACTIONS(1039),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1039),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1039),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1039),
    [anon_sym_BSLASH] = ACTIONS(1037),
    [anon_sym_LBRACE] = ACTIONS(1037),
    [anon_sym_RBRACE] = ACTIONS(1037),
    [aux_sym_trait_declaration_token1] = ACTIONS(1039),
    [aux_sym_interface_declaration_token1] = ACTIONS(1039),
    [aux_sym_enum_declaration_token1] = ACTIONS(1039),
    [aux_sym_enum_case_token1] = ACTIONS(1039),
    [aux_sym_class_declaration_token1] = ACTIONS(1039),
    [aux_sym_final_modifier_token1] = ACTIONS(1039),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1039),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1039),
    [sym_var_modifier] = ACTIONS(1039),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1039),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1039),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1039),
    [anon_sym_LPAREN2] = ACTIONS(1037),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1039),
    [aux_sym_cast_type_token1] = ACTIONS(1039),
    [aux_sym_echo_statement_token1] = ACTIONS(1039),
    [aux_sym_exit_statement_token1] = ACTIONS(1039),
    [anon_sym_unset] = ACTIONS(1039),
    [aux_sym_declare_statement_token1] = ACTIONS(1039),
    [aux_sym_declare_statement_token2] = ACTIONS(1039),
    [sym_float] = ACTIONS(1039),
    [aux_sym_try_statement_token1] = ACTIONS(1039),
    [aux_sym_catch_clause_token1] = ACTIONS(1041),
    [aux_sym_finally_clause_token1] = ACTIONS(1044),
    [aux_sym_goto_statement_token1] = ACTIONS(1039),
    [aux_sym_continue_statement_token1] = ACTIONS(1039),
    [aux_sym_break_statement_token1] = ACTIONS(1039),
    [sym_integer] = ACTIONS(1039),
    [aux_sym_return_statement_token1] = ACTIONS(1039),
    [aux_sym_throw_expression_token1] = ACTIONS(1039),
    [aux_sym_while_statement_token1] = ACTIONS(1039),
    [aux_sym_while_statement_token2] = ACTIONS(1039),
    [aux_sym_do_statement_token1] = ACTIONS(1039),
    [aux_sym_for_statement_token1] = ACTIONS(1039),
    [aux_sym_for_statement_token2] = ACTIONS(1039),
    [aux_sym_foreach_statement_token1] = ACTIONS(1039),
    [aux_sym_foreach_statement_token2] = ACTIONS(1039),
    [aux_sym_if_statement_token1] = ACTIONS(1039),
    [aux_sym_if_statement_token2] = ACTIONS(1039),
    [aux_sym_else_if_clause_token1] = ACTIONS(1039),
    [aux_sym_else_clause_token1] = ACTIONS(1039),
    [aux_sym_match_expression_token1] = ACTIONS(1039),
    [aux_sym_match_default_expression_token1] = ACTIONS(1039),
    [aux_sym_switch_statement_token1] = ACTIONS(1039),
    [aux_sym_switch_block_token1] = ACTIONS(1039),
    [anon_sym_PLUS] = ACTIONS(1039),
    [anon_sym_DASH] = ACTIONS(1039),
    [anon_sym_TILDE] = ACTIONS(1037),
    [anon_sym_BANG] = ACTIONS(1037),
    [anon_sym_AT] = ACTIONS(1037),
    [aux_sym_clone_expression_token1] = ACTIONS(1039),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1039),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1039),
    [anon_sym_DASH_DASH] = ACTIONS(1037),
    [anon_sym_PLUS_PLUS] = ACTIONS(1037),
    [aux_sym__list_destructing_token1] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1037),
    [anon_sym_self] = ACTIONS(1039),
    [anon_sym_parent] = ACTIONS(1039),
    [aux_sym__argument_name_token1] = ACTIONS(1039),
    [aux_sym__argument_name_token2] = ACTIONS(1039),
    [anon_sym_POUND_LBRACK] = ACTIONS(1037),
    [aux_sym_encapsed_string_token1] = ACTIONS(1037),
    [anon_sym_DQUOTE] = ACTIONS(1037),
    [aux_sym_string_token1] = ACTIONS(1037),
    [anon_sym_SQUOTE] = ACTIONS(1037),
    [anon_sym_LT_LT_LT] = ACTIONS(1037),
    [anon_sym_BQUOTE] = ACTIONS(1037),
    [anon_sym_DOLLAR] = ACTIONS(1037),
    [aux_sym_yield_expression_token1] = ACTIONS(1039),
    [aux_sym_include_expression_token1] = ACTIONS(1039),
    [aux_sym_include_once_expression_token1] = ACTIONS(1039),
    [aux_sym_require_expression_token1] = ACTIONS(1039),
    [aux_sym_require_once_expression_token1] = ACTIONS(1039),
    [sym_comment] = ACTIONS(3),
  },
  [445] = {
    [sym_else_if_clause] = STATE(494),
    [sym_else_clause] = STATE(541),
    [aux_sym_if_statement_repeat1] = STATE(452),
    [ts_builtin_sym_end] = ACTIONS(1047),
    [sym_name] = ACTIONS(1049),
    [anon_sym_QMARK_GT] = ACTIONS(1047),
    [anon_sym_SEMI] = ACTIONS(1047),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1049),
    [aux_sym_global_declaration_token1] = ACTIONS(1049),
    [aux_sym_namespace_definition_token1] = ACTIONS(1049),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1049),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1049),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1049),
    [anon_sym_BSLASH] = ACTIONS(1047),
    [anon_sym_LBRACE] = ACTIONS(1047),
    [anon_sym_RBRACE] = ACTIONS(1047),
    [aux_sym_trait_declaration_token1] = ACTIONS(1049),
    [aux_sym_interface_declaration_token1] = ACTIONS(1049),
    [aux_sym_enum_declaration_token1] = ACTIONS(1049),
    [aux_sym_enum_case_token1] = ACTIONS(1049),
    [aux_sym_class_declaration_token1] = ACTIONS(1049),
    [aux_sym_final_modifier_token1] = ACTIONS(1049),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1049),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1049),
    [sym_var_modifier] = ACTIONS(1049),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1049),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1049),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1049),
    [anon_sym_LPAREN2] = ACTIONS(1047),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1049),
    [aux_sym_cast_type_token1] = ACTIONS(1049),
    [aux_sym_echo_statement_token1] = ACTIONS(1049),
    [aux_sym_exit_statement_token1] = ACTIONS(1049),
    [anon_sym_unset] = ACTIONS(1049),
    [aux_sym_declare_statement_token1] = ACTIONS(1049),
    [aux_sym_declare_statement_token2] = ACTIONS(1049),
    [sym_float] = ACTIONS(1049),
    [aux_sym_try_statement_token1] = ACTIONS(1049),
    [aux_sym_goto_statement_token1] = ACTIONS(1049),
    [aux_sym_continue_statement_token1] = ACTIONS(1049),
    [aux_sym_break_statement_token1] = ACTIONS(1049),
    [sym_integer] = ACTIONS(1049),
    [aux_sym_return_statement_token1] = ACTIONS(1049),
    [aux_sym_throw_expression_token1] = ACTIONS(1049),
    [aux_sym_while_statement_token1] = ACTIONS(1049),
    [aux_sym_while_statement_token2] = ACTIONS(1049),
    [aux_sym_do_statement_token1] = ACTIONS(1049),
    [aux_sym_for_statement_token1] = ACTIONS(1049),
    [aux_sym_for_statement_token2] = ACTIONS(1049),
    [aux_sym_foreach_statement_token1] = ACTIONS(1049),
    [aux_sym_foreach_statement_token2] = ACTIONS(1049),
    [aux_sym_if_statement_token1] = ACTIONS(1049),
    [aux_sym_if_statement_token2] = ACTIONS(1049),
    [aux_sym_else_if_clause_token1] = ACTIONS(1051),
    [aux_sym_else_clause_token1] = ACTIONS(1054),
    [aux_sym_match_expression_token1] = ACTIONS(1049),
    [aux_sym_match_default_expression_token1] = ACTIONS(1049),
    [aux_sym_switch_statement_token1] = ACTIONS(1049),
    [aux_sym_switch_block_token1] = ACTIONS(1049),
    [anon_sym_PLUS] = ACTIONS(1049),
    [anon_sym_DASH] = ACTIONS(1049),
    [anon_sym_TILDE] = ACTIONS(1047),
    [anon_sym_BANG] = ACTIONS(1047),
    [anon_sym_AT] = ACTIONS(1047),
    [aux_sym_clone_expression_token1] = ACTIONS(1049),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1049),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1049),
    [anon_sym_DASH_DASH] = ACTIONS(1047),
    [anon_sym_PLUS_PLUS] = ACTIONS(1047),
    [aux_sym__list_destructing_token1] = ACTIONS(1049),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_self] = ACTIONS(1049),
    [anon_sym_parent] = ACTIONS(1049),
    [aux_sym__argument_name_token1] = ACTIONS(1049),
    [aux_sym__argument_name_token2] = ACTIONS(1049),
    [anon_sym_POUND_LBRACK] = ACTIONS(1047),
    [aux_sym_encapsed_string_token1] = ACTIONS(1047),
    [anon_sym_DQUOTE] = ACTIONS(1047),
    [aux_sym_string_token1] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1047),
    [anon_sym_LT_LT_LT] = ACTIONS(1047),
    [anon_sym_BQUOTE] = ACTIONS(1047),
    [anon_sym_DOLLAR] = ACTIONS(1047),
    [aux_sym_yield_expression_token1] = ACTIONS(1049),
    [aux_sym_include_expression_token1] = ACTIONS(1049),
    [aux_sym_include_once_expression_token1] = ACTIONS(1049),
    [aux_sym_require_expression_token1] = ACTIONS(1049),
    [aux_sym_require_once_expression_token1] = ACTIONS(1049),
    [sym_comment] = ACTIONS(3),
  },
  [446] = {
    [sym_else_if_clause] = STATE(494),
    [sym_else_clause] = STATE(495),
    [aux_sym_if_statement_repeat1] = STATE(445),
    [ts_builtin_sym_end] = ACTIONS(1057),
    [sym_name] = ACTIONS(1059),
    [anon_sym_QMARK_GT] = ACTIONS(1057),
    [anon_sym_SEMI] = ACTIONS(1057),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1059),
    [aux_sym_global_declaration_token1] = ACTIONS(1059),
    [aux_sym_namespace_definition_token1] = ACTIONS(1059),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1059),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1059),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1059),
    [anon_sym_BSLASH] = ACTIONS(1057),
    [anon_sym_LBRACE] = ACTIONS(1057),
    [anon_sym_RBRACE] = ACTIONS(1057),
    [aux_sym_trait_declaration_token1] = ACTIONS(1059),
    [aux_sym_interface_declaration_token1] = ACTIONS(1059),
    [aux_sym_enum_declaration_token1] = ACTIONS(1059),
    [aux_sym_enum_case_token1] = ACTIONS(1059),
    [aux_sym_class_declaration_token1] = ACTIONS(1059),
    [aux_sym_final_modifier_token1] = ACTIONS(1059),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1059),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1059),
    [sym_var_modifier] = ACTIONS(1059),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1059),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1059),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1059),
    [anon_sym_LPAREN2] = ACTIONS(1057),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1059),
    [aux_sym_cast_type_token1] = ACTIONS(1059),
    [aux_sym_echo_statement_token1] = ACTIONS(1059),
    [aux_sym_exit_statement_token1] = ACTIONS(1059),
    [anon_sym_unset] = ACTIONS(1059),
    [aux_sym_declare_statement_token1] = ACTIONS(1059),
    [aux_sym_declare_statement_token2] = ACTIONS(1059),
    [sym_float] = ACTIONS(1059),
    [aux_sym_try_statement_token1] = ACTIONS(1059),
    [aux_sym_goto_statement_token1] = ACTIONS(1059),
    [aux_sym_continue_statement_token1] = ACTIONS(1059),
    [aux_sym_break_statement_token1] = ACTIONS(1059),
    [sym_integer] = ACTIONS(1059),
    [aux_sym_return_statement_token1] = ACTIONS(1059),
    [aux_sym_throw_expression_token1] = ACTIONS(1059),
    [aux_sym_while_statement_token1] = ACTIONS(1059),
    [aux_sym_while_statement_token2] = ACTIONS(1059),
    [aux_sym_do_statement_token1] = ACTIONS(1059),
    [aux_sym_for_statement_token1] = ACTIONS(1059),
    [aux_sym_for_statement_token2] = ACTIONS(1059),
    [aux_sym_foreach_statement_token1] = ACTIONS(1059),
    [aux_sym_foreach_statement_token2] = ACTIONS(1059),
    [aux_sym_if_statement_token1] = ACTIONS(1059),
    [aux_sym_if_statement_token2] = ACTIONS(1059),
    [aux_sym_else_if_clause_token1] = ACTIONS(1061),
    [aux_sym_else_clause_token1] = ACTIONS(1064),
    [aux_sym_match_expression_token1] = ACTIONS(1059),
    [aux_sym_match_default_expression_token1] = ACTIONS(1059),
    [aux_sym_switch_statement_token1] = ACTIONS(1059),
    [aux_sym_switch_block_token1] = ACTIONS(1059),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(1059),
    [anon_sym_TILDE] = ACTIONS(1057),
    [anon_sym_BANG] = ACTIONS(1057),
    [anon_sym_AT] = ACTIONS(1057),
    [aux_sym_clone_expression_token1] = ACTIONS(1059),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1059),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1059),
    [anon_sym_DASH_DASH] = ACTIONS(1057),
    [anon_sym_PLUS_PLUS] = ACTIONS(1057),
    [aux_sym__list_destructing_token1] = ACTIONS(1059),
    [anon_sym_LBRACK] = ACTIONS(1057),
    [anon_sym_self] = ACTIONS(1059),
    [anon_sym_parent] = ACTIONS(1059),
    [aux_sym__argument_name_token1] = ACTIONS(1059),
    [aux_sym__argument_name_token2] = ACTIONS(1059),
    [anon_sym_POUND_LBRACK] = ACTIONS(1057),
    [aux_sym_encapsed_string_token1] = ACTIONS(1057),
    [anon_sym_DQUOTE] = ACTIONS(1057),
    [aux_sym_string_token1] = ACTIONS(1057),
    [anon_sym_SQUOTE] = ACTIONS(1057),
    [anon_sym_LT_LT_LT] = ACTIONS(1057),
    [anon_sym_BQUOTE] = ACTIONS(1057),
    [anon_sym_DOLLAR] = ACTIONS(1057),
    [aux_sym_yield_expression_token1] = ACTIONS(1059),
    [aux_sym_include_expression_token1] = ACTIONS(1059),
    [aux_sym_include_once_expression_token1] = ACTIONS(1059),
    [aux_sym_require_expression_token1] = ACTIONS(1059),
    [aux_sym_require_once_expression_token1] = ACTIONS(1059),
    [sym_comment] = ACTIONS(3),
  },
  [447] = {
    [sym_else_if_clause] = STATE(494),
    [sym_else_clause] = STATE(541),
    [aux_sym_if_statement_repeat1] = STATE(452),
    [ts_builtin_sym_end] = ACTIONS(1047),
    [sym_name] = ACTIONS(1049),
    [anon_sym_QMARK_GT] = ACTIONS(1047),
    [anon_sym_SEMI] = ACTIONS(1047),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1049),
    [aux_sym_global_declaration_token1] = ACTIONS(1049),
    [aux_sym_namespace_definition_token1] = ACTIONS(1049),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1049),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1049),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1049),
    [anon_sym_BSLASH] = ACTIONS(1047),
    [anon_sym_LBRACE] = ACTIONS(1047),
    [anon_sym_RBRACE] = ACTIONS(1047),
    [aux_sym_trait_declaration_token1] = ACTIONS(1049),
    [aux_sym_interface_declaration_token1] = ACTIONS(1049),
    [aux_sym_enum_declaration_token1] = ACTIONS(1049),
    [aux_sym_enum_case_token1] = ACTIONS(1049),
    [aux_sym_class_declaration_token1] = ACTIONS(1049),
    [aux_sym_final_modifier_token1] = ACTIONS(1049),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1049),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1049),
    [sym_var_modifier] = ACTIONS(1049),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1049),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1049),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1049),
    [anon_sym_LPAREN2] = ACTIONS(1047),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1049),
    [aux_sym_cast_type_token1] = ACTIONS(1049),
    [aux_sym_echo_statement_token1] = ACTIONS(1049),
    [aux_sym_exit_statement_token1] = ACTIONS(1049),
    [anon_sym_unset] = ACTIONS(1049),
    [aux_sym_declare_statement_token1] = ACTIONS(1049),
    [aux_sym_declare_statement_token2] = ACTIONS(1049),
    [sym_float] = ACTIONS(1049),
    [aux_sym_try_statement_token1] = ACTIONS(1049),
    [aux_sym_goto_statement_token1] = ACTIONS(1049),
    [aux_sym_continue_statement_token1] = ACTIONS(1049),
    [aux_sym_break_statement_token1] = ACTIONS(1049),
    [sym_integer] = ACTIONS(1049),
    [aux_sym_return_statement_token1] = ACTIONS(1049),
    [aux_sym_throw_expression_token1] = ACTIONS(1049),
    [aux_sym_while_statement_token1] = ACTIONS(1049),
    [aux_sym_while_statement_token2] = ACTIONS(1049),
    [aux_sym_do_statement_token1] = ACTIONS(1049),
    [aux_sym_for_statement_token1] = ACTIONS(1049),
    [aux_sym_for_statement_token2] = ACTIONS(1049),
    [aux_sym_foreach_statement_token1] = ACTIONS(1049),
    [aux_sym_foreach_statement_token2] = ACTIONS(1049),
    [aux_sym_if_statement_token1] = ACTIONS(1049),
    [aux_sym_if_statement_token2] = ACTIONS(1049),
    [aux_sym_else_if_clause_token1] = ACTIONS(1067),
    [aux_sym_else_clause_token1] = ACTIONS(1069),
    [aux_sym_match_expression_token1] = ACTIONS(1049),
    [aux_sym_match_default_expression_token1] = ACTIONS(1049),
    [aux_sym_switch_statement_token1] = ACTIONS(1049),
    [aux_sym_switch_block_token1] = ACTIONS(1049),
    [anon_sym_PLUS] = ACTIONS(1049),
    [anon_sym_DASH] = ACTIONS(1049),
    [anon_sym_TILDE] = ACTIONS(1047),
    [anon_sym_BANG] = ACTIONS(1047),
    [anon_sym_AT] = ACTIONS(1047),
    [aux_sym_clone_expression_token1] = ACTIONS(1049),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1049),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1049),
    [anon_sym_DASH_DASH] = ACTIONS(1047),
    [anon_sym_PLUS_PLUS] = ACTIONS(1047),
    [aux_sym__list_destructing_token1] = ACTIONS(1049),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_self] = ACTIONS(1049),
    [anon_sym_parent] = ACTIONS(1049),
    [aux_sym__argument_name_token1] = ACTIONS(1049),
    [aux_sym__argument_name_token2] = ACTIONS(1049),
    [anon_sym_POUND_LBRACK] = ACTIONS(1047),
    [aux_sym_encapsed_string_token1] = ACTIONS(1047),
    [anon_sym_DQUOTE] = ACTIONS(1047),
    [aux_sym_string_token1] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1047),
    [anon_sym_LT_LT_LT] = ACTIONS(1047),
    [anon_sym_BQUOTE] = ACTIONS(1047),
    [anon_sym_DOLLAR] = ACTIONS(1047),
    [aux_sym_yield_expression_token1] = ACTIONS(1049),
    [aux_sym_include_expression_token1] = ACTIONS(1049),
    [aux_sym_include_once_expression_token1] = ACTIONS(1049),
    [aux_sym_require_expression_token1] = ACTIONS(1049),
    [aux_sym_require_once_expression_token1] = ACTIONS(1049),
    [sym_comment] = ACTIONS(3),
  },
  [448] = {
    [sym_else_if_clause] = STATE(494),
    [sym_else_clause] = STATE(495),
    [aux_sym_if_statement_repeat1] = STATE(447),
    [ts_builtin_sym_end] = ACTIONS(1057),
    [sym_name] = ACTIONS(1059),
    [anon_sym_QMARK_GT] = ACTIONS(1057),
    [anon_sym_SEMI] = ACTIONS(1057),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1059),
    [aux_sym_global_declaration_token1] = ACTIONS(1059),
    [aux_sym_namespace_definition_token1] = ACTIONS(1059),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1059),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1059),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1059),
    [anon_sym_BSLASH] = ACTIONS(1057),
    [anon_sym_LBRACE] = ACTIONS(1057),
    [anon_sym_RBRACE] = ACTIONS(1057),
    [aux_sym_trait_declaration_token1] = ACTIONS(1059),
    [aux_sym_interface_declaration_token1] = ACTIONS(1059),
    [aux_sym_enum_declaration_token1] = ACTIONS(1059),
    [aux_sym_enum_case_token1] = ACTIONS(1059),
    [aux_sym_class_declaration_token1] = ACTIONS(1059),
    [aux_sym_final_modifier_token1] = ACTIONS(1059),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1059),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1059),
    [sym_var_modifier] = ACTIONS(1059),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1059),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1059),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1059),
    [anon_sym_LPAREN2] = ACTIONS(1057),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1059),
    [aux_sym_cast_type_token1] = ACTIONS(1059),
    [aux_sym_echo_statement_token1] = ACTIONS(1059),
    [aux_sym_exit_statement_token1] = ACTIONS(1059),
    [anon_sym_unset] = ACTIONS(1059),
    [aux_sym_declare_statement_token1] = ACTIONS(1059),
    [aux_sym_declare_statement_token2] = ACTIONS(1059),
    [sym_float] = ACTIONS(1059),
    [aux_sym_try_statement_token1] = ACTIONS(1059),
    [aux_sym_goto_statement_token1] = ACTIONS(1059),
    [aux_sym_continue_statement_token1] = ACTIONS(1059),
    [aux_sym_break_statement_token1] = ACTIONS(1059),
    [sym_integer] = ACTIONS(1059),
    [aux_sym_return_statement_token1] = ACTIONS(1059),
    [aux_sym_throw_expression_token1] = ACTIONS(1059),
    [aux_sym_while_statement_token1] = ACTIONS(1059),
    [aux_sym_while_statement_token2] = ACTIONS(1059),
    [aux_sym_do_statement_token1] = ACTIONS(1059),
    [aux_sym_for_statement_token1] = ACTIONS(1059),
    [aux_sym_for_statement_token2] = ACTIONS(1059),
    [aux_sym_foreach_statement_token1] = ACTIONS(1059),
    [aux_sym_foreach_statement_token2] = ACTIONS(1059),
    [aux_sym_if_statement_token1] = ACTIONS(1059),
    [aux_sym_if_statement_token2] = ACTIONS(1059),
    [aux_sym_else_if_clause_token1] = ACTIONS(1067),
    [aux_sym_else_clause_token1] = ACTIONS(1069),
    [aux_sym_match_expression_token1] = ACTIONS(1059),
    [aux_sym_match_default_expression_token1] = ACTIONS(1059),
    [aux_sym_switch_statement_token1] = ACTIONS(1059),
    [aux_sym_switch_block_token1] = ACTIONS(1059),
    [anon_sym_PLUS] = ACTIONS(1059),
    [anon_sym_DASH] = ACTIONS(1059),
    [anon_sym_TILDE] = ACTIONS(1057),
    [anon_sym_BANG] = ACTIONS(1057),
    [anon_sym_AT] = ACTIONS(1057),
    [aux_sym_clone_expression_token1] = ACTIONS(1059),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1059),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1059),
    [anon_sym_DASH_DASH] = ACTIONS(1057),
    [anon_sym_PLUS_PLUS] = ACTIONS(1057),
    [aux_sym__list_destructing_token1] = ACTIONS(1059),
    [anon_sym_LBRACK] = ACTIONS(1057),
    [anon_sym_self] = ACTIONS(1059),
    [anon_sym_parent] = ACTIONS(1059),
    [aux_sym__argument_name_token1] = ACTIONS(1059),
    [aux_sym__argument_name_token2] = ACTIONS(1059),
    [anon_sym_POUND_LBRACK] = ACTIONS(1057),
    [aux_sym_encapsed_string_token1] = ACTIONS(1057),
    [anon_sym_DQUOTE] = ACTIONS(1057),
    [aux_sym_string_token1] = ACTIONS(1057),
    [anon_sym_SQUOTE] = ACTIONS(1057),
    [anon_sym_LT_LT_LT] = ACTIONS(1057),
    [anon_sym_BQUOTE] = ACTIONS(1057),
    [anon_sym_DOLLAR] = ACTIONS(1057),
    [aux_sym_yield_expression_token1] = ACTIONS(1059),
    [aux_sym_include_expression_token1] = ACTIONS(1059),
    [aux_sym_include_once_expression_token1] = ACTIONS(1059),
    [aux_sym_require_expression_token1] = ACTIONS(1059),
    [aux_sym_require_once_expression_token1] = ACTIONS(1059),
    [sym_comment] = ACTIONS(3),
  },
  [449] = {
    [ts_builtin_sym_end] = ACTIONS(1071),
    [sym_name] = ACTIONS(1073),
    [anon_sym_QMARK_GT] = ACTIONS(1071),
    [anon_sym_SEMI] = ACTIONS(1071),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1073),
    [aux_sym_global_declaration_token1] = ACTIONS(1073),
    [aux_sym_namespace_definition_token1] = ACTIONS(1073),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1073),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1073),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1073),
    [anon_sym_BSLASH] = ACTIONS(1071),
    [anon_sym_LBRACE] = ACTIONS(1071),
    [anon_sym_RBRACE] = ACTIONS(1071),
    [aux_sym_trait_declaration_token1] = ACTIONS(1073),
    [aux_sym_interface_declaration_token1] = ACTIONS(1073),
    [aux_sym_enum_declaration_token1] = ACTIONS(1073),
    [aux_sym_enum_case_token1] = ACTIONS(1073),
    [aux_sym_class_declaration_token1] = ACTIONS(1073),
    [aux_sym_final_modifier_token1] = ACTIONS(1073),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1073),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1073),
    [sym_var_modifier] = ACTIONS(1073),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1073),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1073),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1073),
    [anon_sym_LPAREN2] = ACTIONS(1071),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1073),
    [aux_sym_cast_type_token1] = ACTIONS(1073),
    [aux_sym_echo_statement_token1] = ACTIONS(1073),
    [aux_sym_exit_statement_token1] = ACTIONS(1073),
    [anon_sym_unset] = ACTIONS(1073),
    [aux_sym_declare_statement_token1] = ACTIONS(1073),
    [aux_sym_declare_statement_token2] = ACTIONS(1073),
    [sym_float] = ACTIONS(1073),
    [aux_sym_try_statement_token1] = ACTIONS(1073),
    [aux_sym_catch_clause_token1] = ACTIONS(1073),
    [aux_sym_finally_clause_token1] = ACTIONS(1073),
    [aux_sym_goto_statement_token1] = ACTIONS(1073),
    [aux_sym_continue_statement_token1] = ACTIONS(1073),
    [aux_sym_break_statement_token1] = ACTIONS(1073),
    [sym_integer] = ACTIONS(1073),
    [aux_sym_return_statement_token1] = ACTIONS(1073),
    [aux_sym_throw_expression_token1] = ACTIONS(1073),
    [aux_sym_while_statement_token1] = ACTIONS(1073),
    [aux_sym_while_statement_token2] = ACTIONS(1073),
    [aux_sym_do_statement_token1] = ACTIONS(1073),
    [aux_sym_for_statement_token1] = ACTIONS(1073),
    [aux_sym_for_statement_token2] = ACTIONS(1073),
    [aux_sym_foreach_statement_token1] = ACTIONS(1073),
    [aux_sym_foreach_statement_token2] = ACTIONS(1073),
    [aux_sym_if_statement_token1] = ACTIONS(1073),
    [aux_sym_if_statement_token2] = ACTIONS(1073),
    [aux_sym_else_if_clause_token1] = ACTIONS(1073),
    [aux_sym_else_clause_token1] = ACTIONS(1073),
    [aux_sym_match_expression_token1] = ACTIONS(1073),
    [aux_sym_match_default_expression_token1] = ACTIONS(1073),
    [aux_sym_switch_statement_token1] = ACTIONS(1073),
    [aux_sym_switch_block_token1] = ACTIONS(1073),
    [anon_sym_PLUS] = ACTIONS(1073),
    [anon_sym_DASH] = ACTIONS(1073),
    [anon_sym_TILDE] = ACTIONS(1071),
    [anon_sym_BANG] = ACTIONS(1071),
    [anon_sym_AT] = ACTIONS(1071),
    [aux_sym_clone_expression_token1] = ACTIONS(1073),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1073),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1073),
    [anon_sym_DASH_DASH] = ACTIONS(1071),
    [anon_sym_PLUS_PLUS] = ACTIONS(1071),
    [aux_sym__list_destructing_token1] = ACTIONS(1073),
    [anon_sym_LBRACK] = ACTIONS(1071),
    [anon_sym_self] = ACTIONS(1073),
    [anon_sym_parent] = ACTIONS(1073),
    [aux_sym__argument_name_token1] = ACTIONS(1073),
    [aux_sym__argument_name_token2] = ACTIONS(1073),
    [anon_sym_POUND_LBRACK] = ACTIONS(1071),
    [aux_sym_encapsed_string_token1] = ACTIONS(1071),
    [anon_sym_DQUOTE] = ACTIONS(1071),
    [aux_sym_string_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1071),
    [anon_sym_LT_LT_LT] = ACTIONS(1071),
    [anon_sym_BQUOTE] = ACTIONS(1071),
    [anon_sym_DOLLAR] = ACTIONS(1071),
    [aux_sym_yield_expression_token1] = ACTIONS(1073),
    [aux_sym_include_expression_token1] = ACTIONS(1073),
    [aux_sym_include_once_expression_token1] = ACTIONS(1073),
    [aux_sym_require_expression_token1] = ACTIONS(1073),
    [aux_sym_require_once_expression_token1] = ACTIONS(1073),
    [sym_comment] = ACTIONS(3),
  },
  [450] = {
    [ts_builtin_sym_end] = ACTIONS(1075),
    [sym_name] = ACTIONS(1077),
    [anon_sym_QMARK_GT] = ACTIONS(1075),
    [anon_sym_SEMI] = ACTIONS(1075),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1077),
    [aux_sym_global_declaration_token1] = ACTIONS(1077),
    [aux_sym_namespace_definition_token1] = ACTIONS(1077),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1077),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1077),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1077),
    [anon_sym_BSLASH] = ACTIONS(1075),
    [anon_sym_LBRACE] = ACTIONS(1075),
    [anon_sym_RBRACE] = ACTIONS(1075),
    [aux_sym_trait_declaration_token1] = ACTIONS(1077),
    [aux_sym_interface_declaration_token1] = ACTIONS(1077),
    [aux_sym_enum_declaration_token1] = ACTIONS(1077),
    [aux_sym_enum_case_token1] = ACTIONS(1077),
    [aux_sym_class_declaration_token1] = ACTIONS(1077),
    [aux_sym_final_modifier_token1] = ACTIONS(1077),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1077),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1077),
    [sym_var_modifier] = ACTIONS(1077),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1077),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1077),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1077),
    [anon_sym_LPAREN2] = ACTIONS(1075),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1077),
    [aux_sym_cast_type_token1] = ACTIONS(1077),
    [aux_sym_echo_statement_token1] = ACTIONS(1077),
    [aux_sym_exit_statement_token1] = ACTIONS(1077),
    [anon_sym_unset] = ACTIONS(1077),
    [aux_sym_declare_statement_token1] = ACTIONS(1077),
    [aux_sym_declare_statement_token2] = ACTIONS(1077),
    [sym_float] = ACTIONS(1077),
    [aux_sym_try_statement_token1] = ACTIONS(1077),
    [aux_sym_catch_clause_token1] = ACTIONS(1077),
    [aux_sym_finally_clause_token1] = ACTIONS(1077),
    [aux_sym_goto_statement_token1] = ACTIONS(1077),
    [aux_sym_continue_statement_token1] = ACTIONS(1077),
    [aux_sym_break_statement_token1] = ACTIONS(1077),
    [sym_integer] = ACTIONS(1077),
    [aux_sym_return_statement_token1] = ACTIONS(1077),
    [aux_sym_throw_expression_token1] = ACTIONS(1077),
    [aux_sym_while_statement_token1] = ACTIONS(1077),
    [aux_sym_while_statement_token2] = ACTIONS(1077),
    [aux_sym_do_statement_token1] = ACTIONS(1077),
    [aux_sym_for_statement_token1] = ACTIONS(1077),
    [aux_sym_for_statement_token2] = ACTIONS(1077),
    [aux_sym_foreach_statement_token1] = ACTIONS(1077),
    [aux_sym_foreach_statement_token2] = ACTIONS(1077),
    [aux_sym_if_statement_token1] = ACTIONS(1077),
    [aux_sym_if_statement_token2] = ACTIONS(1077),
    [aux_sym_else_if_clause_token1] = ACTIONS(1077),
    [aux_sym_else_clause_token1] = ACTIONS(1077),
    [aux_sym_match_expression_token1] = ACTIONS(1077),
    [aux_sym_match_default_expression_token1] = ACTIONS(1077),
    [aux_sym_switch_statement_token1] = ACTIONS(1077),
    [aux_sym_switch_block_token1] = ACTIONS(1077),
    [anon_sym_PLUS] = ACTIONS(1077),
    [anon_sym_DASH] = ACTIONS(1077),
    [anon_sym_TILDE] = ACTIONS(1075),
    [anon_sym_BANG] = ACTIONS(1075),
    [anon_sym_AT] = ACTIONS(1075),
    [aux_sym_clone_expression_token1] = ACTIONS(1077),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1077),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1077),
    [anon_sym_DASH_DASH] = ACTIONS(1075),
    [anon_sym_PLUS_PLUS] = ACTIONS(1075),
    [aux_sym__list_destructing_token1] = ACTIONS(1077),
    [anon_sym_LBRACK] = ACTIONS(1075),
    [anon_sym_self] = ACTIONS(1077),
    [anon_sym_parent] = ACTIONS(1077),
    [aux_sym__argument_name_token1] = ACTIONS(1077),
    [aux_sym__argument_name_token2] = ACTIONS(1077),
    [anon_sym_POUND_LBRACK] = ACTIONS(1075),
    [aux_sym_encapsed_string_token1] = ACTIONS(1075),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [aux_sym_string_token1] = ACTIONS(1075),
    [anon_sym_SQUOTE] = ACTIONS(1075),
    [anon_sym_LT_LT_LT] = ACTIONS(1075),
    [anon_sym_BQUOTE] = ACTIONS(1075),
    [anon_sym_DOLLAR] = ACTIONS(1075),
    [aux_sym_yield_expression_token1] = ACTIONS(1077),
    [aux_sym_include_expression_token1] = ACTIONS(1077),
    [aux_sym_include_once_expression_token1] = ACTIONS(1077),
    [aux_sym_require_expression_token1] = ACTIONS(1077),
    [aux_sym_require_once_expression_token1] = ACTIONS(1077),
    [sym_comment] = ACTIONS(3),
  },
  [451] = {
    [ts_builtin_sym_end] = ACTIONS(1079),
    [sym_name] = ACTIONS(1081),
    [anon_sym_QMARK_GT] = ACTIONS(1079),
    [anon_sym_SEMI] = ACTIONS(1079),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1081),
    [aux_sym_global_declaration_token1] = ACTIONS(1081),
    [aux_sym_namespace_definition_token1] = ACTIONS(1081),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1081),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1081),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1081),
    [anon_sym_BSLASH] = ACTIONS(1079),
    [anon_sym_LBRACE] = ACTIONS(1079),
    [anon_sym_RBRACE] = ACTIONS(1079),
    [aux_sym_trait_declaration_token1] = ACTIONS(1081),
    [aux_sym_interface_declaration_token1] = ACTIONS(1081),
    [aux_sym_enum_declaration_token1] = ACTIONS(1081),
    [aux_sym_enum_case_token1] = ACTIONS(1081),
    [aux_sym_class_declaration_token1] = ACTIONS(1081),
    [aux_sym_final_modifier_token1] = ACTIONS(1081),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1081),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1081),
    [sym_var_modifier] = ACTIONS(1081),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1081),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1081),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1081),
    [anon_sym_LPAREN2] = ACTIONS(1079),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1081),
    [aux_sym_cast_type_token1] = ACTIONS(1081),
    [aux_sym_echo_statement_token1] = ACTIONS(1081),
    [aux_sym_exit_statement_token1] = ACTIONS(1081),
    [anon_sym_unset] = ACTIONS(1081),
    [aux_sym_declare_statement_token1] = ACTIONS(1081),
    [aux_sym_declare_statement_token2] = ACTIONS(1081),
    [sym_float] = ACTIONS(1081),
    [aux_sym_try_statement_token1] = ACTIONS(1081),
    [aux_sym_catch_clause_token1] = ACTIONS(1081),
    [aux_sym_finally_clause_token1] = ACTIONS(1081),
    [aux_sym_goto_statement_token1] = ACTIONS(1081),
    [aux_sym_continue_statement_token1] = ACTIONS(1081),
    [aux_sym_break_statement_token1] = ACTIONS(1081),
    [sym_integer] = ACTIONS(1081),
    [aux_sym_return_statement_token1] = ACTIONS(1081),
    [aux_sym_throw_expression_token1] = ACTIONS(1081),
    [aux_sym_while_statement_token1] = ACTIONS(1081),
    [aux_sym_while_statement_token2] = ACTIONS(1081),
    [aux_sym_do_statement_token1] = ACTIONS(1081),
    [aux_sym_for_statement_token1] = ACTIONS(1081),
    [aux_sym_for_statement_token2] = ACTIONS(1081),
    [aux_sym_foreach_statement_token1] = ACTIONS(1081),
    [aux_sym_foreach_statement_token2] = ACTIONS(1081),
    [aux_sym_if_statement_token1] = ACTIONS(1081),
    [aux_sym_if_statement_token2] = ACTIONS(1081),
    [aux_sym_else_if_clause_token1] = ACTIONS(1081),
    [aux_sym_else_clause_token1] = ACTIONS(1081),
    [aux_sym_match_expression_token1] = ACTIONS(1081),
    [aux_sym_match_default_expression_token1] = ACTIONS(1081),
    [aux_sym_switch_statement_token1] = ACTIONS(1081),
    [aux_sym_switch_block_token1] = ACTIONS(1081),
    [anon_sym_PLUS] = ACTIONS(1081),
    [anon_sym_DASH] = ACTIONS(1081),
    [anon_sym_TILDE] = ACTIONS(1079),
    [anon_sym_BANG] = ACTIONS(1079),
    [anon_sym_AT] = ACTIONS(1079),
    [aux_sym_clone_expression_token1] = ACTIONS(1081),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1081),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1081),
    [anon_sym_DASH_DASH] = ACTIONS(1079),
    [anon_sym_PLUS_PLUS] = ACTIONS(1079),
    [aux_sym__list_destructing_token1] = ACTIONS(1081),
    [anon_sym_LBRACK] = ACTIONS(1079),
    [anon_sym_self] = ACTIONS(1081),
    [anon_sym_parent] = ACTIONS(1081),
    [aux_sym__argument_name_token1] = ACTIONS(1081),
    [aux_sym__argument_name_token2] = ACTIONS(1081),
    [anon_sym_POUND_LBRACK] = ACTIONS(1079),
    [aux_sym_encapsed_string_token1] = ACTIONS(1079),
    [anon_sym_DQUOTE] = ACTIONS(1079),
    [aux_sym_string_token1] = ACTIONS(1079),
    [anon_sym_SQUOTE] = ACTIONS(1079),
    [anon_sym_LT_LT_LT] = ACTIONS(1079),
    [anon_sym_BQUOTE] = ACTIONS(1079),
    [anon_sym_DOLLAR] = ACTIONS(1079),
    [aux_sym_yield_expression_token1] = ACTIONS(1081),
    [aux_sym_include_expression_token1] = ACTIONS(1081),
    [aux_sym_include_once_expression_token1] = ACTIONS(1081),
    [aux_sym_require_expression_token1] = ACTIONS(1081),
    [aux_sym_require_once_expression_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(3),
  },
  [452] = {
    [sym_else_if_clause] = STATE(494),
    [aux_sym_if_statement_repeat1] = STATE(452),
    [ts_builtin_sym_end] = ACTIONS(1083),
    [sym_name] = ACTIONS(1085),
    [anon_sym_QMARK_GT] = ACTIONS(1083),
    [anon_sym_SEMI] = ACTIONS(1083),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1085),
    [aux_sym_global_declaration_token1] = ACTIONS(1085),
    [aux_sym_namespace_definition_token1] = ACTIONS(1085),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1085),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1085),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1085),
    [anon_sym_BSLASH] = ACTIONS(1083),
    [anon_sym_LBRACE] = ACTIONS(1083),
    [anon_sym_RBRACE] = ACTIONS(1083),
    [aux_sym_trait_declaration_token1] = ACTIONS(1085),
    [aux_sym_interface_declaration_token1] = ACTIONS(1085),
    [aux_sym_enum_declaration_token1] = ACTIONS(1085),
    [aux_sym_enum_case_token1] = ACTIONS(1085),
    [aux_sym_class_declaration_token1] = ACTIONS(1085),
    [aux_sym_final_modifier_token1] = ACTIONS(1085),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1085),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1085),
    [sym_var_modifier] = ACTIONS(1085),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1085),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1085),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1085),
    [anon_sym_LPAREN2] = ACTIONS(1083),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1085),
    [aux_sym_cast_type_token1] = ACTIONS(1085),
    [aux_sym_echo_statement_token1] = ACTIONS(1085),
    [aux_sym_exit_statement_token1] = ACTIONS(1085),
    [anon_sym_unset] = ACTIONS(1085),
    [aux_sym_declare_statement_token1] = ACTIONS(1085),
    [aux_sym_declare_statement_token2] = ACTIONS(1085),
    [sym_float] = ACTIONS(1085),
    [aux_sym_try_statement_token1] = ACTIONS(1085),
    [aux_sym_goto_statement_token1] = ACTIONS(1085),
    [aux_sym_continue_statement_token1] = ACTIONS(1085),
    [aux_sym_break_statement_token1] = ACTIONS(1085),
    [sym_integer] = ACTIONS(1085),
    [aux_sym_return_statement_token1] = ACTIONS(1085),
    [aux_sym_throw_expression_token1] = ACTIONS(1085),
    [aux_sym_while_statement_token1] = ACTIONS(1085),
    [aux_sym_while_statement_token2] = ACTIONS(1085),
    [aux_sym_do_statement_token1] = ACTIONS(1085),
    [aux_sym_for_statement_token1] = ACTIONS(1085),
    [aux_sym_for_statement_token2] = ACTIONS(1085),
    [aux_sym_foreach_statement_token1] = ACTIONS(1085),
    [aux_sym_foreach_statement_token2] = ACTIONS(1085),
    [aux_sym_if_statement_token1] = ACTIONS(1085),
    [aux_sym_if_statement_token2] = ACTIONS(1085),
    [aux_sym_else_if_clause_token1] = ACTIONS(1087),
    [aux_sym_else_clause_token1] = ACTIONS(1085),
    [aux_sym_match_expression_token1] = ACTIONS(1085),
    [aux_sym_match_default_expression_token1] = ACTIONS(1085),
    [aux_sym_switch_statement_token1] = ACTIONS(1085),
    [aux_sym_switch_block_token1] = ACTIONS(1085),
    [anon_sym_PLUS] = ACTIONS(1085),
    [anon_sym_DASH] = ACTIONS(1085),
    [anon_sym_TILDE] = ACTIONS(1083),
    [anon_sym_BANG] = ACTIONS(1083),
    [anon_sym_AT] = ACTIONS(1083),
    [aux_sym_clone_expression_token1] = ACTIONS(1085),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1085),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1085),
    [anon_sym_DASH_DASH] = ACTIONS(1083),
    [anon_sym_PLUS_PLUS] = ACTIONS(1083),
    [aux_sym__list_destructing_token1] = ACTIONS(1085),
    [anon_sym_LBRACK] = ACTIONS(1083),
    [anon_sym_self] = ACTIONS(1085),
    [anon_sym_parent] = ACTIONS(1085),
    [aux_sym__argument_name_token1] = ACTIONS(1085),
    [aux_sym__argument_name_token2] = ACTIONS(1085),
    [anon_sym_POUND_LBRACK] = ACTIONS(1083),
    [aux_sym_encapsed_string_token1] = ACTIONS(1083),
    [anon_sym_DQUOTE] = ACTIONS(1083),
    [aux_sym_string_token1] = ACTIONS(1083),
    [anon_sym_SQUOTE] = ACTIONS(1083),
    [anon_sym_LT_LT_LT] = ACTIONS(1083),
    [anon_sym_BQUOTE] = ACTIONS(1083),
    [anon_sym_DOLLAR] = ACTIONS(1083),
    [aux_sym_yield_expression_token1] = ACTIONS(1085),
    [aux_sym_include_expression_token1] = ACTIONS(1085),
    [aux_sym_include_once_expression_token1] = ACTIONS(1085),
    [aux_sym_require_expression_token1] = ACTIONS(1085),
    [aux_sym_require_once_expression_token1] = ACTIONS(1085),
    [sym_comment] = ACTIONS(3),
  },
  [453] = {
    [ts_builtin_sym_end] = ACTIONS(1090),
    [sym_name] = ACTIONS(1092),
    [anon_sym_QMARK_GT] = ACTIONS(1090),
    [anon_sym_SEMI] = ACTIONS(1090),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1092),
    [aux_sym_global_declaration_token1] = ACTIONS(1092),
    [aux_sym_namespace_definition_token1] = ACTIONS(1092),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1092),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1092),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1092),
    [anon_sym_BSLASH] = ACTIONS(1090),
    [anon_sym_LBRACE] = ACTIONS(1090),
    [anon_sym_RBRACE] = ACTIONS(1090),
    [aux_sym_trait_declaration_token1] = ACTIONS(1092),
    [aux_sym_interface_declaration_token1] = ACTIONS(1092),
    [aux_sym_enum_declaration_token1] = ACTIONS(1092),
    [aux_sym_enum_case_token1] = ACTIONS(1092),
    [aux_sym_class_declaration_token1] = ACTIONS(1092),
    [aux_sym_final_modifier_token1] = ACTIONS(1092),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1092),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1092),
    [sym_var_modifier] = ACTIONS(1092),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1092),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1092),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1092),
    [anon_sym_LPAREN2] = ACTIONS(1090),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1092),
    [aux_sym_cast_type_token1] = ACTIONS(1092),
    [aux_sym_echo_statement_token1] = ACTIONS(1092),
    [aux_sym_exit_statement_token1] = ACTIONS(1092),
    [anon_sym_unset] = ACTIONS(1092),
    [aux_sym_declare_statement_token1] = ACTIONS(1092),
    [aux_sym_declare_statement_token2] = ACTIONS(1092),
    [sym_float] = ACTIONS(1092),
    [aux_sym_try_statement_token1] = ACTIONS(1092),
    [aux_sym_catch_clause_token1] = ACTIONS(1092),
    [aux_sym_finally_clause_token1] = ACTIONS(1092),
    [aux_sym_goto_statement_token1] = ACTIONS(1092),
    [aux_sym_continue_statement_token1] = ACTIONS(1092),
    [aux_sym_break_statement_token1] = ACTIONS(1092),
    [sym_integer] = ACTIONS(1092),
    [aux_sym_return_statement_token1] = ACTIONS(1092),
    [aux_sym_throw_expression_token1] = ACTIONS(1092),
    [aux_sym_while_statement_token1] = ACTIONS(1092),
    [aux_sym_while_statement_token2] = ACTIONS(1092),
    [aux_sym_do_statement_token1] = ACTIONS(1092),
    [aux_sym_for_statement_token1] = ACTIONS(1092),
    [aux_sym_for_statement_token2] = ACTIONS(1092),
    [aux_sym_foreach_statement_token1] = ACTIONS(1092),
    [aux_sym_foreach_statement_token2] = ACTIONS(1092),
    [aux_sym_if_statement_token1] = ACTIONS(1092),
    [aux_sym_if_statement_token2] = ACTIONS(1092),
    [aux_sym_else_if_clause_token1] = ACTIONS(1092),
    [aux_sym_else_clause_token1] = ACTIONS(1092),
    [aux_sym_match_expression_token1] = ACTIONS(1092),
    [aux_sym_match_default_expression_token1] = ACTIONS(1092),
    [aux_sym_switch_statement_token1] = ACTIONS(1092),
    [aux_sym_switch_block_token1] = ACTIONS(1092),
    [anon_sym_PLUS] = ACTIONS(1092),
    [anon_sym_DASH] = ACTIONS(1092),
    [anon_sym_TILDE] = ACTIONS(1090),
    [anon_sym_BANG] = ACTIONS(1090),
    [anon_sym_AT] = ACTIONS(1090),
    [aux_sym_clone_expression_token1] = ACTIONS(1092),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1092),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1092),
    [anon_sym_DASH_DASH] = ACTIONS(1090),
    [anon_sym_PLUS_PLUS] = ACTIONS(1090),
    [aux_sym__list_destructing_token1] = ACTIONS(1092),
    [anon_sym_LBRACK] = ACTIONS(1090),
    [anon_sym_self] = ACTIONS(1092),
    [anon_sym_parent] = ACTIONS(1092),
    [aux_sym__argument_name_token1] = ACTIONS(1092),
    [aux_sym__argument_name_token2] = ACTIONS(1092),
    [anon_sym_POUND_LBRACK] = ACTIONS(1090),
    [aux_sym_encapsed_string_token1] = ACTIONS(1090),
    [anon_sym_DQUOTE] = ACTIONS(1090),
    [aux_sym_string_token1] = ACTIONS(1090),
    [anon_sym_SQUOTE] = ACTIONS(1090),
    [anon_sym_LT_LT_LT] = ACTIONS(1090),
    [anon_sym_BQUOTE] = ACTIONS(1090),
    [anon_sym_DOLLAR] = ACTIONS(1090),
    [aux_sym_yield_expression_token1] = ACTIONS(1092),
    [aux_sym_include_expression_token1] = ACTIONS(1092),
    [aux_sym_include_once_expression_token1] = ACTIONS(1092),
    [aux_sym_require_expression_token1] = ACTIONS(1092),
    [aux_sym_require_once_expression_token1] = ACTIONS(1092),
    [sym_comment] = ACTIONS(3),
  },
  [454] = {
    [ts_builtin_sym_end] = ACTIONS(1094),
    [sym_name] = ACTIONS(1096),
    [anon_sym_QMARK_GT] = ACTIONS(1094),
    [anon_sym_SEMI] = ACTIONS(1094),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1096),
    [aux_sym_global_declaration_token1] = ACTIONS(1096),
    [aux_sym_namespace_definition_token1] = ACTIONS(1096),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1096),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1096),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1096),
    [anon_sym_BSLASH] = ACTIONS(1094),
    [anon_sym_LBRACE] = ACTIONS(1094),
    [anon_sym_RBRACE] = ACTIONS(1094),
    [aux_sym_trait_declaration_token1] = ACTIONS(1096),
    [aux_sym_interface_declaration_token1] = ACTIONS(1096),
    [aux_sym_enum_declaration_token1] = ACTIONS(1096),
    [aux_sym_enum_case_token1] = ACTIONS(1096),
    [aux_sym_class_declaration_token1] = ACTIONS(1096),
    [aux_sym_final_modifier_token1] = ACTIONS(1096),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1096),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1096),
    [sym_var_modifier] = ACTIONS(1096),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1096),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1096),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1096),
    [anon_sym_LPAREN2] = ACTIONS(1094),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1096),
    [aux_sym_cast_type_token1] = ACTIONS(1096),
    [aux_sym_echo_statement_token1] = ACTIONS(1096),
    [aux_sym_exit_statement_token1] = ACTIONS(1096),
    [anon_sym_unset] = ACTIONS(1096),
    [aux_sym_declare_statement_token1] = ACTIONS(1096),
    [aux_sym_declare_statement_token2] = ACTIONS(1096),
    [sym_float] = ACTIONS(1096),
    [aux_sym_try_statement_token1] = ACTIONS(1096),
    [aux_sym_catch_clause_token1] = ACTIONS(1096),
    [aux_sym_finally_clause_token1] = ACTIONS(1096),
    [aux_sym_goto_statement_token1] = ACTIONS(1096),
    [aux_sym_continue_statement_token1] = ACTIONS(1096),
    [aux_sym_break_statement_token1] = ACTIONS(1096),
    [sym_integer] = ACTIONS(1096),
    [aux_sym_return_statement_token1] = ACTIONS(1096),
    [aux_sym_throw_expression_token1] = ACTIONS(1096),
    [aux_sym_while_statement_token1] = ACTIONS(1096),
    [aux_sym_while_statement_token2] = ACTIONS(1096),
    [aux_sym_do_statement_token1] = ACTIONS(1096),
    [aux_sym_for_statement_token1] = ACTIONS(1096),
    [aux_sym_for_statement_token2] = ACTIONS(1096),
    [aux_sym_foreach_statement_token1] = ACTIONS(1096),
    [aux_sym_foreach_statement_token2] = ACTIONS(1096),
    [aux_sym_if_statement_token1] = ACTIONS(1096),
    [aux_sym_if_statement_token2] = ACTIONS(1096),
    [aux_sym_else_if_clause_token1] = ACTIONS(1096),
    [aux_sym_else_clause_token1] = ACTIONS(1096),
    [aux_sym_match_expression_token1] = ACTIONS(1096),
    [aux_sym_match_default_expression_token1] = ACTIONS(1096),
    [aux_sym_switch_statement_token1] = ACTIONS(1096),
    [aux_sym_switch_block_token1] = ACTIONS(1096),
    [anon_sym_PLUS] = ACTIONS(1096),
    [anon_sym_DASH] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1094),
    [anon_sym_BANG] = ACTIONS(1094),
    [anon_sym_AT] = ACTIONS(1094),
    [aux_sym_clone_expression_token1] = ACTIONS(1096),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1096),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1096),
    [anon_sym_DASH_DASH] = ACTIONS(1094),
    [anon_sym_PLUS_PLUS] = ACTIONS(1094),
    [aux_sym__list_destructing_token1] = ACTIONS(1096),
    [anon_sym_LBRACK] = ACTIONS(1094),
    [anon_sym_self] = ACTIONS(1096),
    [anon_sym_parent] = ACTIONS(1096),
    [aux_sym__argument_name_token1] = ACTIONS(1096),
    [aux_sym__argument_name_token2] = ACTIONS(1096),
    [anon_sym_POUND_LBRACK] = ACTIONS(1094),
    [aux_sym_encapsed_string_token1] = ACTIONS(1094),
    [anon_sym_DQUOTE] = ACTIONS(1094),
    [aux_sym_string_token1] = ACTIONS(1094),
    [anon_sym_SQUOTE] = ACTIONS(1094),
    [anon_sym_LT_LT_LT] = ACTIONS(1094),
    [anon_sym_BQUOTE] = ACTIONS(1094),
    [anon_sym_DOLLAR] = ACTIONS(1094),
    [aux_sym_yield_expression_token1] = ACTIONS(1096),
    [aux_sym_include_expression_token1] = ACTIONS(1096),
    [aux_sym_include_once_expression_token1] = ACTIONS(1096),
    [aux_sym_require_expression_token1] = ACTIONS(1096),
    [aux_sym_require_once_expression_token1] = ACTIONS(1096),
    [sym_comment] = ACTIONS(3),
  },
  [455] = {
    [ts_builtin_sym_end] = ACTIONS(1098),
    [sym_name] = ACTIONS(1100),
    [anon_sym_QMARK_GT] = ACTIONS(1098),
    [anon_sym_SEMI] = ACTIONS(1098),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1100),
    [aux_sym_global_declaration_token1] = ACTIONS(1100),
    [aux_sym_namespace_definition_token1] = ACTIONS(1100),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1100),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1100),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1100),
    [anon_sym_BSLASH] = ACTIONS(1098),
    [anon_sym_LBRACE] = ACTIONS(1098),
    [anon_sym_RBRACE] = ACTIONS(1098),
    [aux_sym_trait_declaration_token1] = ACTIONS(1100),
    [aux_sym_interface_declaration_token1] = ACTIONS(1100),
    [aux_sym_enum_declaration_token1] = ACTIONS(1100),
    [aux_sym_enum_case_token1] = ACTIONS(1100),
    [aux_sym_class_declaration_token1] = ACTIONS(1100),
    [aux_sym_final_modifier_token1] = ACTIONS(1100),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1100),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1100),
    [sym_var_modifier] = ACTIONS(1100),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1100),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1100),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1100),
    [anon_sym_LPAREN2] = ACTIONS(1098),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1100),
    [aux_sym_cast_type_token1] = ACTIONS(1100),
    [aux_sym_echo_statement_token1] = ACTIONS(1100),
    [aux_sym_exit_statement_token1] = ACTIONS(1100),
    [anon_sym_unset] = ACTIONS(1100),
    [aux_sym_declare_statement_token1] = ACTIONS(1100),
    [aux_sym_declare_statement_token2] = ACTIONS(1100),
    [sym_float] = ACTIONS(1100),
    [aux_sym_try_statement_token1] = ACTIONS(1100),
    [aux_sym_goto_statement_token1] = ACTIONS(1100),
    [aux_sym_continue_statement_token1] = ACTIONS(1100),
    [aux_sym_break_statement_token1] = ACTIONS(1100),
    [sym_integer] = ACTIONS(1100),
    [aux_sym_return_statement_token1] = ACTIONS(1100),
    [aux_sym_throw_expression_token1] = ACTIONS(1100),
    [aux_sym_while_statement_token1] = ACTIONS(1100),
    [aux_sym_while_statement_token2] = ACTIONS(1100),
    [aux_sym_do_statement_token1] = ACTIONS(1100),
    [aux_sym_for_statement_token1] = ACTIONS(1100),
    [aux_sym_for_statement_token2] = ACTIONS(1100),
    [aux_sym_foreach_statement_token1] = ACTIONS(1100),
    [aux_sym_foreach_statement_token2] = ACTIONS(1100),
    [aux_sym_if_statement_token1] = ACTIONS(1100),
    [aux_sym_if_statement_token2] = ACTIONS(1100),
    [aux_sym_else_if_clause_token1] = ACTIONS(1100),
    [aux_sym_else_clause_token1] = ACTIONS(1100),
    [aux_sym_match_expression_token1] = ACTIONS(1100),
    [aux_sym_match_default_expression_token1] = ACTIONS(1100),
    [aux_sym_switch_statement_token1] = ACTIONS(1100),
    [aux_sym_switch_block_token1] = ACTIONS(1100),
    [anon_sym_PLUS] = ACTIONS(1100),
    [anon_sym_DASH] = ACTIONS(1100),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_BANG] = ACTIONS(1098),
    [anon_sym_AT] = ACTIONS(1098),
    [aux_sym_clone_expression_token1] = ACTIONS(1100),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1100),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1100),
    [anon_sym_DASH_DASH] = ACTIONS(1098),
    [anon_sym_PLUS_PLUS] = ACTIONS(1098),
    [aux_sym__list_destructing_token1] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1098),
    [anon_sym_self] = ACTIONS(1100),
    [anon_sym_parent] = ACTIONS(1100),
    [aux_sym__argument_name_token1] = ACTIONS(1100),
    [aux_sym__argument_name_token2] = ACTIONS(1100),
    [anon_sym_POUND_LBRACK] = ACTIONS(1098),
    [aux_sym_encapsed_string_token1] = ACTIONS(1098),
    [anon_sym_DQUOTE] = ACTIONS(1098),
    [aux_sym_string_token1] = ACTIONS(1098),
    [anon_sym_SQUOTE] = ACTIONS(1098),
    [anon_sym_LT_LT_LT] = ACTIONS(1098),
    [anon_sym_BQUOTE] = ACTIONS(1098),
    [anon_sym_DOLLAR] = ACTIONS(1098),
    [aux_sym_yield_expression_token1] = ACTIONS(1100),
    [aux_sym_include_expression_token1] = ACTIONS(1100),
    [aux_sym_include_once_expression_token1] = ACTIONS(1100),
    [aux_sym_require_expression_token1] = ACTIONS(1100),
    [aux_sym_require_once_expression_token1] = ACTIONS(1100),
    [sym_comment] = ACTIONS(3),
  },
  [456] = {
    [ts_builtin_sym_end] = ACTIONS(1102),
    [sym_name] = ACTIONS(1104),
    [anon_sym_QMARK_GT] = ACTIONS(1102),
    [anon_sym_SEMI] = ACTIONS(1102),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1104),
    [aux_sym_global_declaration_token1] = ACTIONS(1104),
    [aux_sym_namespace_definition_token1] = ACTIONS(1104),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1104),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1104),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1104),
    [anon_sym_BSLASH] = ACTIONS(1102),
    [anon_sym_LBRACE] = ACTIONS(1102),
    [anon_sym_RBRACE] = ACTIONS(1102),
    [aux_sym_trait_declaration_token1] = ACTIONS(1104),
    [aux_sym_interface_declaration_token1] = ACTIONS(1104),
    [aux_sym_enum_declaration_token1] = ACTIONS(1104),
    [aux_sym_enum_case_token1] = ACTIONS(1104),
    [aux_sym_class_declaration_token1] = ACTIONS(1104),
    [aux_sym_final_modifier_token1] = ACTIONS(1104),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1104),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1104),
    [sym_var_modifier] = ACTIONS(1104),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1104),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1104),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1104),
    [anon_sym_LPAREN2] = ACTIONS(1102),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1104),
    [aux_sym_cast_type_token1] = ACTIONS(1104),
    [aux_sym_echo_statement_token1] = ACTIONS(1104),
    [aux_sym_exit_statement_token1] = ACTIONS(1104),
    [anon_sym_unset] = ACTIONS(1104),
    [aux_sym_declare_statement_token1] = ACTIONS(1104),
    [aux_sym_declare_statement_token2] = ACTIONS(1104),
    [sym_float] = ACTIONS(1104),
    [aux_sym_try_statement_token1] = ACTIONS(1104),
    [aux_sym_goto_statement_token1] = ACTIONS(1104),
    [aux_sym_continue_statement_token1] = ACTIONS(1104),
    [aux_sym_break_statement_token1] = ACTIONS(1104),
    [sym_integer] = ACTIONS(1104),
    [aux_sym_return_statement_token1] = ACTIONS(1104),
    [aux_sym_throw_expression_token1] = ACTIONS(1104),
    [aux_sym_while_statement_token1] = ACTIONS(1104),
    [aux_sym_while_statement_token2] = ACTIONS(1104),
    [aux_sym_do_statement_token1] = ACTIONS(1104),
    [aux_sym_for_statement_token1] = ACTIONS(1104),
    [aux_sym_for_statement_token2] = ACTIONS(1104),
    [aux_sym_foreach_statement_token1] = ACTIONS(1104),
    [aux_sym_foreach_statement_token2] = ACTIONS(1104),
    [aux_sym_if_statement_token1] = ACTIONS(1104),
    [aux_sym_if_statement_token2] = ACTIONS(1104),
    [aux_sym_else_if_clause_token1] = ACTIONS(1104),
    [aux_sym_else_clause_token1] = ACTIONS(1104),
    [aux_sym_match_expression_token1] = ACTIONS(1104),
    [aux_sym_match_default_expression_token1] = ACTIONS(1104),
    [aux_sym_switch_statement_token1] = ACTIONS(1104),
    [aux_sym_switch_block_token1] = ACTIONS(1104),
    [anon_sym_PLUS] = ACTIONS(1104),
    [anon_sym_DASH] = ACTIONS(1104),
    [anon_sym_TILDE] = ACTIONS(1102),
    [anon_sym_BANG] = ACTIONS(1102),
    [anon_sym_AT] = ACTIONS(1102),
    [aux_sym_clone_expression_token1] = ACTIONS(1104),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1104),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1104),
    [anon_sym_DASH_DASH] = ACTIONS(1102),
    [anon_sym_PLUS_PLUS] = ACTIONS(1102),
    [aux_sym__list_destructing_token1] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1102),
    [anon_sym_self] = ACTIONS(1104),
    [anon_sym_parent] = ACTIONS(1104),
    [aux_sym__argument_name_token1] = ACTIONS(1104),
    [aux_sym__argument_name_token2] = ACTIONS(1104),
    [anon_sym_POUND_LBRACK] = ACTIONS(1102),
    [aux_sym_encapsed_string_token1] = ACTIONS(1102),
    [anon_sym_DQUOTE] = ACTIONS(1102),
    [aux_sym_string_token1] = ACTIONS(1102),
    [anon_sym_SQUOTE] = ACTIONS(1102),
    [anon_sym_LT_LT_LT] = ACTIONS(1102),
    [anon_sym_BQUOTE] = ACTIONS(1102),
    [anon_sym_DOLLAR] = ACTIONS(1102),
    [aux_sym_yield_expression_token1] = ACTIONS(1104),
    [aux_sym_include_expression_token1] = ACTIONS(1104),
    [aux_sym_include_once_expression_token1] = ACTIONS(1104),
    [aux_sym_require_expression_token1] = ACTIONS(1104),
    [aux_sym_require_once_expression_token1] = ACTIONS(1104),
    [sym_comment] = ACTIONS(3),
  },
  [457] = {
    [ts_builtin_sym_end] = ACTIONS(1106),
    [sym_name] = ACTIONS(1108),
    [anon_sym_QMARK_GT] = ACTIONS(1106),
    [anon_sym_SEMI] = ACTIONS(1106),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1108),
    [aux_sym_global_declaration_token1] = ACTIONS(1108),
    [aux_sym_namespace_definition_token1] = ACTIONS(1108),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1108),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1108),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1108),
    [anon_sym_BSLASH] = ACTIONS(1106),
    [anon_sym_LBRACE] = ACTIONS(1106),
    [anon_sym_RBRACE] = ACTIONS(1106),
    [aux_sym_trait_declaration_token1] = ACTIONS(1108),
    [aux_sym_interface_declaration_token1] = ACTIONS(1108),
    [aux_sym_enum_declaration_token1] = ACTIONS(1108),
    [aux_sym_enum_case_token1] = ACTIONS(1108),
    [aux_sym_class_declaration_token1] = ACTIONS(1108),
    [aux_sym_final_modifier_token1] = ACTIONS(1108),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1108),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1108),
    [sym_var_modifier] = ACTIONS(1108),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1108),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1108),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1108),
    [anon_sym_LPAREN2] = ACTIONS(1106),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1108),
    [aux_sym_cast_type_token1] = ACTIONS(1108),
    [aux_sym_echo_statement_token1] = ACTIONS(1108),
    [aux_sym_exit_statement_token1] = ACTIONS(1108),
    [anon_sym_unset] = ACTIONS(1108),
    [aux_sym_declare_statement_token1] = ACTIONS(1108),
    [aux_sym_declare_statement_token2] = ACTIONS(1108),
    [sym_float] = ACTIONS(1108),
    [aux_sym_try_statement_token1] = ACTIONS(1108),
    [aux_sym_goto_statement_token1] = ACTIONS(1108),
    [aux_sym_continue_statement_token1] = ACTIONS(1108),
    [aux_sym_break_statement_token1] = ACTIONS(1108),
    [sym_integer] = ACTIONS(1108),
    [aux_sym_return_statement_token1] = ACTIONS(1108),
    [aux_sym_throw_expression_token1] = ACTIONS(1108),
    [aux_sym_while_statement_token1] = ACTIONS(1108),
    [aux_sym_while_statement_token2] = ACTIONS(1108),
    [aux_sym_do_statement_token1] = ACTIONS(1108),
    [aux_sym_for_statement_token1] = ACTIONS(1108),
    [aux_sym_for_statement_token2] = ACTIONS(1108),
    [aux_sym_foreach_statement_token1] = ACTIONS(1108),
    [aux_sym_foreach_statement_token2] = ACTIONS(1108),
    [aux_sym_if_statement_token1] = ACTIONS(1108),
    [aux_sym_if_statement_token2] = ACTIONS(1108),
    [aux_sym_else_if_clause_token1] = ACTIONS(1108),
    [aux_sym_else_clause_token1] = ACTIONS(1108),
    [aux_sym_match_expression_token1] = ACTIONS(1108),
    [aux_sym_match_default_expression_token1] = ACTIONS(1108),
    [aux_sym_switch_statement_token1] = ACTIONS(1108),
    [aux_sym_switch_block_token1] = ACTIONS(1108),
    [anon_sym_PLUS] = ACTIONS(1108),
    [anon_sym_DASH] = ACTIONS(1108),
    [anon_sym_TILDE] = ACTIONS(1106),
    [anon_sym_BANG] = ACTIONS(1106),
    [anon_sym_AT] = ACTIONS(1106),
    [aux_sym_clone_expression_token1] = ACTIONS(1108),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1108),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1108),
    [anon_sym_DASH_DASH] = ACTIONS(1106),
    [anon_sym_PLUS_PLUS] = ACTIONS(1106),
    [aux_sym__list_destructing_token1] = ACTIONS(1108),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_self] = ACTIONS(1108),
    [anon_sym_parent] = ACTIONS(1108),
    [aux_sym__argument_name_token1] = ACTIONS(1108),
    [aux_sym__argument_name_token2] = ACTIONS(1108),
    [anon_sym_POUND_LBRACK] = ACTIONS(1106),
    [aux_sym_encapsed_string_token1] = ACTIONS(1106),
    [anon_sym_DQUOTE] = ACTIONS(1106),
    [aux_sym_string_token1] = ACTIONS(1106),
    [anon_sym_SQUOTE] = ACTIONS(1106),
    [anon_sym_LT_LT_LT] = ACTIONS(1106),
    [anon_sym_BQUOTE] = ACTIONS(1106),
    [anon_sym_DOLLAR] = ACTIONS(1106),
    [aux_sym_yield_expression_token1] = ACTIONS(1108),
    [aux_sym_include_expression_token1] = ACTIONS(1108),
    [aux_sym_include_once_expression_token1] = ACTIONS(1108),
    [aux_sym_require_expression_token1] = ACTIONS(1108),
    [aux_sym_require_once_expression_token1] = ACTIONS(1108),
    [sym_comment] = ACTIONS(3),
  },
  [458] = {
    [ts_builtin_sym_end] = ACTIONS(1110),
    [sym_name] = ACTIONS(1112),
    [anon_sym_QMARK_GT] = ACTIONS(1110),
    [anon_sym_SEMI] = ACTIONS(1110),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1112),
    [aux_sym_global_declaration_token1] = ACTIONS(1112),
    [aux_sym_namespace_definition_token1] = ACTIONS(1112),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1112),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1112),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1112),
    [anon_sym_BSLASH] = ACTIONS(1110),
    [anon_sym_LBRACE] = ACTIONS(1110),
    [anon_sym_RBRACE] = ACTIONS(1110),
    [aux_sym_trait_declaration_token1] = ACTIONS(1112),
    [aux_sym_interface_declaration_token1] = ACTIONS(1112),
    [aux_sym_enum_declaration_token1] = ACTIONS(1112),
    [aux_sym_enum_case_token1] = ACTIONS(1112),
    [aux_sym_class_declaration_token1] = ACTIONS(1112),
    [aux_sym_final_modifier_token1] = ACTIONS(1112),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1112),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1112),
    [sym_var_modifier] = ACTIONS(1112),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1112),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1112),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1112),
    [anon_sym_LPAREN2] = ACTIONS(1110),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1112),
    [aux_sym_cast_type_token1] = ACTIONS(1112),
    [aux_sym_echo_statement_token1] = ACTIONS(1112),
    [aux_sym_exit_statement_token1] = ACTIONS(1112),
    [anon_sym_unset] = ACTIONS(1112),
    [aux_sym_declare_statement_token1] = ACTIONS(1112),
    [aux_sym_declare_statement_token2] = ACTIONS(1112),
    [sym_float] = ACTIONS(1112),
    [aux_sym_try_statement_token1] = ACTIONS(1112),
    [aux_sym_goto_statement_token1] = ACTIONS(1112),
    [aux_sym_continue_statement_token1] = ACTIONS(1112),
    [aux_sym_break_statement_token1] = ACTIONS(1112),
    [sym_integer] = ACTIONS(1112),
    [aux_sym_return_statement_token1] = ACTIONS(1112),
    [aux_sym_throw_expression_token1] = ACTIONS(1112),
    [aux_sym_while_statement_token1] = ACTIONS(1112),
    [aux_sym_while_statement_token2] = ACTIONS(1112),
    [aux_sym_do_statement_token1] = ACTIONS(1112),
    [aux_sym_for_statement_token1] = ACTIONS(1112),
    [aux_sym_for_statement_token2] = ACTIONS(1112),
    [aux_sym_foreach_statement_token1] = ACTIONS(1112),
    [aux_sym_foreach_statement_token2] = ACTIONS(1112),
    [aux_sym_if_statement_token1] = ACTIONS(1112),
    [aux_sym_if_statement_token2] = ACTIONS(1112),
    [aux_sym_else_if_clause_token1] = ACTIONS(1112),
    [aux_sym_else_clause_token1] = ACTIONS(1112),
    [aux_sym_match_expression_token1] = ACTIONS(1112),
    [aux_sym_match_default_expression_token1] = ACTIONS(1112),
    [aux_sym_switch_statement_token1] = ACTIONS(1112),
    [aux_sym_switch_block_token1] = ACTIONS(1112),
    [anon_sym_PLUS] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_TILDE] = ACTIONS(1110),
    [anon_sym_BANG] = ACTIONS(1110),
    [anon_sym_AT] = ACTIONS(1110),
    [aux_sym_clone_expression_token1] = ACTIONS(1112),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1112),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1112),
    [anon_sym_DASH_DASH] = ACTIONS(1110),
    [anon_sym_PLUS_PLUS] = ACTIONS(1110),
    [aux_sym__list_destructing_token1] = ACTIONS(1112),
    [anon_sym_LBRACK] = ACTIONS(1110),
    [anon_sym_self] = ACTIONS(1112),
    [anon_sym_parent] = ACTIONS(1112),
    [aux_sym__argument_name_token1] = ACTIONS(1112),
    [aux_sym__argument_name_token2] = ACTIONS(1112),
    [anon_sym_POUND_LBRACK] = ACTIONS(1110),
    [aux_sym_encapsed_string_token1] = ACTIONS(1110),
    [anon_sym_DQUOTE] = ACTIONS(1110),
    [aux_sym_string_token1] = ACTIONS(1110),
    [anon_sym_SQUOTE] = ACTIONS(1110),
    [anon_sym_LT_LT_LT] = ACTIONS(1110),
    [anon_sym_BQUOTE] = ACTIONS(1110),
    [anon_sym_DOLLAR] = ACTIONS(1110),
    [aux_sym_yield_expression_token1] = ACTIONS(1112),
    [aux_sym_include_expression_token1] = ACTIONS(1112),
    [aux_sym_include_once_expression_token1] = ACTIONS(1112),
    [aux_sym_require_expression_token1] = ACTIONS(1112),
    [aux_sym_require_once_expression_token1] = ACTIONS(1112),
    [sym_comment] = ACTIONS(3),
  },
  [459] = {
    [ts_builtin_sym_end] = ACTIONS(1114),
    [sym_name] = ACTIONS(1116),
    [anon_sym_QMARK_GT] = ACTIONS(1114),
    [anon_sym_SEMI] = ACTIONS(1114),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1116),
    [aux_sym_global_declaration_token1] = ACTIONS(1116),
    [aux_sym_namespace_definition_token1] = ACTIONS(1116),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1116),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1116),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1116),
    [anon_sym_BSLASH] = ACTIONS(1114),
    [anon_sym_LBRACE] = ACTIONS(1114),
    [anon_sym_RBRACE] = ACTIONS(1114),
    [aux_sym_trait_declaration_token1] = ACTIONS(1116),
    [aux_sym_interface_declaration_token1] = ACTIONS(1116),
    [aux_sym_enum_declaration_token1] = ACTIONS(1116),
    [aux_sym_enum_case_token1] = ACTIONS(1116),
    [aux_sym_class_declaration_token1] = ACTIONS(1116),
    [aux_sym_final_modifier_token1] = ACTIONS(1116),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1116),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1116),
    [sym_var_modifier] = ACTIONS(1116),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1116),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1116),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1116),
    [anon_sym_LPAREN2] = ACTIONS(1114),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1116),
    [aux_sym_cast_type_token1] = ACTIONS(1116),
    [aux_sym_echo_statement_token1] = ACTIONS(1116),
    [aux_sym_exit_statement_token1] = ACTIONS(1116),
    [anon_sym_unset] = ACTIONS(1116),
    [aux_sym_declare_statement_token1] = ACTIONS(1116),
    [aux_sym_declare_statement_token2] = ACTIONS(1116),
    [sym_float] = ACTIONS(1116),
    [aux_sym_try_statement_token1] = ACTIONS(1116),
    [aux_sym_goto_statement_token1] = ACTIONS(1116),
    [aux_sym_continue_statement_token1] = ACTIONS(1116),
    [aux_sym_break_statement_token1] = ACTIONS(1116),
    [sym_integer] = ACTIONS(1116),
    [aux_sym_return_statement_token1] = ACTIONS(1116),
    [aux_sym_throw_expression_token1] = ACTIONS(1116),
    [aux_sym_while_statement_token1] = ACTIONS(1116),
    [aux_sym_while_statement_token2] = ACTIONS(1116),
    [aux_sym_do_statement_token1] = ACTIONS(1116),
    [aux_sym_for_statement_token1] = ACTIONS(1116),
    [aux_sym_for_statement_token2] = ACTIONS(1116),
    [aux_sym_foreach_statement_token1] = ACTIONS(1116),
    [aux_sym_foreach_statement_token2] = ACTIONS(1116),
    [aux_sym_if_statement_token1] = ACTIONS(1116),
    [aux_sym_if_statement_token2] = ACTIONS(1116),
    [aux_sym_else_if_clause_token1] = ACTIONS(1116),
    [aux_sym_else_clause_token1] = ACTIONS(1116),
    [aux_sym_match_expression_token1] = ACTIONS(1116),
    [aux_sym_match_default_expression_token1] = ACTIONS(1116),
    [aux_sym_switch_statement_token1] = ACTIONS(1116),
    [aux_sym_switch_block_token1] = ACTIONS(1116),
    [anon_sym_PLUS] = ACTIONS(1116),
    [anon_sym_DASH] = ACTIONS(1116),
    [anon_sym_TILDE] = ACTIONS(1114),
    [anon_sym_BANG] = ACTIONS(1114),
    [anon_sym_AT] = ACTIONS(1114),
    [aux_sym_clone_expression_token1] = ACTIONS(1116),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1116),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1116),
    [anon_sym_DASH_DASH] = ACTIONS(1114),
    [anon_sym_PLUS_PLUS] = ACTIONS(1114),
    [aux_sym__list_destructing_token1] = ACTIONS(1116),
    [anon_sym_LBRACK] = ACTIONS(1114),
    [anon_sym_self] = ACTIONS(1116),
    [anon_sym_parent] = ACTIONS(1116),
    [aux_sym__argument_name_token1] = ACTIONS(1116),
    [aux_sym__argument_name_token2] = ACTIONS(1116),
    [anon_sym_POUND_LBRACK] = ACTIONS(1114),
    [aux_sym_encapsed_string_token1] = ACTIONS(1114),
    [anon_sym_DQUOTE] = ACTIONS(1114),
    [aux_sym_string_token1] = ACTIONS(1114),
    [anon_sym_SQUOTE] = ACTIONS(1114),
    [anon_sym_LT_LT_LT] = ACTIONS(1114),
    [anon_sym_BQUOTE] = ACTIONS(1114),
    [anon_sym_DOLLAR] = ACTIONS(1114),
    [aux_sym_yield_expression_token1] = ACTIONS(1116),
    [aux_sym_include_expression_token1] = ACTIONS(1116),
    [aux_sym_include_once_expression_token1] = ACTIONS(1116),
    [aux_sym_require_expression_token1] = ACTIONS(1116),
    [aux_sym_require_once_expression_token1] = ACTIONS(1116),
    [sym_comment] = ACTIONS(3),
  },
  [460] = {
    [ts_builtin_sym_end] = ACTIONS(1118),
    [sym_name] = ACTIONS(1120),
    [anon_sym_QMARK_GT] = ACTIONS(1118),
    [anon_sym_SEMI] = ACTIONS(1118),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1120),
    [aux_sym_global_declaration_token1] = ACTIONS(1120),
    [aux_sym_namespace_definition_token1] = ACTIONS(1120),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1120),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1120),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1120),
    [anon_sym_BSLASH] = ACTIONS(1118),
    [anon_sym_LBRACE] = ACTIONS(1118),
    [anon_sym_RBRACE] = ACTIONS(1118),
    [aux_sym_trait_declaration_token1] = ACTIONS(1120),
    [aux_sym_interface_declaration_token1] = ACTIONS(1120),
    [aux_sym_enum_declaration_token1] = ACTIONS(1120),
    [aux_sym_enum_case_token1] = ACTIONS(1120),
    [aux_sym_class_declaration_token1] = ACTIONS(1120),
    [aux_sym_final_modifier_token1] = ACTIONS(1120),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1120),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1120),
    [sym_var_modifier] = ACTIONS(1120),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1120),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1120),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1120),
    [anon_sym_LPAREN2] = ACTIONS(1118),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1120),
    [aux_sym_cast_type_token1] = ACTIONS(1120),
    [aux_sym_echo_statement_token1] = ACTIONS(1120),
    [aux_sym_exit_statement_token1] = ACTIONS(1120),
    [anon_sym_unset] = ACTIONS(1120),
    [aux_sym_declare_statement_token1] = ACTIONS(1120),
    [aux_sym_declare_statement_token2] = ACTIONS(1120),
    [sym_float] = ACTIONS(1120),
    [aux_sym_try_statement_token1] = ACTIONS(1120),
    [aux_sym_goto_statement_token1] = ACTIONS(1120),
    [aux_sym_continue_statement_token1] = ACTIONS(1120),
    [aux_sym_break_statement_token1] = ACTIONS(1120),
    [sym_integer] = ACTIONS(1120),
    [aux_sym_return_statement_token1] = ACTIONS(1120),
    [aux_sym_throw_expression_token1] = ACTIONS(1120),
    [aux_sym_while_statement_token1] = ACTIONS(1120),
    [aux_sym_while_statement_token2] = ACTIONS(1120),
    [aux_sym_do_statement_token1] = ACTIONS(1120),
    [aux_sym_for_statement_token1] = ACTIONS(1120),
    [aux_sym_for_statement_token2] = ACTIONS(1120),
    [aux_sym_foreach_statement_token1] = ACTIONS(1120),
    [aux_sym_foreach_statement_token2] = ACTIONS(1120),
    [aux_sym_if_statement_token1] = ACTIONS(1120),
    [aux_sym_if_statement_token2] = ACTIONS(1120),
    [aux_sym_else_if_clause_token1] = ACTIONS(1120),
    [aux_sym_else_clause_token1] = ACTIONS(1120),
    [aux_sym_match_expression_token1] = ACTIONS(1120),
    [aux_sym_match_default_expression_token1] = ACTIONS(1120),
    [aux_sym_switch_statement_token1] = ACTIONS(1120),
    [aux_sym_switch_block_token1] = ACTIONS(1120),
    [anon_sym_PLUS] = ACTIONS(1120),
    [anon_sym_DASH] = ACTIONS(1120),
    [anon_sym_TILDE] = ACTIONS(1118),
    [anon_sym_BANG] = ACTIONS(1118),
    [anon_sym_AT] = ACTIONS(1118),
    [aux_sym_clone_expression_token1] = ACTIONS(1120),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1120),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1120),
    [anon_sym_DASH_DASH] = ACTIONS(1118),
    [anon_sym_PLUS_PLUS] = ACTIONS(1118),
    [aux_sym__list_destructing_token1] = ACTIONS(1120),
    [anon_sym_LBRACK] = ACTIONS(1118),
    [anon_sym_self] = ACTIONS(1120),
    [anon_sym_parent] = ACTIONS(1120),
    [aux_sym__argument_name_token1] = ACTIONS(1120),
    [aux_sym__argument_name_token2] = ACTIONS(1120),
    [anon_sym_POUND_LBRACK] = ACTIONS(1118),
    [aux_sym_encapsed_string_token1] = ACTIONS(1118),
    [anon_sym_DQUOTE] = ACTIONS(1118),
    [aux_sym_string_token1] = ACTIONS(1118),
    [anon_sym_SQUOTE] = ACTIONS(1118),
    [anon_sym_LT_LT_LT] = ACTIONS(1118),
    [anon_sym_BQUOTE] = ACTIONS(1118),
    [anon_sym_DOLLAR] = ACTIONS(1118),
    [aux_sym_yield_expression_token1] = ACTIONS(1120),
    [aux_sym_include_expression_token1] = ACTIONS(1120),
    [aux_sym_include_once_expression_token1] = ACTIONS(1120),
    [aux_sym_require_expression_token1] = ACTIONS(1120),
    [aux_sym_require_once_expression_token1] = ACTIONS(1120),
    [sym_comment] = ACTIONS(3),
  },
  [461] = {
    [ts_builtin_sym_end] = ACTIONS(1079),
    [sym_name] = ACTIONS(1081),
    [anon_sym_QMARK_GT] = ACTIONS(1079),
    [anon_sym_SEMI] = ACTIONS(1079),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1081),
    [aux_sym_global_declaration_token1] = ACTIONS(1081),
    [aux_sym_namespace_definition_token1] = ACTIONS(1081),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1081),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1081),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1081),
    [anon_sym_BSLASH] = ACTIONS(1079),
    [anon_sym_LBRACE] = ACTIONS(1079),
    [anon_sym_RBRACE] = ACTIONS(1079),
    [aux_sym_trait_declaration_token1] = ACTIONS(1081),
    [aux_sym_interface_declaration_token1] = ACTIONS(1081),
    [aux_sym_enum_declaration_token1] = ACTIONS(1081),
    [aux_sym_enum_case_token1] = ACTIONS(1081),
    [aux_sym_class_declaration_token1] = ACTIONS(1081),
    [aux_sym_final_modifier_token1] = ACTIONS(1081),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1081),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1081),
    [sym_var_modifier] = ACTIONS(1081),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1081),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1081),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1081),
    [anon_sym_LPAREN2] = ACTIONS(1079),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1081),
    [aux_sym_cast_type_token1] = ACTIONS(1081),
    [aux_sym_echo_statement_token1] = ACTIONS(1081),
    [aux_sym_exit_statement_token1] = ACTIONS(1081),
    [anon_sym_unset] = ACTIONS(1081),
    [aux_sym_declare_statement_token1] = ACTIONS(1081),
    [aux_sym_declare_statement_token2] = ACTIONS(1081),
    [sym_float] = ACTIONS(1081),
    [aux_sym_try_statement_token1] = ACTIONS(1081),
    [aux_sym_goto_statement_token1] = ACTIONS(1081),
    [aux_sym_continue_statement_token1] = ACTIONS(1081),
    [aux_sym_break_statement_token1] = ACTIONS(1081),
    [sym_integer] = ACTIONS(1081),
    [aux_sym_return_statement_token1] = ACTIONS(1081),
    [aux_sym_throw_expression_token1] = ACTIONS(1081),
    [aux_sym_while_statement_token1] = ACTIONS(1081),
    [aux_sym_while_statement_token2] = ACTIONS(1081),
    [aux_sym_do_statement_token1] = ACTIONS(1081),
    [aux_sym_for_statement_token1] = ACTIONS(1081),
    [aux_sym_for_statement_token2] = ACTIONS(1081),
    [aux_sym_foreach_statement_token1] = ACTIONS(1081),
    [aux_sym_foreach_statement_token2] = ACTIONS(1081),
    [aux_sym_if_statement_token1] = ACTIONS(1081),
    [aux_sym_if_statement_token2] = ACTIONS(1081),
    [aux_sym_else_if_clause_token1] = ACTIONS(1081),
    [aux_sym_else_clause_token1] = ACTIONS(1081),
    [aux_sym_match_expression_token1] = ACTIONS(1081),
    [aux_sym_match_default_expression_token1] = ACTIONS(1081),
    [aux_sym_switch_statement_token1] = ACTIONS(1081),
    [aux_sym_switch_block_token1] = ACTIONS(1081),
    [anon_sym_PLUS] = ACTIONS(1081),
    [anon_sym_DASH] = ACTIONS(1081),
    [anon_sym_TILDE] = ACTIONS(1079),
    [anon_sym_BANG] = ACTIONS(1079),
    [anon_sym_AT] = ACTIONS(1079),
    [aux_sym_clone_expression_token1] = ACTIONS(1081),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1081),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1081),
    [anon_sym_DASH_DASH] = ACTIONS(1079),
    [anon_sym_PLUS_PLUS] = ACTIONS(1079),
    [aux_sym__list_destructing_token1] = ACTIONS(1081),
    [anon_sym_LBRACK] = ACTIONS(1079),
    [anon_sym_self] = ACTIONS(1081),
    [anon_sym_parent] = ACTIONS(1081),
    [aux_sym__argument_name_token1] = ACTIONS(1081),
    [aux_sym__argument_name_token2] = ACTIONS(1081),
    [anon_sym_POUND_LBRACK] = ACTIONS(1079),
    [aux_sym_encapsed_string_token1] = ACTIONS(1079),
    [anon_sym_DQUOTE] = ACTIONS(1079),
    [aux_sym_string_token1] = ACTIONS(1079),
    [anon_sym_SQUOTE] = ACTIONS(1079),
    [anon_sym_LT_LT_LT] = ACTIONS(1079),
    [anon_sym_BQUOTE] = ACTIONS(1079),
    [anon_sym_DOLLAR] = ACTIONS(1079),
    [aux_sym_yield_expression_token1] = ACTIONS(1081),
    [aux_sym_include_expression_token1] = ACTIONS(1081),
    [aux_sym_include_once_expression_token1] = ACTIONS(1081),
    [aux_sym_require_expression_token1] = ACTIONS(1081),
    [aux_sym_require_once_expression_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(3),
  },
  [462] = {
    [ts_builtin_sym_end] = ACTIONS(1122),
    [sym_name] = ACTIONS(1124),
    [anon_sym_QMARK_GT] = ACTIONS(1122),
    [anon_sym_SEMI] = ACTIONS(1122),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1124),
    [aux_sym_global_declaration_token1] = ACTIONS(1124),
    [aux_sym_namespace_definition_token1] = ACTIONS(1124),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1124),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1124),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1124),
    [anon_sym_BSLASH] = ACTIONS(1122),
    [anon_sym_LBRACE] = ACTIONS(1122),
    [anon_sym_RBRACE] = ACTIONS(1122),
    [aux_sym_trait_declaration_token1] = ACTIONS(1124),
    [aux_sym_interface_declaration_token1] = ACTIONS(1124),
    [aux_sym_enum_declaration_token1] = ACTIONS(1124),
    [aux_sym_enum_case_token1] = ACTIONS(1124),
    [aux_sym_class_declaration_token1] = ACTIONS(1124),
    [aux_sym_final_modifier_token1] = ACTIONS(1124),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1124),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1124),
    [sym_var_modifier] = ACTIONS(1124),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1124),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1124),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1124),
    [anon_sym_LPAREN2] = ACTIONS(1122),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1124),
    [aux_sym_cast_type_token1] = ACTIONS(1124),
    [aux_sym_echo_statement_token1] = ACTIONS(1124),
    [aux_sym_exit_statement_token1] = ACTIONS(1124),
    [anon_sym_unset] = ACTIONS(1124),
    [aux_sym_declare_statement_token1] = ACTIONS(1124),
    [aux_sym_declare_statement_token2] = ACTIONS(1124),
    [sym_float] = ACTIONS(1124),
    [aux_sym_try_statement_token1] = ACTIONS(1124),
    [aux_sym_goto_statement_token1] = ACTIONS(1124),
    [aux_sym_continue_statement_token1] = ACTIONS(1124),
    [aux_sym_break_statement_token1] = ACTIONS(1124),
    [sym_integer] = ACTIONS(1124),
    [aux_sym_return_statement_token1] = ACTIONS(1124),
    [aux_sym_throw_expression_token1] = ACTIONS(1124),
    [aux_sym_while_statement_token1] = ACTIONS(1124),
    [aux_sym_while_statement_token2] = ACTIONS(1124),
    [aux_sym_do_statement_token1] = ACTIONS(1124),
    [aux_sym_for_statement_token1] = ACTIONS(1124),
    [aux_sym_for_statement_token2] = ACTIONS(1124),
    [aux_sym_foreach_statement_token1] = ACTIONS(1124),
    [aux_sym_foreach_statement_token2] = ACTIONS(1124),
    [aux_sym_if_statement_token1] = ACTIONS(1124),
    [aux_sym_if_statement_token2] = ACTIONS(1124),
    [aux_sym_else_if_clause_token1] = ACTIONS(1124),
    [aux_sym_else_clause_token1] = ACTIONS(1124),
    [aux_sym_match_expression_token1] = ACTIONS(1124),
    [aux_sym_match_default_expression_token1] = ACTIONS(1124),
    [aux_sym_switch_statement_token1] = ACTIONS(1124),
    [aux_sym_switch_block_token1] = ACTIONS(1124),
    [anon_sym_PLUS] = ACTIONS(1124),
    [anon_sym_DASH] = ACTIONS(1124),
    [anon_sym_TILDE] = ACTIONS(1122),
    [anon_sym_BANG] = ACTIONS(1122),
    [anon_sym_AT] = ACTIONS(1122),
    [aux_sym_clone_expression_token1] = ACTIONS(1124),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1124),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1124),
    [anon_sym_DASH_DASH] = ACTIONS(1122),
    [anon_sym_PLUS_PLUS] = ACTIONS(1122),
    [aux_sym__list_destructing_token1] = ACTIONS(1124),
    [anon_sym_LBRACK] = ACTIONS(1122),
    [anon_sym_self] = ACTIONS(1124),
    [anon_sym_parent] = ACTIONS(1124),
    [aux_sym__argument_name_token1] = ACTIONS(1124),
    [aux_sym__argument_name_token2] = ACTIONS(1124),
    [anon_sym_POUND_LBRACK] = ACTIONS(1122),
    [aux_sym_encapsed_string_token1] = ACTIONS(1122),
    [anon_sym_DQUOTE] = ACTIONS(1122),
    [aux_sym_string_token1] = ACTIONS(1122),
    [anon_sym_SQUOTE] = ACTIONS(1122),
    [anon_sym_LT_LT_LT] = ACTIONS(1122),
    [anon_sym_BQUOTE] = ACTIONS(1122),
    [anon_sym_DOLLAR] = ACTIONS(1122),
    [aux_sym_yield_expression_token1] = ACTIONS(1124),
    [aux_sym_include_expression_token1] = ACTIONS(1124),
    [aux_sym_include_once_expression_token1] = ACTIONS(1124),
    [aux_sym_require_expression_token1] = ACTIONS(1124),
    [aux_sym_require_once_expression_token1] = ACTIONS(1124),
    [sym_comment] = ACTIONS(3),
  },
  [463] = {
    [ts_builtin_sym_end] = ACTIONS(1126),
    [sym_name] = ACTIONS(1128),
    [anon_sym_QMARK_GT] = ACTIONS(1126),
    [anon_sym_SEMI] = ACTIONS(1126),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1128),
    [aux_sym_global_declaration_token1] = ACTIONS(1128),
    [aux_sym_namespace_definition_token1] = ACTIONS(1128),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1128),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1128),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1128),
    [anon_sym_BSLASH] = ACTIONS(1126),
    [anon_sym_LBRACE] = ACTIONS(1126),
    [anon_sym_RBRACE] = ACTIONS(1126),
    [aux_sym_trait_declaration_token1] = ACTIONS(1128),
    [aux_sym_interface_declaration_token1] = ACTIONS(1128),
    [aux_sym_enum_declaration_token1] = ACTIONS(1128),
    [aux_sym_enum_case_token1] = ACTIONS(1128),
    [aux_sym_class_declaration_token1] = ACTIONS(1128),
    [aux_sym_final_modifier_token1] = ACTIONS(1128),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1128),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1128),
    [sym_var_modifier] = ACTIONS(1128),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1128),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1128),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1128),
    [anon_sym_LPAREN2] = ACTIONS(1126),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1128),
    [aux_sym_cast_type_token1] = ACTIONS(1128),
    [aux_sym_echo_statement_token1] = ACTIONS(1128),
    [aux_sym_exit_statement_token1] = ACTIONS(1128),
    [anon_sym_unset] = ACTIONS(1128),
    [aux_sym_declare_statement_token1] = ACTIONS(1128),
    [aux_sym_declare_statement_token2] = ACTIONS(1128),
    [sym_float] = ACTIONS(1128),
    [aux_sym_try_statement_token1] = ACTIONS(1128),
    [aux_sym_goto_statement_token1] = ACTIONS(1128),
    [aux_sym_continue_statement_token1] = ACTIONS(1128),
    [aux_sym_break_statement_token1] = ACTIONS(1128),
    [sym_integer] = ACTIONS(1128),
    [aux_sym_return_statement_token1] = ACTIONS(1128),
    [aux_sym_throw_expression_token1] = ACTIONS(1128),
    [aux_sym_while_statement_token1] = ACTIONS(1128),
    [aux_sym_while_statement_token2] = ACTIONS(1128),
    [aux_sym_do_statement_token1] = ACTIONS(1128),
    [aux_sym_for_statement_token1] = ACTIONS(1128),
    [aux_sym_for_statement_token2] = ACTIONS(1128),
    [aux_sym_foreach_statement_token1] = ACTIONS(1128),
    [aux_sym_foreach_statement_token2] = ACTIONS(1128),
    [aux_sym_if_statement_token1] = ACTIONS(1128),
    [aux_sym_if_statement_token2] = ACTIONS(1128),
    [aux_sym_else_if_clause_token1] = ACTIONS(1128),
    [aux_sym_else_clause_token1] = ACTIONS(1128),
    [aux_sym_match_expression_token1] = ACTIONS(1128),
    [aux_sym_match_default_expression_token1] = ACTIONS(1128),
    [aux_sym_switch_statement_token1] = ACTIONS(1128),
    [aux_sym_switch_block_token1] = ACTIONS(1128),
    [anon_sym_PLUS] = ACTIONS(1128),
    [anon_sym_DASH] = ACTIONS(1128),
    [anon_sym_TILDE] = ACTIONS(1126),
    [anon_sym_BANG] = ACTIONS(1126),
    [anon_sym_AT] = ACTIONS(1126),
    [aux_sym_clone_expression_token1] = ACTIONS(1128),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1128),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1128),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [anon_sym_PLUS_PLUS] = ACTIONS(1126),
    [aux_sym__list_destructing_token1] = ACTIONS(1128),
    [anon_sym_LBRACK] = ACTIONS(1126),
    [anon_sym_self] = ACTIONS(1128),
    [anon_sym_parent] = ACTIONS(1128),
    [aux_sym__argument_name_token1] = ACTIONS(1128),
    [aux_sym__argument_name_token2] = ACTIONS(1128),
    [anon_sym_POUND_LBRACK] = ACTIONS(1126),
    [aux_sym_encapsed_string_token1] = ACTIONS(1126),
    [anon_sym_DQUOTE] = ACTIONS(1126),
    [aux_sym_string_token1] = ACTIONS(1126),
    [anon_sym_SQUOTE] = ACTIONS(1126),
    [anon_sym_LT_LT_LT] = ACTIONS(1126),
    [anon_sym_BQUOTE] = ACTIONS(1126),
    [anon_sym_DOLLAR] = ACTIONS(1126),
    [aux_sym_yield_expression_token1] = ACTIONS(1128),
    [aux_sym_include_expression_token1] = ACTIONS(1128),
    [aux_sym_include_once_expression_token1] = ACTIONS(1128),
    [aux_sym_require_expression_token1] = ACTIONS(1128),
    [aux_sym_require_once_expression_token1] = ACTIONS(1128),
    [sym_comment] = ACTIONS(3),
  },
  [464] = {
    [ts_builtin_sym_end] = ACTIONS(1130),
    [sym_name] = ACTIONS(1132),
    [anon_sym_QMARK_GT] = ACTIONS(1130),
    [anon_sym_SEMI] = ACTIONS(1130),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1132),
    [aux_sym_global_declaration_token1] = ACTIONS(1132),
    [aux_sym_namespace_definition_token1] = ACTIONS(1132),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1132),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1132),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1132),
    [anon_sym_BSLASH] = ACTIONS(1130),
    [anon_sym_LBRACE] = ACTIONS(1130),
    [anon_sym_RBRACE] = ACTIONS(1130),
    [aux_sym_trait_declaration_token1] = ACTIONS(1132),
    [aux_sym_interface_declaration_token1] = ACTIONS(1132),
    [aux_sym_enum_declaration_token1] = ACTIONS(1132),
    [aux_sym_enum_case_token1] = ACTIONS(1132),
    [aux_sym_class_declaration_token1] = ACTIONS(1132),
    [aux_sym_final_modifier_token1] = ACTIONS(1132),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1132),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1132),
    [sym_var_modifier] = ACTIONS(1132),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1132),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1132),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1132),
    [anon_sym_LPAREN2] = ACTIONS(1130),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1132),
    [aux_sym_cast_type_token1] = ACTIONS(1132),
    [aux_sym_echo_statement_token1] = ACTIONS(1132),
    [aux_sym_exit_statement_token1] = ACTIONS(1132),
    [anon_sym_unset] = ACTIONS(1132),
    [aux_sym_declare_statement_token1] = ACTIONS(1132),
    [aux_sym_declare_statement_token2] = ACTIONS(1132),
    [sym_float] = ACTIONS(1132),
    [aux_sym_try_statement_token1] = ACTIONS(1132),
    [aux_sym_goto_statement_token1] = ACTIONS(1132),
    [aux_sym_continue_statement_token1] = ACTIONS(1132),
    [aux_sym_break_statement_token1] = ACTIONS(1132),
    [sym_integer] = ACTIONS(1132),
    [aux_sym_return_statement_token1] = ACTIONS(1132),
    [aux_sym_throw_expression_token1] = ACTIONS(1132),
    [aux_sym_while_statement_token1] = ACTIONS(1132),
    [aux_sym_while_statement_token2] = ACTIONS(1132),
    [aux_sym_do_statement_token1] = ACTIONS(1132),
    [aux_sym_for_statement_token1] = ACTIONS(1132),
    [aux_sym_for_statement_token2] = ACTIONS(1132),
    [aux_sym_foreach_statement_token1] = ACTIONS(1132),
    [aux_sym_foreach_statement_token2] = ACTIONS(1132),
    [aux_sym_if_statement_token1] = ACTIONS(1132),
    [aux_sym_if_statement_token2] = ACTIONS(1132),
    [aux_sym_else_if_clause_token1] = ACTIONS(1132),
    [aux_sym_else_clause_token1] = ACTIONS(1132),
    [aux_sym_match_expression_token1] = ACTIONS(1132),
    [aux_sym_match_default_expression_token1] = ACTIONS(1132),
    [aux_sym_switch_statement_token1] = ACTIONS(1132),
    [aux_sym_switch_block_token1] = ACTIONS(1132),
    [anon_sym_PLUS] = ACTIONS(1132),
    [anon_sym_DASH] = ACTIONS(1132),
    [anon_sym_TILDE] = ACTIONS(1130),
    [anon_sym_BANG] = ACTIONS(1130),
    [anon_sym_AT] = ACTIONS(1130),
    [aux_sym_clone_expression_token1] = ACTIONS(1132),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1132),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1132),
    [anon_sym_DASH_DASH] = ACTIONS(1130),
    [anon_sym_PLUS_PLUS] = ACTIONS(1130),
    [aux_sym__list_destructing_token1] = ACTIONS(1132),
    [anon_sym_LBRACK] = ACTIONS(1130),
    [anon_sym_self] = ACTIONS(1132),
    [anon_sym_parent] = ACTIONS(1132),
    [aux_sym__argument_name_token1] = ACTIONS(1132),
    [aux_sym__argument_name_token2] = ACTIONS(1132),
    [anon_sym_POUND_LBRACK] = ACTIONS(1130),
    [aux_sym_encapsed_string_token1] = ACTIONS(1130),
    [anon_sym_DQUOTE] = ACTIONS(1130),
    [aux_sym_string_token1] = ACTIONS(1130),
    [anon_sym_SQUOTE] = ACTIONS(1130),
    [anon_sym_LT_LT_LT] = ACTIONS(1130),
    [anon_sym_BQUOTE] = ACTIONS(1130),
    [anon_sym_DOLLAR] = ACTIONS(1130),
    [aux_sym_yield_expression_token1] = ACTIONS(1132),
    [aux_sym_include_expression_token1] = ACTIONS(1132),
    [aux_sym_include_once_expression_token1] = ACTIONS(1132),
    [aux_sym_require_expression_token1] = ACTIONS(1132),
    [aux_sym_require_once_expression_token1] = ACTIONS(1132),
    [sym_comment] = ACTIONS(3),
  },
  [465] = {
    [ts_builtin_sym_end] = ACTIONS(1134),
    [sym_name] = ACTIONS(1136),
    [anon_sym_QMARK_GT] = ACTIONS(1134),
    [anon_sym_SEMI] = ACTIONS(1134),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1136),
    [aux_sym_global_declaration_token1] = ACTIONS(1136),
    [aux_sym_namespace_definition_token1] = ACTIONS(1136),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1136),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1136),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1136),
    [anon_sym_BSLASH] = ACTIONS(1134),
    [anon_sym_LBRACE] = ACTIONS(1134),
    [anon_sym_RBRACE] = ACTIONS(1134),
    [aux_sym_trait_declaration_token1] = ACTIONS(1136),
    [aux_sym_interface_declaration_token1] = ACTIONS(1136),
    [aux_sym_enum_declaration_token1] = ACTIONS(1136),
    [aux_sym_enum_case_token1] = ACTIONS(1136),
    [aux_sym_class_declaration_token1] = ACTIONS(1136),
    [aux_sym_final_modifier_token1] = ACTIONS(1136),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1136),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1136),
    [sym_var_modifier] = ACTIONS(1136),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1136),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1136),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1136),
    [anon_sym_LPAREN2] = ACTIONS(1134),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1136),
    [aux_sym_cast_type_token1] = ACTIONS(1136),
    [aux_sym_echo_statement_token1] = ACTIONS(1136),
    [aux_sym_exit_statement_token1] = ACTIONS(1136),
    [anon_sym_unset] = ACTIONS(1136),
    [aux_sym_declare_statement_token1] = ACTIONS(1136),
    [aux_sym_declare_statement_token2] = ACTIONS(1136),
    [sym_float] = ACTIONS(1136),
    [aux_sym_try_statement_token1] = ACTIONS(1136),
    [aux_sym_goto_statement_token1] = ACTIONS(1136),
    [aux_sym_continue_statement_token1] = ACTIONS(1136),
    [aux_sym_break_statement_token1] = ACTIONS(1136),
    [sym_integer] = ACTIONS(1136),
    [aux_sym_return_statement_token1] = ACTIONS(1136),
    [aux_sym_throw_expression_token1] = ACTIONS(1136),
    [aux_sym_while_statement_token1] = ACTIONS(1136),
    [aux_sym_while_statement_token2] = ACTIONS(1136),
    [aux_sym_do_statement_token1] = ACTIONS(1136),
    [aux_sym_for_statement_token1] = ACTIONS(1136),
    [aux_sym_for_statement_token2] = ACTIONS(1136),
    [aux_sym_foreach_statement_token1] = ACTIONS(1136),
    [aux_sym_foreach_statement_token2] = ACTIONS(1136),
    [aux_sym_if_statement_token1] = ACTIONS(1136),
    [aux_sym_if_statement_token2] = ACTIONS(1136),
    [aux_sym_else_if_clause_token1] = ACTIONS(1136),
    [aux_sym_else_clause_token1] = ACTIONS(1136),
    [aux_sym_match_expression_token1] = ACTIONS(1136),
    [aux_sym_match_default_expression_token1] = ACTIONS(1136),
    [aux_sym_switch_statement_token1] = ACTIONS(1136),
    [aux_sym_switch_block_token1] = ACTIONS(1136),
    [anon_sym_PLUS] = ACTIONS(1136),
    [anon_sym_DASH] = ACTIONS(1136),
    [anon_sym_TILDE] = ACTIONS(1134),
    [anon_sym_BANG] = ACTIONS(1134),
    [anon_sym_AT] = ACTIONS(1134),
    [aux_sym_clone_expression_token1] = ACTIONS(1136),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1136),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1136),
    [anon_sym_DASH_DASH] = ACTIONS(1134),
    [anon_sym_PLUS_PLUS] = ACTIONS(1134),
    [aux_sym__list_destructing_token1] = ACTIONS(1136),
    [anon_sym_LBRACK] = ACTIONS(1134),
    [anon_sym_self] = ACTIONS(1136),
    [anon_sym_parent] = ACTIONS(1136),
    [aux_sym__argument_name_token1] = ACTIONS(1136),
    [aux_sym__argument_name_token2] = ACTIONS(1136),
    [anon_sym_POUND_LBRACK] = ACTIONS(1134),
    [aux_sym_encapsed_string_token1] = ACTIONS(1134),
    [anon_sym_DQUOTE] = ACTIONS(1134),
    [aux_sym_string_token1] = ACTIONS(1134),
    [anon_sym_SQUOTE] = ACTIONS(1134),
    [anon_sym_LT_LT_LT] = ACTIONS(1134),
    [anon_sym_BQUOTE] = ACTIONS(1134),
    [anon_sym_DOLLAR] = ACTIONS(1134),
    [aux_sym_yield_expression_token1] = ACTIONS(1136),
    [aux_sym_include_expression_token1] = ACTIONS(1136),
    [aux_sym_include_once_expression_token1] = ACTIONS(1136),
    [aux_sym_require_expression_token1] = ACTIONS(1136),
    [aux_sym_require_once_expression_token1] = ACTIONS(1136),
    [sym_comment] = ACTIONS(3),
  },
  [466] = {
    [ts_builtin_sym_end] = ACTIONS(1138),
    [sym_name] = ACTIONS(1140),
    [anon_sym_QMARK_GT] = ACTIONS(1138),
    [anon_sym_SEMI] = ACTIONS(1138),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1140),
    [aux_sym_global_declaration_token1] = ACTIONS(1140),
    [aux_sym_namespace_definition_token1] = ACTIONS(1140),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1140),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1140),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1140),
    [anon_sym_BSLASH] = ACTIONS(1138),
    [anon_sym_LBRACE] = ACTIONS(1138),
    [anon_sym_RBRACE] = ACTIONS(1138),
    [aux_sym_trait_declaration_token1] = ACTIONS(1140),
    [aux_sym_interface_declaration_token1] = ACTIONS(1140),
    [aux_sym_enum_declaration_token1] = ACTIONS(1140),
    [aux_sym_enum_case_token1] = ACTIONS(1140),
    [aux_sym_class_declaration_token1] = ACTIONS(1140),
    [aux_sym_final_modifier_token1] = ACTIONS(1140),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1140),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1140),
    [sym_var_modifier] = ACTIONS(1140),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1140),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1140),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1140),
    [anon_sym_LPAREN2] = ACTIONS(1138),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1140),
    [aux_sym_cast_type_token1] = ACTIONS(1140),
    [aux_sym_echo_statement_token1] = ACTIONS(1140),
    [aux_sym_exit_statement_token1] = ACTIONS(1140),
    [anon_sym_unset] = ACTIONS(1140),
    [aux_sym_declare_statement_token1] = ACTIONS(1140),
    [aux_sym_declare_statement_token2] = ACTIONS(1140),
    [sym_float] = ACTIONS(1140),
    [aux_sym_try_statement_token1] = ACTIONS(1140),
    [aux_sym_goto_statement_token1] = ACTIONS(1140),
    [aux_sym_continue_statement_token1] = ACTIONS(1140),
    [aux_sym_break_statement_token1] = ACTIONS(1140),
    [sym_integer] = ACTIONS(1140),
    [aux_sym_return_statement_token1] = ACTIONS(1140),
    [aux_sym_throw_expression_token1] = ACTIONS(1140),
    [aux_sym_while_statement_token1] = ACTIONS(1140),
    [aux_sym_while_statement_token2] = ACTIONS(1140),
    [aux_sym_do_statement_token1] = ACTIONS(1140),
    [aux_sym_for_statement_token1] = ACTIONS(1140),
    [aux_sym_for_statement_token2] = ACTIONS(1140),
    [aux_sym_foreach_statement_token1] = ACTIONS(1140),
    [aux_sym_foreach_statement_token2] = ACTIONS(1140),
    [aux_sym_if_statement_token1] = ACTIONS(1140),
    [aux_sym_if_statement_token2] = ACTIONS(1140),
    [aux_sym_else_if_clause_token1] = ACTIONS(1140),
    [aux_sym_else_clause_token1] = ACTIONS(1140),
    [aux_sym_match_expression_token1] = ACTIONS(1140),
    [aux_sym_match_default_expression_token1] = ACTIONS(1140),
    [aux_sym_switch_statement_token1] = ACTIONS(1140),
    [aux_sym_switch_block_token1] = ACTIONS(1140),
    [anon_sym_PLUS] = ACTIONS(1140),
    [anon_sym_DASH] = ACTIONS(1140),
    [anon_sym_TILDE] = ACTIONS(1138),
    [anon_sym_BANG] = ACTIONS(1138),
    [anon_sym_AT] = ACTIONS(1138),
    [aux_sym_clone_expression_token1] = ACTIONS(1140),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1140),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1140),
    [anon_sym_DASH_DASH] = ACTIONS(1138),
    [anon_sym_PLUS_PLUS] = ACTIONS(1138),
    [aux_sym__list_destructing_token1] = ACTIONS(1140),
    [anon_sym_LBRACK] = ACTIONS(1138),
    [anon_sym_self] = ACTIONS(1140),
    [anon_sym_parent] = ACTIONS(1140),
    [aux_sym__argument_name_token1] = ACTIONS(1140),
    [aux_sym__argument_name_token2] = ACTIONS(1140),
    [anon_sym_POUND_LBRACK] = ACTIONS(1138),
    [aux_sym_encapsed_string_token1] = ACTIONS(1138),
    [anon_sym_DQUOTE] = ACTIONS(1138),
    [aux_sym_string_token1] = ACTIONS(1138),
    [anon_sym_SQUOTE] = ACTIONS(1138),
    [anon_sym_LT_LT_LT] = ACTIONS(1138),
    [anon_sym_BQUOTE] = ACTIONS(1138),
    [anon_sym_DOLLAR] = ACTIONS(1138),
    [aux_sym_yield_expression_token1] = ACTIONS(1140),
    [aux_sym_include_expression_token1] = ACTIONS(1140),
    [aux_sym_include_once_expression_token1] = ACTIONS(1140),
    [aux_sym_require_expression_token1] = ACTIONS(1140),
    [aux_sym_require_once_expression_token1] = ACTIONS(1140),
    [sym_comment] = ACTIONS(3),
  },
  [467] = {
    [ts_builtin_sym_end] = ACTIONS(1142),
    [sym_name] = ACTIONS(1144),
    [anon_sym_QMARK_GT] = ACTIONS(1142),
    [anon_sym_SEMI] = ACTIONS(1142),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1144),
    [aux_sym_global_declaration_token1] = ACTIONS(1144),
    [aux_sym_namespace_definition_token1] = ACTIONS(1144),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1144),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1144),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1144),
    [anon_sym_BSLASH] = ACTIONS(1142),
    [anon_sym_LBRACE] = ACTIONS(1142),
    [anon_sym_RBRACE] = ACTIONS(1142),
    [aux_sym_trait_declaration_token1] = ACTIONS(1144),
    [aux_sym_interface_declaration_token1] = ACTIONS(1144),
    [aux_sym_enum_declaration_token1] = ACTIONS(1144),
    [aux_sym_enum_case_token1] = ACTIONS(1144),
    [aux_sym_class_declaration_token1] = ACTIONS(1144),
    [aux_sym_final_modifier_token1] = ACTIONS(1144),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1144),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1144),
    [sym_var_modifier] = ACTIONS(1144),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1144),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1144),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1144),
    [anon_sym_LPAREN2] = ACTIONS(1142),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1144),
    [aux_sym_cast_type_token1] = ACTIONS(1144),
    [aux_sym_echo_statement_token1] = ACTIONS(1144),
    [aux_sym_exit_statement_token1] = ACTIONS(1144),
    [anon_sym_unset] = ACTIONS(1144),
    [aux_sym_declare_statement_token1] = ACTIONS(1144),
    [aux_sym_declare_statement_token2] = ACTIONS(1144),
    [sym_float] = ACTIONS(1144),
    [aux_sym_try_statement_token1] = ACTIONS(1144),
    [aux_sym_goto_statement_token1] = ACTIONS(1144),
    [aux_sym_continue_statement_token1] = ACTIONS(1144),
    [aux_sym_break_statement_token1] = ACTIONS(1144),
    [sym_integer] = ACTIONS(1144),
    [aux_sym_return_statement_token1] = ACTIONS(1144),
    [aux_sym_throw_expression_token1] = ACTIONS(1144),
    [aux_sym_while_statement_token1] = ACTIONS(1144),
    [aux_sym_while_statement_token2] = ACTIONS(1144),
    [aux_sym_do_statement_token1] = ACTIONS(1144),
    [aux_sym_for_statement_token1] = ACTIONS(1144),
    [aux_sym_for_statement_token2] = ACTIONS(1144),
    [aux_sym_foreach_statement_token1] = ACTIONS(1144),
    [aux_sym_foreach_statement_token2] = ACTIONS(1144),
    [aux_sym_if_statement_token1] = ACTIONS(1144),
    [aux_sym_if_statement_token2] = ACTIONS(1144),
    [aux_sym_else_if_clause_token1] = ACTIONS(1144),
    [aux_sym_else_clause_token1] = ACTIONS(1144),
    [aux_sym_match_expression_token1] = ACTIONS(1144),
    [aux_sym_match_default_expression_token1] = ACTIONS(1144),
    [aux_sym_switch_statement_token1] = ACTIONS(1144),
    [aux_sym_switch_block_token1] = ACTIONS(1144),
    [anon_sym_PLUS] = ACTIONS(1144),
    [anon_sym_DASH] = ACTIONS(1144),
    [anon_sym_TILDE] = ACTIONS(1142),
    [anon_sym_BANG] = ACTIONS(1142),
    [anon_sym_AT] = ACTIONS(1142),
    [aux_sym_clone_expression_token1] = ACTIONS(1144),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1144),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1144),
    [anon_sym_DASH_DASH] = ACTIONS(1142),
    [anon_sym_PLUS_PLUS] = ACTIONS(1142),
    [aux_sym__list_destructing_token1] = ACTIONS(1144),
    [anon_sym_LBRACK] = ACTIONS(1142),
    [anon_sym_self] = ACTIONS(1144),
    [anon_sym_parent] = ACTIONS(1144),
    [aux_sym__argument_name_token1] = ACTIONS(1144),
    [aux_sym__argument_name_token2] = ACTIONS(1144),
    [anon_sym_POUND_LBRACK] = ACTIONS(1142),
    [aux_sym_encapsed_string_token1] = ACTIONS(1142),
    [anon_sym_DQUOTE] = ACTIONS(1142),
    [aux_sym_string_token1] = ACTIONS(1142),
    [anon_sym_SQUOTE] = ACTIONS(1142),
    [anon_sym_LT_LT_LT] = ACTIONS(1142),
    [anon_sym_BQUOTE] = ACTIONS(1142),
    [anon_sym_DOLLAR] = ACTIONS(1142),
    [aux_sym_yield_expression_token1] = ACTIONS(1144),
    [aux_sym_include_expression_token1] = ACTIONS(1144),
    [aux_sym_include_once_expression_token1] = ACTIONS(1144),
    [aux_sym_require_expression_token1] = ACTIONS(1144),
    [aux_sym_require_once_expression_token1] = ACTIONS(1144),
    [sym_comment] = ACTIONS(3),
  },
  [468] = {
    [ts_builtin_sym_end] = ACTIONS(1146),
    [sym_name] = ACTIONS(1148),
    [anon_sym_QMARK_GT] = ACTIONS(1146),
    [anon_sym_SEMI] = ACTIONS(1146),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1148),
    [aux_sym_global_declaration_token1] = ACTIONS(1148),
    [aux_sym_namespace_definition_token1] = ACTIONS(1148),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1148),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1148),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1148),
    [anon_sym_BSLASH] = ACTIONS(1146),
    [anon_sym_LBRACE] = ACTIONS(1146),
    [anon_sym_RBRACE] = ACTIONS(1146),
    [aux_sym_trait_declaration_token1] = ACTIONS(1148),
    [aux_sym_interface_declaration_token1] = ACTIONS(1148),
    [aux_sym_enum_declaration_token1] = ACTIONS(1148),
    [aux_sym_enum_case_token1] = ACTIONS(1148),
    [aux_sym_class_declaration_token1] = ACTIONS(1148),
    [aux_sym_final_modifier_token1] = ACTIONS(1148),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1148),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1148),
    [sym_var_modifier] = ACTIONS(1148),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1148),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1148),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1148),
    [anon_sym_LPAREN2] = ACTIONS(1146),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1148),
    [aux_sym_cast_type_token1] = ACTIONS(1148),
    [aux_sym_echo_statement_token1] = ACTIONS(1148),
    [aux_sym_exit_statement_token1] = ACTIONS(1148),
    [anon_sym_unset] = ACTIONS(1148),
    [aux_sym_declare_statement_token1] = ACTIONS(1148),
    [aux_sym_declare_statement_token2] = ACTIONS(1148),
    [sym_float] = ACTIONS(1148),
    [aux_sym_try_statement_token1] = ACTIONS(1148),
    [aux_sym_goto_statement_token1] = ACTIONS(1148),
    [aux_sym_continue_statement_token1] = ACTIONS(1148),
    [aux_sym_break_statement_token1] = ACTIONS(1148),
    [sym_integer] = ACTIONS(1148),
    [aux_sym_return_statement_token1] = ACTIONS(1148),
    [aux_sym_throw_expression_token1] = ACTIONS(1148),
    [aux_sym_while_statement_token1] = ACTIONS(1148),
    [aux_sym_while_statement_token2] = ACTIONS(1148),
    [aux_sym_do_statement_token1] = ACTIONS(1148),
    [aux_sym_for_statement_token1] = ACTIONS(1148),
    [aux_sym_for_statement_token2] = ACTIONS(1148),
    [aux_sym_foreach_statement_token1] = ACTIONS(1148),
    [aux_sym_foreach_statement_token2] = ACTIONS(1148),
    [aux_sym_if_statement_token1] = ACTIONS(1148),
    [aux_sym_if_statement_token2] = ACTIONS(1148),
    [aux_sym_else_if_clause_token1] = ACTIONS(1148),
    [aux_sym_else_clause_token1] = ACTIONS(1148),
    [aux_sym_match_expression_token1] = ACTIONS(1148),
    [aux_sym_match_default_expression_token1] = ACTIONS(1148),
    [aux_sym_switch_statement_token1] = ACTIONS(1148),
    [aux_sym_switch_block_token1] = ACTIONS(1148),
    [anon_sym_PLUS] = ACTIONS(1148),
    [anon_sym_DASH] = ACTIONS(1148),
    [anon_sym_TILDE] = ACTIONS(1146),
    [anon_sym_BANG] = ACTIONS(1146),
    [anon_sym_AT] = ACTIONS(1146),
    [aux_sym_clone_expression_token1] = ACTIONS(1148),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1148),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1148),
    [anon_sym_DASH_DASH] = ACTIONS(1146),
    [anon_sym_PLUS_PLUS] = ACTIONS(1146),
    [aux_sym__list_destructing_token1] = ACTIONS(1148),
    [anon_sym_LBRACK] = ACTIONS(1146),
    [anon_sym_self] = ACTIONS(1148),
    [anon_sym_parent] = ACTIONS(1148),
    [aux_sym__argument_name_token1] = ACTIONS(1148),
    [aux_sym__argument_name_token2] = ACTIONS(1148),
    [anon_sym_POUND_LBRACK] = ACTIONS(1146),
    [aux_sym_encapsed_string_token1] = ACTIONS(1146),
    [anon_sym_DQUOTE] = ACTIONS(1146),
    [aux_sym_string_token1] = ACTIONS(1146),
    [anon_sym_SQUOTE] = ACTIONS(1146),
    [anon_sym_LT_LT_LT] = ACTIONS(1146),
    [anon_sym_BQUOTE] = ACTIONS(1146),
    [anon_sym_DOLLAR] = ACTIONS(1146),
    [aux_sym_yield_expression_token1] = ACTIONS(1148),
    [aux_sym_include_expression_token1] = ACTIONS(1148),
    [aux_sym_include_once_expression_token1] = ACTIONS(1148),
    [aux_sym_require_expression_token1] = ACTIONS(1148),
    [aux_sym_require_once_expression_token1] = ACTIONS(1148),
    [sym_comment] = ACTIONS(3),
  },
  [469] = {
    [ts_builtin_sym_end] = ACTIONS(1150),
    [sym_name] = ACTIONS(1152),
    [anon_sym_QMARK_GT] = ACTIONS(1150),
    [anon_sym_SEMI] = ACTIONS(1150),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1152),
    [aux_sym_global_declaration_token1] = ACTIONS(1152),
    [aux_sym_namespace_definition_token1] = ACTIONS(1152),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1152),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1152),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1152),
    [anon_sym_BSLASH] = ACTIONS(1150),
    [anon_sym_LBRACE] = ACTIONS(1150),
    [anon_sym_RBRACE] = ACTIONS(1150),
    [aux_sym_trait_declaration_token1] = ACTIONS(1152),
    [aux_sym_interface_declaration_token1] = ACTIONS(1152),
    [aux_sym_enum_declaration_token1] = ACTIONS(1152),
    [aux_sym_enum_case_token1] = ACTIONS(1152),
    [aux_sym_class_declaration_token1] = ACTIONS(1152),
    [aux_sym_final_modifier_token1] = ACTIONS(1152),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1152),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1152),
    [sym_var_modifier] = ACTIONS(1152),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1152),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1152),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1152),
    [anon_sym_LPAREN2] = ACTIONS(1150),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1152),
    [aux_sym_cast_type_token1] = ACTIONS(1152),
    [aux_sym_echo_statement_token1] = ACTIONS(1152),
    [aux_sym_exit_statement_token1] = ACTIONS(1152),
    [anon_sym_unset] = ACTIONS(1152),
    [aux_sym_declare_statement_token1] = ACTIONS(1152),
    [aux_sym_declare_statement_token2] = ACTIONS(1152),
    [sym_float] = ACTIONS(1152),
    [aux_sym_try_statement_token1] = ACTIONS(1152),
    [aux_sym_goto_statement_token1] = ACTIONS(1152),
    [aux_sym_continue_statement_token1] = ACTIONS(1152),
    [aux_sym_break_statement_token1] = ACTIONS(1152),
    [sym_integer] = ACTIONS(1152),
    [aux_sym_return_statement_token1] = ACTIONS(1152),
    [aux_sym_throw_expression_token1] = ACTIONS(1152),
    [aux_sym_while_statement_token1] = ACTIONS(1152),
    [aux_sym_while_statement_token2] = ACTIONS(1152),
    [aux_sym_do_statement_token1] = ACTIONS(1152),
    [aux_sym_for_statement_token1] = ACTIONS(1152),
    [aux_sym_for_statement_token2] = ACTIONS(1152),
    [aux_sym_foreach_statement_token1] = ACTIONS(1152),
    [aux_sym_foreach_statement_token2] = ACTIONS(1152),
    [aux_sym_if_statement_token1] = ACTIONS(1152),
    [aux_sym_if_statement_token2] = ACTIONS(1152),
    [aux_sym_else_if_clause_token1] = ACTIONS(1152),
    [aux_sym_else_clause_token1] = ACTIONS(1152),
    [aux_sym_match_expression_token1] = ACTIONS(1152),
    [aux_sym_match_default_expression_token1] = ACTIONS(1152),
    [aux_sym_switch_statement_token1] = ACTIONS(1152),
    [aux_sym_switch_block_token1] = ACTIONS(1152),
    [anon_sym_PLUS] = ACTIONS(1152),
    [anon_sym_DASH] = ACTIONS(1152),
    [anon_sym_TILDE] = ACTIONS(1150),
    [anon_sym_BANG] = ACTIONS(1150),
    [anon_sym_AT] = ACTIONS(1150),
    [aux_sym_clone_expression_token1] = ACTIONS(1152),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1152),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1152),
    [anon_sym_DASH_DASH] = ACTIONS(1150),
    [anon_sym_PLUS_PLUS] = ACTIONS(1150),
    [aux_sym__list_destructing_token1] = ACTIONS(1152),
    [anon_sym_LBRACK] = ACTIONS(1150),
    [anon_sym_self] = ACTIONS(1152),
    [anon_sym_parent] = ACTIONS(1152),
    [aux_sym__argument_name_token1] = ACTIONS(1152),
    [aux_sym__argument_name_token2] = ACTIONS(1152),
    [anon_sym_POUND_LBRACK] = ACTIONS(1150),
    [aux_sym_encapsed_string_token1] = ACTIONS(1150),
    [anon_sym_DQUOTE] = ACTIONS(1150),
    [aux_sym_string_token1] = ACTIONS(1150),
    [anon_sym_SQUOTE] = ACTIONS(1150),
    [anon_sym_LT_LT_LT] = ACTIONS(1150),
    [anon_sym_BQUOTE] = ACTIONS(1150),
    [anon_sym_DOLLAR] = ACTIONS(1150),
    [aux_sym_yield_expression_token1] = ACTIONS(1152),
    [aux_sym_include_expression_token1] = ACTIONS(1152),
    [aux_sym_include_once_expression_token1] = ACTIONS(1152),
    [aux_sym_require_expression_token1] = ACTIONS(1152),
    [aux_sym_require_once_expression_token1] = ACTIONS(1152),
    [sym_comment] = ACTIONS(3),
  },
  [470] = {
    [ts_builtin_sym_end] = ACTIONS(1154),
    [sym_name] = ACTIONS(1156),
    [anon_sym_QMARK_GT] = ACTIONS(1154),
    [anon_sym_SEMI] = ACTIONS(1154),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1156),
    [aux_sym_global_declaration_token1] = ACTIONS(1156),
    [aux_sym_namespace_definition_token1] = ACTIONS(1156),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1156),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1156),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1156),
    [anon_sym_BSLASH] = ACTIONS(1154),
    [anon_sym_LBRACE] = ACTIONS(1154),
    [anon_sym_RBRACE] = ACTIONS(1154),
    [aux_sym_trait_declaration_token1] = ACTIONS(1156),
    [aux_sym_interface_declaration_token1] = ACTIONS(1156),
    [aux_sym_enum_declaration_token1] = ACTIONS(1156),
    [aux_sym_enum_case_token1] = ACTIONS(1156),
    [aux_sym_class_declaration_token1] = ACTIONS(1156),
    [aux_sym_final_modifier_token1] = ACTIONS(1156),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1156),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1156),
    [sym_var_modifier] = ACTIONS(1156),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1156),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1156),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1156),
    [anon_sym_LPAREN2] = ACTIONS(1154),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1156),
    [aux_sym_cast_type_token1] = ACTIONS(1156),
    [aux_sym_echo_statement_token1] = ACTIONS(1156),
    [aux_sym_exit_statement_token1] = ACTIONS(1156),
    [anon_sym_unset] = ACTIONS(1156),
    [aux_sym_declare_statement_token1] = ACTIONS(1156),
    [aux_sym_declare_statement_token2] = ACTIONS(1156),
    [sym_float] = ACTIONS(1156),
    [aux_sym_try_statement_token1] = ACTIONS(1156),
    [aux_sym_goto_statement_token1] = ACTIONS(1156),
    [aux_sym_continue_statement_token1] = ACTIONS(1156),
    [aux_sym_break_statement_token1] = ACTIONS(1156),
    [sym_integer] = ACTIONS(1156),
    [aux_sym_return_statement_token1] = ACTIONS(1156),
    [aux_sym_throw_expression_token1] = ACTIONS(1156),
    [aux_sym_while_statement_token1] = ACTIONS(1156),
    [aux_sym_while_statement_token2] = ACTIONS(1156),
    [aux_sym_do_statement_token1] = ACTIONS(1156),
    [aux_sym_for_statement_token1] = ACTIONS(1156),
    [aux_sym_for_statement_token2] = ACTIONS(1156),
    [aux_sym_foreach_statement_token1] = ACTIONS(1156),
    [aux_sym_foreach_statement_token2] = ACTIONS(1156),
    [aux_sym_if_statement_token1] = ACTIONS(1156),
    [aux_sym_if_statement_token2] = ACTIONS(1156),
    [aux_sym_else_if_clause_token1] = ACTIONS(1156),
    [aux_sym_else_clause_token1] = ACTIONS(1156),
    [aux_sym_match_expression_token1] = ACTIONS(1156),
    [aux_sym_match_default_expression_token1] = ACTIONS(1156),
    [aux_sym_switch_statement_token1] = ACTIONS(1156),
    [aux_sym_switch_block_token1] = ACTIONS(1156),
    [anon_sym_PLUS] = ACTIONS(1156),
    [anon_sym_DASH] = ACTIONS(1156),
    [anon_sym_TILDE] = ACTIONS(1154),
    [anon_sym_BANG] = ACTIONS(1154),
    [anon_sym_AT] = ACTIONS(1154),
    [aux_sym_clone_expression_token1] = ACTIONS(1156),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1156),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1156),
    [anon_sym_DASH_DASH] = ACTIONS(1154),
    [anon_sym_PLUS_PLUS] = ACTIONS(1154),
    [aux_sym__list_destructing_token1] = ACTIONS(1156),
    [anon_sym_LBRACK] = ACTIONS(1154),
    [anon_sym_self] = ACTIONS(1156),
    [anon_sym_parent] = ACTIONS(1156),
    [aux_sym__argument_name_token1] = ACTIONS(1156),
    [aux_sym__argument_name_token2] = ACTIONS(1156),
    [anon_sym_POUND_LBRACK] = ACTIONS(1154),
    [aux_sym_encapsed_string_token1] = ACTIONS(1154),
    [anon_sym_DQUOTE] = ACTIONS(1154),
    [aux_sym_string_token1] = ACTIONS(1154),
    [anon_sym_SQUOTE] = ACTIONS(1154),
    [anon_sym_LT_LT_LT] = ACTIONS(1154),
    [anon_sym_BQUOTE] = ACTIONS(1154),
    [anon_sym_DOLLAR] = ACTIONS(1154),
    [aux_sym_yield_expression_token1] = ACTIONS(1156),
    [aux_sym_include_expression_token1] = ACTIONS(1156),
    [aux_sym_include_once_expression_token1] = ACTIONS(1156),
    [aux_sym_require_expression_token1] = ACTIONS(1156),
    [aux_sym_require_once_expression_token1] = ACTIONS(1156),
    [sym_comment] = ACTIONS(3),
  },
  [471] = {
    [ts_builtin_sym_end] = ACTIONS(1158),
    [sym_name] = ACTIONS(1160),
    [anon_sym_QMARK_GT] = ACTIONS(1158),
    [anon_sym_SEMI] = ACTIONS(1158),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1160),
    [aux_sym_global_declaration_token1] = ACTIONS(1160),
    [aux_sym_namespace_definition_token1] = ACTIONS(1160),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1160),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1160),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1160),
    [anon_sym_BSLASH] = ACTIONS(1158),
    [anon_sym_LBRACE] = ACTIONS(1158),
    [anon_sym_RBRACE] = ACTIONS(1158),
    [aux_sym_trait_declaration_token1] = ACTIONS(1160),
    [aux_sym_interface_declaration_token1] = ACTIONS(1160),
    [aux_sym_enum_declaration_token1] = ACTIONS(1160),
    [aux_sym_enum_case_token1] = ACTIONS(1160),
    [aux_sym_class_declaration_token1] = ACTIONS(1160),
    [aux_sym_final_modifier_token1] = ACTIONS(1160),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1160),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1160),
    [sym_var_modifier] = ACTIONS(1160),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1160),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1160),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1160),
    [anon_sym_LPAREN2] = ACTIONS(1158),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1160),
    [aux_sym_cast_type_token1] = ACTIONS(1160),
    [aux_sym_echo_statement_token1] = ACTIONS(1160),
    [aux_sym_exit_statement_token1] = ACTIONS(1160),
    [anon_sym_unset] = ACTIONS(1160),
    [aux_sym_declare_statement_token1] = ACTIONS(1160),
    [aux_sym_declare_statement_token2] = ACTIONS(1160),
    [sym_float] = ACTIONS(1160),
    [aux_sym_try_statement_token1] = ACTIONS(1160),
    [aux_sym_goto_statement_token1] = ACTIONS(1160),
    [aux_sym_continue_statement_token1] = ACTIONS(1160),
    [aux_sym_break_statement_token1] = ACTIONS(1160),
    [sym_integer] = ACTIONS(1160),
    [aux_sym_return_statement_token1] = ACTIONS(1160),
    [aux_sym_throw_expression_token1] = ACTIONS(1160),
    [aux_sym_while_statement_token1] = ACTIONS(1160),
    [aux_sym_while_statement_token2] = ACTIONS(1160),
    [aux_sym_do_statement_token1] = ACTIONS(1160),
    [aux_sym_for_statement_token1] = ACTIONS(1160),
    [aux_sym_for_statement_token2] = ACTIONS(1160),
    [aux_sym_foreach_statement_token1] = ACTIONS(1160),
    [aux_sym_foreach_statement_token2] = ACTIONS(1160),
    [aux_sym_if_statement_token1] = ACTIONS(1160),
    [aux_sym_if_statement_token2] = ACTIONS(1160),
    [aux_sym_else_if_clause_token1] = ACTIONS(1160),
    [aux_sym_else_clause_token1] = ACTIONS(1160),
    [aux_sym_match_expression_token1] = ACTIONS(1160),
    [aux_sym_match_default_expression_token1] = ACTIONS(1160),
    [aux_sym_switch_statement_token1] = ACTIONS(1160),
    [aux_sym_switch_block_token1] = ACTIONS(1160),
    [anon_sym_PLUS] = ACTIONS(1160),
    [anon_sym_DASH] = ACTIONS(1160),
    [anon_sym_TILDE] = ACTIONS(1158),
    [anon_sym_BANG] = ACTIONS(1158),
    [anon_sym_AT] = ACTIONS(1158),
    [aux_sym_clone_expression_token1] = ACTIONS(1160),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1160),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1160),
    [anon_sym_DASH_DASH] = ACTIONS(1158),
    [anon_sym_PLUS_PLUS] = ACTIONS(1158),
    [aux_sym__list_destructing_token1] = ACTIONS(1160),
    [anon_sym_LBRACK] = ACTIONS(1158),
    [anon_sym_self] = ACTIONS(1160),
    [anon_sym_parent] = ACTIONS(1160),
    [aux_sym__argument_name_token1] = ACTIONS(1160),
    [aux_sym__argument_name_token2] = ACTIONS(1160),
    [anon_sym_POUND_LBRACK] = ACTIONS(1158),
    [aux_sym_encapsed_string_token1] = ACTIONS(1158),
    [anon_sym_DQUOTE] = ACTIONS(1158),
    [aux_sym_string_token1] = ACTIONS(1158),
    [anon_sym_SQUOTE] = ACTIONS(1158),
    [anon_sym_LT_LT_LT] = ACTIONS(1158),
    [anon_sym_BQUOTE] = ACTIONS(1158),
    [anon_sym_DOLLAR] = ACTIONS(1158),
    [aux_sym_yield_expression_token1] = ACTIONS(1160),
    [aux_sym_include_expression_token1] = ACTIONS(1160),
    [aux_sym_include_once_expression_token1] = ACTIONS(1160),
    [aux_sym_require_expression_token1] = ACTIONS(1160),
    [aux_sym_require_once_expression_token1] = ACTIONS(1160),
    [sym_comment] = ACTIONS(3),
  },
  [472] = {
    [ts_builtin_sym_end] = ACTIONS(1162),
    [sym_name] = ACTIONS(1164),
    [anon_sym_QMARK_GT] = ACTIONS(1162),
    [anon_sym_SEMI] = ACTIONS(1162),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1164),
    [aux_sym_global_declaration_token1] = ACTIONS(1164),
    [aux_sym_namespace_definition_token1] = ACTIONS(1164),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1164),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1164),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1164),
    [anon_sym_BSLASH] = ACTIONS(1162),
    [anon_sym_LBRACE] = ACTIONS(1162),
    [anon_sym_RBRACE] = ACTIONS(1162),
    [aux_sym_trait_declaration_token1] = ACTIONS(1164),
    [aux_sym_interface_declaration_token1] = ACTIONS(1164),
    [aux_sym_enum_declaration_token1] = ACTIONS(1164),
    [aux_sym_enum_case_token1] = ACTIONS(1164),
    [aux_sym_class_declaration_token1] = ACTIONS(1164),
    [aux_sym_final_modifier_token1] = ACTIONS(1164),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1164),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1164),
    [sym_var_modifier] = ACTIONS(1164),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1164),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1164),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1164),
    [anon_sym_LPAREN2] = ACTIONS(1162),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1164),
    [aux_sym_cast_type_token1] = ACTIONS(1164),
    [aux_sym_echo_statement_token1] = ACTIONS(1164),
    [aux_sym_exit_statement_token1] = ACTIONS(1164),
    [anon_sym_unset] = ACTIONS(1164),
    [aux_sym_declare_statement_token1] = ACTIONS(1164),
    [aux_sym_declare_statement_token2] = ACTIONS(1164),
    [sym_float] = ACTIONS(1164),
    [aux_sym_try_statement_token1] = ACTIONS(1164),
    [aux_sym_goto_statement_token1] = ACTIONS(1164),
    [aux_sym_continue_statement_token1] = ACTIONS(1164),
    [aux_sym_break_statement_token1] = ACTIONS(1164),
    [sym_integer] = ACTIONS(1164),
    [aux_sym_return_statement_token1] = ACTIONS(1164),
    [aux_sym_throw_expression_token1] = ACTIONS(1164),
    [aux_sym_while_statement_token1] = ACTIONS(1164),
    [aux_sym_while_statement_token2] = ACTIONS(1164),
    [aux_sym_do_statement_token1] = ACTIONS(1164),
    [aux_sym_for_statement_token1] = ACTIONS(1164),
    [aux_sym_for_statement_token2] = ACTIONS(1164),
    [aux_sym_foreach_statement_token1] = ACTIONS(1164),
    [aux_sym_foreach_statement_token2] = ACTIONS(1164),
    [aux_sym_if_statement_token1] = ACTIONS(1164),
    [aux_sym_if_statement_token2] = ACTIONS(1164),
    [aux_sym_else_if_clause_token1] = ACTIONS(1164),
    [aux_sym_else_clause_token1] = ACTIONS(1164),
    [aux_sym_match_expression_token1] = ACTIONS(1164),
    [aux_sym_match_default_expression_token1] = ACTIONS(1164),
    [aux_sym_switch_statement_token1] = ACTIONS(1164),
    [aux_sym_switch_block_token1] = ACTIONS(1164),
    [anon_sym_PLUS] = ACTIONS(1164),
    [anon_sym_DASH] = ACTIONS(1164),
    [anon_sym_TILDE] = ACTIONS(1162),
    [anon_sym_BANG] = ACTIONS(1162),
    [anon_sym_AT] = ACTIONS(1162),
    [aux_sym_clone_expression_token1] = ACTIONS(1164),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1164),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1164),
    [anon_sym_DASH_DASH] = ACTIONS(1162),
    [anon_sym_PLUS_PLUS] = ACTIONS(1162),
    [aux_sym__list_destructing_token1] = ACTIONS(1164),
    [anon_sym_LBRACK] = ACTIONS(1162),
    [anon_sym_self] = ACTIONS(1164),
    [anon_sym_parent] = ACTIONS(1164),
    [aux_sym__argument_name_token1] = ACTIONS(1164),
    [aux_sym__argument_name_token2] = ACTIONS(1164),
    [anon_sym_POUND_LBRACK] = ACTIONS(1162),
    [aux_sym_encapsed_string_token1] = ACTIONS(1162),
    [anon_sym_DQUOTE] = ACTIONS(1162),
    [aux_sym_string_token1] = ACTIONS(1162),
    [anon_sym_SQUOTE] = ACTIONS(1162),
    [anon_sym_LT_LT_LT] = ACTIONS(1162),
    [anon_sym_BQUOTE] = ACTIONS(1162),
    [anon_sym_DOLLAR] = ACTIONS(1162),
    [aux_sym_yield_expression_token1] = ACTIONS(1164),
    [aux_sym_include_expression_token1] = ACTIONS(1164),
    [aux_sym_include_once_expression_token1] = ACTIONS(1164),
    [aux_sym_require_expression_token1] = ACTIONS(1164),
    [aux_sym_require_once_expression_token1] = ACTIONS(1164),
    [sym_comment] = ACTIONS(3),
  },
  [473] = {
    [ts_builtin_sym_end] = ACTIONS(1166),
    [sym_name] = ACTIONS(1168),
    [anon_sym_QMARK_GT] = ACTIONS(1166),
    [anon_sym_SEMI] = ACTIONS(1166),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1168),
    [aux_sym_global_declaration_token1] = ACTIONS(1168),
    [aux_sym_namespace_definition_token1] = ACTIONS(1168),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1168),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1168),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1168),
    [anon_sym_BSLASH] = ACTIONS(1166),
    [anon_sym_LBRACE] = ACTIONS(1166),
    [anon_sym_RBRACE] = ACTIONS(1166),
    [aux_sym_trait_declaration_token1] = ACTIONS(1168),
    [aux_sym_interface_declaration_token1] = ACTIONS(1168),
    [aux_sym_enum_declaration_token1] = ACTIONS(1168),
    [aux_sym_enum_case_token1] = ACTIONS(1168),
    [aux_sym_class_declaration_token1] = ACTIONS(1168),
    [aux_sym_final_modifier_token1] = ACTIONS(1168),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1168),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1168),
    [sym_var_modifier] = ACTIONS(1168),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1168),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1168),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1168),
    [anon_sym_LPAREN2] = ACTIONS(1166),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1168),
    [aux_sym_cast_type_token1] = ACTIONS(1168),
    [aux_sym_echo_statement_token1] = ACTIONS(1168),
    [aux_sym_exit_statement_token1] = ACTIONS(1168),
    [anon_sym_unset] = ACTIONS(1168),
    [aux_sym_declare_statement_token1] = ACTIONS(1168),
    [aux_sym_declare_statement_token2] = ACTIONS(1168),
    [sym_float] = ACTIONS(1168),
    [aux_sym_try_statement_token1] = ACTIONS(1168),
    [aux_sym_goto_statement_token1] = ACTIONS(1168),
    [aux_sym_continue_statement_token1] = ACTIONS(1168),
    [aux_sym_break_statement_token1] = ACTIONS(1168),
    [sym_integer] = ACTIONS(1168),
    [aux_sym_return_statement_token1] = ACTIONS(1168),
    [aux_sym_throw_expression_token1] = ACTIONS(1168),
    [aux_sym_while_statement_token1] = ACTIONS(1168),
    [aux_sym_while_statement_token2] = ACTIONS(1168),
    [aux_sym_do_statement_token1] = ACTIONS(1168),
    [aux_sym_for_statement_token1] = ACTIONS(1168),
    [aux_sym_for_statement_token2] = ACTIONS(1168),
    [aux_sym_foreach_statement_token1] = ACTIONS(1168),
    [aux_sym_foreach_statement_token2] = ACTIONS(1168),
    [aux_sym_if_statement_token1] = ACTIONS(1168),
    [aux_sym_if_statement_token2] = ACTIONS(1168),
    [aux_sym_else_if_clause_token1] = ACTIONS(1168),
    [aux_sym_else_clause_token1] = ACTIONS(1168),
    [aux_sym_match_expression_token1] = ACTIONS(1168),
    [aux_sym_match_default_expression_token1] = ACTIONS(1168),
    [aux_sym_switch_statement_token1] = ACTIONS(1168),
    [aux_sym_switch_block_token1] = ACTIONS(1168),
    [anon_sym_PLUS] = ACTIONS(1168),
    [anon_sym_DASH] = ACTIONS(1168),
    [anon_sym_TILDE] = ACTIONS(1166),
    [anon_sym_BANG] = ACTIONS(1166),
    [anon_sym_AT] = ACTIONS(1166),
    [aux_sym_clone_expression_token1] = ACTIONS(1168),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1168),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1168),
    [anon_sym_DASH_DASH] = ACTIONS(1166),
    [anon_sym_PLUS_PLUS] = ACTIONS(1166),
    [aux_sym__list_destructing_token1] = ACTIONS(1168),
    [anon_sym_LBRACK] = ACTIONS(1166),
    [anon_sym_self] = ACTIONS(1168),
    [anon_sym_parent] = ACTIONS(1168),
    [aux_sym__argument_name_token1] = ACTIONS(1168),
    [aux_sym__argument_name_token2] = ACTIONS(1168),
    [anon_sym_POUND_LBRACK] = ACTIONS(1166),
    [aux_sym_encapsed_string_token1] = ACTIONS(1166),
    [anon_sym_DQUOTE] = ACTIONS(1166),
    [aux_sym_string_token1] = ACTIONS(1166),
    [anon_sym_SQUOTE] = ACTIONS(1166),
    [anon_sym_LT_LT_LT] = ACTIONS(1166),
    [anon_sym_BQUOTE] = ACTIONS(1166),
    [anon_sym_DOLLAR] = ACTIONS(1166),
    [aux_sym_yield_expression_token1] = ACTIONS(1168),
    [aux_sym_include_expression_token1] = ACTIONS(1168),
    [aux_sym_include_once_expression_token1] = ACTIONS(1168),
    [aux_sym_require_expression_token1] = ACTIONS(1168),
    [aux_sym_require_once_expression_token1] = ACTIONS(1168),
    [sym_comment] = ACTIONS(3),
  },
  [474] = {
    [ts_builtin_sym_end] = ACTIONS(1170),
    [sym_name] = ACTIONS(1172),
    [anon_sym_QMARK_GT] = ACTIONS(1170),
    [anon_sym_SEMI] = ACTIONS(1170),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1172),
    [aux_sym_global_declaration_token1] = ACTIONS(1172),
    [aux_sym_namespace_definition_token1] = ACTIONS(1172),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1172),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1172),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1172),
    [anon_sym_BSLASH] = ACTIONS(1170),
    [anon_sym_LBRACE] = ACTIONS(1170),
    [anon_sym_RBRACE] = ACTIONS(1170),
    [aux_sym_trait_declaration_token1] = ACTIONS(1172),
    [aux_sym_interface_declaration_token1] = ACTIONS(1172),
    [aux_sym_enum_declaration_token1] = ACTIONS(1172),
    [aux_sym_enum_case_token1] = ACTIONS(1172),
    [aux_sym_class_declaration_token1] = ACTIONS(1172),
    [aux_sym_final_modifier_token1] = ACTIONS(1172),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1172),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1172),
    [sym_var_modifier] = ACTIONS(1172),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1172),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1172),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1172),
    [anon_sym_LPAREN2] = ACTIONS(1170),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1172),
    [aux_sym_cast_type_token1] = ACTIONS(1172),
    [aux_sym_echo_statement_token1] = ACTIONS(1172),
    [aux_sym_exit_statement_token1] = ACTIONS(1172),
    [anon_sym_unset] = ACTIONS(1172),
    [aux_sym_declare_statement_token1] = ACTIONS(1172),
    [aux_sym_declare_statement_token2] = ACTIONS(1172),
    [sym_float] = ACTIONS(1172),
    [aux_sym_try_statement_token1] = ACTIONS(1172),
    [aux_sym_goto_statement_token1] = ACTIONS(1172),
    [aux_sym_continue_statement_token1] = ACTIONS(1172),
    [aux_sym_break_statement_token1] = ACTIONS(1172),
    [sym_integer] = ACTIONS(1172),
    [aux_sym_return_statement_token1] = ACTIONS(1172),
    [aux_sym_throw_expression_token1] = ACTIONS(1172),
    [aux_sym_while_statement_token1] = ACTIONS(1172),
    [aux_sym_while_statement_token2] = ACTIONS(1172),
    [aux_sym_do_statement_token1] = ACTIONS(1172),
    [aux_sym_for_statement_token1] = ACTIONS(1172),
    [aux_sym_for_statement_token2] = ACTIONS(1172),
    [aux_sym_foreach_statement_token1] = ACTIONS(1172),
    [aux_sym_foreach_statement_token2] = ACTIONS(1172),
    [aux_sym_if_statement_token1] = ACTIONS(1172),
    [aux_sym_if_statement_token2] = ACTIONS(1172),
    [aux_sym_else_if_clause_token1] = ACTIONS(1172),
    [aux_sym_else_clause_token1] = ACTIONS(1172),
    [aux_sym_match_expression_token1] = ACTIONS(1172),
    [aux_sym_match_default_expression_token1] = ACTIONS(1172),
    [aux_sym_switch_statement_token1] = ACTIONS(1172),
    [aux_sym_switch_block_token1] = ACTIONS(1172),
    [anon_sym_PLUS] = ACTIONS(1172),
    [anon_sym_DASH] = ACTIONS(1172),
    [anon_sym_TILDE] = ACTIONS(1170),
    [anon_sym_BANG] = ACTIONS(1170),
    [anon_sym_AT] = ACTIONS(1170),
    [aux_sym_clone_expression_token1] = ACTIONS(1172),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1172),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1172),
    [anon_sym_DASH_DASH] = ACTIONS(1170),
    [anon_sym_PLUS_PLUS] = ACTIONS(1170),
    [aux_sym__list_destructing_token1] = ACTIONS(1172),
    [anon_sym_LBRACK] = ACTIONS(1170),
    [anon_sym_self] = ACTIONS(1172),
    [anon_sym_parent] = ACTIONS(1172),
    [aux_sym__argument_name_token1] = ACTIONS(1172),
    [aux_sym__argument_name_token2] = ACTIONS(1172),
    [anon_sym_POUND_LBRACK] = ACTIONS(1170),
    [aux_sym_encapsed_string_token1] = ACTIONS(1170),
    [anon_sym_DQUOTE] = ACTIONS(1170),
    [aux_sym_string_token1] = ACTIONS(1170),
    [anon_sym_SQUOTE] = ACTIONS(1170),
    [anon_sym_LT_LT_LT] = ACTIONS(1170),
    [anon_sym_BQUOTE] = ACTIONS(1170),
    [anon_sym_DOLLAR] = ACTIONS(1170),
    [aux_sym_yield_expression_token1] = ACTIONS(1172),
    [aux_sym_include_expression_token1] = ACTIONS(1172),
    [aux_sym_include_once_expression_token1] = ACTIONS(1172),
    [aux_sym_require_expression_token1] = ACTIONS(1172),
    [aux_sym_require_once_expression_token1] = ACTIONS(1172),
    [sym_comment] = ACTIONS(3),
  },
  [475] = {
    [ts_builtin_sym_end] = ACTIONS(1174),
    [sym_name] = ACTIONS(1176),
    [anon_sym_QMARK_GT] = ACTIONS(1174),
    [anon_sym_SEMI] = ACTIONS(1174),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1176),
    [aux_sym_global_declaration_token1] = ACTIONS(1176),
    [aux_sym_namespace_definition_token1] = ACTIONS(1176),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1176),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1176),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1176),
    [anon_sym_BSLASH] = ACTIONS(1174),
    [anon_sym_LBRACE] = ACTIONS(1174),
    [anon_sym_RBRACE] = ACTIONS(1174),
    [aux_sym_trait_declaration_token1] = ACTIONS(1176),
    [aux_sym_interface_declaration_token1] = ACTIONS(1176),
    [aux_sym_enum_declaration_token1] = ACTIONS(1176),
    [aux_sym_enum_case_token1] = ACTIONS(1176),
    [aux_sym_class_declaration_token1] = ACTIONS(1176),
    [aux_sym_final_modifier_token1] = ACTIONS(1176),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1176),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1176),
    [sym_var_modifier] = ACTIONS(1176),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1176),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1176),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1176),
    [anon_sym_LPAREN2] = ACTIONS(1174),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1176),
    [aux_sym_cast_type_token1] = ACTIONS(1176),
    [aux_sym_echo_statement_token1] = ACTIONS(1176),
    [aux_sym_exit_statement_token1] = ACTIONS(1176),
    [anon_sym_unset] = ACTIONS(1176),
    [aux_sym_declare_statement_token1] = ACTIONS(1176),
    [aux_sym_declare_statement_token2] = ACTIONS(1176),
    [sym_float] = ACTIONS(1176),
    [aux_sym_try_statement_token1] = ACTIONS(1176),
    [aux_sym_goto_statement_token1] = ACTIONS(1176),
    [aux_sym_continue_statement_token1] = ACTIONS(1176),
    [aux_sym_break_statement_token1] = ACTIONS(1176),
    [sym_integer] = ACTIONS(1176),
    [aux_sym_return_statement_token1] = ACTIONS(1176),
    [aux_sym_throw_expression_token1] = ACTIONS(1176),
    [aux_sym_while_statement_token1] = ACTIONS(1176),
    [aux_sym_while_statement_token2] = ACTIONS(1176),
    [aux_sym_do_statement_token1] = ACTIONS(1176),
    [aux_sym_for_statement_token1] = ACTIONS(1176),
    [aux_sym_for_statement_token2] = ACTIONS(1176),
    [aux_sym_foreach_statement_token1] = ACTIONS(1176),
    [aux_sym_foreach_statement_token2] = ACTIONS(1176),
    [aux_sym_if_statement_token1] = ACTIONS(1176),
    [aux_sym_if_statement_token2] = ACTIONS(1176),
    [aux_sym_else_if_clause_token1] = ACTIONS(1176),
    [aux_sym_else_clause_token1] = ACTIONS(1176),
    [aux_sym_match_expression_token1] = ACTIONS(1176),
    [aux_sym_match_default_expression_token1] = ACTIONS(1176),
    [aux_sym_switch_statement_token1] = ACTIONS(1176),
    [aux_sym_switch_block_token1] = ACTIONS(1176),
    [anon_sym_PLUS] = ACTIONS(1176),
    [anon_sym_DASH] = ACTIONS(1176),
    [anon_sym_TILDE] = ACTIONS(1174),
    [anon_sym_BANG] = ACTIONS(1174),
    [anon_sym_AT] = ACTIONS(1174),
    [aux_sym_clone_expression_token1] = ACTIONS(1176),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1176),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1176),
    [anon_sym_DASH_DASH] = ACTIONS(1174),
    [anon_sym_PLUS_PLUS] = ACTIONS(1174),
    [aux_sym__list_destructing_token1] = ACTIONS(1176),
    [anon_sym_LBRACK] = ACTIONS(1174),
    [anon_sym_self] = ACTIONS(1176),
    [anon_sym_parent] = ACTIONS(1176),
    [aux_sym__argument_name_token1] = ACTIONS(1176),
    [aux_sym__argument_name_token2] = ACTIONS(1176),
    [anon_sym_POUND_LBRACK] = ACTIONS(1174),
    [aux_sym_encapsed_string_token1] = ACTIONS(1174),
    [anon_sym_DQUOTE] = ACTIONS(1174),
    [aux_sym_string_token1] = ACTIONS(1174),
    [anon_sym_SQUOTE] = ACTIONS(1174),
    [anon_sym_LT_LT_LT] = ACTIONS(1174),
    [anon_sym_BQUOTE] = ACTIONS(1174),
    [anon_sym_DOLLAR] = ACTIONS(1174),
    [aux_sym_yield_expression_token1] = ACTIONS(1176),
    [aux_sym_include_expression_token1] = ACTIONS(1176),
    [aux_sym_include_once_expression_token1] = ACTIONS(1176),
    [aux_sym_require_expression_token1] = ACTIONS(1176),
    [aux_sym_require_once_expression_token1] = ACTIONS(1176),
    [sym_comment] = ACTIONS(3),
  },
  [476] = {
    [ts_builtin_sym_end] = ACTIONS(1178),
    [sym_name] = ACTIONS(1180),
    [anon_sym_QMARK_GT] = ACTIONS(1178),
    [anon_sym_SEMI] = ACTIONS(1178),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1180),
    [aux_sym_global_declaration_token1] = ACTIONS(1180),
    [aux_sym_namespace_definition_token1] = ACTIONS(1180),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1180),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1180),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1180),
    [anon_sym_BSLASH] = ACTIONS(1178),
    [anon_sym_LBRACE] = ACTIONS(1178),
    [anon_sym_RBRACE] = ACTIONS(1178),
    [aux_sym_trait_declaration_token1] = ACTIONS(1180),
    [aux_sym_interface_declaration_token1] = ACTIONS(1180),
    [aux_sym_enum_declaration_token1] = ACTIONS(1180),
    [aux_sym_enum_case_token1] = ACTIONS(1180),
    [aux_sym_class_declaration_token1] = ACTIONS(1180),
    [aux_sym_final_modifier_token1] = ACTIONS(1180),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1180),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1180),
    [sym_var_modifier] = ACTIONS(1180),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1180),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1180),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1180),
    [anon_sym_LPAREN2] = ACTIONS(1178),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1180),
    [aux_sym_cast_type_token1] = ACTIONS(1180),
    [aux_sym_echo_statement_token1] = ACTIONS(1180),
    [aux_sym_exit_statement_token1] = ACTIONS(1180),
    [anon_sym_unset] = ACTIONS(1180),
    [aux_sym_declare_statement_token1] = ACTIONS(1180),
    [aux_sym_declare_statement_token2] = ACTIONS(1180),
    [sym_float] = ACTIONS(1180),
    [aux_sym_try_statement_token1] = ACTIONS(1180),
    [aux_sym_goto_statement_token1] = ACTIONS(1180),
    [aux_sym_continue_statement_token1] = ACTIONS(1180),
    [aux_sym_break_statement_token1] = ACTIONS(1180),
    [sym_integer] = ACTIONS(1180),
    [aux_sym_return_statement_token1] = ACTIONS(1180),
    [aux_sym_throw_expression_token1] = ACTIONS(1180),
    [aux_sym_while_statement_token1] = ACTIONS(1180),
    [aux_sym_while_statement_token2] = ACTIONS(1180),
    [aux_sym_do_statement_token1] = ACTIONS(1180),
    [aux_sym_for_statement_token1] = ACTIONS(1180),
    [aux_sym_for_statement_token2] = ACTIONS(1180),
    [aux_sym_foreach_statement_token1] = ACTIONS(1180),
    [aux_sym_foreach_statement_token2] = ACTIONS(1180),
    [aux_sym_if_statement_token1] = ACTIONS(1180),
    [aux_sym_if_statement_token2] = ACTIONS(1180),
    [aux_sym_else_if_clause_token1] = ACTIONS(1180),
    [aux_sym_else_clause_token1] = ACTIONS(1180),
    [aux_sym_match_expression_token1] = ACTIONS(1180),
    [aux_sym_match_default_expression_token1] = ACTIONS(1180),
    [aux_sym_switch_statement_token1] = ACTIONS(1180),
    [aux_sym_switch_block_token1] = ACTIONS(1180),
    [anon_sym_PLUS] = ACTIONS(1180),
    [anon_sym_DASH] = ACTIONS(1180),
    [anon_sym_TILDE] = ACTIONS(1178),
    [anon_sym_BANG] = ACTIONS(1178),
    [anon_sym_AT] = ACTIONS(1178),
    [aux_sym_clone_expression_token1] = ACTIONS(1180),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1180),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1180),
    [anon_sym_DASH_DASH] = ACTIONS(1178),
    [anon_sym_PLUS_PLUS] = ACTIONS(1178),
    [aux_sym__list_destructing_token1] = ACTIONS(1180),
    [anon_sym_LBRACK] = ACTIONS(1178),
    [anon_sym_self] = ACTIONS(1180),
    [anon_sym_parent] = ACTIONS(1180),
    [aux_sym__argument_name_token1] = ACTIONS(1180),
    [aux_sym__argument_name_token2] = ACTIONS(1180),
    [anon_sym_POUND_LBRACK] = ACTIONS(1178),
    [aux_sym_encapsed_string_token1] = ACTIONS(1178),
    [anon_sym_DQUOTE] = ACTIONS(1178),
    [aux_sym_string_token1] = ACTIONS(1178),
    [anon_sym_SQUOTE] = ACTIONS(1178),
    [anon_sym_LT_LT_LT] = ACTIONS(1178),
    [anon_sym_BQUOTE] = ACTIONS(1178),
    [anon_sym_DOLLAR] = ACTIONS(1178),
    [aux_sym_yield_expression_token1] = ACTIONS(1180),
    [aux_sym_include_expression_token1] = ACTIONS(1180),
    [aux_sym_include_once_expression_token1] = ACTIONS(1180),
    [aux_sym_require_expression_token1] = ACTIONS(1180),
    [aux_sym_require_once_expression_token1] = ACTIONS(1180),
    [sym_comment] = ACTIONS(3),
  },
  [477] = {
    [ts_builtin_sym_end] = ACTIONS(1182),
    [sym_name] = ACTIONS(1184),
    [anon_sym_QMARK_GT] = ACTIONS(1182),
    [anon_sym_SEMI] = ACTIONS(1182),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1184),
    [aux_sym_global_declaration_token1] = ACTIONS(1184),
    [aux_sym_namespace_definition_token1] = ACTIONS(1184),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1184),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1184),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1184),
    [anon_sym_BSLASH] = ACTIONS(1182),
    [anon_sym_LBRACE] = ACTIONS(1182),
    [anon_sym_RBRACE] = ACTIONS(1182),
    [aux_sym_trait_declaration_token1] = ACTIONS(1184),
    [aux_sym_interface_declaration_token1] = ACTIONS(1184),
    [aux_sym_enum_declaration_token1] = ACTIONS(1184),
    [aux_sym_enum_case_token1] = ACTIONS(1184),
    [aux_sym_class_declaration_token1] = ACTIONS(1184),
    [aux_sym_final_modifier_token1] = ACTIONS(1184),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1184),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1184),
    [sym_var_modifier] = ACTIONS(1184),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1184),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1184),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1184),
    [anon_sym_LPAREN2] = ACTIONS(1182),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1184),
    [aux_sym_cast_type_token1] = ACTIONS(1184),
    [aux_sym_echo_statement_token1] = ACTIONS(1184),
    [aux_sym_exit_statement_token1] = ACTIONS(1184),
    [anon_sym_unset] = ACTIONS(1184),
    [aux_sym_declare_statement_token1] = ACTIONS(1184),
    [aux_sym_declare_statement_token2] = ACTIONS(1184),
    [sym_float] = ACTIONS(1184),
    [aux_sym_try_statement_token1] = ACTIONS(1184),
    [aux_sym_goto_statement_token1] = ACTIONS(1184),
    [aux_sym_continue_statement_token1] = ACTIONS(1184),
    [aux_sym_break_statement_token1] = ACTIONS(1184),
    [sym_integer] = ACTIONS(1184),
    [aux_sym_return_statement_token1] = ACTIONS(1184),
    [aux_sym_throw_expression_token1] = ACTIONS(1184),
    [aux_sym_while_statement_token1] = ACTIONS(1184),
    [aux_sym_while_statement_token2] = ACTIONS(1184),
    [aux_sym_do_statement_token1] = ACTIONS(1184),
    [aux_sym_for_statement_token1] = ACTIONS(1184),
    [aux_sym_for_statement_token2] = ACTIONS(1184),
    [aux_sym_foreach_statement_token1] = ACTIONS(1184),
    [aux_sym_foreach_statement_token2] = ACTIONS(1184),
    [aux_sym_if_statement_token1] = ACTIONS(1184),
    [aux_sym_if_statement_token2] = ACTIONS(1184),
    [aux_sym_else_if_clause_token1] = ACTIONS(1184),
    [aux_sym_else_clause_token1] = ACTIONS(1184),
    [aux_sym_match_expression_token1] = ACTIONS(1184),
    [aux_sym_match_default_expression_token1] = ACTIONS(1184),
    [aux_sym_switch_statement_token1] = ACTIONS(1184),
    [aux_sym_switch_block_token1] = ACTIONS(1184),
    [anon_sym_PLUS] = ACTIONS(1184),
    [anon_sym_DASH] = ACTIONS(1184),
    [anon_sym_TILDE] = ACTIONS(1182),
    [anon_sym_BANG] = ACTIONS(1182),
    [anon_sym_AT] = ACTIONS(1182),
    [aux_sym_clone_expression_token1] = ACTIONS(1184),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1184),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1184),
    [anon_sym_DASH_DASH] = ACTIONS(1182),
    [anon_sym_PLUS_PLUS] = ACTIONS(1182),
    [aux_sym__list_destructing_token1] = ACTIONS(1184),
    [anon_sym_LBRACK] = ACTIONS(1182),
    [anon_sym_self] = ACTIONS(1184),
    [anon_sym_parent] = ACTIONS(1184),
    [aux_sym__argument_name_token1] = ACTIONS(1184),
    [aux_sym__argument_name_token2] = ACTIONS(1184),
    [anon_sym_POUND_LBRACK] = ACTIONS(1182),
    [aux_sym_encapsed_string_token1] = ACTIONS(1182),
    [anon_sym_DQUOTE] = ACTIONS(1182),
    [aux_sym_string_token1] = ACTIONS(1182),
    [anon_sym_SQUOTE] = ACTIONS(1182),
    [anon_sym_LT_LT_LT] = ACTIONS(1182),
    [anon_sym_BQUOTE] = ACTIONS(1182),
    [anon_sym_DOLLAR] = ACTIONS(1182),
    [aux_sym_yield_expression_token1] = ACTIONS(1184),
    [aux_sym_include_expression_token1] = ACTIONS(1184),
    [aux_sym_include_once_expression_token1] = ACTIONS(1184),
    [aux_sym_require_expression_token1] = ACTIONS(1184),
    [aux_sym_require_once_expression_token1] = ACTIONS(1184),
    [sym_comment] = ACTIONS(3),
  },
  [478] = {
    [ts_builtin_sym_end] = ACTIONS(1186),
    [sym_name] = ACTIONS(1188),
    [anon_sym_QMARK_GT] = ACTIONS(1186),
    [anon_sym_SEMI] = ACTIONS(1186),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1188),
    [aux_sym_global_declaration_token1] = ACTIONS(1188),
    [aux_sym_namespace_definition_token1] = ACTIONS(1188),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1188),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1188),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1188),
    [anon_sym_BSLASH] = ACTIONS(1186),
    [anon_sym_LBRACE] = ACTIONS(1186),
    [anon_sym_RBRACE] = ACTIONS(1186),
    [aux_sym_trait_declaration_token1] = ACTIONS(1188),
    [aux_sym_interface_declaration_token1] = ACTIONS(1188),
    [aux_sym_enum_declaration_token1] = ACTIONS(1188),
    [aux_sym_enum_case_token1] = ACTIONS(1188),
    [aux_sym_class_declaration_token1] = ACTIONS(1188),
    [aux_sym_final_modifier_token1] = ACTIONS(1188),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1188),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1188),
    [sym_var_modifier] = ACTIONS(1188),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1188),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1188),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1188),
    [anon_sym_LPAREN2] = ACTIONS(1186),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1188),
    [aux_sym_cast_type_token1] = ACTIONS(1188),
    [aux_sym_echo_statement_token1] = ACTIONS(1188),
    [aux_sym_exit_statement_token1] = ACTIONS(1188),
    [anon_sym_unset] = ACTIONS(1188),
    [aux_sym_declare_statement_token1] = ACTIONS(1188),
    [aux_sym_declare_statement_token2] = ACTIONS(1188),
    [sym_float] = ACTIONS(1188),
    [aux_sym_try_statement_token1] = ACTIONS(1188),
    [aux_sym_goto_statement_token1] = ACTIONS(1188),
    [aux_sym_continue_statement_token1] = ACTIONS(1188),
    [aux_sym_break_statement_token1] = ACTIONS(1188),
    [sym_integer] = ACTIONS(1188),
    [aux_sym_return_statement_token1] = ACTIONS(1188),
    [aux_sym_throw_expression_token1] = ACTIONS(1188),
    [aux_sym_while_statement_token1] = ACTIONS(1188),
    [aux_sym_while_statement_token2] = ACTIONS(1188),
    [aux_sym_do_statement_token1] = ACTIONS(1188),
    [aux_sym_for_statement_token1] = ACTIONS(1188),
    [aux_sym_for_statement_token2] = ACTIONS(1188),
    [aux_sym_foreach_statement_token1] = ACTIONS(1188),
    [aux_sym_foreach_statement_token2] = ACTIONS(1188),
    [aux_sym_if_statement_token1] = ACTIONS(1188),
    [aux_sym_if_statement_token2] = ACTIONS(1188),
    [aux_sym_else_if_clause_token1] = ACTIONS(1188),
    [aux_sym_else_clause_token1] = ACTIONS(1188),
    [aux_sym_match_expression_token1] = ACTIONS(1188),
    [aux_sym_match_default_expression_token1] = ACTIONS(1188),
    [aux_sym_switch_statement_token1] = ACTIONS(1188),
    [aux_sym_switch_block_token1] = ACTIONS(1188),
    [anon_sym_PLUS] = ACTIONS(1188),
    [anon_sym_DASH] = ACTIONS(1188),
    [anon_sym_TILDE] = ACTIONS(1186),
    [anon_sym_BANG] = ACTIONS(1186),
    [anon_sym_AT] = ACTIONS(1186),
    [aux_sym_clone_expression_token1] = ACTIONS(1188),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1188),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1188),
    [anon_sym_DASH_DASH] = ACTIONS(1186),
    [anon_sym_PLUS_PLUS] = ACTIONS(1186),
    [aux_sym__list_destructing_token1] = ACTIONS(1188),
    [anon_sym_LBRACK] = ACTIONS(1186),
    [anon_sym_self] = ACTIONS(1188),
    [anon_sym_parent] = ACTIONS(1188),
    [aux_sym__argument_name_token1] = ACTIONS(1188),
    [aux_sym__argument_name_token2] = ACTIONS(1188),
    [anon_sym_POUND_LBRACK] = ACTIONS(1186),
    [aux_sym_encapsed_string_token1] = ACTIONS(1186),
    [anon_sym_DQUOTE] = ACTIONS(1186),
    [aux_sym_string_token1] = ACTIONS(1186),
    [anon_sym_SQUOTE] = ACTIONS(1186),
    [anon_sym_LT_LT_LT] = ACTIONS(1186),
    [anon_sym_BQUOTE] = ACTIONS(1186),
    [anon_sym_DOLLAR] = ACTIONS(1186),
    [aux_sym_yield_expression_token1] = ACTIONS(1188),
    [aux_sym_include_expression_token1] = ACTIONS(1188),
    [aux_sym_include_once_expression_token1] = ACTIONS(1188),
    [aux_sym_require_expression_token1] = ACTIONS(1188),
    [aux_sym_require_once_expression_token1] = ACTIONS(1188),
    [sym_comment] = ACTIONS(3),
  },
  [479] = {
    [ts_builtin_sym_end] = ACTIONS(1190),
    [sym_name] = ACTIONS(1192),
    [anon_sym_QMARK_GT] = ACTIONS(1190),
    [anon_sym_SEMI] = ACTIONS(1190),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1192),
    [aux_sym_global_declaration_token1] = ACTIONS(1192),
    [aux_sym_namespace_definition_token1] = ACTIONS(1192),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1192),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1192),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1192),
    [anon_sym_BSLASH] = ACTIONS(1190),
    [anon_sym_LBRACE] = ACTIONS(1190),
    [anon_sym_RBRACE] = ACTIONS(1190),
    [aux_sym_trait_declaration_token1] = ACTIONS(1192),
    [aux_sym_interface_declaration_token1] = ACTIONS(1192),
    [aux_sym_enum_declaration_token1] = ACTIONS(1192),
    [aux_sym_enum_case_token1] = ACTIONS(1192),
    [aux_sym_class_declaration_token1] = ACTIONS(1192),
    [aux_sym_final_modifier_token1] = ACTIONS(1192),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1192),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1192),
    [sym_var_modifier] = ACTIONS(1192),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1192),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1192),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1192),
    [anon_sym_LPAREN2] = ACTIONS(1190),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1192),
    [aux_sym_cast_type_token1] = ACTIONS(1192),
    [aux_sym_echo_statement_token1] = ACTIONS(1192),
    [aux_sym_exit_statement_token1] = ACTIONS(1192),
    [anon_sym_unset] = ACTIONS(1192),
    [aux_sym_declare_statement_token1] = ACTIONS(1192),
    [aux_sym_declare_statement_token2] = ACTIONS(1192),
    [sym_float] = ACTIONS(1192),
    [aux_sym_try_statement_token1] = ACTIONS(1192),
    [aux_sym_goto_statement_token1] = ACTIONS(1192),
    [aux_sym_continue_statement_token1] = ACTIONS(1192),
    [aux_sym_break_statement_token1] = ACTIONS(1192),
    [sym_integer] = ACTIONS(1192),
    [aux_sym_return_statement_token1] = ACTIONS(1192),
    [aux_sym_throw_expression_token1] = ACTIONS(1192),
    [aux_sym_while_statement_token1] = ACTIONS(1192),
    [aux_sym_while_statement_token2] = ACTIONS(1192),
    [aux_sym_do_statement_token1] = ACTIONS(1192),
    [aux_sym_for_statement_token1] = ACTIONS(1192),
    [aux_sym_for_statement_token2] = ACTIONS(1192),
    [aux_sym_foreach_statement_token1] = ACTIONS(1192),
    [aux_sym_foreach_statement_token2] = ACTIONS(1192),
    [aux_sym_if_statement_token1] = ACTIONS(1192),
    [aux_sym_if_statement_token2] = ACTIONS(1192),
    [aux_sym_else_if_clause_token1] = ACTIONS(1192),
    [aux_sym_else_clause_token1] = ACTIONS(1192),
    [aux_sym_match_expression_token1] = ACTIONS(1192),
    [aux_sym_match_default_expression_token1] = ACTIONS(1192),
    [aux_sym_switch_statement_token1] = ACTIONS(1192),
    [aux_sym_switch_block_token1] = ACTIONS(1192),
    [anon_sym_PLUS] = ACTIONS(1192),
    [anon_sym_DASH] = ACTIONS(1192),
    [anon_sym_TILDE] = ACTIONS(1190),
    [anon_sym_BANG] = ACTIONS(1190),
    [anon_sym_AT] = ACTIONS(1190),
    [aux_sym_clone_expression_token1] = ACTIONS(1192),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1192),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1192),
    [anon_sym_DASH_DASH] = ACTIONS(1190),
    [anon_sym_PLUS_PLUS] = ACTIONS(1190),
    [aux_sym__list_destructing_token1] = ACTIONS(1192),
    [anon_sym_LBRACK] = ACTIONS(1190),
    [anon_sym_self] = ACTIONS(1192),
    [anon_sym_parent] = ACTIONS(1192),
    [aux_sym__argument_name_token1] = ACTIONS(1192),
    [aux_sym__argument_name_token2] = ACTIONS(1192),
    [anon_sym_POUND_LBRACK] = ACTIONS(1190),
    [aux_sym_encapsed_string_token1] = ACTIONS(1190),
    [anon_sym_DQUOTE] = ACTIONS(1190),
    [aux_sym_string_token1] = ACTIONS(1190),
    [anon_sym_SQUOTE] = ACTIONS(1190),
    [anon_sym_LT_LT_LT] = ACTIONS(1190),
    [anon_sym_BQUOTE] = ACTIONS(1190),
    [anon_sym_DOLLAR] = ACTIONS(1190),
    [aux_sym_yield_expression_token1] = ACTIONS(1192),
    [aux_sym_include_expression_token1] = ACTIONS(1192),
    [aux_sym_include_once_expression_token1] = ACTIONS(1192),
    [aux_sym_require_expression_token1] = ACTIONS(1192),
    [aux_sym_require_once_expression_token1] = ACTIONS(1192),
    [sym_comment] = ACTIONS(3),
  },
  [480] = {
    [ts_builtin_sym_end] = ACTIONS(1194),
    [sym_name] = ACTIONS(1196),
    [anon_sym_QMARK_GT] = ACTIONS(1194),
    [anon_sym_SEMI] = ACTIONS(1194),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1196),
    [aux_sym_global_declaration_token1] = ACTIONS(1196),
    [aux_sym_namespace_definition_token1] = ACTIONS(1196),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1196),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1196),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1196),
    [anon_sym_BSLASH] = ACTIONS(1194),
    [anon_sym_LBRACE] = ACTIONS(1194),
    [anon_sym_RBRACE] = ACTIONS(1194),
    [aux_sym_trait_declaration_token1] = ACTIONS(1196),
    [aux_sym_interface_declaration_token1] = ACTIONS(1196),
    [aux_sym_enum_declaration_token1] = ACTIONS(1196),
    [aux_sym_enum_case_token1] = ACTIONS(1196),
    [aux_sym_class_declaration_token1] = ACTIONS(1196),
    [aux_sym_final_modifier_token1] = ACTIONS(1196),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1196),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1196),
    [sym_var_modifier] = ACTIONS(1196),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1196),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1196),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1196),
    [anon_sym_LPAREN2] = ACTIONS(1194),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1196),
    [aux_sym_cast_type_token1] = ACTIONS(1196),
    [aux_sym_echo_statement_token1] = ACTIONS(1196),
    [aux_sym_exit_statement_token1] = ACTIONS(1196),
    [anon_sym_unset] = ACTIONS(1196),
    [aux_sym_declare_statement_token1] = ACTIONS(1196),
    [aux_sym_declare_statement_token2] = ACTIONS(1196),
    [sym_float] = ACTIONS(1196),
    [aux_sym_try_statement_token1] = ACTIONS(1196),
    [aux_sym_goto_statement_token1] = ACTIONS(1196),
    [aux_sym_continue_statement_token1] = ACTIONS(1196),
    [aux_sym_break_statement_token1] = ACTIONS(1196),
    [sym_integer] = ACTIONS(1196),
    [aux_sym_return_statement_token1] = ACTIONS(1196),
    [aux_sym_throw_expression_token1] = ACTIONS(1196),
    [aux_sym_while_statement_token1] = ACTIONS(1196),
    [aux_sym_while_statement_token2] = ACTIONS(1196),
    [aux_sym_do_statement_token1] = ACTIONS(1196),
    [aux_sym_for_statement_token1] = ACTIONS(1196),
    [aux_sym_for_statement_token2] = ACTIONS(1196),
    [aux_sym_foreach_statement_token1] = ACTIONS(1196),
    [aux_sym_foreach_statement_token2] = ACTIONS(1196),
    [aux_sym_if_statement_token1] = ACTIONS(1196),
    [aux_sym_if_statement_token2] = ACTIONS(1196),
    [aux_sym_else_if_clause_token1] = ACTIONS(1196),
    [aux_sym_else_clause_token1] = ACTIONS(1196),
    [aux_sym_match_expression_token1] = ACTIONS(1196),
    [aux_sym_match_default_expression_token1] = ACTIONS(1196),
    [aux_sym_switch_statement_token1] = ACTIONS(1196),
    [aux_sym_switch_block_token1] = ACTIONS(1196),
    [anon_sym_PLUS] = ACTIONS(1196),
    [anon_sym_DASH] = ACTIONS(1196),
    [anon_sym_TILDE] = ACTIONS(1194),
    [anon_sym_BANG] = ACTIONS(1194),
    [anon_sym_AT] = ACTIONS(1194),
    [aux_sym_clone_expression_token1] = ACTIONS(1196),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1196),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1196),
    [anon_sym_DASH_DASH] = ACTIONS(1194),
    [anon_sym_PLUS_PLUS] = ACTIONS(1194),
    [aux_sym__list_destructing_token1] = ACTIONS(1196),
    [anon_sym_LBRACK] = ACTIONS(1194),
    [anon_sym_self] = ACTIONS(1196),
    [anon_sym_parent] = ACTIONS(1196),
    [aux_sym__argument_name_token1] = ACTIONS(1196),
    [aux_sym__argument_name_token2] = ACTIONS(1196),
    [anon_sym_POUND_LBRACK] = ACTIONS(1194),
    [aux_sym_encapsed_string_token1] = ACTIONS(1194),
    [anon_sym_DQUOTE] = ACTIONS(1194),
    [aux_sym_string_token1] = ACTIONS(1194),
    [anon_sym_SQUOTE] = ACTIONS(1194),
    [anon_sym_LT_LT_LT] = ACTIONS(1194),
    [anon_sym_BQUOTE] = ACTIONS(1194),
    [anon_sym_DOLLAR] = ACTIONS(1194),
    [aux_sym_yield_expression_token1] = ACTIONS(1196),
    [aux_sym_include_expression_token1] = ACTIONS(1196),
    [aux_sym_include_once_expression_token1] = ACTIONS(1196),
    [aux_sym_require_expression_token1] = ACTIONS(1196),
    [aux_sym_require_once_expression_token1] = ACTIONS(1196),
    [sym_comment] = ACTIONS(3),
  },
  [481] = {
    [ts_builtin_sym_end] = ACTIONS(1198),
    [sym_name] = ACTIONS(1200),
    [anon_sym_QMARK_GT] = ACTIONS(1198),
    [anon_sym_SEMI] = ACTIONS(1198),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1200),
    [aux_sym_global_declaration_token1] = ACTIONS(1200),
    [aux_sym_namespace_definition_token1] = ACTIONS(1200),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1200),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1200),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1200),
    [anon_sym_BSLASH] = ACTIONS(1198),
    [anon_sym_LBRACE] = ACTIONS(1198),
    [anon_sym_RBRACE] = ACTIONS(1198),
    [aux_sym_trait_declaration_token1] = ACTIONS(1200),
    [aux_sym_interface_declaration_token1] = ACTIONS(1200),
    [aux_sym_enum_declaration_token1] = ACTIONS(1200),
    [aux_sym_enum_case_token1] = ACTIONS(1200),
    [aux_sym_class_declaration_token1] = ACTIONS(1200),
    [aux_sym_final_modifier_token1] = ACTIONS(1200),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1200),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1200),
    [sym_var_modifier] = ACTIONS(1200),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1200),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1200),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1200),
    [anon_sym_LPAREN2] = ACTIONS(1198),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1200),
    [aux_sym_cast_type_token1] = ACTIONS(1200),
    [aux_sym_echo_statement_token1] = ACTIONS(1200),
    [aux_sym_exit_statement_token1] = ACTIONS(1200),
    [anon_sym_unset] = ACTIONS(1200),
    [aux_sym_declare_statement_token1] = ACTIONS(1200),
    [aux_sym_declare_statement_token2] = ACTIONS(1200),
    [sym_float] = ACTIONS(1200),
    [aux_sym_try_statement_token1] = ACTIONS(1200),
    [aux_sym_goto_statement_token1] = ACTIONS(1200),
    [aux_sym_continue_statement_token1] = ACTIONS(1200),
    [aux_sym_break_statement_token1] = ACTIONS(1200),
    [sym_integer] = ACTIONS(1200),
    [aux_sym_return_statement_token1] = ACTIONS(1200),
    [aux_sym_throw_expression_token1] = ACTIONS(1200),
    [aux_sym_while_statement_token1] = ACTIONS(1200),
    [aux_sym_while_statement_token2] = ACTIONS(1200),
    [aux_sym_do_statement_token1] = ACTIONS(1200),
    [aux_sym_for_statement_token1] = ACTIONS(1200),
    [aux_sym_for_statement_token2] = ACTIONS(1200),
    [aux_sym_foreach_statement_token1] = ACTIONS(1200),
    [aux_sym_foreach_statement_token2] = ACTIONS(1200),
    [aux_sym_if_statement_token1] = ACTIONS(1200),
    [aux_sym_if_statement_token2] = ACTIONS(1200),
    [aux_sym_else_if_clause_token1] = ACTIONS(1200),
    [aux_sym_else_clause_token1] = ACTIONS(1200),
    [aux_sym_match_expression_token1] = ACTIONS(1200),
    [aux_sym_match_default_expression_token1] = ACTIONS(1200),
    [aux_sym_switch_statement_token1] = ACTIONS(1200),
    [aux_sym_switch_block_token1] = ACTIONS(1200),
    [anon_sym_PLUS] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1200),
    [anon_sym_TILDE] = ACTIONS(1198),
    [anon_sym_BANG] = ACTIONS(1198),
    [anon_sym_AT] = ACTIONS(1198),
    [aux_sym_clone_expression_token1] = ACTIONS(1200),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1200),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1200),
    [anon_sym_DASH_DASH] = ACTIONS(1198),
    [anon_sym_PLUS_PLUS] = ACTIONS(1198),
    [aux_sym__list_destructing_token1] = ACTIONS(1200),
    [anon_sym_LBRACK] = ACTIONS(1198),
    [anon_sym_self] = ACTIONS(1200),
    [anon_sym_parent] = ACTIONS(1200),
    [aux_sym__argument_name_token1] = ACTIONS(1200),
    [aux_sym__argument_name_token2] = ACTIONS(1200),
    [anon_sym_POUND_LBRACK] = ACTIONS(1198),
    [aux_sym_encapsed_string_token1] = ACTIONS(1198),
    [anon_sym_DQUOTE] = ACTIONS(1198),
    [aux_sym_string_token1] = ACTIONS(1198),
    [anon_sym_SQUOTE] = ACTIONS(1198),
    [anon_sym_LT_LT_LT] = ACTIONS(1198),
    [anon_sym_BQUOTE] = ACTIONS(1198),
    [anon_sym_DOLLAR] = ACTIONS(1198),
    [aux_sym_yield_expression_token1] = ACTIONS(1200),
    [aux_sym_include_expression_token1] = ACTIONS(1200),
    [aux_sym_include_once_expression_token1] = ACTIONS(1200),
    [aux_sym_require_expression_token1] = ACTIONS(1200),
    [aux_sym_require_once_expression_token1] = ACTIONS(1200),
    [sym_comment] = ACTIONS(3),
  },
  [482] = {
    [ts_builtin_sym_end] = ACTIONS(1202),
    [sym_name] = ACTIONS(1204),
    [anon_sym_QMARK_GT] = ACTIONS(1202),
    [anon_sym_SEMI] = ACTIONS(1202),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1204),
    [aux_sym_global_declaration_token1] = ACTIONS(1204),
    [aux_sym_namespace_definition_token1] = ACTIONS(1204),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1204),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1204),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1204),
    [anon_sym_BSLASH] = ACTIONS(1202),
    [anon_sym_LBRACE] = ACTIONS(1202),
    [anon_sym_RBRACE] = ACTIONS(1202),
    [aux_sym_trait_declaration_token1] = ACTIONS(1204),
    [aux_sym_interface_declaration_token1] = ACTIONS(1204),
    [aux_sym_enum_declaration_token1] = ACTIONS(1204),
    [aux_sym_enum_case_token1] = ACTIONS(1204),
    [aux_sym_class_declaration_token1] = ACTIONS(1204),
    [aux_sym_final_modifier_token1] = ACTIONS(1204),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1204),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1204),
    [sym_var_modifier] = ACTIONS(1204),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1204),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1204),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1204),
    [anon_sym_LPAREN2] = ACTIONS(1202),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1204),
    [aux_sym_cast_type_token1] = ACTIONS(1204),
    [aux_sym_echo_statement_token1] = ACTIONS(1204),
    [aux_sym_exit_statement_token1] = ACTIONS(1204),
    [anon_sym_unset] = ACTIONS(1204),
    [aux_sym_declare_statement_token1] = ACTIONS(1204),
    [aux_sym_declare_statement_token2] = ACTIONS(1204),
    [sym_float] = ACTIONS(1204),
    [aux_sym_try_statement_token1] = ACTIONS(1204),
    [aux_sym_goto_statement_token1] = ACTIONS(1204),
    [aux_sym_continue_statement_token1] = ACTIONS(1204),
    [aux_sym_break_statement_token1] = ACTIONS(1204),
    [sym_integer] = ACTIONS(1204),
    [aux_sym_return_statement_token1] = ACTIONS(1204),
    [aux_sym_throw_expression_token1] = ACTIONS(1204),
    [aux_sym_while_statement_token1] = ACTIONS(1204),
    [aux_sym_while_statement_token2] = ACTIONS(1204),
    [aux_sym_do_statement_token1] = ACTIONS(1204),
    [aux_sym_for_statement_token1] = ACTIONS(1204),
    [aux_sym_for_statement_token2] = ACTIONS(1204),
    [aux_sym_foreach_statement_token1] = ACTIONS(1204),
    [aux_sym_foreach_statement_token2] = ACTIONS(1204),
    [aux_sym_if_statement_token1] = ACTIONS(1204),
    [aux_sym_if_statement_token2] = ACTIONS(1204),
    [aux_sym_else_if_clause_token1] = ACTIONS(1204),
    [aux_sym_else_clause_token1] = ACTIONS(1204),
    [aux_sym_match_expression_token1] = ACTIONS(1204),
    [aux_sym_match_default_expression_token1] = ACTIONS(1204),
    [aux_sym_switch_statement_token1] = ACTIONS(1204),
    [aux_sym_switch_block_token1] = ACTIONS(1204),
    [anon_sym_PLUS] = ACTIONS(1204),
    [anon_sym_DASH] = ACTIONS(1204),
    [anon_sym_TILDE] = ACTIONS(1202),
    [anon_sym_BANG] = ACTIONS(1202),
    [anon_sym_AT] = ACTIONS(1202),
    [aux_sym_clone_expression_token1] = ACTIONS(1204),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1204),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1204),
    [anon_sym_DASH_DASH] = ACTIONS(1202),
    [anon_sym_PLUS_PLUS] = ACTIONS(1202),
    [aux_sym__list_destructing_token1] = ACTIONS(1204),
    [anon_sym_LBRACK] = ACTIONS(1202),
    [anon_sym_self] = ACTIONS(1204),
    [anon_sym_parent] = ACTIONS(1204),
    [aux_sym__argument_name_token1] = ACTIONS(1204),
    [aux_sym__argument_name_token2] = ACTIONS(1204),
    [anon_sym_POUND_LBRACK] = ACTIONS(1202),
    [aux_sym_encapsed_string_token1] = ACTIONS(1202),
    [anon_sym_DQUOTE] = ACTIONS(1202),
    [aux_sym_string_token1] = ACTIONS(1202),
    [anon_sym_SQUOTE] = ACTIONS(1202),
    [anon_sym_LT_LT_LT] = ACTIONS(1202),
    [anon_sym_BQUOTE] = ACTIONS(1202),
    [anon_sym_DOLLAR] = ACTIONS(1202),
    [aux_sym_yield_expression_token1] = ACTIONS(1204),
    [aux_sym_include_expression_token1] = ACTIONS(1204),
    [aux_sym_include_once_expression_token1] = ACTIONS(1204),
    [aux_sym_require_expression_token1] = ACTIONS(1204),
    [aux_sym_require_once_expression_token1] = ACTIONS(1204),
    [sym_comment] = ACTIONS(3),
  },
  [483] = {
    [ts_builtin_sym_end] = ACTIONS(1206),
    [sym_name] = ACTIONS(1208),
    [anon_sym_QMARK_GT] = ACTIONS(1206),
    [anon_sym_SEMI] = ACTIONS(1206),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1208),
    [aux_sym_global_declaration_token1] = ACTIONS(1208),
    [aux_sym_namespace_definition_token1] = ACTIONS(1208),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1208),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1208),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1208),
    [anon_sym_BSLASH] = ACTIONS(1206),
    [anon_sym_LBRACE] = ACTIONS(1206),
    [anon_sym_RBRACE] = ACTIONS(1206),
    [aux_sym_trait_declaration_token1] = ACTIONS(1208),
    [aux_sym_interface_declaration_token1] = ACTIONS(1208),
    [aux_sym_enum_declaration_token1] = ACTIONS(1208),
    [aux_sym_enum_case_token1] = ACTIONS(1208),
    [aux_sym_class_declaration_token1] = ACTIONS(1208),
    [aux_sym_final_modifier_token1] = ACTIONS(1208),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1208),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1208),
    [sym_var_modifier] = ACTIONS(1208),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1208),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1208),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1208),
    [anon_sym_LPAREN2] = ACTIONS(1206),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1208),
    [aux_sym_cast_type_token1] = ACTIONS(1208),
    [aux_sym_echo_statement_token1] = ACTIONS(1208),
    [aux_sym_exit_statement_token1] = ACTIONS(1208),
    [anon_sym_unset] = ACTIONS(1208),
    [aux_sym_declare_statement_token1] = ACTIONS(1208),
    [aux_sym_declare_statement_token2] = ACTIONS(1208),
    [sym_float] = ACTIONS(1208),
    [aux_sym_try_statement_token1] = ACTIONS(1208),
    [aux_sym_goto_statement_token1] = ACTIONS(1208),
    [aux_sym_continue_statement_token1] = ACTIONS(1208),
    [aux_sym_break_statement_token1] = ACTIONS(1208),
    [sym_integer] = ACTIONS(1208),
    [aux_sym_return_statement_token1] = ACTIONS(1208),
    [aux_sym_throw_expression_token1] = ACTIONS(1208),
    [aux_sym_while_statement_token1] = ACTIONS(1208),
    [aux_sym_while_statement_token2] = ACTIONS(1208),
    [aux_sym_do_statement_token1] = ACTIONS(1208),
    [aux_sym_for_statement_token1] = ACTIONS(1208),
    [aux_sym_for_statement_token2] = ACTIONS(1208),
    [aux_sym_foreach_statement_token1] = ACTIONS(1208),
    [aux_sym_foreach_statement_token2] = ACTIONS(1208),
    [aux_sym_if_statement_token1] = ACTIONS(1208),
    [aux_sym_if_statement_token2] = ACTIONS(1208),
    [aux_sym_else_if_clause_token1] = ACTIONS(1208),
    [aux_sym_else_clause_token1] = ACTIONS(1208),
    [aux_sym_match_expression_token1] = ACTIONS(1208),
    [aux_sym_match_default_expression_token1] = ACTIONS(1208),
    [aux_sym_switch_statement_token1] = ACTIONS(1208),
    [aux_sym_switch_block_token1] = ACTIONS(1208),
    [anon_sym_PLUS] = ACTIONS(1208),
    [anon_sym_DASH] = ACTIONS(1208),
    [anon_sym_TILDE] = ACTIONS(1206),
    [anon_sym_BANG] = ACTIONS(1206),
    [anon_sym_AT] = ACTIONS(1206),
    [aux_sym_clone_expression_token1] = ACTIONS(1208),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1208),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1208),
    [anon_sym_DASH_DASH] = ACTIONS(1206),
    [anon_sym_PLUS_PLUS] = ACTIONS(1206),
    [aux_sym__list_destructing_token1] = ACTIONS(1208),
    [anon_sym_LBRACK] = ACTIONS(1206),
    [anon_sym_self] = ACTIONS(1208),
    [anon_sym_parent] = ACTIONS(1208),
    [aux_sym__argument_name_token1] = ACTIONS(1208),
    [aux_sym__argument_name_token2] = ACTIONS(1208),
    [anon_sym_POUND_LBRACK] = ACTIONS(1206),
    [aux_sym_encapsed_string_token1] = ACTIONS(1206),
    [anon_sym_DQUOTE] = ACTIONS(1206),
    [aux_sym_string_token1] = ACTIONS(1206),
    [anon_sym_SQUOTE] = ACTIONS(1206),
    [anon_sym_LT_LT_LT] = ACTIONS(1206),
    [anon_sym_BQUOTE] = ACTIONS(1206),
    [anon_sym_DOLLAR] = ACTIONS(1206),
    [aux_sym_yield_expression_token1] = ACTIONS(1208),
    [aux_sym_include_expression_token1] = ACTIONS(1208),
    [aux_sym_include_once_expression_token1] = ACTIONS(1208),
    [aux_sym_require_expression_token1] = ACTIONS(1208),
    [aux_sym_require_once_expression_token1] = ACTIONS(1208),
    [sym_comment] = ACTIONS(3),
  },
  [484] = {
    [ts_builtin_sym_end] = ACTIONS(1210),
    [sym_name] = ACTIONS(1212),
    [anon_sym_QMARK_GT] = ACTIONS(1210),
    [anon_sym_SEMI] = ACTIONS(1210),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1212),
    [aux_sym_global_declaration_token1] = ACTIONS(1212),
    [aux_sym_namespace_definition_token1] = ACTIONS(1212),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1212),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1212),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1212),
    [anon_sym_BSLASH] = ACTIONS(1210),
    [anon_sym_LBRACE] = ACTIONS(1210),
    [anon_sym_RBRACE] = ACTIONS(1210),
    [aux_sym_trait_declaration_token1] = ACTIONS(1212),
    [aux_sym_interface_declaration_token1] = ACTIONS(1212),
    [aux_sym_enum_declaration_token1] = ACTIONS(1212),
    [aux_sym_enum_case_token1] = ACTIONS(1212),
    [aux_sym_class_declaration_token1] = ACTIONS(1212),
    [aux_sym_final_modifier_token1] = ACTIONS(1212),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1212),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1212),
    [sym_var_modifier] = ACTIONS(1212),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1212),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1212),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1212),
    [anon_sym_LPAREN2] = ACTIONS(1210),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1212),
    [aux_sym_cast_type_token1] = ACTIONS(1212),
    [aux_sym_echo_statement_token1] = ACTIONS(1212),
    [aux_sym_exit_statement_token1] = ACTIONS(1212),
    [anon_sym_unset] = ACTIONS(1212),
    [aux_sym_declare_statement_token1] = ACTIONS(1212),
    [aux_sym_declare_statement_token2] = ACTIONS(1212),
    [sym_float] = ACTIONS(1212),
    [aux_sym_try_statement_token1] = ACTIONS(1212),
    [aux_sym_goto_statement_token1] = ACTIONS(1212),
    [aux_sym_continue_statement_token1] = ACTIONS(1212),
    [aux_sym_break_statement_token1] = ACTIONS(1212),
    [sym_integer] = ACTIONS(1212),
    [aux_sym_return_statement_token1] = ACTIONS(1212),
    [aux_sym_throw_expression_token1] = ACTIONS(1212),
    [aux_sym_while_statement_token1] = ACTIONS(1212),
    [aux_sym_while_statement_token2] = ACTIONS(1212),
    [aux_sym_do_statement_token1] = ACTIONS(1212),
    [aux_sym_for_statement_token1] = ACTIONS(1212),
    [aux_sym_for_statement_token2] = ACTIONS(1212),
    [aux_sym_foreach_statement_token1] = ACTIONS(1212),
    [aux_sym_foreach_statement_token2] = ACTIONS(1212),
    [aux_sym_if_statement_token1] = ACTIONS(1212),
    [aux_sym_if_statement_token2] = ACTIONS(1212),
    [aux_sym_else_if_clause_token1] = ACTIONS(1212),
    [aux_sym_else_clause_token1] = ACTIONS(1212),
    [aux_sym_match_expression_token1] = ACTIONS(1212),
    [aux_sym_match_default_expression_token1] = ACTIONS(1212),
    [aux_sym_switch_statement_token1] = ACTIONS(1212),
    [aux_sym_switch_block_token1] = ACTIONS(1212),
    [anon_sym_PLUS] = ACTIONS(1212),
    [anon_sym_DASH] = ACTIONS(1212),
    [anon_sym_TILDE] = ACTIONS(1210),
    [anon_sym_BANG] = ACTIONS(1210),
    [anon_sym_AT] = ACTIONS(1210),
    [aux_sym_clone_expression_token1] = ACTIONS(1212),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1212),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1212),
    [anon_sym_DASH_DASH] = ACTIONS(1210),
    [anon_sym_PLUS_PLUS] = ACTIONS(1210),
    [aux_sym__list_destructing_token1] = ACTIONS(1212),
    [anon_sym_LBRACK] = ACTIONS(1210),
    [anon_sym_self] = ACTIONS(1212),
    [anon_sym_parent] = ACTIONS(1212),
    [aux_sym__argument_name_token1] = ACTIONS(1212),
    [aux_sym__argument_name_token2] = ACTIONS(1212),
    [anon_sym_POUND_LBRACK] = ACTIONS(1210),
    [aux_sym_encapsed_string_token1] = ACTIONS(1210),
    [anon_sym_DQUOTE] = ACTIONS(1210),
    [aux_sym_string_token1] = ACTIONS(1210),
    [anon_sym_SQUOTE] = ACTIONS(1210),
    [anon_sym_LT_LT_LT] = ACTIONS(1210),
    [anon_sym_BQUOTE] = ACTIONS(1210),
    [anon_sym_DOLLAR] = ACTIONS(1210),
    [aux_sym_yield_expression_token1] = ACTIONS(1212),
    [aux_sym_include_expression_token1] = ACTIONS(1212),
    [aux_sym_include_once_expression_token1] = ACTIONS(1212),
    [aux_sym_require_expression_token1] = ACTIONS(1212),
    [aux_sym_require_once_expression_token1] = ACTIONS(1212),
    [sym_comment] = ACTIONS(3),
  },
  [485] = {
    [ts_builtin_sym_end] = ACTIONS(1214),
    [sym_name] = ACTIONS(1216),
    [anon_sym_QMARK_GT] = ACTIONS(1214),
    [anon_sym_SEMI] = ACTIONS(1214),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1216),
    [aux_sym_global_declaration_token1] = ACTIONS(1216),
    [aux_sym_namespace_definition_token1] = ACTIONS(1216),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1216),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1216),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1216),
    [anon_sym_BSLASH] = ACTIONS(1214),
    [anon_sym_LBRACE] = ACTIONS(1214),
    [anon_sym_RBRACE] = ACTIONS(1214),
    [aux_sym_trait_declaration_token1] = ACTIONS(1216),
    [aux_sym_interface_declaration_token1] = ACTIONS(1216),
    [aux_sym_enum_declaration_token1] = ACTIONS(1216),
    [aux_sym_enum_case_token1] = ACTIONS(1216),
    [aux_sym_class_declaration_token1] = ACTIONS(1216),
    [aux_sym_final_modifier_token1] = ACTIONS(1216),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1216),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1216),
    [sym_var_modifier] = ACTIONS(1216),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1216),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1216),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1216),
    [anon_sym_LPAREN2] = ACTIONS(1214),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1216),
    [aux_sym_cast_type_token1] = ACTIONS(1216),
    [aux_sym_echo_statement_token1] = ACTIONS(1216),
    [aux_sym_exit_statement_token1] = ACTIONS(1216),
    [anon_sym_unset] = ACTIONS(1216),
    [aux_sym_declare_statement_token1] = ACTIONS(1216),
    [aux_sym_declare_statement_token2] = ACTIONS(1216),
    [sym_float] = ACTIONS(1216),
    [aux_sym_try_statement_token1] = ACTIONS(1216),
    [aux_sym_goto_statement_token1] = ACTIONS(1216),
    [aux_sym_continue_statement_token1] = ACTIONS(1216),
    [aux_sym_break_statement_token1] = ACTIONS(1216),
    [sym_integer] = ACTIONS(1216),
    [aux_sym_return_statement_token1] = ACTIONS(1216),
    [aux_sym_throw_expression_token1] = ACTIONS(1216),
    [aux_sym_while_statement_token1] = ACTIONS(1216),
    [aux_sym_while_statement_token2] = ACTIONS(1216),
    [aux_sym_do_statement_token1] = ACTIONS(1216),
    [aux_sym_for_statement_token1] = ACTIONS(1216),
    [aux_sym_for_statement_token2] = ACTIONS(1216),
    [aux_sym_foreach_statement_token1] = ACTIONS(1216),
    [aux_sym_foreach_statement_token2] = ACTIONS(1216),
    [aux_sym_if_statement_token1] = ACTIONS(1216),
    [aux_sym_if_statement_token2] = ACTIONS(1216),
    [aux_sym_else_if_clause_token1] = ACTIONS(1216),
    [aux_sym_else_clause_token1] = ACTIONS(1216),
    [aux_sym_match_expression_token1] = ACTIONS(1216),
    [aux_sym_match_default_expression_token1] = ACTIONS(1216),
    [aux_sym_switch_statement_token1] = ACTIONS(1216),
    [aux_sym_switch_block_token1] = ACTIONS(1216),
    [anon_sym_PLUS] = ACTIONS(1216),
    [anon_sym_DASH] = ACTIONS(1216),
    [anon_sym_TILDE] = ACTIONS(1214),
    [anon_sym_BANG] = ACTIONS(1214),
    [anon_sym_AT] = ACTIONS(1214),
    [aux_sym_clone_expression_token1] = ACTIONS(1216),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1216),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1216),
    [anon_sym_DASH_DASH] = ACTIONS(1214),
    [anon_sym_PLUS_PLUS] = ACTIONS(1214),
    [aux_sym__list_destructing_token1] = ACTIONS(1216),
    [anon_sym_LBRACK] = ACTIONS(1214),
    [anon_sym_self] = ACTIONS(1216),
    [anon_sym_parent] = ACTIONS(1216),
    [aux_sym__argument_name_token1] = ACTIONS(1216),
    [aux_sym__argument_name_token2] = ACTIONS(1216),
    [anon_sym_POUND_LBRACK] = ACTIONS(1214),
    [aux_sym_encapsed_string_token1] = ACTIONS(1214),
    [anon_sym_DQUOTE] = ACTIONS(1214),
    [aux_sym_string_token1] = ACTIONS(1214),
    [anon_sym_SQUOTE] = ACTIONS(1214),
    [anon_sym_LT_LT_LT] = ACTIONS(1214),
    [anon_sym_BQUOTE] = ACTIONS(1214),
    [anon_sym_DOLLAR] = ACTIONS(1214),
    [aux_sym_yield_expression_token1] = ACTIONS(1216),
    [aux_sym_include_expression_token1] = ACTIONS(1216),
    [aux_sym_include_once_expression_token1] = ACTIONS(1216),
    [aux_sym_require_expression_token1] = ACTIONS(1216),
    [aux_sym_require_once_expression_token1] = ACTIONS(1216),
    [sym_comment] = ACTIONS(3),
  },
  [486] = {
    [ts_builtin_sym_end] = ACTIONS(1218),
    [sym_name] = ACTIONS(1220),
    [anon_sym_QMARK_GT] = ACTIONS(1218),
    [anon_sym_SEMI] = ACTIONS(1218),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1220),
    [aux_sym_global_declaration_token1] = ACTIONS(1220),
    [aux_sym_namespace_definition_token1] = ACTIONS(1220),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1220),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1220),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1220),
    [anon_sym_BSLASH] = ACTIONS(1218),
    [anon_sym_LBRACE] = ACTIONS(1218),
    [anon_sym_RBRACE] = ACTIONS(1218),
    [aux_sym_trait_declaration_token1] = ACTIONS(1220),
    [aux_sym_interface_declaration_token1] = ACTIONS(1220),
    [aux_sym_enum_declaration_token1] = ACTIONS(1220),
    [aux_sym_enum_case_token1] = ACTIONS(1220),
    [aux_sym_class_declaration_token1] = ACTIONS(1220),
    [aux_sym_final_modifier_token1] = ACTIONS(1220),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1220),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1220),
    [sym_var_modifier] = ACTIONS(1220),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1220),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1220),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1220),
    [anon_sym_LPAREN2] = ACTIONS(1218),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1220),
    [aux_sym_cast_type_token1] = ACTIONS(1220),
    [aux_sym_echo_statement_token1] = ACTIONS(1220),
    [aux_sym_exit_statement_token1] = ACTIONS(1220),
    [anon_sym_unset] = ACTIONS(1220),
    [aux_sym_declare_statement_token1] = ACTIONS(1220),
    [aux_sym_declare_statement_token2] = ACTIONS(1220),
    [sym_float] = ACTIONS(1220),
    [aux_sym_try_statement_token1] = ACTIONS(1220),
    [aux_sym_goto_statement_token1] = ACTIONS(1220),
    [aux_sym_continue_statement_token1] = ACTIONS(1220),
    [aux_sym_break_statement_token1] = ACTIONS(1220),
    [sym_integer] = ACTIONS(1220),
    [aux_sym_return_statement_token1] = ACTIONS(1220),
    [aux_sym_throw_expression_token1] = ACTIONS(1220),
    [aux_sym_while_statement_token1] = ACTIONS(1220),
    [aux_sym_while_statement_token2] = ACTIONS(1220),
    [aux_sym_do_statement_token1] = ACTIONS(1220),
    [aux_sym_for_statement_token1] = ACTIONS(1220),
    [aux_sym_for_statement_token2] = ACTIONS(1220),
    [aux_sym_foreach_statement_token1] = ACTIONS(1220),
    [aux_sym_foreach_statement_token2] = ACTIONS(1220),
    [aux_sym_if_statement_token1] = ACTIONS(1220),
    [aux_sym_if_statement_token2] = ACTIONS(1220),
    [aux_sym_else_if_clause_token1] = ACTIONS(1220),
    [aux_sym_else_clause_token1] = ACTIONS(1220),
    [aux_sym_match_expression_token1] = ACTIONS(1220),
    [aux_sym_match_default_expression_token1] = ACTIONS(1220),
    [aux_sym_switch_statement_token1] = ACTIONS(1220),
    [aux_sym_switch_block_token1] = ACTIONS(1220),
    [anon_sym_PLUS] = ACTIONS(1220),
    [anon_sym_DASH] = ACTIONS(1220),
    [anon_sym_TILDE] = ACTIONS(1218),
    [anon_sym_BANG] = ACTIONS(1218),
    [anon_sym_AT] = ACTIONS(1218),
    [aux_sym_clone_expression_token1] = ACTIONS(1220),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1220),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1220),
    [anon_sym_DASH_DASH] = ACTIONS(1218),
    [anon_sym_PLUS_PLUS] = ACTIONS(1218),
    [aux_sym__list_destructing_token1] = ACTIONS(1220),
    [anon_sym_LBRACK] = ACTIONS(1218),
    [anon_sym_self] = ACTIONS(1220),
    [anon_sym_parent] = ACTIONS(1220),
    [aux_sym__argument_name_token1] = ACTIONS(1220),
    [aux_sym__argument_name_token2] = ACTIONS(1220),
    [anon_sym_POUND_LBRACK] = ACTIONS(1218),
    [aux_sym_encapsed_string_token1] = ACTIONS(1218),
    [anon_sym_DQUOTE] = ACTIONS(1218),
    [aux_sym_string_token1] = ACTIONS(1218),
    [anon_sym_SQUOTE] = ACTIONS(1218),
    [anon_sym_LT_LT_LT] = ACTIONS(1218),
    [anon_sym_BQUOTE] = ACTIONS(1218),
    [anon_sym_DOLLAR] = ACTIONS(1218),
    [aux_sym_yield_expression_token1] = ACTIONS(1220),
    [aux_sym_include_expression_token1] = ACTIONS(1220),
    [aux_sym_include_once_expression_token1] = ACTIONS(1220),
    [aux_sym_require_expression_token1] = ACTIONS(1220),
    [aux_sym_require_once_expression_token1] = ACTIONS(1220),
    [sym_comment] = ACTIONS(3),
  },
  [487] = {
    [ts_builtin_sym_end] = ACTIONS(1222),
    [sym_name] = ACTIONS(1224),
    [anon_sym_QMARK_GT] = ACTIONS(1222),
    [anon_sym_SEMI] = ACTIONS(1222),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1224),
    [aux_sym_global_declaration_token1] = ACTIONS(1224),
    [aux_sym_namespace_definition_token1] = ACTIONS(1224),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1224),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1224),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1224),
    [anon_sym_BSLASH] = ACTIONS(1222),
    [anon_sym_LBRACE] = ACTIONS(1222),
    [anon_sym_RBRACE] = ACTIONS(1222),
    [aux_sym_trait_declaration_token1] = ACTIONS(1224),
    [aux_sym_interface_declaration_token1] = ACTIONS(1224),
    [aux_sym_enum_declaration_token1] = ACTIONS(1224),
    [aux_sym_enum_case_token1] = ACTIONS(1224),
    [aux_sym_class_declaration_token1] = ACTIONS(1224),
    [aux_sym_final_modifier_token1] = ACTIONS(1224),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1224),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1224),
    [sym_var_modifier] = ACTIONS(1224),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1224),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1224),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1224),
    [anon_sym_LPAREN2] = ACTIONS(1222),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1224),
    [aux_sym_cast_type_token1] = ACTIONS(1224),
    [aux_sym_echo_statement_token1] = ACTIONS(1224),
    [aux_sym_exit_statement_token1] = ACTIONS(1224),
    [anon_sym_unset] = ACTIONS(1224),
    [aux_sym_declare_statement_token1] = ACTIONS(1224),
    [aux_sym_declare_statement_token2] = ACTIONS(1224),
    [sym_float] = ACTIONS(1224),
    [aux_sym_try_statement_token1] = ACTIONS(1224),
    [aux_sym_goto_statement_token1] = ACTIONS(1224),
    [aux_sym_continue_statement_token1] = ACTIONS(1224),
    [aux_sym_break_statement_token1] = ACTIONS(1224),
    [sym_integer] = ACTIONS(1224),
    [aux_sym_return_statement_token1] = ACTIONS(1224),
    [aux_sym_throw_expression_token1] = ACTIONS(1224),
    [aux_sym_while_statement_token1] = ACTIONS(1224),
    [aux_sym_while_statement_token2] = ACTIONS(1224),
    [aux_sym_do_statement_token1] = ACTIONS(1224),
    [aux_sym_for_statement_token1] = ACTIONS(1224),
    [aux_sym_for_statement_token2] = ACTIONS(1224),
    [aux_sym_foreach_statement_token1] = ACTIONS(1224),
    [aux_sym_foreach_statement_token2] = ACTIONS(1224),
    [aux_sym_if_statement_token1] = ACTIONS(1224),
    [aux_sym_if_statement_token2] = ACTIONS(1224),
    [aux_sym_else_if_clause_token1] = ACTIONS(1224),
    [aux_sym_else_clause_token1] = ACTIONS(1224),
    [aux_sym_match_expression_token1] = ACTIONS(1224),
    [aux_sym_match_default_expression_token1] = ACTIONS(1224),
    [aux_sym_switch_statement_token1] = ACTIONS(1224),
    [aux_sym_switch_block_token1] = ACTIONS(1224),
    [anon_sym_PLUS] = ACTIONS(1224),
    [anon_sym_DASH] = ACTIONS(1224),
    [anon_sym_TILDE] = ACTIONS(1222),
    [anon_sym_BANG] = ACTIONS(1222),
    [anon_sym_AT] = ACTIONS(1222),
    [aux_sym_clone_expression_token1] = ACTIONS(1224),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1224),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1224),
    [anon_sym_DASH_DASH] = ACTIONS(1222),
    [anon_sym_PLUS_PLUS] = ACTIONS(1222),
    [aux_sym__list_destructing_token1] = ACTIONS(1224),
    [anon_sym_LBRACK] = ACTIONS(1222),
    [anon_sym_self] = ACTIONS(1224),
    [anon_sym_parent] = ACTIONS(1224),
    [aux_sym__argument_name_token1] = ACTIONS(1224),
    [aux_sym__argument_name_token2] = ACTIONS(1224),
    [anon_sym_POUND_LBRACK] = ACTIONS(1222),
    [aux_sym_encapsed_string_token1] = ACTIONS(1222),
    [anon_sym_DQUOTE] = ACTIONS(1222),
    [aux_sym_string_token1] = ACTIONS(1222),
    [anon_sym_SQUOTE] = ACTIONS(1222),
    [anon_sym_LT_LT_LT] = ACTIONS(1222),
    [anon_sym_BQUOTE] = ACTIONS(1222),
    [anon_sym_DOLLAR] = ACTIONS(1222),
    [aux_sym_yield_expression_token1] = ACTIONS(1224),
    [aux_sym_include_expression_token1] = ACTIONS(1224),
    [aux_sym_include_once_expression_token1] = ACTIONS(1224),
    [aux_sym_require_expression_token1] = ACTIONS(1224),
    [aux_sym_require_once_expression_token1] = ACTIONS(1224),
    [sym_comment] = ACTIONS(3),
  },
  [488] = {
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_name] = ACTIONS(1228),
    [anon_sym_QMARK_GT] = ACTIONS(1226),
    [anon_sym_SEMI] = ACTIONS(1226),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1228),
    [aux_sym_global_declaration_token1] = ACTIONS(1228),
    [aux_sym_namespace_definition_token1] = ACTIONS(1228),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1228),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1228),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1228),
    [anon_sym_BSLASH] = ACTIONS(1226),
    [anon_sym_LBRACE] = ACTIONS(1226),
    [anon_sym_RBRACE] = ACTIONS(1226),
    [aux_sym_trait_declaration_token1] = ACTIONS(1228),
    [aux_sym_interface_declaration_token1] = ACTIONS(1228),
    [aux_sym_enum_declaration_token1] = ACTIONS(1228),
    [aux_sym_enum_case_token1] = ACTIONS(1228),
    [aux_sym_class_declaration_token1] = ACTIONS(1228),
    [aux_sym_final_modifier_token1] = ACTIONS(1228),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1228),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1228),
    [sym_var_modifier] = ACTIONS(1228),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1228),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1228),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1228),
    [anon_sym_LPAREN2] = ACTIONS(1226),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1228),
    [aux_sym_cast_type_token1] = ACTIONS(1228),
    [aux_sym_echo_statement_token1] = ACTIONS(1228),
    [aux_sym_exit_statement_token1] = ACTIONS(1228),
    [anon_sym_unset] = ACTIONS(1228),
    [aux_sym_declare_statement_token1] = ACTIONS(1228),
    [aux_sym_declare_statement_token2] = ACTIONS(1228),
    [sym_float] = ACTIONS(1228),
    [aux_sym_try_statement_token1] = ACTIONS(1228),
    [aux_sym_goto_statement_token1] = ACTIONS(1228),
    [aux_sym_continue_statement_token1] = ACTIONS(1228),
    [aux_sym_break_statement_token1] = ACTIONS(1228),
    [sym_integer] = ACTIONS(1228),
    [aux_sym_return_statement_token1] = ACTIONS(1228),
    [aux_sym_throw_expression_token1] = ACTIONS(1228),
    [aux_sym_while_statement_token1] = ACTIONS(1228),
    [aux_sym_while_statement_token2] = ACTIONS(1228),
    [aux_sym_do_statement_token1] = ACTIONS(1228),
    [aux_sym_for_statement_token1] = ACTIONS(1228),
    [aux_sym_for_statement_token2] = ACTIONS(1228),
    [aux_sym_foreach_statement_token1] = ACTIONS(1228),
    [aux_sym_foreach_statement_token2] = ACTIONS(1228),
    [aux_sym_if_statement_token1] = ACTIONS(1228),
    [aux_sym_if_statement_token2] = ACTIONS(1228),
    [aux_sym_else_if_clause_token1] = ACTIONS(1228),
    [aux_sym_else_clause_token1] = ACTIONS(1228),
    [aux_sym_match_expression_token1] = ACTIONS(1228),
    [aux_sym_match_default_expression_token1] = ACTIONS(1228),
    [aux_sym_switch_statement_token1] = ACTIONS(1228),
    [aux_sym_switch_block_token1] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1226),
    [anon_sym_BANG] = ACTIONS(1226),
    [anon_sym_AT] = ACTIONS(1226),
    [aux_sym_clone_expression_token1] = ACTIONS(1228),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1228),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1228),
    [anon_sym_DASH_DASH] = ACTIONS(1226),
    [anon_sym_PLUS_PLUS] = ACTIONS(1226),
    [aux_sym__list_destructing_token1] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1226),
    [anon_sym_self] = ACTIONS(1228),
    [anon_sym_parent] = ACTIONS(1228),
    [aux_sym__argument_name_token1] = ACTIONS(1228),
    [aux_sym__argument_name_token2] = ACTIONS(1228),
    [anon_sym_POUND_LBRACK] = ACTIONS(1226),
    [aux_sym_encapsed_string_token1] = ACTIONS(1226),
    [anon_sym_DQUOTE] = ACTIONS(1226),
    [aux_sym_string_token1] = ACTIONS(1226),
    [anon_sym_SQUOTE] = ACTIONS(1226),
    [anon_sym_LT_LT_LT] = ACTIONS(1226),
    [anon_sym_BQUOTE] = ACTIONS(1226),
    [anon_sym_DOLLAR] = ACTIONS(1226),
    [aux_sym_yield_expression_token1] = ACTIONS(1228),
    [aux_sym_include_expression_token1] = ACTIONS(1228),
    [aux_sym_include_once_expression_token1] = ACTIONS(1228),
    [aux_sym_require_expression_token1] = ACTIONS(1228),
    [aux_sym_require_once_expression_token1] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
  },
  [489] = {
    [ts_builtin_sym_end] = ACTIONS(1230),
    [sym_name] = ACTIONS(1232),
    [anon_sym_QMARK_GT] = ACTIONS(1230),
    [anon_sym_SEMI] = ACTIONS(1230),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1232),
    [aux_sym_global_declaration_token1] = ACTIONS(1232),
    [aux_sym_namespace_definition_token1] = ACTIONS(1232),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1232),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1232),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1232),
    [anon_sym_BSLASH] = ACTIONS(1230),
    [anon_sym_LBRACE] = ACTIONS(1230),
    [anon_sym_RBRACE] = ACTIONS(1230),
    [aux_sym_trait_declaration_token1] = ACTIONS(1232),
    [aux_sym_interface_declaration_token1] = ACTIONS(1232),
    [aux_sym_enum_declaration_token1] = ACTIONS(1232),
    [aux_sym_enum_case_token1] = ACTIONS(1232),
    [aux_sym_class_declaration_token1] = ACTIONS(1232),
    [aux_sym_final_modifier_token1] = ACTIONS(1232),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1232),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1232),
    [sym_var_modifier] = ACTIONS(1232),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1232),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1232),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1232),
    [anon_sym_LPAREN2] = ACTIONS(1230),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1232),
    [aux_sym_cast_type_token1] = ACTIONS(1232),
    [aux_sym_echo_statement_token1] = ACTIONS(1232),
    [aux_sym_exit_statement_token1] = ACTIONS(1232),
    [anon_sym_unset] = ACTIONS(1232),
    [aux_sym_declare_statement_token1] = ACTIONS(1232),
    [aux_sym_declare_statement_token2] = ACTIONS(1232),
    [sym_float] = ACTIONS(1232),
    [aux_sym_try_statement_token1] = ACTIONS(1232),
    [aux_sym_goto_statement_token1] = ACTIONS(1232),
    [aux_sym_continue_statement_token1] = ACTIONS(1232),
    [aux_sym_break_statement_token1] = ACTIONS(1232),
    [sym_integer] = ACTIONS(1232),
    [aux_sym_return_statement_token1] = ACTIONS(1232),
    [aux_sym_throw_expression_token1] = ACTIONS(1232),
    [aux_sym_while_statement_token1] = ACTIONS(1232),
    [aux_sym_while_statement_token2] = ACTIONS(1232),
    [aux_sym_do_statement_token1] = ACTIONS(1232),
    [aux_sym_for_statement_token1] = ACTIONS(1232),
    [aux_sym_for_statement_token2] = ACTIONS(1232),
    [aux_sym_foreach_statement_token1] = ACTIONS(1232),
    [aux_sym_foreach_statement_token2] = ACTIONS(1232),
    [aux_sym_if_statement_token1] = ACTIONS(1232),
    [aux_sym_if_statement_token2] = ACTIONS(1232),
    [aux_sym_else_if_clause_token1] = ACTIONS(1232),
    [aux_sym_else_clause_token1] = ACTIONS(1232),
    [aux_sym_match_expression_token1] = ACTIONS(1232),
    [aux_sym_match_default_expression_token1] = ACTIONS(1232),
    [aux_sym_switch_statement_token1] = ACTIONS(1232),
    [aux_sym_switch_block_token1] = ACTIONS(1232),
    [anon_sym_PLUS] = ACTIONS(1232),
    [anon_sym_DASH] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1230),
    [anon_sym_BANG] = ACTIONS(1230),
    [anon_sym_AT] = ACTIONS(1230),
    [aux_sym_clone_expression_token1] = ACTIONS(1232),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1232),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1232),
    [anon_sym_DASH_DASH] = ACTIONS(1230),
    [anon_sym_PLUS_PLUS] = ACTIONS(1230),
    [aux_sym__list_destructing_token1] = ACTIONS(1232),
    [anon_sym_LBRACK] = ACTIONS(1230),
    [anon_sym_self] = ACTIONS(1232),
    [anon_sym_parent] = ACTIONS(1232),
    [aux_sym__argument_name_token1] = ACTIONS(1232),
    [aux_sym__argument_name_token2] = ACTIONS(1232),
    [anon_sym_POUND_LBRACK] = ACTIONS(1230),
    [aux_sym_encapsed_string_token1] = ACTIONS(1230),
    [anon_sym_DQUOTE] = ACTIONS(1230),
    [aux_sym_string_token1] = ACTIONS(1230),
    [anon_sym_SQUOTE] = ACTIONS(1230),
    [anon_sym_LT_LT_LT] = ACTIONS(1230),
    [anon_sym_BQUOTE] = ACTIONS(1230),
    [anon_sym_DOLLAR] = ACTIONS(1230),
    [aux_sym_yield_expression_token1] = ACTIONS(1232),
    [aux_sym_include_expression_token1] = ACTIONS(1232),
    [aux_sym_include_once_expression_token1] = ACTIONS(1232),
    [aux_sym_require_expression_token1] = ACTIONS(1232),
    [aux_sym_require_once_expression_token1] = ACTIONS(1232),
    [sym_comment] = ACTIONS(3),
  },
  [490] = {
    [ts_builtin_sym_end] = ACTIONS(1234),
    [sym_name] = ACTIONS(1236),
    [anon_sym_QMARK_GT] = ACTIONS(1234),
    [anon_sym_SEMI] = ACTIONS(1234),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1236),
    [aux_sym_global_declaration_token1] = ACTIONS(1236),
    [aux_sym_namespace_definition_token1] = ACTIONS(1236),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1236),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1236),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1236),
    [anon_sym_BSLASH] = ACTIONS(1234),
    [anon_sym_LBRACE] = ACTIONS(1234),
    [anon_sym_RBRACE] = ACTIONS(1234),
    [aux_sym_trait_declaration_token1] = ACTIONS(1236),
    [aux_sym_interface_declaration_token1] = ACTIONS(1236),
    [aux_sym_enum_declaration_token1] = ACTIONS(1236),
    [aux_sym_enum_case_token1] = ACTIONS(1236),
    [aux_sym_class_declaration_token1] = ACTIONS(1236),
    [aux_sym_final_modifier_token1] = ACTIONS(1236),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1236),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1236),
    [sym_var_modifier] = ACTIONS(1236),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1236),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1236),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1236),
    [anon_sym_LPAREN2] = ACTIONS(1234),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1236),
    [aux_sym_cast_type_token1] = ACTIONS(1236),
    [aux_sym_echo_statement_token1] = ACTIONS(1236),
    [aux_sym_exit_statement_token1] = ACTIONS(1236),
    [anon_sym_unset] = ACTIONS(1236),
    [aux_sym_declare_statement_token1] = ACTIONS(1236),
    [aux_sym_declare_statement_token2] = ACTIONS(1236),
    [sym_float] = ACTIONS(1236),
    [aux_sym_try_statement_token1] = ACTIONS(1236),
    [aux_sym_goto_statement_token1] = ACTIONS(1236),
    [aux_sym_continue_statement_token1] = ACTIONS(1236),
    [aux_sym_break_statement_token1] = ACTIONS(1236),
    [sym_integer] = ACTIONS(1236),
    [aux_sym_return_statement_token1] = ACTIONS(1236),
    [aux_sym_throw_expression_token1] = ACTIONS(1236),
    [aux_sym_while_statement_token1] = ACTIONS(1236),
    [aux_sym_while_statement_token2] = ACTIONS(1236),
    [aux_sym_do_statement_token1] = ACTIONS(1236),
    [aux_sym_for_statement_token1] = ACTIONS(1236),
    [aux_sym_for_statement_token2] = ACTIONS(1236),
    [aux_sym_foreach_statement_token1] = ACTIONS(1236),
    [aux_sym_foreach_statement_token2] = ACTIONS(1236),
    [aux_sym_if_statement_token1] = ACTIONS(1236),
    [aux_sym_if_statement_token2] = ACTIONS(1236),
    [aux_sym_else_if_clause_token1] = ACTIONS(1236),
    [aux_sym_else_clause_token1] = ACTIONS(1236),
    [aux_sym_match_expression_token1] = ACTIONS(1236),
    [aux_sym_match_default_expression_token1] = ACTIONS(1236),
    [aux_sym_switch_statement_token1] = ACTIONS(1236),
    [aux_sym_switch_block_token1] = ACTIONS(1236),
    [anon_sym_PLUS] = ACTIONS(1236),
    [anon_sym_DASH] = ACTIONS(1236),
    [anon_sym_TILDE] = ACTIONS(1234),
    [anon_sym_BANG] = ACTIONS(1234),
    [anon_sym_AT] = ACTIONS(1234),
    [aux_sym_clone_expression_token1] = ACTIONS(1236),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1236),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1236),
    [anon_sym_DASH_DASH] = ACTIONS(1234),
    [anon_sym_PLUS_PLUS] = ACTIONS(1234),
    [aux_sym__list_destructing_token1] = ACTIONS(1236),
    [anon_sym_LBRACK] = ACTIONS(1234),
    [anon_sym_self] = ACTIONS(1236),
    [anon_sym_parent] = ACTIONS(1236),
    [aux_sym__argument_name_token1] = ACTIONS(1236),
    [aux_sym__argument_name_token2] = ACTIONS(1236),
    [anon_sym_POUND_LBRACK] = ACTIONS(1234),
    [aux_sym_encapsed_string_token1] = ACTIONS(1234),
    [anon_sym_DQUOTE] = ACTIONS(1234),
    [aux_sym_string_token1] = ACTIONS(1234),
    [anon_sym_SQUOTE] = ACTIONS(1234),
    [anon_sym_LT_LT_LT] = ACTIONS(1234),
    [anon_sym_BQUOTE] = ACTIONS(1234),
    [anon_sym_DOLLAR] = ACTIONS(1234),
    [aux_sym_yield_expression_token1] = ACTIONS(1236),
    [aux_sym_include_expression_token1] = ACTIONS(1236),
    [aux_sym_include_once_expression_token1] = ACTIONS(1236),
    [aux_sym_require_expression_token1] = ACTIONS(1236),
    [aux_sym_require_once_expression_token1] = ACTIONS(1236),
    [sym_comment] = ACTIONS(3),
  },
  [491] = {
    [ts_builtin_sym_end] = ACTIONS(1238),
    [sym_name] = ACTIONS(1240),
    [anon_sym_QMARK_GT] = ACTIONS(1238),
    [anon_sym_SEMI] = ACTIONS(1238),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1240),
    [aux_sym_global_declaration_token1] = ACTIONS(1240),
    [aux_sym_namespace_definition_token1] = ACTIONS(1240),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1240),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1240),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1240),
    [anon_sym_BSLASH] = ACTIONS(1238),
    [anon_sym_LBRACE] = ACTIONS(1238),
    [anon_sym_RBRACE] = ACTIONS(1238),
    [aux_sym_trait_declaration_token1] = ACTIONS(1240),
    [aux_sym_interface_declaration_token1] = ACTIONS(1240),
    [aux_sym_enum_declaration_token1] = ACTIONS(1240),
    [aux_sym_enum_case_token1] = ACTIONS(1240),
    [aux_sym_class_declaration_token1] = ACTIONS(1240),
    [aux_sym_final_modifier_token1] = ACTIONS(1240),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1240),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1240),
    [sym_var_modifier] = ACTIONS(1240),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1240),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1240),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1240),
    [anon_sym_LPAREN2] = ACTIONS(1238),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1240),
    [aux_sym_cast_type_token1] = ACTIONS(1240),
    [aux_sym_echo_statement_token1] = ACTIONS(1240),
    [aux_sym_exit_statement_token1] = ACTIONS(1240),
    [anon_sym_unset] = ACTIONS(1240),
    [aux_sym_declare_statement_token1] = ACTIONS(1240),
    [aux_sym_declare_statement_token2] = ACTIONS(1240),
    [sym_float] = ACTIONS(1240),
    [aux_sym_try_statement_token1] = ACTIONS(1240),
    [aux_sym_goto_statement_token1] = ACTIONS(1240),
    [aux_sym_continue_statement_token1] = ACTIONS(1240),
    [aux_sym_break_statement_token1] = ACTIONS(1240),
    [sym_integer] = ACTIONS(1240),
    [aux_sym_return_statement_token1] = ACTIONS(1240),
    [aux_sym_throw_expression_token1] = ACTIONS(1240),
    [aux_sym_while_statement_token1] = ACTIONS(1240),
    [aux_sym_while_statement_token2] = ACTIONS(1240),
    [aux_sym_do_statement_token1] = ACTIONS(1240),
    [aux_sym_for_statement_token1] = ACTIONS(1240),
    [aux_sym_for_statement_token2] = ACTIONS(1240),
    [aux_sym_foreach_statement_token1] = ACTIONS(1240),
    [aux_sym_foreach_statement_token2] = ACTIONS(1240),
    [aux_sym_if_statement_token1] = ACTIONS(1240),
    [aux_sym_if_statement_token2] = ACTIONS(1240),
    [aux_sym_else_if_clause_token1] = ACTIONS(1240),
    [aux_sym_else_clause_token1] = ACTIONS(1240),
    [aux_sym_match_expression_token1] = ACTIONS(1240),
    [aux_sym_match_default_expression_token1] = ACTIONS(1240),
    [aux_sym_switch_statement_token1] = ACTIONS(1240),
    [aux_sym_switch_block_token1] = ACTIONS(1240),
    [anon_sym_PLUS] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1240),
    [anon_sym_TILDE] = ACTIONS(1238),
    [anon_sym_BANG] = ACTIONS(1238),
    [anon_sym_AT] = ACTIONS(1238),
    [aux_sym_clone_expression_token1] = ACTIONS(1240),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1240),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1240),
    [anon_sym_DASH_DASH] = ACTIONS(1238),
    [anon_sym_PLUS_PLUS] = ACTIONS(1238),
    [aux_sym__list_destructing_token1] = ACTIONS(1240),
    [anon_sym_LBRACK] = ACTIONS(1238),
    [anon_sym_self] = ACTIONS(1240),
    [anon_sym_parent] = ACTIONS(1240),
    [aux_sym__argument_name_token1] = ACTIONS(1240),
    [aux_sym__argument_name_token2] = ACTIONS(1240),
    [anon_sym_POUND_LBRACK] = ACTIONS(1238),
    [aux_sym_encapsed_string_token1] = ACTIONS(1238),
    [anon_sym_DQUOTE] = ACTIONS(1238),
    [aux_sym_string_token1] = ACTIONS(1238),
    [anon_sym_SQUOTE] = ACTIONS(1238),
    [anon_sym_LT_LT_LT] = ACTIONS(1238),
    [anon_sym_BQUOTE] = ACTIONS(1238),
    [anon_sym_DOLLAR] = ACTIONS(1238),
    [aux_sym_yield_expression_token1] = ACTIONS(1240),
    [aux_sym_include_expression_token1] = ACTIONS(1240),
    [aux_sym_include_once_expression_token1] = ACTIONS(1240),
    [aux_sym_require_expression_token1] = ACTIONS(1240),
    [aux_sym_require_once_expression_token1] = ACTIONS(1240),
    [sym_comment] = ACTIONS(3),
  },
  [492] = {
    [ts_builtin_sym_end] = ACTIONS(1242),
    [sym_name] = ACTIONS(1244),
    [anon_sym_QMARK_GT] = ACTIONS(1242),
    [anon_sym_SEMI] = ACTIONS(1242),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1244),
    [aux_sym_global_declaration_token1] = ACTIONS(1244),
    [aux_sym_namespace_definition_token1] = ACTIONS(1244),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1244),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1244),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1244),
    [anon_sym_BSLASH] = ACTIONS(1242),
    [anon_sym_LBRACE] = ACTIONS(1242),
    [anon_sym_RBRACE] = ACTIONS(1242),
    [aux_sym_trait_declaration_token1] = ACTIONS(1244),
    [aux_sym_interface_declaration_token1] = ACTIONS(1244),
    [aux_sym_enum_declaration_token1] = ACTIONS(1244),
    [aux_sym_enum_case_token1] = ACTIONS(1244),
    [aux_sym_class_declaration_token1] = ACTIONS(1244),
    [aux_sym_final_modifier_token1] = ACTIONS(1244),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1244),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1244),
    [sym_var_modifier] = ACTIONS(1244),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1244),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1244),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1244),
    [anon_sym_LPAREN2] = ACTIONS(1242),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1244),
    [aux_sym_cast_type_token1] = ACTIONS(1244),
    [aux_sym_echo_statement_token1] = ACTIONS(1244),
    [aux_sym_exit_statement_token1] = ACTIONS(1244),
    [anon_sym_unset] = ACTIONS(1244),
    [aux_sym_declare_statement_token1] = ACTIONS(1244),
    [aux_sym_declare_statement_token2] = ACTIONS(1244),
    [sym_float] = ACTIONS(1244),
    [aux_sym_try_statement_token1] = ACTIONS(1244),
    [aux_sym_goto_statement_token1] = ACTIONS(1244),
    [aux_sym_continue_statement_token1] = ACTIONS(1244),
    [aux_sym_break_statement_token1] = ACTIONS(1244),
    [sym_integer] = ACTIONS(1244),
    [aux_sym_return_statement_token1] = ACTIONS(1244),
    [aux_sym_throw_expression_token1] = ACTIONS(1244),
    [aux_sym_while_statement_token1] = ACTIONS(1244),
    [aux_sym_while_statement_token2] = ACTIONS(1244),
    [aux_sym_do_statement_token1] = ACTIONS(1244),
    [aux_sym_for_statement_token1] = ACTIONS(1244),
    [aux_sym_for_statement_token2] = ACTIONS(1244),
    [aux_sym_foreach_statement_token1] = ACTIONS(1244),
    [aux_sym_foreach_statement_token2] = ACTIONS(1244),
    [aux_sym_if_statement_token1] = ACTIONS(1244),
    [aux_sym_if_statement_token2] = ACTIONS(1244),
    [aux_sym_else_if_clause_token1] = ACTIONS(1244),
    [aux_sym_else_clause_token1] = ACTIONS(1244),
    [aux_sym_match_expression_token1] = ACTIONS(1244),
    [aux_sym_match_default_expression_token1] = ACTIONS(1244),
    [aux_sym_switch_statement_token1] = ACTIONS(1244),
    [aux_sym_switch_block_token1] = ACTIONS(1244),
    [anon_sym_PLUS] = ACTIONS(1244),
    [anon_sym_DASH] = ACTIONS(1244),
    [anon_sym_TILDE] = ACTIONS(1242),
    [anon_sym_BANG] = ACTIONS(1242),
    [anon_sym_AT] = ACTIONS(1242),
    [aux_sym_clone_expression_token1] = ACTIONS(1244),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1244),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1244),
    [anon_sym_DASH_DASH] = ACTIONS(1242),
    [anon_sym_PLUS_PLUS] = ACTIONS(1242),
    [aux_sym__list_destructing_token1] = ACTIONS(1244),
    [anon_sym_LBRACK] = ACTIONS(1242),
    [anon_sym_self] = ACTIONS(1244),
    [anon_sym_parent] = ACTIONS(1244),
    [aux_sym__argument_name_token1] = ACTIONS(1244),
    [aux_sym__argument_name_token2] = ACTIONS(1244),
    [anon_sym_POUND_LBRACK] = ACTIONS(1242),
    [aux_sym_encapsed_string_token1] = ACTIONS(1242),
    [anon_sym_DQUOTE] = ACTIONS(1242),
    [aux_sym_string_token1] = ACTIONS(1242),
    [anon_sym_SQUOTE] = ACTIONS(1242),
    [anon_sym_LT_LT_LT] = ACTIONS(1242),
    [anon_sym_BQUOTE] = ACTIONS(1242),
    [anon_sym_DOLLAR] = ACTIONS(1242),
    [aux_sym_yield_expression_token1] = ACTIONS(1244),
    [aux_sym_include_expression_token1] = ACTIONS(1244),
    [aux_sym_include_once_expression_token1] = ACTIONS(1244),
    [aux_sym_require_expression_token1] = ACTIONS(1244),
    [aux_sym_require_once_expression_token1] = ACTIONS(1244),
    [sym_comment] = ACTIONS(3),
  },
  [493] = {
    [ts_builtin_sym_end] = ACTIONS(1246),
    [sym_name] = ACTIONS(1248),
    [anon_sym_QMARK_GT] = ACTIONS(1246),
    [anon_sym_SEMI] = ACTIONS(1246),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1248),
    [aux_sym_global_declaration_token1] = ACTIONS(1248),
    [aux_sym_namespace_definition_token1] = ACTIONS(1248),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1248),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1248),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1248),
    [anon_sym_BSLASH] = ACTIONS(1246),
    [anon_sym_LBRACE] = ACTIONS(1246),
    [anon_sym_RBRACE] = ACTIONS(1246),
    [aux_sym_trait_declaration_token1] = ACTIONS(1248),
    [aux_sym_interface_declaration_token1] = ACTIONS(1248),
    [aux_sym_enum_declaration_token1] = ACTIONS(1248),
    [aux_sym_enum_case_token1] = ACTIONS(1248),
    [aux_sym_class_declaration_token1] = ACTIONS(1248),
    [aux_sym_final_modifier_token1] = ACTIONS(1248),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1248),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1248),
    [sym_var_modifier] = ACTIONS(1248),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1248),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1248),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1248),
    [anon_sym_LPAREN2] = ACTIONS(1246),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1248),
    [aux_sym_cast_type_token1] = ACTIONS(1248),
    [aux_sym_echo_statement_token1] = ACTIONS(1248),
    [aux_sym_exit_statement_token1] = ACTIONS(1248),
    [anon_sym_unset] = ACTIONS(1248),
    [aux_sym_declare_statement_token1] = ACTIONS(1248),
    [aux_sym_declare_statement_token2] = ACTIONS(1248),
    [sym_float] = ACTIONS(1248),
    [aux_sym_try_statement_token1] = ACTIONS(1248),
    [aux_sym_goto_statement_token1] = ACTIONS(1248),
    [aux_sym_continue_statement_token1] = ACTIONS(1248),
    [aux_sym_break_statement_token1] = ACTIONS(1248),
    [sym_integer] = ACTIONS(1248),
    [aux_sym_return_statement_token1] = ACTIONS(1248),
    [aux_sym_throw_expression_token1] = ACTIONS(1248),
    [aux_sym_while_statement_token1] = ACTIONS(1248),
    [aux_sym_while_statement_token2] = ACTIONS(1248),
    [aux_sym_do_statement_token1] = ACTIONS(1248),
    [aux_sym_for_statement_token1] = ACTIONS(1248),
    [aux_sym_for_statement_token2] = ACTIONS(1248),
    [aux_sym_foreach_statement_token1] = ACTIONS(1248),
    [aux_sym_foreach_statement_token2] = ACTIONS(1248),
    [aux_sym_if_statement_token1] = ACTIONS(1248),
    [aux_sym_if_statement_token2] = ACTIONS(1248),
    [aux_sym_else_if_clause_token1] = ACTIONS(1248),
    [aux_sym_else_clause_token1] = ACTIONS(1248),
    [aux_sym_match_expression_token1] = ACTIONS(1248),
    [aux_sym_match_default_expression_token1] = ACTIONS(1248),
    [aux_sym_switch_statement_token1] = ACTIONS(1248),
    [aux_sym_switch_block_token1] = ACTIONS(1248),
    [anon_sym_PLUS] = ACTIONS(1248),
    [anon_sym_DASH] = ACTIONS(1248),
    [anon_sym_TILDE] = ACTIONS(1246),
    [anon_sym_BANG] = ACTIONS(1246),
    [anon_sym_AT] = ACTIONS(1246),
    [aux_sym_clone_expression_token1] = ACTIONS(1248),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1248),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1248),
    [anon_sym_DASH_DASH] = ACTIONS(1246),
    [anon_sym_PLUS_PLUS] = ACTIONS(1246),
    [aux_sym__list_destructing_token1] = ACTIONS(1248),
    [anon_sym_LBRACK] = ACTIONS(1246),
    [anon_sym_self] = ACTIONS(1248),
    [anon_sym_parent] = ACTIONS(1248),
    [aux_sym__argument_name_token1] = ACTIONS(1248),
    [aux_sym__argument_name_token2] = ACTIONS(1248),
    [anon_sym_POUND_LBRACK] = ACTIONS(1246),
    [aux_sym_encapsed_string_token1] = ACTIONS(1246),
    [anon_sym_DQUOTE] = ACTIONS(1246),
    [aux_sym_string_token1] = ACTIONS(1246),
    [anon_sym_SQUOTE] = ACTIONS(1246),
    [anon_sym_LT_LT_LT] = ACTIONS(1246),
    [anon_sym_BQUOTE] = ACTIONS(1246),
    [anon_sym_DOLLAR] = ACTIONS(1246),
    [aux_sym_yield_expression_token1] = ACTIONS(1248),
    [aux_sym_include_expression_token1] = ACTIONS(1248),
    [aux_sym_include_once_expression_token1] = ACTIONS(1248),
    [aux_sym_require_expression_token1] = ACTIONS(1248),
    [aux_sym_require_once_expression_token1] = ACTIONS(1248),
    [sym_comment] = ACTIONS(3),
  },
  [494] = {
    [ts_builtin_sym_end] = ACTIONS(1250),
    [sym_name] = ACTIONS(1252),
    [anon_sym_QMARK_GT] = ACTIONS(1250),
    [anon_sym_SEMI] = ACTIONS(1250),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1252),
    [aux_sym_global_declaration_token1] = ACTIONS(1252),
    [aux_sym_namespace_definition_token1] = ACTIONS(1252),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1252),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1252),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1252),
    [anon_sym_BSLASH] = ACTIONS(1250),
    [anon_sym_LBRACE] = ACTIONS(1250),
    [anon_sym_RBRACE] = ACTIONS(1250),
    [aux_sym_trait_declaration_token1] = ACTIONS(1252),
    [aux_sym_interface_declaration_token1] = ACTIONS(1252),
    [aux_sym_enum_declaration_token1] = ACTIONS(1252),
    [aux_sym_enum_case_token1] = ACTIONS(1252),
    [aux_sym_class_declaration_token1] = ACTIONS(1252),
    [aux_sym_final_modifier_token1] = ACTIONS(1252),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1252),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1252),
    [sym_var_modifier] = ACTIONS(1252),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1252),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1252),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1252),
    [anon_sym_LPAREN2] = ACTIONS(1250),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1252),
    [aux_sym_cast_type_token1] = ACTIONS(1252),
    [aux_sym_echo_statement_token1] = ACTIONS(1252),
    [aux_sym_exit_statement_token1] = ACTIONS(1252),
    [anon_sym_unset] = ACTIONS(1252),
    [aux_sym_declare_statement_token1] = ACTIONS(1252),
    [aux_sym_declare_statement_token2] = ACTIONS(1252),
    [sym_float] = ACTIONS(1252),
    [aux_sym_try_statement_token1] = ACTIONS(1252),
    [aux_sym_goto_statement_token1] = ACTIONS(1252),
    [aux_sym_continue_statement_token1] = ACTIONS(1252),
    [aux_sym_break_statement_token1] = ACTIONS(1252),
    [sym_integer] = ACTIONS(1252),
    [aux_sym_return_statement_token1] = ACTIONS(1252),
    [aux_sym_throw_expression_token1] = ACTIONS(1252),
    [aux_sym_while_statement_token1] = ACTIONS(1252),
    [aux_sym_while_statement_token2] = ACTIONS(1252),
    [aux_sym_do_statement_token1] = ACTIONS(1252),
    [aux_sym_for_statement_token1] = ACTIONS(1252),
    [aux_sym_for_statement_token2] = ACTIONS(1252),
    [aux_sym_foreach_statement_token1] = ACTIONS(1252),
    [aux_sym_foreach_statement_token2] = ACTIONS(1252),
    [aux_sym_if_statement_token1] = ACTIONS(1252),
    [aux_sym_if_statement_token2] = ACTIONS(1252),
    [aux_sym_else_if_clause_token1] = ACTIONS(1252),
    [aux_sym_else_clause_token1] = ACTIONS(1252),
    [aux_sym_match_expression_token1] = ACTIONS(1252),
    [aux_sym_match_default_expression_token1] = ACTIONS(1252),
    [aux_sym_switch_statement_token1] = ACTIONS(1252),
    [aux_sym_switch_block_token1] = ACTIONS(1252),
    [anon_sym_PLUS] = ACTIONS(1252),
    [anon_sym_DASH] = ACTIONS(1252),
    [anon_sym_TILDE] = ACTIONS(1250),
    [anon_sym_BANG] = ACTIONS(1250),
    [anon_sym_AT] = ACTIONS(1250),
    [aux_sym_clone_expression_token1] = ACTIONS(1252),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1252),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1250),
    [anon_sym_PLUS_PLUS] = ACTIONS(1250),
    [aux_sym__list_destructing_token1] = ACTIONS(1252),
    [anon_sym_LBRACK] = ACTIONS(1250),
    [anon_sym_self] = ACTIONS(1252),
    [anon_sym_parent] = ACTIONS(1252),
    [aux_sym__argument_name_token1] = ACTIONS(1252),
    [aux_sym__argument_name_token2] = ACTIONS(1252),
    [anon_sym_POUND_LBRACK] = ACTIONS(1250),
    [aux_sym_encapsed_string_token1] = ACTIONS(1250),
    [anon_sym_DQUOTE] = ACTIONS(1250),
    [aux_sym_string_token1] = ACTIONS(1250),
    [anon_sym_SQUOTE] = ACTIONS(1250),
    [anon_sym_LT_LT_LT] = ACTIONS(1250),
    [anon_sym_BQUOTE] = ACTIONS(1250),
    [anon_sym_DOLLAR] = ACTIONS(1250),
    [aux_sym_yield_expression_token1] = ACTIONS(1252),
    [aux_sym_include_expression_token1] = ACTIONS(1252),
    [aux_sym_include_once_expression_token1] = ACTIONS(1252),
    [aux_sym_require_expression_token1] = ACTIONS(1252),
    [aux_sym_require_once_expression_token1] = ACTIONS(1252),
    [sym_comment] = ACTIONS(3),
  },
  [495] = {
    [ts_builtin_sym_end] = ACTIONS(1254),
    [sym_name] = ACTIONS(1256),
    [anon_sym_QMARK_GT] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1254),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1256),
    [aux_sym_global_declaration_token1] = ACTIONS(1256),
    [aux_sym_namespace_definition_token1] = ACTIONS(1256),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1256),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1256),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1256),
    [anon_sym_BSLASH] = ACTIONS(1254),
    [anon_sym_LBRACE] = ACTIONS(1254),
    [anon_sym_RBRACE] = ACTIONS(1254),
    [aux_sym_trait_declaration_token1] = ACTIONS(1256),
    [aux_sym_interface_declaration_token1] = ACTIONS(1256),
    [aux_sym_enum_declaration_token1] = ACTIONS(1256),
    [aux_sym_enum_case_token1] = ACTIONS(1256),
    [aux_sym_class_declaration_token1] = ACTIONS(1256),
    [aux_sym_final_modifier_token1] = ACTIONS(1256),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1256),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1256),
    [sym_var_modifier] = ACTIONS(1256),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1256),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1256),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1256),
    [anon_sym_LPAREN2] = ACTIONS(1254),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1256),
    [aux_sym_cast_type_token1] = ACTIONS(1256),
    [aux_sym_echo_statement_token1] = ACTIONS(1256),
    [aux_sym_exit_statement_token1] = ACTIONS(1256),
    [anon_sym_unset] = ACTIONS(1256),
    [aux_sym_declare_statement_token1] = ACTIONS(1256),
    [aux_sym_declare_statement_token2] = ACTIONS(1256),
    [sym_float] = ACTIONS(1256),
    [aux_sym_try_statement_token1] = ACTIONS(1256),
    [aux_sym_goto_statement_token1] = ACTIONS(1256),
    [aux_sym_continue_statement_token1] = ACTIONS(1256),
    [aux_sym_break_statement_token1] = ACTIONS(1256),
    [sym_integer] = ACTIONS(1256),
    [aux_sym_return_statement_token1] = ACTIONS(1256),
    [aux_sym_throw_expression_token1] = ACTIONS(1256),
    [aux_sym_while_statement_token1] = ACTIONS(1256),
    [aux_sym_while_statement_token2] = ACTIONS(1256),
    [aux_sym_do_statement_token1] = ACTIONS(1256),
    [aux_sym_for_statement_token1] = ACTIONS(1256),
    [aux_sym_for_statement_token2] = ACTIONS(1256),
    [aux_sym_foreach_statement_token1] = ACTIONS(1256),
    [aux_sym_foreach_statement_token2] = ACTIONS(1256),
    [aux_sym_if_statement_token1] = ACTIONS(1256),
    [aux_sym_if_statement_token2] = ACTIONS(1256),
    [aux_sym_else_if_clause_token1] = ACTIONS(1256),
    [aux_sym_else_clause_token1] = ACTIONS(1256),
    [aux_sym_match_expression_token1] = ACTIONS(1256),
    [aux_sym_match_default_expression_token1] = ACTIONS(1256),
    [aux_sym_switch_statement_token1] = ACTIONS(1256),
    [aux_sym_switch_block_token1] = ACTIONS(1256),
    [anon_sym_PLUS] = ACTIONS(1256),
    [anon_sym_DASH] = ACTIONS(1256),
    [anon_sym_TILDE] = ACTIONS(1254),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [aux_sym_clone_expression_token1] = ACTIONS(1256),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1256),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1256),
    [anon_sym_DASH_DASH] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1254),
    [aux_sym__list_destructing_token1] = ACTIONS(1256),
    [anon_sym_LBRACK] = ACTIONS(1254),
    [anon_sym_self] = ACTIONS(1256),
    [anon_sym_parent] = ACTIONS(1256),
    [aux_sym__argument_name_token1] = ACTIONS(1256),
    [aux_sym__argument_name_token2] = ACTIONS(1256),
    [anon_sym_POUND_LBRACK] = ACTIONS(1254),
    [aux_sym_encapsed_string_token1] = ACTIONS(1254),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [aux_sym_string_token1] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_LT_LT_LT] = ACTIONS(1254),
    [anon_sym_BQUOTE] = ACTIONS(1254),
    [anon_sym_DOLLAR] = ACTIONS(1254),
    [aux_sym_yield_expression_token1] = ACTIONS(1256),
    [aux_sym_include_expression_token1] = ACTIONS(1256),
    [aux_sym_include_once_expression_token1] = ACTIONS(1256),
    [aux_sym_require_expression_token1] = ACTIONS(1256),
    [aux_sym_require_once_expression_token1] = ACTIONS(1256),
    [sym_comment] = ACTIONS(3),
  },
  [496] = {
    [ts_builtin_sym_end] = ACTIONS(1258),
    [sym_name] = ACTIONS(1260),
    [anon_sym_QMARK_GT] = ACTIONS(1258),
    [anon_sym_SEMI] = ACTIONS(1258),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1260),
    [aux_sym_global_declaration_token1] = ACTIONS(1260),
    [aux_sym_namespace_definition_token1] = ACTIONS(1260),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1260),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1260),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1260),
    [anon_sym_BSLASH] = ACTIONS(1258),
    [anon_sym_LBRACE] = ACTIONS(1258),
    [anon_sym_RBRACE] = ACTIONS(1258),
    [aux_sym_trait_declaration_token1] = ACTIONS(1260),
    [aux_sym_interface_declaration_token1] = ACTIONS(1260),
    [aux_sym_enum_declaration_token1] = ACTIONS(1260),
    [aux_sym_enum_case_token1] = ACTIONS(1260),
    [aux_sym_class_declaration_token1] = ACTIONS(1260),
    [aux_sym_final_modifier_token1] = ACTIONS(1260),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1260),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1260),
    [sym_var_modifier] = ACTIONS(1260),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1260),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1260),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1260),
    [anon_sym_LPAREN2] = ACTIONS(1258),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1260),
    [aux_sym_cast_type_token1] = ACTIONS(1260),
    [aux_sym_echo_statement_token1] = ACTIONS(1260),
    [aux_sym_exit_statement_token1] = ACTIONS(1260),
    [anon_sym_unset] = ACTIONS(1260),
    [aux_sym_declare_statement_token1] = ACTIONS(1260),
    [aux_sym_declare_statement_token2] = ACTIONS(1260),
    [sym_float] = ACTIONS(1260),
    [aux_sym_try_statement_token1] = ACTIONS(1260),
    [aux_sym_goto_statement_token1] = ACTIONS(1260),
    [aux_sym_continue_statement_token1] = ACTIONS(1260),
    [aux_sym_break_statement_token1] = ACTIONS(1260),
    [sym_integer] = ACTIONS(1260),
    [aux_sym_return_statement_token1] = ACTIONS(1260),
    [aux_sym_throw_expression_token1] = ACTIONS(1260),
    [aux_sym_while_statement_token1] = ACTIONS(1260),
    [aux_sym_while_statement_token2] = ACTIONS(1260),
    [aux_sym_do_statement_token1] = ACTIONS(1260),
    [aux_sym_for_statement_token1] = ACTIONS(1260),
    [aux_sym_for_statement_token2] = ACTIONS(1260),
    [aux_sym_foreach_statement_token1] = ACTIONS(1260),
    [aux_sym_foreach_statement_token2] = ACTIONS(1260),
    [aux_sym_if_statement_token1] = ACTIONS(1260),
    [aux_sym_if_statement_token2] = ACTIONS(1260),
    [aux_sym_else_if_clause_token1] = ACTIONS(1260),
    [aux_sym_else_clause_token1] = ACTIONS(1260),
    [aux_sym_match_expression_token1] = ACTIONS(1260),
    [aux_sym_match_default_expression_token1] = ACTIONS(1260),
    [aux_sym_switch_statement_token1] = ACTIONS(1260),
    [aux_sym_switch_block_token1] = ACTIONS(1260),
    [anon_sym_PLUS] = ACTIONS(1260),
    [anon_sym_DASH] = ACTIONS(1260),
    [anon_sym_TILDE] = ACTIONS(1258),
    [anon_sym_BANG] = ACTIONS(1258),
    [anon_sym_AT] = ACTIONS(1258),
    [aux_sym_clone_expression_token1] = ACTIONS(1260),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1260),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1258),
    [anon_sym_PLUS_PLUS] = ACTIONS(1258),
    [aux_sym__list_destructing_token1] = ACTIONS(1260),
    [anon_sym_LBRACK] = ACTIONS(1258),
    [anon_sym_self] = ACTIONS(1260),
    [anon_sym_parent] = ACTIONS(1260),
    [aux_sym__argument_name_token1] = ACTIONS(1260),
    [aux_sym__argument_name_token2] = ACTIONS(1260),
    [anon_sym_POUND_LBRACK] = ACTIONS(1258),
    [aux_sym_encapsed_string_token1] = ACTIONS(1258),
    [anon_sym_DQUOTE] = ACTIONS(1258),
    [aux_sym_string_token1] = ACTIONS(1258),
    [anon_sym_SQUOTE] = ACTIONS(1258),
    [anon_sym_LT_LT_LT] = ACTIONS(1258),
    [anon_sym_BQUOTE] = ACTIONS(1258),
    [anon_sym_DOLLAR] = ACTIONS(1258),
    [aux_sym_yield_expression_token1] = ACTIONS(1260),
    [aux_sym_include_expression_token1] = ACTIONS(1260),
    [aux_sym_include_once_expression_token1] = ACTIONS(1260),
    [aux_sym_require_expression_token1] = ACTIONS(1260),
    [aux_sym_require_once_expression_token1] = ACTIONS(1260),
    [sym_comment] = ACTIONS(3),
  },
  [497] = {
    [ts_builtin_sym_end] = ACTIONS(1262),
    [sym_name] = ACTIONS(1264),
    [anon_sym_QMARK_GT] = ACTIONS(1262),
    [anon_sym_SEMI] = ACTIONS(1262),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1264),
    [aux_sym_global_declaration_token1] = ACTIONS(1264),
    [aux_sym_namespace_definition_token1] = ACTIONS(1264),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1264),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1264),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1264),
    [anon_sym_BSLASH] = ACTIONS(1262),
    [anon_sym_LBRACE] = ACTIONS(1262),
    [anon_sym_RBRACE] = ACTIONS(1262),
    [aux_sym_trait_declaration_token1] = ACTIONS(1264),
    [aux_sym_interface_declaration_token1] = ACTIONS(1264),
    [aux_sym_enum_declaration_token1] = ACTIONS(1264),
    [aux_sym_enum_case_token1] = ACTIONS(1264),
    [aux_sym_class_declaration_token1] = ACTIONS(1264),
    [aux_sym_final_modifier_token1] = ACTIONS(1264),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1264),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1264),
    [sym_var_modifier] = ACTIONS(1264),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1264),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1264),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1264),
    [anon_sym_LPAREN2] = ACTIONS(1262),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1264),
    [aux_sym_cast_type_token1] = ACTIONS(1264),
    [aux_sym_echo_statement_token1] = ACTIONS(1264),
    [aux_sym_exit_statement_token1] = ACTIONS(1264),
    [anon_sym_unset] = ACTIONS(1264),
    [aux_sym_declare_statement_token1] = ACTIONS(1264),
    [aux_sym_declare_statement_token2] = ACTIONS(1264),
    [sym_float] = ACTIONS(1264),
    [aux_sym_try_statement_token1] = ACTIONS(1264),
    [aux_sym_goto_statement_token1] = ACTIONS(1264),
    [aux_sym_continue_statement_token1] = ACTIONS(1264),
    [aux_sym_break_statement_token1] = ACTIONS(1264),
    [sym_integer] = ACTIONS(1264),
    [aux_sym_return_statement_token1] = ACTIONS(1264),
    [aux_sym_throw_expression_token1] = ACTIONS(1264),
    [aux_sym_while_statement_token1] = ACTIONS(1264),
    [aux_sym_while_statement_token2] = ACTIONS(1264),
    [aux_sym_do_statement_token1] = ACTIONS(1264),
    [aux_sym_for_statement_token1] = ACTIONS(1264),
    [aux_sym_for_statement_token2] = ACTIONS(1264),
    [aux_sym_foreach_statement_token1] = ACTIONS(1264),
    [aux_sym_foreach_statement_token2] = ACTIONS(1264),
    [aux_sym_if_statement_token1] = ACTIONS(1264),
    [aux_sym_if_statement_token2] = ACTIONS(1264),
    [aux_sym_else_if_clause_token1] = ACTIONS(1264),
    [aux_sym_else_clause_token1] = ACTIONS(1264),
    [aux_sym_match_expression_token1] = ACTIONS(1264),
    [aux_sym_match_default_expression_token1] = ACTIONS(1264),
    [aux_sym_switch_statement_token1] = ACTIONS(1264),
    [aux_sym_switch_block_token1] = ACTIONS(1264),
    [anon_sym_PLUS] = ACTIONS(1264),
    [anon_sym_DASH] = ACTIONS(1264),
    [anon_sym_TILDE] = ACTIONS(1262),
    [anon_sym_BANG] = ACTIONS(1262),
    [anon_sym_AT] = ACTIONS(1262),
    [aux_sym_clone_expression_token1] = ACTIONS(1264),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1264),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1264),
    [anon_sym_DASH_DASH] = ACTIONS(1262),
    [anon_sym_PLUS_PLUS] = ACTIONS(1262),
    [aux_sym__list_destructing_token1] = ACTIONS(1264),
    [anon_sym_LBRACK] = ACTIONS(1262),
    [anon_sym_self] = ACTIONS(1264),
    [anon_sym_parent] = ACTIONS(1264),
    [aux_sym__argument_name_token1] = ACTIONS(1264),
    [aux_sym__argument_name_token2] = ACTIONS(1264),
    [anon_sym_POUND_LBRACK] = ACTIONS(1262),
    [aux_sym_encapsed_string_token1] = ACTIONS(1262),
    [anon_sym_DQUOTE] = ACTIONS(1262),
    [aux_sym_string_token1] = ACTIONS(1262),
    [anon_sym_SQUOTE] = ACTIONS(1262),
    [anon_sym_LT_LT_LT] = ACTIONS(1262),
    [anon_sym_BQUOTE] = ACTIONS(1262),
    [anon_sym_DOLLAR] = ACTIONS(1262),
    [aux_sym_yield_expression_token1] = ACTIONS(1264),
    [aux_sym_include_expression_token1] = ACTIONS(1264),
    [aux_sym_include_once_expression_token1] = ACTIONS(1264),
    [aux_sym_require_expression_token1] = ACTIONS(1264),
    [aux_sym_require_once_expression_token1] = ACTIONS(1264),
    [sym_comment] = ACTIONS(3),
  },
  [498] = {
    [ts_builtin_sym_end] = ACTIONS(1266),
    [sym_name] = ACTIONS(1268),
    [anon_sym_QMARK_GT] = ACTIONS(1266),
    [anon_sym_SEMI] = ACTIONS(1266),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1268),
    [aux_sym_global_declaration_token1] = ACTIONS(1268),
    [aux_sym_namespace_definition_token1] = ACTIONS(1268),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1268),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1268),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1268),
    [anon_sym_BSLASH] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1266),
    [anon_sym_RBRACE] = ACTIONS(1266),
    [aux_sym_trait_declaration_token1] = ACTIONS(1268),
    [aux_sym_interface_declaration_token1] = ACTIONS(1268),
    [aux_sym_enum_declaration_token1] = ACTIONS(1268),
    [aux_sym_enum_case_token1] = ACTIONS(1268),
    [aux_sym_class_declaration_token1] = ACTIONS(1268),
    [aux_sym_final_modifier_token1] = ACTIONS(1268),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1268),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1268),
    [sym_var_modifier] = ACTIONS(1268),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1268),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1268),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1268),
    [anon_sym_LPAREN2] = ACTIONS(1266),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1268),
    [aux_sym_cast_type_token1] = ACTIONS(1268),
    [aux_sym_echo_statement_token1] = ACTIONS(1268),
    [aux_sym_exit_statement_token1] = ACTIONS(1268),
    [anon_sym_unset] = ACTIONS(1268),
    [aux_sym_declare_statement_token1] = ACTIONS(1268),
    [aux_sym_declare_statement_token2] = ACTIONS(1268),
    [sym_float] = ACTIONS(1268),
    [aux_sym_try_statement_token1] = ACTIONS(1268),
    [aux_sym_goto_statement_token1] = ACTIONS(1268),
    [aux_sym_continue_statement_token1] = ACTIONS(1268),
    [aux_sym_break_statement_token1] = ACTIONS(1268),
    [sym_integer] = ACTIONS(1268),
    [aux_sym_return_statement_token1] = ACTIONS(1268),
    [aux_sym_throw_expression_token1] = ACTIONS(1268),
    [aux_sym_while_statement_token1] = ACTIONS(1268),
    [aux_sym_while_statement_token2] = ACTIONS(1268),
    [aux_sym_do_statement_token1] = ACTIONS(1268),
    [aux_sym_for_statement_token1] = ACTIONS(1268),
    [aux_sym_for_statement_token2] = ACTIONS(1268),
    [aux_sym_foreach_statement_token1] = ACTIONS(1268),
    [aux_sym_foreach_statement_token2] = ACTIONS(1268),
    [aux_sym_if_statement_token1] = ACTIONS(1268),
    [aux_sym_if_statement_token2] = ACTIONS(1268),
    [aux_sym_else_if_clause_token1] = ACTIONS(1268),
    [aux_sym_else_clause_token1] = ACTIONS(1268),
    [aux_sym_match_expression_token1] = ACTIONS(1268),
    [aux_sym_match_default_expression_token1] = ACTIONS(1268),
    [aux_sym_switch_statement_token1] = ACTIONS(1268),
    [aux_sym_switch_block_token1] = ACTIONS(1268),
    [anon_sym_PLUS] = ACTIONS(1268),
    [anon_sym_DASH] = ACTIONS(1268),
    [anon_sym_TILDE] = ACTIONS(1266),
    [anon_sym_BANG] = ACTIONS(1266),
    [anon_sym_AT] = ACTIONS(1266),
    [aux_sym_clone_expression_token1] = ACTIONS(1268),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1268),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1268),
    [anon_sym_DASH_DASH] = ACTIONS(1266),
    [anon_sym_PLUS_PLUS] = ACTIONS(1266),
    [aux_sym__list_destructing_token1] = ACTIONS(1268),
    [anon_sym_LBRACK] = ACTIONS(1266),
    [anon_sym_self] = ACTIONS(1268),
    [anon_sym_parent] = ACTIONS(1268),
    [aux_sym__argument_name_token1] = ACTIONS(1268),
    [aux_sym__argument_name_token2] = ACTIONS(1268),
    [anon_sym_POUND_LBRACK] = ACTIONS(1266),
    [aux_sym_encapsed_string_token1] = ACTIONS(1266),
    [anon_sym_DQUOTE] = ACTIONS(1266),
    [aux_sym_string_token1] = ACTIONS(1266),
    [anon_sym_SQUOTE] = ACTIONS(1266),
    [anon_sym_LT_LT_LT] = ACTIONS(1266),
    [anon_sym_BQUOTE] = ACTIONS(1266),
    [anon_sym_DOLLAR] = ACTIONS(1266),
    [aux_sym_yield_expression_token1] = ACTIONS(1268),
    [aux_sym_include_expression_token1] = ACTIONS(1268),
    [aux_sym_include_once_expression_token1] = ACTIONS(1268),
    [aux_sym_require_expression_token1] = ACTIONS(1268),
    [aux_sym_require_once_expression_token1] = ACTIONS(1268),
    [sym_comment] = ACTIONS(3),
  },
  [499] = {
    [ts_builtin_sym_end] = ACTIONS(1270),
    [sym_name] = ACTIONS(1272),
    [anon_sym_QMARK_GT] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1270),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1272),
    [aux_sym_global_declaration_token1] = ACTIONS(1272),
    [aux_sym_namespace_definition_token1] = ACTIONS(1272),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1272),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1272),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1272),
    [anon_sym_BSLASH] = ACTIONS(1270),
    [anon_sym_LBRACE] = ACTIONS(1270),
    [anon_sym_RBRACE] = ACTIONS(1270),
    [aux_sym_trait_declaration_token1] = ACTIONS(1272),
    [aux_sym_interface_declaration_token1] = ACTIONS(1272),
    [aux_sym_enum_declaration_token1] = ACTIONS(1272),
    [aux_sym_enum_case_token1] = ACTIONS(1272),
    [aux_sym_class_declaration_token1] = ACTIONS(1272),
    [aux_sym_final_modifier_token1] = ACTIONS(1272),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1272),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1272),
    [sym_var_modifier] = ACTIONS(1272),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1272),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1272),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1272),
    [anon_sym_LPAREN2] = ACTIONS(1270),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1272),
    [aux_sym_cast_type_token1] = ACTIONS(1272),
    [aux_sym_echo_statement_token1] = ACTIONS(1272),
    [aux_sym_exit_statement_token1] = ACTIONS(1272),
    [anon_sym_unset] = ACTIONS(1272),
    [aux_sym_declare_statement_token1] = ACTIONS(1272),
    [aux_sym_declare_statement_token2] = ACTIONS(1272),
    [sym_float] = ACTIONS(1272),
    [aux_sym_try_statement_token1] = ACTIONS(1272),
    [aux_sym_goto_statement_token1] = ACTIONS(1272),
    [aux_sym_continue_statement_token1] = ACTIONS(1272),
    [aux_sym_break_statement_token1] = ACTIONS(1272),
    [sym_integer] = ACTIONS(1272),
    [aux_sym_return_statement_token1] = ACTIONS(1272),
    [aux_sym_throw_expression_token1] = ACTIONS(1272),
    [aux_sym_while_statement_token1] = ACTIONS(1272),
    [aux_sym_while_statement_token2] = ACTIONS(1272),
    [aux_sym_do_statement_token1] = ACTIONS(1272),
    [aux_sym_for_statement_token1] = ACTIONS(1272),
    [aux_sym_for_statement_token2] = ACTIONS(1272),
    [aux_sym_foreach_statement_token1] = ACTIONS(1272),
    [aux_sym_foreach_statement_token2] = ACTIONS(1272),
    [aux_sym_if_statement_token1] = ACTIONS(1272),
    [aux_sym_if_statement_token2] = ACTIONS(1272),
    [aux_sym_else_if_clause_token1] = ACTIONS(1272),
    [aux_sym_else_clause_token1] = ACTIONS(1272),
    [aux_sym_match_expression_token1] = ACTIONS(1272),
    [aux_sym_match_default_expression_token1] = ACTIONS(1272),
    [aux_sym_switch_statement_token1] = ACTIONS(1272),
    [aux_sym_switch_block_token1] = ACTIONS(1272),
    [anon_sym_PLUS] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1272),
    [anon_sym_TILDE] = ACTIONS(1270),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [aux_sym_clone_expression_token1] = ACTIONS(1272),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1272),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1272),
    [anon_sym_DASH_DASH] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1270),
    [aux_sym__list_destructing_token1] = ACTIONS(1272),
    [anon_sym_LBRACK] = ACTIONS(1270),
    [anon_sym_self] = ACTIONS(1272),
    [anon_sym_parent] = ACTIONS(1272),
    [aux_sym__argument_name_token1] = ACTIONS(1272),
    [aux_sym__argument_name_token2] = ACTIONS(1272),
    [anon_sym_POUND_LBRACK] = ACTIONS(1270),
    [aux_sym_encapsed_string_token1] = ACTIONS(1270),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [aux_sym_string_token1] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_LT_LT_LT] = ACTIONS(1270),
    [anon_sym_BQUOTE] = ACTIONS(1270),
    [anon_sym_DOLLAR] = ACTIONS(1270),
    [aux_sym_yield_expression_token1] = ACTIONS(1272),
    [aux_sym_include_expression_token1] = ACTIONS(1272),
    [aux_sym_include_once_expression_token1] = ACTIONS(1272),
    [aux_sym_require_expression_token1] = ACTIONS(1272),
    [aux_sym_require_once_expression_token1] = ACTIONS(1272),
    [sym_comment] = ACTIONS(3),
  },
  [500] = {
    [ts_builtin_sym_end] = ACTIONS(1274),
    [sym_name] = ACTIONS(1276),
    [anon_sym_QMARK_GT] = ACTIONS(1274),
    [anon_sym_SEMI] = ACTIONS(1274),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1276),
    [aux_sym_global_declaration_token1] = ACTIONS(1276),
    [aux_sym_namespace_definition_token1] = ACTIONS(1276),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1276),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1276),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1276),
    [anon_sym_BSLASH] = ACTIONS(1274),
    [anon_sym_LBRACE] = ACTIONS(1274),
    [anon_sym_RBRACE] = ACTIONS(1274),
    [aux_sym_trait_declaration_token1] = ACTIONS(1276),
    [aux_sym_interface_declaration_token1] = ACTIONS(1276),
    [aux_sym_enum_declaration_token1] = ACTIONS(1276),
    [aux_sym_enum_case_token1] = ACTIONS(1276),
    [aux_sym_class_declaration_token1] = ACTIONS(1276),
    [aux_sym_final_modifier_token1] = ACTIONS(1276),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1276),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1276),
    [sym_var_modifier] = ACTIONS(1276),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1276),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1276),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1276),
    [anon_sym_LPAREN2] = ACTIONS(1274),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1276),
    [aux_sym_cast_type_token1] = ACTIONS(1276),
    [aux_sym_echo_statement_token1] = ACTIONS(1276),
    [aux_sym_exit_statement_token1] = ACTIONS(1276),
    [anon_sym_unset] = ACTIONS(1276),
    [aux_sym_declare_statement_token1] = ACTIONS(1276),
    [aux_sym_declare_statement_token2] = ACTIONS(1276),
    [sym_float] = ACTIONS(1276),
    [aux_sym_try_statement_token1] = ACTIONS(1276),
    [aux_sym_goto_statement_token1] = ACTIONS(1276),
    [aux_sym_continue_statement_token1] = ACTIONS(1276),
    [aux_sym_break_statement_token1] = ACTIONS(1276),
    [sym_integer] = ACTIONS(1276),
    [aux_sym_return_statement_token1] = ACTIONS(1276),
    [aux_sym_throw_expression_token1] = ACTIONS(1276),
    [aux_sym_while_statement_token1] = ACTIONS(1276),
    [aux_sym_while_statement_token2] = ACTIONS(1276),
    [aux_sym_do_statement_token1] = ACTIONS(1276),
    [aux_sym_for_statement_token1] = ACTIONS(1276),
    [aux_sym_for_statement_token2] = ACTIONS(1276),
    [aux_sym_foreach_statement_token1] = ACTIONS(1276),
    [aux_sym_foreach_statement_token2] = ACTIONS(1276),
    [aux_sym_if_statement_token1] = ACTIONS(1276),
    [aux_sym_if_statement_token2] = ACTIONS(1276),
    [aux_sym_else_if_clause_token1] = ACTIONS(1276),
    [aux_sym_else_clause_token1] = ACTIONS(1276),
    [aux_sym_match_expression_token1] = ACTIONS(1276),
    [aux_sym_match_default_expression_token1] = ACTIONS(1276),
    [aux_sym_switch_statement_token1] = ACTIONS(1276),
    [aux_sym_switch_block_token1] = ACTIONS(1276),
    [anon_sym_PLUS] = ACTIONS(1276),
    [anon_sym_DASH] = ACTIONS(1276),
    [anon_sym_TILDE] = ACTIONS(1274),
    [anon_sym_BANG] = ACTIONS(1274),
    [anon_sym_AT] = ACTIONS(1274),
    [aux_sym_clone_expression_token1] = ACTIONS(1276),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1276),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1276),
    [anon_sym_DASH_DASH] = ACTIONS(1274),
    [anon_sym_PLUS_PLUS] = ACTIONS(1274),
    [aux_sym__list_destructing_token1] = ACTIONS(1276),
    [anon_sym_LBRACK] = ACTIONS(1274),
    [anon_sym_self] = ACTIONS(1276),
    [anon_sym_parent] = ACTIONS(1276),
    [aux_sym__argument_name_token1] = ACTIONS(1276),
    [aux_sym__argument_name_token2] = ACTIONS(1276),
    [anon_sym_POUND_LBRACK] = ACTIONS(1274),
    [aux_sym_encapsed_string_token1] = ACTIONS(1274),
    [anon_sym_DQUOTE] = ACTIONS(1274),
    [aux_sym_string_token1] = ACTIONS(1274),
    [anon_sym_SQUOTE] = ACTIONS(1274),
    [anon_sym_LT_LT_LT] = ACTIONS(1274),
    [anon_sym_BQUOTE] = ACTIONS(1274),
    [anon_sym_DOLLAR] = ACTIONS(1274),
    [aux_sym_yield_expression_token1] = ACTIONS(1276),
    [aux_sym_include_expression_token1] = ACTIONS(1276),
    [aux_sym_include_once_expression_token1] = ACTIONS(1276),
    [aux_sym_require_expression_token1] = ACTIONS(1276),
    [aux_sym_require_once_expression_token1] = ACTIONS(1276),
    [sym_comment] = ACTIONS(3),
  },
  [501] = {
    [ts_builtin_sym_end] = ACTIONS(1278),
    [sym_name] = ACTIONS(1280),
    [anon_sym_QMARK_GT] = ACTIONS(1278),
    [anon_sym_SEMI] = ACTIONS(1278),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1280),
    [aux_sym_global_declaration_token1] = ACTIONS(1280),
    [aux_sym_namespace_definition_token1] = ACTIONS(1280),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1280),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1280),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1280),
    [anon_sym_BSLASH] = ACTIONS(1278),
    [anon_sym_LBRACE] = ACTIONS(1278),
    [anon_sym_RBRACE] = ACTIONS(1278),
    [aux_sym_trait_declaration_token1] = ACTIONS(1280),
    [aux_sym_interface_declaration_token1] = ACTIONS(1280),
    [aux_sym_enum_declaration_token1] = ACTIONS(1280),
    [aux_sym_enum_case_token1] = ACTIONS(1280),
    [aux_sym_class_declaration_token1] = ACTIONS(1280),
    [aux_sym_final_modifier_token1] = ACTIONS(1280),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1280),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1280),
    [sym_var_modifier] = ACTIONS(1280),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1280),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1280),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1280),
    [anon_sym_LPAREN2] = ACTIONS(1278),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1280),
    [aux_sym_cast_type_token1] = ACTIONS(1280),
    [aux_sym_echo_statement_token1] = ACTIONS(1280),
    [aux_sym_exit_statement_token1] = ACTIONS(1280),
    [anon_sym_unset] = ACTIONS(1280),
    [aux_sym_declare_statement_token1] = ACTIONS(1280),
    [aux_sym_declare_statement_token2] = ACTIONS(1280),
    [sym_float] = ACTIONS(1280),
    [aux_sym_try_statement_token1] = ACTIONS(1280),
    [aux_sym_goto_statement_token1] = ACTIONS(1280),
    [aux_sym_continue_statement_token1] = ACTIONS(1280),
    [aux_sym_break_statement_token1] = ACTIONS(1280),
    [sym_integer] = ACTIONS(1280),
    [aux_sym_return_statement_token1] = ACTIONS(1280),
    [aux_sym_throw_expression_token1] = ACTIONS(1280),
    [aux_sym_while_statement_token1] = ACTIONS(1280),
    [aux_sym_while_statement_token2] = ACTIONS(1280),
    [aux_sym_do_statement_token1] = ACTIONS(1280),
    [aux_sym_for_statement_token1] = ACTIONS(1280),
    [aux_sym_for_statement_token2] = ACTIONS(1280),
    [aux_sym_foreach_statement_token1] = ACTIONS(1280),
    [aux_sym_foreach_statement_token2] = ACTIONS(1280),
    [aux_sym_if_statement_token1] = ACTIONS(1280),
    [aux_sym_if_statement_token2] = ACTIONS(1280),
    [aux_sym_else_if_clause_token1] = ACTIONS(1280),
    [aux_sym_else_clause_token1] = ACTIONS(1280),
    [aux_sym_match_expression_token1] = ACTIONS(1280),
    [aux_sym_match_default_expression_token1] = ACTIONS(1280),
    [aux_sym_switch_statement_token1] = ACTIONS(1280),
    [aux_sym_switch_block_token1] = ACTIONS(1280),
    [anon_sym_PLUS] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1280),
    [anon_sym_TILDE] = ACTIONS(1278),
    [anon_sym_BANG] = ACTIONS(1278),
    [anon_sym_AT] = ACTIONS(1278),
    [aux_sym_clone_expression_token1] = ACTIONS(1280),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1280),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1280),
    [anon_sym_DASH_DASH] = ACTIONS(1278),
    [anon_sym_PLUS_PLUS] = ACTIONS(1278),
    [aux_sym__list_destructing_token1] = ACTIONS(1280),
    [anon_sym_LBRACK] = ACTIONS(1278),
    [anon_sym_self] = ACTIONS(1280),
    [anon_sym_parent] = ACTIONS(1280),
    [aux_sym__argument_name_token1] = ACTIONS(1280),
    [aux_sym__argument_name_token2] = ACTIONS(1280),
    [anon_sym_POUND_LBRACK] = ACTIONS(1278),
    [aux_sym_encapsed_string_token1] = ACTIONS(1278),
    [anon_sym_DQUOTE] = ACTIONS(1278),
    [aux_sym_string_token1] = ACTIONS(1278),
    [anon_sym_SQUOTE] = ACTIONS(1278),
    [anon_sym_LT_LT_LT] = ACTIONS(1278),
    [anon_sym_BQUOTE] = ACTIONS(1278),
    [anon_sym_DOLLAR] = ACTIONS(1278),
    [aux_sym_yield_expression_token1] = ACTIONS(1280),
    [aux_sym_include_expression_token1] = ACTIONS(1280),
    [aux_sym_include_once_expression_token1] = ACTIONS(1280),
    [aux_sym_require_expression_token1] = ACTIONS(1280),
    [aux_sym_require_once_expression_token1] = ACTIONS(1280),
    [sym_comment] = ACTIONS(3),
  },
  [502] = {
    [ts_builtin_sym_end] = ACTIONS(1282),
    [sym_name] = ACTIONS(1284),
    [anon_sym_QMARK_GT] = ACTIONS(1282),
    [anon_sym_SEMI] = ACTIONS(1282),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1284),
    [aux_sym_global_declaration_token1] = ACTIONS(1284),
    [aux_sym_namespace_definition_token1] = ACTIONS(1284),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1284),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1284),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1284),
    [anon_sym_BSLASH] = ACTIONS(1282),
    [anon_sym_LBRACE] = ACTIONS(1282),
    [anon_sym_RBRACE] = ACTIONS(1282),
    [aux_sym_trait_declaration_token1] = ACTIONS(1284),
    [aux_sym_interface_declaration_token1] = ACTIONS(1284),
    [aux_sym_enum_declaration_token1] = ACTIONS(1284),
    [aux_sym_enum_case_token1] = ACTIONS(1284),
    [aux_sym_class_declaration_token1] = ACTIONS(1284),
    [aux_sym_final_modifier_token1] = ACTIONS(1284),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1284),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1284),
    [sym_var_modifier] = ACTIONS(1284),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1284),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1284),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1284),
    [anon_sym_LPAREN2] = ACTIONS(1282),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1284),
    [aux_sym_cast_type_token1] = ACTIONS(1284),
    [aux_sym_echo_statement_token1] = ACTIONS(1284),
    [aux_sym_exit_statement_token1] = ACTIONS(1284),
    [anon_sym_unset] = ACTIONS(1284),
    [aux_sym_declare_statement_token1] = ACTIONS(1284),
    [aux_sym_declare_statement_token2] = ACTIONS(1284),
    [sym_float] = ACTIONS(1284),
    [aux_sym_try_statement_token1] = ACTIONS(1284),
    [aux_sym_goto_statement_token1] = ACTIONS(1284),
    [aux_sym_continue_statement_token1] = ACTIONS(1284),
    [aux_sym_break_statement_token1] = ACTIONS(1284),
    [sym_integer] = ACTIONS(1284),
    [aux_sym_return_statement_token1] = ACTIONS(1284),
    [aux_sym_throw_expression_token1] = ACTIONS(1284),
    [aux_sym_while_statement_token1] = ACTIONS(1284),
    [aux_sym_while_statement_token2] = ACTIONS(1284),
    [aux_sym_do_statement_token1] = ACTIONS(1284),
    [aux_sym_for_statement_token1] = ACTIONS(1284),
    [aux_sym_for_statement_token2] = ACTIONS(1284),
    [aux_sym_foreach_statement_token1] = ACTIONS(1284),
    [aux_sym_foreach_statement_token2] = ACTIONS(1284),
    [aux_sym_if_statement_token1] = ACTIONS(1284),
    [aux_sym_if_statement_token2] = ACTIONS(1284),
    [aux_sym_else_if_clause_token1] = ACTIONS(1284),
    [aux_sym_else_clause_token1] = ACTIONS(1284),
    [aux_sym_match_expression_token1] = ACTIONS(1284),
    [aux_sym_match_default_expression_token1] = ACTIONS(1284),
    [aux_sym_switch_statement_token1] = ACTIONS(1284),
    [aux_sym_switch_block_token1] = ACTIONS(1284),
    [anon_sym_PLUS] = ACTIONS(1284),
    [anon_sym_DASH] = ACTIONS(1284),
    [anon_sym_TILDE] = ACTIONS(1282),
    [anon_sym_BANG] = ACTIONS(1282),
    [anon_sym_AT] = ACTIONS(1282),
    [aux_sym_clone_expression_token1] = ACTIONS(1284),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1284),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1284),
    [anon_sym_DASH_DASH] = ACTIONS(1282),
    [anon_sym_PLUS_PLUS] = ACTIONS(1282),
    [aux_sym__list_destructing_token1] = ACTIONS(1284),
    [anon_sym_LBRACK] = ACTIONS(1282),
    [anon_sym_self] = ACTIONS(1284),
    [anon_sym_parent] = ACTIONS(1284),
    [aux_sym__argument_name_token1] = ACTIONS(1284),
    [aux_sym__argument_name_token2] = ACTIONS(1284),
    [anon_sym_POUND_LBRACK] = ACTIONS(1282),
    [aux_sym_encapsed_string_token1] = ACTIONS(1282),
    [anon_sym_DQUOTE] = ACTIONS(1282),
    [aux_sym_string_token1] = ACTIONS(1282),
    [anon_sym_SQUOTE] = ACTIONS(1282),
    [anon_sym_LT_LT_LT] = ACTIONS(1282),
    [anon_sym_BQUOTE] = ACTIONS(1282),
    [anon_sym_DOLLAR] = ACTIONS(1282),
    [aux_sym_yield_expression_token1] = ACTIONS(1284),
    [aux_sym_include_expression_token1] = ACTIONS(1284),
    [aux_sym_include_once_expression_token1] = ACTIONS(1284),
    [aux_sym_require_expression_token1] = ACTIONS(1284),
    [aux_sym_require_once_expression_token1] = ACTIONS(1284),
    [sym_comment] = ACTIONS(3),
  },
  [503] = {
    [ts_builtin_sym_end] = ACTIONS(1286),
    [sym_name] = ACTIONS(1288),
    [anon_sym_QMARK_GT] = ACTIONS(1286),
    [anon_sym_SEMI] = ACTIONS(1286),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1288),
    [aux_sym_global_declaration_token1] = ACTIONS(1288),
    [aux_sym_namespace_definition_token1] = ACTIONS(1288),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1288),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1288),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1288),
    [anon_sym_BSLASH] = ACTIONS(1286),
    [anon_sym_LBRACE] = ACTIONS(1286),
    [anon_sym_RBRACE] = ACTIONS(1286),
    [aux_sym_trait_declaration_token1] = ACTIONS(1288),
    [aux_sym_interface_declaration_token1] = ACTIONS(1288),
    [aux_sym_enum_declaration_token1] = ACTIONS(1288),
    [aux_sym_enum_case_token1] = ACTIONS(1288),
    [aux_sym_class_declaration_token1] = ACTIONS(1288),
    [aux_sym_final_modifier_token1] = ACTIONS(1288),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1288),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1288),
    [sym_var_modifier] = ACTIONS(1288),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1288),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1288),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1288),
    [anon_sym_LPAREN2] = ACTIONS(1286),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1288),
    [aux_sym_cast_type_token1] = ACTIONS(1288),
    [aux_sym_echo_statement_token1] = ACTIONS(1288),
    [aux_sym_exit_statement_token1] = ACTIONS(1288),
    [anon_sym_unset] = ACTIONS(1288),
    [aux_sym_declare_statement_token1] = ACTIONS(1288),
    [aux_sym_declare_statement_token2] = ACTIONS(1288),
    [sym_float] = ACTIONS(1288),
    [aux_sym_try_statement_token1] = ACTIONS(1288),
    [aux_sym_goto_statement_token1] = ACTIONS(1288),
    [aux_sym_continue_statement_token1] = ACTIONS(1288),
    [aux_sym_break_statement_token1] = ACTIONS(1288),
    [sym_integer] = ACTIONS(1288),
    [aux_sym_return_statement_token1] = ACTIONS(1288),
    [aux_sym_throw_expression_token1] = ACTIONS(1288),
    [aux_sym_while_statement_token1] = ACTIONS(1288),
    [aux_sym_while_statement_token2] = ACTIONS(1288),
    [aux_sym_do_statement_token1] = ACTIONS(1288),
    [aux_sym_for_statement_token1] = ACTIONS(1288),
    [aux_sym_for_statement_token2] = ACTIONS(1288),
    [aux_sym_foreach_statement_token1] = ACTIONS(1288),
    [aux_sym_foreach_statement_token2] = ACTIONS(1288),
    [aux_sym_if_statement_token1] = ACTIONS(1288),
    [aux_sym_if_statement_token2] = ACTIONS(1288),
    [aux_sym_else_if_clause_token1] = ACTIONS(1288),
    [aux_sym_else_clause_token1] = ACTIONS(1288),
    [aux_sym_match_expression_token1] = ACTIONS(1288),
    [aux_sym_match_default_expression_token1] = ACTIONS(1288),
    [aux_sym_switch_statement_token1] = ACTIONS(1288),
    [aux_sym_switch_block_token1] = ACTIONS(1288),
    [anon_sym_PLUS] = ACTIONS(1288),
    [anon_sym_DASH] = ACTIONS(1288),
    [anon_sym_TILDE] = ACTIONS(1286),
    [anon_sym_BANG] = ACTIONS(1286),
    [anon_sym_AT] = ACTIONS(1286),
    [aux_sym_clone_expression_token1] = ACTIONS(1288),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1288),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1288),
    [anon_sym_DASH_DASH] = ACTIONS(1286),
    [anon_sym_PLUS_PLUS] = ACTIONS(1286),
    [aux_sym__list_destructing_token1] = ACTIONS(1288),
    [anon_sym_LBRACK] = ACTIONS(1286),
    [anon_sym_self] = ACTIONS(1288),
    [anon_sym_parent] = ACTIONS(1288),
    [aux_sym__argument_name_token1] = ACTIONS(1288),
    [aux_sym__argument_name_token2] = ACTIONS(1288),
    [anon_sym_POUND_LBRACK] = ACTIONS(1286),
    [aux_sym_encapsed_string_token1] = ACTIONS(1286),
    [anon_sym_DQUOTE] = ACTIONS(1286),
    [aux_sym_string_token1] = ACTIONS(1286),
    [anon_sym_SQUOTE] = ACTIONS(1286),
    [anon_sym_LT_LT_LT] = ACTIONS(1286),
    [anon_sym_BQUOTE] = ACTIONS(1286),
    [anon_sym_DOLLAR] = ACTIONS(1286),
    [aux_sym_yield_expression_token1] = ACTIONS(1288),
    [aux_sym_include_expression_token1] = ACTIONS(1288),
    [aux_sym_include_once_expression_token1] = ACTIONS(1288),
    [aux_sym_require_expression_token1] = ACTIONS(1288),
    [aux_sym_require_once_expression_token1] = ACTIONS(1288),
    [sym_comment] = ACTIONS(3),
  },
  [504] = {
    [ts_builtin_sym_end] = ACTIONS(1290),
    [sym_name] = ACTIONS(1292),
    [anon_sym_QMARK_GT] = ACTIONS(1290),
    [anon_sym_SEMI] = ACTIONS(1290),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1292),
    [aux_sym_global_declaration_token1] = ACTIONS(1292),
    [aux_sym_namespace_definition_token1] = ACTIONS(1292),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1292),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1292),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1292),
    [anon_sym_BSLASH] = ACTIONS(1290),
    [anon_sym_LBRACE] = ACTIONS(1290),
    [anon_sym_RBRACE] = ACTIONS(1290),
    [aux_sym_trait_declaration_token1] = ACTIONS(1292),
    [aux_sym_interface_declaration_token1] = ACTIONS(1292),
    [aux_sym_enum_declaration_token1] = ACTIONS(1292),
    [aux_sym_enum_case_token1] = ACTIONS(1292),
    [aux_sym_class_declaration_token1] = ACTIONS(1292),
    [aux_sym_final_modifier_token1] = ACTIONS(1292),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1292),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1292),
    [sym_var_modifier] = ACTIONS(1292),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1292),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1292),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1292),
    [anon_sym_LPAREN2] = ACTIONS(1290),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1292),
    [aux_sym_cast_type_token1] = ACTIONS(1292),
    [aux_sym_echo_statement_token1] = ACTIONS(1292),
    [aux_sym_exit_statement_token1] = ACTIONS(1292),
    [anon_sym_unset] = ACTIONS(1292),
    [aux_sym_declare_statement_token1] = ACTIONS(1292),
    [aux_sym_declare_statement_token2] = ACTIONS(1292),
    [sym_float] = ACTIONS(1292),
    [aux_sym_try_statement_token1] = ACTIONS(1292),
    [aux_sym_goto_statement_token1] = ACTIONS(1292),
    [aux_sym_continue_statement_token1] = ACTIONS(1292),
    [aux_sym_break_statement_token1] = ACTIONS(1292),
    [sym_integer] = ACTIONS(1292),
    [aux_sym_return_statement_token1] = ACTIONS(1292),
    [aux_sym_throw_expression_token1] = ACTIONS(1292),
    [aux_sym_while_statement_token1] = ACTIONS(1292),
    [aux_sym_while_statement_token2] = ACTIONS(1292),
    [aux_sym_do_statement_token1] = ACTIONS(1292),
    [aux_sym_for_statement_token1] = ACTIONS(1292),
    [aux_sym_for_statement_token2] = ACTIONS(1292),
    [aux_sym_foreach_statement_token1] = ACTIONS(1292),
    [aux_sym_foreach_statement_token2] = ACTIONS(1292),
    [aux_sym_if_statement_token1] = ACTIONS(1292),
    [aux_sym_if_statement_token2] = ACTIONS(1292),
    [aux_sym_else_if_clause_token1] = ACTIONS(1292),
    [aux_sym_else_clause_token1] = ACTIONS(1292),
    [aux_sym_match_expression_token1] = ACTIONS(1292),
    [aux_sym_match_default_expression_token1] = ACTIONS(1292),
    [aux_sym_switch_statement_token1] = ACTIONS(1292),
    [aux_sym_switch_block_token1] = ACTIONS(1292),
    [anon_sym_PLUS] = ACTIONS(1292),
    [anon_sym_DASH] = ACTIONS(1292),
    [anon_sym_TILDE] = ACTIONS(1290),
    [anon_sym_BANG] = ACTIONS(1290),
    [anon_sym_AT] = ACTIONS(1290),
    [aux_sym_clone_expression_token1] = ACTIONS(1292),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1292),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1292),
    [anon_sym_DASH_DASH] = ACTIONS(1290),
    [anon_sym_PLUS_PLUS] = ACTIONS(1290),
    [aux_sym__list_destructing_token1] = ACTIONS(1292),
    [anon_sym_LBRACK] = ACTIONS(1290),
    [anon_sym_self] = ACTIONS(1292),
    [anon_sym_parent] = ACTIONS(1292),
    [aux_sym__argument_name_token1] = ACTIONS(1292),
    [aux_sym__argument_name_token2] = ACTIONS(1292),
    [anon_sym_POUND_LBRACK] = ACTIONS(1290),
    [aux_sym_encapsed_string_token1] = ACTIONS(1290),
    [anon_sym_DQUOTE] = ACTIONS(1290),
    [aux_sym_string_token1] = ACTIONS(1290),
    [anon_sym_SQUOTE] = ACTIONS(1290),
    [anon_sym_LT_LT_LT] = ACTIONS(1290),
    [anon_sym_BQUOTE] = ACTIONS(1290),
    [anon_sym_DOLLAR] = ACTIONS(1290),
    [aux_sym_yield_expression_token1] = ACTIONS(1292),
    [aux_sym_include_expression_token1] = ACTIONS(1292),
    [aux_sym_include_once_expression_token1] = ACTIONS(1292),
    [aux_sym_require_expression_token1] = ACTIONS(1292),
    [aux_sym_require_once_expression_token1] = ACTIONS(1292),
    [sym_comment] = ACTIONS(3),
  },
  [505] = {
    [ts_builtin_sym_end] = ACTIONS(1294),
    [sym_name] = ACTIONS(1296),
    [anon_sym_QMARK_GT] = ACTIONS(1294),
    [anon_sym_SEMI] = ACTIONS(1294),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1296),
    [aux_sym_global_declaration_token1] = ACTIONS(1296),
    [aux_sym_namespace_definition_token1] = ACTIONS(1296),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1296),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1296),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1296),
    [anon_sym_BSLASH] = ACTIONS(1294),
    [anon_sym_LBRACE] = ACTIONS(1294),
    [anon_sym_RBRACE] = ACTIONS(1294),
    [aux_sym_trait_declaration_token1] = ACTIONS(1296),
    [aux_sym_interface_declaration_token1] = ACTIONS(1296),
    [aux_sym_enum_declaration_token1] = ACTIONS(1296),
    [aux_sym_enum_case_token1] = ACTIONS(1296),
    [aux_sym_class_declaration_token1] = ACTIONS(1296),
    [aux_sym_final_modifier_token1] = ACTIONS(1296),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1296),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1296),
    [sym_var_modifier] = ACTIONS(1296),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1296),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1296),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1296),
    [anon_sym_LPAREN2] = ACTIONS(1294),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1296),
    [aux_sym_cast_type_token1] = ACTIONS(1296),
    [aux_sym_echo_statement_token1] = ACTIONS(1296),
    [aux_sym_exit_statement_token1] = ACTIONS(1296),
    [anon_sym_unset] = ACTIONS(1296),
    [aux_sym_declare_statement_token1] = ACTIONS(1296),
    [aux_sym_declare_statement_token2] = ACTIONS(1296),
    [sym_float] = ACTIONS(1296),
    [aux_sym_try_statement_token1] = ACTIONS(1296),
    [aux_sym_goto_statement_token1] = ACTIONS(1296),
    [aux_sym_continue_statement_token1] = ACTIONS(1296),
    [aux_sym_break_statement_token1] = ACTIONS(1296),
    [sym_integer] = ACTIONS(1296),
    [aux_sym_return_statement_token1] = ACTIONS(1296),
    [aux_sym_throw_expression_token1] = ACTIONS(1296),
    [aux_sym_while_statement_token1] = ACTIONS(1296),
    [aux_sym_while_statement_token2] = ACTIONS(1296),
    [aux_sym_do_statement_token1] = ACTIONS(1296),
    [aux_sym_for_statement_token1] = ACTIONS(1296),
    [aux_sym_for_statement_token2] = ACTIONS(1296),
    [aux_sym_foreach_statement_token1] = ACTIONS(1296),
    [aux_sym_foreach_statement_token2] = ACTIONS(1296),
    [aux_sym_if_statement_token1] = ACTIONS(1296),
    [aux_sym_if_statement_token2] = ACTIONS(1296),
    [aux_sym_else_if_clause_token1] = ACTIONS(1296),
    [aux_sym_else_clause_token1] = ACTIONS(1296),
    [aux_sym_match_expression_token1] = ACTIONS(1296),
    [aux_sym_match_default_expression_token1] = ACTIONS(1296),
    [aux_sym_switch_statement_token1] = ACTIONS(1296),
    [aux_sym_switch_block_token1] = ACTIONS(1296),
    [anon_sym_PLUS] = ACTIONS(1296),
    [anon_sym_DASH] = ACTIONS(1296),
    [anon_sym_TILDE] = ACTIONS(1294),
    [anon_sym_BANG] = ACTIONS(1294),
    [anon_sym_AT] = ACTIONS(1294),
    [aux_sym_clone_expression_token1] = ACTIONS(1296),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1296),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1296),
    [anon_sym_DASH_DASH] = ACTIONS(1294),
    [anon_sym_PLUS_PLUS] = ACTIONS(1294),
    [aux_sym__list_destructing_token1] = ACTIONS(1296),
    [anon_sym_LBRACK] = ACTIONS(1294),
    [anon_sym_self] = ACTIONS(1296),
    [anon_sym_parent] = ACTIONS(1296),
    [aux_sym__argument_name_token1] = ACTIONS(1296),
    [aux_sym__argument_name_token2] = ACTIONS(1296),
    [anon_sym_POUND_LBRACK] = ACTIONS(1294),
    [aux_sym_encapsed_string_token1] = ACTIONS(1294),
    [anon_sym_DQUOTE] = ACTIONS(1294),
    [aux_sym_string_token1] = ACTIONS(1294),
    [anon_sym_SQUOTE] = ACTIONS(1294),
    [anon_sym_LT_LT_LT] = ACTIONS(1294),
    [anon_sym_BQUOTE] = ACTIONS(1294),
    [anon_sym_DOLLAR] = ACTIONS(1294),
    [aux_sym_yield_expression_token1] = ACTIONS(1296),
    [aux_sym_include_expression_token1] = ACTIONS(1296),
    [aux_sym_include_once_expression_token1] = ACTIONS(1296),
    [aux_sym_require_expression_token1] = ACTIONS(1296),
    [aux_sym_require_once_expression_token1] = ACTIONS(1296),
    [sym_comment] = ACTIONS(3),
  },
  [506] = {
    [ts_builtin_sym_end] = ACTIONS(1298),
    [sym_name] = ACTIONS(1300),
    [anon_sym_QMARK_GT] = ACTIONS(1298),
    [anon_sym_SEMI] = ACTIONS(1298),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1300),
    [aux_sym_global_declaration_token1] = ACTIONS(1300),
    [aux_sym_namespace_definition_token1] = ACTIONS(1300),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1300),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1300),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1300),
    [anon_sym_BSLASH] = ACTIONS(1298),
    [anon_sym_LBRACE] = ACTIONS(1298),
    [anon_sym_RBRACE] = ACTIONS(1298),
    [aux_sym_trait_declaration_token1] = ACTIONS(1300),
    [aux_sym_interface_declaration_token1] = ACTIONS(1300),
    [aux_sym_enum_declaration_token1] = ACTIONS(1300),
    [aux_sym_enum_case_token1] = ACTIONS(1300),
    [aux_sym_class_declaration_token1] = ACTIONS(1300),
    [aux_sym_final_modifier_token1] = ACTIONS(1300),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1300),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1300),
    [sym_var_modifier] = ACTIONS(1300),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1300),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1300),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1300),
    [anon_sym_LPAREN2] = ACTIONS(1298),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1300),
    [aux_sym_cast_type_token1] = ACTIONS(1300),
    [aux_sym_echo_statement_token1] = ACTIONS(1300),
    [aux_sym_exit_statement_token1] = ACTIONS(1300),
    [anon_sym_unset] = ACTIONS(1300),
    [aux_sym_declare_statement_token1] = ACTIONS(1300),
    [aux_sym_declare_statement_token2] = ACTIONS(1300),
    [sym_float] = ACTIONS(1300),
    [aux_sym_try_statement_token1] = ACTIONS(1300),
    [aux_sym_goto_statement_token1] = ACTIONS(1300),
    [aux_sym_continue_statement_token1] = ACTIONS(1300),
    [aux_sym_break_statement_token1] = ACTIONS(1300),
    [sym_integer] = ACTIONS(1300),
    [aux_sym_return_statement_token1] = ACTIONS(1300),
    [aux_sym_throw_expression_token1] = ACTIONS(1300),
    [aux_sym_while_statement_token1] = ACTIONS(1300),
    [aux_sym_while_statement_token2] = ACTIONS(1300),
    [aux_sym_do_statement_token1] = ACTIONS(1300),
    [aux_sym_for_statement_token1] = ACTIONS(1300),
    [aux_sym_for_statement_token2] = ACTIONS(1300),
    [aux_sym_foreach_statement_token1] = ACTIONS(1300),
    [aux_sym_foreach_statement_token2] = ACTIONS(1300),
    [aux_sym_if_statement_token1] = ACTIONS(1300),
    [aux_sym_if_statement_token2] = ACTIONS(1300),
    [aux_sym_else_if_clause_token1] = ACTIONS(1300),
    [aux_sym_else_clause_token1] = ACTIONS(1300),
    [aux_sym_match_expression_token1] = ACTIONS(1300),
    [aux_sym_match_default_expression_token1] = ACTIONS(1300),
    [aux_sym_switch_statement_token1] = ACTIONS(1300),
    [aux_sym_switch_block_token1] = ACTIONS(1300),
    [anon_sym_PLUS] = ACTIONS(1300),
    [anon_sym_DASH] = ACTIONS(1300),
    [anon_sym_TILDE] = ACTIONS(1298),
    [anon_sym_BANG] = ACTIONS(1298),
    [anon_sym_AT] = ACTIONS(1298),
    [aux_sym_clone_expression_token1] = ACTIONS(1300),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1300),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1300),
    [anon_sym_DASH_DASH] = ACTIONS(1298),
    [anon_sym_PLUS_PLUS] = ACTIONS(1298),
    [aux_sym__list_destructing_token1] = ACTIONS(1300),
    [anon_sym_LBRACK] = ACTIONS(1298),
    [anon_sym_self] = ACTIONS(1300),
    [anon_sym_parent] = ACTIONS(1300),
    [aux_sym__argument_name_token1] = ACTIONS(1300),
    [aux_sym__argument_name_token2] = ACTIONS(1300),
    [anon_sym_POUND_LBRACK] = ACTIONS(1298),
    [aux_sym_encapsed_string_token1] = ACTIONS(1298),
    [anon_sym_DQUOTE] = ACTIONS(1298),
    [aux_sym_string_token1] = ACTIONS(1298),
    [anon_sym_SQUOTE] = ACTIONS(1298),
    [anon_sym_LT_LT_LT] = ACTIONS(1298),
    [anon_sym_BQUOTE] = ACTIONS(1298),
    [anon_sym_DOLLAR] = ACTIONS(1298),
    [aux_sym_yield_expression_token1] = ACTIONS(1300),
    [aux_sym_include_expression_token1] = ACTIONS(1300),
    [aux_sym_include_once_expression_token1] = ACTIONS(1300),
    [aux_sym_require_expression_token1] = ACTIONS(1300),
    [aux_sym_require_once_expression_token1] = ACTIONS(1300),
    [sym_comment] = ACTIONS(3),
  },
  [507] = {
    [ts_builtin_sym_end] = ACTIONS(1302),
    [sym_name] = ACTIONS(1304),
    [anon_sym_QMARK_GT] = ACTIONS(1302),
    [anon_sym_SEMI] = ACTIONS(1302),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1304),
    [aux_sym_global_declaration_token1] = ACTIONS(1304),
    [aux_sym_namespace_definition_token1] = ACTIONS(1304),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1304),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1304),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1304),
    [anon_sym_BSLASH] = ACTIONS(1302),
    [anon_sym_LBRACE] = ACTIONS(1302),
    [anon_sym_RBRACE] = ACTIONS(1302),
    [aux_sym_trait_declaration_token1] = ACTIONS(1304),
    [aux_sym_interface_declaration_token1] = ACTIONS(1304),
    [aux_sym_enum_declaration_token1] = ACTIONS(1304),
    [aux_sym_enum_case_token1] = ACTIONS(1304),
    [aux_sym_class_declaration_token1] = ACTIONS(1304),
    [aux_sym_final_modifier_token1] = ACTIONS(1304),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1304),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1304),
    [sym_var_modifier] = ACTIONS(1304),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1304),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1304),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1304),
    [anon_sym_LPAREN2] = ACTIONS(1302),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1304),
    [aux_sym_cast_type_token1] = ACTIONS(1304),
    [aux_sym_echo_statement_token1] = ACTIONS(1304),
    [aux_sym_exit_statement_token1] = ACTIONS(1304),
    [anon_sym_unset] = ACTIONS(1304),
    [aux_sym_declare_statement_token1] = ACTIONS(1304),
    [aux_sym_declare_statement_token2] = ACTIONS(1304),
    [sym_float] = ACTIONS(1304),
    [aux_sym_try_statement_token1] = ACTIONS(1304),
    [aux_sym_goto_statement_token1] = ACTIONS(1304),
    [aux_sym_continue_statement_token1] = ACTIONS(1304),
    [aux_sym_break_statement_token1] = ACTIONS(1304),
    [sym_integer] = ACTIONS(1304),
    [aux_sym_return_statement_token1] = ACTIONS(1304),
    [aux_sym_throw_expression_token1] = ACTIONS(1304),
    [aux_sym_while_statement_token1] = ACTIONS(1304),
    [aux_sym_while_statement_token2] = ACTIONS(1304),
    [aux_sym_do_statement_token1] = ACTIONS(1304),
    [aux_sym_for_statement_token1] = ACTIONS(1304),
    [aux_sym_for_statement_token2] = ACTIONS(1304),
    [aux_sym_foreach_statement_token1] = ACTIONS(1304),
    [aux_sym_foreach_statement_token2] = ACTIONS(1304),
    [aux_sym_if_statement_token1] = ACTIONS(1304),
    [aux_sym_if_statement_token2] = ACTIONS(1304),
    [aux_sym_else_if_clause_token1] = ACTIONS(1304),
    [aux_sym_else_clause_token1] = ACTIONS(1304),
    [aux_sym_match_expression_token1] = ACTIONS(1304),
    [aux_sym_match_default_expression_token1] = ACTIONS(1304),
    [aux_sym_switch_statement_token1] = ACTIONS(1304),
    [aux_sym_switch_block_token1] = ACTIONS(1304),
    [anon_sym_PLUS] = ACTIONS(1304),
    [anon_sym_DASH] = ACTIONS(1304),
    [anon_sym_TILDE] = ACTIONS(1302),
    [anon_sym_BANG] = ACTIONS(1302),
    [anon_sym_AT] = ACTIONS(1302),
    [aux_sym_clone_expression_token1] = ACTIONS(1304),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1304),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1304),
    [anon_sym_DASH_DASH] = ACTIONS(1302),
    [anon_sym_PLUS_PLUS] = ACTIONS(1302),
    [aux_sym__list_destructing_token1] = ACTIONS(1304),
    [anon_sym_LBRACK] = ACTIONS(1302),
    [anon_sym_self] = ACTIONS(1304),
    [anon_sym_parent] = ACTIONS(1304),
    [aux_sym__argument_name_token1] = ACTIONS(1304),
    [aux_sym__argument_name_token2] = ACTIONS(1304),
    [anon_sym_POUND_LBRACK] = ACTIONS(1302),
    [aux_sym_encapsed_string_token1] = ACTIONS(1302),
    [anon_sym_DQUOTE] = ACTIONS(1302),
    [aux_sym_string_token1] = ACTIONS(1302),
    [anon_sym_SQUOTE] = ACTIONS(1302),
    [anon_sym_LT_LT_LT] = ACTIONS(1302),
    [anon_sym_BQUOTE] = ACTIONS(1302),
    [anon_sym_DOLLAR] = ACTIONS(1302),
    [aux_sym_yield_expression_token1] = ACTIONS(1304),
    [aux_sym_include_expression_token1] = ACTIONS(1304),
    [aux_sym_include_once_expression_token1] = ACTIONS(1304),
    [aux_sym_require_expression_token1] = ACTIONS(1304),
    [aux_sym_require_once_expression_token1] = ACTIONS(1304),
    [sym_comment] = ACTIONS(3),
  },
  [508] = {
    [ts_builtin_sym_end] = ACTIONS(1306),
    [sym_name] = ACTIONS(1308),
    [anon_sym_QMARK_GT] = ACTIONS(1306),
    [anon_sym_SEMI] = ACTIONS(1306),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1308),
    [aux_sym_global_declaration_token1] = ACTIONS(1308),
    [aux_sym_namespace_definition_token1] = ACTIONS(1308),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1308),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1308),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1308),
    [anon_sym_BSLASH] = ACTIONS(1306),
    [anon_sym_LBRACE] = ACTIONS(1306),
    [anon_sym_RBRACE] = ACTIONS(1306),
    [aux_sym_trait_declaration_token1] = ACTIONS(1308),
    [aux_sym_interface_declaration_token1] = ACTIONS(1308),
    [aux_sym_enum_declaration_token1] = ACTIONS(1308),
    [aux_sym_enum_case_token1] = ACTIONS(1308),
    [aux_sym_class_declaration_token1] = ACTIONS(1308),
    [aux_sym_final_modifier_token1] = ACTIONS(1308),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1308),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1308),
    [sym_var_modifier] = ACTIONS(1308),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1308),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1308),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1308),
    [anon_sym_LPAREN2] = ACTIONS(1306),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1308),
    [aux_sym_cast_type_token1] = ACTIONS(1308),
    [aux_sym_echo_statement_token1] = ACTIONS(1308),
    [aux_sym_exit_statement_token1] = ACTIONS(1308),
    [anon_sym_unset] = ACTIONS(1308),
    [aux_sym_declare_statement_token1] = ACTIONS(1308),
    [aux_sym_declare_statement_token2] = ACTIONS(1308),
    [sym_float] = ACTIONS(1308),
    [aux_sym_try_statement_token1] = ACTIONS(1308),
    [aux_sym_goto_statement_token1] = ACTIONS(1308),
    [aux_sym_continue_statement_token1] = ACTIONS(1308),
    [aux_sym_break_statement_token1] = ACTIONS(1308),
    [sym_integer] = ACTIONS(1308),
    [aux_sym_return_statement_token1] = ACTIONS(1308),
    [aux_sym_throw_expression_token1] = ACTIONS(1308),
    [aux_sym_while_statement_token1] = ACTIONS(1308),
    [aux_sym_while_statement_token2] = ACTIONS(1308),
    [aux_sym_do_statement_token1] = ACTIONS(1308),
    [aux_sym_for_statement_token1] = ACTIONS(1308),
    [aux_sym_for_statement_token2] = ACTIONS(1308),
    [aux_sym_foreach_statement_token1] = ACTIONS(1308),
    [aux_sym_foreach_statement_token2] = ACTIONS(1308),
    [aux_sym_if_statement_token1] = ACTIONS(1308),
    [aux_sym_if_statement_token2] = ACTIONS(1308),
    [aux_sym_else_if_clause_token1] = ACTIONS(1308),
    [aux_sym_else_clause_token1] = ACTIONS(1308),
    [aux_sym_match_expression_token1] = ACTIONS(1308),
    [aux_sym_match_default_expression_token1] = ACTIONS(1308),
    [aux_sym_switch_statement_token1] = ACTIONS(1308),
    [aux_sym_switch_block_token1] = ACTIONS(1308),
    [anon_sym_PLUS] = ACTIONS(1308),
    [anon_sym_DASH] = ACTIONS(1308),
    [anon_sym_TILDE] = ACTIONS(1306),
    [anon_sym_BANG] = ACTIONS(1306),
    [anon_sym_AT] = ACTIONS(1306),
    [aux_sym_clone_expression_token1] = ACTIONS(1308),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1308),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1306),
    [anon_sym_PLUS_PLUS] = ACTIONS(1306),
    [aux_sym__list_destructing_token1] = ACTIONS(1308),
    [anon_sym_LBRACK] = ACTIONS(1306),
    [anon_sym_self] = ACTIONS(1308),
    [anon_sym_parent] = ACTIONS(1308),
    [aux_sym__argument_name_token1] = ACTIONS(1308),
    [aux_sym__argument_name_token2] = ACTIONS(1308),
    [anon_sym_POUND_LBRACK] = ACTIONS(1306),
    [aux_sym_encapsed_string_token1] = ACTIONS(1306),
    [anon_sym_DQUOTE] = ACTIONS(1306),
    [aux_sym_string_token1] = ACTIONS(1306),
    [anon_sym_SQUOTE] = ACTIONS(1306),
    [anon_sym_LT_LT_LT] = ACTIONS(1306),
    [anon_sym_BQUOTE] = ACTIONS(1306),
    [anon_sym_DOLLAR] = ACTIONS(1306),
    [aux_sym_yield_expression_token1] = ACTIONS(1308),
    [aux_sym_include_expression_token1] = ACTIONS(1308),
    [aux_sym_include_once_expression_token1] = ACTIONS(1308),
    [aux_sym_require_expression_token1] = ACTIONS(1308),
    [aux_sym_require_once_expression_token1] = ACTIONS(1308),
    [sym_comment] = ACTIONS(3),
  },
  [509] = {
    [ts_builtin_sym_end] = ACTIONS(1310),
    [sym_name] = ACTIONS(1312),
    [anon_sym_QMARK_GT] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1310),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1312),
    [aux_sym_global_declaration_token1] = ACTIONS(1312),
    [aux_sym_namespace_definition_token1] = ACTIONS(1312),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1312),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1312),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1312),
    [anon_sym_BSLASH] = ACTIONS(1310),
    [anon_sym_LBRACE] = ACTIONS(1310),
    [anon_sym_RBRACE] = ACTIONS(1310),
    [aux_sym_trait_declaration_token1] = ACTIONS(1312),
    [aux_sym_interface_declaration_token1] = ACTIONS(1312),
    [aux_sym_enum_declaration_token1] = ACTIONS(1312),
    [aux_sym_enum_case_token1] = ACTIONS(1312),
    [aux_sym_class_declaration_token1] = ACTIONS(1312),
    [aux_sym_final_modifier_token1] = ACTIONS(1312),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1312),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1312),
    [sym_var_modifier] = ACTIONS(1312),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1312),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1312),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1312),
    [anon_sym_LPAREN2] = ACTIONS(1310),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1312),
    [aux_sym_cast_type_token1] = ACTIONS(1312),
    [aux_sym_echo_statement_token1] = ACTIONS(1312),
    [aux_sym_exit_statement_token1] = ACTIONS(1312),
    [anon_sym_unset] = ACTIONS(1312),
    [aux_sym_declare_statement_token1] = ACTIONS(1312),
    [aux_sym_declare_statement_token2] = ACTIONS(1312),
    [sym_float] = ACTIONS(1312),
    [aux_sym_try_statement_token1] = ACTIONS(1312),
    [aux_sym_goto_statement_token1] = ACTIONS(1312),
    [aux_sym_continue_statement_token1] = ACTIONS(1312),
    [aux_sym_break_statement_token1] = ACTIONS(1312),
    [sym_integer] = ACTIONS(1312),
    [aux_sym_return_statement_token1] = ACTIONS(1312),
    [aux_sym_throw_expression_token1] = ACTIONS(1312),
    [aux_sym_while_statement_token1] = ACTIONS(1312),
    [aux_sym_while_statement_token2] = ACTIONS(1312),
    [aux_sym_do_statement_token1] = ACTIONS(1312),
    [aux_sym_for_statement_token1] = ACTIONS(1312),
    [aux_sym_for_statement_token2] = ACTIONS(1312),
    [aux_sym_foreach_statement_token1] = ACTIONS(1312),
    [aux_sym_foreach_statement_token2] = ACTIONS(1312),
    [aux_sym_if_statement_token1] = ACTIONS(1312),
    [aux_sym_if_statement_token2] = ACTIONS(1312),
    [aux_sym_else_if_clause_token1] = ACTIONS(1312),
    [aux_sym_else_clause_token1] = ACTIONS(1312),
    [aux_sym_match_expression_token1] = ACTIONS(1312),
    [aux_sym_match_default_expression_token1] = ACTIONS(1312),
    [aux_sym_switch_statement_token1] = ACTIONS(1312),
    [aux_sym_switch_block_token1] = ACTIONS(1312),
    [anon_sym_PLUS] = ACTIONS(1312),
    [anon_sym_DASH] = ACTIONS(1312),
    [anon_sym_TILDE] = ACTIONS(1310),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [aux_sym_clone_expression_token1] = ACTIONS(1312),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1312),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1312),
    [anon_sym_DASH_DASH] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1310),
    [aux_sym__list_destructing_token1] = ACTIONS(1312),
    [anon_sym_LBRACK] = ACTIONS(1310),
    [anon_sym_self] = ACTIONS(1312),
    [anon_sym_parent] = ACTIONS(1312),
    [aux_sym__argument_name_token1] = ACTIONS(1312),
    [aux_sym__argument_name_token2] = ACTIONS(1312),
    [anon_sym_POUND_LBRACK] = ACTIONS(1310),
    [aux_sym_encapsed_string_token1] = ACTIONS(1310),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [aux_sym_string_token1] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_LT_LT_LT] = ACTIONS(1310),
    [anon_sym_BQUOTE] = ACTIONS(1310),
    [anon_sym_DOLLAR] = ACTIONS(1310),
    [aux_sym_yield_expression_token1] = ACTIONS(1312),
    [aux_sym_include_expression_token1] = ACTIONS(1312),
    [aux_sym_include_once_expression_token1] = ACTIONS(1312),
    [aux_sym_require_expression_token1] = ACTIONS(1312),
    [aux_sym_require_once_expression_token1] = ACTIONS(1312),
    [sym_comment] = ACTIONS(3),
  },
  [510] = {
    [ts_builtin_sym_end] = ACTIONS(1314),
    [sym_name] = ACTIONS(1316),
    [anon_sym_QMARK_GT] = ACTIONS(1314),
    [anon_sym_SEMI] = ACTIONS(1314),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1316),
    [aux_sym_global_declaration_token1] = ACTIONS(1316),
    [aux_sym_namespace_definition_token1] = ACTIONS(1316),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1316),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1316),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1316),
    [anon_sym_BSLASH] = ACTIONS(1314),
    [anon_sym_LBRACE] = ACTIONS(1314),
    [anon_sym_RBRACE] = ACTIONS(1314),
    [aux_sym_trait_declaration_token1] = ACTIONS(1316),
    [aux_sym_interface_declaration_token1] = ACTIONS(1316),
    [aux_sym_enum_declaration_token1] = ACTIONS(1316),
    [aux_sym_enum_case_token1] = ACTIONS(1316),
    [aux_sym_class_declaration_token1] = ACTIONS(1316),
    [aux_sym_final_modifier_token1] = ACTIONS(1316),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1316),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1316),
    [sym_var_modifier] = ACTIONS(1316),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1316),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1316),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1316),
    [anon_sym_LPAREN2] = ACTIONS(1314),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1316),
    [aux_sym_cast_type_token1] = ACTIONS(1316),
    [aux_sym_echo_statement_token1] = ACTIONS(1316),
    [aux_sym_exit_statement_token1] = ACTIONS(1316),
    [anon_sym_unset] = ACTIONS(1316),
    [aux_sym_declare_statement_token1] = ACTIONS(1316),
    [aux_sym_declare_statement_token2] = ACTIONS(1316),
    [sym_float] = ACTIONS(1316),
    [aux_sym_try_statement_token1] = ACTIONS(1316),
    [aux_sym_goto_statement_token1] = ACTIONS(1316),
    [aux_sym_continue_statement_token1] = ACTIONS(1316),
    [aux_sym_break_statement_token1] = ACTIONS(1316),
    [sym_integer] = ACTIONS(1316),
    [aux_sym_return_statement_token1] = ACTIONS(1316),
    [aux_sym_throw_expression_token1] = ACTIONS(1316),
    [aux_sym_while_statement_token1] = ACTIONS(1316),
    [aux_sym_while_statement_token2] = ACTIONS(1316),
    [aux_sym_do_statement_token1] = ACTIONS(1316),
    [aux_sym_for_statement_token1] = ACTIONS(1316),
    [aux_sym_for_statement_token2] = ACTIONS(1316),
    [aux_sym_foreach_statement_token1] = ACTIONS(1316),
    [aux_sym_foreach_statement_token2] = ACTIONS(1316),
    [aux_sym_if_statement_token1] = ACTIONS(1316),
    [aux_sym_if_statement_token2] = ACTIONS(1316),
    [aux_sym_else_if_clause_token1] = ACTIONS(1316),
    [aux_sym_else_clause_token1] = ACTIONS(1316),
    [aux_sym_match_expression_token1] = ACTIONS(1316),
    [aux_sym_match_default_expression_token1] = ACTIONS(1316),
    [aux_sym_switch_statement_token1] = ACTIONS(1316),
    [aux_sym_switch_block_token1] = ACTIONS(1316),
    [anon_sym_PLUS] = ACTIONS(1316),
    [anon_sym_DASH] = ACTIONS(1316),
    [anon_sym_TILDE] = ACTIONS(1314),
    [anon_sym_BANG] = ACTIONS(1314),
    [anon_sym_AT] = ACTIONS(1314),
    [aux_sym_clone_expression_token1] = ACTIONS(1316),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1316),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1316),
    [anon_sym_DASH_DASH] = ACTIONS(1314),
    [anon_sym_PLUS_PLUS] = ACTIONS(1314),
    [aux_sym__list_destructing_token1] = ACTIONS(1316),
    [anon_sym_LBRACK] = ACTIONS(1314),
    [anon_sym_self] = ACTIONS(1316),
    [anon_sym_parent] = ACTIONS(1316),
    [aux_sym__argument_name_token1] = ACTIONS(1316),
    [aux_sym__argument_name_token2] = ACTIONS(1316),
    [anon_sym_POUND_LBRACK] = ACTIONS(1314),
    [aux_sym_encapsed_string_token1] = ACTIONS(1314),
    [anon_sym_DQUOTE] = ACTIONS(1314),
    [aux_sym_string_token1] = ACTIONS(1314),
    [anon_sym_SQUOTE] = ACTIONS(1314),
    [anon_sym_LT_LT_LT] = ACTIONS(1314),
    [anon_sym_BQUOTE] = ACTIONS(1314),
    [anon_sym_DOLLAR] = ACTIONS(1314),
    [aux_sym_yield_expression_token1] = ACTIONS(1316),
    [aux_sym_include_expression_token1] = ACTIONS(1316),
    [aux_sym_include_once_expression_token1] = ACTIONS(1316),
    [aux_sym_require_expression_token1] = ACTIONS(1316),
    [aux_sym_require_once_expression_token1] = ACTIONS(1316),
    [sym_comment] = ACTIONS(3),
  },
  [511] = {
    [ts_builtin_sym_end] = ACTIONS(1318),
    [sym_name] = ACTIONS(1320),
    [anon_sym_QMARK_GT] = ACTIONS(1318),
    [anon_sym_SEMI] = ACTIONS(1318),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1320),
    [aux_sym_global_declaration_token1] = ACTIONS(1320),
    [aux_sym_namespace_definition_token1] = ACTIONS(1320),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1320),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1320),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1320),
    [anon_sym_BSLASH] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1318),
    [anon_sym_RBRACE] = ACTIONS(1318),
    [aux_sym_trait_declaration_token1] = ACTIONS(1320),
    [aux_sym_interface_declaration_token1] = ACTIONS(1320),
    [aux_sym_enum_declaration_token1] = ACTIONS(1320),
    [aux_sym_enum_case_token1] = ACTIONS(1320),
    [aux_sym_class_declaration_token1] = ACTIONS(1320),
    [aux_sym_final_modifier_token1] = ACTIONS(1320),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1320),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1320),
    [sym_var_modifier] = ACTIONS(1320),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1320),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1320),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1320),
    [anon_sym_LPAREN2] = ACTIONS(1318),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1320),
    [aux_sym_cast_type_token1] = ACTIONS(1320),
    [aux_sym_echo_statement_token1] = ACTIONS(1320),
    [aux_sym_exit_statement_token1] = ACTIONS(1320),
    [anon_sym_unset] = ACTIONS(1320),
    [aux_sym_declare_statement_token1] = ACTIONS(1320),
    [aux_sym_declare_statement_token2] = ACTIONS(1320),
    [sym_float] = ACTIONS(1320),
    [aux_sym_try_statement_token1] = ACTIONS(1320),
    [aux_sym_goto_statement_token1] = ACTIONS(1320),
    [aux_sym_continue_statement_token1] = ACTIONS(1320),
    [aux_sym_break_statement_token1] = ACTIONS(1320),
    [sym_integer] = ACTIONS(1320),
    [aux_sym_return_statement_token1] = ACTIONS(1320),
    [aux_sym_throw_expression_token1] = ACTIONS(1320),
    [aux_sym_while_statement_token1] = ACTIONS(1320),
    [aux_sym_while_statement_token2] = ACTIONS(1320),
    [aux_sym_do_statement_token1] = ACTIONS(1320),
    [aux_sym_for_statement_token1] = ACTIONS(1320),
    [aux_sym_for_statement_token2] = ACTIONS(1320),
    [aux_sym_foreach_statement_token1] = ACTIONS(1320),
    [aux_sym_foreach_statement_token2] = ACTIONS(1320),
    [aux_sym_if_statement_token1] = ACTIONS(1320),
    [aux_sym_if_statement_token2] = ACTIONS(1320),
    [aux_sym_else_if_clause_token1] = ACTIONS(1320),
    [aux_sym_else_clause_token1] = ACTIONS(1320),
    [aux_sym_match_expression_token1] = ACTIONS(1320),
    [aux_sym_match_default_expression_token1] = ACTIONS(1320),
    [aux_sym_switch_statement_token1] = ACTIONS(1320),
    [aux_sym_switch_block_token1] = ACTIONS(1320),
    [anon_sym_PLUS] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1320),
    [anon_sym_TILDE] = ACTIONS(1318),
    [anon_sym_BANG] = ACTIONS(1318),
    [anon_sym_AT] = ACTIONS(1318),
    [aux_sym_clone_expression_token1] = ACTIONS(1320),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1320),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1320),
    [anon_sym_DASH_DASH] = ACTIONS(1318),
    [anon_sym_PLUS_PLUS] = ACTIONS(1318),
    [aux_sym__list_destructing_token1] = ACTIONS(1320),
    [anon_sym_LBRACK] = ACTIONS(1318),
    [anon_sym_self] = ACTIONS(1320),
    [anon_sym_parent] = ACTIONS(1320),
    [aux_sym__argument_name_token1] = ACTIONS(1320),
    [aux_sym__argument_name_token2] = ACTIONS(1320),
    [anon_sym_POUND_LBRACK] = ACTIONS(1318),
    [aux_sym_encapsed_string_token1] = ACTIONS(1318),
    [anon_sym_DQUOTE] = ACTIONS(1318),
    [aux_sym_string_token1] = ACTIONS(1318),
    [anon_sym_SQUOTE] = ACTIONS(1318),
    [anon_sym_LT_LT_LT] = ACTIONS(1318),
    [anon_sym_BQUOTE] = ACTIONS(1318),
    [anon_sym_DOLLAR] = ACTIONS(1318),
    [aux_sym_yield_expression_token1] = ACTIONS(1320),
    [aux_sym_include_expression_token1] = ACTIONS(1320),
    [aux_sym_include_once_expression_token1] = ACTIONS(1320),
    [aux_sym_require_expression_token1] = ACTIONS(1320),
    [aux_sym_require_once_expression_token1] = ACTIONS(1320),
    [sym_comment] = ACTIONS(3),
  },
  [512] = {
    [ts_builtin_sym_end] = ACTIONS(1322),
    [sym_name] = ACTIONS(1324),
    [anon_sym_QMARK_GT] = ACTIONS(1322),
    [anon_sym_SEMI] = ACTIONS(1322),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1324),
    [aux_sym_global_declaration_token1] = ACTIONS(1324),
    [aux_sym_namespace_definition_token1] = ACTIONS(1324),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1324),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1324),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1324),
    [anon_sym_BSLASH] = ACTIONS(1322),
    [anon_sym_LBRACE] = ACTIONS(1322),
    [anon_sym_RBRACE] = ACTIONS(1322),
    [aux_sym_trait_declaration_token1] = ACTIONS(1324),
    [aux_sym_interface_declaration_token1] = ACTIONS(1324),
    [aux_sym_enum_declaration_token1] = ACTIONS(1324),
    [aux_sym_enum_case_token1] = ACTIONS(1324),
    [aux_sym_class_declaration_token1] = ACTIONS(1324),
    [aux_sym_final_modifier_token1] = ACTIONS(1324),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1324),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1324),
    [sym_var_modifier] = ACTIONS(1324),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1324),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1324),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1324),
    [anon_sym_LPAREN2] = ACTIONS(1322),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1324),
    [aux_sym_cast_type_token1] = ACTIONS(1324),
    [aux_sym_echo_statement_token1] = ACTIONS(1324),
    [aux_sym_exit_statement_token1] = ACTIONS(1324),
    [anon_sym_unset] = ACTIONS(1324),
    [aux_sym_declare_statement_token1] = ACTIONS(1324),
    [aux_sym_declare_statement_token2] = ACTIONS(1324),
    [sym_float] = ACTIONS(1324),
    [aux_sym_try_statement_token1] = ACTIONS(1324),
    [aux_sym_goto_statement_token1] = ACTIONS(1324),
    [aux_sym_continue_statement_token1] = ACTIONS(1324),
    [aux_sym_break_statement_token1] = ACTIONS(1324),
    [sym_integer] = ACTIONS(1324),
    [aux_sym_return_statement_token1] = ACTIONS(1324),
    [aux_sym_throw_expression_token1] = ACTIONS(1324),
    [aux_sym_while_statement_token1] = ACTIONS(1324),
    [aux_sym_while_statement_token2] = ACTIONS(1324),
    [aux_sym_do_statement_token1] = ACTIONS(1324),
    [aux_sym_for_statement_token1] = ACTIONS(1324),
    [aux_sym_for_statement_token2] = ACTIONS(1324),
    [aux_sym_foreach_statement_token1] = ACTIONS(1324),
    [aux_sym_foreach_statement_token2] = ACTIONS(1324),
    [aux_sym_if_statement_token1] = ACTIONS(1324),
    [aux_sym_if_statement_token2] = ACTIONS(1324),
    [aux_sym_else_if_clause_token1] = ACTIONS(1324),
    [aux_sym_else_clause_token1] = ACTIONS(1324),
    [aux_sym_match_expression_token1] = ACTIONS(1324),
    [aux_sym_match_default_expression_token1] = ACTIONS(1324),
    [aux_sym_switch_statement_token1] = ACTIONS(1324),
    [aux_sym_switch_block_token1] = ACTIONS(1324),
    [anon_sym_PLUS] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1324),
    [anon_sym_TILDE] = ACTIONS(1322),
    [anon_sym_BANG] = ACTIONS(1322),
    [anon_sym_AT] = ACTIONS(1322),
    [aux_sym_clone_expression_token1] = ACTIONS(1324),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1324),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1324),
    [anon_sym_DASH_DASH] = ACTIONS(1322),
    [anon_sym_PLUS_PLUS] = ACTIONS(1322),
    [aux_sym__list_destructing_token1] = ACTIONS(1324),
    [anon_sym_LBRACK] = ACTIONS(1322),
    [anon_sym_self] = ACTIONS(1324),
    [anon_sym_parent] = ACTIONS(1324),
    [aux_sym__argument_name_token1] = ACTIONS(1324),
    [aux_sym__argument_name_token2] = ACTIONS(1324),
    [anon_sym_POUND_LBRACK] = ACTIONS(1322),
    [aux_sym_encapsed_string_token1] = ACTIONS(1322),
    [anon_sym_DQUOTE] = ACTIONS(1322),
    [aux_sym_string_token1] = ACTIONS(1322),
    [anon_sym_SQUOTE] = ACTIONS(1322),
    [anon_sym_LT_LT_LT] = ACTIONS(1322),
    [anon_sym_BQUOTE] = ACTIONS(1322),
    [anon_sym_DOLLAR] = ACTIONS(1322),
    [aux_sym_yield_expression_token1] = ACTIONS(1324),
    [aux_sym_include_expression_token1] = ACTIONS(1324),
    [aux_sym_include_once_expression_token1] = ACTIONS(1324),
    [aux_sym_require_expression_token1] = ACTIONS(1324),
    [aux_sym_require_once_expression_token1] = ACTIONS(1324),
    [sym_comment] = ACTIONS(3),
  },
  [513] = {
    [ts_builtin_sym_end] = ACTIONS(1326),
    [sym_name] = ACTIONS(1328),
    [anon_sym_QMARK_GT] = ACTIONS(1326),
    [anon_sym_SEMI] = ACTIONS(1326),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1328),
    [aux_sym_global_declaration_token1] = ACTIONS(1328),
    [aux_sym_namespace_definition_token1] = ACTIONS(1328),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1328),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1328),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1328),
    [anon_sym_BSLASH] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1326),
    [anon_sym_RBRACE] = ACTIONS(1326),
    [aux_sym_trait_declaration_token1] = ACTIONS(1328),
    [aux_sym_interface_declaration_token1] = ACTIONS(1328),
    [aux_sym_enum_declaration_token1] = ACTIONS(1328),
    [aux_sym_enum_case_token1] = ACTIONS(1328),
    [aux_sym_class_declaration_token1] = ACTIONS(1328),
    [aux_sym_final_modifier_token1] = ACTIONS(1328),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1328),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1328),
    [sym_var_modifier] = ACTIONS(1328),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1328),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1328),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1328),
    [anon_sym_LPAREN2] = ACTIONS(1326),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1328),
    [aux_sym_cast_type_token1] = ACTIONS(1328),
    [aux_sym_echo_statement_token1] = ACTIONS(1328),
    [aux_sym_exit_statement_token1] = ACTIONS(1328),
    [anon_sym_unset] = ACTIONS(1328),
    [aux_sym_declare_statement_token1] = ACTIONS(1328),
    [aux_sym_declare_statement_token2] = ACTIONS(1328),
    [sym_float] = ACTIONS(1328),
    [aux_sym_try_statement_token1] = ACTIONS(1328),
    [aux_sym_goto_statement_token1] = ACTIONS(1328),
    [aux_sym_continue_statement_token1] = ACTIONS(1328),
    [aux_sym_break_statement_token1] = ACTIONS(1328),
    [sym_integer] = ACTIONS(1328),
    [aux_sym_return_statement_token1] = ACTIONS(1328),
    [aux_sym_throw_expression_token1] = ACTIONS(1328),
    [aux_sym_while_statement_token1] = ACTIONS(1328),
    [aux_sym_while_statement_token2] = ACTIONS(1328),
    [aux_sym_do_statement_token1] = ACTIONS(1328),
    [aux_sym_for_statement_token1] = ACTIONS(1328),
    [aux_sym_for_statement_token2] = ACTIONS(1328),
    [aux_sym_foreach_statement_token1] = ACTIONS(1328),
    [aux_sym_foreach_statement_token2] = ACTIONS(1328),
    [aux_sym_if_statement_token1] = ACTIONS(1328),
    [aux_sym_if_statement_token2] = ACTIONS(1328),
    [aux_sym_else_if_clause_token1] = ACTIONS(1328),
    [aux_sym_else_clause_token1] = ACTIONS(1328),
    [aux_sym_match_expression_token1] = ACTIONS(1328),
    [aux_sym_match_default_expression_token1] = ACTIONS(1328),
    [aux_sym_switch_statement_token1] = ACTIONS(1328),
    [aux_sym_switch_block_token1] = ACTIONS(1328),
    [anon_sym_PLUS] = ACTIONS(1328),
    [anon_sym_DASH] = ACTIONS(1328),
    [anon_sym_TILDE] = ACTIONS(1326),
    [anon_sym_BANG] = ACTIONS(1326),
    [anon_sym_AT] = ACTIONS(1326),
    [aux_sym_clone_expression_token1] = ACTIONS(1328),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1328),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1328),
    [anon_sym_DASH_DASH] = ACTIONS(1326),
    [anon_sym_PLUS_PLUS] = ACTIONS(1326),
    [aux_sym__list_destructing_token1] = ACTIONS(1328),
    [anon_sym_LBRACK] = ACTIONS(1326),
    [anon_sym_self] = ACTIONS(1328),
    [anon_sym_parent] = ACTIONS(1328),
    [aux_sym__argument_name_token1] = ACTIONS(1328),
    [aux_sym__argument_name_token2] = ACTIONS(1328),
    [anon_sym_POUND_LBRACK] = ACTIONS(1326),
    [aux_sym_encapsed_string_token1] = ACTIONS(1326),
    [anon_sym_DQUOTE] = ACTIONS(1326),
    [aux_sym_string_token1] = ACTIONS(1326),
    [anon_sym_SQUOTE] = ACTIONS(1326),
    [anon_sym_LT_LT_LT] = ACTIONS(1326),
    [anon_sym_BQUOTE] = ACTIONS(1326),
    [anon_sym_DOLLAR] = ACTIONS(1326),
    [aux_sym_yield_expression_token1] = ACTIONS(1328),
    [aux_sym_include_expression_token1] = ACTIONS(1328),
    [aux_sym_include_once_expression_token1] = ACTIONS(1328),
    [aux_sym_require_expression_token1] = ACTIONS(1328),
    [aux_sym_require_once_expression_token1] = ACTIONS(1328),
    [sym_comment] = ACTIONS(3),
  },
  [514] = {
    [ts_builtin_sym_end] = ACTIONS(1330),
    [sym_name] = ACTIONS(1332),
    [anon_sym_QMARK_GT] = ACTIONS(1330),
    [anon_sym_SEMI] = ACTIONS(1330),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1332),
    [aux_sym_global_declaration_token1] = ACTIONS(1332),
    [aux_sym_namespace_definition_token1] = ACTIONS(1332),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1332),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1332),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1332),
    [anon_sym_BSLASH] = ACTIONS(1330),
    [anon_sym_LBRACE] = ACTIONS(1330),
    [anon_sym_RBRACE] = ACTIONS(1330),
    [aux_sym_trait_declaration_token1] = ACTIONS(1332),
    [aux_sym_interface_declaration_token1] = ACTIONS(1332),
    [aux_sym_enum_declaration_token1] = ACTIONS(1332),
    [aux_sym_enum_case_token1] = ACTIONS(1332),
    [aux_sym_class_declaration_token1] = ACTIONS(1332),
    [aux_sym_final_modifier_token1] = ACTIONS(1332),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1332),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1332),
    [sym_var_modifier] = ACTIONS(1332),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1332),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1332),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1332),
    [anon_sym_LPAREN2] = ACTIONS(1330),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1332),
    [aux_sym_cast_type_token1] = ACTIONS(1332),
    [aux_sym_echo_statement_token1] = ACTIONS(1332),
    [aux_sym_exit_statement_token1] = ACTIONS(1332),
    [anon_sym_unset] = ACTIONS(1332),
    [aux_sym_declare_statement_token1] = ACTIONS(1332),
    [aux_sym_declare_statement_token2] = ACTIONS(1332),
    [sym_float] = ACTIONS(1332),
    [aux_sym_try_statement_token1] = ACTIONS(1332),
    [aux_sym_goto_statement_token1] = ACTIONS(1332),
    [aux_sym_continue_statement_token1] = ACTIONS(1332),
    [aux_sym_break_statement_token1] = ACTIONS(1332),
    [sym_integer] = ACTIONS(1332),
    [aux_sym_return_statement_token1] = ACTIONS(1332),
    [aux_sym_throw_expression_token1] = ACTIONS(1332),
    [aux_sym_while_statement_token1] = ACTIONS(1332),
    [aux_sym_while_statement_token2] = ACTIONS(1332),
    [aux_sym_do_statement_token1] = ACTIONS(1332),
    [aux_sym_for_statement_token1] = ACTIONS(1332),
    [aux_sym_for_statement_token2] = ACTIONS(1332),
    [aux_sym_foreach_statement_token1] = ACTIONS(1332),
    [aux_sym_foreach_statement_token2] = ACTIONS(1332),
    [aux_sym_if_statement_token1] = ACTIONS(1332),
    [aux_sym_if_statement_token2] = ACTIONS(1332),
    [aux_sym_else_if_clause_token1] = ACTIONS(1332),
    [aux_sym_else_clause_token1] = ACTIONS(1332),
    [aux_sym_match_expression_token1] = ACTIONS(1332),
    [aux_sym_match_default_expression_token1] = ACTIONS(1332),
    [aux_sym_switch_statement_token1] = ACTIONS(1332),
    [aux_sym_switch_block_token1] = ACTIONS(1332),
    [anon_sym_PLUS] = ACTIONS(1332),
    [anon_sym_DASH] = ACTIONS(1332),
    [anon_sym_TILDE] = ACTIONS(1330),
    [anon_sym_BANG] = ACTIONS(1330),
    [anon_sym_AT] = ACTIONS(1330),
    [aux_sym_clone_expression_token1] = ACTIONS(1332),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1332),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1332),
    [anon_sym_DASH_DASH] = ACTIONS(1330),
    [anon_sym_PLUS_PLUS] = ACTIONS(1330),
    [aux_sym__list_destructing_token1] = ACTIONS(1332),
    [anon_sym_LBRACK] = ACTIONS(1330),
    [anon_sym_self] = ACTIONS(1332),
    [anon_sym_parent] = ACTIONS(1332),
    [aux_sym__argument_name_token1] = ACTIONS(1332),
    [aux_sym__argument_name_token2] = ACTIONS(1332),
    [anon_sym_POUND_LBRACK] = ACTIONS(1330),
    [aux_sym_encapsed_string_token1] = ACTIONS(1330),
    [anon_sym_DQUOTE] = ACTIONS(1330),
    [aux_sym_string_token1] = ACTIONS(1330),
    [anon_sym_SQUOTE] = ACTIONS(1330),
    [anon_sym_LT_LT_LT] = ACTIONS(1330),
    [anon_sym_BQUOTE] = ACTIONS(1330),
    [anon_sym_DOLLAR] = ACTIONS(1330),
    [aux_sym_yield_expression_token1] = ACTIONS(1332),
    [aux_sym_include_expression_token1] = ACTIONS(1332),
    [aux_sym_include_once_expression_token1] = ACTIONS(1332),
    [aux_sym_require_expression_token1] = ACTIONS(1332),
    [aux_sym_require_once_expression_token1] = ACTIONS(1332),
    [sym_comment] = ACTIONS(3),
  },
  [515] = {
    [ts_builtin_sym_end] = ACTIONS(1334),
    [sym_name] = ACTIONS(1336),
    [anon_sym_QMARK_GT] = ACTIONS(1334),
    [anon_sym_SEMI] = ACTIONS(1334),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1336),
    [aux_sym_global_declaration_token1] = ACTIONS(1336),
    [aux_sym_namespace_definition_token1] = ACTIONS(1336),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1336),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1336),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1336),
    [anon_sym_BSLASH] = ACTIONS(1334),
    [anon_sym_LBRACE] = ACTIONS(1334),
    [anon_sym_RBRACE] = ACTIONS(1334),
    [aux_sym_trait_declaration_token1] = ACTIONS(1336),
    [aux_sym_interface_declaration_token1] = ACTIONS(1336),
    [aux_sym_enum_declaration_token1] = ACTIONS(1336),
    [aux_sym_enum_case_token1] = ACTIONS(1336),
    [aux_sym_class_declaration_token1] = ACTIONS(1336),
    [aux_sym_final_modifier_token1] = ACTIONS(1336),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1336),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1336),
    [sym_var_modifier] = ACTIONS(1336),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1336),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1336),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1336),
    [anon_sym_LPAREN2] = ACTIONS(1334),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1336),
    [aux_sym_cast_type_token1] = ACTIONS(1336),
    [aux_sym_echo_statement_token1] = ACTIONS(1336),
    [aux_sym_exit_statement_token1] = ACTIONS(1336),
    [anon_sym_unset] = ACTIONS(1336),
    [aux_sym_declare_statement_token1] = ACTIONS(1336),
    [aux_sym_declare_statement_token2] = ACTIONS(1336),
    [sym_float] = ACTIONS(1336),
    [aux_sym_try_statement_token1] = ACTIONS(1336),
    [aux_sym_goto_statement_token1] = ACTIONS(1336),
    [aux_sym_continue_statement_token1] = ACTIONS(1336),
    [aux_sym_break_statement_token1] = ACTIONS(1336),
    [sym_integer] = ACTIONS(1336),
    [aux_sym_return_statement_token1] = ACTIONS(1336),
    [aux_sym_throw_expression_token1] = ACTIONS(1336),
    [aux_sym_while_statement_token1] = ACTIONS(1336),
    [aux_sym_while_statement_token2] = ACTIONS(1336),
    [aux_sym_do_statement_token1] = ACTIONS(1336),
    [aux_sym_for_statement_token1] = ACTIONS(1336),
    [aux_sym_for_statement_token2] = ACTIONS(1336),
    [aux_sym_foreach_statement_token1] = ACTIONS(1336),
    [aux_sym_foreach_statement_token2] = ACTIONS(1336),
    [aux_sym_if_statement_token1] = ACTIONS(1336),
    [aux_sym_if_statement_token2] = ACTIONS(1336),
    [aux_sym_else_if_clause_token1] = ACTIONS(1336),
    [aux_sym_else_clause_token1] = ACTIONS(1336),
    [aux_sym_match_expression_token1] = ACTIONS(1336),
    [aux_sym_match_default_expression_token1] = ACTIONS(1336),
    [aux_sym_switch_statement_token1] = ACTIONS(1336),
    [aux_sym_switch_block_token1] = ACTIONS(1336),
    [anon_sym_PLUS] = ACTIONS(1336),
    [anon_sym_DASH] = ACTIONS(1336),
    [anon_sym_TILDE] = ACTIONS(1334),
    [anon_sym_BANG] = ACTIONS(1334),
    [anon_sym_AT] = ACTIONS(1334),
    [aux_sym_clone_expression_token1] = ACTIONS(1336),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1336),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1336),
    [anon_sym_DASH_DASH] = ACTIONS(1334),
    [anon_sym_PLUS_PLUS] = ACTIONS(1334),
    [aux_sym__list_destructing_token1] = ACTIONS(1336),
    [anon_sym_LBRACK] = ACTIONS(1334),
    [anon_sym_self] = ACTIONS(1336),
    [anon_sym_parent] = ACTIONS(1336),
    [aux_sym__argument_name_token1] = ACTIONS(1336),
    [aux_sym__argument_name_token2] = ACTIONS(1336),
    [anon_sym_POUND_LBRACK] = ACTIONS(1334),
    [aux_sym_encapsed_string_token1] = ACTIONS(1334),
    [anon_sym_DQUOTE] = ACTIONS(1334),
    [aux_sym_string_token1] = ACTIONS(1334),
    [anon_sym_SQUOTE] = ACTIONS(1334),
    [anon_sym_LT_LT_LT] = ACTIONS(1334),
    [anon_sym_BQUOTE] = ACTIONS(1334),
    [anon_sym_DOLLAR] = ACTIONS(1334),
    [aux_sym_yield_expression_token1] = ACTIONS(1336),
    [aux_sym_include_expression_token1] = ACTIONS(1336),
    [aux_sym_include_once_expression_token1] = ACTIONS(1336),
    [aux_sym_require_expression_token1] = ACTIONS(1336),
    [aux_sym_require_once_expression_token1] = ACTIONS(1336),
    [sym_comment] = ACTIONS(3),
  },
  [516] = {
    [ts_builtin_sym_end] = ACTIONS(1338),
    [sym_name] = ACTIONS(1340),
    [anon_sym_QMARK_GT] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1338),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1340),
    [aux_sym_global_declaration_token1] = ACTIONS(1340),
    [aux_sym_namespace_definition_token1] = ACTIONS(1340),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1340),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1340),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1340),
    [anon_sym_BSLASH] = ACTIONS(1338),
    [anon_sym_LBRACE] = ACTIONS(1338),
    [anon_sym_RBRACE] = ACTIONS(1338),
    [aux_sym_trait_declaration_token1] = ACTIONS(1340),
    [aux_sym_interface_declaration_token1] = ACTIONS(1340),
    [aux_sym_enum_declaration_token1] = ACTIONS(1340),
    [aux_sym_enum_case_token1] = ACTIONS(1340),
    [aux_sym_class_declaration_token1] = ACTIONS(1340),
    [aux_sym_final_modifier_token1] = ACTIONS(1340),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1340),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1340),
    [sym_var_modifier] = ACTIONS(1340),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1340),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1340),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1340),
    [anon_sym_LPAREN2] = ACTIONS(1338),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1340),
    [aux_sym_cast_type_token1] = ACTIONS(1340),
    [aux_sym_echo_statement_token1] = ACTIONS(1340),
    [aux_sym_exit_statement_token1] = ACTIONS(1340),
    [anon_sym_unset] = ACTIONS(1340),
    [aux_sym_declare_statement_token1] = ACTIONS(1340),
    [aux_sym_declare_statement_token2] = ACTIONS(1340),
    [sym_float] = ACTIONS(1340),
    [aux_sym_try_statement_token1] = ACTIONS(1340),
    [aux_sym_goto_statement_token1] = ACTIONS(1340),
    [aux_sym_continue_statement_token1] = ACTIONS(1340),
    [aux_sym_break_statement_token1] = ACTIONS(1340),
    [sym_integer] = ACTIONS(1340),
    [aux_sym_return_statement_token1] = ACTIONS(1340),
    [aux_sym_throw_expression_token1] = ACTIONS(1340),
    [aux_sym_while_statement_token1] = ACTIONS(1340),
    [aux_sym_while_statement_token2] = ACTIONS(1340),
    [aux_sym_do_statement_token1] = ACTIONS(1340),
    [aux_sym_for_statement_token1] = ACTIONS(1340),
    [aux_sym_for_statement_token2] = ACTIONS(1340),
    [aux_sym_foreach_statement_token1] = ACTIONS(1340),
    [aux_sym_foreach_statement_token2] = ACTIONS(1340),
    [aux_sym_if_statement_token1] = ACTIONS(1340),
    [aux_sym_if_statement_token2] = ACTIONS(1340),
    [aux_sym_else_if_clause_token1] = ACTIONS(1340),
    [aux_sym_else_clause_token1] = ACTIONS(1340),
    [aux_sym_match_expression_token1] = ACTIONS(1340),
    [aux_sym_match_default_expression_token1] = ACTIONS(1340),
    [aux_sym_switch_statement_token1] = ACTIONS(1340),
    [aux_sym_switch_block_token1] = ACTIONS(1340),
    [anon_sym_PLUS] = ACTIONS(1340),
    [anon_sym_DASH] = ACTIONS(1340),
    [anon_sym_TILDE] = ACTIONS(1338),
    [anon_sym_BANG] = ACTIONS(1338),
    [anon_sym_AT] = ACTIONS(1338),
    [aux_sym_clone_expression_token1] = ACTIONS(1340),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1340),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1340),
    [anon_sym_DASH_DASH] = ACTIONS(1338),
    [anon_sym_PLUS_PLUS] = ACTIONS(1338),
    [aux_sym__list_destructing_token1] = ACTIONS(1340),
    [anon_sym_LBRACK] = ACTIONS(1338),
    [anon_sym_self] = ACTIONS(1340),
    [anon_sym_parent] = ACTIONS(1340),
    [aux_sym__argument_name_token1] = ACTIONS(1340),
    [aux_sym__argument_name_token2] = ACTIONS(1340),
    [anon_sym_POUND_LBRACK] = ACTIONS(1338),
    [aux_sym_encapsed_string_token1] = ACTIONS(1338),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [aux_sym_string_token1] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_LT_LT_LT] = ACTIONS(1338),
    [anon_sym_BQUOTE] = ACTIONS(1338),
    [anon_sym_DOLLAR] = ACTIONS(1338),
    [aux_sym_yield_expression_token1] = ACTIONS(1340),
    [aux_sym_include_expression_token1] = ACTIONS(1340),
    [aux_sym_include_once_expression_token1] = ACTIONS(1340),
    [aux_sym_require_expression_token1] = ACTIONS(1340),
    [aux_sym_require_once_expression_token1] = ACTIONS(1340),
    [sym_comment] = ACTIONS(3),
  },
  [517] = {
    [ts_builtin_sym_end] = ACTIONS(1342),
    [sym_name] = ACTIONS(1344),
    [anon_sym_QMARK_GT] = ACTIONS(1342),
    [anon_sym_SEMI] = ACTIONS(1342),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1344),
    [aux_sym_global_declaration_token1] = ACTIONS(1344),
    [aux_sym_namespace_definition_token1] = ACTIONS(1344),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1344),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1344),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1344),
    [anon_sym_BSLASH] = ACTIONS(1342),
    [anon_sym_LBRACE] = ACTIONS(1342),
    [anon_sym_RBRACE] = ACTIONS(1342),
    [aux_sym_trait_declaration_token1] = ACTIONS(1344),
    [aux_sym_interface_declaration_token1] = ACTIONS(1344),
    [aux_sym_enum_declaration_token1] = ACTIONS(1344),
    [aux_sym_enum_case_token1] = ACTIONS(1344),
    [aux_sym_class_declaration_token1] = ACTIONS(1344),
    [aux_sym_final_modifier_token1] = ACTIONS(1344),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1344),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1344),
    [sym_var_modifier] = ACTIONS(1344),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1344),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1344),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1344),
    [anon_sym_LPAREN2] = ACTIONS(1342),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1344),
    [aux_sym_cast_type_token1] = ACTIONS(1344),
    [aux_sym_echo_statement_token1] = ACTIONS(1344),
    [aux_sym_exit_statement_token1] = ACTIONS(1344),
    [anon_sym_unset] = ACTIONS(1344),
    [aux_sym_declare_statement_token1] = ACTIONS(1344),
    [aux_sym_declare_statement_token2] = ACTIONS(1344),
    [sym_float] = ACTIONS(1344),
    [aux_sym_try_statement_token1] = ACTIONS(1344),
    [aux_sym_goto_statement_token1] = ACTIONS(1344),
    [aux_sym_continue_statement_token1] = ACTIONS(1344),
    [aux_sym_break_statement_token1] = ACTIONS(1344),
    [sym_integer] = ACTIONS(1344),
    [aux_sym_return_statement_token1] = ACTIONS(1344),
    [aux_sym_throw_expression_token1] = ACTIONS(1344),
    [aux_sym_while_statement_token1] = ACTIONS(1344),
    [aux_sym_while_statement_token2] = ACTIONS(1344),
    [aux_sym_do_statement_token1] = ACTIONS(1344),
    [aux_sym_for_statement_token1] = ACTIONS(1344),
    [aux_sym_for_statement_token2] = ACTIONS(1344),
    [aux_sym_foreach_statement_token1] = ACTIONS(1344),
    [aux_sym_foreach_statement_token2] = ACTIONS(1344),
    [aux_sym_if_statement_token1] = ACTIONS(1344),
    [aux_sym_if_statement_token2] = ACTIONS(1344),
    [aux_sym_else_if_clause_token1] = ACTIONS(1344),
    [aux_sym_else_clause_token1] = ACTIONS(1344),
    [aux_sym_match_expression_token1] = ACTIONS(1344),
    [aux_sym_match_default_expression_token1] = ACTIONS(1344),
    [aux_sym_switch_statement_token1] = ACTIONS(1344),
    [aux_sym_switch_block_token1] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_TILDE] = ACTIONS(1342),
    [anon_sym_BANG] = ACTIONS(1342),
    [anon_sym_AT] = ACTIONS(1342),
    [aux_sym_clone_expression_token1] = ACTIONS(1344),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1344),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1344),
    [anon_sym_DASH_DASH] = ACTIONS(1342),
    [anon_sym_PLUS_PLUS] = ACTIONS(1342),
    [aux_sym__list_destructing_token1] = ACTIONS(1344),
    [anon_sym_LBRACK] = ACTIONS(1342),
    [anon_sym_self] = ACTIONS(1344),
    [anon_sym_parent] = ACTIONS(1344),
    [aux_sym__argument_name_token1] = ACTIONS(1344),
    [aux_sym__argument_name_token2] = ACTIONS(1344),
    [anon_sym_POUND_LBRACK] = ACTIONS(1342),
    [aux_sym_encapsed_string_token1] = ACTIONS(1342),
    [anon_sym_DQUOTE] = ACTIONS(1342),
    [aux_sym_string_token1] = ACTIONS(1342),
    [anon_sym_SQUOTE] = ACTIONS(1342),
    [anon_sym_LT_LT_LT] = ACTIONS(1342),
    [anon_sym_BQUOTE] = ACTIONS(1342),
    [anon_sym_DOLLAR] = ACTIONS(1342),
    [aux_sym_yield_expression_token1] = ACTIONS(1344),
    [aux_sym_include_expression_token1] = ACTIONS(1344),
    [aux_sym_include_once_expression_token1] = ACTIONS(1344),
    [aux_sym_require_expression_token1] = ACTIONS(1344),
    [aux_sym_require_once_expression_token1] = ACTIONS(1344),
    [sym_comment] = ACTIONS(3),
  },
  [518] = {
    [ts_builtin_sym_end] = ACTIONS(1346),
    [sym_name] = ACTIONS(1348),
    [anon_sym_QMARK_GT] = ACTIONS(1346),
    [anon_sym_SEMI] = ACTIONS(1346),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1348),
    [aux_sym_global_declaration_token1] = ACTIONS(1348),
    [aux_sym_namespace_definition_token1] = ACTIONS(1348),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1348),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1348),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1348),
    [anon_sym_BSLASH] = ACTIONS(1346),
    [anon_sym_LBRACE] = ACTIONS(1346),
    [anon_sym_RBRACE] = ACTIONS(1346),
    [aux_sym_trait_declaration_token1] = ACTIONS(1348),
    [aux_sym_interface_declaration_token1] = ACTIONS(1348),
    [aux_sym_enum_declaration_token1] = ACTIONS(1348),
    [aux_sym_enum_case_token1] = ACTIONS(1348),
    [aux_sym_class_declaration_token1] = ACTIONS(1348),
    [aux_sym_final_modifier_token1] = ACTIONS(1348),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1348),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1348),
    [sym_var_modifier] = ACTIONS(1348),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1348),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1348),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1348),
    [anon_sym_LPAREN2] = ACTIONS(1346),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1348),
    [aux_sym_cast_type_token1] = ACTIONS(1348),
    [aux_sym_echo_statement_token1] = ACTIONS(1348),
    [aux_sym_exit_statement_token1] = ACTIONS(1348),
    [anon_sym_unset] = ACTIONS(1348),
    [aux_sym_declare_statement_token1] = ACTIONS(1348),
    [aux_sym_declare_statement_token2] = ACTIONS(1348),
    [sym_float] = ACTIONS(1348),
    [aux_sym_try_statement_token1] = ACTIONS(1348),
    [aux_sym_goto_statement_token1] = ACTIONS(1348),
    [aux_sym_continue_statement_token1] = ACTIONS(1348),
    [aux_sym_break_statement_token1] = ACTIONS(1348),
    [sym_integer] = ACTIONS(1348),
    [aux_sym_return_statement_token1] = ACTIONS(1348),
    [aux_sym_throw_expression_token1] = ACTIONS(1348),
    [aux_sym_while_statement_token1] = ACTIONS(1348),
    [aux_sym_while_statement_token2] = ACTIONS(1348),
    [aux_sym_do_statement_token1] = ACTIONS(1348),
    [aux_sym_for_statement_token1] = ACTIONS(1348),
    [aux_sym_for_statement_token2] = ACTIONS(1348),
    [aux_sym_foreach_statement_token1] = ACTIONS(1348),
    [aux_sym_foreach_statement_token2] = ACTIONS(1348),
    [aux_sym_if_statement_token1] = ACTIONS(1348),
    [aux_sym_if_statement_token2] = ACTIONS(1348),
    [aux_sym_else_if_clause_token1] = ACTIONS(1348),
    [aux_sym_else_clause_token1] = ACTIONS(1348),
    [aux_sym_match_expression_token1] = ACTIONS(1348),
    [aux_sym_match_default_expression_token1] = ACTIONS(1348),
    [aux_sym_switch_statement_token1] = ACTIONS(1348),
    [aux_sym_switch_block_token1] = ACTIONS(1348),
    [anon_sym_PLUS] = ACTIONS(1348),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_TILDE] = ACTIONS(1346),
    [anon_sym_BANG] = ACTIONS(1346),
    [anon_sym_AT] = ACTIONS(1346),
    [aux_sym_clone_expression_token1] = ACTIONS(1348),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1348),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1348),
    [anon_sym_DASH_DASH] = ACTIONS(1346),
    [anon_sym_PLUS_PLUS] = ACTIONS(1346),
    [aux_sym__list_destructing_token1] = ACTIONS(1348),
    [anon_sym_LBRACK] = ACTIONS(1346),
    [anon_sym_self] = ACTIONS(1348),
    [anon_sym_parent] = ACTIONS(1348),
    [aux_sym__argument_name_token1] = ACTIONS(1348),
    [aux_sym__argument_name_token2] = ACTIONS(1348),
    [anon_sym_POUND_LBRACK] = ACTIONS(1346),
    [aux_sym_encapsed_string_token1] = ACTIONS(1346),
    [anon_sym_DQUOTE] = ACTIONS(1346),
    [aux_sym_string_token1] = ACTIONS(1346),
    [anon_sym_SQUOTE] = ACTIONS(1346),
    [anon_sym_LT_LT_LT] = ACTIONS(1346),
    [anon_sym_BQUOTE] = ACTIONS(1346),
    [anon_sym_DOLLAR] = ACTIONS(1346),
    [aux_sym_yield_expression_token1] = ACTIONS(1348),
    [aux_sym_include_expression_token1] = ACTIONS(1348),
    [aux_sym_include_once_expression_token1] = ACTIONS(1348),
    [aux_sym_require_expression_token1] = ACTIONS(1348),
    [aux_sym_require_once_expression_token1] = ACTIONS(1348),
    [sym_comment] = ACTIONS(3),
  },
  [519] = {
    [ts_builtin_sym_end] = ACTIONS(1350),
    [sym_name] = ACTIONS(1352),
    [anon_sym_QMARK_GT] = ACTIONS(1350),
    [anon_sym_SEMI] = ACTIONS(1350),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1352),
    [aux_sym_global_declaration_token1] = ACTIONS(1352),
    [aux_sym_namespace_definition_token1] = ACTIONS(1352),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1352),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1352),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1352),
    [anon_sym_BSLASH] = ACTIONS(1350),
    [anon_sym_LBRACE] = ACTIONS(1350),
    [anon_sym_RBRACE] = ACTIONS(1350),
    [aux_sym_trait_declaration_token1] = ACTIONS(1352),
    [aux_sym_interface_declaration_token1] = ACTIONS(1352),
    [aux_sym_enum_declaration_token1] = ACTIONS(1352),
    [aux_sym_enum_case_token1] = ACTIONS(1352),
    [aux_sym_class_declaration_token1] = ACTIONS(1352),
    [aux_sym_final_modifier_token1] = ACTIONS(1352),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1352),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1352),
    [sym_var_modifier] = ACTIONS(1352),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1352),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1352),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1352),
    [anon_sym_LPAREN2] = ACTIONS(1350),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1352),
    [aux_sym_cast_type_token1] = ACTIONS(1352),
    [aux_sym_echo_statement_token1] = ACTIONS(1352),
    [aux_sym_exit_statement_token1] = ACTIONS(1352),
    [anon_sym_unset] = ACTIONS(1352),
    [aux_sym_declare_statement_token1] = ACTIONS(1352),
    [aux_sym_declare_statement_token2] = ACTIONS(1352),
    [sym_float] = ACTIONS(1352),
    [aux_sym_try_statement_token1] = ACTIONS(1352),
    [aux_sym_goto_statement_token1] = ACTIONS(1352),
    [aux_sym_continue_statement_token1] = ACTIONS(1352),
    [aux_sym_break_statement_token1] = ACTIONS(1352),
    [sym_integer] = ACTIONS(1352),
    [aux_sym_return_statement_token1] = ACTIONS(1352),
    [aux_sym_throw_expression_token1] = ACTIONS(1352),
    [aux_sym_while_statement_token1] = ACTIONS(1352),
    [aux_sym_while_statement_token2] = ACTIONS(1352),
    [aux_sym_do_statement_token1] = ACTIONS(1352),
    [aux_sym_for_statement_token1] = ACTIONS(1352),
    [aux_sym_for_statement_token2] = ACTIONS(1352),
    [aux_sym_foreach_statement_token1] = ACTIONS(1352),
    [aux_sym_foreach_statement_token2] = ACTIONS(1352),
    [aux_sym_if_statement_token1] = ACTIONS(1352),
    [aux_sym_if_statement_token2] = ACTIONS(1352),
    [aux_sym_else_if_clause_token1] = ACTIONS(1352),
    [aux_sym_else_clause_token1] = ACTIONS(1352),
    [aux_sym_match_expression_token1] = ACTIONS(1352),
    [aux_sym_match_default_expression_token1] = ACTIONS(1352),
    [aux_sym_switch_statement_token1] = ACTIONS(1352),
    [aux_sym_switch_block_token1] = ACTIONS(1352),
    [anon_sym_PLUS] = ACTIONS(1352),
    [anon_sym_DASH] = ACTIONS(1352),
    [anon_sym_TILDE] = ACTIONS(1350),
    [anon_sym_BANG] = ACTIONS(1350),
    [anon_sym_AT] = ACTIONS(1350),
    [aux_sym_clone_expression_token1] = ACTIONS(1352),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1352),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1352),
    [anon_sym_DASH_DASH] = ACTIONS(1350),
    [anon_sym_PLUS_PLUS] = ACTIONS(1350),
    [aux_sym__list_destructing_token1] = ACTIONS(1352),
    [anon_sym_LBRACK] = ACTIONS(1350),
    [anon_sym_self] = ACTIONS(1352),
    [anon_sym_parent] = ACTIONS(1352),
    [aux_sym__argument_name_token1] = ACTIONS(1352),
    [aux_sym__argument_name_token2] = ACTIONS(1352),
    [anon_sym_POUND_LBRACK] = ACTIONS(1350),
    [aux_sym_encapsed_string_token1] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1350),
    [aux_sym_string_token1] = ACTIONS(1350),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_LT_LT_LT] = ACTIONS(1350),
    [anon_sym_BQUOTE] = ACTIONS(1350),
    [anon_sym_DOLLAR] = ACTIONS(1350),
    [aux_sym_yield_expression_token1] = ACTIONS(1352),
    [aux_sym_include_expression_token1] = ACTIONS(1352),
    [aux_sym_include_once_expression_token1] = ACTIONS(1352),
    [aux_sym_require_expression_token1] = ACTIONS(1352),
    [aux_sym_require_once_expression_token1] = ACTIONS(1352),
    [sym_comment] = ACTIONS(3),
  },
  [520] = {
    [ts_builtin_sym_end] = ACTIONS(1354),
    [sym_name] = ACTIONS(1356),
    [anon_sym_QMARK_GT] = ACTIONS(1354),
    [anon_sym_SEMI] = ACTIONS(1354),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1356),
    [aux_sym_global_declaration_token1] = ACTIONS(1356),
    [aux_sym_namespace_definition_token1] = ACTIONS(1356),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1356),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1356),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1356),
    [anon_sym_BSLASH] = ACTIONS(1354),
    [anon_sym_LBRACE] = ACTIONS(1354),
    [anon_sym_RBRACE] = ACTIONS(1354),
    [aux_sym_trait_declaration_token1] = ACTIONS(1356),
    [aux_sym_interface_declaration_token1] = ACTIONS(1356),
    [aux_sym_enum_declaration_token1] = ACTIONS(1356),
    [aux_sym_enum_case_token1] = ACTIONS(1356),
    [aux_sym_class_declaration_token1] = ACTIONS(1356),
    [aux_sym_final_modifier_token1] = ACTIONS(1356),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1356),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1356),
    [sym_var_modifier] = ACTIONS(1356),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1356),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1356),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1356),
    [anon_sym_LPAREN2] = ACTIONS(1354),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1356),
    [aux_sym_cast_type_token1] = ACTIONS(1356),
    [aux_sym_echo_statement_token1] = ACTIONS(1356),
    [aux_sym_exit_statement_token1] = ACTIONS(1356),
    [anon_sym_unset] = ACTIONS(1356),
    [aux_sym_declare_statement_token1] = ACTIONS(1356),
    [aux_sym_declare_statement_token2] = ACTIONS(1356),
    [sym_float] = ACTIONS(1356),
    [aux_sym_try_statement_token1] = ACTIONS(1356),
    [aux_sym_goto_statement_token1] = ACTIONS(1356),
    [aux_sym_continue_statement_token1] = ACTIONS(1356),
    [aux_sym_break_statement_token1] = ACTIONS(1356),
    [sym_integer] = ACTIONS(1356),
    [aux_sym_return_statement_token1] = ACTIONS(1356),
    [aux_sym_throw_expression_token1] = ACTIONS(1356),
    [aux_sym_while_statement_token1] = ACTIONS(1356),
    [aux_sym_while_statement_token2] = ACTIONS(1356),
    [aux_sym_do_statement_token1] = ACTIONS(1356),
    [aux_sym_for_statement_token1] = ACTIONS(1356),
    [aux_sym_for_statement_token2] = ACTIONS(1356),
    [aux_sym_foreach_statement_token1] = ACTIONS(1356),
    [aux_sym_foreach_statement_token2] = ACTIONS(1356),
    [aux_sym_if_statement_token1] = ACTIONS(1356),
    [aux_sym_if_statement_token2] = ACTIONS(1356),
    [aux_sym_else_if_clause_token1] = ACTIONS(1356),
    [aux_sym_else_clause_token1] = ACTIONS(1356),
    [aux_sym_match_expression_token1] = ACTIONS(1356),
    [aux_sym_match_default_expression_token1] = ACTIONS(1356),
    [aux_sym_switch_statement_token1] = ACTIONS(1356),
    [aux_sym_switch_block_token1] = ACTIONS(1356),
    [anon_sym_PLUS] = ACTIONS(1356),
    [anon_sym_DASH] = ACTIONS(1356),
    [anon_sym_TILDE] = ACTIONS(1354),
    [anon_sym_BANG] = ACTIONS(1354),
    [anon_sym_AT] = ACTIONS(1354),
    [aux_sym_clone_expression_token1] = ACTIONS(1356),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1356),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1356),
    [anon_sym_DASH_DASH] = ACTIONS(1354),
    [anon_sym_PLUS_PLUS] = ACTIONS(1354),
    [aux_sym__list_destructing_token1] = ACTIONS(1356),
    [anon_sym_LBRACK] = ACTIONS(1354),
    [anon_sym_self] = ACTIONS(1356),
    [anon_sym_parent] = ACTIONS(1356),
    [aux_sym__argument_name_token1] = ACTIONS(1356),
    [aux_sym__argument_name_token2] = ACTIONS(1356),
    [anon_sym_POUND_LBRACK] = ACTIONS(1354),
    [aux_sym_encapsed_string_token1] = ACTIONS(1354),
    [anon_sym_DQUOTE] = ACTIONS(1354),
    [aux_sym_string_token1] = ACTIONS(1354),
    [anon_sym_SQUOTE] = ACTIONS(1354),
    [anon_sym_LT_LT_LT] = ACTIONS(1354),
    [anon_sym_BQUOTE] = ACTIONS(1354),
    [anon_sym_DOLLAR] = ACTIONS(1354),
    [aux_sym_yield_expression_token1] = ACTIONS(1356),
    [aux_sym_include_expression_token1] = ACTIONS(1356),
    [aux_sym_include_once_expression_token1] = ACTIONS(1356),
    [aux_sym_require_expression_token1] = ACTIONS(1356),
    [aux_sym_require_once_expression_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(3),
  },
  [521] = {
    [ts_builtin_sym_end] = ACTIONS(1358),
    [sym_name] = ACTIONS(1360),
    [anon_sym_QMARK_GT] = ACTIONS(1358),
    [anon_sym_SEMI] = ACTIONS(1358),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1360),
    [aux_sym_global_declaration_token1] = ACTIONS(1360),
    [aux_sym_namespace_definition_token1] = ACTIONS(1360),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1360),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1360),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1360),
    [anon_sym_BSLASH] = ACTIONS(1358),
    [anon_sym_LBRACE] = ACTIONS(1358),
    [anon_sym_RBRACE] = ACTIONS(1358),
    [aux_sym_trait_declaration_token1] = ACTIONS(1360),
    [aux_sym_interface_declaration_token1] = ACTIONS(1360),
    [aux_sym_enum_declaration_token1] = ACTIONS(1360),
    [aux_sym_enum_case_token1] = ACTIONS(1360),
    [aux_sym_class_declaration_token1] = ACTIONS(1360),
    [aux_sym_final_modifier_token1] = ACTIONS(1360),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1360),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1360),
    [sym_var_modifier] = ACTIONS(1360),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1360),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1360),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1360),
    [anon_sym_LPAREN2] = ACTIONS(1358),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1360),
    [aux_sym_cast_type_token1] = ACTIONS(1360),
    [aux_sym_echo_statement_token1] = ACTIONS(1360),
    [aux_sym_exit_statement_token1] = ACTIONS(1360),
    [anon_sym_unset] = ACTIONS(1360),
    [aux_sym_declare_statement_token1] = ACTIONS(1360),
    [aux_sym_declare_statement_token2] = ACTIONS(1360),
    [sym_float] = ACTIONS(1360),
    [aux_sym_try_statement_token1] = ACTIONS(1360),
    [aux_sym_goto_statement_token1] = ACTIONS(1360),
    [aux_sym_continue_statement_token1] = ACTIONS(1360),
    [aux_sym_break_statement_token1] = ACTIONS(1360),
    [sym_integer] = ACTIONS(1360),
    [aux_sym_return_statement_token1] = ACTIONS(1360),
    [aux_sym_throw_expression_token1] = ACTIONS(1360),
    [aux_sym_while_statement_token1] = ACTIONS(1360),
    [aux_sym_while_statement_token2] = ACTIONS(1360),
    [aux_sym_do_statement_token1] = ACTIONS(1360),
    [aux_sym_for_statement_token1] = ACTIONS(1360),
    [aux_sym_for_statement_token2] = ACTIONS(1360),
    [aux_sym_foreach_statement_token1] = ACTIONS(1360),
    [aux_sym_foreach_statement_token2] = ACTIONS(1360),
    [aux_sym_if_statement_token1] = ACTIONS(1360),
    [aux_sym_if_statement_token2] = ACTIONS(1360),
    [aux_sym_else_if_clause_token1] = ACTIONS(1360),
    [aux_sym_else_clause_token1] = ACTIONS(1360),
    [aux_sym_match_expression_token1] = ACTIONS(1360),
    [aux_sym_match_default_expression_token1] = ACTIONS(1360),
    [aux_sym_switch_statement_token1] = ACTIONS(1360),
    [aux_sym_switch_block_token1] = ACTIONS(1360),
    [anon_sym_PLUS] = ACTIONS(1360),
    [anon_sym_DASH] = ACTIONS(1360),
    [anon_sym_TILDE] = ACTIONS(1358),
    [anon_sym_BANG] = ACTIONS(1358),
    [anon_sym_AT] = ACTIONS(1358),
    [aux_sym_clone_expression_token1] = ACTIONS(1360),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1360),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1360),
    [anon_sym_DASH_DASH] = ACTIONS(1358),
    [anon_sym_PLUS_PLUS] = ACTIONS(1358),
    [aux_sym__list_destructing_token1] = ACTIONS(1360),
    [anon_sym_LBRACK] = ACTIONS(1358),
    [anon_sym_self] = ACTIONS(1360),
    [anon_sym_parent] = ACTIONS(1360),
    [aux_sym__argument_name_token1] = ACTIONS(1360),
    [aux_sym__argument_name_token2] = ACTIONS(1360),
    [anon_sym_POUND_LBRACK] = ACTIONS(1358),
    [aux_sym_encapsed_string_token1] = ACTIONS(1358),
    [anon_sym_DQUOTE] = ACTIONS(1358),
    [aux_sym_string_token1] = ACTIONS(1358),
    [anon_sym_SQUOTE] = ACTIONS(1358),
    [anon_sym_LT_LT_LT] = ACTIONS(1358),
    [anon_sym_BQUOTE] = ACTIONS(1358),
    [anon_sym_DOLLAR] = ACTIONS(1358),
    [aux_sym_yield_expression_token1] = ACTIONS(1360),
    [aux_sym_include_expression_token1] = ACTIONS(1360),
    [aux_sym_include_once_expression_token1] = ACTIONS(1360),
    [aux_sym_require_expression_token1] = ACTIONS(1360),
    [aux_sym_require_once_expression_token1] = ACTIONS(1360),
    [sym_comment] = ACTIONS(3),
  },
  [522] = {
    [ts_builtin_sym_end] = ACTIONS(1362),
    [sym_name] = ACTIONS(1364),
    [anon_sym_QMARK_GT] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1362),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1364),
    [aux_sym_global_declaration_token1] = ACTIONS(1364),
    [aux_sym_namespace_definition_token1] = ACTIONS(1364),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1364),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1364),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1364),
    [anon_sym_BSLASH] = ACTIONS(1362),
    [anon_sym_LBRACE] = ACTIONS(1362),
    [anon_sym_RBRACE] = ACTIONS(1362),
    [aux_sym_trait_declaration_token1] = ACTIONS(1364),
    [aux_sym_interface_declaration_token1] = ACTIONS(1364),
    [aux_sym_enum_declaration_token1] = ACTIONS(1364),
    [aux_sym_enum_case_token1] = ACTIONS(1364),
    [aux_sym_class_declaration_token1] = ACTIONS(1364),
    [aux_sym_final_modifier_token1] = ACTIONS(1364),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1364),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1364),
    [sym_var_modifier] = ACTIONS(1364),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1364),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1364),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1364),
    [anon_sym_LPAREN2] = ACTIONS(1362),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1364),
    [aux_sym_cast_type_token1] = ACTIONS(1364),
    [aux_sym_echo_statement_token1] = ACTIONS(1364),
    [aux_sym_exit_statement_token1] = ACTIONS(1364),
    [anon_sym_unset] = ACTIONS(1364),
    [aux_sym_declare_statement_token1] = ACTIONS(1364),
    [aux_sym_declare_statement_token2] = ACTIONS(1364),
    [sym_float] = ACTIONS(1364),
    [aux_sym_try_statement_token1] = ACTIONS(1364),
    [aux_sym_goto_statement_token1] = ACTIONS(1364),
    [aux_sym_continue_statement_token1] = ACTIONS(1364),
    [aux_sym_break_statement_token1] = ACTIONS(1364),
    [sym_integer] = ACTIONS(1364),
    [aux_sym_return_statement_token1] = ACTIONS(1364),
    [aux_sym_throw_expression_token1] = ACTIONS(1364),
    [aux_sym_while_statement_token1] = ACTIONS(1364),
    [aux_sym_while_statement_token2] = ACTIONS(1364),
    [aux_sym_do_statement_token1] = ACTIONS(1364),
    [aux_sym_for_statement_token1] = ACTIONS(1364),
    [aux_sym_for_statement_token2] = ACTIONS(1364),
    [aux_sym_foreach_statement_token1] = ACTIONS(1364),
    [aux_sym_foreach_statement_token2] = ACTIONS(1364),
    [aux_sym_if_statement_token1] = ACTIONS(1364),
    [aux_sym_if_statement_token2] = ACTIONS(1364),
    [aux_sym_else_if_clause_token1] = ACTIONS(1364),
    [aux_sym_else_clause_token1] = ACTIONS(1364),
    [aux_sym_match_expression_token1] = ACTIONS(1364),
    [aux_sym_match_default_expression_token1] = ACTIONS(1364),
    [aux_sym_switch_statement_token1] = ACTIONS(1364),
    [aux_sym_switch_block_token1] = ACTIONS(1364),
    [anon_sym_PLUS] = ACTIONS(1364),
    [anon_sym_DASH] = ACTIONS(1364),
    [anon_sym_TILDE] = ACTIONS(1362),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_AT] = ACTIONS(1362),
    [aux_sym_clone_expression_token1] = ACTIONS(1364),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1364),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1364),
    [anon_sym_DASH_DASH] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1362),
    [aux_sym__list_destructing_token1] = ACTIONS(1364),
    [anon_sym_LBRACK] = ACTIONS(1362),
    [anon_sym_self] = ACTIONS(1364),
    [anon_sym_parent] = ACTIONS(1364),
    [aux_sym__argument_name_token1] = ACTIONS(1364),
    [aux_sym__argument_name_token2] = ACTIONS(1364),
    [anon_sym_POUND_LBRACK] = ACTIONS(1362),
    [aux_sym_encapsed_string_token1] = ACTIONS(1362),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [aux_sym_string_token1] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_LT_LT_LT] = ACTIONS(1362),
    [anon_sym_BQUOTE] = ACTIONS(1362),
    [anon_sym_DOLLAR] = ACTIONS(1362),
    [aux_sym_yield_expression_token1] = ACTIONS(1364),
    [aux_sym_include_expression_token1] = ACTIONS(1364),
    [aux_sym_include_once_expression_token1] = ACTIONS(1364),
    [aux_sym_require_expression_token1] = ACTIONS(1364),
    [aux_sym_require_once_expression_token1] = ACTIONS(1364),
    [sym_comment] = ACTIONS(3),
  },
  [523] = {
    [ts_builtin_sym_end] = ACTIONS(1366),
    [sym_name] = ACTIONS(1368),
    [anon_sym_QMARK_GT] = ACTIONS(1366),
    [anon_sym_SEMI] = ACTIONS(1366),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1368),
    [aux_sym_global_declaration_token1] = ACTIONS(1368),
    [aux_sym_namespace_definition_token1] = ACTIONS(1368),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1368),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1368),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1368),
    [anon_sym_BSLASH] = ACTIONS(1366),
    [anon_sym_LBRACE] = ACTIONS(1366),
    [anon_sym_RBRACE] = ACTIONS(1366),
    [aux_sym_trait_declaration_token1] = ACTIONS(1368),
    [aux_sym_interface_declaration_token1] = ACTIONS(1368),
    [aux_sym_enum_declaration_token1] = ACTIONS(1368),
    [aux_sym_enum_case_token1] = ACTIONS(1368),
    [aux_sym_class_declaration_token1] = ACTIONS(1368),
    [aux_sym_final_modifier_token1] = ACTIONS(1368),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1368),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1368),
    [sym_var_modifier] = ACTIONS(1368),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1368),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1368),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1368),
    [anon_sym_LPAREN2] = ACTIONS(1366),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1368),
    [aux_sym_cast_type_token1] = ACTIONS(1368),
    [aux_sym_echo_statement_token1] = ACTIONS(1368),
    [aux_sym_exit_statement_token1] = ACTIONS(1368),
    [anon_sym_unset] = ACTIONS(1368),
    [aux_sym_declare_statement_token1] = ACTIONS(1368),
    [aux_sym_declare_statement_token2] = ACTIONS(1368),
    [sym_float] = ACTIONS(1368),
    [aux_sym_try_statement_token1] = ACTIONS(1368),
    [aux_sym_goto_statement_token1] = ACTIONS(1368),
    [aux_sym_continue_statement_token1] = ACTIONS(1368),
    [aux_sym_break_statement_token1] = ACTIONS(1368),
    [sym_integer] = ACTIONS(1368),
    [aux_sym_return_statement_token1] = ACTIONS(1368),
    [aux_sym_throw_expression_token1] = ACTIONS(1368),
    [aux_sym_while_statement_token1] = ACTIONS(1368),
    [aux_sym_while_statement_token2] = ACTIONS(1368),
    [aux_sym_do_statement_token1] = ACTIONS(1368),
    [aux_sym_for_statement_token1] = ACTIONS(1368),
    [aux_sym_for_statement_token2] = ACTIONS(1368),
    [aux_sym_foreach_statement_token1] = ACTIONS(1368),
    [aux_sym_foreach_statement_token2] = ACTIONS(1368),
    [aux_sym_if_statement_token1] = ACTIONS(1368),
    [aux_sym_if_statement_token2] = ACTIONS(1368),
    [aux_sym_else_if_clause_token1] = ACTIONS(1368),
    [aux_sym_else_clause_token1] = ACTIONS(1368),
    [aux_sym_match_expression_token1] = ACTIONS(1368),
    [aux_sym_match_default_expression_token1] = ACTIONS(1368),
    [aux_sym_switch_statement_token1] = ACTIONS(1368),
    [aux_sym_switch_block_token1] = ACTIONS(1368),
    [anon_sym_PLUS] = ACTIONS(1368),
    [anon_sym_DASH] = ACTIONS(1368),
    [anon_sym_TILDE] = ACTIONS(1366),
    [anon_sym_BANG] = ACTIONS(1366),
    [anon_sym_AT] = ACTIONS(1366),
    [aux_sym_clone_expression_token1] = ACTIONS(1368),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1368),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1368),
    [anon_sym_DASH_DASH] = ACTIONS(1366),
    [anon_sym_PLUS_PLUS] = ACTIONS(1366),
    [aux_sym__list_destructing_token1] = ACTIONS(1368),
    [anon_sym_LBRACK] = ACTIONS(1366),
    [anon_sym_self] = ACTIONS(1368),
    [anon_sym_parent] = ACTIONS(1368),
    [aux_sym__argument_name_token1] = ACTIONS(1368),
    [aux_sym__argument_name_token2] = ACTIONS(1368),
    [anon_sym_POUND_LBRACK] = ACTIONS(1366),
    [aux_sym_encapsed_string_token1] = ACTIONS(1366),
    [anon_sym_DQUOTE] = ACTIONS(1366),
    [aux_sym_string_token1] = ACTIONS(1366),
    [anon_sym_SQUOTE] = ACTIONS(1366),
    [anon_sym_LT_LT_LT] = ACTIONS(1366),
    [anon_sym_BQUOTE] = ACTIONS(1366),
    [anon_sym_DOLLAR] = ACTIONS(1366),
    [aux_sym_yield_expression_token1] = ACTIONS(1368),
    [aux_sym_include_expression_token1] = ACTIONS(1368),
    [aux_sym_include_once_expression_token1] = ACTIONS(1368),
    [aux_sym_require_expression_token1] = ACTIONS(1368),
    [aux_sym_require_once_expression_token1] = ACTIONS(1368),
    [sym_comment] = ACTIONS(3),
  },
  [524] = {
    [ts_builtin_sym_end] = ACTIONS(1370),
    [sym_name] = ACTIONS(1372),
    [anon_sym_QMARK_GT] = ACTIONS(1370),
    [anon_sym_SEMI] = ACTIONS(1370),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1372),
    [aux_sym_global_declaration_token1] = ACTIONS(1372),
    [aux_sym_namespace_definition_token1] = ACTIONS(1372),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1372),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1372),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1372),
    [anon_sym_BSLASH] = ACTIONS(1370),
    [anon_sym_LBRACE] = ACTIONS(1370),
    [anon_sym_RBRACE] = ACTIONS(1370),
    [aux_sym_trait_declaration_token1] = ACTIONS(1372),
    [aux_sym_interface_declaration_token1] = ACTIONS(1372),
    [aux_sym_enum_declaration_token1] = ACTIONS(1372),
    [aux_sym_enum_case_token1] = ACTIONS(1372),
    [aux_sym_class_declaration_token1] = ACTIONS(1372),
    [aux_sym_final_modifier_token1] = ACTIONS(1372),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1372),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1372),
    [sym_var_modifier] = ACTIONS(1372),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1372),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1372),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1372),
    [anon_sym_LPAREN2] = ACTIONS(1370),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1372),
    [aux_sym_cast_type_token1] = ACTIONS(1372),
    [aux_sym_echo_statement_token1] = ACTIONS(1372),
    [aux_sym_exit_statement_token1] = ACTIONS(1372),
    [anon_sym_unset] = ACTIONS(1372),
    [aux_sym_declare_statement_token1] = ACTIONS(1372),
    [aux_sym_declare_statement_token2] = ACTIONS(1372),
    [sym_float] = ACTIONS(1372),
    [aux_sym_try_statement_token1] = ACTIONS(1372),
    [aux_sym_goto_statement_token1] = ACTIONS(1372),
    [aux_sym_continue_statement_token1] = ACTIONS(1372),
    [aux_sym_break_statement_token1] = ACTIONS(1372),
    [sym_integer] = ACTIONS(1372),
    [aux_sym_return_statement_token1] = ACTIONS(1372),
    [aux_sym_throw_expression_token1] = ACTIONS(1372),
    [aux_sym_while_statement_token1] = ACTIONS(1372),
    [aux_sym_while_statement_token2] = ACTIONS(1372),
    [aux_sym_do_statement_token1] = ACTIONS(1372),
    [aux_sym_for_statement_token1] = ACTIONS(1372),
    [aux_sym_for_statement_token2] = ACTIONS(1372),
    [aux_sym_foreach_statement_token1] = ACTIONS(1372),
    [aux_sym_foreach_statement_token2] = ACTIONS(1372),
    [aux_sym_if_statement_token1] = ACTIONS(1372),
    [aux_sym_if_statement_token2] = ACTIONS(1372),
    [aux_sym_else_if_clause_token1] = ACTIONS(1372),
    [aux_sym_else_clause_token1] = ACTIONS(1372),
    [aux_sym_match_expression_token1] = ACTIONS(1372),
    [aux_sym_match_default_expression_token1] = ACTIONS(1372),
    [aux_sym_switch_statement_token1] = ACTIONS(1372),
    [aux_sym_switch_block_token1] = ACTIONS(1372),
    [anon_sym_PLUS] = ACTIONS(1372),
    [anon_sym_DASH] = ACTIONS(1372),
    [anon_sym_TILDE] = ACTIONS(1370),
    [anon_sym_BANG] = ACTIONS(1370),
    [anon_sym_AT] = ACTIONS(1370),
    [aux_sym_clone_expression_token1] = ACTIONS(1372),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1372),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1372),
    [anon_sym_DASH_DASH] = ACTIONS(1370),
    [anon_sym_PLUS_PLUS] = ACTIONS(1370),
    [aux_sym__list_destructing_token1] = ACTIONS(1372),
    [anon_sym_LBRACK] = ACTIONS(1370),
    [anon_sym_self] = ACTIONS(1372),
    [anon_sym_parent] = ACTIONS(1372),
    [aux_sym__argument_name_token1] = ACTIONS(1372),
    [aux_sym__argument_name_token2] = ACTIONS(1372),
    [anon_sym_POUND_LBRACK] = ACTIONS(1370),
    [aux_sym_encapsed_string_token1] = ACTIONS(1370),
    [anon_sym_DQUOTE] = ACTIONS(1370),
    [aux_sym_string_token1] = ACTIONS(1370),
    [anon_sym_SQUOTE] = ACTIONS(1370),
    [anon_sym_LT_LT_LT] = ACTIONS(1370),
    [anon_sym_BQUOTE] = ACTIONS(1370),
    [anon_sym_DOLLAR] = ACTIONS(1370),
    [aux_sym_yield_expression_token1] = ACTIONS(1372),
    [aux_sym_include_expression_token1] = ACTIONS(1372),
    [aux_sym_include_once_expression_token1] = ACTIONS(1372),
    [aux_sym_require_expression_token1] = ACTIONS(1372),
    [aux_sym_require_once_expression_token1] = ACTIONS(1372),
    [sym_comment] = ACTIONS(3),
  },
  [525] = {
    [ts_builtin_sym_end] = ACTIONS(1374),
    [sym_name] = ACTIONS(1376),
    [anon_sym_QMARK_GT] = ACTIONS(1374),
    [anon_sym_SEMI] = ACTIONS(1374),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1376),
    [aux_sym_global_declaration_token1] = ACTIONS(1376),
    [aux_sym_namespace_definition_token1] = ACTIONS(1376),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1376),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1376),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1376),
    [anon_sym_BSLASH] = ACTIONS(1374),
    [anon_sym_LBRACE] = ACTIONS(1374),
    [anon_sym_RBRACE] = ACTIONS(1374),
    [aux_sym_trait_declaration_token1] = ACTIONS(1376),
    [aux_sym_interface_declaration_token1] = ACTIONS(1376),
    [aux_sym_enum_declaration_token1] = ACTIONS(1376),
    [aux_sym_enum_case_token1] = ACTIONS(1376),
    [aux_sym_class_declaration_token1] = ACTIONS(1376),
    [aux_sym_final_modifier_token1] = ACTIONS(1376),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1376),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1376),
    [sym_var_modifier] = ACTIONS(1376),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1376),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1376),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1376),
    [anon_sym_LPAREN2] = ACTIONS(1374),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1376),
    [aux_sym_cast_type_token1] = ACTIONS(1376),
    [aux_sym_echo_statement_token1] = ACTIONS(1376),
    [aux_sym_exit_statement_token1] = ACTIONS(1376),
    [anon_sym_unset] = ACTIONS(1376),
    [aux_sym_declare_statement_token1] = ACTIONS(1376),
    [aux_sym_declare_statement_token2] = ACTIONS(1376),
    [sym_float] = ACTIONS(1376),
    [aux_sym_try_statement_token1] = ACTIONS(1376),
    [aux_sym_goto_statement_token1] = ACTIONS(1376),
    [aux_sym_continue_statement_token1] = ACTIONS(1376),
    [aux_sym_break_statement_token1] = ACTIONS(1376),
    [sym_integer] = ACTIONS(1376),
    [aux_sym_return_statement_token1] = ACTIONS(1376),
    [aux_sym_throw_expression_token1] = ACTIONS(1376),
    [aux_sym_while_statement_token1] = ACTIONS(1376),
    [aux_sym_while_statement_token2] = ACTIONS(1376),
    [aux_sym_do_statement_token1] = ACTIONS(1376),
    [aux_sym_for_statement_token1] = ACTIONS(1376),
    [aux_sym_for_statement_token2] = ACTIONS(1376),
    [aux_sym_foreach_statement_token1] = ACTIONS(1376),
    [aux_sym_foreach_statement_token2] = ACTIONS(1376),
    [aux_sym_if_statement_token1] = ACTIONS(1376),
    [aux_sym_if_statement_token2] = ACTIONS(1376),
    [aux_sym_else_if_clause_token1] = ACTIONS(1376),
    [aux_sym_else_clause_token1] = ACTIONS(1376),
    [aux_sym_match_expression_token1] = ACTIONS(1376),
    [aux_sym_match_default_expression_token1] = ACTIONS(1376),
    [aux_sym_switch_statement_token1] = ACTIONS(1376),
    [aux_sym_switch_block_token1] = ACTIONS(1376),
    [anon_sym_PLUS] = ACTIONS(1376),
    [anon_sym_DASH] = ACTIONS(1376),
    [anon_sym_TILDE] = ACTIONS(1374),
    [anon_sym_BANG] = ACTIONS(1374),
    [anon_sym_AT] = ACTIONS(1374),
    [aux_sym_clone_expression_token1] = ACTIONS(1376),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1376),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1376),
    [anon_sym_DASH_DASH] = ACTIONS(1374),
    [anon_sym_PLUS_PLUS] = ACTIONS(1374),
    [aux_sym__list_destructing_token1] = ACTIONS(1376),
    [anon_sym_LBRACK] = ACTIONS(1374),
    [anon_sym_self] = ACTIONS(1376),
    [anon_sym_parent] = ACTIONS(1376),
    [aux_sym__argument_name_token1] = ACTIONS(1376),
    [aux_sym__argument_name_token2] = ACTIONS(1376),
    [anon_sym_POUND_LBRACK] = ACTIONS(1374),
    [aux_sym_encapsed_string_token1] = ACTIONS(1374),
    [anon_sym_DQUOTE] = ACTIONS(1374),
    [aux_sym_string_token1] = ACTIONS(1374),
    [anon_sym_SQUOTE] = ACTIONS(1374),
    [anon_sym_LT_LT_LT] = ACTIONS(1374),
    [anon_sym_BQUOTE] = ACTIONS(1374),
    [anon_sym_DOLLAR] = ACTIONS(1374),
    [aux_sym_yield_expression_token1] = ACTIONS(1376),
    [aux_sym_include_expression_token1] = ACTIONS(1376),
    [aux_sym_include_once_expression_token1] = ACTIONS(1376),
    [aux_sym_require_expression_token1] = ACTIONS(1376),
    [aux_sym_require_once_expression_token1] = ACTIONS(1376),
    [sym_comment] = ACTIONS(3),
  },
  [526] = {
    [ts_builtin_sym_end] = ACTIONS(1378),
    [sym_name] = ACTIONS(1380),
    [anon_sym_QMARK_GT] = ACTIONS(1378),
    [anon_sym_SEMI] = ACTIONS(1378),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1380),
    [aux_sym_global_declaration_token1] = ACTIONS(1380),
    [aux_sym_namespace_definition_token1] = ACTIONS(1380),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1380),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1380),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1380),
    [anon_sym_BSLASH] = ACTIONS(1378),
    [anon_sym_LBRACE] = ACTIONS(1378),
    [anon_sym_RBRACE] = ACTIONS(1378),
    [aux_sym_trait_declaration_token1] = ACTIONS(1380),
    [aux_sym_interface_declaration_token1] = ACTIONS(1380),
    [aux_sym_enum_declaration_token1] = ACTIONS(1380),
    [aux_sym_enum_case_token1] = ACTIONS(1380),
    [aux_sym_class_declaration_token1] = ACTIONS(1380),
    [aux_sym_final_modifier_token1] = ACTIONS(1380),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1380),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1380),
    [sym_var_modifier] = ACTIONS(1380),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1380),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1380),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1380),
    [anon_sym_LPAREN2] = ACTIONS(1378),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1380),
    [aux_sym_cast_type_token1] = ACTIONS(1380),
    [aux_sym_echo_statement_token1] = ACTIONS(1380),
    [aux_sym_exit_statement_token1] = ACTIONS(1380),
    [anon_sym_unset] = ACTIONS(1380),
    [aux_sym_declare_statement_token1] = ACTIONS(1380),
    [aux_sym_declare_statement_token2] = ACTIONS(1380),
    [sym_float] = ACTIONS(1380),
    [aux_sym_try_statement_token1] = ACTIONS(1380),
    [aux_sym_goto_statement_token1] = ACTIONS(1380),
    [aux_sym_continue_statement_token1] = ACTIONS(1380),
    [aux_sym_break_statement_token1] = ACTIONS(1380),
    [sym_integer] = ACTIONS(1380),
    [aux_sym_return_statement_token1] = ACTIONS(1380),
    [aux_sym_throw_expression_token1] = ACTIONS(1380),
    [aux_sym_while_statement_token1] = ACTIONS(1380),
    [aux_sym_while_statement_token2] = ACTIONS(1380),
    [aux_sym_do_statement_token1] = ACTIONS(1380),
    [aux_sym_for_statement_token1] = ACTIONS(1380),
    [aux_sym_for_statement_token2] = ACTIONS(1380),
    [aux_sym_foreach_statement_token1] = ACTIONS(1380),
    [aux_sym_foreach_statement_token2] = ACTIONS(1380),
    [aux_sym_if_statement_token1] = ACTIONS(1380),
    [aux_sym_if_statement_token2] = ACTIONS(1380),
    [aux_sym_else_if_clause_token1] = ACTIONS(1380),
    [aux_sym_else_clause_token1] = ACTIONS(1380),
    [aux_sym_match_expression_token1] = ACTIONS(1380),
    [aux_sym_match_default_expression_token1] = ACTIONS(1380),
    [aux_sym_switch_statement_token1] = ACTIONS(1380),
    [aux_sym_switch_block_token1] = ACTIONS(1380),
    [anon_sym_PLUS] = ACTIONS(1380),
    [anon_sym_DASH] = ACTIONS(1380),
    [anon_sym_TILDE] = ACTIONS(1378),
    [anon_sym_BANG] = ACTIONS(1378),
    [anon_sym_AT] = ACTIONS(1378),
    [aux_sym_clone_expression_token1] = ACTIONS(1380),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1380),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1380),
    [anon_sym_DASH_DASH] = ACTIONS(1378),
    [anon_sym_PLUS_PLUS] = ACTIONS(1378),
    [aux_sym__list_destructing_token1] = ACTIONS(1380),
    [anon_sym_LBRACK] = ACTIONS(1378),
    [anon_sym_self] = ACTIONS(1380),
    [anon_sym_parent] = ACTIONS(1380),
    [aux_sym__argument_name_token1] = ACTIONS(1380),
    [aux_sym__argument_name_token2] = ACTIONS(1380),
    [anon_sym_POUND_LBRACK] = ACTIONS(1378),
    [aux_sym_encapsed_string_token1] = ACTIONS(1378),
    [anon_sym_DQUOTE] = ACTIONS(1378),
    [aux_sym_string_token1] = ACTIONS(1378),
    [anon_sym_SQUOTE] = ACTIONS(1378),
    [anon_sym_LT_LT_LT] = ACTIONS(1378),
    [anon_sym_BQUOTE] = ACTIONS(1378),
    [anon_sym_DOLLAR] = ACTIONS(1378),
    [aux_sym_yield_expression_token1] = ACTIONS(1380),
    [aux_sym_include_expression_token1] = ACTIONS(1380),
    [aux_sym_include_once_expression_token1] = ACTIONS(1380),
    [aux_sym_require_expression_token1] = ACTIONS(1380),
    [aux_sym_require_once_expression_token1] = ACTIONS(1380),
    [sym_comment] = ACTIONS(3),
  },
  [527] = {
    [ts_builtin_sym_end] = ACTIONS(1382),
    [sym_name] = ACTIONS(1384),
    [anon_sym_QMARK_GT] = ACTIONS(1382),
    [anon_sym_SEMI] = ACTIONS(1382),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1384),
    [aux_sym_global_declaration_token1] = ACTIONS(1384),
    [aux_sym_namespace_definition_token1] = ACTIONS(1384),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1384),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1384),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1384),
    [anon_sym_BSLASH] = ACTIONS(1382),
    [anon_sym_LBRACE] = ACTIONS(1382),
    [anon_sym_RBRACE] = ACTIONS(1382),
    [aux_sym_trait_declaration_token1] = ACTIONS(1384),
    [aux_sym_interface_declaration_token1] = ACTIONS(1384),
    [aux_sym_enum_declaration_token1] = ACTIONS(1384),
    [aux_sym_enum_case_token1] = ACTIONS(1384),
    [aux_sym_class_declaration_token1] = ACTIONS(1384),
    [aux_sym_final_modifier_token1] = ACTIONS(1384),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1384),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1384),
    [sym_var_modifier] = ACTIONS(1384),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1384),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1384),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1384),
    [anon_sym_LPAREN2] = ACTIONS(1382),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1384),
    [aux_sym_cast_type_token1] = ACTIONS(1384),
    [aux_sym_echo_statement_token1] = ACTIONS(1384),
    [aux_sym_exit_statement_token1] = ACTIONS(1384),
    [anon_sym_unset] = ACTIONS(1384),
    [aux_sym_declare_statement_token1] = ACTIONS(1384),
    [aux_sym_declare_statement_token2] = ACTIONS(1384),
    [sym_float] = ACTIONS(1384),
    [aux_sym_try_statement_token1] = ACTIONS(1384),
    [aux_sym_goto_statement_token1] = ACTIONS(1384),
    [aux_sym_continue_statement_token1] = ACTIONS(1384),
    [aux_sym_break_statement_token1] = ACTIONS(1384),
    [sym_integer] = ACTIONS(1384),
    [aux_sym_return_statement_token1] = ACTIONS(1384),
    [aux_sym_throw_expression_token1] = ACTIONS(1384),
    [aux_sym_while_statement_token1] = ACTIONS(1384),
    [aux_sym_while_statement_token2] = ACTIONS(1384),
    [aux_sym_do_statement_token1] = ACTIONS(1384),
    [aux_sym_for_statement_token1] = ACTIONS(1384),
    [aux_sym_for_statement_token2] = ACTIONS(1384),
    [aux_sym_foreach_statement_token1] = ACTIONS(1384),
    [aux_sym_foreach_statement_token2] = ACTIONS(1384),
    [aux_sym_if_statement_token1] = ACTIONS(1384),
    [aux_sym_if_statement_token2] = ACTIONS(1384),
    [aux_sym_else_if_clause_token1] = ACTIONS(1384),
    [aux_sym_else_clause_token1] = ACTIONS(1384),
    [aux_sym_match_expression_token1] = ACTIONS(1384),
    [aux_sym_match_default_expression_token1] = ACTIONS(1384),
    [aux_sym_switch_statement_token1] = ACTIONS(1384),
    [aux_sym_switch_block_token1] = ACTIONS(1384),
    [anon_sym_PLUS] = ACTIONS(1384),
    [anon_sym_DASH] = ACTIONS(1384),
    [anon_sym_TILDE] = ACTIONS(1382),
    [anon_sym_BANG] = ACTIONS(1382),
    [anon_sym_AT] = ACTIONS(1382),
    [aux_sym_clone_expression_token1] = ACTIONS(1384),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1384),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1384),
    [anon_sym_DASH_DASH] = ACTIONS(1382),
    [anon_sym_PLUS_PLUS] = ACTIONS(1382),
    [aux_sym__list_destructing_token1] = ACTIONS(1384),
    [anon_sym_LBRACK] = ACTIONS(1382),
    [anon_sym_self] = ACTIONS(1384),
    [anon_sym_parent] = ACTIONS(1384),
    [aux_sym__argument_name_token1] = ACTIONS(1384),
    [aux_sym__argument_name_token2] = ACTIONS(1384),
    [anon_sym_POUND_LBRACK] = ACTIONS(1382),
    [aux_sym_encapsed_string_token1] = ACTIONS(1382),
    [anon_sym_DQUOTE] = ACTIONS(1382),
    [aux_sym_string_token1] = ACTIONS(1382),
    [anon_sym_SQUOTE] = ACTIONS(1382),
    [anon_sym_LT_LT_LT] = ACTIONS(1382),
    [anon_sym_BQUOTE] = ACTIONS(1382),
    [anon_sym_DOLLAR] = ACTIONS(1382),
    [aux_sym_yield_expression_token1] = ACTIONS(1384),
    [aux_sym_include_expression_token1] = ACTIONS(1384),
    [aux_sym_include_once_expression_token1] = ACTIONS(1384),
    [aux_sym_require_expression_token1] = ACTIONS(1384),
    [aux_sym_require_once_expression_token1] = ACTIONS(1384),
    [sym_comment] = ACTIONS(3),
  },
  [528] = {
    [ts_builtin_sym_end] = ACTIONS(1090),
    [sym_name] = ACTIONS(1092),
    [anon_sym_QMARK_GT] = ACTIONS(1090),
    [anon_sym_SEMI] = ACTIONS(1090),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1092),
    [aux_sym_global_declaration_token1] = ACTIONS(1092),
    [aux_sym_namespace_definition_token1] = ACTIONS(1092),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1092),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1092),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1092),
    [anon_sym_BSLASH] = ACTIONS(1090),
    [anon_sym_LBRACE] = ACTIONS(1090),
    [anon_sym_RBRACE] = ACTIONS(1090),
    [aux_sym_trait_declaration_token1] = ACTIONS(1092),
    [aux_sym_interface_declaration_token1] = ACTIONS(1092),
    [aux_sym_enum_declaration_token1] = ACTIONS(1092),
    [aux_sym_enum_case_token1] = ACTIONS(1092),
    [aux_sym_class_declaration_token1] = ACTIONS(1092),
    [aux_sym_final_modifier_token1] = ACTIONS(1092),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1092),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1092),
    [sym_var_modifier] = ACTIONS(1092),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1092),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1092),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1092),
    [anon_sym_LPAREN2] = ACTIONS(1090),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1092),
    [aux_sym_cast_type_token1] = ACTIONS(1092),
    [aux_sym_echo_statement_token1] = ACTIONS(1092),
    [aux_sym_exit_statement_token1] = ACTIONS(1092),
    [anon_sym_unset] = ACTIONS(1092),
    [aux_sym_declare_statement_token1] = ACTIONS(1092),
    [aux_sym_declare_statement_token2] = ACTIONS(1092),
    [sym_float] = ACTIONS(1092),
    [aux_sym_try_statement_token1] = ACTIONS(1092),
    [aux_sym_goto_statement_token1] = ACTIONS(1092),
    [aux_sym_continue_statement_token1] = ACTIONS(1092),
    [aux_sym_break_statement_token1] = ACTIONS(1092),
    [sym_integer] = ACTIONS(1092),
    [aux_sym_return_statement_token1] = ACTIONS(1092),
    [aux_sym_throw_expression_token1] = ACTIONS(1092),
    [aux_sym_while_statement_token1] = ACTIONS(1092),
    [aux_sym_while_statement_token2] = ACTIONS(1092),
    [aux_sym_do_statement_token1] = ACTIONS(1092),
    [aux_sym_for_statement_token1] = ACTIONS(1092),
    [aux_sym_for_statement_token2] = ACTIONS(1092),
    [aux_sym_foreach_statement_token1] = ACTIONS(1092),
    [aux_sym_foreach_statement_token2] = ACTIONS(1092),
    [aux_sym_if_statement_token1] = ACTIONS(1092),
    [aux_sym_if_statement_token2] = ACTIONS(1092),
    [aux_sym_else_if_clause_token1] = ACTIONS(1092),
    [aux_sym_else_clause_token1] = ACTIONS(1092),
    [aux_sym_match_expression_token1] = ACTIONS(1092),
    [aux_sym_match_default_expression_token1] = ACTIONS(1092),
    [aux_sym_switch_statement_token1] = ACTIONS(1092),
    [aux_sym_switch_block_token1] = ACTIONS(1092),
    [anon_sym_PLUS] = ACTIONS(1092),
    [anon_sym_DASH] = ACTIONS(1092),
    [anon_sym_TILDE] = ACTIONS(1090),
    [anon_sym_BANG] = ACTIONS(1090),
    [anon_sym_AT] = ACTIONS(1090),
    [aux_sym_clone_expression_token1] = ACTIONS(1092),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1092),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1092),
    [anon_sym_DASH_DASH] = ACTIONS(1090),
    [anon_sym_PLUS_PLUS] = ACTIONS(1090),
    [aux_sym__list_destructing_token1] = ACTIONS(1092),
    [anon_sym_LBRACK] = ACTIONS(1090),
    [anon_sym_self] = ACTIONS(1092),
    [anon_sym_parent] = ACTIONS(1092),
    [aux_sym__argument_name_token1] = ACTIONS(1092),
    [aux_sym__argument_name_token2] = ACTIONS(1092),
    [anon_sym_POUND_LBRACK] = ACTIONS(1090),
    [aux_sym_encapsed_string_token1] = ACTIONS(1090),
    [anon_sym_DQUOTE] = ACTIONS(1090),
    [aux_sym_string_token1] = ACTIONS(1090),
    [anon_sym_SQUOTE] = ACTIONS(1090),
    [anon_sym_LT_LT_LT] = ACTIONS(1090),
    [anon_sym_BQUOTE] = ACTIONS(1090),
    [anon_sym_DOLLAR] = ACTIONS(1090),
    [aux_sym_yield_expression_token1] = ACTIONS(1092),
    [aux_sym_include_expression_token1] = ACTIONS(1092),
    [aux_sym_include_once_expression_token1] = ACTIONS(1092),
    [aux_sym_require_expression_token1] = ACTIONS(1092),
    [aux_sym_require_once_expression_token1] = ACTIONS(1092),
    [sym_comment] = ACTIONS(3),
  },
  [529] = {
    [ts_builtin_sym_end] = ACTIONS(1386),
    [sym_name] = ACTIONS(1388),
    [anon_sym_QMARK_GT] = ACTIONS(1386),
    [anon_sym_SEMI] = ACTIONS(1386),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1388),
    [aux_sym_global_declaration_token1] = ACTIONS(1388),
    [aux_sym_namespace_definition_token1] = ACTIONS(1388),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1388),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1388),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1388),
    [anon_sym_BSLASH] = ACTIONS(1386),
    [anon_sym_LBRACE] = ACTIONS(1386),
    [anon_sym_RBRACE] = ACTIONS(1386),
    [aux_sym_trait_declaration_token1] = ACTIONS(1388),
    [aux_sym_interface_declaration_token1] = ACTIONS(1388),
    [aux_sym_enum_declaration_token1] = ACTIONS(1388),
    [aux_sym_enum_case_token1] = ACTIONS(1388),
    [aux_sym_class_declaration_token1] = ACTIONS(1388),
    [aux_sym_final_modifier_token1] = ACTIONS(1388),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1388),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1388),
    [sym_var_modifier] = ACTIONS(1388),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1388),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1388),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1388),
    [anon_sym_LPAREN2] = ACTIONS(1386),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1388),
    [aux_sym_cast_type_token1] = ACTIONS(1388),
    [aux_sym_echo_statement_token1] = ACTIONS(1388),
    [aux_sym_exit_statement_token1] = ACTIONS(1388),
    [anon_sym_unset] = ACTIONS(1388),
    [aux_sym_declare_statement_token1] = ACTIONS(1388),
    [aux_sym_declare_statement_token2] = ACTIONS(1388),
    [sym_float] = ACTIONS(1388),
    [aux_sym_try_statement_token1] = ACTIONS(1388),
    [aux_sym_goto_statement_token1] = ACTIONS(1388),
    [aux_sym_continue_statement_token1] = ACTIONS(1388),
    [aux_sym_break_statement_token1] = ACTIONS(1388),
    [sym_integer] = ACTIONS(1388),
    [aux_sym_return_statement_token1] = ACTIONS(1388),
    [aux_sym_throw_expression_token1] = ACTIONS(1388),
    [aux_sym_while_statement_token1] = ACTIONS(1388),
    [aux_sym_while_statement_token2] = ACTIONS(1388),
    [aux_sym_do_statement_token1] = ACTIONS(1388),
    [aux_sym_for_statement_token1] = ACTIONS(1388),
    [aux_sym_for_statement_token2] = ACTIONS(1388),
    [aux_sym_foreach_statement_token1] = ACTIONS(1388),
    [aux_sym_foreach_statement_token2] = ACTIONS(1388),
    [aux_sym_if_statement_token1] = ACTIONS(1388),
    [aux_sym_if_statement_token2] = ACTIONS(1388),
    [aux_sym_else_if_clause_token1] = ACTIONS(1388),
    [aux_sym_else_clause_token1] = ACTIONS(1388),
    [aux_sym_match_expression_token1] = ACTIONS(1388),
    [aux_sym_match_default_expression_token1] = ACTIONS(1388),
    [aux_sym_switch_statement_token1] = ACTIONS(1388),
    [aux_sym_switch_block_token1] = ACTIONS(1388),
    [anon_sym_PLUS] = ACTIONS(1388),
    [anon_sym_DASH] = ACTIONS(1388),
    [anon_sym_TILDE] = ACTIONS(1386),
    [anon_sym_BANG] = ACTIONS(1386),
    [anon_sym_AT] = ACTIONS(1386),
    [aux_sym_clone_expression_token1] = ACTIONS(1388),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1388),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1388),
    [anon_sym_DASH_DASH] = ACTIONS(1386),
    [anon_sym_PLUS_PLUS] = ACTIONS(1386),
    [aux_sym__list_destructing_token1] = ACTIONS(1388),
    [anon_sym_LBRACK] = ACTIONS(1386),
    [anon_sym_self] = ACTIONS(1388),
    [anon_sym_parent] = ACTIONS(1388),
    [aux_sym__argument_name_token1] = ACTIONS(1388),
    [aux_sym__argument_name_token2] = ACTIONS(1388),
    [anon_sym_POUND_LBRACK] = ACTIONS(1386),
    [aux_sym_encapsed_string_token1] = ACTIONS(1386),
    [anon_sym_DQUOTE] = ACTIONS(1386),
    [aux_sym_string_token1] = ACTIONS(1386),
    [anon_sym_SQUOTE] = ACTIONS(1386),
    [anon_sym_LT_LT_LT] = ACTIONS(1386),
    [anon_sym_BQUOTE] = ACTIONS(1386),
    [anon_sym_DOLLAR] = ACTIONS(1386),
    [aux_sym_yield_expression_token1] = ACTIONS(1388),
    [aux_sym_include_expression_token1] = ACTIONS(1388),
    [aux_sym_include_once_expression_token1] = ACTIONS(1388),
    [aux_sym_require_expression_token1] = ACTIONS(1388),
    [aux_sym_require_once_expression_token1] = ACTIONS(1388),
    [sym_comment] = ACTIONS(3),
  },
  [530] = {
    [ts_builtin_sym_end] = ACTIONS(1390),
    [sym_name] = ACTIONS(1392),
    [anon_sym_QMARK_GT] = ACTIONS(1390),
    [anon_sym_SEMI] = ACTIONS(1390),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1392),
    [aux_sym_global_declaration_token1] = ACTIONS(1392),
    [aux_sym_namespace_definition_token1] = ACTIONS(1392),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1392),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1392),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1392),
    [anon_sym_BSLASH] = ACTIONS(1390),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(1390),
    [aux_sym_trait_declaration_token1] = ACTIONS(1392),
    [aux_sym_interface_declaration_token1] = ACTIONS(1392),
    [aux_sym_enum_declaration_token1] = ACTIONS(1392),
    [aux_sym_enum_case_token1] = ACTIONS(1392),
    [aux_sym_class_declaration_token1] = ACTIONS(1392),
    [aux_sym_final_modifier_token1] = ACTIONS(1392),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1392),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1392),
    [sym_var_modifier] = ACTIONS(1392),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1392),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1392),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1392),
    [anon_sym_LPAREN2] = ACTIONS(1390),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1392),
    [aux_sym_cast_type_token1] = ACTIONS(1392),
    [aux_sym_echo_statement_token1] = ACTIONS(1392),
    [aux_sym_exit_statement_token1] = ACTIONS(1392),
    [anon_sym_unset] = ACTIONS(1392),
    [aux_sym_declare_statement_token1] = ACTIONS(1392),
    [aux_sym_declare_statement_token2] = ACTIONS(1392),
    [sym_float] = ACTIONS(1392),
    [aux_sym_try_statement_token1] = ACTIONS(1392),
    [aux_sym_goto_statement_token1] = ACTIONS(1392),
    [aux_sym_continue_statement_token1] = ACTIONS(1392),
    [aux_sym_break_statement_token1] = ACTIONS(1392),
    [sym_integer] = ACTIONS(1392),
    [aux_sym_return_statement_token1] = ACTIONS(1392),
    [aux_sym_throw_expression_token1] = ACTIONS(1392),
    [aux_sym_while_statement_token1] = ACTIONS(1392),
    [aux_sym_while_statement_token2] = ACTIONS(1392),
    [aux_sym_do_statement_token1] = ACTIONS(1392),
    [aux_sym_for_statement_token1] = ACTIONS(1392),
    [aux_sym_for_statement_token2] = ACTIONS(1392),
    [aux_sym_foreach_statement_token1] = ACTIONS(1392),
    [aux_sym_foreach_statement_token2] = ACTIONS(1392),
    [aux_sym_if_statement_token1] = ACTIONS(1392),
    [aux_sym_if_statement_token2] = ACTIONS(1392),
    [aux_sym_else_if_clause_token1] = ACTIONS(1392),
    [aux_sym_else_clause_token1] = ACTIONS(1392),
    [aux_sym_match_expression_token1] = ACTIONS(1392),
    [aux_sym_match_default_expression_token1] = ACTIONS(1392),
    [aux_sym_switch_statement_token1] = ACTIONS(1392),
    [aux_sym_switch_block_token1] = ACTIONS(1392),
    [anon_sym_PLUS] = ACTIONS(1392),
    [anon_sym_DASH] = ACTIONS(1392),
    [anon_sym_TILDE] = ACTIONS(1390),
    [anon_sym_BANG] = ACTIONS(1390),
    [anon_sym_AT] = ACTIONS(1390),
    [aux_sym_clone_expression_token1] = ACTIONS(1392),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1392),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1392),
    [anon_sym_DASH_DASH] = ACTIONS(1390),
    [anon_sym_PLUS_PLUS] = ACTIONS(1390),
    [aux_sym__list_destructing_token1] = ACTIONS(1392),
    [anon_sym_LBRACK] = ACTIONS(1390),
    [anon_sym_self] = ACTIONS(1392),
    [anon_sym_parent] = ACTIONS(1392),
    [aux_sym__argument_name_token1] = ACTIONS(1392),
    [aux_sym__argument_name_token2] = ACTIONS(1392),
    [anon_sym_POUND_LBRACK] = ACTIONS(1390),
    [aux_sym_encapsed_string_token1] = ACTIONS(1390),
    [anon_sym_DQUOTE] = ACTIONS(1390),
    [aux_sym_string_token1] = ACTIONS(1390),
    [anon_sym_SQUOTE] = ACTIONS(1390),
    [anon_sym_LT_LT_LT] = ACTIONS(1390),
    [anon_sym_BQUOTE] = ACTIONS(1390),
    [anon_sym_DOLLAR] = ACTIONS(1390),
    [aux_sym_yield_expression_token1] = ACTIONS(1392),
    [aux_sym_include_expression_token1] = ACTIONS(1392),
    [aux_sym_include_once_expression_token1] = ACTIONS(1392),
    [aux_sym_require_expression_token1] = ACTIONS(1392),
    [aux_sym_require_once_expression_token1] = ACTIONS(1392),
    [sym_comment] = ACTIONS(3),
  },
  [531] = {
    [ts_builtin_sym_end] = ACTIONS(1394),
    [sym_name] = ACTIONS(1396),
    [anon_sym_QMARK_GT] = ACTIONS(1394),
    [anon_sym_SEMI] = ACTIONS(1394),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1396),
    [aux_sym_global_declaration_token1] = ACTIONS(1396),
    [aux_sym_namespace_definition_token1] = ACTIONS(1396),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1396),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1396),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1396),
    [anon_sym_BSLASH] = ACTIONS(1394),
    [anon_sym_LBRACE] = ACTIONS(1394),
    [anon_sym_RBRACE] = ACTIONS(1394),
    [aux_sym_trait_declaration_token1] = ACTIONS(1396),
    [aux_sym_interface_declaration_token1] = ACTIONS(1396),
    [aux_sym_enum_declaration_token1] = ACTIONS(1396),
    [aux_sym_enum_case_token1] = ACTIONS(1396),
    [aux_sym_class_declaration_token1] = ACTIONS(1396),
    [aux_sym_final_modifier_token1] = ACTIONS(1396),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1396),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1396),
    [sym_var_modifier] = ACTIONS(1396),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1396),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1396),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1396),
    [anon_sym_LPAREN2] = ACTIONS(1394),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1396),
    [aux_sym_cast_type_token1] = ACTIONS(1396),
    [aux_sym_echo_statement_token1] = ACTIONS(1396),
    [aux_sym_exit_statement_token1] = ACTIONS(1396),
    [anon_sym_unset] = ACTIONS(1396),
    [aux_sym_declare_statement_token1] = ACTIONS(1396),
    [aux_sym_declare_statement_token2] = ACTIONS(1396),
    [sym_float] = ACTIONS(1396),
    [aux_sym_try_statement_token1] = ACTIONS(1396),
    [aux_sym_goto_statement_token1] = ACTIONS(1396),
    [aux_sym_continue_statement_token1] = ACTIONS(1396),
    [aux_sym_break_statement_token1] = ACTIONS(1396),
    [sym_integer] = ACTIONS(1396),
    [aux_sym_return_statement_token1] = ACTIONS(1396),
    [aux_sym_throw_expression_token1] = ACTIONS(1396),
    [aux_sym_while_statement_token1] = ACTIONS(1396),
    [aux_sym_while_statement_token2] = ACTIONS(1396),
    [aux_sym_do_statement_token1] = ACTIONS(1396),
    [aux_sym_for_statement_token1] = ACTIONS(1396),
    [aux_sym_for_statement_token2] = ACTIONS(1396),
    [aux_sym_foreach_statement_token1] = ACTIONS(1396),
    [aux_sym_foreach_statement_token2] = ACTIONS(1396),
    [aux_sym_if_statement_token1] = ACTIONS(1396),
    [aux_sym_if_statement_token2] = ACTIONS(1396),
    [aux_sym_else_if_clause_token1] = ACTIONS(1396),
    [aux_sym_else_clause_token1] = ACTIONS(1396),
    [aux_sym_match_expression_token1] = ACTIONS(1396),
    [aux_sym_match_default_expression_token1] = ACTIONS(1396),
    [aux_sym_switch_statement_token1] = ACTIONS(1396),
    [aux_sym_switch_block_token1] = ACTIONS(1396),
    [anon_sym_PLUS] = ACTIONS(1396),
    [anon_sym_DASH] = ACTIONS(1396),
    [anon_sym_TILDE] = ACTIONS(1394),
    [anon_sym_BANG] = ACTIONS(1394),
    [anon_sym_AT] = ACTIONS(1394),
    [aux_sym_clone_expression_token1] = ACTIONS(1396),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1396),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1396),
    [anon_sym_DASH_DASH] = ACTIONS(1394),
    [anon_sym_PLUS_PLUS] = ACTIONS(1394),
    [aux_sym__list_destructing_token1] = ACTIONS(1396),
    [anon_sym_LBRACK] = ACTIONS(1394),
    [anon_sym_self] = ACTIONS(1396),
    [anon_sym_parent] = ACTIONS(1396),
    [aux_sym__argument_name_token1] = ACTIONS(1396),
    [aux_sym__argument_name_token2] = ACTIONS(1396),
    [anon_sym_POUND_LBRACK] = ACTIONS(1394),
    [aux_sym_encapsed_string_token1] = ACTIONS(1394),
    [anon_sym_DQUOTE] = ACTIONS(1394),
    [aux_sym_string_token1] = ACTIONS(1394),
    [anon_sym_SQUOTE] = ACTIONS(1394),
    [anon_sym_LT_LT_LT] = ACTIONS(1394),
    [anon_sym_BQUOTE] = ACTIONS(1394),
    [anon_sym_DOLLAR] = ACTIONS(1394),
    [aux_sym_yield_expression_token1] = ACTIONS(1396),
    [aux_sym_include_expression_token1] = ACTIONS(1396),
    [aux_sym_include_once_expression_token1] = ACTIONS(1396),
    [aux_sym_require_expression_token1] = ACTIONS(1396),
    [aux_sym_require_once_expression_token1] = ACTIONS(1396),
    [sym_comment] = ACTIONS(3),
  },
  [532] = {
    [ts_builtin_sym_end] = ACTIONS(1398),
    [sym_name] = ACTIONS(1400),
    [anon_sym_QMARK_GT] = ACTIONS(1398),
    [anon_sym_SEMI] = ACTIONS(1398),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1400),
    [aux_sym_global_declaration_token1] = ACTIONS(1400),
    [aux_sym_namespace_definition_token1] = ACTIONS(1400),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1400),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1400),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1400),
    [anon_sym_BSLASH] = ACTIONS(1398),
    [anon_sym_LBRACE] = ACTIONS(1398),
    [anon_sym_RBRACE] = ACTIONS(1398),
    [aux_sym_trait_declaration_token1] = ACTIONS(1400),
    [aux_sym_interface_declaration_token1] = ACTIONS(1400),
    [aux_sym_enum_declaration_token1] = ACTIONS(1400),
    [aux_sym_enum_case_token1] = ACTIONS(1400),
    [aux_sym_class_declaration_token1] = ACTIONS(1400),
    [aux_sym_final_modifier_token1] = ACTIONS(1400),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1400),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1400),
    [sym_var_modifier] = ACTIONS(1400),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1400),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1400),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1400),
    [anon_sym_LPAREN2] = ACTIONS(1398),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1400),
    [aux_sym_cast_type_token1] = ACTIONS(1400),
    [aux_sym_echo_statement_token1] = ACTIONS(1400),
    [aux_sym_exit_statement_token1] = ACTIONS(1400),
    [anon_sym_unset] = ACTIONS(1400),
    [aux_sym_declare_statement_token1] = ACTIONS(1400),
    [aux_sym_declare_statement_token2] = ACTIONS(1400),
    [sym_float] = ACTIONS(1400),
    [aux_sym_try_statement_token1] = ACTIONS(1400),
    [aux_sym_goto_statement_token1] = ACTIONS(1400),
    [aux_sym_continue_statement_token1] = ACTIONS(1400),
    [aux_sym_break_statement_token1] = ACTIONS(1400),
    [sym_integer] = ACTIONS(1400),
    [aux_sym_return_statement_token1] = ACTIONS(1400),
    [aux_sym_throw_expression_token1] = ACTIONS(1400),
    [aux_sym_while_statement_token1] = ACTIONS(1400),
    [aux_sym_while_statement_token2] = ACTIONS(1400),
    [aux_sym_do_statement_token1] = ACTIONS(1400),
    [aux_sym_for_statement_token1] = ACTIONS(1400),
    [aux_sym_for_statement_token2] = ACTIONS(1400),
    [aux_sym_foreach_statement_token1] = ACTIONS(1400),
    [aux_sym_foreach_statement_token2] = ACTIONS(1400),
    [aux_sym_if_statement_token1] = ACTIONS(1400),
    [aux_sym_if_statement_token2] = ACTIONS(1400),
    [aux_sym_else_if_clause_token1] = ACTIONS(1400),
    [aux_sym_else_clause_token1] = ACTIONS(1400),
    [aux_sym_match_expression_token1] = ACTIONS(1400),
    [aux_sym_match_default_expression_token1] = ACTIONS(1400),
    [aux_sym_switch_statement_token1] = ACTIONS(1400),
    [aux_sym_switch_block_token1] = ACTIONS(1400),
    [anon_sym_PLUS] = ACTIONS(1400),
    [anon_sym_DASH] = ACTIONS(1400),
    [anon_sym_TILDE] = ACTIONS(1398),
    [anon_sym_BANG] = ACTIONS(1398),
    [anon_sym_AT] = ACTIONS(1398),
    [aux_sym_clone_expression_token1] = ACTIONS(1400),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1400),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1400),
    [anon_sym_DASH_DASH] = ACTIONS(1398),
    [anon_sym_PLUS_PLUS] = ACTIONS(1398),
    [aux_sym__list_destructing_token1] = ACTIONS(1400),
    [anon_sym_LBRACK] = ACTIONS(1398),
    [anon_sym_self] = ACTIONS(1400),
    [anon_sym_parent] = ACTIONS(1400),
    [aux_sym__argument_name_token1] = ACTIONS(1400),
    [aux_sym__argument_name_token2] = ACTIONS(1400),
    [anon_sym_POUND_LBRACK] = ACTIONS(1398),
    [aux_sym_encapsed_string_token1] = ACTIONS(1398),
    [anon_sym_DQUOTE] = ACTIONS(1398),
    [aux_sym_string_token1] = ACTIONS(1398),
    [anon_sym_SQUOTE] = ACTIONS(1398),
    [anon_sym_LT_LT_LT] = ACTIONS(1398),
    [anon_sym_BQUOTE] = ACTIONS(1398),
    [anon_sym_DOLLAR] = ACTIONS(1398),
    [aux_sym_yield_expression_token1] = ACTIONS(1400),
    [aux_sym_include_expression_token1] = ACTIONS(1400),
    [aux_sym_include_once_expression_token1] = ACTIONS(1400),
    [aux_sym_require_expression_token1] = ACTIONS(1400),
    [aux_sym_require_once_expression_token1] = ACTIONS(1400),
    [sym_comment] = ACTIONS(3),
  },
  [533] = {
    [ts_builtin_sym_end] = ACTIONS(1402),
    [sym_name] = ACTIONS(1404),
    [anon_sym_QMARK_GT] = ACTIONS(1402),
    [anon_sym_SEMI] = ACTIONS(1402),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1404),
    [aux_sym_global_declaration_token1] = ACTIONS(1404),
    [aux_sym_namespace_definition_token1] = ACTIONS(1404),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1404),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1404),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1404),
    [anon_sym_BSLASH] = ACTIONS(1402),
    [anon_sym_LBRACE] = ACTIONS(1402),
    [anon_sym_RBRACE] = ACTIONS(1402),
    [aux_sym_trait_declaration_token1] = ACTIONS(1404),
    [aux_sym_interface_declaration_token1] = ACTIONS(1404),
    [aux_sym_enum_declaration_token1] = ACTIONS(1404),
    [aux_sym_enum_case_token1] = ACTIONS(1404),
    [aux_sym_class_declaration_token1] = ACTIONS(1404),
    [aux_sym_final_modifier_token1] = ACTIONS(1404),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1404),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1404),
    [sym_var_modifier] = ACTIONS(1404),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1404),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1404),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1404),
    [anon_sym_LPAREN2] = ACTIONS(1402),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1404),
    [aux_sym_cast_type_token1] = ACTIONS(1404),
    [aux_sym_echo_statement_token1] = ACTIONS(1404),
    [aux_sym_exit_statement_token1] = ACTIONS(1404),
    [anon_sym_unset] = ACTIONS(1404),
    [aux_sym_declare_statement_token1] = ACTIONS(1404),
    [aux_sym_declare_statement_token2] = ACTIONS(1404),
    [sym_float] = ACTIONS(1404),
    [aux_sym_try_statement_token1] = ACTIONS(1404),
    [aux_sym_goto_statement_token1] = ACTIONS(1404),
    [aux_sym_continue_statement_token1] = ACTIONS(1404),
    [aux_sym_break_statement_token1] = ACTIONS(1404),
    [sym_integer] = ACTIONS(1404),
    [aux_sym_return_statement_token1] = ACTIONS(1404),
    [aux_sym_throw_expression_token1] = ACTIONS(1404),
    [aux_sym_while_statement_token1] = ACTIONS(1404),
    [aux_sym_while_statement_token2] = ACTIONS(1404),
    [aux_sym_do_statement_token1] = ACTIONS(1404),
    [aux_sym_for_statement_token1] = ACTIONS(1404),
    [aux_sym_for_statement_token2] = ACTIONS(1404),
    [aux_sym_foreach_statement_token1] = ACTIONS(1404),
    [aux_sym_foreach_statement_token2] = ACTIONS(1404),
    [aux_sym_if_statement_token1] = ACTIONS(1404),
    [aux_sym_if_statement_token2] = ACTIONS(1404),
    [aux_sym_else_if_clause_token1] = ACTIONS(1404),
    [aux_sym_else_clause_token1] = ACTIONS(1404),
    [aux_sym_match_expression_token1] = ACTIONS(1404),
    [aux_sym_match_default_expression_token1] = ACTIONS(1404),
    [aux_sym_switch_statement_token1] = ACTIONS(1404),
    [aux_sym_switch_block_token1] = ACTIONS(1404),
    [anon_sym_PLUS] = ACTIONS(1404),
    [anon_sym_DASH] = ACTIONS(1404),
    [anon_sym_TILDE] = ACTIONS(1402),
    [anon_sym_BANG] = ACTIONS(1402),
    [anon_sym_AT] = ACTIONS(1402),
    [aux_sym_clone_expression_token1] = ACTIONS(1404),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1404),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1404),
    [anon_sym_DASH_DASH] = ACTIONS(1402),
    [anon_sym_PLUS_PLUS] = ACTIONS(1402),
    [aux_sym__list_destructing_token1] = ACTIONS(1404),
    [anon_sym_LBRACK] = ACTIONS(1402),
    [anon_sym_self] = ACTIONS(1404),
    [anon_sym_parent] = ACTIONS(1404),
    [aux_sym__argument_name_token1] = ACTIONS(1404),
    [aux_sym__argument_name_token2] = ACTIONS(1404),
    [anon_sym_POUND_LBRACK] = ACTIONS(1402),
    [aux_sym_encapsed_string_token1] = ACTIONS(1402),
    [anon_sym_DQUOTE] = ACTIONS(1402),
    [aux_sym_string_token1] = ACTIONS(1402),
    [anon_sym_SQUOTE] = ACTIONS(1402),
    [anon_sym_LT_LT_LT] = ACTIONS(1402),
    [anon_sym_BQUOTE] = ACTIONS(1402),
    [anon_sym_DOLLAR] = ACTIONS(1402),
    [aux_sym_yield_expression_token1] = ACTIONS(1404),
    [aux_sym_include_expression_token1] = ACTIONS(1404),
    [aux_sym_include_once_expression_token1] = ACTIONS(1404),
    [aux_sym_require_expression_token1] = ACTIONS(1404),
    [aux_sym_require_once_expression_token1] = ACTIONS(1404),
    [sym_comment] = ACTIONS(3),
  },
  [534] = {
    [ts_builtin_sym_end] = ACTIONS(1406),
    [sym_name] = ACTIONS(1408),
    [anon_sym_QMARK_GT] = ACTIONS(1406),
    [anon_sym_SEMI] = ACTIONS(1406),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1408),
    [aux_sym_global_declaration_token1] = ACTIONS(1408),
    [aux_sym_namespace_definition_token1] = ACTIONS(1408),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1408),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1408),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1408),
    [anon_sym_BSLASH] = ACTIONS(1406),
    [anon_sym_LBRACE] = ACTIONS(1406),
    [anon_sym_RBRACE] = ACTIONS(1406),
    [aux_sym_trait_declaration_token1] = ACTIONS(1408),
    [aux_sym_interface_declaration_token1] = ACTIONS(1408),
    [aux_sym_enum_declaration_token1] = ACTIONS(1408),
    [aux_sym_enum_case_token1] = ACTIONS(1408),
    [aux_sym_class_declaration_token1] = ACTIONS(1408),
    [aux_sym_final_modifier_token1] = ACTIONS(1408),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1408),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1408),
    [sym_var_modifier] = ACTIONS(1408),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1408),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1408),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1408),
    [anon_sym_LPAREN2] = ACTIONS(1406),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1408),
    [aux_sym_cast_type_token1] = ACTIONS(1408),
    [aux_sym_echo_statement_token1] = ACTIONS(1408),
    [aux_sym_exit_statement_token1] = ACTIONS(1408),
    [anon_sym_unset] = ACTIONS(1408),
    [aux_sym_declare_statement_token1] = ACTIONS(1408),
    [aux_sym_declare_statement_token2] = ACTIONS(1408),
    [sym_float] = ACTIONS(1408),
    [aux_sym_try_statement_token1] = ACTIONS(1408),
    [aux_sym_goto_statement_token1] = ACTIONS(1408),
    [aux_sym_continue_statement_token1] = ACTIONS(1408),
    [aux_sym_break_statement_token1] = ACTIONS(1408),
    [sym_integer] = ACTIONS(1408),
    [aux_sym_return_statement_token1] = ACTIONS(1408),
    [aux_sym_throw_expression_token1] = ACTIONS(1408),
    [aux_sym_while_statement_token1] = ACTIONS(1408),
    [aux_sym_while_statement_token2] = ACTIONS(1408),
    [aux_sym_do_statement_token1] = ACTIONS(1408),
    [aux_sym_for_statement_token1] = ACTIONS(1408),
    [aux_sym_for_statement_token2] = ACTIONS(1408),
    [aux_sym_foreach_statement_token1] = ACTIONS(1408),
    [aux_sym_foreach_statement_token2] = ACTIONS(1408),
    [aux_sym_if_statement_token1] = ACTIONS(1408),
    [aux_sym_if_statement_token2] = ACTIONS(1408),
    [aux_sym_else_if_clause_token1] = ACTIONS(1408),
    [aux_sym_else_clause_token1] = ACTIONS(1408),
    [aux_sym_match_expression_token1] = ACTIONS(1408),
    [aux_sym_match_default_expression_token1] = ACTIONS(1408),
    [aux_sym_switch_statement_token1] = ACTIONS(1408),
    [aux_sym_switch_block_token1] = ACTIONS(1408),
    [anon_sym_PLUS] = ACTIONS(1408),
    [anon_sym_DASH] = ACTIONS(1408),
    [anon_sym_TILDE] = ACTIONS(1406),
    [anon_sym_BANG] = ACTIONS(1406),
    [anon_sym_AT] = ACTIONS(1406),
    [aux_sym_clone_expression_token1] = ACTIONS(1408),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1408),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1408),
    [anon_sym_DASH_DASH] = ACTIONS(1406),
    [anon_sym_PLUS_PLUS] = ACTIONS(1406),
    [aux_sym__list_destructing_token1] = ACTIONS(1408),
    [anon_sym_LBRACK] = ACTIONS(1406),
    [anon_sym_self] = ACTIONS(1408),
    [anon_sym_parent] = ACTIONS(1408),
    [aux_sym__argument_name_token1] = ACTIONS(1408),
    [aux_sym__argument_name_token2] = ACTIONS(1408),
    [anon_sym_POUND_LBRACK] = ACTIONS(1406),
    [aux_sym_encapsed_string_token1] = ACTIONS(1406),
    [anon_sym_DQUOTE] = ACTIONS(1406),
    [aux_sym_string_token1] = ACTIONS(1406),
    [anon_sym_SQUOTE] = ACTIONS(1406),
    [anon_sym_LT_LT_LT] = ACTIONS(1406),
    [anon_sym_BQUOTE] = ACTIONS(1406),
    [anon_sym_DOLLAR] = ACTIONS(1406),
    [aux_sym_yield_expression_token1] = ACTIONS(1408),
    [aux_sym_include_expression_token1] = ACTIONS(1408),
    [aux_sym_include_once_expression_token1] = ACTIONS(1408),
    [aux_sym_require_expression_token1] = ACTIONS(1408),
    [aux_sym_require_once_expression_token1] = ACTIONS(1408),
    [sym_comment] = ACTIONS(3),
  },
  [535] = {
    [ts_builtin_sym_end] = ACTIONS(1410),
    [sym_name] = ACTIONS(1412),
    [anon_sym_QMARK_GT] = ACTIONS(1410),
    [anon_sym_SEMI] = ACTIONS(1410),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1412),
    [aux_sym_global_declaration_token1] = ACTIONS(1412),
    [aux_sym_namespace_definition_token1] = ACTIONS(1412),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1412),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1412),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1412),
    [anon_sym_BSLASH] = ACTIONS(1410),
    [anon_sym_LBRACE] = ACTIONS(1410),
    [anon_sym_RBRACE] = ACTIONS(1410),
    [aux_sym_trait_declaration_token1] = ACTIONS(1412),
    [aux_sym_interface_declaration_token1] = ACTIONS(1412),
    [aux_sym_enum_declaration_token1] = ACTIONS(1412),
    [aux_sym_enum_case_token1] = ACTIONS(1412),
    [aux_sym_class_declaration_token1] = ACTIONS(1412),
    [aux_sym_final_modifier_token1] = ACTIONS(1412),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1412),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1412),
    [sym_var_modifier] = ACTIONS(1412),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1412),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1412),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1412),
    [anon_sym_LPAREN2] = ACTIONS(1410),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1412),
    [aux_sym_cast_type_token1] = ACTIONS(1412),
    [aux_sym_echo_statement_token1] = ACTIONS(1412),
    [aux_sym_exit_statement_token1] = ACTIONS(1412),
    [anon_sym_unset] = ACTIONS(1412),
    [aux_sym_declare_statement_token1] = ACTIONS(1412),
    [aux_sym_declare_statement_token2] = ACTIONS(1412),
    [sym_float] = ACTIONS(1412),
    [aux_sym_try_statement_token1] = ACTIONS(1412),
    [aux_sym_goto_statement_token1] = ACTIONS(1412),
    [aux_sym_continue_statement_token1] = ACTIONS(1412),
    [aux_sym_break_statement_token1] = ACTIONS(1412),
    [sym_integer] = ACTIONS(1412),
    [aux_sym_return_statement_token1] = ACTIONS(1412),
    [aux_sym_throw_expression_token1] = ACTIONS(1412),
    [aux_sym_while_statement_token1] = ACTIONS(1412),
    [aux_sym_while_statement_token2] = ACTIONS(1412),
    [aux_sym_do_statement_token1] = ACTIONS(1412),
    [aux_sym_for_statement_token1] = ACTIONS(1412),
    [aux_sym_for_statement_token2] = ACTIONS(1412),
    [aux_sym_foreach_statement_token1] = ACTIONS(1412),
    [aux_sym_foreach_statement_token2] = ACTIONS(1412),
    [aux_sym_if_statement_token1] = ACTIONS(1412),
    [aux_sym_if_statement_token2] = ACTIONS(1412),
    [aux_sym_else_if_clause_token1] = ACTIONS(1412),
    [aux_sym_else_clause_token1] = ACTIONS(1412),
    [aux_sym_match_expression_token1] = ACTIONS(1412),
    [aux_sym_match_default_expression_token1] = ACTIONS(1412),
    [aux_sym_switch_statement_token1] = ACTIONS(1412),
    [aux_sym_switch_block_token1] = ACTIONS(1412),
    [anon_sym_PLUS] = ACTIONS(1412),
    [anon_sym_DASH] = ACTIONS(1412),
    [anon_sym_TILDE] = ACTIONS(1410),
    [anon_sym_BANG] = ACTIONS(1410),
    [anon_sym_AT] = ACTIONS(1410),
    [aux_sym_clone_expression_token1] = ACTIONS(1412),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1412),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1412),
    [anon_sym_DASH_DASH] = ACTIONS(1410),
    [anon_sym_PLUS_PLUS] = ACTIONS(1410),
    [aux_sym__list_destructing_token1] = ACTIONS(1412),
    [anon_sym_LBRACK] = ACTIONS(1410),
    [anon_sym_self] = ACTIONS(1412),
    [anon_sym_parent] = ACTIONS(1412),
    [aux_sym__argument_name_token1] = ACTIONS(1412),
    [aux_sym__argument_name_token2] = ACTIONS(1412),
    [anon_sym_POUND_LBRACK] = ACTIONS(1410),
    [aux_sym_encapsed_string_token1] = ACTIONS(1410),
    [anon_sym_DQUOTE] = ACTIONS(1410),
    [aux_sym_string_token1] = ACTIONS(1410),
    [anon_sym_SQUOTE] = ACTIONS(1410),
    [anon_sym_LT_LT_LT] = ACTIONS(1410),
    [anon_sym_BQUOTE] = ACTIONS(1410),
    [anon_sym_DOLLAR] = ACTIONS(1410),
    [aux_sym_yield_expression_token1] = ACTIONS(1412),
    [aux_sym_include_expression_token1] = ACTIONS(1412),
    [aux_sym_include_once_expression_token1] = ACTIONS(1412),
    [aux_sym_require_expression_token1] = ACTIONS(1412),
    [aux_sym_require_once_expression_token1] = ACTIONS(1412),
    [sym_comment] = ACTIONS(3),
  },
  [536] = {
    [ts_builtin_sym_end] = ACTIONS(1414),
    [sym_name] = ACTIONS(1416),
    [anon_sym_QMARK_GT] = ACTIONS(1414),
    [anon_sym_SEMI] = ACTIONS(1414),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1416),
    [aux_sym_global_declaration_token1] = ACTIONS(1416),
    [aux_sym_namespace_definition_token1] = ACTIONS(1416),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1416),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1416),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1416),
    [anon_sym_BSLASH] = ACTIONS(1414),
    [anon_sym_LBRACE] = ACTIONS(1414),
    [anon_sym_RBRACE] = ACTIONS(1414),
    [aux_sym_trait_declaration_token1] = ACTIONS(1416),
    [aux_sym_interface_declaration_token1] = ACTIONS(1416),
    [aux_sym_enum_declaration_token1] = ACTIONS(1416),
    [aux_sym_enum_case_token1] = ACTIONS(1416),
    [aux_sym_class_declaration_token1] = ACTIONS(1416),
    [aux_sym_final_modifier_token1] = ACTIONS(1416),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1416),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1416),
    [sym_var_modifier] = ACTIONS(1416),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1416),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1416),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1416),
    [anon_sym_LPAREN2] = ACTIONS(1414),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1416),
    [aux_sym_cast_type_token1] = ACTIONS(1416),
    [aux_sym_echo_statement_token1] = ACTIONS(1416),
    [aux_sym_exit_statement_token1] = ACTIONS(1416),
    [anon_sym_unset] = ACTIONS(1416),
    [aux_sym_declare_statement_token1] = ACTIONS(1416),
    [aux_sym_declare_statement_token2] = ACTIONS(1416),
    [sym_float] = ACTIONS(1416),
    [aux_sym_try_statement_token1] = ACTIONS(1416),
    [aux_sym_goto_statement_token1] = ACTIONS(1416),
    [aux_sym_continue_statement_token1] = ACTIONS(1416),
    [aux_sym_break_statement_token1] = ACTIONS(1416),
    [sym_integer] = ACTIONS(1416),
    [aux_sym_return_statement_token1] = ACTIONS(1416),
    [aux_sym_throw_expression_token1] = ACTIONS(1416),
    [aux_sym_while_statement_token1] = ACTIONS(1416),
    [aux_sym_while_statement_token2] = ACTIONS(1416),
    [aux_sym_do_statement_token1] = ACTIONS(1416),
    [aux_sym_for_statement_token1] = ACTIONS(1416),
    [aux_sym_for_statement_token2] = ACTIONS(1416),
    [aux_sym_foreach_statement_token1] = ACTIONS(1416),
    [aux_sym_foreach_statement_token2] = ACTIONS(1416),
    [aux_sym_if_statement_token1] = ACTIONS(1416),
    [aux_sym_if_statement_token2] = ACTIONS(1416),
    [aux_sym_else_if_clause_token1] = ACTIONS(1416),
    [aux_sym_else_clause_token1] = ACTIONS(1416),
    [aux_sym_match_expression_token1] = ACTIONS(1416),
    [aux_sym_match_default_expression_token1] = ACTIONS(1416),
    [aux_sym_switch_statement_token1] = ACTIONS(1416),
    [aux_sym_switch_block_token1] = ACTIONS(1416),
    [anon_sym_PLUS] = ACTIONS(1416),
    [anon_sym_DASH] = ACTIONS(1416),
    [anon_sym_TILDE] = ACTIONS(1414),
    [anon_sym_BANG] = ACTIONS(1414),
    [anon_sym_AT] = ACTIONS(1414),
    [aux_sym_clone_expression_token1] = ACTIONS(1416),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1416),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1416),
    [anon_sym_DASH_DASH] = ACTIONS(1414),
    [anon_sym_PLUS_PLUS] = ACTIONS(1414),
    [aux_sym__list_destructing_token1] = ACTIONS(1416),
    [anon_sym_LBRACK] = ACTIONS(1414),
    [anon_sym_self] = ACTIONS(1416),
    [anon_sym_parent] = ACTIONS(1416),
    [aux_sym__argument_name_token1] = ACTIONS(1416),
    [aux_sym__argument_name_token2] = ACTIONS(1416),
    [anon_sym_POUND_LBRACK] = ACTIONS(1414),
    [aux_sym_encapsed_string_token1] = ACTIONS(1414),
    [anon_sym_DQUOTE] = ACTIONS(1414),
    [aux_sym_string_token1] = ACTIONS(1414),
    [anon_sym_SQUOTE] = ACTIONS(1414),
    [anon_sym_LT_LT_LT] = ACTIONS(1414),
    [anon_sym_BQUOTE] = ACTIONS(1414),
    [anon_sym_DOLLAR] = ACTIONS(1414),
    [aux_sym_yield_expression_token1] = ACTIONS(1416),
    [aux_sym_include_expression_token1] = ACTIONS(1416),
    [aux_sym_include_once_expression_token1] = ACTIONS(1416),
    [aux_sym_require_expression_token1] = ACTIONS(1416),
    [aux_sym_require_once_expression_token1] = ACTIONS(1416),
    [sym_comment] = ACTIONS(3),
  },
  [537] = {
    [ts_builtin_sym_end] = ACTIONS(1418),
    [sym_name] = ACTIONS(1420),
    [anon_sym_QMARK_GT] = ACTIONS(1418),
    [anon_sym_SEMI] = ACTIONS(1418),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1420),
    [aux_sym_global_declaration_token1] = ACTIONS(1420),
    [aux_sym_namespace_definition_token1] = ACTIONS(1420),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1420),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1420),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1420),
    [anon_sym_BSLASH] = ACTIONS(1418),
    [anon_sym_LBRACE] = ACTIONS(1418),
    [anon_sym_RBRACE] = ACTIONS(1418),
    [aux_sym_trait_declaration_token1] = ACTIONS(1420),
    [aux_sym_interface_declaration_token1] = ACTIONS(1420),
    [aux_sym_enum_declaration_token1] = ACTIONS(1420),
    [aux_sym_enum_case_token1] = ACTIONS(1420),
    [aux_sym_class_declaration_token1] = ACTIONS(1420),
    [aux_sym_final_modifier_token1] = ACTIONS(1420),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1420),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1420),
    [sym_var_modifier] = ACTIONS(1420),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1420),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1420),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1420),
    [anon_sym_LPAREN2] = ACTIONS(1418),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1420),
    [aux_sym_cast_type_token1] = ACTIONS(1420),
    [aux_sym_echo_statement_token1] = ACTIONS(1420),
    [aux_sym_exit_statement_token1] = ACTIONS(1420),
    [anon_sym_unset] = ACTIONS(1420),
    [aux_sym_declare_statement_token1] = ACTIONS(1420),
    [aux_sym_declare_statement_token2] = ACTIONS(1420),
    [sym_float] = ACTIONS(1420),
    [aux_sym_try_statement_token1] = ACTIONS(1420),
    [aux_sym_goto_statement_token1] = ACTIONS(1420),
    [aux_sym_continue_statement_token1] = ACTIONS(1420),
    [aux_sym_break_statement_token1] = ACTIONS(1420),
    [sym_integer] = ACTIONS(1420),
    [aux_sym_return_statement_token1] = ACTIONS(1420),
    [aux_sym_throw_expression_token1] = ACTIONS(1420),
    [aux_sym_while_statement_token1] = ACTIONS(1420),
    [aux_sym_while_statement_token2] = ACTIONS(1420),
    [aux_sym_do_statement_token1] = ACTIONS(1420),
    [aux_sym_for_statement_token1] = ACTIONS(1420),
    [aux_sym_for_statement_token2] = ACTIONS(1420),
    [aux_sym_foreach_statement_token1] = ACTIONS(1420),
    [aux_sym_foreach_statement_token2] = ACTIONS(1420),
    [aux_sym_if_statement_token1] = ACTIONS(1420),
    [aux_sym_if_statement_token2] = ACTIONS(1420),
    [aux_sym_else_if_clause_token1] = ACTIONS(1420),
    [aux_sym_else_clause_token1] = ACTIONS(1420),
    [aux_sym_match_expression_token1] = ACTIONS(1420),
    [aux_sym_match_default_expression_token1] = ACTIONS(1420),
    [aux_sym_switch_statement_token1] = ACTIONS(1420),
    [aux_sym_switch_block_token1] = ACTIONS(1420),
    [anon_sym_PLUS] = ACTIONS(1420),
    [anon_sym_DASH] = ACTIONS(1420),
    [anon_sym_TILDE] = ACTIONS(1418),
    [anon_sym_BANG] = ACTIONS(1418),
    [anon_sym_AT] = ACTIONS(1418),
    [aux_sym_clone_expression_token1] = ACTIONS(1420),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1420),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1420),
    [anon_sym_DASH_DASH] = ACTIONS(1418),
    [anon_sym_PLUS_PLUS] = ACTIONS(1418),
    [aux_sym__list_destructing_token1] = ACTIONS(1420),
    [anon_sym_LBRACK] = ACTIONS(1418),
    [anon_sym_self] = ACTIONS(1420),
    [anon_sym_parent] = ACTIONS(1420),
    [aux_sym__argument_name_token1] = ACTIONS(1420),
    [aux_sym__argument_name_token2] = ACTIONS(1420),
    [anon_sym_POUND_LBRACK] = ACTIONS(1418),
    [aux_sym_encapsed_string_token1] = ACTIONS(1418),
    [anon_sym_DQUOTE] = ACTIONS(1418),
    [aux_sym_string_token1] = ACTIONS(1418),
    [anon_sym_SQUOTE] = ACTIONS(1418),
    [anon_sym_LT_LT_LT] = ACTIONS(1418),
    [anon_sym_BQUOTE] = ACTIONS(1418),
    [anon_sym_DOLLAR] = ACTIONS(1418),
    [aux_sym_yield_expression_token1] = ACTIONS(1420),
    [aux_sym_include_expression_token1] = ACTIONS(1420),
    [aux_sym_include_once_expression_token1] = ACTIONS(1420),
    [aux_sym_require_expression_token1] = ACTIONS(1420),
    [aux_sym_require_once_expression_token1] = ACTIONS(1420),
    [sym_comment] = ACTIONS(3),
  },
  [538] = {
    [ts_builtin_sym_end] = ACTIONS(1422),
    [sym_name] = ACTIONS(1424),
    [anon_sym_QMARK_GT] = ACTIONS(1422),
    [anon_sym_SEMI] = ACTIONS(1422),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1424),
    [aux_sym_global_declaration_token1] = ACTIONS(1424),
    [aux_sym_namespace_definition_token1] = ACTIONS(1424),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1424),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1424),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1424),
    [anon_sym_BSLASH] = ACTIONS(1422),
    [anon_sym_LBRACE] = ACTIONS(1422),
    [anon_sym_RBRACE] = ACTIONS(1422),
    [aux_sym_trait_declaration_token1] = ACTIONS(1424),
    [aux_sym_interface_declaration_token1] = ACTIONS(1424),
    [aux_sym_enum_declaration_token1] = ACTIONS(1424),
    [aux_sym_enum_case_token1] = ACTIONS(1424),
    [aux_sym_class_declaration_token1] = ACTIONS(1424),
    [aux_sym_final_modifier_token1] = ACTIONS(1424),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1424),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1424),
    [sym_var_modifier] = ACTIONS(1424),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1424),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1424),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1424),
    [anon_sym_LPAREN2] = ACTIONS(1422),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1424),
    [aux_sym_cast_type_token1] = ACTIONS(1424),
    [aux_sym_echo_statement_token1] = ACTIONS(1424),
    [aux_sym_exit_statement_token1] = ACTIONS(1424),
    [anon_sym_unset] = ACTIONS(1424),
    [aux_sym_declare_statement_token1] = ACTIONS(1424),
    [aux_sym_declare_statement_token2] = ACTIONS(1424),
    [sym_float] = ACTIONS(1424),
    [aux_sym_try_statement_token1] = ACTIONS(1424),
    [aux_sym_goto_statement_token1] = ACTIONS(1424),
    [aux_sym_continue_statement_token1] = ACTIONS(1424),
    [aux_sym_break_statement_token1] = ACTIONS(1424),
    [sym_integer] = ACTIONS(1424),
    [aux_sym_return_statement_token1] = ACTIONS(1424),
    [aux_sym_throw_expression_token1] = ACTIONS(1424),
    [aux_sym_while_statement_token1] = ACTIONS(1424),
    [aux_sym_while_statement_token2] = ACTIONS(1424),
    [aux_sym_do_statement_token1] = ACTIONS(1424),
    [aux_sym_for_statement_token1] = ACTIONS(1424),
    [aux_sym_for_statement_token2] = ACTIONS(1424),
    [aux_sym_foreach_statement_token1] = ACTIONS(1424),
    [aux_sym_foreach_statement_token2] = ACTIONS(1424),
    [aux_sym_if_statement_token1] = ACTIONS(1424),
    [aux_sym_if_statement_token2] = ACTIONS(1424),
    [aux_sym_else_if_clause_token1] = ACTIONS(1424),
    [aux_sym_else_clause_token1] = ACTIONS(1424),
    [aux_sym_match_expression_token1] = ACTIONS(1424),
    [aux_sym_match_default_expression_token1] = ACTIONS(1424),
    [aux_sym_switch_statement_token1] = ACTIONS(1424),
    [aux_sym_switch_block_token1] = ACTIONS(1424),
    [anon_sym_PLUS] = ACTIONS(1424),
    [anon_sym_DASH] = ACTIONS(1424),
    [anon_sym_TILDE] = ACTIONS(1422),
    [anon_sym_BANG] = ACTIONS(1422),
    [anon_sym_AT] = ACTIONS(1422),
    [aux_sym_clone_expression_token1] = ACTIONS(1424),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1424),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1424),
    [anon_sym_DASH_DASH] = ACTIONS(1422),
    [anon_sym_PLUS_PLUS] = ACTIONS(1422),
    [aux_sym__list_destructing_token1] = ACTIONS(1424),
    [anon_sym_LBRACK] = ACTIONS(1422),
    [anon_sym_self] = ACTIONS(1424),
    [anon_sym_parent] = ACTIONS(1424),
    [aux_sym__argument_name_token1] = ACTIONS(1424),
    [aux_sym__argument_name_token2] = ACTIONS(1424),
    [anon_sym_POUND_LBRACK] = ACTIONS(1422),
    [aux_sym_encapsed_string_token1] = ACTIONS(1422),
    [anon_sym_DQUOTE] = ACTIONS(1422),
    [aux_sym_string_token1] = ACTIONS(1422),
    [anon_sym_SQUOTE] = ACTIONS(1422),
    [anon_sym_LT_LT_LT] = ACTIONS(1422),
    [anon_sym_BQUOTE] = ACTIONS(1422),
    [anon_sym_DOLLAR] = ACTIONS(1422),
    [aux_sym_yield_expression_token1] = ACTIONS(1424),
    [aux_sym_include_expression_token1] = ACTIONS(1424),
    [aux_sym_include_once_expression_token1] = ACTIONS(1424),
    [aux_sym_require_expression_token1] = ACTIONS(1424),
    [aux_sym_require_once_expression_token1] = ACTIONS(1424),
    [sym_comment] = ACTIONS(3),
  },
  [539] = {
    [ts_builtin_sym_end] = ACTIONS(1426),
    [sym_name] = ACTIONS(1428),
    [anon_sym_QMARK_GT] = ACTIONS(1426),
    [anon_sym_SEMI] = ACTIONS(1426),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1428),
    [aux_sym_global_declaration_token1] = ACTIONS(1428),
    [aux_sym_namespace_definition_token1] = ACTIONS(1428),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1428),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1428),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1428),
    [anon_sym_BSLASH] = ACTIONS(1426),
    [anon_sym_LBRACE] = ACTIONS(1426),
    [anon_sym_RBRACE] = ACTIONS(1426),
    [aux_sym_trait_declaration_token1] = ACTIONS(1428),
    [aux_sym_interface_declaration_token1] = ACTIONS(1428),
    [aux_sym_enum_declaration_token1] = ACTIONS(1428),
    [aux_sym_enum_case_token1] = ACTIONS(1428),
    [aux_sym_class_declaration_token1] = ACTIONS(1428),
    [aux_sym_final_modifier_token1] = ACTIONS(1428),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1428),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1428),
    [sym_var_modifier] = ACTIONS(1428),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1428),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1428),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1428),
    [anon_sym_LPAREN2] = ACTIONS(1426),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1428),
    [aux_sym_cast_type_token1] = ACTIONS(1428),
    [aux_sym_echo_statement_token1] = ACTIONS(1428),
    [aux_sym_exit_statement_token1] = ACTIONS(1428),
    [anon_sym_unset] = ACTIONS(1428),
    [aux_sym_declare_statement_token1] = ACTIONS(1428),
    [aux_sym_declare_statement_token2] = ACTIONS(1428),
    [sym_float] = ACTIONS(1428),
    [aux_sym_try_statement_token1] = ACTIONS(1428),
    [aux_sym_goto_statement_token1] = ACTIONS(1428),
    [aux_sym_continue_statement_token1] = ACTIONS(1428),
    [aux_sym_break_statement_token1] = ACTIONS(1428),
    [sym_integer] = ACTIONS(1428),
    [aux_sym_return_statement_token1] = ACTIONS(1428),
    [aux_sym_throw_expression_token1] = ACTIONS(1428),
    [aux_sym_while_statement_token1] = ACTIONS(1428),
    [aux_sym_while_statement_token2] = ACTIONS(1428),
    [aux_sym_do_statement_token1] = ACTIONS(1428),
    [aux_sym_for_statement_token1] = ACTIONS(1428),
    [aux_sym_for_statement_token2] = ACTIONS(1428),
    [aux_sym_foreach_statement_token1] = ACTIONS(1428),
    [aux_sym_foreach_statement_token2] = ACTIONS(1428),
    [aux_sym_if_statement_token1] = ACTIONS(1428),
    [aux_sym_if_statement_token2] = ACTIONS(1428),
    [aux_sym_else_if_clause_token1] = ACTIONS(1428),
    [aux_sym_else_clause_token1] = ACTIONS(1428),
    [aux_sym_match_expression_token1] = ACTIONS(1428),
    [aux_sym_match_default_expression_token1] = ACTIONS(1428),
    [aux_sym_switch_statement_token1] = ACTIONS(1428),
    [aux_sym_switch_block_token1] = ACTIONS(1428),
    [anon_sym_PLUS] = ACTIONS(1428),
    [anon_sym_DASH] = ACTIONS(1428),
    [anon_sym_TILDE] = ACTIONS(1426),
    [anon_sym_BANG] = ACTIONS(1426),
    [anon_sym_AT] = ACTIONS(1426),
    [aux_sym_clone_expression_token1] = ACTIONS(1428),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1428),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1428),
    [anon_sym_DASH_DASH] = ACTIONS(1426),
    [anon_sym_PLUS_PLUS] = ACTIONS(1426),
    [aux_sym__list_destructing_token1] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1426),
    [anon_sym_self] = ACTIONS(1428),
    [anon_sym_parent] = ACTIONS(1428),
    [aux_sym__argument_name_token1] = ACTIONS(1428),
    [aux_sym__argument_name_token2] = ACTIONS(1428),
    [anon_sym_POUND_LBRACK] = ACTIONS(1426),
    [aux_sym_encapsed_string_token1] = ACTIONS(1426),
    [anon_sym_DQUOTE] = ACTIONS(1426),
    [aux_sym_string_token1] = ACTIONS(1426),
    [anon_sym_SQUOTE] = ACTIONS(1426),
    [anon_sym_LT_LT_LT] = ACTIONS(1426),
    [anon_sym_BQUOTE] = ACTIONS(1426),
    [anon_sym_DOLLAR] = ACTIONS(1426),
    [aux_sym_yield_expression_token1] = ACTIONS(1428),
    [aux_sym_include_expression_token1] = ACTIONS(1428),
    [aux_sym_include_once_expression_token1] = ACTIONS(1428),
    [aux_sym_require_expression_token1] = ACTIONS(1428),
    [aux_sym_require_once_expression_token1] = ACTIONS(1428),
    [sym_comment] = ACTIONS(3),
  },
  [540] = {
    [ts_builtin_sym_end] = ACTIONS(1430),
    [sym_name] = ACTIONS(1432),
    [anon_sym_QMARK_GT] = ACTIONS(1430),
    [anon_sym_SEMI] = ACTIONS(1430),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1432),
    [aux_sym_global_declaration_token1] = ACTIONS(1432),
    [aux_sym_namespace_definition_token1] = ACTIONS(1432),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1432),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1432),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1432),
    [anon_sym_BSLASH] = ACTIONS(1430),
    [anon_sym_LBRACE] = ACTIONS(1430),
    [anon_sym_RBRACE] = ACTIONS(1430),
    [aux_sym_trait_declaration_token1] = ACTIONS(1432),
    [aux_sym_interface_declaration_token1] = ACTIONS(1432),
    [aux_sym_enum_declaration_token1] = ACTIONS(1432),
    [aux_sym_enum_case_token1] = ACTIONS(1432),
    [aux_sym_class_declaration_token1] = ACTIONS(1432),
    [aux_sym_final_modifier_token1] = ACTIONS(1432),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1432),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1432),
    [sym_var_modifier] = ACTIONS(1432),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1432),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1432),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1432),
    [anon_sym_LPAREN2] = ACTIONS(1430),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1432),
    [aux_sym_cast_type_token1] = ACTIONS(1432),
    [aux_sym_echo_statement_token1] = ACTIONS(1432),
    [aux_sym_exit_statement_token1] = ACTIONS(1432),
    [anon_sym_unset] = ACTIONS(1432),
    [aux_sym_declare_statement_token1] = ACTIONS(1432),
    [aux_sym_declare_statement_token2] = ACTIONS(1432),
    [sym_float] = ACTIONS(1432),
    [aux_sym_try_statement_token1] = ACTIONS(1432),
    [aux_sym_goto_statement_token1] = ACTIONS(1432),
    [aux_sym_continue_statement_token1] = ACTIONS(1432),
    [aux_sym_break_statement_token1] = ACTIONS(1432),
    [sym_integer] = ACTIONS(1432),
    [aux_sym_return_statement_token1] = ACTIONS(1432),
    [aux_sym_throw_expression_token1] = ACTIONS(1432),
    [aux_sym_while_statement_token1] = ACTIONS(1432),
    [aux_sym_while_statement_token2] = ACTIONS(1432),
    [aux_sym_do_statement_token1] = ACTIONS(1432),
    [aux_sym_for_statement_token1] = ACTIONS(1432),
    [aux_sym_for_statement_token2] = ACTIONS(1432),
    [aux_sym_foreach_statement_token1] = ACTIONS(1432),
    [aux_sym_foreach_statement_token2] = ACTIONS(1432),
    [aux_sym_if_statement_token1] = ACTIONS(1432),
    [aux_sym_if_statement_token2] = ACTIONS(1432),
    [aux_sym_else_if_clause_token1] = ACTIONS(1432),
    [aux_sym_else_clause_token1] = ACTIONS(1432),
    [aux_sym_match_expression_token1] = ACTIONS(1432),
    [aux_sym_match_default_expression_token1] = ACTIONS(1432),
    [aux_sym_switch_statement_token1] = ACTIONS(1432),
    [aux_sym_switch_block_token1] = ACTIONS(1432),
    [anon_sym_PLUS] = ACTIONS(1432),
    [anon_sym_DASH] = ACTIONS(1432),
    [anon_sym_TILDE] = ACTIONS(1430),
    [anon_sym_BANG] = ACTIONS(1430),
    [anon_sym_AT] = ACTIONS(1430),
    [aux_sym_clone_expression_token1] = ACTIONS(1432),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1432),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1432),
    [anon_sym_DASH_DASH] = ACTIONS(1430),
    [anon_sym_PLUS_PLUS] = ACTIONS(1430),
    [aux_sym__list_destructing_token1] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1430),
    [anon_sym_self] = ACTIONS(1432),
    [anon_sym_parent] = ACTIONS(1432),
    [aux_sym__argument_name_token1] = ACTIONS(1432),
    [aux_sym__argument_name_token2] = ACTIONS(1432),
    [anon_sym_POUND_LBRACK] = ACTIONS(1430),
    [aux_sym_encapsed_string_token1] = ACTIONS(1430),
    [anon_sym_DQUOTE] = ACTIONS(1430),
    [aux_sym_string_token1] = ACTIONS(1430),
    [anon_sym_SQUOTE] = ACTIONS(1430),
    [anon_sym_LT_LT_LT] = ACTIONS(1430),
    [anon_sym_BQUOTE] = ACTIONS(1430),
    [anon_sym_DOLLAR] = ACTIONS(1430),
    [aux_sym_yield_expression_token1] = ACTIONS(1432),
    [aux_sym_include_expression_token1] = ACTIONS(1432),
    [aux_sym_include_once_expression_token1] = ACTIONS(1432),
    [aux_sym_require_expression_token1] = ACTIONS(1432),
    [aux_sym_require_once_expression_token1] = ACTIONS(1432),
    [sym_comment] = ACTIONS(3),
  },
  [541] = {
    [ts_builtin_sym_end] = ACTIONS(1434),
    [sym_name] = ACTIONS(1436),
    [anon_sym_QMARK_GT] = ACTIONS(1434),
    [anon_sym_SEMI] = ACTIONS(1434),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1436),
    [aux_sym_global_declaration_token1] = ACTIONS(1436),
    [aux_sym_namespace_definition_token1] = ACTIONS(1436),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1436),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1436),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1436),
    [anon_sym_BSLASH] = ACTIONS(1434),
    [anon_sym_LBRACE] = ACTIONS(1434),
    [anon_sym_RBRACE] = ACTIONS(1434),
    [aux_sym_trait_declaration_token1] = ACTIONS(1436),
    [aux_sym_interface_declaration_token1] = ACTIONS(1436),
    [aux_sym_enum_declaration_token1] = ACTIONS(1436),
    [aux_sym_enum_case_token1] = ACTIONS(1436),
    [aux_sym_class_declaration_token1] = ACTIONS(1436),
    [aux_sym_final_modifier_token1] = ACTIONS(1436),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1436),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1436),
    [sym_var_modifier] = ACTIONS(1436),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1436),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1436),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1436),
    [anon_sym_LPAREN2] = ACTIONS(1434),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1436),
    [aux_sym_cast_type_token1] = ACTIONS(1436),
    [aux_sym_echo_statement_token1] = ACTIONS(1436),
    [aux_sym_exit_statement_token1] = ACTIONS(1436),
    [anon_sym_unset] = ACTIONS(1436),
    [aux_sym_declare_statement_token1] = ACTIONS(1436),
    [aux_sym_declare_statement_token2] = ACTIONS(1436),
    [sym_float] = ACTIONS(1436),
    [aux_sym_try_statement_token1] = ACTIONS(1436),
    [aux_sym_goto_statement_token1] = ACTIONS(1436),
    [aux_sym_continue_statement_token1] = ACTIONS(1436),
    [aux_sym_break_statement_token1] = ACTIONS(1436),
    [sym_integer] = ACTIONS(1436),
    [aux_sym_return_statement_token1] = ACTIONS(1436),
    [aux_sym_throw_expression_token1] = ACTIONS(1436),
    [aux_sym_while_statement_token1] = ACTIONS(1436),
    [aux_sym_while_statement_token2] = ACTIONS(1436),
    [aux_sym_do_statement_token1] = ACTIONS(1436),
    [aux_sym_for_statement_token1] = ACTIONS(1436),
    [aux_sym_for_statement_token2] = ACTIONS(1436),
    [aux_sym_foreach_statement_token1] = ACTIONS(1436),
    [aux_sym_foreach_statement_token2] = ACTIONS(1436),
    [aux_sym_if_statement_token1] = ACTIONS(1436),
    [aux_sym_if_statement_token2] = ACTIONS(1436),
    [aux_sym_else_if_clause_token1] = ACTIONS(1436),
    [aux_sym_else_clause_token1] = ACTIONS(1436),
    [aux_sym_match_expression_token1] = ACTIONS(1436),
    [aux_sym_match_default_expression_token1] = ACTIONS(1436),
    [aux_sym_switch_statement_token1] = ACTIONS(1436),
    [aux_sym_switch_block_token1] = ACTIONS(1436),
    [anon_sym_PLUS] = ACTIONS(1436),
    [anon_sym_DASH] = ACTIONS(1436),
    [anon_sym_TILDE] = ACTIONS(1434),
    [anon_sym_BANG] = ACTIONS(1434),
    [anon_sym_AT] = ACTIONS(1434),
    [aux_sym_clone_expression_token1] = ACTIONS(1436),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1436),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1436),
    [anon_sym_DASH_DASH] = ACTIONS(1434),
    [anon_sym_PLUS_PLUS] = ACTIONS(1434),
    [aux_sym__list_destructing_token1] = ACTIONS(1436),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_self] = ACTIONS(1436),
    [anon_sym_parent] = ACTIONS(1436),
    [aux_sym__argument_name_token1] = ACTIONS(1436),
    [aux_sym__argument_name_token2] = ACTIONS(1436),
    [anon_sym_POUND_LBRACK] = ACTIONS(1434),
    [aux_sym_encapsed_string_token1] = ACTIONS(1434),
    [anon_sym_DQUOTE] = ACTIONS(1434),
    [aux_sym_string_token1] = ACTIONS(1434),
    [anon_sym_SQUOTE] = ACTIONS(1434),
    [anon_sym_LT_LT_LT] = ACTIONS(1434),
    [anon_sym_BQUOTE] = ACTIONS(1434),
    [anon_sym_DOLLAR] = ACTIONS(1434),
    [aux_sym_yield_expression_token1] = ACTIONS(1436),
    [aux_sym_include_expression_token1] = ACTIONS(1436),
    [aux_sym_include_once_expression_token1] = ACTIONS(1436),
    [aux_sym_require_expression_token1] = ACTIONS(1436),
    [aux_sym_require_once_expression_token1] = ACTIONS(1436),
    [sym_comment] = ACTIONS(3),
  },
  [542] = {
    [ts_builtin_sym_end] = ACTIONS(1438),
    [sym_name] = ACTIONS(1440),
    [anon_sym_QMARK_GT] = ACTIONS(1438),
    [anon_sym_SEMI] = ACTIONS(1438),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1440),
    [aux_sym_global_declaration_token1] = ACTIONS(1440),
    [aux_sym_namespace_definition_token1] = ACTIONS(1440),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1440),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1440),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1440),
    [anon_sym_BSLASH] = ACTIONS(1438),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_RBRACE] = ACTIONS(1438),
    [aux_sym_trait_declaration_token1] = ACTIONS(1440),
    [aux_sym_interface_declaration_token1] = ACTIONS(1440),
    [aux_sym_enum_declaration_token1] = ACTIONS(1440),
    [aux_sym_enum_case_token1] = ACTIONS(1440),
    [aux_sym_class_declaration_token1] = ACTIONS(1440),
    [aux_sym_final_modifier_token1] = ACTIONS(1440),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1440),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1440),
    [sym_var_modifier] = ACTIONS(1440),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1440),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1440),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1440),
    [anon_sym_LPAREN2] = ACTIONS(1438),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1440),
    [aux_sym_cast_type_token1] = ACTIONS(1440),
    [aux_sym_echo_statement_token1] = ACTIONS(1440),
    [aux_sym_exit_statement_token1] = ACTIONS(1440),
    [anon_sym_unset] = ACTIONS(1440),
    [aux_sym_declare_statement_token1] = ACTIONS(1440),
    [aux_sym_declare_statement_token2] = ACTIONS(1440),
    [sym_float] = ACTIONS(1440),
    [aux_sym_try_statement_token1] = ACTIONS(1440),
    [aux_sym_goto_statement_token1] = ACTIONS(1440),
    [aux_sym_continue_statement_token1] = ACTIONS(1440),
    [aux_sym_break_statement_token1] = ACTIONS(1440),
    [sym_integer] = ACTIONS(1440),
    [aux_sym_return_statement_token1] = ACTIONS(1440),
    [aux_sym_throw_expression_token1] = ACTIONS(1440),
    [aux_sym_while_statement_token1] = ACTIONS(1440),
    [aux_sym_while_statement_token2] = ACTIONS(1440),
    [aux_sym_do_statement_token1] = ACTIONS(1440),
    [aux_sym_for_statement_token1] = ACTIONS(1440),
    [aux_sym_for_statement_token2] = ACTIONS(1440),
    [aux_sym_foreach_statement_token1] = ACTIONS(1440),
    [aux_sym_foreach_statement_token2] = ACTIONS(1440),
    [aux_sym_if_statement_token1] = ACTIONS(1440),
    [aux_sym_if_statement_token2] = ACTIONS(1440),
    [aux_sym_else_if_clause_token1] = ACTIONS(1440),
    [aux_sym_else_clause_token1] = ACTIONS(1440),
    [aux_sym_match_expression_token1] = ACTIONS(1440),
    [aux_sym_match_default_expression_token1] = ACTIONS(1440),
    [aux_sym_switch_statement_token1] = ACTIONS(1440),
    [aux_sym_switch_block_token1] = ACTIONS(1440),
    [anon_sym_PLUS] = ACTIONS(1440),
    [anon_sym_DASH] = ACTIONS(1440),
    [anon_sym_TILDE] = ACTIONS(1438),
    [anon_sym_BANG] = ACTIONS(1438),
    [anon_sym_AT] = ACTIONS(1438),
    [aux_sym_clone_expression_token1] = ACTIONS(1440),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1440),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1440),
    [anon_sym_DASH_DASH] = ACTIONS(1438),
    [anon_sym_PLUS_PLUS] = ACTIONS(1438),
    [aux_sym__list_destructing_token1] = ACTIONS(1440),
    [anon_sym_LBRACK] = ACTIONS(1438),
    [anon_sym_self] = ACTIONS(1440),
    [anon_sym_parent] = ACTIONS(1440),
    [aux_sym__argument_name_token1] = ACTIONS(1440),
    [aux_sym__argument_name_token2] = ACTIONS(1440),
    [anon_sym_POUND_LBRACK] = ACTIONS(1438),
    [aux_sym_encapsed_string_token1] = ACTIONS(1438),
    [anon_sym_DQUOTE] = ACTIONS(1438),
    [aux_sym_string_token1] = ACTIONS(1438),
    [anon_sym_SQUOTE] = ACTIONS(1438),
    [anon_sym_LT_LT_LT] = ACTIONS(1438),
    [anon_sym_BQUOTE] = ACTIONS(1438),
    [anon_sym_DOLLAR] = ACTIONS(1438),
    [aux_sym_yield_expression_token1] = ACTIONS(1440),
    [aux_sym_include_expression_token1] = ACTIONS(1440),
    [aux_sym_include_once_expression_token1] = ACTIONS(1440),
    [aux_sym_require_expression_token1] = ACTIONS(1440),
    [aux_sym_require_once_expression_token1] = ACTIONS(1440),
    [sym_comment] = ACTIONS(3),
  },
  [543] = {
    [ts_builtin_sym_end] = ACTIONS(1442),
    [sym_name] = ACTIONS(1444),
    [anon_sym_QMARK_GT] = ACTIONS(1442),
    [anon_sym_SEMI] = ACTIONS(1442),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1444),
    [aux_sym_global_declaration_token1] = ACTIONS(1444),
    [aux_sym_namespace_definition_token1] = ACTIONS(1444),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1444),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1444),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1444),
    [anon_sym_BSLASH] = ACTIONS(1442),
    [anon_sym_LBRACE] = ACTIONS(1442),
    [anon_sym_RBRACE] = ACTIONS(1442),
    [aux_sym_trait_declaration_token1] = ACTIONS(1444),
    [aux_sym_interface_declaration_token1] = ACTIONS(1444),
    [aux_sym_enum_declaration_token1] = ACTIONS(1444),
    [aux_sym_enum_case_token1] = ACTIONS(1444),
    [aux_sym_class_declaration_token1] = ACTIONS(1444),
    [aux_sym_final_modifier_token1] = ACTIONS(1444),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1444),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1444),
    [sym_var_modifier] = ACTIONS(1444),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1444),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1444),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1444),
    [anon_sym_LPAREN2] = ACTIONS(1442),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1444),
    [aux_sym_cast_type_token1] = ACTIONS(1444),
    [aux_sym_echo_statement_token1] = ACTIONS(1444),
    [aux_sym_exit_statement_token1] = ACTIONS(1444),
    [anon_sym_unset] = ACTIONS(1444),
    [aux_sym_declare_statement_token1] = ACTIONS(1444),
    [aux_sym_declare_statement_token2] = ACTIONS(1444),
    [sym_float] = ACTIONS(1444),
    [aux_sym_try_statement_token1] = ACTIONS(1444),
    [aux_sym_goto_statement_token1] = ACTIONS(1444),
    [aux_sym_continue_statement_token1] = ACTIONS(1444),
    [aux_sym_break_statement_token1] = ACTIONS(1444),
    [sym_integer] = ACTIONS(1444),
    [aux_sym_return_statement_token1] = ACTIONS(1444),
    [aux_sym_throw_expression_token1] = ACTIONS(1444),
    [aux_sym_while_statement_token1] = ACTIONS(1444),
    [aux_sym_while_statement_token2] = ACTIONS(1444),
    [aux_sym_do_statement_token1] = ACTIONS(1444),
    [aux_sym_for_statement_token1] = ACTIONS(1444),
    [aux_sym_for_statement_token2] = ACTIONS(1444),
    [aux_sym_foreach_statement_token1] = ACTIONS(1444),
    [aux_sym_foreach_statement_token2] = ACTIONS(1444),
    [aux_sym_if_statement_token1] = ACTIONS(1444),
    [aux_sym_if_statement_token2] = ACTIONS(1444),
    [aux_sym_else_if_clause_token1] = ACTIONS(1444),
    [aux_sym_else_clause_token1] = ACTIONS(1444),
    [aux_sym_match_expression_token1] = ACTIONS(1444),
    [aux_sym_match_default_expression_token1] = ACTIONS(1444),
    [aux_sym_switch_statement_token1] = ACTIONS(1444),
    [aux_sym_switch_block_token1] = ACTIONS(1444),
    [anon_sym_PLUS] = ACTIONS(1444),
    [anon_sym_DASH] = ACTIONS(1444),
    [anon_sym_TILDE] = ACTIONS(1442),
    [anon_sym_BANG] = ACTIONS(1442),
    [anon_sym_AT] = ACTIONS(1442),
    [aux_sym_clone_expression_token1] = ACTIONS(1444),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1444),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1444),
    [anon_sym_DASH_DASH] = ACTIONS(1442),
    [anon_sym_PLUS_PLUS] = ACTIONS(1442),
    [aux_sym__list_destructing_token1] = ACTIONS(1444),
    [anon_sym_LBRACK] = ACTIONS(1442),
    [anon_sym_self] = ACTIONS(1444),
    [anon_sym_parent] = ACTIONS(1444),
    [aux_sym__argument_name_token1] = ACTIONS(1444),
    [aux_sym__argument_name_token2] = ACTIONS(1444),
    [anon_sym_POUND_LBRACK] = ACTIONS(1442),
    [aux_sym_encapsed_string_token1] = ACTIONS(1442),
    [anon_sym_DQUOTE] = ACTIONS(1442),
    [aux_sym_string_token1] = ACTIONS(1442),
    [anon_sym_SQUOTE] = ACTIONS(1442),
    [anon_sym_LT_LT_LT] = ACTIONS(1442),
    [anon_sym_BQUOTE] = ACTIONS(1442),
    [anon_sym_DOLLAR] = ACTIONS(1442),
    [aux_sym_yield_expression_token1] = ACTIONS(1444),
    [aux_sym_include_expression_token1] = ACTIONS(1444),
    [aux_sym_include_once_expression_token1] = ACTIONS(1444),
    [aux_sym_require_expression_token1] = ACTIONS(1444),
    [aux_sym_require_once_expression_token1] = ACTIONS(1444),
    [sym_comment] = ACTIONS(3),
  },
  [544] = {
    [ts_builtin_sym_end] = ACTIONS(1446),
    [sym_name] = ACTIONS(1448),
    [anon_sym_QMARK_GT] = ACTIONS(1446),
    [anon_sym_SEMI] = ACTIONS(1446),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1448),
    [aux_sym_global_declaration_token1] = ACTIONS(1448),
    [aux_sym_namespace_definition_token1] = ACTIONS(1448),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1448),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1448),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1448),
    [anon_sym_BSLASH] = ACTIONS(1446),
    [anon_sym_LBRACE] = ACTIONS(1446),
    [anon_sym_RBRACE] = ACTIONS(1446),
    [aux_sym_trait_declaration_token1] = ACTIONS(1448),
    [aux_sym_interface_declaration_token1] = ACTIONS(1448),
    [aux_sym_enum_declaration_token1] = ACTIONS(1448),
    [aux_sym_enum_case_token1] = ACTIONS(1448),
    [aux_sym_class_declaration_token1] = ACTIONS(1448),
    [aux_sym_final_modifier_token1] = ACTIONS(1448),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1448),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1448),
    [sym_var_modifier] = ACTIONS(1448),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1448),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1448),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1448),
    [anon_sym_LPAREN2] = ACTIONS(1446),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1448),
    [aux_sym_cast_type_token1] = ACTIONS(1448),
    [aux_sym_echo_statement_token1] = ACTIONS(1448),
    [aux_sym_exit_statement_token1] = ACTIONS(1448),
    [anon_sym_unset] = ACTIONS(1448),
    [aux_sym_declare_statement_token1] = ACTIONS(1448),
    [aux_sym_declare_statement_token2] = ACTIONS(1448),
    [sym_float] = ACTIONS(1448),
    [aux_sym_try_statement_token1] = ACTIONS(1448),
    [aux_sym_goto_statement_token1] = ACTIONS(1448),
    [aux_sym_continue_statement_token1] = ACTIONS(1448),
    [aux_sym_break_statement_token1] = ACTIONS(1448),
    [sym_integer] = ACTIONS(1448),
    [aux_sym_return_statement_token1] = ACTIONS(1448),
    [aux_sym_throw_expression_token1] = ACTIONS(1448),
    [aux_sym_while_statement_token1] = ACTIONS(1448),
    [aux_sym_while_statement_token2] = ACTIONS(1448),
    [aux_sym_do_statement_token1] = ACTIONS(1448),
    [aux_sym_for_statement_token1] = ACTIONS(1448),
    [aux_sym_for_statement_token2] = ACTIONS(1448),
    [aux_sym_foreach_statement_token1] = ACTIONS(1448),
    [aux_sym_foreach_statement_token2] = ACTIONS(1448),
    [aux_sym_if_statement_token1] = ACTIONS(1448),
    [aux_sym_if_statement_token2] = ACTIONS(1448),
    [aux_sym_else_if_clause_token1] = ACTIONS(1448),
    [aux_sym_else_clause_token1] = ACTIONS(1448),
    [aux_sym_match_expression_token1] = ACTIONS(1448),
    [aux_sym_match_default_expression_token1] = ACTIONS(1448),
    [aux_sym_switch_statement_token1] = ACTIONS(1448),
    [aux_sym_switch_block_token1] = ACTIONS(1448),
    [anon_sym_PLUS] = ACTIONS(1448),
    [anon_sym_DASH] = ACTIONS(1448),
    [anon_sym_TILDE] = ACTIONS(1446),
    [anon_sym_BANG] = ACTIONS(1446),
    [anon_sym_AT] = ACTIONS(1446),
    [aux_sym_clone_expression_token1] = ACTIONS(1448),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1448),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1448),
    [anon_sym_DASH_DASH] = ACTIONS(1446),
    [anon_sym_PLUS_PLUS] = ACTIONS(1446),
    [aux_sym__list_destructing_token1] = ACTIONS(1448),
    [anon_sym_LBRACK] = ACTIONS(1446),
    [anon_sym_self] = ACTIONS(1448),
    [anon_sym_parent] = ACTIONS(1448),
    [aux_sym__argument_name_token1] = ACTIONS(1448),
    [aux_sym__argument_name_token2] = ACTIONS(1448),
    [anon_sym_POUND_LBRACK] = ACTIONS(1446),
    [aux_sym_encapsed_string_token1] = ACTIONS(1446),
    [anon_sym_DQUOTE] = ACTIONS(1446),
    [aux_sym_string_token1] = ACTIONS(1446),
    [anon_sym_SQUOTE] = ACTIONS(1446),
    [anon_sym_LT_LT_LT] = ACTIONS(1446),
    [anon_sym_BQUOTE] = ACTIONS(1446),
    [anon_sym_DOLLAR] = ACTIONS(1446),
    [aux_sym_yield_expression_token1] = ACTIONS(1448),
    [aux_sym_include_expression_token1] = ACTIONS(1448),
    [aux_sym_include_once_expression_token1] = ACTIONS(1448),
    [aux_sym_require_expression_token1] = ACTIONS(1448),
    [aux_sym_require_once_expression_token1] = ACTIONS(1448),
    [sym_comment] = ACTIONS(3),
  },
  [545] = {
    [ts_builtin_sym_end] = ACTIONS(1450),
    [sym_name] = ACTIONS(1452),
    [anon_sym_QMARK_GT] = ACTIONS(1450),
    [anon_sym_SEMI] = ACTIONS(1450),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1452),
    [aux_sym_global_declaration_token1] = ACTIONS(1452),
    [aux_sym_namespace_definition_token1] = ACTIONS(1452),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1452),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1452),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1452),
    [anon_sym_BSLASH] = ACTIONS(1450),
    [anon_sym_LBRACE] = ACTIONS(1450),
    [anon_sym_RBRACE] = ACTIONS(1450),
    [aux_sym_trait_declaration_token1] = ACTIONS(1452),
    [aux_sym_interface_declaration_token1] = ACTIONS(1452),
    [aux_sym_enum_declaration_token1] = ACTIONS(1452),
    [aux_sym_enum_case_token1] = ACTIONS(1452),
    [aux_sym_class_declaration_token1] = ACTIONS(1452),
    [aux_sym_final_modifier_token1] = ACTIONS(1452),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1452),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1452),
    [sym_var_modifier] = ACTIONS(1452),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1452),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1452),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1452),
    [anon_sym_LPAREN2] = ACTIONS(1450),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1452),
    [aux_sym_cast_type_token1] = ACTIONS(1452),
    [aux_sym_echo_statement_token1] = ACTIONS(1452),
    [aux_sym_exit_statement_token1] = ACTIONS(1452),
    [anon_sym_unset] = ACTIONS(1452),
    [aux_sym_declare_statement_token1] = ACTIONS(1452),
    [aux_sym_declare_statement_token2] = ACTIONS(1452),
    [sym_float] = ACTIONS(1452),
    [aux_sym_try_statement_token1] = ACTIONS(1452),
    [aux_sym_goto_statement_token1] = ACTIONS(1452),
    [aux_sym_continue_statement_token1] = ACTIONS(1452),
    [aux_sym_break_statement_token1] = ACTIONS(1452),
    [sym_integer] = ACTIONS(1452),
    [aux_sym_return_statement_token1] = ACTIONS(1452),
    [aux_sym_throw_expression_token1] = ACTIONS(1452),
    [aux_sym_while_statement_token1] = ACTIONS(1452),
    [aux_sym_while_statement_token2] = ACTIONS(1452),
    [aux_sym_do_statement_token1] = ACTIONS(1452),
    [aux_sym_for_statement_token1] = ACTIONS(1452),
    [aux_sym_for_statement_token2] = ACTIONS(1452),
    [aux_sym_foreach_statement_token1] = ACTIONS(1452),
    [aux_sym_foreach_statement_token2] = ACTIONS(1452),
    [aux_sym_if_statement_token1] = ACTIONS(1452),
    [aux_sym_if_statement_token2] = ACTIONS(1452),
    [aux_sym_else_if_clause_token1] = ACTIONS(1452),
    [aux_sym_else_clause_token1] = ACTIONS(1452),
    [aux_sym_match_expression_token1] = ACTIONS(1452),
    [aux_sym_match_default_expression_token1] = ACTIONS(1452),
    [aux_sym_switch_statement_token1] = ACTIONS(1452),
    [aux_sym_switch_block_token1] = ACTIONS(1452),
    [anon_sym_PLUS] = ACTIONS(1452),
    [anon_sym_DASH] = ACTIONS(1452),
    [anon_sym_TILDE] = ACTIONS(1450),
    [anon_sym_BANG] = ACTIONS(1450),
    [anon_sym_AT] = ACTIONS(1450),
    [aux_sym_clone_expression_token1] = ACTIONS(1452),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1452),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1452),
    [anon_sym_DASH_DASH] = ACTIONS(1450),
    [anon_sym_PLUS_PLUS] = ACTIONS(1450),
    [aux_sym__list_destructing_token1] = ACTIONS(1452),
    [anon_sym_LBRACK] = ACTIONS(1450),
    [anon_sym_self] = ACTIONS(1452),
    [anon_sym_parent] = ACTIONS(1452),
    [aux_sym__argument_name_token1] = ACTIONS(1452),
    [aux_sym__argument_name_token2] = ACTIONS(1452),
    [anon_sym_POUND_LBRACK] = ACTIONS(1450),
    [aux_sym_encapsed_string_token1] = ACTIONS(1450),
    [anon_sym_DQUOTE] = ACTIONS(1450),
    [aux_sym_string_token1] = ACTIONS(1450),
    [anon_sym_SQUOTE] = ACTIONS(1450),
    [anon_sym_LT_LT_LT] = ACTIONS(1450),
    [anon_sym_BQUOTE] = ACTIONS(1450),
    [anon_sym_DOLLAR] = ACTIONS(1450),
    [aux_sym_yield_expression_token1] = ACTIONS(1452),
    [aux_sym_include_expression_token1] = ACTIONS(1452),
    [aux_sym_include_once_expression_token1] = ACTIONS(1452),
    [aux_sym_require_expression_token1] = ACTIONS(1452),
    [aux_sym_require_once_expression_token1] = ACTIONS(1452),
    [sym_comment] = ACTIONS(3),
  },
  [546] = {
    [ts_builtin_sym_end] = ACTIONS(1454),
    [sym_name] = ACTIONS(1456),
    [anon_sym_QMARK_GT] = ACTIONS(1454),
    [anon_sym_SEMI] = ACTIONS(1454),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1456),
    [aux_sym_global_declaration_token1] = ACTIONS(1456),
    [aux_sym_namespace_definition_token1] = ACTIONS(1456),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1456),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1456),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1456),
    [anon_sym_BSLASH] = ACTIONS(1454),
    [anon_sym_LBRACE] = ACTIONS(1454),
    [anon_sym_RBRACE] = ACTIONS(1454),
    [aux_sym_trait_declaration_token1] = ACTIONS(1456),
    [aux_sym_interface_declaration_token1] = ACTIONS(1456),
    [aux_sym_enum_declaration_token1] = ACTIONS(1456),
    [aux_sym_enum_case_token1] = ACTIONS(1456),
    [aux_sym_class_declaration_token1] = ACTIONS(1456),
    [aux_sym_final_modifier_token1] = ACTIONS(1456),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1456),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1456),
    [sym_var_modifier] = ACTIONS(1456),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1456),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1456),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1456),
    [anon_sym_LPAREN2] = ACTIONS(1454),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1456),
    [aux_sym_cast_type_token1] = ACTIONS(1456),
    [aux_sym_echo_statement_token1] = ACTIONS(1456),
    [aux_sym_exit_statement_token1] = ACTIONS(1456),
    [anon_sym_unset] = ACTIONS(1456),
    [aux_sym_declare_statement_token1] = ACTIONS(1456),
    [aux_sym_declare_statement_token2] = ACTIONS(1456),
    [sym_float] = ACTIONS(1456),
    [aux_sym_try_statement_token1] = ACTIONS(1456),
    [aux_sym_goto_statement_token1] = ACTIONS(1456),
    [aux_sym_continue_statement_token1] = ACTIONS(1456),
    [aux_sym_break_statement_token1] = ACTIONS(1456),
    [sym_integer] = ACTIONS(1456),
    [aux_sym_return_statement_token1] = ACTIONS(1456),
    [aux_sym_throw_expression_token1] = ACTIONS(1456),
    [aux_sym_while_statement_token1] = ACTIONS(1456),
    [aux_sym_while_statement_token2] = ACTIONS(1456),
    [aux_sym_do_statement_token1] = ACTIONS(1456),
    [aux_sym_for_statement_token1] = ACTIONS(1456),
    [aux_sym_for_statement_token2] = ACTIONS(1456),
    [aux_sym_foreach_statement_token1] = ACTIONS(1456),
    [aux_sym_foreach_statement_token2] = ACTIONS(1456),
    [aux_sym_if_statement_token1] = ACTIONS(1456),
    [aux_sym_if_statement_token2] = ACTIONS(1456),
    [aux_sym_else_if_clause_token1] = ACTIONS(1456),
    [aux_sym_else_clause_token1] = ACTIONS(1456),
    [aux_sym_match_expression_token1] = ACTIONS(1456),
    [aux_sym_match_default_expression_token1] = ACTIONS(1456),
    [aux_sym_switch_statement_token1] = ACTIONS(1456),
    [aux_sym_switch_block_token1] = ACTIONS(1456),
    [anon_sym_PLUS] = ACTIONS(1456),
    [anon_sym_DASH] = ACTIONS(1456),
    [anon_sym_TILDE] = ACTIONS(1454),
    [anon_sym_BANG] = ACTIONS(1454),
    [anon_sym_AT] = ACTIONS(1454),
    [aux_sym_clone_expression_token1] = ACTIONS(1456),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1456),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1456),
    [anon_sym_DASH_DASH] = ACTIONS(1454),
    [anon_sym_PLUS_PLUS] = ACTIONS(1454),
    [aux_sym__list_destructing_token1] = ACTIONS(1456),
    [anon_sym_LBRACK] = ACTIONS(1454),
    [anon_sym_self] = ACTIONS(1456),
    [anon_sym_parent] = ACTIONS(1456),
    [aux_sym__argument_name_token1] = ACTIONS(1456),
    [aux_sym__argument_name_token2] = ACTIONS(1456),
    [anon_sym_POUND_LBRACK] = ACTIONS(1454),
    [aux_sym_encapsed_string_token1] = ACTIONS(1454),
    [anon_sym_DQUOTE] = ACTIONS(1454),
    [aux_sym_string_token1] = ACTIONS(1454),
    [anon_sym_SQUOTE] = ACTIONS(1454),
    [anon_sym_LT_LT_LT] = ACTIONS(1454),
    [anon_sym_BQUOTE] = ACTIONS(1454),
    [anon_sym_DOLLAR] = ACTIONS(1454),
    [aux_sym_yield_expression_token1] = ACTIONS(1456),
    [aux_sym_include_expression_token1] = ACTIONS(1456),
    [aux_sym_include_once_expression_token1] = ACTIONS(1456),
    [aux_sym_require_expression_token1] = ACTIONS(1456),
    [aux_sym_require_once_expression_token1] = ACTIONS(1456),
    [sym_comment] = ACTIONS(3),
  },
  [547] = {
    [ts_builtin_sym_end] = ACTIONS(1458),
    [sym_name] = ACTIONS(1460),
    [anon_sym_QMARK_GT] = ACTIONS(1458),
    [anon_sym_SEMI] = ACTIONS(1458),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1460),
    [aux_sym_global_declaration_token1] = ACTIONS(1460),
    [aux_sym_namespace_definition_token1] = ACTIONS(1460),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1460),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1460),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1460),
    [anon_sym_BSLASH] = ACTIONS(1458),
    [anon_sym_LBRACE] = ACTIONS(1458),
    [anon_sym_RBRACE] = ACTIONS(1458),
    [aux_sym_trait_declaration_token1] = ACTIONS(1460),
    [aux_sym_interface_declaration_token1] = ACTIONS(1460),
    [aux_sym_enum_declaration_token1] = ACTIONS(1460),
    [aux_sym_enum_case_token1] = ACTIONS(1460),
    [aux_sym_class_declaration_token1] = ACTIONS(1460),
    [aux_sym_final_modifier_token1] = ACTIONS(1460),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1460),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1460),
    [sym_var_modifier] = ACTIONS(1460),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1460),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1460),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1460),
    [anon_sym_LPAREN2] = ACTIONS(1458),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1460),
    [aux_sym_cast_type_token1] = ACTIONS(1460),
    [aux_sym_echo_statement_token1] = ACTIONS(1460),
    [aux_sym_exit_statement_token1] = ACTIONS(1460),
    [anon_sym_unset] = ACTIONS(1460),
    [aux_sym_declare_statement_token1] = ACTIONS(1460),
    [aux_sym_declare_statement_token2] = ACTIONS(1460),
    [sym_float] = ACTIONS(1460),
    [aux_sym_try_statement_token1] = ACTIONS(1460),
    [aux_sym_goto_statement_token1] = ACTIONS(1460),
    [aux_sym_continue_statement_token1] = ACTIONS(1460),
    [aux_sym_break_statement_token1] = ACTIONS(1460),
    [sym_integer] = ACTIONS(1460),
    [aux_sym_return_statement_token1] = ACTIONS(1460),
    [aux_sym_throw_expression_token1] = ACTIONS(1460),
    [aux_sym_while_statement_token1] = ACTIONS(1460),
    [aux_sym_while_statement_token2] = ACTIONS(1460),
    [aux_sym_do_statement_token1] = ACTIONS(1460),
    [aux_sym_for_statement_token1] = ACTIONS(1460),
    [aux_sym_for_statement_token2] = ACTIONS(1460),
    [aux_sym_foreach_statement_token1] = ACTIONS(1460),
    [aux_sym_foreach_statement_token2] = ACTIONS(1460),
    [aux_sym_if_statement_token1] = ACTIONS(1460),
    [aux_sym_if_statement_token2] = ACTIONS(1460),
    [aux_sym_else_if_clause_token1] = ACTIONS(1460),
    [aux_sym_else_clause_token1] = ACTIONS(1460),
    [aux_sym_match_expression_token1] = ACTIONS(1460),
    [aux_sym_match_default_expression_token1] = ACTIONS(1460),
    [aux_sym_switch_statement_token1] = ACTIONS(1460),
    [aux_sym_switch_block_token1] = ACTIONS(1460),
    [anon_sym_PLUS] = ACTIONS(1460),
    [anon_sym_DASH] = ACTIONS(1460),
    [anon_sym_TILDE] = ACTIONS(1458),
    [anon_sym_BANG] = ACTIONS(1458),
    [anon_sym_AT] = ACTIONS(1458),
    [aux_sym_clone_expression_token1] = ACTIONS(1460),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1460),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1460),
    [anon_sym_DASH_DASH] = ACTIONS(1458),
    [anon_sym_PLUS_PLUS] = ACTIONS(1458),
    [aux_sym__list_destructing_token1] = ACTIONS(1460),
    [anon_sym_LBRACK] = ACTIONS(1458),
    [anon_sym_self] = ACTIONS(1460),
    [anon_sym_parent] = ACTIONS(1460),
    [aux_sym__argument_name_token1] = ACTIONS(1460),
    [aux_sym__argument_name_token2] = ACTIONS(1460),
    [anon_sym_POUND_LBRACK] = ACTIONS(1458),
    [aux_sym_encapsed_string_token1] = ACTIONS(1458),
    [anon_sym_DQUOTE] = ACTIONS(1458),
    [aux_sym_string_token1] = ACTIONS(1458),
    [anon_sym_SQUOTE] = ACTIONS(1458),
    [anon_sym_LT_LT_LT] = ACTIONS(1458),
    [anon_sym_BQUOTE] = ACTIONS(1458),
    [anon_sym_DOLLAR] = ACTIONS(1458),
    [aux_sym_yield_expression_token1] = ACTIONS(1460),
    [aux_sym_include_expression_token1] = ACTIONS(1460),
    [aux_sym_include_once_expression_token1] = ACTIONS(1460),
    [aux_sym_require_expression_token1] = ACTIONS(1460),
    [aux_sym_require_once_expression_token1] = ACTIONS(1460),
    [sym_comment] = ACTIONS(3),
  },
  [548] = {
    [ts_builtin_sym_end] = ACTIONS(1462),
    [sym_name] = ACTIONS(1464),
    [anon_sym_QMARK_GT] = ACTIONS(1462),
    [anon_sym_SEMI] = ACTIONS(1462),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1464),
    [aux_sym_global_declaration_token1] = ACTIONS(1464),
    [aux_sym_namespace_definition_token1] = ACTIONS(1464),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1464),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1464),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1464),
    [anon_sym_BSLASH] = ACTIONS(1462),
    [anon_sym_LBRACE] = ACTIONS(1462),
    [anon_sym_RBRACE] = ACTIONS(1462),
    [aux_sym_trait_declaration_token1] = ACTIONS(1464),
    [aux_sym_interface_declaration_token1] = ACTIONS(1464),
    [aux_sym_enum_declaration_token1] = ACTIONS(1464),
    [aux_sym_enum_case_token1] = ACTIONS(1464),
    [aux_sym_class_declaration_token1] = ACTIONS(1464),
    [aux_sym_final_modifier_token1] = ACTIONS(1464),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1464),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1464),
    [sym_var_modifier] = ACTIONS(1464),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1464),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1464),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1464),
    [anon_sym_LPAREN2] = ACTIONS(1462),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1464),
    [aux_sym_cast_type_token1] = ACTIONS(1464),
    [aux_sym_echo_statement_token1] = ACTIONS(1464),
    [aux_sym_exit_statement_token1] = ACTIONS(1464),
    [anon_sym_unset] = ACTIONS(1464),
    [aux_sym_declare_statement_token1] = ACTIONS(1464),
    [aux_sym_declare_statement_token2] = ACTIONS(1464),
    [sym_float] = ACTIONS(1464),
    [aux_sym_try_statement_token1] = ACTIONS(1464),
    [aux_sym_goto_statement_token1] = ACTIONS(1464),
    [aux_sym_continue_statement_token1] = ACTIONS(1464),
    [aux_sym_break_statement_token1] = ACTIONS(1464),
    [sym_integer] = ACTIONS(1464),
    [aux_sym_return_statement_token1] = ACTIONS(1464),
    [aux_sym_throw_expression_token1] = ACTIONS(1464),
    [aux_sym_while_statement_token1] = ACTIONS(1464),
    [aux_sym_while_statement_token2] = ACTIONS(1464),
    [aux_sym_do_statement_token1] = ACTIONS(1464),
    [aux_sym_for_statement_token1] = ACTIONS(1464),
    [aux_sym_for_statement_token2] = ACTIONS(1464),
    [aux_sym_foreach_statement_token1] = ACTIONS(1464),
    [aux_sym_foreach_statement_token2] = ACTIONS(1464),
    [aux_sym_if_statement_token1] = ACTIONS(1464),
    [aux_sym_if_statement_token2] = ACTIONS(1464),
    [aux_sym_else_if_clause_token1] = ACTIONS(1464),
    [aux_sym_else_clause_token1] = ACTIONS(1464),
    [aux_sym_match_expression_token1] = ACTIONS(1464),
    [aux_sym_match_default_expression_token1] = ACTIONS(1464),
    [aux_sym_switch_statement_token1] = ACTIONS(1464),
    [aux_sym_switch_block_token1] = ACTIONS(1464),
    [anon_sym_PLUS] = ACTIONS(1464),
    [anon_sym_DASH] = ACTIONS(1464),
    [anon_sym_TILDE] = ACTIONS(1462),
    [anon_sym_BANG] = ACTIONS(1462),
    [anon_sym_AT] = ACTIONS(1462),
    [aux_sym_clone_expression_token1] = ACTIONS(1464),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1464),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1464),
    [anon_sym_DASH_DASH] = ACTIONS(1462),
    [anon_sym_PLUS_PLUS] = ACTIONS(1462),
    [aux_sym__list_destructing_token1] = ACTIONS(1464),
    [anon_sym_LBRACK] = ACTIONS(1462),
    [anon_sym_self] = ACTIONS(1464),
    [anon_sym_parent] = ACTIONS(1464),
    [aux_sym__argument_name_token1] = ACTIONS(1464),
    [aux_sym__argument_name_token2] = ACTIONS(1464),
    [anon_sym_POUND_LBRACK] = ACTIONS(1462),
    [aux_sym_encapsed_string_token1] = ACTIONS(1462),
    [anon_sym_DQUOTE] = ACTIONS(1462),
    [aux_sym_string_token1] = ACTIONS(1462),
    [anon_sym_SQUOTE] = ACTIONS(1462),
    [anon_sym_LT_LT_LT] = ACTIONS(1462),
    [anon_sym_BQUOTE] = ACTIONS(1462),
    [anon_sym_DOLLAR] = ACTIONS(1462),
    [aux_sym_yield_expression_token1] = ACTIONS(1464),
    [aux_sym_include_expression_token1] = ACTIONS(1464),
    [aux_sym_include_once_expression_token1] = ACTIONS(1464),
    [aux_sym_require_expression_token1] = ACTIONS(1464),
    [aux_sym_require_once_expression_token1] = ACTIONS(1464),
    [sym_comment] = ACTIONS(3),
  },
  [549] = {
    [ts_builtin_sym_end] = ACTIONS(1466),
    [sym_name] = ACTIONS(1468),
    [anon_sym_QMARK_GT] = ACTIONS(1466),
    [anon_sym_SEMI] = ACTIONS(1466),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1468),
    [aux_sym_global_declaration_token1] = ACTIONS(1468),
    [aux_sym_namespace_definition_token1] = ACTIONS(1468),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1468),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1468),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1468),
    [anon_sym_BSLASH] = ACTIONS(1466),
    [anon_sym_LBRACE] = ACTIONS(1466),
    [anon_sym_RBRACE] = ACTIONS(1466),
    [aux_sym_trait_declaration_token1] = ACTIONS(1468),
    [aux_sym_interface_declaration_token1] = ACTIONS(1468),
    [aux_sym_enum_declaration_token1] = ACTIONS(1468),
    [aux_sym_enum_case_token1] = ACTIONS(1468),
    [aux_sym_class_declaration_token1] = ACTIONS(1468),
    [aux_sym_final_modifier_token1] = ACTIONS(1468),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1468),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1468),
    [sym_var_modifier] = ACTIONS(1468),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1468),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1468),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1468),
    [anon_sym_LPAREN2] = ACTIONS(1466),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1468),
    [aux_sym_cast_type_token1] = ACTIONS(1468),
    [aux_sym_echo_statement_token1] = ACTIONS(1468),
    [aux_sym_exit_statement_token1] = ACTIONS(1468),
    [anon_sym_unset] = ACTIONS(1468),
    [aux_sym_declare_statement_token1] = ACTIONS(1468),
    [aux_sym_declare_statement_token2] = ACTIONS(1468),
    [sym_float] = ACTIONS(1468),
    [aux_sym_try_statement_token1] = ACTIONS(1468),
    [aux_sym_goto_statement_token1] = ACTIONS(1468),
    [aux_sym_continue_statement_token1] = ACTIONS(1468),
    [aux_sym_break_statement_token1] = ACTIONS(1468),
    [sym_integer] = ACTIONS(1468),
    [aux_sym_return_statement_token1] = ACTIONS(1468),
    [aux_sym_throw_expression_token1] = ACTIONS(1468),
    [aux_sym_while_statement_token1] = ACTIONS(1468),
    [aux_sym_while_statement_token2] = ACTIONS(1468),
    [aux_sym_do_statement_token1] = ACTIONS(1468),
    [aux_sym_for_statement_token1] = ACTIONS(1468),
    [aux_sym_for_statement_token2] = ACTIONS(1468),
    [aux_sym_foreach_statement_token1] = ACTIONS(1468),
    [aux_sym_foreach_statement_token2] = ACTIONS(1468),
    [aux_sym_if_statement_token1] = ACTIONS(1468),
    [aux_sym_if_statement_token2] = ACTIONS(1468),
    [aux_sym_else_if_clause_token1] = ACTIONS(1468),
    [aux_sym_else_clause_token1] = ACTIONS(1468),
    [aux_sym_match_expression_token1] = ACTIONS(1468),
    [aux_sym_match_default_expression_token1] = ACTIONS(1468),
    [aux_sym_switch_statement_token1] = ACTIONS(1468),
    [aux_sym_switch_block_token1] = ACTIONS(1468),
    [anon_sym_PLUS] = ACTIONS(1468),
    [anon_sym_DASH] = ACTIONS(1468),
    [anon_sym_TILDE] = ACTIONS(1466),
    [anon_sym_BANG] = ACTIONS(1466),
    [anon_sym_AT] = ACTIONS(1466),
    [aux_sym_clone_expression_token1] = ACTIONS(1468),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1468),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1468),
    [anon_sym_DASH_DASH] = ACTIONS(1466),
    [anon_sym_PLUS_PLUS] = ACTIONS(1466),
    [aux_sym__list_destructing_token1] = ACTIONS(1468),
    [anon_sym_LBRACK] = ACTIONS(1466),
    [anon_sym_self] = ACTIONS(1468),
    [anon_sym_parent] = ACTIONS(1468),
    [aux_sym__argument_name_token1] = ACTIONS(1468),
    [aux_sym__argument_name_token2] = ACTIONS(1468),
    [anon_sym_POUND_LBRACK] = ACTIONS(1466),
    [aux_sym_encapsed_string_token1] = ACTIONS(1466),
    [anon_sym_DQUOTE] = ACTIONS(1466),
    [aux_sym_string_token1] = ACTIONS(1466),
    [anon_sym_SQUOTE] = ACTIONS(1466),
    [anon_sym_LT_LT_LT] = ACTIONS(1466),
    [anon_sym_BQUOTE] = ACTIONS(1466),
    [anon_sym_DOLLAR] = ACTIONS(1466),
    [aux_sym_yield_expression_token1] = ACTIONS(1468),
    [aux_sym_include_expression_token1] = ACTIONS(1468),
    [aux_sym_include_once_expression_token1] = ACTIONS(1468),
    [aux_sym_require_expression_token1] = ACTIONS(1468),
    [aux_sym_require_once_expression_token1] = ACTIONS(1468),
    [sym_comment] = ACTIONS(3),
  },
  [550] = {
    [ts_builtin_sym_end] = ACTIONS(1470),
    [sym_name] = ACTIONS(1472),
    [anon_sym_QMARK_GT] = ACTIONS(1470),
    [anon_sym_SEMI] = ACTIONS(1470),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1472),
    [aux_sym_global_declaration_token1] = ACTIONS(1472),
    [aux_sym_namespace_definition_token1] = ACTIONS(1472),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1472),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1472),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1472),
    [anon_sym_BSLASH] = ACTIONS(1470),
    [anon_sym_LBRACE] = ACTIONS(1470),
    [anon_sym_RBRACE] = ACTIONS(1470),
    [aux_sym_trait_declaration_token1] = ACTIONS(1472),
    [aux_sym_interface_declaration_token1] = ACTIONS(1472),
    [aux_sym_enum_declaration_token1] = ACTIONS(1472),
    [aux_sym_enum_case_token1] = ACTIONS(1472),
    [aux_sym_class_declaration_token1] = ACTIONS(1472),
    [aux_sym_final_modifier_token1] = ACTIONS(1472),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1472),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1472),
    [sym_var_modifier] = ACTIONS(1472),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1472),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1472),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1472),
    [anon_sym_LPAREN2] = ACTIONS(1470),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1472),
    [aux_sym_cast_type_token1] = ACTIONS(1472),
    [aux_sym_echo_statement_token1] = ACTIONS(1472),
    [aux_sym_exit_statement_token1] = ACTIONS(1472),
    [anon_sym_unset] = ACTIONS(1472),
    [aux_sym_declare_statement_token1] = ACTIONS(1472),
    [aux_sym_declare_statement_token2] = ACTIONS(1472),
    [sym_float] = ACTIONS(1472),
    [aux_sym_try_statement_token1] = ACTIONS(1472),
    [aux_sym_goto_statement_token1] = ACTIONS(1472),
    [aux_sym_continue_statement_token1] = ACTIONS(1472),
    [aux_sym_break_statement_token1] = ACTIONS(1472),
    [sym_integer] = ACTIONS(1472),
    [aux_sym_return_statement_token1] = ACTIONS(1472),
    [aux_sym_throw_expression_token1] = ACTIONS(1472),
    [aux_sym_while_statement_token1] = ACTIONS(1472),
    [aux_sym_while_statement_token2] = ACTIONS(1472),
    [aux_sym_do_statement_token1] = ACTIONS(1472),
    [aux_sym_for_statement_token1] = ACTIONS(1472),
    [aux_sym_for_statement_token2] = ACTIONS(1472),
    [aux_sym_foreach_statement_token1] = ACTIONS(1472),
    [aux_sym_foreach_statement_token2] = ACTIONS(1472),
    [aux_sym_if_statement_token1] = ACTIONS(1472),
    [aux_sym_if_statement_token2] = ACTIONS(1472),
    [aux_sym_else_if_clause_token1] = ACTIONS(1472),
    [aux_sym_else_clause_token1] = ACTIONS(1472),
    [aux_sym_match_expression_token1] = ACTIONS(1472),
    [aux_sym_match_default_expression_token1] = ACTIONS(1472),
    [aux_sym_switch_statement_token1] = ACTIONS(1472),
    [aux_sym_switch_block_token1] = ACTIONS(1472),
    [anon_sym_PLUS] = ACTIONS(1472),
    [anon_sym_DASH] = ACTIONS(1472),
    [anon_sym_TILDE] = ACTIONS(1470),
    [anon_sym_BANG] = ACTIONS(1470),
    [anon_sym_AT] = ACTIONS(1470),
    [aux_sym_clone_expression_token1] = ACTIONS(1472),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1472),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1472),
    [anon_sym_DASH_DASH] = ACTIONS(1470),
    [anon_sym_PLUS_PLUS] = ACTIONS(1470),
    [aux_sym__list_destructing_token1] = ACTIONS(1472),
    [anon_sym_LBRACK] = ACTIONS(1470),
    [anon_sym_self] = ACTIONS(1472),
    [anon_sym_parent] = ACTIONS(1472),
    [aux_sym__argument_name_token1] = ACTIONS(1472),
    [aux_sym__argument_name_token2] = ACTIONS(1472),
    [anon_sym_POUND_LBRACK] = ACTIONS(1470),
    [aux_sym_encapsed_string_token1] = ACTIONS(1470),
    [anon_sym_DQUOTE] = ACTIONS(1470),
    [aux_sym_string_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1470),
    [anon_sym_LT_LT_LT] = ACTIONS(1470),
    [anon_sym_BQUOTE] = ACTIONS(1470),
    [anon_sym_DOLLAR] = ACTIONS(1470),
    [aux_sym_yield_expression_token1] = ACTIONS(1472),
    [aux_sym_include_expression_token1] = ACTIONS(1472),
    [aux_sym_include_once_expression_token1] = ACTIONS(1472),
    [aux_sym_require_expression_token1] = ACTIONS(1472),
    [aux_sym_require_once_expression_token1] = ACTIONS(1472),
    [sym_comment] = ACTIONS(3),
  },
  [551] = {
    [ts_builtin_sym_end] = ACTIONS(1474),
    [sym_name] = ACTIONS(1476),
    [anon_sym_QMARK_GT] = ACTIONS(1474),
    [anon_sym_SEMI] = ACTIONS(1474),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1476),
    [aux_sym_global_declaration_token1] = ACTIONS(1476),
    [aux_sym_namespace_definition_token1] = ACTIONS(1476),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1476),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1476),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1476),
    [anon_sym_BSLASH] = ACTIONS(1474),
    [anon_sym_LBRACE] = ACTIONS(1474),
    [anon_sym_RBRACE] = ACTIONS(1474),
    [aux_sym_trait_declaration_token1] = ACTIONS(1476),
    [aux_sym_interface_declaration_token1] = ACTIONS(1476),
    [aux_sym_enum_declaration_token1] = ACTIONS(1476),
    [aux_sym_enum_case_token1] = ACTIONS(1476),
    [aux_sym_class_declaration_token1] = ACTIONS(1476),
    [aux_sym_final_modifier_token1] = ACTIONS(1476),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1476),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1476),
    [sym_var_modifier] = ACTIONS(1476),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1476),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1476),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1476),
    [anon_sym_LPAREN2] = ACTIONS(1474),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1476),
    [aux_sym_cast_type_token1] = ACTIONS(1476),
    [aux_sym_echo_statement_token1] = ACTIONS(1476),
    [aux_sym_exit_statement_token1] = ACTIONS(1476),
    [anon_sym_unset] = ACTIONS(1476),
    [aux_sym_declare_statement_token1] = ACTIONS(1476),
    [aux_sym_declare_statement_token2] = ACTIONS(1476),
    [sym_float] = ACTIONS(1476),
    [aux_sym_try_statement_token1] = ACTIONS(1476),
    [aux_sym_goto_statement_token1] = ACTIONS(1476),
    [aux_sym_continue_statement_token1] = ACTIONS(1476),
    [aux_sym_break_statement_token1] = ACTIONS(1476),
    [sym_integer] = ACTIONS(1476),
    [aux_sym_return_statement_token1] = ACTIONS(1476),
    [aux_sym_throw_expression_token1] = ACTIONS(1476),
    [aux_sym_while_statement_token1] = ACTIONS(1476),
    [aux_sym_while_statement_token2] = ACTIONS(1476),
    [aux_sym_do_statement_token1] = ACTIONS(1476),
    [aux_sym_for_statement_token1] = ACTIONS(1476),
    [aux_sym_for_statement_token2] = ACTIONS(1476),
    [aux_sym_foreach_statement_token1] = ACTIONS(1476),
    [aux_sym_foreach_statement_token2] = ACTIONS(1476),
    [aux_sym_if_statement_token1] = ACTIONS(1476),
    [aux_sym_if_statement_token2] = ACTIONS(1476),
    [aux_sym_else_if_clause_token1] = ACTIONS(1476),
    [aux_sym_else_clause_token1] = ACTIONS(1476),
    [aux_sym_match_expression_token1] = ACTIONS(1476),
    [aux_sym_match_default_expression_token1] = ACTIONS(1476),
    [aux_sym_switch_statement_token1] = ACTIONS(1476),
    [aux_sym_switch_block_token1] = ACTIONS(1476),
    [anon_sym_PLUS] = ACTIONS(1476),
    [anon_sym_DASH] = ACTIONS(1476),
    [anon_sym_TILDE] = ACTIONS(1474),
    [anon_sym_BANG] = ACTIONS(1474),
    [anon_sym_AT] = ACTIONS(1474),
    [aux_sym_clone_expression_token1] = ACTIONS(1476),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1476),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1476),
    [anon_sym_DASH_DASH] = ACTIONS(1474),
    [anon_sym_PLUS_PLUS] = ACTIONS(1474),
    [aux_sym__list_destructing_token1] = ACTIONS(1476),
    [anon_sym_LBRACK] = ACTIONS(1474),
    [anon_sym_self] = ACTIONS(1476),
    [anon_sym_parent] = ACTIONS(1476),
    [aux_sym__argument_name_token1] = ACTIONS(1476),
    [aux_sym__argument_name_token2] = ACTIONS(1476),
    [anon_sym_POUND_LBRACK] = ACTIONS(1474),
    [aux_sym_encapsed_string_token1] = ACTIONS(1474),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [aux_sym_string_token1] = ACTIONS(1474),
    [anon_sym_SQUOTE] = ACTIONS(1474),
    [anon_sym_LT_LT_LT] = ACTIONS(1474),
    [anon_sym_BQUOTE] = ACTIONS(1474),
    [anon_sym_DOLLAR] = ACTIONS(1474),
    [aux_sym_yield_expression_token1] = ACTIONS(1476),
    [aux_sym_include_expression_token1] = ACTIONS(1476),
    [aux_sym_include_once_expression_token1] = ACTIONS(1476),
    [aux_sym_require_expression_token1] = ACTIONS(1476),
    [aux_sym_require_once_expression_token1] = ACTIONS(1476),
    [sym_comment] = ACTIONS(3),
  },
  [552] = {
    [ts_builtin_sym_end] = ACTIONS(1478),
    [sym_name] = ACTIONS(1480),
    [anon_sym_QMARK_GT] = ACTIONS(1478),
    [anon_sym_SEMI] = ACTIONS(1478),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1480),
    [aux_sym_global_declaration_token1] = ACTIONS(1480),
    [aux_sym_namespace_definition_token1] = ACTIONS(1480),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1480),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1480),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1480),
    [anon_sym_BSLASH] = ACTIONS(1478),
    [anon_sym_LBRACE] = ACTIONS(1478),
    [anon_sym_RBRACE] = ACTIONS(1478),
    [aux_sym_trait_declaration_token1] = ACTIONS(1480),
    [aux_sym_interface_declaration_token1] = ACTIONS(1480),
    [aux_sym_enum_declaration_token1] = ACTIONS(1480),
    [aux_sym_enum_case_token1] = ACTIONS(1480),
    [aux_sym_class_declaration_token1] = ACTIONS(1480),
    [aux_sym_final_modifier_token1] = ACTIONS(1480),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1480),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1480),
    [sym_var_modifier] = ACTIONS(1480),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1480),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1480),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1480),
    [anon_sym_LPAREN2] = ACTIONS(1478),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1480),
    [aux_sym_cast_type_token1] = ACTIONS(1480),
    [aux_sym_echo_statement_token1] = ACTIONS(1480),
    [aux_sym_exit_statement_token1] = ACTIONS(1480),
    [anon_sym_unset] = ACTIONS(1480),
    [aux_sym_declare_statement_token1] = ACTIONS(1480),
    [aux_sym_declare_statement_token2] = ACTIONS(1480),
    [sym_float] = ACTIONS(1480),
    [aux_sym_try_statement_token1] = ACTIONS(1480),
    [aux_sym_goto_statement_token1] = ACTIONS(1480),
    [aux_sym_continue_statement_token1] = ACTIONS(1480),
    [aux_sym_break_statement_token1] = ACTIONS(1480),
    [sym_integer] = ACTIONS(1480),
    [aux_sym_return_statement_token1] = ACTIONS(1480),
    [aux_sym_throw_expression_token1] = ACTIONS(1480),
    [aux_sym_while_statement_token1] = ACTIONS(1480),
    [aux_sym_while_statement_token2] = ACTIONS(1480),
    [aux_sym_do_statement_token1] = ACTIONS(1480),
    [aux_sym_for_statement_token1] = ACTIONS(1480),
    [aux_sym_for_statement_token2] = ACTIONS(1480),
    [aux_sym_foreach_statement_token1] = ACTIONS(1480),
    [aux_sym_foreach_statement_token2] = ACTIONS(1480),
    [aux_sym_if_statement_token1] = ACTIONS(1480),
    [aux_sym_if_statement_token2] = ACTIONS(1480),
    [aux_sym_else_if_clause_token1] = ACTIONS(1480),
    [aux_sym_else_clause_token1] = ACTIONS(1480),
    [aux_sym_match_expression_token1] = ACTIONS(1480),
    [aux_sym_match_default_expression_token1] = ACTIONS(1480),
    [aux_sym_switch_statement_token1] = ACTIONS(1480),
    [aux_sym_switch_block_token1] = ACTIONS(1480),
    [anon_sym_PLUS] = ACTIONS(1480),
    [anon_sym_DASH] = ACTIONS(1480),
    [anon_sym_TILDE] = ACTIONS(1478),
    [anon_sym_BANG] = ACTIONS(1478),
    [anon_sym_AT] = ACTIONS(1478),
    [aux_sym_clone_expression_token1] = ACTIONS(1480),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1480),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1480),
    [anon_sym_DASH_DASH] = ACTIONS(1478),
    [anon_sym_PLUS_PLUS] = ACTIONS(1478),
    [aux_sym__list_destructing_token1] = ACTIONS(1480),
    [anon_sym_LBRACK] = ACTIONS(1478),
    [anon_sym_self] = ACTIONS(1480),
    [anon_sym_parent] = ACTIONS(1480),
    [aux_sym__argument_name_token1] = ACTIONS(1480),
    [aux_sym__argument_name_token2] = ACTIONS(1480),
    [anon_sym_POUND_LBRACK] = ACTIONS(1478),
    [aux_sym_encapsed_string_token1] = ACTIONS(1478),
    [anon_sym_DQUOTE] = ACTIONS(1478),
    [aux_sym_string_token1] = ACTIONS(1478),
    [anon_sym_SQUOTE] = ACTIONS(1478),
    [anon_sym_LT_LT_LT] = ACTIONS(1478),
    [anon_sym_BQUOTE] = ACTIONS(1478),
    [anon_sym_DOLLAR] = ACTIONS(1478),
    [aux_sym_yield_expression_token1] = ACTIONS(1480),
    [aux_sym_include_expression_token1] = ACTIONS(1480),
    [aux_sym_include_once_expression_token1] = ACTIONS(1480),
    [aux_sym_require_expression_token1] = ACTIONS(1480),
    [aux_sym_require_once_expression_token1] = ACTIONS(1480),
    [sym_comment] = ACTIONS(3),
  },
  [553] = {
    [ts_builtin_sym_end] = ACTIONS(1482),
    [sym_name] = ACTIONS(1484),
    [anon_sym_QMARK_GT] = ACTIONS(1482),
    [anon_sym_SEMI] = ACTIONS(1482),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1484),
    [aux_sym_global_declaration_token1] = ACTIONS(1484),
    [aux_sym_namespace_definition_token1] = ACTIONS(1484),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1484),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1484),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1484),
    [anon_sym_BSLASH] = ACTIONS(1482),
    [anon_sym_LBRACE] = ACTIONS(1482),
    [anon_sym_RBRACE] = ACTIONS(1482),
    [aux_sym_trait_declaration_token1] = ACTIONS(1484),
    [aux_sym_interface_declaration_token1] = ACTIONS(1484),
    [aux_sym_enum_declaration_token1] = ACTIONS(1484),
    [aux_sym_enum_case_token1] = ACTIONS(1484),
    [aux_sym_class_declaration_token1] = ACTIONS(1484),
    [aux_sym_final_modifier_token1] = ACTIONS(1484),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1484),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1484),
    [sym_var_modifier] = ACTIONS(1484),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1484),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1484),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1484),
    [anon_sym_LPAREN2] = ACTIONS(1482),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1484),
    [aux_sym_cast_type_token1] = ACTIONS(1484),
    [aux_sym_echo_statement_token1] = ACTIONS(1484),
    [aux_sym_exit_statement_token1] = ACTIONS(1484),
    [anon_sym_unset] = ACTIONS(1484),
    [aux_sym_declare_statement_token1] = ACTIONS(1484),
    [aux_sym_declare_statement_token2] = ACTIONS(1484),
    [sym_float] = ACTIONS(1484),
    [aux_sym_try_statement_token1] = ACTIONS(1484),
    [aux_sym_goto_statement_token1] = ACTIONS(1484),
    [aux_sym_continue_statement_token1] = ACTIONS(1484),
    [aux_sym_break_statement_token1] = ACTIONS(1484),
    [sym_integer] = ACTIONS(1484),
    [aux_sym_return_statement_token1] = ACTIONS(1484),
    [aux_sym_throw_expression_token1] = ACTIONS(1484),
    [aux_sym_while_statement_token1] = ACTIONS(1484),
    [aux_sym_while_statement_token2] = ACTIONS(1484),
    [aux_sym_do_statement_token1] = ACTIONS(1484),
    [aux_sym_for_statement_token1] = ACTIONS(1484),
    [aux_sym_for_statement_token2] = ACTIONS(1484),
    [aux_sym_foreach_statement_token1] = ACTIONS(1484),
    [aux_sym_foreach_statement_token2] = ACTIONS(1484),
    [aux_sym_if_statement_token1] = ACTIONS(1484),
    [aux_sym_if_statement_token2] = ACTIONS(1484),
    [aux_sym_else_if_clause_token1] = ACTIONS(1484),
    [aux_sym_else_clause_token1] = ACTIONS(1484),
    [aux_sym_match_expression_token1] = ACTIONS(1484),
    [aux_sym_match_default_expression_token1] = ACTIONS(1484),
    [aux_sym_switch_statement_token1] = ACTIONS(1484),
    [aux_sym_switch_block_token1] = ACTIONS(1484),
    [anon_sym_PLUS] = ACTIONS(1484),
    [anon_sym_DASH] = ACTIONS(1484),
    [anon_sym_TILDE] = ACTIONS(1482),
    [anon_sym_BANG] = ACTIONS(1482),
    [anon_sym_AT] = ACTIONS(1482),
    [aux_sym_clone_expression_token1] = ACTIONS(1484),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1484),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1484),
    [anon_sym_DASH_DASH] = ACTIONS(1482),
    [anon_sym_PLUS_PLUS] = ACTIONS(1482),
    [aux_sym__list_destructing_token1] = ACTIONS(1484),
    [anon_sym_LBRACK] = ACTIONS(1482),
    [anon_sym_self] = ACTIONS(1484),
    [anon_sym_parent] = ACTIONS(1484),
    [aux_sym__argument_name_token1] = ACTIONS(1484),
    [aux_sym__argument_name_token2] = ACTIONS(1484),
    [anon_sym_POUND_LBRACK] = ACTIONS(1482),
    [aux_sym_encapsed_string_token1] = ACTIONS(1482),
    [anon_sym_DQUOTE] = ACTIONS(1482),
    [aux_sym_string_token1] = ACTIONS(1482),
    [anon_sym_SQUOTE] = ACTIONS(1482),
    [anon_sym_LT_LT_LT] = ACTIONS(1482),
    [anon_sym_BQUOTE] = ACTIONS(1482),
    [anon_sym_DOLLAR] = ACTIONS(1482),
    [aux_sym_yield_expression_token1] = ACTIONS(1484),
    [aux_sym_include_expression_token1] = ACTIONS(1484),
    [aux_sym_include_once_expression_token1] = ACTIONS(1484),
    [aux_sym_require_expression_token1] = ACTIONS(1484),
    [aux_sym_require_once_expression_token1] = ACTIONS(1484),
    [sym_comment] = ACTIONS(3),
  },
  [554] = {
    [ts_builtin_sym_end] = ACTIONS(1486),
    [sym_name] = ACTIONS(1488),
    [anon_sym_QMARK_GT] = ACTIONS(1486),
    [anon_sym_SEMI] = ACTIONS(1486),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1488),
    [aux_sym_global_declaration_token1] = ACTIONS(1488),
    [aux_sym_namespace_definition_token1] = ACTIONS(1488),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1488),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1488),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1488),
    [anon_sym_BSLASH] = ACTIONS(1486),
    [anon_sym_LBRACE] = ACTIONS(1486),
    [anon_sym_RBRACE] = ACTIONS(1486),
    [aux_sym_trait_declaration_token1] = ACTIONS(1488),
    [aux_sym_interface_declaration_token1] = ACTIONS(1488),
    [aux_sym_enum_declaration_token1] = ACTIONS(1488),
    [aux_sym_enum_case_token1] = ACTIONS(1488),
    [aux_sym_class_declaration_token1] = ACTIONS(1488),
    [aux_sym_final_modifier_token1] = ACTIONS(1488),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1488),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1488),
    [sym_var_modifier] = ACTIONS(1488),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1488),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1488),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1488),
    [anon_sym_LPAREN2] = ACTIONS(1486),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1488),
    [aux_sym_cast_type_token1] = ACTIONS(1488),
    [aux_sym_echo_statement_token1] = ACTIONS(1488),
    [aux_sym_exit_statement_token1] = ACTIONS(1488),
    [anon_sym_unset] = ACTIONS(1488),
    [aux_sym_declare_statement_token1] = ACTIONS(1488),
    [aux_sym_declare_statement_token2] = ACTIONS(1488),
    [sym_float] = ACTIONS(1488),
    [aux_sym_try_statement_token1] = ACTIONS(1488),
    [aux_sym_goto_statement_token1] = ACTIONS(1488),
    [aux_sym_continue_statement_token1] = ACTIONS(1488),
    [aux_sym_break_statement_token1] = ACTIONS(1488),
    [sym_integer] = ACTIONS(1488),
    [aux_sym_return_statement_token1] = ACTIONS(1488),
    [aux_sym_throw_expression_token1] = ACTIONS(1488),
    [aux_sym_while_statement_token1] = ACTIONS(1488),
    [aux_sym_while_statement_token2] = ACTIONS(1488),
    [aux_sym_do_statement_token1] = ACTIONS(1488),
    [aux_sym_for_statement_token1] = ACTIONS(1488),
    [aux_sym_for_statement_token2] = ACTIONS(1488),
    [aux_sym_foreach_statement_token1] = ACTIONS(1488),
    [aux_sym_foreach_statement_token2] = ACTIONS(1488),
    [aux_sym_if_statement_token1] = ACTIONS(1488),
    [aux_sym_if_statement_token2] = ACTIONS(1488),
    [aux_sym_else_if_clause_token1] = ACTIONS(1488),
    [aux_sym_else_clause_token1] = ACTIONS(1488),
    [aux_sym_match_expression_token1] = ACTIONS(1488),
    [aux_sym_match_default_expression_token1] = ACTIONS(1488),
    [aux_sym_switch_statement_token1] = ACTIONS(1488),
    [aux_sym_switch_block_token1] = ACTIONS(1488),
    [anon_sym_PLUS] = ACTIONS(1488),
    [anon_sym_DASH] = ACTIONS(1488),
    [anon_sym_TILDE] = ACTIONS(1486),
    [anon_sym_BANG] = ACTIONS(1486),
    [anon_sym_AT] = ACTIONS(1486),
    [aux_sym_clone_expression_token1] = ACTIONS(1488),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1488),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1488),
    [anon_sym_DASH_DASH] = ACTIONS(1486),
    [anon_sym_PLUS_PLUS] = ACTIONS(1486),
    [aux_sym__list_destructing_token1] = ACTIONS(1488),
    [anon_sym_LBRACK] = ACTIONS(1486),
    [anon_sym_self] = ACTIONS(1488),
    [anon_sym_parent] = ACTIONS(1488),
    [aux_sym__argument_name_token1] = ACTIONS(1488),
    [aux_sym__argument_name_token2] = ACTIONS(1488),
    [anon_sym_POUND_LBRACK] = ACTIONS(1486),
    [aux_sym_encapsed_string_token1] = ACTIONS(1486),
    [anon_sym_DQUOTE] = ACTIONS(1486),
    [aux_sym_string_token1] = ACTIONS(1486),
    [anon_sym_SQUOTE] = ACTIONS(1486),
    [anon_sym_LT_LT_LT] = ACTIONS(1486),
    [anon_sym_BQUOTE] = ACTIONS(1486),
    [anon_sym_DOLLAR] = ACTIONS(1486),
    [aux_sym_yield_expression_token1] = ACTIONS(1488),
    [aux_sym_include_expression_token1] = ACTIONS(1488),
    [aux_sym_include_once_expression_token1] = ACTIONS(1488),
    [aux_sym_require_expression_token1] = ACTIONS(1488),
    [aux_sym_require_once_expression_token1] = ACTIONS(1488),
    [sym_comment] = ACTIONS(3),
  },
  [555] = {
    [ts_builtin_sym_end] = ACTIONS(1490),
    [sym_name] = ACTIONS(1492),
    [anon_sym_QMARK_GT] = ACTIONS(1490),
    [anon_sym_SEMI] = ACTIONS(1490),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1492),
    [aux_sym_global_declaration_token1] = ACTIONS(1492),
    [aux_sym_namespace_definition_token1] = ACTIONS(1492),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1492),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1492),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1492),
    [anon_sym_BSLASH] = ACTIONS(1490),
    [anon_sym_LBRACE] = ACTIONS(1490),
    [anon_sym_RBRACE] = ACTIONS(1490),
    [aux_sym_trait_declaration_token1] = ACTIONS(1492),
    [aux_sym_interface_declaration_token1] = ACTIONS(1492),
    [aux_sym_enum_declaration_token1] = ACTIONS(1492),
    [aux_sym_enum_case_token1] = ACTIONS(1492),
    [aux_sym_class_declaration_token1] = ACTIONS(1492),
    [aux_sym_final_modifier_token1] = ACTIONS(1492),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1492),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1492),
    [sym_var_modifier] = ACTIONS(1492),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1492),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1492),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1492),
    [anon_sym_LPAREN2] = ACTIONS(1490),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1492),
    [aux_sym_cast_type_token1] = ACTIONS(1492),
    [aux_sym_echo_statement_token1] = ACTIONS(1492),
    [aux_sym_exit_statement_token1] = ACTIONS(1492),
    [anon_sym_unset] = ACTIONS(1492),
    [aux_sym_declare_statement_token1] = ACTIONS(1492),
    [aux_sym_declare_statement_token2] = ACTIONS(1492),
    [sym_float] = ACTIONS(1492),
    [aux_sym_try_statement_token1] = ACTIONS(1492),
    [aux_sym_goto_statement_token1] = ACTIONS(1492),
    [aux_sym_continue_statement_token1] = ACTIONS(1492),
    [aux_sym_break_statement_token1] = ACTIONS(1492),
    [sym_integer] = ACTIONS(1492),
    [aux_sym_return_statement_token1] = ACTIONS(1492),
    [aux_sym_throw_expression_token1] = ACTIONS(1492),
    [aux_sym_while_statement_token1] = ACTIONS(1492),
    [aux_sym_while_statement_token2] = ACTIONS(1492),
    [aux_sym_do_statement_token1] = ACTIONS(1492),
    [aux_sym_for_statement_token1] = ACTIONS(1492),
    [aux_sym_for_statement_token2] = ACTIONS(1492),
    [aux_sym_foreach_statement_token1] = ACTIONS(1492),
    [aux_sym_foreach_statement_token2] = ACTIONS(1492),
    [aux_sym_if_statement_token1] = ACTIONS(1492),
    [aux_sym_if_statement_token2] = ACTIONS(1492),
    [aux_sym_else_if_clause_token1] = ACTIONS(1492),
    [aux_sym_else_clause_token1] = ACTIONS(1492),
    [aux_sym_match_expression_token1] = ACTIONS(1492),
    [aux_sym_match_default_expression_token1] = ACTIONS(1492),
    [aux_sym_switch_statement_token1] = ACTIONS(1492),
    [aux_sym_switch_block_token1] = ACTIONS(1492),
    [anon_sym_PLUS] = ACTIONS(1492),
    [anon_sym_DASH] = ACTIONS(1492),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_BANG] = ACTIONS(1490),
    [anon_sym_AT] = ACTIONS(1490),
    [aux_sym_clone_expression_token1] = ACTIONS(1492),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1492),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1492),
    [anon_sym_DASH_DASH] = ACTIONS(1490),
    [anon_sym_PLUS_PLUS] = ACTIONS(1490),
    [aux_sym__list_destructing_token1] = ACTIONS(1492),
    [anon_sym_LBRACK] = ACTIONS(1490),
    [anon_sym_self] = ACTIONS(1492),
    [anon_sym_parent] = ACTIONS(1492),
    [aux_sym__argument_name_token1] = ACTIONS(1492),
    [aux_sym__argument_name_token2] = ACTIONS(1492),
    [anon_sym_POUND_LBRACK] = ACTIONS(1490),
    [aux_sym_encapsed_string_token1] = ACTIONS(1490),
    [anon_sym_DQUOTE] = ACTIONS(1490),
    [aux_sym_string_token1] = ACTIONS(1490),
    [anon_sym_SQUOTE] = ACTIONS(1490),
    [anon_sym_LT_LT_LT] = ACTIONS(1490),
    [anon_sym_BQUOTE] = ACTIONS(1490),
    [anon_sym_DOLLAR] = ACTIONS(1490),
    [aux_sym_yield_expression_token1] = ACTIONS(1492),
    [aux_sym_include_expression_token1] = ACTIONS(1492),
    [aux_sym_include_once_expression_token1] = ACTIONS(1492),
    [aux_sym_require_expression_token1] = ACTIONS(1492),
    [aux_sym_require_once_expression_token1] = ACTIONS(1492),
    [sym_comment] = ACTIONS(3),
  },
  [556] = {
    [ts_builtin_sym_end] = ACTIONS(1494),
    [sym_name] = ACTIONS(1496),
    [anon_sym_QMARK_GT] = ACTIONS(1494),
    [anon_sym_SEMI] = ACTIONS(1494),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1496),
    [aux_sym_global_declaration_token1] = ACTIONS(1496),
    [aux_sym_namespace_definition_token1] = ACTIONS(1496),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1496),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1496),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1496),
    [anon_sym_BSLASH] = ACTIONS(1494),
    [anon_sym_LBRACE] = ACTIONS(1494),
    [anon_sym_RBRACE] = ACTIONS(1494),
    [aux_sym_trait_declaration_token1] = ACTIONS(1496),
    [aux_sym_interface_declaration_token1] = ACTIONS(1496),
    [aux_sym_enum_declaration_token1] = ACTIONS(1496),
    [aux_sym_enum_case_token1] = ACTIONS(1496),
    [aux_sym_class_declaration_token1] = ACTIONS(1496),
    [aux_sym_final_modifier_token1] = ACTIONS(1496),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1496),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1496),
    [sym_var_modifier] = ACTIONS(1496),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1496),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1496),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1496),
    [anon_sym_LPAREN2] = ACTIONS(1494),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1496),
    [aux_sym_cast_type_token1] = ACTIONS(1496),
    [aux_sym_echo_statement_token1] = ACTIONS(1496),
    [aux_sym_exit_statement_token1] = ACTIONS(1496),
    [anon_sym_unset] = ACTIONS(1496),
    [aux_sym_declare_statement_token1] = ACTIONS(1496),
    [aux_sym_declare_statement_token2] = ACTIONS(1496),
    [sym_float] = ACTIONS(1496),
    [aux_sym_try_statement_token1] = ACTIONS(1496),
    [aux_sym_goto_statement_token1] = ACTIONS(1496),
    [aux_sym_continue_statement_token1] = ACTIONS(1496),
    [aux_sym_break_statement_token1] = ACTIONS(1496),
    [sym_integer] = ACTIONS(1496),
    [aux_sym_return_statement_token1] = ACTIONS(1496),
    [aux_sym_throw_expression_token1] = ACTIONS(1496),
    [aux_sym_while_statement_token1] = ACTIONS(1496),
    [aux_sym_while_statement_token2] = ACTIONS(1496),
    [aux_sym_do_statement_token1] = ACTIONS(1496),
    [aux_sym_for_statement_token1] = ACTIONS(1496),
    [aux_sym_for_statement_token2] = ACTIONS(1496),
    [aux_sym_foreach_statement_token1] = ACTIONS(1496),
    [aux_sym_foreach_statement_token2] = ACTIONS(1496),
    [aux_sym_if_statement_token1] = ACTIONS(1496),
    [aux_sym_if_statement_token2] = ACTIONS(1496),
    [aux_sym_else_if_clause_token1] = ACTIONS(1496),
    [aux_sym_else_clause_token1] = ACTIONS(1496),
    [aux_sym_match_expression_token1] = ACTIONS(1496),
    [aux_sym_match_default_expression_token1] = ACTIONS(1496),
    [aux_sym_switch_statement_token1] = ACTIONS(1496),
    [aux_sym_switch_block_token1] = ACTIONS(1496),
    [anon_sym_PLUS] = ACTIONS(1496),
    [anon_sym_DASH] = ACTIONS(1496),
    [anon_sym_TILDE] = ACTIONS(1494),
    [anon_sym_BANG] = ACTIONS(1494),
    [anon_sym_AT] = ACTIONS(1494),
    [aux_sym_clone_expression_token1] = ACTIONS(1496),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1496),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1496),
    [anon_sym_DASH_DASH] = ACTIONS(1494),
    [anon_sym_PLUS_PLUS] = ACTIONS(1494),
    [aux_sym__list_destructing_token1] = ACTIONS(1496),
    [anon_sym_LBRACK] = ACTIONS(1494),
    [anon_sym_self] = ACTIONS(1496),
    [anon_sym_parent] = ACTIONS(1496),
    [aux_sym__argument_name_token1] = ACTIONS(1496),
    [aux_sym__argument_name_token2] = ACTIONS(1496),
    [anon_sym_POUND_LBRACK] = ACTIONS(1494),
    [aux_sym_encapsed_string_token1] = ACTIONS(1494),
    [anon_sym_DQUOTE] = ACTIONS(1494),
    [aux_sym_string_token1] = ACTIONS(1494),
    [anon_sym_SQUOTE] = ACTIONS(1494),
    [anon_sym_LT_LT_LT] = ACTIONS(1494),
    [anon_sym_BQUOTE] = ACTIONS(1494),
    [anon_sym_DOLLAR] = ACTIONS(1494),
    [aux_sym_yield_expression_token1] = ACTIONS(1496),
    [aux_sym_include_expression_token1] = ACTIONS(1496),
    [aux_sym_include_once_expression_token1] = ACTIONS(1496),
    [aux_sym_require_expression_token1] = ACTIONS(1496),
    [aux_sym_require_once_expression_token1] = ACTIONS(1496),
    [sym_comment] = ACTIONS(3),
  },
  [557] = {
    [ts_builtin_sym_end] = ACTIONS(1498),
    [sym_name] = ACTIONS(1500),
    [anon_sym_QMARK_GT] = ACTIONS(1498),
    [anon_sym_SEMI] = ACTIONS(1498),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1500),
    [aux_sym_global_declaration_token1] = ACTIONS(1500),
    [aux_sym_namespace_definition_token1] = ACTIONS(1500),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1500),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1500),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1500),
    [anon_sym_BSLASH] = ACTIONS(1498),
    [anon_sym_LBRACE] = ACTIONS(1498),
    [anon_sym_RBRACE] = ACTIONS(1498),
    [aux_sym_trait_declaration_token1] = ACTIONS(1500),
    [aux_sym_interface_declaration_token1] = ACTIONS(1500),
    [aux_sym_enum_declaration_token1] = ACTIONS(1500),
    [aux_sym_enum_case_token1] = ACTIONS(1500),
    [aux_sym_class_declaration_token1] = ACTIONS(1500),
    [aux_sym_final_modifier_token1] = ACTIONS(1500),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1500),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1500),
    [sym_var_modifier] = ACTIONS(1500),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1500),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1500),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1500),
    [anon_sym_LPAREN2] = ACTIONS(1498),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1500),
    [aux_sym_cast_type_token1] = ACTIONS(1500),
    [aux_sym_echo_statement_token1] = ACTIONS(1500),
    [aux_sym_exit_statement_token1] = ACTIONS(1500),
    [anon_sym_unset] = ACTIONS(1500),
    [aux_sym_declare_statement_token1] = ACTIONS(1500),
    [aux_sym_declare_statement_token2] = ACTIONS(1500),
    [sym_float] = ACTIONS(1500),
    [aux_sym_try_statement_token1] = ACTIONS(1500),
    [aux_sym_goto_statement_token1] = ACTIONS(1500),
    [aux_sym_continue_statement_token1] = ACTIONS(1500),
    [aux_sym_break_statement_token1] = ACTIONS(1500),
    [sym_integer] = ACTIONS(1500),
    [aux_sym_return_statement_token1] = ACTIONS(1500),
    [aux_sym_throw_expression_token1] = ACTIONS(1500),
    [aux_sym_while_statement_token1] = ACTIONS(1500),
    [aux_sym_while_statement_token2] = ACTIONS(1500),
    [aux_sym_do_statement_token1] = ACTIONS(1500),
    [aux_sym_for_statement_token1] = ACTIONS(1500),
    [aux_sym_for_statement_token2] = ACTIONS(1500),
    [aux_sym_foreach_statement_token1] = ACTIONS(1500),
    [aux_sym_foreach_statement_token2] = ACTIONS(1500),
    [aux_sym_if_statement_token1] = ACTIONS(1500),
    [aux_sym_if_statement_token2] = ACTIONS(1500),
    [aux_sym_else_if_clause_token1] = ACTIONS(1500),
    [aux_sym_else_clause_token1] = ACTIONS(1500),
    [aux_sym_match_expression_token1] = ACTIONS(1500),
    [aux_sym_match_default_expression_token1] = ACTIONS(1500),
    [aux_sym_switch_statement_token1] = ACTIONS(1500),
    [aux_sym_switch_block_token1] = ACTIONS(1500),
    [anon_sym_PLUS] = ACTIONS(1500),
    [anon_sym_DASH] = ACTIONS(1500),
    [anon_sym_TILDE] = ACTIONS(1498),
    [anon_sym_BANG] = ACTIONS(1498),
    [anon_sym_AT] = ACTIONS(1498),
    [aux_sym_clone_expression_token1] = ACTIONS(1500),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1500),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1500),
    [anon_sym_DASH_DASH] = ACTIONS(1498),
    [anon_sym_PLUS_PLUS] = ACTIONS(1498),
    [aux_sym__list_destructing_token1] = ACTIONS(1500),
    [anon_sym_LBRACK] = ACTIONS(1498),
    [anon_sym_self] = ACTIONS(1500),
    [anon_sym_parent] = ACTIONS(1500),
    [aux_sym__argument_name_token1] = ACTIONS(1500),
    [aux_sym__argument_name_token2] = ACTIONS(1500),
    [anon_sym_POUND_LBRACK] = ACTIONS(1498),
    [aux_sym_encapsed_string_token1] = ACTIONS(1498),
    [anon_sym_DQUOTE] = ACTIONS(1498),
    [aux_sym_string_token1] = ACTIONS(1498),
    [anon_sym_SQUOTE] = ACTIONS(1498),
    [anon_sym_LT_LT_LT] = ACTIONS(1498),
    [anon_sym_BQUOTE] = ACTIONS(1498),
    [anon_sym_DOLLAR] = ACTIONS(1498),
    [aux_sym_yield_expression_token1] = ACTIONS(1500),
    [aux_sym_include_expression_token1] = ACTIONS(1500),
    [aux_sym_include_once_expression_token1] = ACTIONS(1500),
    [aux_sym_require_expression_token1] = ACTIONS(1500),
    [aux_sym_require_once_expression_token1] = ACTIONS(1500),
    [sym_comment] = ACTIONS(3),
  },
  [558] = {
    [ts_builtin_sym_end] = ACTIONS(1502),
    [sym_name] = ACTIONS(1504),
    [anon_sym_QMARK_GT] = ACTIONS(1502),
    [anon_sym_SEMI] = ACTIONS(1502),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1504),
    [aux_sym_global_declaration_token1] = ACTIONS(1504),
    [aux_sym_namespace_definition_token1] = ACTIONS(1504),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1504),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1504),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1504),
    [anon_sym_BSLASH] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1502),
    [anon_sym_RBRACE] = ACTIONS(1502),
    [aux_sym_trait_declaration_token1] = ACTIONS(1504),
    [aux_sym_interface_declaration_token1] = ACTIONS(1504),
    [aux_sym_enum_declaration_token1] = ACTIONS(1504),
    [aux_sym_enum_case_token1] = ACTIONS(1504),
    [aux_sym_class_declaration_token1] = ACTIONS(1504),
    [aux_sym_final_modifier_token1] = ACTIONS(1504),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1504),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1504),
    [sym_var_modifier] = ACTIONS(1504),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1504),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1504),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1504),
    [anon_sym_LPAREN2] = ACTIONS(1502),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1504),
    [aux_sym_cast_type_token1] = ACTIONS(1504),
    [aux_sym_echo_statement_token1] = ACTIONS(1504),
    [aux_sym_exit_statement_token1] = ACTIONS(1504),
    [anon_sym_unset] = ACTIONS(1504),
    [aux_sym_declare_statement_token1] = ACTIONS(1504),
    [aux_sym_declare_statement_token2] = ACTIONS(1504),
    [sym_float] = ACTIONS(1504),
    [aux_sym_try_statement_token1] = ACTIONS(1504),
    [aux_sym_goto_statement_token1] = ACTIONS(1504),
    [aux_sym_continue_statement_token1] = ACTIONS(1504),
    [aux_sym_break_statement_token1] = ACTIONS(1504),
    [sym_integer] = ACTIONS(1504),
    [aux_sym_return_statement_token1] = ACTIONS(1504),
    [aux_sym_throw_expression_token1] = ACTIONS(1504),
    [aux_sym_while_statement_token1] = ACTIONS(1504),
    [aux_sym_while_statement_token2] = ACTIONS(1504),
    [aux_sym_do_statement_token1] = ACTIONS(1504),
    [aux_sym_for_statement_token1] = ACTIONS(1504),
    [aux_sym_for_statement_token2] = ACTIONS(1504),
    [aux_sym_foreach_statement_token1] = ACTIONS(1504),
    [aux_sym_foreach_statement_token2] = ACTIONS(1504),
    [aux_sym_if_statement_token1] = ACTIONS(1504),
    [aux_sym_if_statement_token2] = ACTIONS(1504),
    [aux_sym_else_if_clause_token1] = ACTIONS(1504),
    [aux_sym_else_clause_token1] = ACTIONS(1504),
    [aux_sym_match_expression_token1] = ACTIONS(1504),
    [aux_sym_match_default_expression_token1] = ACTIONS(1504),
    [aux_sym_switch_statement_token1] = ACTIONS(1504),
    [aux_sym_switch_block_token1] = ACTIONS(1504),
    [anon_sym_PLUS] = ACTIONS(1504),
    [anon_sym_DASH] = ACTIONS(1504),
    [anon_sym_TILDE] = ACTIONS(1502),
    [anon_sym_BANG] = ACTIONS(1502),
    [anon_sym_AT] = ACTIONS(1502),
    [aux_sym_clone_expression_token1] = ACTIONS(1504),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1504),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1504),
    [anon_sym_DASH_DASH] = ACTIONS(1502),
    [anon_sym_PLUS_PLUS] = ACTIONS(1502),
    [aux_sym__list_destructing_token1] = ACTIONS(1504),
    [anon_sym_LBRACK] = ACTIONS(1502),
    [anon_sym_self] = ACTIONS(1504),
    [anon_sym_parent] = ACTIONS(1504),
    [aux_sym__argument_name_token1] = ACTIONS(1504),
    [aux_sym__argument_name_token2] = ACTIONS(1504),
    [anon_sym_POUND_LBRACK] = ACTIONS(1502),
    [aux_sym_encapsed_string_token1] = ACTIONS(1502),
    [anon_sym_DQUOTE] = ACTIONS(1502),
    [aux_sym_string_token1] = ACTIONS(1502),
    [anon_sym_SQUOTE] = ACTIONS(1502),
    [anon_sym_LT_LT_LT] = ACTIONS(1502),
    [anon_sym_BQUOTE] = ACTIONS(1502),
    [anon_sym_DOLLAR] = ACTIONS(1502),
    [aux_sym_yield_expression_token1] = ACTIONS(1504),
    [aux_sym_include_expression_token1] = ACTIONS(1504),
    [aux_sym_include_once_expression_token1] = ACTIONS(1504),
    [aux_sym_require_expression_token1] = ACTIONS(1504),
    [aux_sym_require_once_expression_token1] = ACTIONS(1504),
    [sym_comment] = ACTIONS(3),
  },
  [559] = {
    [ts_builtin_sym_end] = ACTIONS(1506),
    [sym_name] = ACTIONS(1508),
    [anon_sym_QMARK_GT] = ACTIONS(1506),
    [anon_sym_SEMI] = ACTIONS(1506),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1508),
    [aux_sym_global_declaration_token1] = ACTIONS(1508),
    [aux_sym_namespace_definition_token1] = ACTIONS(1508),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1508),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1508),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1508),
    [anon_sym_BSLASH] = ACTIONS(1506),
    [anon_sym_LBRACE] = ACTIONS(1506),
    [anon_sym_RBRACE] = ACTIONS(1506),
    [aux_sym_trait_declaration_token1] = ACTIONS(1508),
    [aux_sym_interface_declaration_token1] = ACTIONS(1508),
    [aux_sym_enum_declaration_token1] = ACTIONS(1508),
    [aux_sym_enum_case_token1] = ACTIONS(1508),
    [aux_sym_class_declaration_token1] = ACTIONS(1508),
    [aux_sym_final_modifier_token1] = ACTIONS(1508),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1508),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1508),
    [sym_var_modifier] = ACTIONS(1508),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1508),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1508),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1508),
    [anon_sym_LPAREN2] = ACTIONS(1506),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1508),
    [aux_sym_cast_type_token1] = ACTIONS(1508),
    [aux_sym_echo_statement_token1] = ACTIONS(1508),
    [aux_sym_exit_statement_token1] = ACTIONS(1508),
    [anon_sym_unset] = ACTIONS(1508),
    [aux_sym_declare_statement_token1] = ACTIONS(1508),
    [aux_sym_declare_statement_token2] = ACTIONS(1508),
    [sym_float] = ACTIONS(1508),
    [aux_sym_try_statement_token1] = ACTIONS(1508),
    [aux_sym_goto_statement_token1] = ACTIONS(1508),
    [aux_sym_continue_statement_token1] = ACTIONS(1508),
    [aux_sym_break_statement_token1] = ACTIONS(1508),
    [sym_integer] = ACTIONS(1508),
    [aux_sym_return_statement_token1] = ACTIONS(1508),
    [aux_sym_throw_expression_token1] = ACTIONS(1508),
    [aux_sym_while_statement_token1] = ACTIONS(1508),
    [aux_sym_while_statement_token2] = ACTIONS(1508),
    [aux_sym_do_statement_token1] = ACTIONS(1508),
    [aux_sym_for_statement_token1] = ACTIONS(1508),
    [aux_sym_for_statement_token2] = ACTIONS(1508),
    [aux_sym_foreach_statement_token1] = ACTIONS(1508),
    [aux_sym_foreach_statement_token2] = ACTIONS(1508),
    [aux_sym_if_statement_token1] = ACTIONS(1508),
    [aux_sym_if_statement_token2] = ACTIONS(1508),
    [aux_sym_else_if_clause_token1] = ACTIONS(1508),
    [aux_sym_else_clause_token1] = ACTIONS(1508),
    [aux_sym_match_expression_token1] = ACTIONS(1508),
    [aux_sym_match_default_expression_token1] = ACTIONS(1508),
    [aux_sym_switch_statement_token1] = ACTIONS(1508),
    [aux_sym_switch_block_token1] = ACTIONS(1508),
    [anon_sym_PLUS] = ACTIONS(1508),
    [anon_sym_DASH] = ACTIONS(1508),
    [anon_sym_TILDE] = ACTIONS(1506),
    [anon_sym_BANG] = ACTIONS(1506),
    [anon_sym_AT] = ACTIONS(1506),
    [aux_sym_clone_expression_token1] = ACTIONS(1508),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1508),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1508),
    [anon_sym_DASH_DASH] = ACTIONS(1506),
    [anon_sym_PLUS_PLUS] = ACTIONS(1506),
    [aux_sym__list_destructing_token1] = ACTIONS(1508),
    [anon_sym_LBRACK] = ACTIONS(1506),
    [anon_sym_self] = ACTIONS(1508),
    [anon_sym_parent] = ACTIONS(1508),
    [aux_sym__argument_name_token1] = ACTIONS(1508),
    [aux_sym__argument_name_token2] = ACTIONS(1508),
    [anon_sym_POUND_LBRACK] = ACTIONS(1506),
    [aux_sym_encapsed_string_token1] = ACTIONS(1506),
    [anon_sym_DQUOTE] = ACTIONS(1506),
    [aux_sym_string_token1] = ACTIONS(1506),
    [anon_sym_SQUOTE] = ACTIONS(1506),
    [anon_sym_LT_LT_LT] = ACTIONS(1506),
    [anon_sym_BQUOTE] = ACTIONS(1506),
    [anon_sym_DOLLAR] = ACTIONS(1506),
    [aux_sym_yield_expression_token1] = ACTIONS(1508),
    [aux_sym_include_expression_token1] = ACTIONS(1508),
    [aux_sym_include_once_expression_token1] = ACTIONS(1508),
    [aux_sym_require_expression_token1] = ACTIONS(1508),
    [aux_sym_require_once_expression_token1] = ACTIONS(1508),
    [sym_comment] = ACTIONS(3),
  },
  [560] = {
    [ts_builtin_sym_end] = ACTIONS(1510),
    [sym_name] = ACTIONS(1512),
    [anon_sym_QMARK_GT] = ACTIONS(1510),
    [anon_sym_SEMI] = ACTIONS(1510),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1512),
    [aux_sym_global_declaration_token1] = ACTIONS(1512),
    [aux_sym_namespace_definition_token1] = ACTIONS(1512),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1512),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1512),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1512),
    [anon_sym_BSLASH] = ACTIONS(1510),
    [anon_sym_LBRACE] = ACTIONS(1510),
    [anon_sym_RBRACE] = ACTIONS(1510),
    [aux_sym_trait_declaration_token1] = ACTIONS(1512),
    [aux_sym_interface_declaration_token1] = ACTIONS(1512),
    [aux_sym_enum_declaration_token1] = ACTIONS(1512),
    [aux_sym_enum_case_token1] = ACTIONS(1512),
    [aux_sym_class_declaration_token1] = ACTIONS(1512),
    [aux_sym_final_modifier_token1] = ACTIONS(1512),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1512),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1512),
    [sym_var_modifier] = ACTIONS(1512),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1512),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1512),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1512),
    [anon_sym_LPAREN2] = ACTIONS(1510),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1512),
    [aux_sym_cast_type_token1] = ACTIONS(1512),
    [aux_sym_echo_statement_token1] = ACTIONS(1512),
    [aux_sym_exit_statement_token1] = ACTIONS(1512),
    [anon_sym_unset] = ACTIONS(1512),
    [aux_sym_declare_statement_token1] = ACTIONS(1512),
    [aux_sym_declare_statement_token2] = ACTIONS(1512),
    [sym_float] = ACTIONS(1512),
    [aux_sym_try_statement_token1] = ACTIONS(1512),
    [aux_sym_goto_statement_token1] = ACTIONS(1512),
    [aux_sym_continue_statement_token1] = ACTIONS(1512),
    [aux_sym_break_statement_token1] = ACTIONS(1512),
    [sym_integer] = ACTIONS(1512),
    [aux_sym_return_statement_token1] = ACTIONS(1512),
    [aux_sym_throw_expression_token1] = ACTIONS(1512),
    [aux_sym_while_statement_token1] = ACTIONS(1512),
    [aux_sym_while_statement_token2] = ACTIONS(1512),
    [aux_sym_do_statement_token1] = ACTIONS(1512),
    [aux_sym_for_statement_token1] = ACTIONS(1512),
    [aux_sym_for_statement_token2] = ACTIONS(1512),
    [aux_sym_foreach_statement_token1] = ACTIONS(1512),
    [aux_sym_foreach_statement_token2] = ACTIONS(1512),
    [aux_sym_if_statement_token1] = ACTIONS(1512),
    [aux_sym_if_statement_token2] = ACTIONS(1512),
    [aux_sym_else_if_clause_token1] = ACTIONS(1512),
    [aux_sym_else_clause_token1] = ACTIONS(1512),
    [aux_sym_match_expression_token1] = ACTIONS(1512),
    [aux_sym_match_default_expression_token1] = ACTIONS(1512),
    [aux_sym_switch_statement_token1] = ACTIONS(1512),
    [aux_sym_switch_block_token1] = ACTIONS(1512),
    [anon_sym_PLUS] = ACTIONS(1512),
    [anon_sym_DASH] = ACTIONS(1512),
    [anon_sym_TILDE] = ACTIONS(1510),
    [anon_sym_BANG] = ACTIONS(1510),
    [anon_sym_AT] = ACTIONS(1510),
    [aux_sym_clone_expression_token1] = ACTIONS(1512),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1512),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1512),
    [anon_sym_DASH_DASH] = ACTIONS(1510),
    [anon_sym_PLUS_PLUS] = ACTIONS(1510),
    [aux_sym__list_destructing_token1] = ACTIONS(1512),
    [anon_sym_LBRACK] = ACTIONS(1510),
    [anon_sym_self] = ACTIONS(1512),
    [anon_sym_parent] = ACTIONS(1512),
    [aux_sym__argument_name_token1] = ACTIONS(1512),
    [aux_sym__argument_name_token2] = ACTIONS(1512),
    [anon_sym_POUND_LBRACK] = ACTIONS(1510),
    [aux_sym_encapsed_string_token1] = ACTIONS(1510),
    [anon_sym_DQUOTE] = ACTIONS(1510),
    [aux_sym_string_token1] = ACTIONS(1510),
    [anon_sym_SQUOTE] = ACTIONS(1510),
    [anon_sym_LT_LT_LT] = ACTIONS(1510),
    [anon_sym_BQUOTE] = ACTIONS(1510),
    [anon_sym_DOLLAR] = ACTIONS(1510),
    [aux_sym_yield_expression_token1] = ACTIONS(1512),
    [aux_sym_include_expression_token1] = ACTIONS(1512),
    [aux_sym_include_once_expression_token1] = ACTIONS(1512),
    [aux_sym_require_expression_token1] = ACTIONS(1512),
    [aux_sym_require_once_expression_token1] = ACTIONS(1512),
    [sym_comment] = ACTIONS(3),
  },
  [561] = {
    [ts_builtin_sym_end] = ACTIONS(1514),
    [sym_name] = ACTIONS(1516),
    [anon_sym_QMARK_GT] = ACTIONS(1514),
    [anon_sym_SEMI] = ACTIONS(1514),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1516),
    [aux_sym_global_declaration_token1] = ACTIONS(1516),
    [aux_sym_namespace_definition_token1] = ACTIONS(1516),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1516),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1516),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1516),
    [anon_sym_BSLASH] = ACTIONS(1514),
    [anon_sym_LBRACE] = ACTIONS(1514),
    [anon_sym_RBRACE] = ACTIONS(1514),
    [aux_sym_trait_declaration_token1] = ACTIONS(1516),
    [aux_sym_interface_declaration_token1] = ACTIONS(1516),
    [aux_sym_enum_declaration_token1] = ACTIONS(1516),
    [aux_sym_enum_case_token1] = ACTIONS(1516),
    [aux_sym_class_declaration_token1] = ACTIONS(1516),
    [aux_sym_final_modifier_token1] = ACTIONS(1516),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1516),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1516),
    [sym_var_modifier] = ACTIONS(1516),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1516),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1516),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1516),
    [anon_sym_LPAREN2] = ACTIONS(1514),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1516),
    [aux_sym_cast_type_token1] = ACTIONS(1516),
    [aux_sym_echo_statement_token1] = ACTIONS(1516),
    [aux_sym_exit_statement_token1] = ACTIONS(1516),
    [anon_sym_unset] = ACTIONS(1516),
    [aux_sym_declare_statement_token1] = ACTIONS(1516),
    [aux_sym_declare_statement_token2] = ACTIONS(1516),
    [sym_float] = ACTIONS(1516),
    [aux_sym_try_statement_token1] = ACTIONS(1516),
    [aux_sym_goto_statement_token1] = ACTIONS(1516),
    [aux_sym_continue_statement_token1] = ACTIONS(1516),
    [aux_sym_break_statement_token1] = ACTIONS(1516),
    [sym_integer] = ACTIONS(1516),
    [aux_sym_return_statement_token1] = ACTIONS(1516),
    [aux_sym_throw_expression_token1] = ACTIONS(1516),
    [aux_sym_while_statement_token1] = ACTIONS(1516),
    [aux_sym_while_statement_token2] = ACTIONS(1516),
    [aux_sym_do_statement_token1] = ACTIONS(1516),
    [aux_sym_for_statement_token1] = ACTIONS(1516),
    [aux_sym_for_statement_token2] = ACTIONS(1516),
    [aux_sym_foreach_statement_token1] = ACTIONS(1516),
    [aux_sym_foreach_statement_token2] = ACTIONS(1516),
    [aux_sym_if_statement_token1] = ACTIONS(1516),
    [aux_sym_if_statement_token2] = ACTIONS(1516),
    [aux_sym_else_if_clause_token1] = ACTIONS(1516),
    [aux_sym_else_clause_token1] = ACTIONS(1516),
    [aux_sym_match_expression_token1] = ACTIONS(1516),
    [aux_sym_match_default_expression_token1] = ACTIONS(1516),
    [aux_sym_switch_statement_token1] = ACTIONS(1516),
    [aux_sym_switch_block_token1] = ACTIONS(1516),
    [anon_sym_PLUS] = ACTIONS(1516),
    [anon_sym_DASH] = ACTIONS(1516),
    [anon_sym_TILDE] = ACTIONS(1514),
    [anon_sym_BANG] = ACTIONS(1514),
    [anon_sym_AT] = ACTIONS(1514),
    [aux_sym_clone_expression_token1] = ACTIONS(1516),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1516),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1516),
    [anon_sym_DASH_DASH] = ACTIONS(1514),
    [anon_sym_PLUS_PLUS] = ACTIONS(1514),
    [aux_sym__list_destructing_token1] = ACTIONS(1516),
    [anon_sym_LBRACK] = ACTIONS(1514),
    [anon_sym_self] = ACTIONS(1516),
    [anon_sym_parent] = ACTIONS(1516),
    [aux_sym__argument_name_token1] = ACTIONS(1516),
    [aux_sym__argument_name_token2] = ACTIONS(1516),
    [anon_sym_POUND_LBRACK] = ACTIONS(1514),
    [aux_sym_encapsed_string_token1] = ACTIONS(1514),
    [anon_sym_DQUOTE] = ACTIONS(1514),
    [aux_sym_string_token1] = ACTIONS(1514),
    [anon_sym_SQUOTE] = ACTIONS(1514),
    [anon_sym_LT_LT_LT] = ACTIONS(1514),
    [anon_sym_BQUOTE] = ACTIONS(1514),
    [anon_sym_DOLLAR] = ACTIONS(1514),
    [aux_sym_yield_expression_token1] = ACTIONS(1516),
    [aux_sym_include_expression_token1] = ACTIONS(1516),
    [aux_sym_include_once_expression_token1] = ACTIONS(1516),
    [aux_sym_require_expression_token1] = ACTIONS(1516),
    [aux_sym_require_once_expression_token1] = ACTIONS(1516),
    [sym_comment] = ACTIONS(3),
  },
  [562] = {
    [ts_builtin_sym_end] = ACTIONS(1518),
    [sym_name] = ACTIONS(1520),
    [anon_sym_QMARK_GT] = ACTIONS(1518),
    [anon_sym_SEMI] = ACTIONS(1518),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1520),
    [aux_sym_global_declaration_token1] = ACTIONS(1520),
    [aux_sym_namespace_definition_token1] = ACTIONS(1520),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1520),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1520),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1520),
    [anon_sym_BSLASH] = ACTIONS(1518),
    [anon_sym_LBRACE] = ACTIONS(1518),
    [anon_sym_RBRACE] = ACTIONS(1518),
    [aux_sym_trait_declaration_token1] = ACTIONS(1520),
    [aux_sym_interface_declaration_token1] = ACTIONS(1520),
    [aux_sym_enum_declaration_token1] = ACTIONS(1520),
    [aux_sym_enum_case_token1] = ACTIONS(1520),
    [aux_sym_class_declaration_token1] = ACTIONS(1520),
    [aux_sym_final_modifier_token1] = ACTIONS(1520),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1520),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1520),
    [sym_var_modifier] = ACTIONS(1520),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1520),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1520),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1520),
    [anon_sym_LPAREN2] = ACTIONS(1518),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1520),
    [aux_sym_cast_type_token1] = ACTIONS(1520),
    [aux_sym_echo_statement_token1] = ACTIONS(1520),
    [aux_sym_exit_statement_token1] = ACTIONS(1520),
    [anon_sym_unset] = ACTIONS(1520),
    [aux_sym_declare_statement_token1] = ACTIONS(1520),
    [aux_sym_declare_statement_token2] = ACTIONS(1520),
    [sym_float] = ACTIONS(1520),
    [aux_sym_try_statement_token1] = ACTIONS(1520),
    [aux_sym_goto_statement_token1] = ACTIONS(1520),
    [aux_sym_continue_statement_token1] = ACTIONS(1520),
    [aux_sym_break_statement_token1] = ACTIONS(1520),
    [sym_integer] = ACTIONS(1520),
    [aux_sym_return_statement_token1] = ACTIONS(1520),
    [aux_sym_throw_expression_token1] = ACTIONS(1520),
    [aux_sym_while_statement_token1] = ACTIONS(1520),
    [aux_sym_while_statement_token2] = ACTIONS(1520),
    [aux_sym_do_statement_token1] = ACTIONS(1520),
    [aux_sym_for_statement_token1] = ACTIONS(1520),
    [aux_sym_for_statement_token2] = ACTIONS(1520),
    [aux_sym_foreach_statement_token1] = ACTIONS(1520),
    [aux_sym_foreach_statement_token2] = ACTIONS(1520),
    [aux_sym_if_statement_token1] = ACTIONS(1520),
    [aux_sym_if_statement_token2] = ACTIONS(1520),
    [aux_sym_else_if_clause_token1] = ACTIONS(1520),
    [aux_sym_else_clause_token1] = ACTIONS(1520),
    [aux_sym_match_expression_token1] = ACTIONS(1520),
    [aux_sym_match_default_expression_token1] = ACTIONS(1520),
    [aux_sym_switch_statement_token1] = ACTIONS(1520),
    [aux_sym_switch_block_token1] = ACTIONS(1520),
    [anon_sym_PLUS] = ACTIONS(1520),
    [anon_sym_DASH] = ACTIONS(1520),
    [anon_sym_TILDE] = ACTIONS(1518),
    [anon_sym_BANG] = ACTIONS(1518),
    [anon_sym_AT] = ACTIONS(1518),
    [aux_sym_clone_expression_token1] = ACTIONS(1520),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1520),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1520),
    [anon_sym_DASH_DASH] = ACTIONS(1518),
    [anon_sym_PLUS_PLUS] = ACTIONS(1518),
    [aux_sym__list_destructing_token1] = ACTIONS(1520),
    [anon_sym_LBRACK] = ACTIONS(1518),
    [anon_sym_self] = ACTIONS(1520),
    [anon_sym_parent] = ACTIONS(1520),
    [aux_sym__argument_name_token1] = ACTIONS(1520),
    [aux_sym__argument_name_token2] = ACTIONS(1520),
    [anon_sym_POUND_LBRACK] = ACTIONS(1518),
    [aux_sym_encapsed_string_token1] = ACTIONS(1518),
    [anon_sym_DQUOTE] = ACTIONS(1518),
    [aux_sym_string_token1] = ACTIONS(1518),
    [anon_sym_SQUOTE] = ACTIONS(1518),
    [anon_sym_LT_LT_LT] = ACTIONS(1518),
    [anon_sym_BQUOTE] = ACTIONS(1518),
    [anon_sym_DOLLAR] = ACTIONS(1518),
    [aux_sym_yield_expression_token1] = ACTIONS(1520),
    [aux_sym_include_expression_token1] = ACTIONS(1520),
    [aux_sym_include_once_expression_token1] = ACTIONS(1520),
    [aux_sym_require_expression_token1] = ACTIONS(1520),
    [aux_sym_require_once_expression_token1] = ACTIONS(1520),
    [sym_comment] = ACTIONS(3),
  },
  [563] = {
    [ts_builtin_sym_end] = ACTIONS(1522),
    [sym_name] = ACTIONS(1524),
    [anon_sym_QMARK_GT] = ACTIONS(1522),
    [anon_sym_SEMI] = ACTIONS(1522),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1524),
    [aux_sym_global_declaration_token1] = ACTIONS(1524),
    [aux_sym_namespace_definition_token1] = ACTIONS(1524),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1524),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1524),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1524),
    [anon_sym_BSLASH] = ACTIONS(1522),
    [anon_sym_LBRACE] = ACTIONS(1522),
    [anon_sym_RBRACE] = ACTIONS(1522),
    [aux_sym_trait_declaration_token1] = ACTIONS(1524),
    [aux_sym_interface_declaration_token1] = ACTIONS(1524),
    [aux_sym_enum_declaration_token1] = ACTIONS(1524),
    [aux_sym_enum_case_token1] = ACTIONS(1524),
    [aux_sym_class_declaration_token1] = ACTIONS(1524),
    [aux_sym_final_modifier_token1] = ACTIONS(1524),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1524),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1524),
    [sym_var_modifier] = ACTIONS(1524),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1524),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1524),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1524),
    [anon_sym_LPAREN2] = ACTIONS(1522),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1524),
    [aux_sym_cast_type_token1] = ACTIONS(1524),
    [aux_sym_echo_statement_token1] = ACTIONS(1524),
    [aux_sym_exit_statement_token1] = ACTIONS(1524),
    [anon_sym_unset] = ACTIONS(1524),
    [aux_sym_declare_statement_token1] = ACTIONS(1524),
    [aux_sym_declare_statement_token2] = ACTIONS(1524),
    [sym_float] = ACTIONS(1524),
    [aux_sym_try_statement_token1] = ACTIONS(1524),
    [aux_sym_goto_statement_token1] = ACTIONS(1524),
    [aux_sym_continue_statement_token1] = ACTIONS(1524),
    [aux_sym_break_statement_token1] = ACTIONS(1524),
    [sym_integer] = ACTIONS(1524),
    [aux_sym_return_statement_token1] = ACTIONS(1524),
    [aux_sym_throw_expression_token1] = ACTIONS(1524),
    [aux_sym_while_statement_token1] = ACTIONS(1524),
    [aux_sym_while_statement_token2] = ACTIONS(1524),
    [aux_sym_do_statement_token1] = ACTIONS(1524),
    [aux_sym_for_statement_token1] = ACTIONS(1524),
    [aux_sym_for_statement_token2] = ACTIONS(1524),
    [aux_sym_foreach_statement_token1] = ACTIONS(1524),
    [aux_sym_foreach_statement_token2] = ACTIONS(1524),
    [aux_sym_if_statement_token1] = ACTIONS(1524),
    [aux_sym_if_statement_token2] = ACTIONS(1524),
    [aux_sym_else_if_clause_token1] = ACTIONS(1524),
    [aux_sym_else_clause_token1] = ACTIONS(1524),
    [aux_sym_match_expression_token1] = ACTIONS(1524),
    [aux_sym_match_default_expression_token1] = ACTIONS(1524),
    [aux_sym_switch_statement_token1] = ACTIONS(1524),
    [aux_sym_switch_block_token1] = ACTIONS(1524),
    [anon_sym_PLUS] = ACTIONS(1524),
    [anon_sym_DASH] = ACTIONS(1524),
    [anon_sym_TILDE] = ACTIONS(1522),
    [anon_sym_BANG] = ACTIONS(1522),
    [anon_sym_AT] = ACTIONS(1522),
    [aux_sym_clone_expression_token1] = ACTIONS(1524),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1524),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1524),
    [anon_sym_DASH_DASH] = ACTIONS(1522),
    [anon_sym_PLUS_PLUS] = ACTIONS(1522),
    [aux_sym__list_destructing_token1] = ACTIONS(1524),
    [anon_sym_LBRACK] = ACTIONS(1522),
    [anon_sym_self] = ACTIONS(1524),
    [anon_sym_parent] = ACTIONS(1524),
    [aux_sym__argument_name_token1] = ACTIONS(1524),
    [aux_sym__argument_name_token2] = ACTIONS(1524),
    [anon_sym_POUND_LBRACK] = ACTIONS(1522),
    [aux_sym_encapsed_string_token1] = ACTIONS(1522),
    [anon_sym_DQUOTE] = ACTIONS(1522),
    [aux_sym_string_token1] = ACTIONS(1522),
    [anon_sym_SQUOTE] = ACTIONS(1522),
    [anon_sym_LT_LT_LT] = ACTIONS(1522),
    [anon_sym_BQUOTE] = ACTIONS(1522),
    [anon_sym_DOLLAR] = ACTIONS(1522),
    [aux_sym_yield_expression_token1] = ACTIONS(1524),
    [aux_sym_include_expression_token1] = ACTIONS(1524),
    [aux_sym_include_once_expression_token1] = ACTIONS(1524),
    [aux_sym_require_expression_token1] = ACTIONS(1524),
    [aux_sym_require_once_expression_token1] = ACTIONS(1524),
    [sym_comment] = ACTIONS(3),
  },
  [564] = {
    [ts_builtin_sym_end] = ACTIONS(1526),
    [sym_name] = ACTIONS(1528),
    [anon_sym_QMARK_GT] = ACTIONS(1526),
    [anon_sym_SEMI] = ACTIONS(1526),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1528),
    [aux_sym_global_declaration_token1] = ACTIONS(1528),
    [aux_sym_namespace_definition_token1] = ACTIONS(1528),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1528),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1528),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1528),
    [anon_sym_BSLASH] = ACTIONS(1526),
    [anon_sym_LBRACE] = ACTIONS(1526),
    [anon_sym_RBRACE] = ACTIONS(1526),
    [aux_sym_trait_declaration_token1] = ACTIONS(1528),
    [aux_sym_interface_declaration_token1] = ACTIONS(1528),
    [aux_sym_enum_declaration_token1] = ACTIONS(1528),
    [aux_sym_enum_case_token1] = ACTIONS(1528),
    [aux_sym_class_declaration_token1] = ACTIONS(1528),
    [aux_sym_final_modifier_token1] = ACTIONS(1528),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1528),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1528),
    [sym_var_modifier] = ACTIONS(1528),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1528),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1528),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1528),
    [anon_sym_LPAREN2] = ACTIONS(1526),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1528),
    [aux_sym_cast_type_token1] = ACTIONS(1528),
    [aux_sym_echo_statement_token1] = ACTIONS(1528),
    [aux_sym_exit_statement_token1] = ACTIONS(1528),
    [anon_sym_unset] = ACTIONS(1528),
    [aux_sym_declare_statement_token1] = ACTIONS(1528),
    [aux_sym_declare_statement_token2] = ACTIONS(1528),
    [sym_float] = ACTIONS(1528),
    [aux_sym_try_statement_token1] = ACTIONS(1528),
    [aux_sym_goto_statement_token1] = ACTIONS(1528),
    [aux_sym_continue_statement_token1] = ACTIONS(1528),
    [aux_sym_break_statement_token1] = ACTIONS(1528),
    [sym_integer] = ACTIONS(1528),
    [aux_sym_return_statement_token1] = ACTIONS(1528),
    [aux_sym_throw_expression_token1] = ACTIONS(1528),
    [aux_sym_while_statement_token1] = ACTIONS(1528),
    [aux_sym_while_statement_token2] = ACTIONS(1528),
    [aux_sym_do_statement_token1] = ACTIONS(1528),
    [aux_sym_for_statement_token1] = ACTIONS(1528),
    [aux_sym_for_statement_token2] = ACTIONS(1528),
    [aux_sym_foreach_statement_token1] = ACTIONS(1528),
    [aux_sym_foreach_statement_token2] = ACTIONS(1528),
    [aux_sym_if_statement_token1] = ACTIONS(1528),
    [aux_sym_if_statement_token2] = ACTIONS(1528),
    [aux_sym_else_if_clause_token1] = ACTIONS(1528),
    [aux_sym_else_clause_token1] = ACTIONS(1528),
    [aux_sym_match_expression_token1] = ACTIONS(1528),
    [aux_sym_match_default_expression_token1] = ACTIONS(1528),
    [aux_sym_switch_statement_token1] = ACTIONS(1528),
    [aux_sym_switch_block_token1] = ACTIONS(1528),
    [anon_sym_PLUS] = ACTIONS(1528),
    [anon_sym_DASH] = ACTIONS(1528),
    [anon_sym_TILDE] = ACTIONS(1526),
    [anon_sym_BANG] = ACTIONS(1526),
    [anon_sym_AT] = ACTIONS(1526),
    [aux_sym_clone_expression_token1] = ACTIONS(1528),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1528),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1528),
    [anon_sym_DASH_DASH] = ACTIONS(1526),
    [anon_sym_PLUS_PLUS] = ACTIONS(1526),
    [aux_sym__list_destructing_token1] = ACTIONS(1528),
    [anon_sym_LBRACK] = ACTIONS(1526),
    [anon_sym_self] = ACTIONS(1528),
    [anon_sym_parent] = ACTIONS(1528),
    [aux_sym__argument_name_token1] = ACTIONS(1528),
    [aux_sym__argument_name_token2] = ACTIONS(1528),
    [anon_sym_POUND_LBRACK] = ACTIONS(1526),
    [aux_sym_encapsed_string_token1] = ACTIONS(1526),
    [anon_sym_DQUOTE] = ACTIONS(1526),
    [aux_sym_string_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1526),
    [anon_sym_LT_LT_LT] = ACTIONS(1526),
    [anon_sym_BQUOTE] = ACTIONS(1526),
    [anon_sym_DOLLAR] = ACTIONS(1526),
    [aux_sym_yield_expression_token1] = ACTIONS(1528),
    [aux_sym_include_expression_token1] = ACTIONS(1528),
    [aux_sym_include_once_expression_token1] = ACTIONS(1528),
    [aux_sym_require_expression_token1] = ACTIONS(1528),
    [aux_sym_require_once_expression_token1] = ACTIONS(1528),
    [sym_comment] = ACTIONS(3),
  },
  [565] = {
    [ts_builtin_sym_end] = ACTIONS(1530),
    [sym_name] = ACTIONS(1532),
    [anon_sym_QMARK_GT] = ACTIONS(1530),
    [anon_sym_SEMI] = ACTIONS(1530),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1532),
    [aux_sym_global_declaration_token1] = ACTIONS(1532),
    [aux_sym_namespace_definition_token1] = ACTIONS(1532),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1532),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1532),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1532),
    [anon_sym_BSLASH] = ACTIONS(1530),
    [anon_sym_LBRACE] = ACTIONS(1530),
    [anon_sym_RBRACE] = ACTIONS(1530),
    [aux_sym_trait_declaration_token1] = ACTIONS(1532),
    [aux_sym_interface_declaration_token1] = ACTIONS(1532),
    [aux_sym_enum_declaration_token1] = ACTIONS(1532),
    [aux_sym_enum_case_token1] = ACTIONS(1532),
    [aux_sym_class_declaration_token1] = ACTIONS(1532),
    [aux_sym_final_modifier_token1] = ACTIONS(1532),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1532),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1532),
    [sym_var_modifier] = ACTIONS(1532),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1532),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1532),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1532),
    [anon_sym_LPAREN2] = ACTIONS(1530),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1532),
    [aux_sym_cast_type_token1] = ACTIONS(1532),
    [aux_sym_echo_statement_token1] = ACTIONS(1532),
    [aux_sym_exit_statement_token1] = ACTIONS(1532),
    [anon_sym_unset] = ACTIONS(1532),
    [aux_sym_declare_statement_token1] = ACTIONS(1532),
    [aux_sym_declare_statement_token2] = ACTIONS(1532),
    [sym_float] = ACTIONS(1532),
    [aux_sym_try_statement_token1] = ACTIONS(1532),
    [aux_sym_goto_statement_token1] = ACTIONS(1532),
    [aux_sym_continue_statement_token1] = ACTIONS(1532),
    [aux_sym_break_statement_token1] = ACTIONS(1532),
    [sym_integer] = ACTIONS(1532),
    [aux_sym_return_statement_token1] = ACTIONS(1532),
    [aux_sym_throw_expression_token1] = ACTIONS(1532),
    [aux_sym_while_statement_token1] = ACTIONS(1532),
    [aux_sym_while_statement_token2] = ACTIONS(1532),
    [aux_sym_do_statement_token1] = ACTIONS(1532),
    [aux_sym_for_statement_token1] = ACTIONS(1532),
    [aux_sym_for_statement_token2] = ACTIONS(1532),
    [aux_sym_foreach_statement_token1] = ACTIONS(1532),
    [aux_sym_foreach_statement_token2] = ACTIONS(1532),
    [aux_sym_if_statement_token1] = ACTIONS(1532),
    [aux_sym_if_statement_token2] = ACTIONS(1532),
    [aux_sym_else_if_clause_token1] = ACTIONS(1532),
    [aux_sym_else_clause_token1] = ACTIONS(1532),
    [aux_sym_match_expression_token1] = ACTIONS(1532),
    [aux_sym_match_default_expression_token1] = ACTIONS(1532),
    [aux_sym_switch_statement_token1] = ACTIONS(1532),
    [aux_sym_switch_block_token1] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1532),
    [anon_sym_DASH] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1530),
    [anon_sym_BANG] = ACTIONS(1530),
    [anon_sym_AT] = ACTIONS(1530),
    [aux_sym_clone_expression_token1] = ACTIONS(1532),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1532),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1532),
    [anon_sym_DASH_DASH] = ACTIONS(1530),
    [anon_sym_PLUS_PLUS] = ACTIONS(1530),
    [aux_sym__list_destructing_token1] = ACTIONS(1532),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_self] = ACTIONS(1532),
    [anon_sym_parent] = ACTIONS(1532),
    [aux_sym__argument_name_token1] = ACTIONS(1532),
    [aux_sym__argument_name_token2] = ACTIONS(1532),
    [anon_sym_POUND_LBRACK] = ACTIONS(1530),
    [aux_sym_encapsed_string_token1] = ACTIONS(1530),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [aux_sym_string_token1] = ACTIONS(1530),
    [anon_sym_SQUOTE] = ACTIONS(1530),
    [anon_sym_LT_LT_LT] = ACTIONS(1530),
    [anon_sym_BQUOTE] = ACTIONS(1530),
    [anon_sym_DOLLAR] = ACTIONS(1530),
    [aux_sym_yield_expression_token1] = ACTIONS(1532),
    [aux_sym_include_expression_token1] = ACTIONS(1532),
    [aux_sym_include_once_expression_token1] = ACTIONS(1532),
    [aux_sym_require_expression_token1] = ACTIONS(1532),
    [aux_sym_require_once_expression_token1] = ACTIONS(1532),
    [sym_comment] = ACTIONS(3),
  },
  [566] = {
    [ts_builtin_sym_end] = ACTIONS(1534),
    [sym_name] = ACTIONS(1536),
    [anon_sym_QMARK_GT] = ACTIONS(1534),
    [anon_sym_SEMI] = ACTIONS(1534),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1536),
    [aux_sym_global_declaration_token1] = ACTIONS(1536),
    [aux_sym_namespace_definition_token1] = ACTIONS(1536),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1536),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1536),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1536),
    [anon_sym_BSLASH] = ACTIONS(1534),
    [anon_sym_LBRACE] = ACTIONS(1534),
    [anon_sym_RBRACE] = ACTIONS(1534),
    [aux_sym_trait_declaration_token1] = ACTIONS(1536),
    [aux_sym_interface_declaration_token1] = ACTIONS(1536),
    [aux_sym_enum_declaration_token1] = ACTIONS(1536),
    [aux_sym_enum_case_token1] = ACTIONS(1536),
    [aux_sym_class_declaration_token1] = ACTIONS(1536),
    [aux_sym_final_modifier_token1] = ACTIONS(1536),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1536),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1536),
    [sym_var_modifier] = ACTIONS(1536),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1536),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1536),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1536),
    [anon_sym_LPAREN2] = ACTIONS(1534),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1536),
    [aux_sym_cast_type_token1] = ACTIONS(1536),
    [aux_sym_echo_statement_token1] = ACTIONS(1536),
    [aux_sym_exit_statement_token1] = ACTIONS(1536),
    [anon_sym_unset] = ACTIONS(1536),
    [aux_sym_declare_statement_token1] = ACTIONS(1536),
    [aux_sym_declare_statement_token2] = ACTIONS(1536),
    [sym_float] = ACTIONS(1536),
    [aux_sym_try_statement_token1] = ACTIONS(1536),
    [aux_sym_goto_statement_token1] = ACTIONS(1536),
    [aux_sym_continue_statement_token1] = ACTIONS(1536),
    [aux_sym_break_statement_token1] = ACTIONS(1536),
    [sym_integer] = ACTIONS(1536),
    [aux_sym_return_statement_token1] = ACTIONS(1536),
    [aux_sym_throw_expression_token1] = ACTIONS(1536),
    [aux_sym_while_statement_token1] = ACTIONS(1536),
    [aux_sym_while_statement_token2] = ACTIONS(1536),
    [aux_sym_do_statement_token1] = ACTIONS(1536),
    [aux_sym_for_statement_token1] = ACTIONS(1536),
    [aux_sym_for_statement_token2] = ACTIONS(1536),
    [aux_sym_foreach_statement_token1] = ACTIONS(1536),
    [aux_sym_foreach_statement_token2] = ACTIONS(1536),
    [aux_sym_if_statement_token1] = ACTIONS(1536),
    [aux_sym_if_statement_token2] = ACTIONS(1536),
    [aux_sym_else_if_clause_token1] = ACTIONS(1536),
    [aux_sym_else_clause_token1] = ACTIONS(1536),
    [aux_sym_match_expression_token1] = ACTIONS(1536),
    [aux_sym_match_default_expression_token1] = ACTIONS(1536),
    [aux_sym_switch_statement_token1] = ACTIONS(1536),
    [aux_sym_switch_block_token1] = ACTIONS(1536),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_TILDE] = ACTIONS(1534),
    [anon_sym_BANG] = ACTIONS(1534),
    [anon_sym_AT] = ACTIONS(1534),
    [aux_sym_clone_expression_token1] = ACTIONS(1536),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1536),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1536),
    [anon_sym_DASH_DASH] = ACTIONS(1534),
    [anon_sym_PLUS_PLUS] = ACTIONS(1534),
    [aux_sym__list_destructing_token1] = ACTIONS(1536),
    [anon_sym_LBRACK] = ACTIONS(1534),
    [anon_sym_self] = ACTIONS(1536),
    [anon_sym_parent] = ACTIONS(1536),
    [aux_sym__argument_name_token1] = ACTIONS(1536),
    [aux_sym__argument_name_token2] = ACTIONS(1536),
    [anon_sym_POUND_LBRACK] = ACTIONS(1534),
    [aux_sym_encapsed_string_token1] = ACTIONS(1534),
    [anon_sym_DQUOTE] = ACTIONS(1534),
    [aux_sym_string_token1] = ACTIONS(1534),
    [anon_sym_SQUOTE] = ACTIONS(1534),
    [anon_sym_LT_LT_LT] = ACTIONS(1534),
    [anon_sym_BQUOTE] = ACTIONS(1534),
    [anon_sym_DOLLAR] = ACTIONS(1534),
    [aux_sym_yield_expression_token1] = ACTIONS(1536),
    [aux_sym_include_expression_token1] = ACTIONS(1536),
    [aux_sym_include_once_expression_token1] = ACTIONS(1536),
    [aux_sym_require_expression_token1] = ACTIONS(1536),
    [aux_sym_require_once_expression_token1] = ACTIONS(1536),
    [sym_comment] = ACTIONS(3),
  },
  [567] = {
    [ts_builtin_sym_end] = ACTIONS(1538),
    [sym_name] = ACTIONS(1540),
    [anon_sym_QMARK_GT] = ACTIONS(1538),
    [anon_sym_SEMI] = ACTIONS(1538),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1540),
    [aux_sym_global_declaration_token1] = ACTIONS(1540),
    [aux_sym_namespace_definition_token1] = ACTIONS(1540),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1540),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1540),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1540),
    [anon_sym_BSLASH] = ACTIONS(1538),
    [anon_sym_LBRACE] = ACTIONS(1538),
    [anon_sym_RBRACE] = ACTIONS(1538),
    [aux_sym_trait_declaration_token1] = ACTIONS(1540),
    [aux_sym_interface_declaration_token1] = ACTIONS(1540),
    [aux_sym_enum_declaration_token1] = ACTIONS(1540),
    [aux_sym_enum_case_token1] = ACTIONS(1540),
    [aux_sym_class_declaration_token1] = ACTIONS(1540),
    [aux_sym_final_modifier_token1] = ACTIONS(1540),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1540),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1540),
    [sym_var_modifier] = ACTIONS(1540),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1540),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1540),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1540),
    [anon_sym_LPAREN2] = ACTIONS(1538),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1540),
    [aux_sym_cast_type_token1] = ACTIONS(1540),
    [aux_sym_echo_statement_token1] = ACTIONS(1540),
    [aux_sym_exit_statement_token1] = ACTIONS(1540),
    [anon_sym_unset] = ACTIONS(1540),
    [aux_sym_declare_statement_token1] = ACTIONS(1540),
    [aux_sym_declare_statement_token2] = ACTIONS(1540),
    [sym_float] = ACTIONS(1540),
    [aux_sym_try_statement_token1] = ACTIONS(1540),
    [aux_sym_goto_statement_token1] = ACTIONS(1540),
    [aux_sym_continue_statement_token1] = ACTIONS(1540),
    [aux_sym_break_statement_token1] = ACTIONS(1540),
    [sym_integer] = ACTIONS(1540),
    [aux_sym_return_statement_token1] = ACTIONS(1540),
    [aux_sym_throw_expression_token1] = ACTIONS(1540),
    [aux_sym_while_statement_token1] = ACTIONS(1540),
    [aux_sym_while_statement_token2] = ACTIONS(1540),
    [aux_sym_do_statement_token1] = ACTIONS(1540),
    [aux_sym_for_statement_token1] = ACTIONS(1540),
    [aux_sym_for_statement_token2] = ACTIONS(1540),
    [aux_sym_foreach_statement_token1] = ACTIONS(1540),
    [aux_sym_foreach_statement_token2] = ACTIONS(1540),
    [aux_sym_if_statement_token1] = ACTIONS(1540),
    [aux_sym_if_statement_token2] = ACTIONS(1540),
    [aux_sym_else_if_clause_token1] = ACTIONS(1540),
    [aux_sym_else_clause_token1] = ACTIONS(1540),
    [aux_sym_match_expression_token1] = ACTIONS(1540),
    [aux_sym_match_default_expression_token1] = ACTIONS(1540),
    [aux_sym_switch_statement_token1] = ACTIONS(1540),
    [aux_sym_switch_block_token1] = ACTIONS(1540),
    [anon_sym_PLUS] = ACTIONS(1540),
    [anon_sym_DASH] = ACTIONS(1540),
    [anon_sym_TILDE] = ACTIONS(1538),
    [anon_sym_BANG] = ACTIONS(1538),
    [anon_sym_AT] = ACTIONS(1538),
    [aux_sym_clone_expression_token1] = ACTIONS(1540),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1540),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1540),
    [anon_sym_DASH_DASH] = ACTIONS(1538),
    [anon_sym_PLUS_PLUS] = ACTIONS(1538),
    [aux_sym__list_destructing_token1] = ACTIONS(1540),
    [anon_sym_LBRACK] = ACTIONS(1538),
    [anon_sym_self] = ACTIONS(1540),
    [anon_sym_parent] = ACTIONS(1540),
    [aux_sym__argument_name_token1] = ACTIONS(1540),
    [aux_sym__argument_name_token2] = ACTIONS(1540),
    [anon_sym_POUND_LBRACK] = ACTIONS(1538),
    [aux_sym_encapsed_string_token1] = ACTIONS(1538),
    [anon_sym_DQUOTE] = ACTIONS(1538),
    [aux_sym_string_token1] = ACTIONS(1538),
    [anon_sym_SQUOTE] = ACTIONS(1538),
    [anon_sym_LT_LT_LT] = ACTIONS(1538),
    [anon_sym_BQUOTE] = ACTIONS(1538),
    [anon_sym_DOLLAR] = ACTIONS(1538),
    [aux_sym_yield_expression_token1] = ACTIONS(1540),
    [aux_sym_include_expression_token1] = ACTIONS(1540),
    [aux_sym_include_once_expression_token1] = ACTIONS(1540),
    [aux_sym_require_expression_token1] = ACTIONS(1540),
    [aux_sym_require_once_expression_token1] = ACTIONS(1540),
    [sym_comment] = ACTIONS(3),
  },
  [568] = {
    [ts_builtin_sym_end] = ACTIONS(1542),
    [sym_name] = ACTIONS(1544),
    [anon_sym_QMARK_GT] = ACTIONS(1542),
    [anon_sym_SEMI] = ACTIONS(1542),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1544),
    [aux_sym_global_declaration_token1] = ACTIONS(1544),
    [aux_sym_namespace_definition_token1] = ACTIONS(1544),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1544),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1544),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1544),
    [anon_sym_BSLASH] = ACTIONS(1542),
    [anon_sym_LBRACE] = ACTIONS(1542),
    [anon_sym_RBRACE] = ACTIONS(1542),
    [aux_sym_trait_declaration_token1] = ACTIONS(1544),
    [aux_sym_interface_declaration_token1] = ACTIONS(1544),
    [aux_sym_enum_declaration_token1] = ACTIONS(1544),
    [aux_sym_enum_case_token1] = ACTIONS(1544),
    [aux_sym_class_declaration_token1] = ACTIONS(1544),
    [aux_sym_final_modifier_token1] = ACTIONS(1544),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1544),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1544),
    [sym_var_modifier] = ACTIONS(1544),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1544),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1544),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1544),
    [anon_sym_LPAREN2] = ACTIONS(1542),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1544),
    [aux_sym_cast_type_token1] = ACTIONS(1544),
    [aux_sym_echo_statement_token1] = ACTIONS(1544),
    [aux_sym_exit_statement_token1] = ACTIONS(1544),
    [anon_sym_unset] = ACTIONS(1544),
    [aux_sym_declare_statement_token1] = ACTIONS(1544),
    [aux_sym_declare_statement_token2] = ACTIONS(1544),
    [sym_float] = ACTIONS(1544),
    [aux_sym_try_statement_token1] = ACTIONS(1544),
    [aux_sym_goto_statement_token1] = ACTIONS(1544),
    [aux_sym_continue_statement_token1] = ACTIONS(1544),
    [aux_sym_break_statement_token1] = ACTIONS(1544),
    [sym_integer] = ACTIONS(1544),
    [aux_sym_return_statement_token1] = ACTIONS(1544),
    [aux_sym_throw_expression_token1] = ACTIONS(1544),
    [aux_sym_while_statement_token1] = ACTIONS(1544),
    [aux_sym_while_statement_token2] = ACTIONS(1544),
    [aux_sym_do_statement_token1] = ACTIONS(1544),
    [aux_sym_for_statement_token1] = ACTIONS(1544),
    [aux_sym_for_statement_token2] = ACTIONS(1544),
    [aux_sym_foreach_statement_token1] = ACTIONS(1544),
    [aux_sym_foreach_statement_token2] = ACTIONS(1544),
    [aux_sym_if_statement_token1] = ACTIONS(1544),
    [aux_sym_if_statement_token2] = ACTIONS(1544),
    [aux_sym_else_if_clause_token1] = ACTIONS(1544),
    [aux_sym_else_clause_token1] = ACTIONS(1544),
    [aux_sym_match_expression_token1] = ACTIONS(1544),
    [aux_sym_match_default_expression_token1] = ACTIONS(1544),
    [aux_sym_switch_statement_token1] = ACTIONS(1544),
    [aux_sym_switch_block_token1] = ACTIONS(1544),
    [anon_sym_PLUS] = ACTIONS(1544),
    [anon_sym_DASH] = ACTIONS(1544),
    [anon_sym_TILDE] = ACTIONS(1542),
    [anon_sym_BANG] = ACTIONS(1542),
    [anon_sym_AT] = ACTIONS(1542),
    [aux_sym_clone_expression_token1] = ACTIONS(1544),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1544),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1544),
    [anon_sym_DASH_DASH] = ACTIONS(1542),
    [anon_sym_PLUS_PLUS] = ACTIONS(1542),
    [aux_sym__list_destructing_token1] = ACTIONS(1544),
    [anon_sym_LBRACK] = ACTIONS(1542),
    [anon_sym_self] = ACTIONS(1544),
    [anon_sym_parent] = ACTIONS(1544),
    [aux_sym__argument_name_token1] = ACTIONS(1544),
    [aux_sym__argument_name_token2] = ACTIONS(1544),
    [anon_sym_POUND_LBRACK] = ACTIONS(1542),
    [aux_sym_encapsed_string_token1] = ACTIONS(1542),
    [anon_sym_DQUOTE] = ACTIONS(1542),
    [aux_sym_string_token1] = ACTIONS(1542),
    [anon_sym_SQUOTE] = ACTIONS(1542),
    [anon_sym_LT_LT_LT] = ACTIONS(1542),
    [anon_sym_BQUOTE] = ACTIONS(1542),
    [anon_sym_DOLLAR] = ACTIONS(1542),
    [aux_sym_yield_expression_token1] = ACTIONS(1544),
    [aux_sym_include_expression_token1] = ACTIONS(1544),
    [aux_sym_include_once_expression_token1] = ACTIONS(1544),
    [aux_sym_require_expression_token1] = ACTIONS(1544),
    [aux_sym_require_once_expression_token1] = ACTIONS(1544),
    [sym_comment] = ACTIONS(3),
  },
  [569] = {
    [ts_builtin_sym_end] = ACTIONS(1546),
    [sym_name] = ACTIONS(1548),
    [anon_sym_QMARK_GT] = ACTIONS(1546),
    [anon_sym_SEMI] = ACTIONS(1546),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1548),
    [aux_sym_global_declaration_token1] = ACTIONS(1548),
    [aux_sym_namespace_definition_token1] = ACTIONS(1548),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1548),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1548),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1548),
    [anon_sym_BSLASH] = ACTIONS(1546),
    [anon_sym_LBRACE] = ACTIONS(1546),
    [anon_sym_RBRACE] = ACTIONS(1546),
    [aux_sym_trait_declaration_token1] = ACTIONS(1548),
    [aux_sym_interface_declaration_token1] = ACTIONS(1548),
    [aux_sym_enum_declaration_token1] = ACTIONS(1548),
    [aux_sym_enum_case_token1] = ACTIONS(1548),
    [aux_sym_class_declaration_token1] = ACTIONS(1548),
    [aux_sym_final_modifier_token1] = ACTIONS(1548),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1548),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1548),
    [sym_var_modifier] = ACTIONS(1548),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1548),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1548),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1548),
    [anon_sym_LPAREN2] = ACTIONS(1546),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1548),
    [aux_sym_cast_type_token1] = ACTIONS(1548),
    [aux_sym_echo_statement_token1] = ACTIONS(1548),
    [aux_sym_exit_statement_token1] = ACTIONS(1548),
    [anon_sym_unset] = ACTIONS(1548),
    [aux_sym_declare_statement_token1] = ACTIONS(1548),
    [aux_sym_declare_statement_token2] = ACTIONS(1548),
    [sym_float] = ACTIONS(1548),
    [aux_sym_try_statement_token1] = ACTIONS(1548),
    [aux_sym_goto_statement_token1] = ACTIONS(1548),
    [aux_sym_continue_statement_token1] = ACTIONS(1548),
    [aux_sym_break_statement_token1] = ACTIONS(1548),
    [sym_integer] = ACTIONS(1548),
    [aux_sym_return_statement_token1] = ACTIONS(1548),
    [aux_sym_throw_expression_token1] = ACTIONS(1548),
    [aux_sym_while_statement_token1] = ACTIONS(1548),
    [aux_sym_while_statement_token2] = ACTIONS(1548),
    [aux_sym_do_statement_token1] = ACTIONS(1548),
    [aux_sym_for_statement_token1] = ACTIONS(1548),
    [aux_sym_for_statement_token2] = ACTIONS(1548),
    [aux_sym_foreach_statement_token1] = ACTIONS(1548),
    [aux_sym_foreach_statement_token2] = ACTIONS(1548),
    [aux_sym_if_statement_token1] = ACTIONS(1548),
    [aux_sym_if_statement_token2] = ACTIONS(1548),
    [aux_sym_else_if_clause_token1] = ACTIONS(1548),
    [aux_sym_else_clause_token1] = ACTIONS(1548),
    [aux_sym_match_expression_token1] = ACTIONS(1548),
    [aux_sym_match_default_expression_token1] = ACTIONS(1548),
    [aux_sym_switch_statement_token1] = ACTIONS(1548),
    [aux_sym_switch_block_token1] = ACTIONS(1548),
    [anon_sym_PLUS] = ACTIONS(1548),
    [anon_sym_DASH] = ACTIONS(1548),
    [anon_sym_TILDE] = ACTIONS(1546),
    [anon_sym_BANG] = ACTIONS(1546),
    [anon_sym_AT] = ACTIONS(1546),
    [aux_sym_clone_expression_token1] = ACTIONS(1548),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1548),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1548),
    [anon_sym_DASH_DASH] = ACTIONS(1546),
    [anon_sym_PLUS_PLUS] = ACTIONS(1546),
    [aux_sym__list_destructing_token1] = ACTIONS(1548),
    [anon_sym_LBRACK] = ACTIONS(1546),
    [anon_sym_self] = ACTIONS(1548),
    [anon_sym_parent] = ACTIONS(1548),
    [aux_sym__argument_name_token1] = ACTIONS(1548),
    [aux_sym__argument_name_token2] = ACTIONS(1548),
    [anon_sym_POUND_LBRACK] = ACTIONS(1546),
    [aux_sym_encapsed_string_token1] = ACTIONS(1546),
    [anon_sym_DQUOTE] = ACTIONS(1546),
    [aux_sym_string_token1] = ACTIONS(1546),
    [anon_sym_SQUOTE] = ACTIONS(1546),
    [anon_sym_LT_LT_LT] = ACTIONS(1546),
    [anon_sym_BQUOTE] = ACTIONS(1546),
    [anon_sym_DOLLAR] = ACTIONS(1546),
    [aux_sym_yield_expression_token1] = ACTIONS(1548),
    [aux_sym_include_expression_token1] = ACTIONS(1548),
    [aux_sym_include_once_expression_token1] = ACTIONS(1548),
    [aux_sym_require_expression_token1] = ACTIONS(1548),
    [aux_sym_require_once_expression_token1] = ACTIONS(1548),
    [sym_comment] = ACTIONS(3),
  },
  [570] = {
    [ts_builtin_sym_end] = ACTIONS(1550),
    [sym_name] = ACTIONS(1552),
    [anon_sym_QMARK_GT] = ACTIONS(1550),
    [anon_sym_SEMI] = ACTIONS(1550),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1552),
    [aux_sym_global_declaration_token1] = ACTIONS(1552),
    [aux_sym_namespace_definition_token1] = ACTIONS(1552),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1552),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1552),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1552),
    [anon_sym_BSLASH] = ACTIONS(1550),
    [anon_sym_LBRACE] = ACTIONS(1550),
    [anon_sym_RBRACE] = ACTIONS(1550),
    [aux_sym_trait_declaration_token1] = ACTIONS(1552),
    [aux_sym_interface_declaration_token1] = ACTIONS(1552),
    [aux_sym_enum_declaration_token1] = ACTIONS(1552),
    [aux_sym_enum_case_token1] = ACTIONS(1552),
    [aux_sym_class_declaration_token1] = ACTIONS(1552),
    [aux_sym_final_modifier_token1] = ACTIONS(1552),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1552),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1552),
    [sym_var_modifier] = ACTIONS(1552),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1552),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1552),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1552),
    [anon_sym_LPAREN2] = ACTIONS(1550),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1552),
    [aux_sym_cast_type_token1] = ACTIONS(1552),
    [aux_sym_echo_statement_token1] = ACTIONS(1552),
    [aux_sym_exit_statement_token1] = ACTIONS(1552),
    [anon_sym_unset] = ACTIONS(1552),
    [aux_sym_declare_statement_token1] = ACTIONS(1552),
    [aux_sym_declare_statement_token2] = ACTIONS(1552),
    [sym_float] = ACTIONS(1552),
    [aux_sym_try_statement_token1] = ACTIONS(1552),
    [aux_sym_goto_statement_token1] = ACTIONS(1552),
    [aux_sym_continue_statement_token1] = ACTIONS(1552),
    [aux_sym_break_statement_token1] = ACTIONS(1552),
    [sym_integer] = ACTIONS(1552),
    [aux_sym_return_statement_token1] = ACTIONS(1552),
    [aux_sym_throw_expression_token1] = ACTIONS(1552),
    [aux_sym_while_statement_token1] = ACTIONS(1552),
    [aux_sym_while_statement_token2] = ACTIONS(1552),
    [aux_sym_do_statement_token1] = ACTIONS(1552),
    [aux_sym_for_statement_token1] = ACTIONS(1552),
    [aux_sym_for_statement_token2] = ACTIONS(1552),
    [aux_sym_foreach_statement_token1] = ACTIONS(1552),
    [aux_sym_foreach_statement_token2] = ACTIONS(1552),
    [aux_sym_if_statement_token1] = ACTIONS(1552),
    [aux_sym_if_statement_token2] = ACTIONS(1552),
    [aux_sym_else_if_clause_token1] = ACTIONS(1552),
    [aux_sym_else_clause_token1] = ACTIONS(1552),
    [aux_sym_match_expression_token1] = ACTIONS(1552),
    [aux_sym_match_default_expression_token1] = ACTIONS(1552),
    [aux_sym_switch_statement_token1] = ACTIONS(1552),
    [aux_sym_switch_block_token1] = ACTIONS(1552),
    [anon_sym_PLUS] = ACTIONS(1552),
    [anon_sym_DASH] = ACTIONS(1552),
    [anon_sym_TILDE] = ACTIONS(1550),
    [anon_sym_BANG] = ACTIONS(1550),
    [anon_sym_AT] = ACTIONS(1550),
    [aux_sym_clone_expression_token1] = ACTIONS(1552),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1552),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1552),
    [anon_sym_DASH_DASH] = ACTIONS(1550),
    [anon_sym_PLUS_PLUS] = ACTIONS(1550),
    [aux_sym__list_destructing_token1] = ACTIONS(1552),
    [anon_sym_LBRACK] = ACTIONS(1550),
    [anon_sym_self] = ACTIONS(1552),
    [anon_sym_parent] = ACTIONS(1552),
    [aux_sym__argument_name_token1] = ACTIONS(1552),
    [aux_sym__argument_name_token2] = ACTIONS(1552),
    [anon_sym_POUND_LBRACK] = ACTIONS(1550),
    [aux_sym_encapsed_string_token1] = ACTIONS(1550),
    [anon_sym_DQUOTE] = ACTIONS(1550),
    [aux_sym_string_token1] = ACTIONS(1550),
    [anon_sym_SQUOTE] = ACTIONS(1550),
    [anon_sym_LT_LT_LT] = ACTIONS(1550),
    [anon_sym_BQUOTE] = ACTIONS(1550),
    [anon_sym_DOLLAR] = ACTIONS(1550),
    [aux_sym_yield_expression_token1] = ACTIONS(1552),
    [aux_sym_include_expression_token1] = ACTIONS(1552),
    [aux_sym_include_once_expression_token1] = ACTIONS(1552),
    [aux_sym_require_expression_token1] = ACTIONS(1552),
    [aux_sym_require_once_expression_token1] = ACTIONS(1552),
    [sym_comment] = ACTIONS(3),
  },
  [571] = {
    [ts_builtin_sym_end] = ACTIONS(1554),
    [sym_name] = ACTIONS(1556),
    [anon_sym_QMARK_GT] = ACTIONS(1554),
    [anon_sym_SEMI] = ACTIONS(1554),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1556),
    [aux_sym_global_declaration_token1] = ACTIONS(1556),
    [aux_sym_namespace_definition_token1] = ACTIONS(1556),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1556),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1556),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1556),
    [anon_sym_BSLASH] = ACTIONS(1554),
    [anon_sym_LBRACE] = ACTIONS(1554),
    [anon_sym_RBRACE] = ACTIONS(1554),
    [aux_sym_trait_declaration_token1] = ACTIONS(1556),
    [aux_sym_interface_declaration_token1] = ACTIONS(1556),
    [aux_sym_enum_declaration_token1] = ACTIONS(1556),
    [aux_sym_enum_case_token1] = ACTIONS(1556),
    [aux_sym_class_declaration_token1] = ACTIONS(1556),
    [aux_sym_final_modifier_token1] = ACTIONS(1556),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1556),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1556),
    [sym_var_modifier] = ACTIONS(1556),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1556),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1556),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1556),
    [anon_sym_LPAREN2] = ACTIONS(1554),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1556),
    [aux_sym_cast_type_token1] = ACTIONS(1556),
    [aux_sym_echo_statement_token1] = ACTIONS(1556),
    [aux_sym_exit_statement_token1] = ACTIONS(1556),
    [anon_sym_unset] = ACTIONS(1556),
    [aux_sym_declare_statement_token1] = ACTIONS(1556),
    [aux_sym_declare_statement_token2] = ACTIONS(1556),
    [sym_float] = ACTIONS(1556),
    [aux_sym_try_statement_token1] = ACTIONS(1556),
    [aux_sym_goto_statement_token1] = ACTIONS(1556),
    [aux_sym_continue_statement_token1] = ACTIONS(1556),
    [aux_sym_break_statement_token1] = ACTIONS(1556),
    [sym_integer] = ACTIONS(1556),
    [aux_sym_return_statement_token1] = ACTIONS(1556),
    [aux_sym_throw_expression_token1] = ACTIONS(1556),
    [aux_sym_while_statement_token1] = ACTIONS(1556),
    [aux_sym_while_statement_token2] = ACTIONS(1556),
    [aux_sym_do_statement_token1] = ACTIONS(1556),
    [aux_sym_for_statement_token1] = ACTIONS(1556),
    [aux_sym_for_statement_token2] = ACTIONS(1556),
    [aux_sym_foreach_statement_token1] = ACTIONS(1556),
    [aux_sym_foreach_statement_token2] = ACTIONS(1556),
    [aux_sym_if_statement_token1] = ACTIONS(1556),
    [aux_sym_if_statement_token2] = ACTIONS(1556),
    [aux_sym_else_if_clause_token1] = ACTIONS(1556),
    [aux_sym_else_clause_token1] = ACTIONS(1556),
    [aux_sym_match_expression_token1] = ACTIONS(1556),
    [aux_sym_match_default_expression_token1] = ACTIONS(1556),
    [aux_sym_switch_statement_token1] = ACTIONS(1556),
    [aux_sym_switch_block_token1] = ACTIONS(1556),
    [anon_sym_PLUS] = ACTIONS(1556),
    [anon_sym_DASH] = ACTIONS(1556),
    [anon_sym_TILDE] = ACTIONS(1554),
    [anon_sym_BANG] = ACTIONS(1554),
    [anon_sym_AT] = ACTIONS(1554),
    [aux_sym_clone_expression_token1] = ACTIONS(1556),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1556),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1556),
    [anon_sym_DASH_DASH] = ACTIONS(1554),
    [anon_sym_PLUS_PLUS] = ACTIONS(1554),
    [aux_sym__list_destructing_token1] = ACTIONS(1556),
    [anon_sym_LBRACK] = ACTIONS(1554),
    [anon_sym_self] = ACTIONS(1556),
    [anon_sym_parent] = ACTIONS(1556),
    [aux_sym__argument_name_token1] = ACTIONS(1556),
    [aux_sym__argument_name_token2] = ACTIONS(1556),
    [anon_sym_POUND_LBRACK] = ACTIONS(1554),
    [aux_sym_encapsed_string_token1] = ACTIONS(1554),
    [anon_sym_DQUOTE] = ACTIONS(1554),
    [aux_sym_string_token1] = ACTIONS(1554),
    [anon_sym_SQUOTE] = ACTIONS(1554),
    [anon_sym_LT_LT_LT] = ACTIONS(1554),
    [anon_sym_BQUOTE] = ACTIONS(1554),
    [anon_sym_DOLLAR] = ACTIONS(1554),
    [aux_sym_yield_expression_token1] = ACTIONS(1556),
    [aux_sym_include_expression_token1] = ACTIONS(1556),
    [aux_sym_include_once_expression_token1] = ACTIONS(1556),
    [aux_sym_require_expression_token1] = ACTIONS(1556),
    [aux_sym_require_once_expression_token1] = ACTIONS(1556),
    [sym_comment] = ACTIONS(3),
  },
  [572] = {
    [ts_builtin_sym_end] = ACTIONS(1558),
    [sym_name] = ACTIONS(1560),
    [anon_sym_QMARK_GT] = ACTIONS(1558),
    [anon_sym_SEMI] = ACTIONS(1558),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1560),
    [aux_sym_global_declaration_token1] = ACTIONS(1560),
    [aux_sym_namespace_definition_token1] = ACTIONS(1560),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1560),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1560),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1560),
    [anon_sym_BSLASH] = ACTIONS(1558),
    [anon_sym_LBRACE] = ACTIONS(1558),
    [anon_sym_RBRACE] = ACTIONS(1558),
    [aux_sym_trait_declaration_token1] = ACTIONS(1560),
    [aux_sym_interface_declaration_token1] = ACTIONS(1560),
    [aux_sym_enum_declaration_token1] = ACTIONS(1560),
    [aux_sym_enum_case_token1] = ACTIONS(1560),
    [aux_sym_class_declaration_token1] = ACTIONS(1560),
    [aux_sym_final_modifier_token1] = ACTIONS(1560),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1560),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1560),
    [sym_var_modifier] = ACTIONS(1560),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1560),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1560),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1560),
    [anon_sym_LPAREN2] = ACTIONS(1558),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1560),
    [aux_sym_cast_type_token1] = ACTIONS(1560),
    [aux_sym_echo_statement_token1] = ACTIONS(1560),
    [aux_sym_exit_statement_token1] = ACTIONS(1560),
    [anon_sym_unset] = ACTIONS(1560),
    [aux_sym_declare_statement_token1] = ACTIONS(1560),
    [aux_sym_declare_statement_token2] = ACTIONS(1560),
    [sym_float] = ACTIONS(1560),
    [aux_sym_try_statement_token1] = ACTIONS(1560),
    [aux_sym_goto_statement_token1] = ACTIONS(1560),
    [aux_sym_continue_statement_token1] = ACTIONS(1560),
    [aux_sym_break_statement_token1] = ACTIONS(1560),
    [sym_integer] = ACTIONS(1560),
    [aux_sym_return_statement_token1] = ACTIONS(1560),
    [aux_sym_throw_expression_token1] = ACTIONS(1560),
    [aux_sym_while_statement_token1] = ACTIONS(1560),
    [aux_sym_while_statement_token2] = ACTIONS(1560),
    [aux_sym_do_statement_token1] = ACTIONS(1560),
    [aux_sym_for_statement_token1] = ACTIONS(1560),
    [aux_sym_for_statement_token2] = ACTIONS(1560),
    [aux_sym_foreach_statement_token1] = ACTIONS(1560),
    [aux_sym_foreach_statement_token2] = ACTIONS(1560),
    [aux_sym_if_statement_token1] = ACTIONS(1560),
    [aux_sym_if_statement_token2] = ACTIONS(1560),
    [aux_sym_else_if_clause_token1] = ACTIONS(1560),
    [aux_sym_else_clause_token1] = ACTIONS(1560),
    [aux_sym_match_expression_token1] = ACTIONS(1560),
    [aux_sym_match_default_expression_token1] = ACTIONS(1560),
    [aux_sym_switch_statement_token1] = ACTIONS(1560),
    [aux_sym_switch_block_token1] = ACTIONS(1560),
    [anon_sym_PLUS] = ACTIONS(1560),
    [anon_sym_DASH] = ACTIONS(1560),
    [anon_sym_TILDE] = ACTIONS(1558),
    [anon_sym_BANG] = ACTIONS(1558),
    [anon_sym_AT] = ACTIONS(1558),
    [aux_sym_clone_expression_token1] = ACTIONS(1560),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1560),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1560),
    [anon_sym_DASH_DASH] = ACTIONS(1558),
    [anon_sym_PLUS_PLUS] = ACTIONS(1558),
    [aux_sym__list_destructing_token1] = ACTIONS(1560),
    [anon_sym_LBRACK] = ACTIONS(1558),
    [anon_sym_self] = ACTIONS(1560),
    [anon_sym_parent] = ACTIONS(1560),
    [aux_sym__argument_name_token1] = ACTIONS(1560),
    [aux_sym__argument_name_token2] = ACTIONS(1560),
    [anon_sym_POUND_LBRACK] = ACTIONS(1558),
    [aux_sym_encapsed_string_token1] = ACTIONS(1558),
    [anon_sym_DQUOTE] = ACTIONS(1558),
    [aux_sym_string_token1] = ACTIONS(1558),
    [anon_sym_SQUOTE] = ACTIONS(1558),
    [anon_sym_LT_LT_LT] = ACTIONS(1558),
    [anon_sym_BQUOTE] = ACTIONS(1558),
    [anon_sym_DOLLAR] = ACTIONS(1558),
    [aux_sym_yield_expression_token1] = ACTIONS(1560),
    [aux_sym_include_expression_token1] = ACTIONS(1560),
    [aux_sym_include_once_expression_token1] = ACTIONS(1560),
    [aux_sym_require_expression_token1] = ACTIONS(1560),
    [aux_sym_require_once_expression_token1] = ACTIONS(1560),
    [sym_comment] = ACTIONS(3),
  },
  [573] = {
    [ts_builtin_sym_end] = ACTIONS(1562),
    [sym_name] = ACTIONS(1564),
    [anon_sym_QMARK_GT] = ACTIONS(1562),
    [anon_sym_SEMI] = ACTIONS(1562),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1564),
    [aux_sym_global_declaration_token1] = ACTIONS(1564),
    [aux_sym_namespace_definition_token1] = ACTIONS(1564),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1564),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1564),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1564),
    [anon_sym_BSLASH] = ACTIONS(1562),
    [anon_sym_LBRACE] = ACTIONS(1562),
    [anon_sym_RBRACE] = ACTIONS(1562),
    [aux_sym_trait_declaration_token1] = ACTIONS(1564),
    [aux_sym_interface_declaration_token1] = ACTIONS(1564),
    [aux_sym_enum_declaration_token1] = ACTIONS(1564),
    [aux_sym_enum_case_token1] = ACTIONS(1564),
    [aux_sym_class_declaration_token1] = ACTIONS(1564),
    [aux_sym_final_modifier_token1] = ACTIONS(1564),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1564),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1564),
    [sym_var_modifier] = ACTIONS(1564),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1564),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1564),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1564),
    [anon_sym_LPAREN2] = ACTIONS(1562),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1564),
    [aux_sym_cast_type_token1] = ACTIONS(1564),
    [aux_sym_echo_statement_token1] = ACTIONS(1564),
    [aux_sym_exit_statement_token1] = ACTIONS(1564),
    [anon_sym_unset] = ACTIONS(1564),
    [aux_sym_declare_statement_token1] = ACTIONS(1564),
    [aux_sym_declare_statement_token2] = ACTIONS(1564),
    [sym_float] = ACTIONS(1564),
    [aux_sym_try_statement_token1] = ACTIONS(1564),
    [aux_sym_goto_statement_token1] = ACTIONS(1564),
    [aux_sym_continue_statement_token1] = ACTIONS(1564),
    [aux_sym_break_statement_token1] = ACTIONS(1564),
    [sym_integer] = ACTIONS(1564),
    [aux_sym_return_statement_token1] = ACTIONS(1564),
    [aux_sym_throw_expression_token1] = ACTIONS(1564),
    [aux_sym_while_statement_token1] = ACTIONS(1564),
    [aux_sym_while_statement_token2] = ACTIONS(1564),
    [aux_sym_do_statement_token1] = ACTIONS(1564),
    [aux_sym_for_statement_token1] = ACTIONS(1564),
    [aux_sym_for_statement_token2] = ACTIONS(1564),
    [aux_sym_foreach_statement_token1] = ACTIONS(1564),
    [aux_sym_foreach_statement_token2] = ACTIONS(1564),
    [aux_sym_if_statement_token1] = ACTIONS(1564),
    [aux_sym_if_statement_token2] = ACTIONS(1564),
    [aux_sym_else_if_clause_token1] = ACTIONS(1564),
    [aux_sym_else_clause_token1] = ACTIONS(1564),
    [aux_sym_match_expression_token1] = ACTIONS(1564),
    [aux_sym_match_default_expression_token1] = ACTIONS(1564),
    [aux_sym_switch_statement_token1] = ACTIONS(1564),
    [aux_sym_switch_block_token1] = ACTIONS(1564),
    [anon_sym_PLUS] = ACTIONS(1564),
    [anon_sym_DASH] = ACTIONS(1564),
    [anon_sym_TILDE] = ACTIONS(1562),
    [anon_sym_BANG] = ACTIONS(1562),
    [anon_sym_AT] = ACTIONS(1562),
    [aux_sym_clone_expression_token1] = ACTIONS(1564),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1564),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1564),
    [anon_sym_DASH_DASH] = ACTIONS(1562),
    [anon_sym_PLUS_PLUS] = ACTIONS(1562),
    [aux_sym__list_destructing_token1] = ACTIONS(1564),
    [anon_sym_LBRACK] = ACTIONS(1562),
    [anon_sym_self] = ACTIONS(1564),
    [anon_sym_parent] = ACTIONS(1564),
    [aux_sym__argument_name_token1] = ACTIONS(1564),
    [aux_sym__argument_name_token2] = ACTIONS(1564),
    [anon_sym_POUND_LBRACK] = ACTIONS(1562),
    [aux_sym_encapsed_string_token1] = ACTIONS(1562),
    [anon_sym_DQUOTE] = ACTIONS(1562),
    [aux_sym_string_token1] = ACTIONS(1562),
    [anon_sym_SQUOTE] = ACTIONS(1562),
    [anon_sym_LT_LT_LT] = ACTIONS(1562),
    [anon_sym_BQUOTE] = ACTIONS(1562),
    [anon_sym_DOLLAR] = ACTIONS(1562),
    [aux_sym_yield_expression_token1] = ACTIONS(1564),
    [aux_sym_include_expression_token1] = ACTIONS(1564),
    [aux_sym_include_once_expression_token1] = ACTIONS(1564),
    [aux_sym_require_expression_token1] = ACTIONS(1564),
    [aux_sym_require_once_expression_token1] = ACTIONS(1564),
    [sym_comment] = ACTIONS(3),
  },
  [574] = {
    [ts_builtin_sym_end] = ACTIONS(1566),
    [sym_name] = ACTIONS(1568),
    [anon_sym_QMARK_GT] = ACTIONS(1566),
    [anon_sym_SEMI] = ACTIONS(1566),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1568),
    [aux_sym_global_declaration_token1] = ACTIONS(1568),
    [aux_sym_namespace_definition_token1] = ACTIONS(1568),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1568),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1568),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1568),
    [anon_sym_BSLASH] = ACTIONS(1566),
    [anon_sym_LBRACE] = ACTIONS(1566),
    [anon_sym_RBRACE] = ACTIONS(1566),
    [aux_sym_trait_declaration_token1] = ACTIONS(1568),
    [aux_sym_interface_declaration_token1] = ACTIONS(1568),
    [aux_sym_enum_declaration_token1] = ACTIONS(1568),
    [aux_sym_enum_case_token1] = ACTIONS(1568),
    [aux_sym_class_declaration_token1] = ACTIONS(1568),
    [aux_sym_final_modifier_token1] = ACTIONS(1568),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1568),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1568),
    [sym_var_modifier] = ACTIONS(1568),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1568),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1568),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1568),
    [anon_sym_LPAREN2] = ACTIONS(1566),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1568),
    [aux_sym_cast_type_token1] = ACTIONS(1568),
    [aux_sym_echo_statement_token1] = ACTIONS(1568),
    [aux_sym_exit_statement_token1] = ACTIONS(1568),
    [anon_sym_unset] = ACTIONS(1568),
    [aux_sym_declare_statement_token1] = ACTIONS(1568),
    [aux_sym_declare_statement_token2] = ACTIONS(1568),
    [sym_float] = ACTIONS(1568),
    [aux_sym_try_statement_token1] = ACTIONS(1568),
    [aux_sym_goto_statement_token1] = ACTIONS(1568),
    [aux_sym_continue_statement_token1] = ACTIONS(1568),
    [aux_sym_break_statement_token1] = ACTIONS(1568),
    [sym_integer] = ACTIONS(1568),
    [aux_sym_return_statement_token1] = ACTIONS(1568),
    [aux_sym_throw_expression_token1] = ACTIONS(1568),
    [aux_sym_while_statement_token1] = ACTIONS(1568),
    [aux_sym_while_statement_token2] = ACTIONS(1568),
    [aux_sym_do_statement_token1] = ACTIONS(1568),
    [aux_sym_for_statement_token1] = ACTIONS(1568),
    [aux_sym_for_statement_token2] = ACTIONS(1568),
    [aux_sym_foreach_statement_token1] = ACTIONS(1568),
    [aux_sym_foreach_statement_token2] = ACTIONS(1568),
    [aux_sym_if_statement_token1] = ACTIONS(1568),
    [aux_sym_if_statement_token2] = ACTIONS(1568),
    [aux_sym_else_if_clause_token1] = ACTIONS(1568),
    [aux_sym_else_clause_token1] = ACTIONS(1568),
    [aux_sym_match_expression_token1] = ACTIONS(1568),
    [aux_sym_match_default_expression_token1] = ACTIONS(1568),
    [aux_sym_switch_statement_token1] = ACTIONS(1568),
    [aux_sym_switch_block_token1] = ACTIONS(1568),
    [anon_sym_PLUS] = ACTIONS(1568),
    [anon_sym_DASH] = ACTIONS(1568),
    [anon_sym_TILDE] = ACTIONS(1566),
    [anon_sym_BANG] = ACTIONS(1566),
    [anon_sym_AT] = ACTIONS(1566),
    [aux_sym_clone_expression_token1] = ACTIONS(1568),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1568),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1568),
    [anon_sym_DASH_DASH] = ACTIONS(1566),
    [anon_sym_PLUS_PLUS] = ACTIONS(1566),
    [aux_sym__list_destructing_token1] = ACTIONS(1568),
    [anon_sym_LBRACK] = ACTIONS(1566),
    [anon_sym_self] = ACTIONS(1568),
    [anon_sym_parent] = ACTIONS(1568),
    [aux_sym__argument_name_token1] = ACTIONS(1568),
    [aux_sym__argument_name_token2] = ACTIONS(1568),
    [anon_sym_POUND_LBRACK] = ACTIONS(1566),
    [aux_sym_encapsed_string_token1] = ACTIONS(1566),
    [anon_sym_DQUOTE] = ACTIONS(1566),
    [aux_sym_string_token1] = ACTIONS(1566),
    [anon_sym_SQUOTE] = ACTIONS(1566),
    [anon_sym_LT_LT_LT] = ACTIONS(1566),
    [anon_sym_BQUOTE] = ACTIONS(1566),
    [anon_sym_DOLLAR] = ACTIONS(1566),
    [aux_sym_yield_expression_token1] = ACTIONS(1568),
    [aux_sym_include_expression_token1] = ACTIONS(1568),
    [aux_sym_include_once_expression_token1] = ACTIONS(1568),
    [aux_sym_require_expression_token1] = ACTIONS(1568),
    [aux_sym_require_once_expression_token1] = ACTIONS(1568),
    [sym_comment] = ACTIONS(3),
  },
  [575] = {
    [ts_builtin_sym_end] = ACTIONS(1570),
    [sym_name] = ACTIONS(1572),
    [anon_sym_QMARK_GT] = ACTIONS(1570),
    [anon_sym_SEMI] = ACTIONS(1570),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1572),
    [aux_sym_global_declaration_token1] = ACTIONS(1572),
    [aux_sym_namespace_definition_token1] = ACTIONS(1572),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1572),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1572),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1572),
    [anon_sym_BSLASH] = ACTIONS(1570),
    [anon_sym_LBRACE] = ACTIONS(1570),
    [anon_sym_RBRACE] = ACTIONS(1570),
    [aux_sym_trait_declaration_token1] = ACTIONS(1572),
    [aux_sym_interface_declaration_token1] = ACTIONS(1572),
    [aux_sym_enum_declaration_token1] = ACTIONS(1572),
    [aux_sym_enum_case_token1] = ACTIONS(1572),
    [aux_sym_class_declaration_token1] = ACTIONS(1572),
    [aux_sym_final_modifier_token1] = ACTIONS(1572),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1572),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1572),
    [sym_var_modifier] = ACTIONS(1572),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1572),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1572),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1572),
    [anon_sym_LPAREN2] = ACTIONS(1570),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1572),
    [aux_sym_cast_type_token1] = ACTIONS(1572),
    [aux_sym_echo_statement_token1] = ACTIONS(1572),
    [aux_sym_exit_statement_token1] = ACTIONS(1572),
    [anon_sym_unset] = ACTIONS(1572),
    [aux_sym_declare_statement_token1] = ACTIONS(1572),
    [aux_sym_declare_statement_token2] = ACTIONS(1572),
    [sym_float] = ACTIONS(1572),
    [aux_sym_try_statement_token1] = ACTIONS(1572),
    [aux_sym_goto_statement_token1] = ACTIONS(1572),
    [aux_sym_continue_statement_token1] = ACTIONS(1572),
    [aux_sym_break_statement_token1] = ACTIONS(1572),
    [sym_integer] = ACTIONS(1572),
    [aux_sym_return_statement_token1] = ACTIONS(1572),
    [aux_sym_throw_expression_token1] = ACTIONS(1572),
    [aux_sym_while_statement_token1] = ACTIONS(1572),
    [aux_sym_while_statement_token2] = ACTIONS(1572),
    [aux_sym_do_statement_token1] = ACTIONS(1572),
    [aux_sym_for_statement_token1] = ACTIONS(1572),
    [aux_sym_for_statement_token2] = ACTIONS(1572),
    [aux_sym_foreach_statement_token1] = ACTIONS(1572),
    [aux_sym_foreach_statement_token2] = ACTIONS(1572),
    [aux_sym_if_statement_token1] = ACTIONS(1572),
    [aux_sym_if_statement_token2] = ACTIONS(1572),
    [aux_sym_else_if_clause_token1] = ACTIONS(1572),
    [aux_sym_else_clause_token1] = ACTIONS(1572),
    [aux_sym_match_expression_token1] = ACTIONS(1572),
    [aux_sym_match_default_expression_token1] = ACTIONS(1572),
    [aux_sym_switch_statement_token1] = ACTIONS(1572),
    [aux_sym_switch_block_token1] = ACTIONS(1572),
    [anon_sym_PLUS] = ACTIONS(1572),
    [anon_sym_DASH] = ACTIONS(1572),
    [anon_sym_TILDE] = ACTIONS(1570),
    [anon_sym_BANG] = ACTIONS(1570),
    [anon_sym_AT] = ACTIONS(1570),
    [aux_sym_clone_expression_token1] = ACTIONS(1572),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1572),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1572),
    [anon_sym_DASH_DASH] = ACTIONS(1570),
    [anon_sym_PLUS_PLUS] = ACTIONS(1570),
    [aux_sym__list_destructing_token1] = ACTIONS(1572),
    [anon_sym_LBRACK] = ACTIONS(1570),
    [anon_sym_self] = ACTIONS(1572),
    [anon_sym_parent] = ACTIONS(1572),
    [aux_sym__argument_name_token1] = ACTIONS(1572),
    [aux_sym__argument_name_token2] = ACTIONS(1572),
    [anon_sym_POUND_LBRACK] = ACTIONS(1570),
    [aux_sym_encapsed_string_token1] = ACTIONS(1570),
    [anon_sym_DQUOTE] = ACTIONS(1570),
    [aux_sym_string_token1] = ACTIONS(1570),
    [anon_sym_SQUOTE] = ACTIONS(1570),
    [anon_sym_LT_LT_LT] = ACTIONS(1570),
    [anon_sym_BQUOTE] = ACTIONS(1570),
    [anon_sym_DOLLAR] = ACTIONS(1570),
    [aux_sym_yield_expression_token1] = ACTIONS(1572),
    [aux_sym_include_expression_token1] = ACTIONS(1572),
    [aux_sym_include_once_expression_token1] = ACTIONS(1572),
    [aux_sym_require_expression_token1] = ACTIONS(1572),
    [aux_sym_require_once_expression_token1] = ACTIONS(1572),
    [sym_comment] = ACTIONS(3),
  },
  [576] = {
    [ts_builtin_sym_end] = ACTIONS(1574),
    [sym_name] = ACTIONS(1576),
    [anon_sym_QMARK_GT] = ACTIONS(1574),
    [anon_sym_SEMI] = ACTIONS(1574),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1576),
    [aux_sym_global_declaration_token1] = ACTIONS(1576),
    [aux_sym_namespace_definition_token1] = ACTIONS(1576),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1576),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1576),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1576),
    [anon_sym_BSLASH] = ACTIONS(1574),
    [anon_sym_LBRACE] = ACTIONS(1574),
    [anon_sym_RBRACE] = ACTIONS(1574),
    [aux_sym_trait_declaration_token1] = ACTIONS(1576),
    [aux_sym_interface_declaration_token1] = ACTIONS(1576),
    [aux_sym_enum_declaration_token1] = ACTIONS(1576),
    [aux_sym_enum_case_token1] = ACTIONS(1576),
    [aux_sym_class_declaration_token1] = ACTIONS(1576),
    [aux_sym_final_modifier_token1] = ACTIONS(1576),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1576),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1576),
    [sym_var_modifier] = ACTIONS(1576),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1576),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1576),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1576),
    [anon_sym_LPAREN2] = ACTIONS(1574),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1576),
    [aux_sym_cast_type_token1] = ACTIONS(1576),
    [aux_sym_echo_statement_token1] = ACTIONS(1576),
    [aux_sym_exit_statement_token1] = ACTIONS(1576),
    [anon_sym_unset] = ACTIONS(1576),
    [aux_sym_declare_statement_token1] = ACTIONS(1576),
    [aux_sym_declare_statement_token2] = ACTIONS(1576),
    [sym_float] = ACTIONS(1576),
    [aux_sym_try_statement_token1] = ACTIONS(1576),
    [aux_sym_goto_statement_token1] = ACTIONS(1576),
    [aux_sym_continue_statement_token1] = ACTIONS(1576),
    [aux_sym_break_statement_token1] = ACTIONS(1576),
    [sym_integer] = ACTIONS(1576),
    [aux_sym_return_statement_token1] = ACTIONS(1576),
    [aux_sym_throw_expression_token1] = ACTIONS(1576),
    [aux_sym_while_statement_token1] = ACTIONS(1576),
    [aux_sym_while_statement_token2] = ACTIONS(1576),
    [aux_sym_do_statement_token1] = ACTIONS(1576),
    [aux_sym_for_statement_token1] = ACTIONS(1576),
    [aux_sym_for_statement_token2] = ACTIONS(1576),
    [aux_sym_foreach_statement_token1] = ACTIONS(1576),
    [aux_sym_foreach_statement_token2] = ACTIONS(1576),
    [aux_sym_if_statement_token1] = ACTIONS(1576),
    [aux_sym_if_statement_token2] = ACTIONS(1576),
    [aux_sym_else_if_clause_token1] = ACTIONS(1576),
    [aux_sym_else_clause_token1] = ACTIONS(1576),
    [aux_sym_match_expression_token1] = ACTIONS(1576),
    [aux_sym_match_default_expression_token1] = ACTIONS(1576),
    [aux_sym_switch_statement_token1] = ACTIONS(1576),
    [aux_sym_switch_block_token1] = ACTIONS(1576),
    [anon_sym_PLUS] = ACTIONS(1576),
    [anon_sym_DASH] = ACTIONS(1576),
    [anon_sym_TILDE] = ACTIONS(1574),
    [anon_sym_BANG] = ACTIONS(1574),
    [anon_sym_AT] = ACTIONS(1574),
    [aux_sym_clone_expression_token1] = ACTIONS(1576),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1576),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1576),
    [anon_sym_DASH_DASH] = ACTIONS(1574),
    [anon_sym_PLUS_PLUS] = ACTIONS(1574),
    [aux_sym__list_destructing_token1] = ACTIONS(1576),
    [anon_sym_LBRACK] = ACTIONS(1574),
    [anon_sym_self] = ACTIONS(1576),
    [anon_sym_parent] = ACTIONS(1576),
    [aux_sym__argument_name_token1] = ACTIONS(1576),
    [aux_sym__argument_name_token2] = ACTIONS(1576),
    [anon_sym_POUND_LBRACK] = ACTIONS(1574),
    [aux_sym_encapsed_string_token1] = ACTIONS(1574),
    [anon_sym_DQUOTE] = ACTIONS(1574),
    [aux_sym_string_token1] = ACTIONS(1574),
    [anon_sym_SQUOTE] = ACTIONS(1574),
    [anon_sym_LT_LT_LT] = ACTIONS(1574),
    [anon_sym_BQUOTE] = ACTIONS(1574),
    [anon_sym_DOLLAR] = ACTIONS(1574),
    [aux_sym_yield_expression_token1] = ACTIONS(1576),
    [aux_sym_include_expression_token1] = ACTIONS(1576),
    [aux_sym_include_once_expression_token1] = ACTIONS(1576),
    [aux_sym_require_expression_token1] = ACTIONS(1576),
    [aux_sym_require_once_expression_token1] = ACTIONS(1576),
    [sym_comment] = ACTIONS(3),
  },
  [577] = {
    [ts_builtin_sym_end] = ACTIONS(1578),
    [sym_name] = ACTIONS(1580),
    [anon_sym_QMARK_GT] = ACTIONS(1578),
    [anon_sym_SEMI] = ACTIONS(1578),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1580),
    [aux_sym_global_declaration_token1] = ACTIONS(1580),
    [aux_sym_namespace_definition_token1] = ACTIONS(1580),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1580),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1580),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1580),
    [anon_sym_BSLASH] = ACTIONS(1578),
    [anon_sym_LBRACE] = ACTIONS(1578),
    [anon_sym_RBRACE] = ACTIONS(1578),
    [aux_sym_trait_declaration_token1] = ACTIONS(1580),
    [aux_sym_interface_declaration_token1] = ACTIONS(1580),
    [aux_sym_enum_declaration_token1] = ACTIONS(1580),
    [aux_sym_enum_case_token1] = ACTIONS(1580),
    [aux_sym_class_declaration_token1] = ACTIONS(1580),
    [aux_sym_final_modifier_token1] = ACTIONS(1580),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1580),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1580),
    [sym_var_modifier] = ACTIONS(1580),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1580),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1580),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1580),
    [anon_sym_LPAREN2] = ACTIONS(1578),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1580),
    [aux_sym_cast_type_token1] = ACTIONS(1580),
    [aux_sym_echo_statement_token1] = ACTIONS(1580),
    [aux_sym_exit_statement_token1] = ACTIONS(1580),
    [anon_sym_unset] = ACTIONS(1580),
    [aux_sym_declare_statement_token1] = ACTIONS(1580),
    [aux_sym_declare_statement_token2] = ACTIONS(1580),
    [sym_float] = ACTIONS(1580),
    [aux_sym_try_statement_token1] = ACTIONS(1580),
    [aux_sym_goto_statement_token1] = ACTIONS(1580),
    [aux_sym_continue_statement_token1] = ACTIONS(1580),
    [aux_sym_break_statement_token1] = ACTIONS(1580),
    [sym_integer] = ACTIONS(1580),
    [aux_sym_return_statement_token1] = ACTIONS(1580),
    [aux_sym_throw_expression_token1] = ACTIONS(1580),
    [aux_sym_while_statement_token1] = ACTIONS(1580),
    [aux_sym_while_statement_token2] = ACTIONS(1580),
    [aux_sym_do_statement_token1] = ACTIONS(1580),
    [aux_sym_for_statement_token1] = ACTIONS(1580),
    [aux_sym_for_statement_token2] = ACTIONS(1580),
    [aux_sym_foreach_statement_token1] = ACTIONS(1580),
    [aux_sym_foreach_statement_token2] = ACTIONS(1580),
    [aux_sym_if_statement_token1] = ACTIONS(1580),
    [aux_sym_if_statement_token2] = ACTIONS(1580),
    [aux_sym_else_if_clause_token1] = ACTIONS(1580),
    [aux_sym_else_clause_token1] = ACTIONS(1580),
    [aux_sym_match_expression_token1] = ACTIONS(1580),
    [aux_sym_match_default_expression_token1] = ACTIONS(1580),
    [aux_sym_switch_statement_token1] = ACTIONS(1580),
    [aux_sym_switch_block_token1] = ACTIONS(1580),
    [anon_sym_PLUS] = ACTIONS(1580),
    [anon_sym_DASH] = ACTIONS(1580),
    [anon_sym_TILDE] = ACTIONS(1578),
    [anon_sym_BANG] = ACTIONS(1578),
    [anon_sym_AT] = ACTIONS(1578),
    [aux_sym_clone_expression_token1] = ACTIONS(1580),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1580),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1580),
    [anon_sym_DASH_DASH] = ACTIONS(1578),
    [anon_sym_PLUS_PLUS] = ACTIONS(1578),
    [aux_sym__list_destructing_token1] = ACTIONS(1580),
    [anon_sym_LBRACK] = ACTIONS(1578),
    [anon_sym_self] = ACTIONS(1580),
    [anon_sym_parent] = ACTIONS(1580),
    [aux_sym__argument_name_token1] = ACTIONS(1580),
    [aux_sym__argument_name_token2] = ACTIONS(1580),
    [anon_sym_POUND_LBRACK] = ACTIONS(1578),
    [aux_sym_encapsed_string_token1] = ACTIONS(1578),
    [anon_sym_DQUOTE] = ACTIONS(1578),
    [aux_sym_string_token1] = ACTIONS(1578),
    [anon_sym_SQUOTE] = ACTIONS(1578),
    [anon_sym_LT_LT_LT] = ACTIONS(1578),
    [anon_sym_BQUOTE] = ACTIONS(1578),
    [anon_sym_DOLLAR] = ACTIONS(1578),
    [aux_sym_yield_expression_token1] = ACTIONS(1580),
    [aux_sym_include_expression_token1] = ACTIONS(1580),
    [aux_sym_include_once_expression_token1] = ACTIONS(1580),
    [aux_sym_require_expression_token1] = ACTIONS(1580),
    [aux_sym_require_once_expression_token1] = ACTIONS(1580),
    [sym_comment] = ACTIONS(3),
  },
  [578] = {
    [ts_builtin_sym_end] = ACTIONS(1582),
    [sym_name] = ACTIONS(1584),
    [anon_sym_QMARK_GT] = ACTIONS(1582),
    [anon_sym_SEMI] = ACTIONS(1582),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1584),
    [aux_sym_global_declaration_token1] = ACTIONS(1584),
    [aux_sym_namespace_definition_token1] = ACTIONS(1584),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1584),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1584),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1584),
    [anon_sym_BSLASH] = ACTIONS(1582),
    [anon_sym_LBRACE] = ACTIONS(1582),
    [anon_sym_RBRACE] = ACTIONS(1582),
    [aux_sym_trait_declaration_token1] = ACTIONS(1584),
    [aux_sym_interface_declaration_token1] = ACTIONS(1584),
    [aux_sym_enum_declaration_token1] = ACTIONS(1584),
    [aux_sym_enum_case_token1] = ACTIONS(1584),
    [aux_sym_class_declaration_token1] = ACTIONS(1584),
    [aux_sym_final_modifier_token1] = ACTIONS(1584),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1584),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1584),
    [sym_var_modifier] = ACTIONS(1584),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1584),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1584),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1584),
    [anon_sym_LPAREN2] = ACTIONS(1582),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1584),
    [aux_sym_cast_type_token1] = ACTIONS(1584),
    [aux_sym_echo_statement_token1] = ACTIONS(1584),
    [aux_sym_exit_statement_token1] = ACTIONS(1584),
    [anon_sym_unset] = ACTIONS(1584),
    [aux_sym_declare_statement_token1] = ACTIONS(1584),
    [aux_sym_declare_statement_token2] = ACTIONS(1584),
    [sym_float] = ACTIONS(1584),
    [aux_sym_try_statement_token1] = ACTIONS(1584),
    [aux_sym_goto_statement_token1] = ACTIONS(1584),
    [aux_sym_continue_statement_token1] = ACTIONS(1584),
    [aux_sym_break_statement_token1] = ACTIONS(1584),
    [sym_integer] = ACTIONS(1584),
    [aux_sym_return_statement_token1] = ACTIONS(1584),
    [aux_sym_throw_expression_token1] = ACTIONS(1584),
    [aux_sym_while_statement_token1] = ACTIONS(1584),
    [aux_sym_while_statement_token2] = ACTIONS(1584),
    [aux_sym_do_statement_token1] = ACTIONS(1584),
    [aux_sym_for_statement_token1] = ACTIONS(1584),
    [aux_sym_for_statement_token2] = ACTIONS(1584),
    [aux_sym_foreach_statement_token1] = ACTIONS(1584),
    [aux_sym_foreach_statement_token2] = ACTIONS(1584),
    [aux_sym_if_statement_token1] = ACTIONS(1584),
    [aux_sym_if_statement_token2] = ACTIONS(1584),
    [aux_sym_else_if_clause_token1] = ACTIONS(1584),
    [aux_sym_else_clause_token1] = ACTIONS(1584),
    [aux_sym_match_expression_token1] = ACTIONS(1584),
    [aux_sym_match_default_expression_token1] = ACTIONS(1584),
    [aux_sym_switch_statement_token1] = ACTIONS(1584),
    [aux_sym_switch_block_token1] = ACTIONS(1584),
    [anon_sym_PLUS] = ACTIONS(1584),
    [anon_sym_DASH] = ACTIONS(1584),
    [anon_sym_TILDE] = ACTIONS(1582),
    [anon_sym_BANG] = ACTIONS(1582),
    [anon_sym_AT] = ACTIONS(1582),
    [aux_sym_clone_expression_token1] = ACTIONS(1584),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1584),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1584),
    [anon_sym_DASH_DASH] = ACTIONS(1582),
    [anon_sym_PLUS_PLUS] = ACTIONS(1582),
    [aux_sym__list_destructing_token1] = ACTIONS(1584),
    [anon_sym_LBRACK] = ACTIONS(1582),
    [anon_sym_self] = ACTIONS(1584),
    [anon_sym_parent] = ACTIONS(1584),
    [aux_sym__argument_name_token1] = ACTIONS(1584),
    [aux_sym__argument_name_token2] = ACTIONS(1584),
    [anon_sym_POUND_LBRACK] = ACTIONS(1582),
    [aux_sym_encapsed_string_token1] = ACTIONS(1582),
    [anon_sym_DQUOTE] = ACTIONS(1582),
    [aux_sym_string_token1] = ACTIONS(1582),
    [anon_sym_SQUOTE] = ACTIONS(1582),
    [anon_sym_LT_LT_LT] = ACTIONS(1582),
    [anon_sym_BQUOTE] = ACTIONS(1582),
    [anon_sym_DOLLAR] = ACTIONS(1582),
    [aux_sym_yield_expression_token1] = ACTIONS(1584),
    [aux_sym_include_expression_token1] = ACTIONS(1584),
    [aux_sym_include_once_expression_token1] = ACTIONS(1584),
    [aux_sym_require_expression_token1] = ACTIONS(1584),
    [aux_sym_require_once_expression_token1] = ACTIONS(1584),
    [sym_comment] = ACTIONS(3),
  },
  [579] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1620),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2460),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_primary_expression] = STATE(981),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_variable] = STATE(746),
    [sym__variable_member_access_expression] = STATE(745),
    [sym_member_access_expression] = STATE(746),
    [sym__variable_nullsafe_member_access_expression] = STATE(739),
    [sym_nullsafe_member_access_expression] = STATE(746),
    [sym__variable_scoped_property_access_expression] = STATE(741),
    [sym_scoped_property_access_expression] = STATE(746),
    [sym_function_call_expression] = STATE(712),
    [sym__callable_expression] = STATE(2237),
    [sym_scoped_call_expression] = STATE(712),
    [sym__scope_resolution_qualifier] = STATE(2494),
    [sym_relative_scope] = STATE(2494),
    [sym_member_call_expression] = STATE(712),
    [sym_nullsafe_member_call_expression] = STATE(712),
    [sym__variable_subscript_expression] = STATE(733),
    [sym__dereferencable_subscript_expression] = STATE(712),
    [sym__dereferencable_expression] = STATE(1845),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(714),
    [sym__simple_variable] = STATE(714),
    [sym__new_variable] = STATE(733),
    [sym__callable_variable] = STATE(712),
    [sym_variable_name] = STATE(714),
    [sym__reserved_identifier] = STATE(1620),
    [sym__identifier] = STATE(1620),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(1586),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(693),
    [aux_sym_print_intrinsic_token1] = ACTIONS(703),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [anon_sym_LBRACK] = ACTIONS(1025),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(1588),
    [sym_comment] = ACTIONS(3),
  },
  [580] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1620),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2595),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_primary_expression] = STATE(981),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_variable] = STATE(746),
    [sym__variable_member_access_expression] = STATE(745),
    [sym_member_access_expression] = STATE(746),
    [sym__variable_nullsafe_member_access_expression] = STATE(739),
    [sym_nullsafe_member_access_expression] = STATE(746),
    [sym__variable_scoped_property_access_expression] = STATE(741),
    [sym_scoped_property_access_expression] = STATE(746),
    [sym_function_call_expression] = STATE(712),
    [sym__callable_expression] = STATE(2237),
    [sym_scoped_call_expression] = STATE(712),
    [sym__scope_resolution_qualifier] = STATE(2494),
    [sym_relative_scope] = STATE(2494),
    [sym_member_call_expression] = STATE(712),
    [sym_nullsafe_member_call_expression] = STATE(712),
    [sym__variable_subscript_expression] = STATE(733),
    [sym__dereferencable_subscript_expression] = STATE(712),
    [sym__dereferencable_expression] = STATE(1845),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(714),
    [sym__simple_variable] = STATE(714),
    [sym__new_variable] = STATE(733),
    [sym__callable_variable] = STATE(712),
    [sym_variable_name] = STATE(714),
    [sym__reserved_identifier] = STATE(1620),
    [sym__identifier] = STATE(1620),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(1586),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(643),
    [aux_sym_print_intrinsic_token1] = ACTIONS(655),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [anon_sym_LBRACK] = ACTIONS(1025),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(1588),
    [sym_comment] = ACTIONS(3),
  },
  [581] = {
    [sym_qualified_name] = STATE(783),
    [sym__name] = STATE(1620),
    [sym_namespace_name] = STATE(2646),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1006),
    [sym__anonymous_function_header] = STATE(2399),
    [sym__arrow_function_header] = STATE(2552),
    [sym_arrow_function] = STATE(1006),
    [sym_literal] = STATE(1006),
    [sym_throw_expression] = STATE(1006),
    [sym_primary_expression] = STATE(981),
    [sym_parenthesized_expression] = STATE(784),
    [sym_class_constant_access_expression] = STATE(836),
    [sym_print_intrinsic] = STATE(1006),
    [sym_object_creation_expression] = STATE(1006),
    [sym__new_non_dereferencable_expression] = STATE(983),
    [sym__new_dereferencable_expression] = STATE(785),
    [sym_update_expression] = STATE(1006),
    [sym_cast_variable] = STATE(746),
    [sym__variable_member_access_expression] = STATE(745),
    [sym_member_access_expression] = STATE(746),
    [sym__variable_nullsafe_member_access_expression] = STATE(739),
    [sym_nullsafe_member_access_expression] = STATE(746),
    [sym__variable_scoped_property_access_expression] = STATE(741),
    [sym_scoped_property_access_expression] = STATE(746),
    [sym_function_call_expression] = STATE(712),
    [sym__callable_expression] = STATE(2237),
    [sym_scoped_call_expression] = STATE(712),
    [sym__scope_resolution_qualifier] = STATE(2494),
    [sym_relative_scope] = STATE(2494),
    [sym_member_call_expression] = STATE(712),
    [sym_nullsafe_member_call_expression] = STATE(712),
    [sym__variable_subscript_expression] = STATE(733),
    [sym__dereferencable_subscript_expression] = STATE(712),
    [sym__dereferencable_expression] = STATE(1845),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(786),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(787),
    [sym_string] = STATE(787),
    [sym_heredoc] = STATE(787),
    [sym_nowdoc] = STATE(787),
    [sym_shell_command_expression] = STATE(1006),
    [sym_boolean] = STATE(1003),
    [sym_null] = STATE(1003),
    [sym__string] = STATE(787),
    [sym_dynamic_variable_name] = STATE(714),
    [sym__simple_variable] = STATE(714),
    [sym__new_variable] = STATE(733),
    [sym__callable_variable] = STATE(712),
    [sym_variable_name] = STATE(714),
    [sym__reserved_identifier] = STATE(1620),
    [sym__identifier] = STATE(1620),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(621),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(633),
    [anon_sym_LPAREN2] = ACTIONS(1586),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(639),
    [sym_float] = ACTIONS(641),
    [sym_integer] = ACTIONS(641),
    [aux_sym_throw_expression_token1] = ACTIONS(731),
    [aux_sym_print_intrinsic_token1] = ACTIONS(741),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(657),
    [anon_sym_DASH_DASH] = ACTIONS(659),
    [anon_sym_PLUS_PLUS] = ACTIONS(659),
    [anon_sym_LBRACK] = ACTIONS(1025),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(663),
    [aux_sym__argument_name_token2] = ACTIONS(665),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(667),
    [anon_sym_DQUOTE] = ACTIONS(667),
    [aux_sym_string_token1] = ACTIONS(669),
    [anon_sym_SQUOTE] = ACTIONS(669),
    [anon_sym_LT_LT_LT] = ACTIONS(671),
    [anon_sym_BQUOTE] = ACTIONS(673),
    [anon_sym_DOLLAR] = ACTIONS(1588),
    [sym_comment] = ACTIONS(3),
  },
  [582] = {
    [sym_qualified_name] = STATE(897),
    [sym__name] = STATE(1616),
    [sym_namespace_name] = STATE(2519),
    [sym_static_modifier] = STATE(2284),
    [sym_anonymous_function] = STATE(1096),
    [sym__anonymous_function_header] = STATE(2232),
    [sym__arrow_function_header] = STATE(2449),
    [sym_arrow_function] = STATE(1096),
    [sym_literal] = STATE(1096),
    [sym_throw_expression] = STATE(1096),
    [sym_primary_expression] = STATE(1108),
    [sym_parenthesized_expression] = STATE(941),
    [sym_class_constant_access_expression] = STATE(979),
    [sym_print_intrinsic] = STATE(1096),
    [sym_object_creation_expression] = STATE(1096),
    [sym__new_non_dereferencable_expression] = STATE(1098),
    [sym__new_dereferencable_expression] = STATE(971),
    [sym_update_expression] = STATE(1096),
    [sym_cast_variable] = STATE(870),
    [sym__variable_member_access_expression] = STATE(888),
    [sym_member_access_expression] = STATE(870),
    [sym__variable_nullsafe_member_access_expression] = STATE(887),
    [sym_nullsafe_member_access_expression] = STATE(870),
    [sym__variable_scoped_property_access_expression] = STATE(886),
    [sym_scoped_property_access_expression] = STATE(870),
    [sym_function_call_expression] = STATE(862),
    [sym__callable_expression] = STATE(2405),
    [sym_scoped_call_expression] = STATE(862),
    [sym__scope_resolution_qualifier] = STATE(2648),
    [sym_relative_scope] = STATE(2648),
    [sym_member_call_expression] = STATE(862),
    [sym_nullsafe_member_call_expression] = STATE(862),
    [sym__variable_subscript_expression] = STATE(875),
    [sym__dereferencable_subscript_expression] = STATE(862),
    [sym__dereferencable_expression] = STATE(1820),
    [sym__dereferencable_scalar] = STATE(1731),
    [sym_array_creation_expression] = STATE(892),
    [sym_attribute_group] = STATE(1325),
    [sym_attribute_list] = STATE(1824),
    [sym_encapsed_string] = STATE(894),
    [sym_string] = STATE(894),
    [sym_heredoc] = STATE(894),
    [sym_nowdoc] = STATE(894),
    [sym_shell_command_expression] = STATE(1096),
    [sym_boolean] = STATE(1095),
    [sym_null] = STATE(1095),
    [sym__string] = STATE(894),
    [sym_dynamic_variable_name] = STATE(853),
    [sym__simple_variable] = STATE(853),
    [sym__new_variable] = STATE(875),
    [sym__callable_variable] = STATE(862),
    [sym_variable_name] = STATE(853),
    [sym__reserved_identifier] = STATE(1616),
    [sym__identifier] = STATE(1616),
    [aux_sym_attribute_list_repeat1] = STATE(1325),
    [sym_name] = ACTIONS(719),
    [aux_sym_function_static_declaration_token1] = ACTIONS(627),
    [aux_sym__namespace_use_type_token1] = ACTIONS(631),
    [anon_sym_BSLASH] = ACTIONS(27),
    [anon_sym_LPAREN2] = ACTIONS(1590),
    [aux_sym__arrow_function_header_token1] = ACTIONS(51),
    [aux_sym_cast_type_token1] = ACTIONS(53),
    [sym_float] = ACTIONS(63),
    [sym_integer] = ACTIONS(63),
    [aux_sym_throw_expression_token1] = ACTIONS(75),
    [aux_sym_print_intrinsic_token1] = ACTIONS(99),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(101),
    [anon_sym_DASH_DASH] = ACTIONS(103),
    [anon_sym_PLUS_PLUS] = ACTIONS(103),
    [anon_sym_LBRACK] = ACTIONS(1027),
    [anon_sym_self] = ACTIONS(109),
    [anon_sym_parent] = ACTIONS(109),
    [aux_sym__argument_name_token1] = ACTIONS(111),
    [aux_sym__argument_name_token2] = ACTIONS(113),
    [anon_sym_POUND_LBRACK] = ACTIONS(115),
    [aux_sym_encapsed_string_token1] = ACTIONS(117),
    [anon_sym_DQUOTE] = ACTIONS(117),
    [aux_sym_string_token1] = ACTIONS(119),
    [anon_sym_SQUOTE] = ACTIONS(119),
    [anon_sym_LT_LT_LT] = ACTIONS(121),
    [anon_sym_BQUOTE] = ACTIONS(123),
    [anon_sym_DOLLAR] = ACTIONS(1592),
    [sym_comment] = ACTIONS(3),
  },
  [583] = {
    [sym_name] = ACTIONS(1594),
    [anon_sym_SEMI] = ACTIONS(1596),
    [aux_sym_function_static_declaration_token1] = ACTIONS(1594),
    [aux_sym_global_declaration_token1] = ACTIONS(1594),
    [aux_sym_namespace_definition_token1] = ACTIONS(1594),
    [aux_sym_namespace_use_declaration_token1] = ACTIONS(1594),
    [aux_sym__namespace_use_type_token1] = ACTIONS(1594),
    [aux_sym__namespace_use_type_token2] = ACTIONS(1594),
    [anon_sym_BSLASH] = ACTIONS(1596),
    [anon_sym_LBRACE] = ACTIONS(1596),
    [aux_sym_trait_declaration_token1] = ACTIONS(1594),
    [aux_sym_interface_declaration_token1] = ACTIONS(1594),
    [aux_sym_enum_declaration_token1] = ACTIONS(1594),
    [anon_sym_COLON] = ACTIONS(1596),
    [aux_sym_class_declaration_token1] = ACTIONS(1594),
    [aux_sym_final_modifier_token1] = ACTIONS(1594),
    [aux_sym_abstract_modifier_token1] = ACTIONS(1594),
    [aux_sym_readonly_modifier_token1] = ACTIONS(1594),
    [sym_var_modifier] = ACTIONS(1594),
    [aux_sym_visibility_modifier_token1] = ACTIONS(1594),
    [aux_sym_visibility_modifier_token2] = ACTIONS(1594),
    [aux_sym_visibility_modifier_token3] = ACTIONS(1594),
    [anon_sym_LPAREN2] = ACTIONS(1596),
    [aux_sym__arrow_function_header_token1] = ACTIONS(1594),
    [aux_sym_cast_type_token1] = ACTIONS(1594),
    [aux_sym_echo_statement_token1] = ACTIONS(1594),
    [aux_sym_exit_statement_token1] = ACTIONS(1594),
    [anon_sym_unset] = ACTIONS(1594),
    [aux_sym_declare_statement_token1] = ACTIONS(1594),
    [sym_float] = ACTIONS(1594),
    [aux_sym_try_statement_token1] = ACTIONS(1594),
    [aux_sym_goto_statement_token1] = ACTIONS(1594),
    [aux_sym_continue_statement_token1] = ACTIONS(1594),
    [aux_sym_break_statement_token1] = ACTIONS(1594),
    [sym_integer] = ACTIONS(1594),
    [aux_sym_return_statement_token1] = ACTIONS(1594),
    [aux_sym_throw_expression_token1] = ACTIONS(1594),
    [aux_sym_while_statement_token1] = ACTIONS(1594),
    [aux_sym_do_statement_token1] = ACTIONS(1594),
    [aux_sym_for_statement_token1] = ACTIONS(1594),
    [aux_sym_foreach_statement_token1] = ACTIONS(1594),
    [aux_sym_if_statement_token1] = ACTIONS(1594),
    [aux_sym_match_expression_token1] = ACTIONS(1594),
    [aux_sym_switch_statement_token1] = ACTIONS(1594),
    [anon_sym_PLUS] = ACTIONS(1594),
    [anon_sym_DASH] = ACTIONS(1594),
    [anon_sym_TILDE] = ACTIONS(1596),
    [anon_sym_BANG] = ACTIONS(1596),
    [anon_sym_AT] = ACTIONS(1596),
    [aux_sym_clone_expression_token1] = ACTIONS(1594),
    [aux_sym_print_intrinsic_token1] = ACTIONS(1594),
    [aux_sym__new_non_dereferencable_expression_token1] = ACTIONS(1594),
    [anon_sym_DASH_DASH] = ACTIONS(1596),
    [anon_sym_PLUS_PLUS] = ACTIONS(1596),
    [aux_sym__list_destructing_token1] = ACTIONS(1594),
    [anon_sym_LBRACK] = ACTIONS(1596),
    [anon_sym_self] = ACTIONS(1594),
    [anon_sym_parent] = ACTIONS(1594),
    [aux_sym__argument_name_token1] = ACTIONS(1594),
    [aux_sym__argument_name_token2] = ACTIONS(1594),
    [anon_sym_POUND_LBRACK] = ACTIONS(1596),
    [aux_sym_encapsed_string_token1] = ACTIONS(1596),
    [anon_sym_DQUOTE] = ACTIONS(1596),
    [aux_sym_string_token1] = ACTIONS(1596),
    [anon_sym_SQUOTE] = ACTIONS(1596),
    [anon_sym_LT_LT_LT] = ACTIONS(1596),
    [anon_sym_BQUOTE] = ACTIONS(1596),
    [anon_sym_DOLLAR] = ACTIONS(1596),
    [aux_sym_yield_expression_token1] = ACTIONS(1594),
    [aux_sym_include_expression_token1] = ACTIONS(1594),
    [aux_sym_include_once_expression_token1] = ACTIONS(1594),
    [aux_sym_require_expression_token1] = ACTIONS(1594),
    [aux_sym_require_once_expression_token1] = ACTIONS(1594),
    [sym_comment] = ACTIONS(3),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1600), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1598), 41,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      aux_sym_base_clause_token1,
      aux_sym_class_interface_clause_token1,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [70] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1606), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1604), 41,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      aux_sym_base_clause_token1,
      aux_sym_class_interface_clause_token1,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [140] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1610), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1608), 41,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      aux_sym_base_clause_token1,
      aux_sym_class_interface_clause_token1,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [210] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1614), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1612), 41,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      aux_sym_base_clause_token1,
      aux_sym_class_interface_clause_token1,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [280] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    STATE(607), 1,
      sym_arguments,
    ACTIONS(1618), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1616), 38,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [353] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    STATE(608), 1,
      sym_arguments,
    ACTIONS(1624), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1622), 38,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [426] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1628), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1626), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [494] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1632), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1630), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [562] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1636), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1634), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [630] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1640), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1638), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [698] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1646), 1,
      anon_sym_LPAREN2,
    ACTIONS(1644), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1642), 38,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [768] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1650), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1648), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [836] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1656), 1,
      anon_sym_LPAREN2,
    ACTIONS(1654), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1652), 38,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [906] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1660), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1658), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [974] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1662), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1656), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1042] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1666), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1664), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1110] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1670), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1668), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1178] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1674), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1672), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1246] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1678), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1676), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1314] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1684), 1,
      anon_sym_EQ,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1692), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 19,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 20,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [1392] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1696), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1694), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1460] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1700), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1698), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1528] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1704), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1702), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1596] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1708), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1706), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1664] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1712), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1710), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1732] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1716), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1714), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1800] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1720), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1718), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1868] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1724), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1722), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [1936] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1700), 21,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1698), 39,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2004] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1684), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1692), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 19,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 20,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [2079] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1684), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1692), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 19,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 20,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [2154] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1726), 1,
      anon_sym_LPAREN2,
    STATE(620), 1,
      sym_arguments,
    ACTIONS(1618), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1616), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2223] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1728), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1730), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 16,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 20,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [2298] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1726), 1,
      anon_sym_LPAREN2,
    STATE(622), 1,
      sym_arguments,
    ACTIONS(1624), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1622), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2367] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1660), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1658), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2431] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1704), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1702), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2495] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1708), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1706), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2559] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1728), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1730), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 16,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 20,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [2631] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1712), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1710), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2695] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1600), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1598), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2759] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1670), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1668), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [2823] = 27,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1732), 1,
      sym_name,
    ACTIONS(1734), 1,
      anon_sym_RBRACE,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(630), 1,
      aux_sym_use_list_repeat1,
    STATE(1643), 1,
      sym_class_constant_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(2128), 2,
      sym_use_instead_of_clause,
      sym_use_as_clause,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1808), 7,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
      sym__variable_subscript_expression,
      sym__dereferencable_expression,
      sym__new_variable,
    STATE(1731), 14,
      sym_parenthesized_expression,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
      sym__callable_variable,
  [2935] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1684), 1,
      anon_sym_EQ,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1740), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1692), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 16,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [3009] = 30,
    ACTIONS(105), 1,
      aux_sym__list_destructing_token1,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2319), 1,
      sym__list_destructing,
    STATE(2321), 1,
      sym_by_ref,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(1636), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1808), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1635), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1553), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [3127] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1678), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1676), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3191] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1614), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1612), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3255] = 27,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1744), 1,
      sym_name,
    ACTIONS(1750), 1,
      anon_sym_BSLASH,
    ACTIONS(1753), 1,
      anon_sym_RBRACE,
    ACTIONS(1755), 1,
      anon_sym_LPAREN2,
    ACTIONS(1758), 1,
      aux_sym_cast_type_token1,
    ACTIONS(1761), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1764), 1,
      anon_sym_LBRACK,
    ACTIONS(1773), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1776), 1,
      anon_sym_DOLLAR,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(630), 1,
      aux_sym_use_list_repeat1,
    STATE(1643), 1,
      sym_class_constant_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(1767), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(1770), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(2128), 2,
      sym_use_instead_of_clause,
      sym_use_as_clause,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(1747), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1808), 7,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
      sym__variable_subscript_expression,
      sym__dereferencable_expression,
      sym__new_variable,
    STATE(1731), 14,
      sym_parenthesized_expression,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
      sym__callable_variable,
  [3367] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1724), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1722), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3431] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1674), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1672), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3495] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1716), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1714), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3559] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1610), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1608), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3623] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1606), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1604), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3687] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1640), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1638), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3751] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1656), 1,
      anon_sym_LPAREN2,
    ACTIONS(1654), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1652), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [3817] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1779), 1,
      anon_sym_EQ,
    ACTIONS(1740), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1781), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 16,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [3891] = 30,
    ACTIONS(105), 1,
      aux_sym__list_destructing_token1,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2126), 1,
      sym__list_destructing,
    STATE(2137), 1,
      sym_by_ref,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(1521), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1808), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(1516), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1504), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [4009] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1646), 1,
      anon_sym_LPAREN2,
    ACTIONS(1644), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1642), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4075] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1662), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1656), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4139] = 27,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1732), 1,
      sym_name,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1783), 1,
      anon_sym_RBRACE,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(625), 1,
      aux_sym_use_list_repeat1,
    STATE(1643), 1,
      sym_class_constant_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(2128), 2,
      sym_use_instead_of_clause,
      sym_use_as_clause,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1808), 7,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
      sym__variable_subscript_expression,
      sym__dereferencable_expression,
      sym__new_variable,
    STATE(1731), 14,
      sym_parenthesized_expression,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
      sym__callable_variable,
  [4251] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1728), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1730), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 16,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 20,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [4323] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1785), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1787), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 16,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [4397] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1696), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1694), 36,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4461] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1785), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1787), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 16,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [4532] = 11,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1728), 1,
      anon_sym_EQ,
    ACTIONS(1789), 1,
      anon_sym_RPAREN2,
    STATE(2130), 1,
      aux_sym__list_destructing_repeat1,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1730), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [4611] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1632), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1630), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4674] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1636), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1634), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4737] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1700), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1698), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4800] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1720), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1718), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4863] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1650), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1648), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4926] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1700), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1698), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [4989] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1684), 1,
      anon_sym_EQ,
    ACTIONS(1740), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1692), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 16,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5060] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1684), 1,
      anon_sym_EQ,
    ACTIONS(1740), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1692), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 16,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5131] = 29,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(885), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2338), 1,
      sym__array_destructing,
    STATE(2367), 1,
      sym_by_ref,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(1588), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1808), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(1585), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1539), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [5246] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1628), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1626), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [5309] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1684), 1,
      anon_sym_EQ,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1692), 16,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_RBRACK,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5382] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1779), 1,
      anon_sym_EQ,
    ACTIONS(1740), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1781), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 16,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5453] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1791), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1793), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 15,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5526] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1666), 20,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1664), 35,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [5589] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1785), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1787), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 16,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5660] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1779), 1,
      anon_sym_EQ,
    ACTIONS(1740), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1781), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 16,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5731] = 10,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1728), 1,
      anon_sym_EQ,
    ACTIONS(1789), 1,
      anon_sym_RPAREN2,
    STATE(2130), 1,
      aux_sym__list_destructing_repeat1,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1730), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5807] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1785), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1795), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1787), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5881] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1791), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1793), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 15,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [5951] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1791), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1793), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1680), 15,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6021] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1728), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1798), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1730), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6095] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1684), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1692), 16,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_RBRACK,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6165] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1728), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1800), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1730), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6239] = 10,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1728), 1,
      anon_sym_EQ,
    ACTIONS(1789), 1,
      anon_sym_RPAREN2,
    STATE(2130), 1,
      aux_sym__list_destructing_repeat1,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1730), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6315] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1684), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1692), 16,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_RBRACK,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6385] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1785), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1795), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1787), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6456] = 27,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(1802), 1,
      anon_sym_RPAREN2,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(1577), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1808), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(1576), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1550), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [6565] = 27,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(1804), 1,
      anon_sym_RPAREN2,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(1577), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1808), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(1576), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1550), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [6674] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1785), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1795), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1787), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6745] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1728), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1800), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1730), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6816] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1728), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1798), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1730), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6887] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1728), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1798), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1730), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [6958] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1785), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1806), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 11,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1787), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [7031] = 27,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(1809), 1,
      anon_sym_RPAREN2,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(1577), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1808), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(1576), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1550), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [7140] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1728), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1800), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 12,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1730), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [7211] = 27,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(1811), 1,
      anon_sym_RPAREN2,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(1577), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1808), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(1576), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1550), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [7320] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1817), 1,
      anon_sym_BSLASH,
    STATE(2411), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1813), 22,
      anon_sym_AMP,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_clause_token1,
      anon_sym_COLON,
      aux_sym_class_interface_clause_token1,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_self,
      anon_sym_parent,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
      sym_name,
    ACTIONS(1815), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOLLAR,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [7384] = 26,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(1514), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1808), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(1561), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1493), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [7490] = 26,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(1820), 1,
      anon_sym_LPAREN2,
    ACTIONS(1822), 1,
      anon_sym_DOLLAR,
    STATE(903), 1,
      sym__variable_member_access_expression,
    STATE(912), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(913), 1,
      sym__variable_scoped_property_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2219), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(998), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1843), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2602), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(914), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(975), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1618), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(964), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [7596] = 26,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(1820), 1,
      anon_sym_LPAREN2,
    ACTIONS(1822), 1,
      anon_sym_DOLLAR,
    STATE(903), 1,
      sym__variable_member_access_expression,
    STATE(912), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(913), 1,
      sym__variable_scoped_property_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2219), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(997), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1843), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2602), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(914), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(995), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1618), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(937), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [7702] = 26,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(756), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1808), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(755), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(747), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [7808] = 26,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1592), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(1824), 1,
      anon_sym_LPAREN2,
    STATE(886), 1,
      sym__variable_scoped_property_access_expression,
    STATE(887), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(888), 1,
      sym__variable_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2405), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(881), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1820), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2450), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(853), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(880), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1616), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(858), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [7914] = 26,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(1826), 1,
      anon_sym_LPAREN2,
    ACTIONS(1828), 1,
      anon_sym_DOLLAR,
    STATE(757), 1,
      sym__variable_member_access_expression,
    STATE(758), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(759), 1,
      sym__variable_scoped_property_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2251), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(829), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1792), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2453), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(760), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(841), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1623), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(754), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [8020] = 26,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(1515), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1808), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(1549), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1506), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [8126] = 26,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(1826), 1,
      anon_sym_LPAREN2,
    ACTIONS(1828), 1,
      anon_sym_DOLLAR,
    STATE(757), 1,
      sym__variable_member_access_expression,
    STATE(758), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(759), 1,
      sym__variable_scoped_property_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2251), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(838), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1792), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2453), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(760), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(828), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1623), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(792), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [8232] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1785), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1806), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 11,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1787), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [8302] = 26,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(1826), 1,
      anon_sym_LPAREN2,
    ACTIONS(1828), 1,
      anon_sym_DOLLAR,
    STATE(757), 1,
      sym__variable_member_access_expression,
    STATE(758), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(759), 1,
      sym__variable_scoped_property_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2251), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(756), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1792), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2453), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(760), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(755), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1623), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(747), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [8408] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1785), 1,
      anon_sym_EQ,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1806), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1680), 11,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(1787), 13,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_DOT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1682), 19,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
  [8478] = 26,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(728), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1808), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(725), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(703), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [8584] = 26,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1736), 1,
      anon_sym_LPAREN2,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2410), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(1577), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1808), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2419), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(594), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(1576), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1591), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1550), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [8690] = 26,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1586), 1,
      anon_sym_LPAREN2,
    ACTIONS(1588), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    STATE(739), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(741), 1,
      sym__variable_scoped_property_access_expression,
    STATE(745), 1,
      sym__variable_member_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2237), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(728), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1845), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2494), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(714), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(725), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1620), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(703), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [8796] = 26,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(639), 1,
      aux_sym_cast_type_token1,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1025), 1,
      anon_sym_LBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1738), 1,
      aux_sym__new_non_dereferencable_expression_token1,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(1820), 1,
      anon_sym_LPAREN2,
    ACTIONS(1822), 1,
      anon_sym_DOLLAR,
    STATE(903), 1,
      sym__variable_member_access_expression,
    STATE(912), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(913), 1,
      sym__variable_scoped_property_access_expression,
    STATE(1730), 1,
      sym__new_dereferencable_expression,
    STATE(2219), 1,
      sym__callable_expression,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(987), 2,
      sym__variable_subscript_expression,
      sym__new_variable,
    STATE(1843), 2,
      sym_class_constant_access_expression,
      sym__dereferencable_expression,
    STATE(2602), 2,
      sym__scope_resolution_qualifier,
      sym_relative_scope,
    ACTIONS(109), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(914), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
    STATE(984), 4,
      sym_cast_variable,
      sym_member_access_expression,
      sym_nullsafe_member_access_expression,
      sym_scoped_property_access_expression,
    STATE(1618), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(926), 6,
      sym_function_call_expression,
      sym_scoped_call_expression,
      sym_member_call_expression,
      sym_nullsafe_member_call_expression,
      sym__dereferencable_subscript_expression,
      sym__callable_variable,
    STATE(1731), 8,
      sym_parenthesized_expression,
      sym__dereferencable_scalar,
      sym_array_creation_expression,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym__string,
  [8902] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1830), 22,
      anon_sym_AMP,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_clause_token1,
      anon_sym_COLON,
      aux_sym_class_interface_clause_token1,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_self,
      anon_sym_parent,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
      sym_name,
    ACTIONS(1832), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOLLAR,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [8960] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1834), 9,
      aux_sym_function_static_declaration_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    ACTIONS(1640), 11,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1638), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9020] = 24,
    ACTIONS(39), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(41), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(43), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1838), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(1840), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(1842), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(1846), 1,
      sym_var_modifier,
    ACTIONS(1848), 1,
      anon_sym_LPAREN2,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    ACTIONS(1852), 1,
      anon_sym_DOLLAR,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1546), 1,
      sym_property_element,
    STATE(1713), 1,
      sym_variable_name,
    STATE(1722), 1,
      sym__function_definition_header,
    STATE(2042), 1,
      sym_type,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(47), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1410), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [9117] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1684), 14,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1692), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9176] = 24,
    ACTIONS(39), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(41), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(43), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1838), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(1840), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(1842), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(1846), 1,
      sym_var_modifier,
    ACTIONS(1848), 1,
      anon_sym_LPAREN2,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    ACTIONS(1852), 1,
      anon_sym_DOLLAR,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1512), 1,
      sym_property_element,
    STATE(1703), 1,
      sym__function_definition_header,
    STATE(1713), 1,
      sym_variable_name,
    STATE(2039), 1,
      sym_type,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(47), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1410), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [9273] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1344), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      aux_sym_else_clause_token1,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1342), 34,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9328] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1854), 1,
      anon_sym_LPAREN2,
    STATE(710), 1,
      sym_arguments,
    ACTIONS(1618), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1616), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9387] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1854), 1,
      anon_sym_LPAREN2,
    STATE(711), 1,
      sym_arguments,
    ACTIONS(1624), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1622), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9446] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1188), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      aux_sym_else_clause_token1,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1186), 34,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9501] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1640), 13,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1638), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      aux_sym_use_instead_of_clause_token1,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9555] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1708), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1706), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9609] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1712), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1710), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9663] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 26,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9723] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1834), 9,
      aux_sym_function_static_declaration_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    ACTIONS(1640), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1638), 27,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9779] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1646), 1,
      anon_sym_LPAREN2,
    ACTIONS(1644), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1642), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9835] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1716), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1714), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9889] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1656), 1,
      anon_sym_LPAREN2,
    ACTIONS(1654), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1652), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9945] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1696), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1694), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [9999] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1662), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1656), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10053] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1600), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1598), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10107] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1674), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1672), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10161] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1614), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1612), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10215] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1724), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1722), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10269] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1606), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1604), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10323] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1670), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1668), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10377] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1684), 14,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1692), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10433] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1704), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1702), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10487] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1640), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1638), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10541] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1684), 14,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1692), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10597] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1610), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1608), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10651] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1856), 1,
      anon_sym_BSLASH,
    STATE(2291), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1859), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10711] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1660), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1658), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10765] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1678), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10819] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 26,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10876] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1650), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1648), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10929] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1628), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1626), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [10982] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1861), 1,
      anon_sym_LPAREN2,
    STATE(772), 1,
      sym_arguments,
    ACTIONS(1618), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1616), 31,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11039] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1722), 1,
      anon_sym_LPAREN2,
    ACTIONS(1865), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1863), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      aux_sym_use_instead_of_clause_token1,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11094] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1666), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11147] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1700), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1698), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11200] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1861), 1,
      anon_sym_LPAREN2,
    STATE(773), 1,
      sym_arguments,
    ACTIONS(1624), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1622), 31,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11257] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1636), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1634), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11310] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1632), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1630), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11363] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1656), 1,
      anon_sym_LPAREN2,
    ACTIONS(1869), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1867), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      aux_sym_use_instead_of_clause_token1,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11418] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1720), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1718), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11471] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1700), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1698), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11524] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1690), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 13,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 26,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11581] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1873), 13,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1871), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11638] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1877), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1875), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11690] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1881), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1879), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11742] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1885), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1883), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11794] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1889), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1887), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11846] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1893), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1891), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11898] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1897), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1895), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [11950] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1684), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1692), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12006] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1873), 13,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1871), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12060] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1873), 13,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1871), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12114] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1700), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1698), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12166] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1700), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1698), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12218] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1636), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1634), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12270] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1646), 1,
      anon_sym_LPAREN2,
    ACTIONS(1644), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1642), 31,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12324] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1674), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1672), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12376] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1610), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1608), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12428] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1650), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1648), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12480] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1656), 1,
      anon_sym_LPAREN2,
    ACTIONS(1654), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1652), 31,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12534] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1662), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1656), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12586] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1670), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1668), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12638] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1628), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1626), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12690] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1666), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12742] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1632), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1630), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12794] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1600), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1598), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12846] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1704), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1702), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12898] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1708), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1706), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [12950] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1712), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1710), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13002] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1716), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1714), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13054] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1720), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1718), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13106] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1614), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1612), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13158] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1724), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1722), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13210] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1606), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1604), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13262] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1660), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1658), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13314] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1678), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13366] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1696), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1694), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13418] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(1901), 1,
      anon_sym_COMMA,
    ACTIONS(1903), 1,
      anon_sym_BSLASH,
    ACTIONS(1905), 1,
      anon_sym_LPAREN2,
    ACTIONS(1907), 1,
      anon_sym_RPAREN2,
    ACTIONS(1909), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1911), 1,
      anon_sym_QMARK,
    ACTIONS(1913), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(1084), 1,
      sym_attribute_list,
    STATE(1181), 1,
      sym_visibility_modifier,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1656), 1,
      sym_type,
    STATE(2057), 1,
      sym_reference_modifier,
    STATE(2171), 1,
      sym_variable_name,
    STATE(2558), 1,
      sym_namespace_name,
    STATE(1300), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(47), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1880), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    STATE(1530), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [13512] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1917), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13566] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13622] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1923), 1,
      anon_sym_LPAREN2,
    ACTIONS(1925), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1921), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1919), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13678] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1927), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13732] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1927), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1931), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1929), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13786] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1935), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1933), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13838] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1939), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1937), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13890] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1945), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1947), 1,
      anon_sym_DASH_GT,
    ACTIONS(1949), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(1951), 1,
      anon_sym_LBRACK,
    ACTIONS(1943), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1941), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [13950] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1955), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1953), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14002] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1959), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1957), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14058] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1968), 1,
      anon_sym_EQ,
    ACTIONS(1965), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1963), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1961), 29,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14114] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1972), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1970), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14166] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1976), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1974), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14218] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(1903), 1,
      anon_sym_BSLASH,
    ACTIONS(1905), 1,
      anon_sym_LPAREN2,
    ACTIONS(1909), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1911), 1,
      anon_sym_QMARK,
    ACTIONS(1913), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(1978), 1,
      anon_sym_COMMA,
    ACTIONS(1980), 1,
      anon_sym_RPAREN2,
    STATE(1084), 1,
      sym_attribute_list,
    STATE(1181), 1,
      sym_visibility_modifier,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1656), 1,
      sym_type,
    STATE(2057), 1,
      sym_reference_modifier,
    STATE(2171), 1,
      sym_variable_name,
    STATE(2558), 1,
      sym_namespace_name,
    STATE(1300), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(47), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(2120), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    STATE(1530), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [14312] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1092), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      aux_sym_else_clause_token1,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1090), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14364] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1594), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1596), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14416] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1984), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1982), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14468] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1963), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1961), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14520] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1988), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1986), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14572] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1992), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1990), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14624] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1996), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1994), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14676] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2000), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1998), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14728] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2004), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2002), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14780] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2008), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2006), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14832] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2012), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2010), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14884] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2016), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2014), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14936] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2020), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2018), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [14988] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2024), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2022), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15040] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2028), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2026), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15092] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2032), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2030), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15144] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2036), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2034), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15196] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2040), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2038), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15248] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2044), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2042), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15300] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2048), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2046), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15352] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2052), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2050), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15404] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2056), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2054), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15456] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2060), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2058), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15508] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2064), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2062), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15560] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2068), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2066), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15612] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2072), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2070), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15664] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2076), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2074), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15716] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2080), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2078), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15768] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2084), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2082), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15820] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2088), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2086), 32,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15872] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1081), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      aux_sym_else_clause_token1,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1079), 33,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15924] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1959), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1957), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [15977] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1684), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1692), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16030] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1856), 1,
      anon_sym_BSLASH,
    STATE(2291), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(2090), 11,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1859), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16085] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2092), 1,
      anon_sym_LPAREN2,
    STATE(848), 1,
      sym_arguments,
    ACTIONS(1618), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1616), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16140] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2092), 1,
      anon_sym_LPAREN2,
    STATE(849), 1,
      sym_arguments,
    ACTIONS(1624), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1622), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16195] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(1903), 1,
      anon_sym_BSLASH,
    ACTIONS(1905), 1,
      anon_sym_LPAREN2,
    ACTIONS(1909), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1911), 1,
      anon_sym_QMARK,
    ACTIONS(1913), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(2094), 1,
      anon_sym_RPAREN2,
    STATE(1084), 1,
      sym_attribute_list,
    STATE(1181), 1,
      sym_visibility_modifier,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1656), 1,
      sym_type,
    STATE(2057), 1,
      sym_reference_modifier,
    STATE(2171), 1,
      sym_variable_name,
    STATE(2558), 1,
      sym_namespace_name,
    STATE(1300), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(47), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(2221), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    STATE(1530), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [16286] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(1903), 1,
      anon_sym_BSLASH,
    ACTIONS(1905), 1,
      anon_sym_LPAREN2,
    ACTIONS(1909), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1911), 1,
      anon_sym_QMARK,
    ACTIONS(1913), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(2096), 1,
      anon_sym_RPAREN2,
    STATE(1084), 1,
      sym_attribute_list,
    STATE(1181), 1,
      sym_visibility_modifier,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1656), 1,
      sym_type,
    STATE(2057), 1,
      sym_reference_modifier,
    STATE(2171), 1,
      sym_variable_name,
    STATE(2558), 1,
      sym_namespace_name,
    STATE(1300), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(47), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(2221), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    STATE(1530), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [16377] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(1903), 1,
      anon_sym_BSLASH,
    ACTIONS(1905), 1,
      anon_sym_LPAREN2,
    ACTIONS(1909), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1911), 1,
      anon_sym_QMARK,
    ACTIONS(1913), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(2098), 1,
      anon_sym_RPAREN2,
    STATE(1084), 1,
      sym_attribute_list,
    STATE(1181), 1,
      sym_visibility_modifier,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1656), 1,
      sym_type,
    STATE(2057), 1,
      sym_reference_modifier,
    STATE(2171), 1,
      sym_variable_name,
    STATE(2558), 1,
      sym_namespace_name,
    STATE(1300), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(47), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(2221), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    STATE(1530), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [16468] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16521] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1640), 12,
      anon_sym_AMP,
      anon_sym_EQ,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1638), 31,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      aux_sym_class_interface_clause_token1,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16572] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1959), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1957), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16625] = 23,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(1903), 1,
      anon_sym_BSLASH,
    ACTIONS(1905), 1,
      anon_sym_LPAREN2,
    ACTIONS(1909), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1911), 1,
      anon_sym_QMARK,
    ACTIONS(1913), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(2100), 1,
      anon_sym_RPAREN2,
    STATE(1084), 1,
      sym_attribute_list,
    STATE(1181), 1,
      sym_visibility_modifier,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1656), 1,
      sym_type,
    STATE(2057), 1,
      sym_reference_modifier,
    STATE(2171), 1,
      sym_variable_name,
    STATE(2558), 1,
      sym_namespace_name,
    STATE(1300), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(47), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(2221), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    STATE(1530), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [16716] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1817), 1,
      anon_sym_BSLASH,
    STATE(2411), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1813), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1815), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16771] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1684), 12,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1692), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16824] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1817), 1,
      anon_sym_BSLASH,
    STATE(2411), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1813), 11,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1815), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16879] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1662), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1656), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16929] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1670), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1668), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [16979] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1696), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1694), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17029] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1600), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1598), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17079] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1704), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1702), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17129] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1708), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1706), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17179] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1712), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1710), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17229] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1716), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1714), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17279] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(41), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(43), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOLLAR,
    ACTIONS(2102), 1,
      sym_name,
    ACTIONS(2104), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2106), 1,
      anon_sym_BSLASH,
    ACTIONS(2108), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2110), 1,
      sym_var_modifier,
    ACTIONS(2112), 1,
      anon_sym_LPAREN2,
    STATE(903), 1,
      sym__variable_member_access_expression,
    STATE(912), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(913), 1,
      sym__variable_scoped_property_access_expression,
    STATE(922), 1,
      sym_anonymous_class,
    STATE(1350), 1,
      sym_attribute_list,
    STATE(2478), 1,
      sym_namespace_name,
    ACTIONS(2114), 2,
      anon_sym_self,
      anon_sym_parent,
    STATE(1011), 2,
      sym_parenthesized_expression,
      sym__class_name_reference,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(47), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1009), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(923), 5,
      sym__variable_subscript_expression,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym__new_variable,
      sym_variable_name,
    STATE(1353), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [17373] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(41), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(43), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1828), 1,
      anon_sym_DOLLAR,
    ACTIONS(2116), 1,
      sym_name,
    ACTIONS(2118), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2120), 1,
      anon_sym_BSLASH,
    ACTIONS(2122), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2124), 1,
      sym_var_modifier,
    ACTIONS(2126), 1,
      anon_sym_LPAREN2,
    STATE(757), 1,
      sym__variable_member_access_expression,
    STATE(758), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(759), 1,
      sym__variable_scoped_property_access_expression,
    STATE(789), 1,
      sym_anonymous_class,
    STATE(1347), 1,
      sym_attribute_list,
    STATE(2565), 1,
      sym_namespace_name,
    ACTIONS(2128), 2,
      anon_sym_self,
      anon_sym_parent,
    STATE(877), 2,
      sym_parenthesized_expression,
      sym__class_name_reference,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(47), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(876), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(790), 5,
      sym__variable_subscript_expression,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym__new_variable,
      sym_variable_name,
    STATE(1348), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [17467] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1646), 1,
      anon_sym_LPAREN2,
    ACTIONS(1644), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1642), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17519] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2132), 18,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_AT,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_LBRACK,
      anon_sym_POUND_LBRACK,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
    ACTIONS(2130), 24,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__arrow_function_header_token1,
      aux_sym_cast_type_token1,
      sym_float,
      sym_integer,
      aux_sym_throw_expression_token1,
      aux_sym_match_expression_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_clone_expression_token1,
      aux_sym_print_intrinsic_token1,
      aux_sym__new_non_dereferencable_expression_token1,
      aux_sym__list_destructing_token1,
      anon_sym_self,
      anon_sym_parent,
      aux_sym__argument_name_token1,
      aux_sym__argument_name_token2,
      aux_sym_yield_expression_token1,
      aux_sym_include_expression_token1,
      aux_sym_include_once_expression_token1,
      aux_sym_require_expression_token1,
      aux_sym_require_once_expression_token1,
      sym_name,
  [17569] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1965), 1,
      anon_sym_RPAREN2,
    ACTIONS(1968), 1,
      anon_sym_EQ,
    ACTIONS(1963), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1961), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17623] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1660), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1658), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17673] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1614), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1612), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17723] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1684), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1692), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17777] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1724), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1722), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17827] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1606), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1604), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17877] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1678), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17927] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1740), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 23,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [17983] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1856), 1,
      anon_sym_BSLASH,
    STATE(2291), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1859), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18039] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1674), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1672), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18089] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(41), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(43), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(2118), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2122), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2124), 1,
      sym_var_modifier,
    ACTIONS(2134), 1,
      anon_sym_LPAREN2,
    STATE(592), 1,
      sym__variable_scoped_property_access_expression,
    STATE(605), 1,
      sym__variable_member_access_expression,
    STATE(612), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(789), 1,
      sym_anonymous_class,
    STATE(1347), 1,
      sym_attribute_list,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(2128), 2,
      anon_sym_self,
      anon_sym_parent,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    STATE(2288), 2,
      sym_parenthesized_expression,
      sym__class_name_reference,
    ACTIONS(47), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(2301), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(1729), 5,
      sym__variable_subscript_expression,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym__new_variable,
      sym_variable_name,
    STATE(1348), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [18183] = 22,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(1903), 1,
      anon_sym_BSLASH,
    ACTIONS(1905), 1,
      anon_sym_LPAREN2,
    ACTIONS(1909), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1911), 1,
      anon_sym_QMARK,
    ACTIONS(1913), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(1084), 1,
      sym_attribute_list,
    STATE(1181), 1,
      sym_visibility_modifier,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1656), 1,
      sym_type,
    STATE(2057), 1,
      sym_reference_modifier,
    STATE(2171), 1,
      sym_variable_name,
    STATE(2558), 1,
      sym_namespace_name,
    STATE(1300), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(47), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(2221), 3,
      sym_property_promotion_parameter,
      sym_simple_parameter,
      sym_variadic_parameter,
    STATE(1530), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [18271] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1610), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1608), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18321] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1656), 1,
      anon_sym_LPAREN2,
    ACTIONS(1654), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1652), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18373] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1640), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1638), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18423] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1740), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 23,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18476] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1830), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1832), 30,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18525] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1666), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18574] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2138), 17,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_AT,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_LBRACK,
      anon_sym_POUND_LBRACK,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
      anon_sym_LT_LT_LT,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
    ACTIONS(2136), 24,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__arrow_function_header_token1,
      aux_sym_cast_type_token1,
      sym_float,
      sym_integer,
      aux_sym_throw_expression_token1,
      aux_sym_match_expression_token1,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_clone_expression_token1,
      aux_sym_print_intrinsic_token1,
      aux_sym__new_non_dereferencable_expression_token1,
      aux_sym__list_destructing_token1,
      anon_sym_self,
      anon_sym_parent,
      aux_sym__argument_name_token1,
      aux_sym__argument_name_token2,
      aux_sym_yield_expression_token1,
      aux_sym_include_expression_token1,
      aux_sym_include_once_expression_token1,
      aux_sym_require_expression_token1,
      aux_sym_require_once_expression_token1,
      sym_name,
  [18623] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1830), 11,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1832), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18672] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1740), 2,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 23,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18725] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1945), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1943), 11,
      anon_sym_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1941), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18776] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1861), 1,
      anon_sym_LPAREN2,
    STATE(802), 1,
      sym_arguments,
    ACTIONS(2142), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2140), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18829] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1632), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1630), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18878] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1720), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1718), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18927] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1684), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1692), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [18978] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1684), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1692), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19029] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1650), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1648), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19078] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(950), 1,
      sym_arguments,
    ACTIONS(1618), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1616), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19131] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(951), 1,
      sym_arguments,
    ACTIONS(1624), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1622), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19184] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1968), 1,
      anon_sym_EQ,
    ACTIONS(1963), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1961), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19235] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1636), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1634), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19284] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1700), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1698), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19333] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1700), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1698), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19382] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1628), 12,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1626), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_COLON_COLON,
      anon_sym_DASH_DASH,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19431] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1889), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1887), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19479] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2020), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2018), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19527] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1927), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19577] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1640), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1638), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19625] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1927), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1931), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1929), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19675] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2028), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2026), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19723] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2044), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2042), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19771] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1917), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19821] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2000), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1998), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19869] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2048), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2046), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19917] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2052), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2050), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [19965] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2004), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2002), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20013] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2008), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2006), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20061] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1700), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1698), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20109] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1856), 1,
      anon_sym_BSLASH,
    ACTIONS(2146), 1,
      anon_sym_COLON,
    STATE(2291), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1859), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 21,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20165] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2056), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2054), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20213] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2060), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2058), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20261] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2064), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2062), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20309] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2068), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2066), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20357] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2072), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2070), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20405] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2032), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2030), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20453] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2036), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2034), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20501] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1700), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1698), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20549] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1636), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1634), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20597] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1646), 1,
      anon_sym_LPAREN2,
    ACTIONS(1644), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1642), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20647] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1897), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1895), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20695] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1963), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1961), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20743] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1656), 1,
      anon_sym_LPAREN2,
    ACTIONS(1869), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1867), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20793] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1968), 1,
      anon_sym_EQ,
    ACTIONS(1963), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1961), 28,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20843] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1722), 1,
      anon_sym_LPAREN2,
    ACTIONS(1865), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1863), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20893] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1660), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1658), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20941] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1678), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1676), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [20989] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1939), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1937), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21037] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2148), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2150), 1,
      anon_sym_DASH_GT,
    ACTIONS(2152), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(2154), 1,
      anon_sym_LBRACK,
    ACTIONS(1943), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1941), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21093] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1856), 1,
      anon_sym_BSLASH,
    ACTIONS(2156), 1,
      anon_sym_COLON,
    STATE(2291), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1859), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 21,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21149] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1881), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1879), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21197] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1959), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1957), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21249] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2040), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2038), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21297] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1856), 1,
      anon_sym_BSLASH,
    STATE(2291), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(2090), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1859), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21349] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1674), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1672), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21397] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1885), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1883), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21445] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1988), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1986), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21493] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2076), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2074), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21541] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1610), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1608), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21589] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2024), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2022), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21637] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1656), 1,
      anon_sym_LPAREN2,
    ACTIONS(1654), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1652), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21687] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1996), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1994), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21735] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1873), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1871), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21787] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1662), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1656), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21835] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1670), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1668), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21883] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1628), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1626), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21931] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [21983] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2012), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2010), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22031] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1188), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1186), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22079] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1666), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1664), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22127] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1972), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1970), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22175] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1632), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1630), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22223] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1696), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1694), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22271] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1600), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1598), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22319] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1704), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1702), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22367] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1708), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1706), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22415] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1712), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1710), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22463] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1976), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1974), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22511] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1594), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1596), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22559] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1984), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1982), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22607] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2080), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2078), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22655] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2084), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2082), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22703] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2088), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2086), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22751] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1935), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1933), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22799] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1955), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1953), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22847] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1344), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1342), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22895] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1716), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1714), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22943] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1877), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1875), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [22991] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1594), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1596), 30,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23039] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1684), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1692), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23091] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1720), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1718), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23139] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1893), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1891), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23187] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1614), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1612), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23235] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1724), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1722), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23283] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1606), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1604), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23331] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1992), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1990), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23379] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1923), 1,
      anon_sym_LPAREN2,
    ACTIONS(1925), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1921), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1919), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23431] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1856), 1,
      anon_sym_BSLASH,
    ACTIONS(2158), 1,
      anon_sym_COLON,
    STATE(2291), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1859), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 21,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23487] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2016), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2014), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23535] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1650), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1648), 29,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23583] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1684), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1692), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23632] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2160), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23679] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2166), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2164), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23726] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2170), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2168), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23773] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1682), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23822] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2174), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2172), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23869] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2178), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2176), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23916] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2182), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2180), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [23963] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1921), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1919), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24010] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1959), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1957), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24059] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2186), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2184), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24106] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1817), 1,
      anon_sym_BSLASH,
    STATE(2411), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1813), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1815), 27,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24157] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1959), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1957), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24206] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2192), 1,
      aux_sym_binary_expression_token1,
    ACTIONS(2190), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2188), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24255] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2196), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2194), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24302] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1968), 1,
      anon_sym_EQ,
    ACTIONS(2198), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1963), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1961), 25,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24353] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2202), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2200), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24400] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2206), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2204), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24447] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2210), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2208), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24494] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2214), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2212), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24541] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1873), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1871), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24590] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2218), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2216), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24637] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1873), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1871), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24686] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1684), 11,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1692), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24735] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2222), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2220), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24782] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2226), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2224), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24829] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2230), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2228), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24876] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2234), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2232), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24923] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1931), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1929), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [24970] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2190), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2188), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25017] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2238), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2236), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25064] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1682), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25111] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2242), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2240), 29,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25158] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1640), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1638), 28,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25204] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2148), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1943), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1941), 26,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25251] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1830), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1832), 27,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON_COLON,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25296] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(970), 1,
      sym_arguments,
    ACTIONS(2142), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2140), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25345] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2244), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [25427] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 24,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [25477] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2286), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [25553] = 21,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 8,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
  [25633] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1594), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1596), 26,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_LPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [25677] = 20,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 9,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [25755] = 17,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 1,
      anon_sym_QMARK,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 12,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
  [25827] = 16,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 1,
      anon_sym_QMARK,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 13,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [25897] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2288), 1,
      anon_sym_EQ_GT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2168), 6,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [25981] = 14,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2162), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 14,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [26047] = 11,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2162), 5,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2160), 18,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [26107] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 19,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [26163] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 22,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [26215] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 20,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_DOT,
  [26269] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2290), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [26351] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2292), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [26433] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2294), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [26515] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2296), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [26597] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2298), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [26679] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2240), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [26761] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2162), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2160), 25,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [26807] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2300), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [26889] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2302), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [26971] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 1,
      anon_sym_QMARK,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [27047] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2304), 7,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [27129] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2306), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [27205] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2310), 1,
      anon_sym_QMARK,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2308), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [27281] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2312), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [27357] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2162), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2160), 25,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27403] = 13,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 3,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 14,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [27467] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2316), 1,
      anon_sym_QMARK,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2314), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [27543] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 10,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [27619] = 15,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2162), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 13,
      anon_sym_SEMI,
      aux_sym_namespace_use_clause_token1,
      anon_sym_RBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [27687] = 10,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2320), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2325), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2328), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2331), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2334), 1,
      sym_var_modifier,
    ACTIONS(2337), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    ACTIONS(2323), 4,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
    ACTIONS(2318), 16,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [27744] = 11,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2162), 5,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2160), 17,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [27803] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 18,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [27858] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2218), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2216), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27901] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2230), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2228), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [27944] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 19,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_DOT,
  [27997] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2162), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2160), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28042] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2160), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28085] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2294), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [28166] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2306), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [28241] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2240), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [28322] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2304), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [28403] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1092), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1090), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28446] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2244), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [28527] = 13,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 3,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 13,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [28590] = 15,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2162), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 12,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [28657] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 23,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [28706] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 1,
      anon_sym_QMARK,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [28781] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2162), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2160), 24,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [28826] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [28901] = 21,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
  [28980] = 20,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [29057] = 17,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 1,
      anon_sym_QMARK,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 11,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
  [29128] = 16,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 1,
      anon_sym_QMARK,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 12,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [29197] = 14,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2162), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 13,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [29262] = 11,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2162), 5,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2160), 17,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [29321] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 18,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [29376] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 21,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [29427] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 19,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_DOT,
  [29480] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2162), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2160), 24,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [29525] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2296), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [29606] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2306), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [29681] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2312), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [29756] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2316), 1,
      anon_sym_QMARK,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2314), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [29831] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2286), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [29906] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2286), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [29981] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2310), 1,
      anon_sym_QMARK,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2308), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [30056] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2296), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [30137] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2420), 1,
      aux_sym_binary_expression_token1,
    ACTIONS(2190), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2188), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [30182] = 18,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(1903), 1,
      anon_sym_BSLASH,
    ACTIONS(1905), 1,
      anon_sym_LPAREN2,
    ACTIONS(1911), 1,
      anon_sym_QMARK,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(2422), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1192), 1,
      sym_visibility_modifier,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1684), 1,
      sym_type,
    STATE(1995), 1,
      sym_reference_modifier,
    STATE(2000), 1,
      sym_variable_name,
    STATE(2558), 1,
      sym_namespace_name,
    ACTIONS(47), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1530), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [30255] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2316), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2314), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [30330] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2292), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [30411] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2312), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [30486] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1081), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1079), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [30529] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2186), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2184), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [30572] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2174), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2172), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [30615] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2300), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [30696] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2234), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2232), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [30739] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2304), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [30820] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2182), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2180), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [30863] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1931), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1929), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [30906] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1682), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1680), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [30949] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2202), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2200), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [30992] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1921), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1919), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [31035] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2226), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2224), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [31078] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2166), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2164), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [31121] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2238), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2236), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [31164] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2298), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [31245] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2210), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2208), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [31288] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2240), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [31369] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2190), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2188), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [31412] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2170), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2168), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [31455] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2310), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2308), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [31530] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2178), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2176), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [31573] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2302), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [31654] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2242), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2240), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [31697] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2244), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [31778] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2206), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2204), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [31821] = 13,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2162), 3,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 13,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [31884] = 15,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2162), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 12,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [31951] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 23,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [32000] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [32075] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2162), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2160), 24,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [32120] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 9,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [32195] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2196), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2194), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [32238] = 21,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 7,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
  [32317] = 20,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 8,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [32394] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2294), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [32475] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2298), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [32556] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2424), 1,
      anon_sym_EQ_GT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2168), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [32639] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2290), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [32720] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2292), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [32801] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2300), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [32882] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2302), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      aux_sym_binary_expression_token1,
  [32963] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2222), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2220), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [33006] = 17,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 11,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
  [33077] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2426), 1,
      anon_sym_EQ_GT,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2168), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      aux_sym_binary_expression_token1,
  [33160] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2214), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2212), 25,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [33203] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2290), 6,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
  [33284] = 16,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 1,
      anon_sym_QMARK,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 12,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [33353] = 14,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2162), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 13,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [33418] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 21,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [33469] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [33543] = 21,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 6,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
  [33621] = 20,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 7,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [33697] = 17,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 1,
      anon_sym_QMARK,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 10,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
  [33767] = 16,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 1,
      anon_sym_QMARK,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 11,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [33835] = 14,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2162), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 12,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [33899] = 11,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2162), 5,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2160), 16,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [33957] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 17,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
  [34011] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 20,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [34061] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 18,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_DOT,
  [34113] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2162), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2160), 23,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [34157] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2306), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [34231] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2312), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [34305] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2316), 1,
      anon_sym_QMARK,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2314), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [34379] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2304), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [34459] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2286), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [34533] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2296), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [34613] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2240), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [34693] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2310), 1,
      anon_sym_QMARK,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2308), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [34767] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2244), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [34847] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2294), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [34927] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2298), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [35007] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2468), 1,
      anon_sym_EQ_GT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2168), 4,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [35089] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2290), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [35169] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2292), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [35249] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2300), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [35329] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2302), 5,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
  [35409] = 13,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 3,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 12,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
  [35471] = 15,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2162), 2,
      anon_sym_QMARK,
      anon_sym_PIPE,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 11,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
  [35537] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2162), 8,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2160), 22,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
  [35585] = 19,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2162), 1,
      anon_sym_QMARK,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2160), 8,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      aux_sym_binary_expression_token1,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
  [35659] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2162), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2160), 23,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [35703] = 24,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    STATE(2394), 1,
      sym_property_hook_list,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2470), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [35786] = 24,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    STATE(2337), 1,
      sym_property_hook_list,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2474), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [35869] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
    ACTIONS(2476), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [35948] = 4,
    ACTIONS(983), 1,
      anon_sym_COLON,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2222), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2220), 22,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [35991] = 24,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    STATE(2414), 1,
      sym_property_hook_list,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2478), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [36074] = 24,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    STATE(2308), 1,
      sym_property_hook_list,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2480), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [36157] = 24,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    STATE(2217), 1,
      sym_property_hook_list,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2482), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [36240] = 4,
    ACTIONS(983), 1,
      anon_sym_COLON,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2226), 10,
      anon_sym_AMP,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2224), 22,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym_binary_expression_token1,
      anon_sym_QMARK_QMARK,
      anon_sym_STAR_STAR,
      aux_sym_binary_expression_token2,
      aux_sym_binary_expression_token3,
      aux_sym_binary_expression_token4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_CARET,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ_GT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT,
      anon_sym_PERCENT,
  [36283] = 24,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    STATE(2393), 1,
      sym_property_hook_list,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [36366] = 24,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    STATE(2296), 1,
      sym_property_hook_list,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2486), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [36449] = 24,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    STATE(2408), 1,
      sym_property_hook_list,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2488), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [36532] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    STATE(2037), 1,
      sym__semicolon,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2490), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [36612] = 16,
    ACTIONS(43), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1848), 1,
      anon_sym_LPAREN2,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(1236), 1,
      sym_readonly_modifier,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1859), 1,
      sym_type,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1657), 2,
      sym_variable_name,
      sym_by_ref,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1633), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [36678] = 14,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1848), 1,
      anon_sym_LPAREN2,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    ACTIONS(2492), 1,
      sym_name,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1529), 1,
      sym_type,
    STATE(1662), 1,
      sym_const_element,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(2480), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1410), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [36740] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2496), 1,
      anon_sym_COMMA,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2494), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [36820] = 14,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1848), 1,
      anon_sym_LPAREN2,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    ACTIONS(2492), 1,
      sym_name,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1562), 1,
      sym_type,
    STATE(1709), 1,
      sym_const_element,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(2480), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1410), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [36882] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2498), 1,
      anon_sym_EQ_GT,
    ACTIONS(2168), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [36962] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    STATE(560), 1,
      sym__semicolon,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2500), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [37042] = 14,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1848), 1,
      anon_sym_LPAREN2,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    ACTIONS(2492), 1,
      sym_name,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1552), 1,
      sym_type,
    STATE(1747), 1,
      sym_const_element,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(2480), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1410), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [37104] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    STATE(573), 1,
      sym__semicolon,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2502), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [37184] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2496), 1,
      anon_sym_COMMA,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2504), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [37264] = 14,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1848), 1,
      anon_sym_LPAREN2,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    ACTIONS(2492), 1,
      sym_name,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1535), 1,
      sym_type,
    STATE(1649), 1,
      sym_const_element,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(2480), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1410), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [37326] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    STATE(1704), 1,
      sym__semicolon,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2506), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [37406] = 16,
    ACTIONS(43), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1848), 1,
      anon_sym_LPAREN2,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(1241), 1,
      sym_readonly_modifier,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1784), 1,
      sym_type,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1687), 2,
      sym_variable_name,
      sym_by_ref,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1633), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [37472] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    STATE(571), 1,
      sym__semicolon,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2508), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [37552] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    STATE(572), 1,
      sym__semicolon,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2510), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [37632] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    STATE(1954), 1,
      sym__semicolon,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2512), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [37712] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    STATE(2033), 1,
      sym__semicolon,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2514), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [37792] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    STATE(2036), 1,
      sym__semicolon,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2516), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [37872] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    STATE(1383), 1,
      sym__semicolon,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2518), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [37952] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2520), 1,
      anon_sym_EQ_GT,
    ACTIONS(2168), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [38032] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    STATE(1388), 1,
      sym__semicolon,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2522), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [38112] = 24,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2524), 1,
      anon_sym_COMMA,
    ACTIONS(2526), 1,
      anon_sym_EQ_GT,
    STATE(2108), 1,
      aux_sym_match_condition_list_repeat1,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [38194] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2528), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [38272] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2342), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2346), 1,
      anon_sym_GT_EQ,
    ACTIONS(2350), 1,
      anon_sym_DOT,
    ACTIONS(2354), 1,
      anon_sym_PERCENT,
    ACTIONS(2356), 1,
      anon_sym_AMP,
    ACTIONS(2358), 1,
      anon_sym_QMARK,
    ACTIONS(2360), 1,
      anon_sym_PIPE,
    ACTIONS(2362), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2364), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2366), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2368), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2370), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2372), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2374), 1,
      anon_sym_CARET,
    ACTIONS(2340), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2348), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2352), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2376), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2344), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2530), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
    ACTIONS(2378), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [38350] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2532), 1,
      anon_sym_EQ_GT,
    ACTIONS(2168), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [38430] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2534), 1,
      anon_sym_EQ_GT,
    ACTIONS(2536), 1,
      anon_sym_RPAREN2,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [38509] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2538), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [38586] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2540), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [38663] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2542), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2545), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2548), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2551), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2554), 1,
      anon_sym_RBRACE,
    ACTIONS(2556), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2559), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2562), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2565), 1,
      sym_var_modifier,
    ACTIONS(2571), 1,
      anon_sym_POUND_LBRACK,
    STATE(1331), 1,
      sym_attribute_list,
    STATE(1396), 1,
      sym__class_const_declaration,
    STATE(1397), 1,
      sym__const_declaration,
    STATE(1653), 1,
      sym__function_definition_header,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2568), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1208), 5,
      sym__member_declaration,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
      aux_sym_declaration_list_repeat1,
    STATE(704), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [38734] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2574), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [38811] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2576), 2,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [38888] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2580), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2584), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2586), 1,
      anon_sym_RBRACE,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2594), 1,
      sym_var_modifier,
    STATE(1331), 1,
      sym_attribute_list,
    STATE(1396), 1,
      sym__class_const_declaration,
    STATE(1397), 1,
      sym__const_declaration,
    STATE(1653), 1,
      sym__function_definition_header,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1208), 5,
      sym__member_declaration,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
      aux_sym_declaration_list_repeat1,
    STATE(704), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [38959] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2598), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [39036] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2600), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [39113] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2602), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [39190] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2604), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [39267] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2580), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2584), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2594), 1,
      sym_var_modifier,
    ACTIONS(2606), 1,
      anon_sym_RBRACE,
    STATE(1331), 1,
      sym_attribute_list,
    STATE(1396), 1,
      sym__class_const_declaration,
    STATE(1397), 1,
      sym__const_declaration,
    STATE(1653), 1,
      sym__function_definition_header,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1208), 5,
      sym__member_declaration,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
      aux_sym_declaration_list_repeat1,
    STATE(704), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [39338] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2608), 2,
      anon_sym_SEMI,
      anon_sym_COLON,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [39415] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [39492] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2504), 1,
      anon_sym_SEMI,
    ACTIONS(2612), 1,
      anon_sym_COMMA,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [39571] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2494), 1,
      anon_sym_RPAREN2,
    ACTIONS(2614), 1,
      anon_sym_COMMA,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [39650] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2616), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [39727] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2618), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [39804] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2580), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2584), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2594), 1,
      sym_var_modifier,
    ACTIONS(2620), 1,
      anon_sym_RBRACE,
    STATE(1331), 1,
      sym_attribute_list,
    STATE(1396), 1,
      sym__class_const_declaration,
    STATE(1397), 1,
      sym__const_declaration,
    STATE(1653), 1,
      sym__function_definition_header,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1208), 5,
      sym__member_declaration,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
      aux_sym_declaration_list_repeat1,
    STATE(704), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [39875] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2622), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [39952] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2504), 1,
      anon_sym_RPAREN2,
    ACTIONS(2614), 1,
      anon_sym_COMMA,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [40031] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2624), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [40108] = 23,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2494), 1,
      anon_sym_SEMI,
    ACTIONS(2612), 1,
      anon_sym_COMMA,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [40187] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2580), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2584), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2594), 1,
      sym_var_modifier,
    ACTIONS(2626), 1,
      anon_sym_RBRACE,
    STATE(1331), 1,
      sym_attribute_list,
    STATE(1396), 1,
      sym__class_const_declaration,
    STATE(1397), 1,
      sym__const_declaration,
    STATE(1653), 1,
      sym__function_definition_header,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1211), 5,
      sym__member_declaration,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
      aux_sym_declaration_list_repeat1,
    STATE(704), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [40258] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2628), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [40335] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2428), 1,
      anon_sym_AMP,
    ACTIONS(2430), 1,
      anon_sym_QMARK,
    ACTIONS(2432), 1,
      anon_sym_PIPE,
    ACTIONS(2436), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2438), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2440), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2442), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2444), 1,
      anon_sym_CARET,
    ACTIONS(2452), 1,
      anon_sym_GT_EQ,
    ACTIONS(2456), 1,
      anon_sym_DOT,
    ACTIONS(2460), 1,
      anon_sym_PERCENT,
    ACTIONS(2462), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2464), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2466), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2446), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2454), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2458), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2630), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(2450), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2448), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [40412] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2580), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2584), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2594), 1,
      sym_var_modifier,
    ACTIONS(2632), 1,
      anon_sym_RBRACE,
    STATE(1331), 1,
      sym_attribute_list,
    STATE(1396), 1,
      sym__class_const_declaration,
    STATE(1397), 1,
      sym__const_declaration,
    STATE(1653), 1,
      sym__function_definition_header,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1223), 5,
      sym__member_declaration,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
      aux_sym_declaration_list_repeat1,
    STATE(704), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [40483] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2580), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2584), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2594), 1,
      sym_var_modifier,
    ACTIONS(2634), 1,
      anon_sym_RBRACE,
    STATE(1331), 1,
      sym_attribute_list,
    STATE(1396), 1,
      sym__class_const_declaration,
    STATE(1397), 1,
      sym__const_declaration,
    STATE(1653), 1,
      sym__function_definition_header,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1216), 5,
      sym__member_declaration,
      sym_property_declaration,
      sym_method_declaration,
      sym_use_declaration,
      aux_sym_declaration_list_repeat1,
    STATE(704), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [40554] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2640), 1,
      anon_sym_LPAREN,
    ACTIONS(2638), 4,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
    ACTIONS(2636), 25,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_LPAREN2,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [40594] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2642), 1,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [40670] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2644), 1,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [40746] = 14,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1848), 1,
      anon_sym_LPAREN2,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1842), 1,
      sym_type,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1678), 2,
      sym_variable_name,
      sym_by_ref,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1633), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [40806] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2646), 1,
      anon_sym_RPAREN2,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [40882] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2648), 1,
      anon_sym_EQ_GT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [40958] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2650), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41034] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2652), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41110] = 14,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1848), 1,
      anon_sym_LPAREN2,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1834), 1,
      sym_type,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1699), 2,
      sym_variable_name,
      sym_by_ref,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1633), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [41170] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2654), 1,
      anon_sym_RPAREN2,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41246] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2656), 1,
      anon_sym_RPAREN2,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41322] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2658), 1,
      anon_sym_RPAREN2,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41398] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2660), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41474] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2662), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41550] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2664), 1,
      anon_sym_COLON,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41626] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2666), 1,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41702] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2668), 1,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41778] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2670), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41854] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2672), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [41930] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2674), 1,
      aux_sym_namespace_use_clause_token1,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42006] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2676), 1,
      aux_sym_namespace_use_clause_token1,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42082] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2678), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42158] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2680), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42234] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2682), 1,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42310] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2684), 1,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42386] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2686), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42462] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2688), 1,
      anon_sym_RPAREN2,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42538] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2690), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42614] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2692), 1,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42690] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2694), 1,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42766] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2696), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42842] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2698), 1,
      anon_sym_RPAREN2,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42918] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2700), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [42994] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2702), 1,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43070] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2704), 1,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43146] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43222] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2708), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43298] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2710), 1,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43374] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2380), 1,
      anon_sym_AMP,
    ACTIONS(2382), 1,
      anon_sym_QMARK,
    ACTIONS(2384), 1,
      anon_sym_PIPE,
    ACTIONS(2388), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2390), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2392), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2394), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2396), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2398), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2400), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2402), 1,
      anon_sym_CARET,
    ACTIONS(2410), 1,
      anon_sym_GT_EQ,
    ACTIONS(2414), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_PERCENT,
    ACTIONS(2712), 1,
      anon_sym_RBRACK,
    ACTIONS(2386), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2404), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2412), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2416), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2408), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2406), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43450] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2714), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43526] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2716), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43602] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2718), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43678] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2720), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43754] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2722), 1,
      anon_sym_EQ_GT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43830] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2724), 1,
      anon_sym_EQ_GT,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43906] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2726), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [43982] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2728), 1,
      anon_sym_COLON,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44058] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2730), 1,
      anon_sym_RBRACE,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44134] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2536), 1,
      anon_sym_RPAREN2,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44210] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2732), 1,
      anon_sym_COLON,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44286] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2734), 1,
      anon_sym_COLON,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44362] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2736), 1,
      aux_sym_namespace_use_clause_token1,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44438] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2738), 1,
      aux_sym_namespace_use_clause_token1,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44514] = 22,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2246), 1,
      anon_sym_AMP,
    ACTIONS(2248), 1,
      anon_sym_QMARK,
    ACTIONS(2250), 1,
      anon_sym_PIPE,
    ACTIONS(2254), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2256), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2258), 1,
      aux_sym_binary_expression_token2,
    ACTIONS(2260), 1,
      aux_sym_binary_expression_token3,
    ACTIONS(2262), 1,
      aux_sym_binary_expression_token4,
    ACTIONS(2264), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2266), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2268), 1,
      anon_sym_CARET,
    ACTIONS(2276), 1,
      anon_sym_GT_EQ,
    ACTIONS(2280), 1,
      anon_sym_DOT,
    ACTIONS(2284), 1,
      anon_sym_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_RPAREN2,
    ACTIONS(2252), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2270), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2278), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2282), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2274), 3,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_LT_EQ,
    ACTIONS(2272), 4,
      anon_sym_LT_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_LT_EQ_GT,
  [44590] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2744), 5,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
    ACTIONS(2742), 24,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [44627] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2580), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2746), 1,
      anon_sym_RBRACE,
    ACTIONS(2748), 1,
      aux_sym_enum_case_token1,
    ACTIONS(2750), 1,
      sym_var_modifier,
    STATE(1334), 1,
      sym_attribute_list,
    STATE(1653), 1,
      sym__function_definition_header,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1294), 5,
      sym__enum_member_declaration,
      sym_enum_case,
      sym_method_declaration,
      sym_use_declaration,
      aux_sym_enum_declaration_list_repeat1,
    STATE(1339), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [44692] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2754), 5,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
    ACTIONS(2752), 24,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [44729] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2580), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2748), 1,
      aux_sym_enum_case_token1,
    ACTIONS(2750), 1,
      sym_var_modifier,
    ACTIONS(2756), 1,
      anon_sym_RBRACE,
    STATE(1334), 1,
      sym_attribute_list,
    STATE(1653), 1,
      sym__function_definition_header,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1294), 5,
      sym__enum_member_declaration,
      sym_enum_case,
      sym_method_declaration,
      sym_use_declaration,
      aux_sym_enum_declaration_list_repeat1,
    STATE(1339), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [44794] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1834), 4,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
    ACTIONS(2758), 25,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [44831] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2762), 5,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
    ACTIONS(2760), 24,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [44868] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2580), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2748), 1,
      aux_sym_enum_case_token1,
    ACTIONS(2750), 1,
      sym_var_modifier,
    ACTIONS(2764), 1,
      anon_sym_RBRACE,
    STATE(1334), 1,
      sym_attribute_list,
    STATE(1653), 1,
      sym__function_definition_header,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1290), 5,
      sym__enum_member_declaration,
      sym_enum_case,
      sym_method_declaration,
      sym_use_declaration,
      aux_sym_enum_declaration_list_repeat1,
    STATE(1339), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [44933] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2766), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2769), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2772), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2775), 1,
      anon_sym_RBRACE,
    ACTIONS(2777), 1,
      aux_sym_enum_case_token1,
    ACTIONS(2780), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2783), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2786), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2789), 1,
      sym_var_modifier,
    ACTIONS(2795), 1,
      anon_sym_POUND_LBRACK,
    STATE(1334), 1,
      sym_attribute_list,
    STATE(1653), 1,
      sym__function_definition_header,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2792), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1294), 5,
      sym__enum_member_declaration,
      sym_enum_case,
      sym_method_declaration,
      sym_use_declaration,
      aux_sym_enum_declaration_list_repeat1,
    STATE(1339), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [44998] = 17,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2580), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2748), 1,
      aux_sym_enum_case_token1,
    ACTIONS(2750), 1,
      sym_var_modifier,
    ACTIONS(2798), 1,
      anon_sym_RBRACE,
    STATE(1334), 1,
      sym_attribute_list,
    STATE(1653), 1,
      sym__function_definition_header,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1288), 5,
      sym__enum_member_declaration,
      sym_enum_case,
      sym_method_declaration,
      sym_use_declaration,
      aux_sym_enum_declaration_list_repeat1,
    STATE(1339), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [45063] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2802), 4,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
    ACTIONS(2800), 24,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_string,
      anon_sym_int,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [45099] = 12,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1848), 1,
      anon_sym_LPAREN2,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    ACTIONS(2804), 1,
      sym_bottom_type,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2328), 1,
      sym_type,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1487), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1410), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [45152] = 12,
    ACTIONS(27), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2806), 1,
      sym_name,
    ACTIONS(2810), 1,
      anon_sym_LPAREN2,
    ACTIONS(2812), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      sym_bottom_type,
    STATE(1693), 1,
      sym_qualified_name,
    STATE(2173), 1,
      sym_type,
    STATE(2519), 1,
      sym_namespace_name,
    STATE(2024), 3,
      sym_union_type,
      sym_intersection_type,
      sym_disjunctive_normal_form_type,
    STATE(1495), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(2808), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [45205] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2820), 1,
      anon_sym_POUND_LBRACK,
    STATE(1299), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2818), 6,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
    ACTIONS(2816), 16,
      anon_sym_string,
      anon_sym_int,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [45242] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1913), 1,
      anon_sym_POUND_LBRACK,
    STATE(1299), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2825), 6,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_QMARK,
      anon_sym_DOLLAR,
    ACTIONS(2823), 16,
      anon_sym_string,
      anon_sym_int,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [45279] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2829), 7,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_QMARK,
      anon_sym_POUND_LBRACK,
      anon_sym_DOLLAR,
    ACTIONS(2827), 16,
      anon_sym_string,
      anon_sym_int,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [45310] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2833), 7,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_QMARK,
      anon_sym_POUND_LBRACK,
      anon_sym_DOLLAR,
    ACTIONS(2831), 16,
      anon_sym_string,
      anon_sym_int,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [45341] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2837), 7,
      anon_sym_AMP,
      anon_sym_BSLASH,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_QMARK,
      anon_sym_POUND_LBRACK,
      anon_sym_DOLLAR,
    ACTIONS(2835), 16,
      anon_sym_string,
      anon_sym_int,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
      sym_name,
  [45372] = 9,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2579), 1,
      sym_intersection_type,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1972), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [45414] = 9,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2573), 1,
      sym_intersection_type,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1972), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [45456] = 9,
    ACTIONS(27), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2806), 1,
      sym_name,
    ACTIONS(2812), 1,
      anon_sym_QMARK,
    ACTIONS(2839), 1,
      anon_sym_LPAREN2,
    STATE(1693), 1,
      sym_qualified_name,
    STATE(2519), 1,
      sym_namespace_name,
    STATE(1804), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(2808), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [45498] = 13,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1828), 1,
      anon_sym_DOLLAR,
    ACTIONS(2116), 1,
      sym_name,
    ACTIONS(2120), 1,
      anon_sym_BSLASH,
    ACTIONS(2126), 1,
      anon_sym_LPAREN2,
    STATE(757), 1,
      sym__variable_member_access_expression,
    STATE(758), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(759), 1,
      sym__variable_scoped_property_access_expression,
    STATE(2565), 1,
      sym_namespace_name,
    STATE(976), 2,
      sym_parenthesized_expression,
      sym__class_name_reference,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(876), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(790), 5,
      sym__variable_subscript_expression,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym__new_variable,
      sym_variable_name,
  [45548] = 13,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1822), 1,
      anon_sym_DOLLAR,
    ACTIONS(2102), 1,
      sym_name,
    ACTIONS(2106), 1,
      anon_sym_BSLASH,
    ACTIONS(2112), 1,
      anon_sym_LPAREN2,
    STATE(903), 1,
      sym__variable_member_access_expression,
    STATE(912), 1,
      sym__variable_nullsafe_member_access_expression,
    STATE(913), 1,
      sym__variable_scoped_property_access_expression,
    STATE(2478), 1,
      sym_namespace_name,
    STATE(1052), 2,
      sym_parenthesized_expression,
      sym__class_name_reference,
    ACTIONS(2114), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1009), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
    STATE(923), 5,
      sym__variable_subscript_expression,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym__new_variable,
      sym_variable_name,
  [45598] = 9,
    ACTIONS(27), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2806), 1,
      sym_name,
    ACTIONS(2812), 1,
      anon_sym_QMARK,
    ACTIONS(2839), 1,
      anon_sym_LPAREN2,
    STATE(1693), 1,
      sym_qualified_name,
    STATE(2519), 1,
      sym_namespace_name,
    STATE(1863), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(2808), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [45640] = 9,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    ACTIONS(2841), 1,
      anon_sym_LPAREN2,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1453), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [45682] = 16,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2843), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2845), 1,
      aux_sym_trait_declaration_token1,
    ACTIONS(2847), 1,
      aux_sym_interface_declaration_token1,
    ACTIONS(2849), 1,
      aux_sym_enum_declaration_token1,
    ACTIONS(2851), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2853), 1,
      sym_var_modifier,
    ACTIONS(2855), 1,
      aux_sym__arrow_function_header_token1,
    STATE(1447), 1,
      sym_static_modifier,
    STATE(2250), 1,
      sym__function_definition_header,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1349), 6,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [45738] = 9,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2509), 1,
      sym_intersection_type,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1972), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [45780] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1903), 1,
      anon_sym_BSLASH,
    ACTIONS(1911), 1,
      anon_sym_QMARK,
    ACTIONS(2841), 1,
      anon_sym_LPAREN2,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2558), 1,
      sym_namespace_name,
    STATE(1453), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [45822] = 9,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2470), 1,
      sym_intersection_type,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1972), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [45864] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1903), 1,
      anon_sym_BSLASH,
    ACTIONS(1911), 1,
      anon_sym_QMARK,
    ACTIONS(2841), 1,
      anon_sym_LPAREN2,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2558), 1,
      sym_namespace_name,
    STATE(1468), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [45906] = 16,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2843), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2855), 1,
      aux_sym__arrow_function_header_token1,
    ACTIONS(2857), 1,
      aux_sym_trait_declaration_token1,
    ACTIONS(2859), 1,
      aux_sym_interface_declaration_token1,
    ACTIONS(2861), 1,
      aux_sym_enum_declaration_token1,
    ACTIONS(2863), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2865), 1,
      sym_var_modifier,
    STATE(1447), 1,
      sym_static_modifier,
    STATE(2279), 1,
      sym__function_definition_header,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1342), 6,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [45962] = 9,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    ACTIONS(2841), 1,
      anon_sym_LPAREN2,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1468), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [46004] = 9,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2634), 1,
      sym_intersection_type,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1972), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [46046] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1903), 1,
      anon_sym_BSLASH,
    ACTIONS(1911), 1,
      anon_sym_QMARK,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2558), 1,
      sym_namespace_name,
    STATE(1476), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [46085] = 8,
    ACTIONS(27), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2806), 1,
      sym_name,
    ACTIONS(2812), 1,
      anon_sym_QMARK,
    STATE(1693), 1,
      sym_qualified_name,
    STATE(2519), 1,
      sym_namespace_name,
    STATE(1783), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(2808), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [46124] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1903), 1,
      anon_sym_BSLASH,
    ACTIONS(1911), 1,
      anon_sym_QMARK,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2558), 1,
      sym_namespace_name,
    STATE(1467), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [46163] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2818), 1,
      anon_sym_AMP,
    ACTIONS(2867), 1,
      anon_sym_POUND_LBRACK,
    STATE(1322), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2816), 17,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_trait_declaration_token1,
      aux_sym_interface_declaration_token1,
      aux_sym_enum_declaration_token1,
      aux_sym_enum_case_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
      sym_name,
  [46196] = 8,
    ACTIONS(27), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2806), 1,
      sym_name,
    ACTIONS(2812), 1,
      anon_sym_QMARK,
    STATE(1693), 1,
      sym_qualified_name,
    STATE(2519), 1,
      sym_namespace_name,
    STATE(1861), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(2808), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [46235] = 8,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1467), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [46274] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(2825), 1,
      anon_sym_AMP,
    STATE(1322), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    ACTIONS(2823), 17,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_trait_declaration_token1,
      aux_sym_interface_declaration_token1,
      aux_sym_enum_declaration_token1,
      aux_sym_enum_case_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
      sym_name,
  [46307] = 8,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1850), 1,
      anon_sym_QMARK,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1476), 4,
      sym__types,
      sym_named_type,
      sym_optional_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [46346] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1852), 1,
      anon_sym_DOLLAR,
    ACTIONS(2870), 1,
      anon_sym_COLON_COLON,
    STATE(1658), 1,
      sym_static_variable_declaration,
    STATE(1844), 1,
      sym_variable_name,
    ACTIONS(1638), 4,
      anon_sym_LPAREN2,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1834), 12,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
  [46382] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1852), 1,
      anon_sym_DOLLAR,
    ACTIONS(2870), 1,
      anon_sym_COLON_COLON,
    STATE(1751), 1,
      sym_static_variable_declaration,
    STATE(1844), 1,
      sym_variable_name,
    ACTIONS(1638), 4,
      anon_sym_LPAREN2,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
    ACTIONS(1834), 12,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
  [46418] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2833), 2,
      anon_sym_AMP,
      anon_sym_POUND_LBRACK,
    ACTIONS(2831), 17,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_trait_declaration_token1,
      aux_sym_interface_declaration_token1,
      aux_sym_enum_declaration_token1,
      aux_sym_enum_case_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
      sym_name,
  [46445] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2829), 2,
      anon_sym_AMP,
      anon_sym_POUND_LBRACK,
    ACTIONS(2827), 17,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_trait_declaration_token1,
      aux_sym_interface_declaration_token1,
      aux_sym_enum_declaration_token1,
      aux_sym_enum_case_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
      sym_name,
  [46472] = 12,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2584), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2872), 1,
      sym_var_modifier,
    STATE(1403), 1,
      sym__const_declaration,
    STATE(1696), 1,
      sym__function_definition_header,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(702), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [46517] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2837), 2,
      anon_sym_AMP,
      anon_sym_POUND_LBRACK,
    ACTIONS(2835), 17,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_trait_declaration_token1,
      aux_sym_interface_declaration_token1,
      aux_sym_enum_declaration_token1,
      aux_sym_enum_case_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      aux_sym__arrow_function_header_token1,
      sym_name,
  [46544] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    ACTIONS(1903), 1,
      anon_sym_BSLASH,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2558), 1,
      sym_namespace_name,
    STATE(1446), 2,
      sym_named_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [46578] = 11,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2874), 1,
      aux_sym_enum_case_token1,
    ACTIONS(2876), 1,
      sym_var_modifier,
    STATE(1696), 1,
      sym__function_definition_header,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1337), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [46620] = 7,
    ACTIONS(27), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2806), 1,
      sym_name,
    STATE(1693), 1,
      sym_qualified_name,
    STATE(2519), 1,
      sym_namespace_name,
    STATE(1706), 2,
      sym_named_type,
      sym_primitive_type,
    ACTIONS(2808), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [46654] = 7,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1836), 1,
      sym_name,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2646), 1,
      sym_namespace_name,
    STATE(1446), 2,
      sym_named_type,
      sym_primitive_type,
    ACTIONS(1844), 12,
      anon_sym_string,
      anon_sym_int,
      anon_sym_array,
      aux_sym_primitive_type_token1,
      anon_sym_iterable,
      anon_sym_bool,
      anon_sym_float,
      anon_sym_void,
      anon_sym_mixed,
      anon_sym_false,
      anon_sym_null,
      anon_sym_true,
  [46688] = 10,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2878), 1,
      sym_var_modifier,
    STATE(1722), 1,
      sym__function_definition_header,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [46727] = 10,
    ACTIONS(641), 1,
      sym_integer,
    ACTIONS(671), 1,
      anon_sym_LT_LT_LT,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2880), 1,
      sym_float,
    ACTIONS(2882), 1,
      aux_sym__argument_name_token1,
    ACTIONS(2884), 1,
      aux_sym__argument_name_token2,
    STATE(2545), 1,
      sym_literal,
    ACTIONS(667), 2,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
    ACTIONS(669), 2,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
    STATE(1003), 7,
      sym_encapsed_string,
      sym_string,
      sym_heredoc,
      sym_nowdoc,
      sym_boolean,
      sym_null,
      sym__string,
  [46766] = 10,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2582), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2878), 1,
      sym_var_modifier,
    STATE(1703), 1,
      sym__function_definition_header,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [46805] = 10,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2878), 1,
      sym_var_modifier,
    ACTIONS(2886), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2888), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [46844] = 10,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2878), 1,
      sym_var_modifier,
    ACTIONS(2890), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2892), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [46883] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2878), 1,
      sym_var_modifier,
    ACTIONS(2894), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [46919] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2878), 1,
      sym_var_modifier,
    ACTIONS(2890), 1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [46955] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2878), 1,
      sym_var_modifier,
    ACTIONS(2896), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [46991] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2878), 1,
      sym_var_modifier,
    ACTIONS(2898), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [47027] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2900), 1,
      anon_sym_LBRACE,
    ACTIONS(2904), 1,
      anon_sym_BSLASHu,
    ACTIONS(2906), 1,
      sym__new_line,
    ACTIONS(2909), 1,
      anon_sym_DOLLAR,
    ACTIONS(2911), 1,
      sym_heredoc_end,
    STATE(1352), 1,
      aux_sym_heredoc_body_repeat1,
    STATE(1457), 1,
      sym_variable_name,
    STATE(1532), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2902), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    STATE(1356), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_string_body_heredoc,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47067] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2913), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2915), 1,
      sym_var_modifier,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1344), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [47103] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2878), 1,
      sym_var_modifier,
    ACTIONS(2917), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [47139] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2878), 1,
      sym_var_modifier,
    ACTIONS(2919), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [47175] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2921), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2923), 1,
      sym_var_modifier,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1345), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [47211] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2925), 6,
      aux_sym_function_static_declaration_token1,
      aux_sym_cast_type_token1,
      aux_sym__new_non_dereferencable_expression_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(1753), 10,
      anon_sym_BSLASH,
      anon_sym_RBRACE,
      anon_sym_LPAREN2,
      anon_sym_LBRACK,
      aux_sym_encapsed_string_token1,
      anon_sym_DQUOTE,
      aux_sym_string_token1,
      anon_sym_SQUOTE,
      anon_sym_LT_LT_LT,
      anon_sym_DOLLAR,
  [47235] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2927), 1,
      anon_sym_LBRACE,
    ACTIONS(2933), 1,
      anon_sym_BSLASHu,
    ACTIONS(2936), 1,
      sym__new_line,
    ACTIONS(2939), 1,
      anon_sym_DOLLAR,
    ACTIONS(2942), 1,
      sym_heredoc_end,
    STATE(1352), 1,
      aux_sym_heredoc_body_repeat1,
    STATE(1457), 1,
      sym_variable_name,
    STATE(1532), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2930), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    STATE(1356), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_string_body_heredoc,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47275] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2588), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(2590), 1,
      aux_sym_abstract_modifier_token1,
    ACTIONS(2592), 1,
      aux_sym_readonly_modifier_token1,
    ACTIONS(2878), 1,
      sym_var_modifier,
    ACTIONS(2944), 1,
      aux_sym_class_declaration_token1,
    ACTIONS(2596), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
    STATE(1045), 7,
      sym_final_modifier,
      sym_abstract_modifier,
      sym_readonly_modifier,
      sym__modifier,
      sym_static_modifier,
      sym_visibility_modifier,
      aux_sym_class_declaration_repeat1,
  [47311] = 10,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2900), 1,
      anon_sym_LBRACE,
    ACTIONS(2904), 1,
      anon_sym_BSLASHu,
    ACTIONS(2909), 1,
      anon_sym_DOLLAR,
    ACTIONS(2946), 1,
      sym__new_line,
    STATE(1346), 1,
      aux_sym_heredoc_body_repeat1,
    STATE(1457), 1,
      sym_variable_name,
    STATE(1532), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2902), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    STATE(1356), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_string_body_heredoc,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47348] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2948), 1,
      anon_sym_LBRACE,
    ACTIONS(2954), 1,
      anon_sym_BSLASHu,
    ACTIONS(2959), 1,
      anon_sym_DOLLAR,
    STATE(1457), 1,
      sym_variable_name,
    STATE(1532), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2951), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    ACTIONS(2957), 2,
      sym_heredoc_end,
      sym__new_line,
    STATE(1355), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_string_body_heredoc,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47383] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2900), 1,
      anon_sym_LBRACE,
    ACTIONS(2909), 1,
      anon_sym_DOLLAR,
    ACTIONS(2964), 1,
      anon_sym_BSLASHu,
    STATE(1457), 1,
      sym_variable_name,
    STATE(1532), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2962), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    ACTIONS(2966), 2,
      sym_heredoc_end,
      sym__new_line,
    STATE(1355), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_string_body_heredoc,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47418] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2900), 1,
      anon_sym_LBRACE,
    ACTIONS(2909), 1,
      anon_sym_DOLLAR,
    ACTIONS(2964), 1,
      anon_sym_BSLASHu,
    STATE(1457), 1,
      sym_variable_name,
    STATE(1532), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2942), 2,
      sym_heredoc_end,
      sym__new_line,
    ACTIONS(2962), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    STATE(1355), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_string_body_heredoc,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47453] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2968), 1,
      anon_sym_LBRACE,
    ACTIONS(2972), 1,
      anon_sym_BSLASHu,
    ACTIONS(2974), 1,
      anon_sym_DQUOTE,
    ACTIONS(2976), 1,
      anon_sym_DOLLAR,
    STATE(1484), 1,
      sym_variable_name,
    STATE(1644), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2970), 2,
      sym_encapsed_string_chars,
      sym_escape_sequence,
    STATE(1372), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_string_body,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47487] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2978), 1,
      anon_sym_LBRACE,
    ACTIONS(2982), 1,
      anon_sym_BSLASHu,
    ACTIONS(2984), 1,
      anon_sym_BQUOTE,
    ACTIONS(2986), 1,
      anon_sym_DOLLAR,
    STATE(1491), 1,
      sym_variable_name,
    STATE(1595), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2980), 2,
      sym_execution_string_chars,
      sym_escape_sequence,
    STATE(1373), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_execution_operator_body,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47521] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2968), 1,
      anon_sym_LBRACE,
    ACTIONS(2976), 1,
      anon_sym_DOLLAR,
    ACTIONS(2990), 1,
      anon_sym_BSLASHu,
    ACTIONS(2992), 1,
      anon_sym_DQUOTE,
    STATE(1484), 1,
      sym_variable_name,
    STATE(1644), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2988), 2,
      sym_encapsed_string_chars,
      sym_escape_sequence,
    STATE(1362), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_string_body,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47555] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2978), 1,
      anon_sym_LBRACE,
    ACTIONS(2986), 1,
      anon_sym_DOLLAR,
    ACTIONS(2996), 1,
      anon_sym_BSLASHu,
    ACTIONS(2998), 1,
      anon_sym_BQUOTE,
    STATE(1491), 1,
      sym_variable_name,
    STATE(1595), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2994), 2,
      sym_execution_string_chars,
      sym_escape_sequence,
    STATE(1363), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_execution_operator_body,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47589] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2968), 1,
      anon_sym_LBRACE,
    ACTIONS(2972), 1,
      anon_sym_BSLASHu,
    ACTIONS(2976), 1,
      anon_sym_DOLLAR,
    ACTIONS(3000), 1,
      anon_sym_DQUOTE,
    STATE(1484), 1,
      sym_variable_name,
    STATE(1644), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2970), 2,
      sym_encapsed_string_chars,
      sym_escape_sequence,
    STATE(1372), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_string_body,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47623] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2978), 1,
      anon_sym_LBRACE,
    ACTIONS(2982), 1,
      anon_sym_BSLASHu,
    ACTIONS(2986), 1,
      anon_sym_DOLLAR,
    ACTIONS(3002), 1,
      anon_sym_BQUOTE,
    STATE(1491), 1,
      sym_variable_name,
    STATE(1595), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(2980), 2,
      sym_execution_string_chars,
      sym_escape_sequence,
    STATE(1373), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_execution_operator_body,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47657] = 9,
    ACTIONS(27), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2102), 1,
      sym_name,
    STATE(1689), 1,
      sym_namespace_use_clause,
    STATE(1996), 1,
      sym__namespace_use_group,
    STATE(2628), 1,
      sym_namespace_name,
    ACTIONS(3004), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2114), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1776), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [47691] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2968), 1,
      anon_sym_LBRACE,
    ACTIONS(2976), 1,
      anon_sym_DOLLAR,
    ACTIONS(3008), 1,
      anon_sym_BSLASHu,
    ACTIONS(3010), 1,
      anon_sym_DQUOTE,
    STATE(1484), 1,
      sym_variable_name,
    STATE(1644), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(3006), 2,
      sym_encapsed_string_chars,
      sym_escape_sequence,
    STATE(1358), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_string_body,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47725] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3012), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [47745] = 9,
    ACTIONS(27), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2102), 1,
      sym_name,
    STATE(1661), 1,
      sym_namespace_use_clause,
    STATE(2182), 1,
      sym__namespace_use_group,
    STATE(2628), 1,
      sym_namespace_name,
    ACTIONS(3004), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2114), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1776), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [47779] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3014), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [47799] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3016), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [47819] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3018), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [47839] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3020), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [47859] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3022), 1,
      anon_sym_LBRACE,
    ACTIONS(3028), 1,
      anon_sym_BSLASHu,
    ACTIONS(3031), 1,
      anon_sym_DQUOTE,
    ACTIONS(3033), 1,
      anon_sym_DOLLAR,
    STATE(1484), 1,
      sym_variable_name,
    STATE(1644), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(3025), 2,
      sym_encapsed_string_chars,
      sym_escape_sequence,
    STATE(1372), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_string_body,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47893] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3036), 1,
      anon_sym_LBRACE,
    ACTIONS(3042), 1,
      anon_sym_BSLASHu,
    ACTIONS(3045), 1,
      anon_sym_BQUOTE,
    ACTIONS(3047), 1,
      anon_sym_DOLLAR,
    STATE(1491), 1,
      sym_variable_name,
    STATE(1595), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(3039), 2,
      sym_execution_string_chars,
      sym_escape_sequence,
    STATE(1373), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_execution_operator_body,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [47927] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3050), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [47947] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3052), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [47967] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3054), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [47987] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3056), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48007] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2978), 1,
      anon_sym_LBRACE,
    ACTIONS(2986), 1,
      anon_sym_DOLLAR,
    ACTIONS(3060), 1,
      anon_sym_BSLASHu,
    ACTIONS(3062), 1,
      anon_sym_BQUOTE,
    STATE(1491), 1,
      sym_variable_name,
    STATE(1595), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(3058), 2,
      sym_execution_string_chars,
      sym_escape_sequence,
    STATE(1359), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_execution_operator_body,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [48041] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3064), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48061] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3066), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48081] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3068), 14,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48101] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3070), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48120] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3072), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48139] = 8,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1742), 1,
      sym_name,
    STATE(1976), 1,
      sym_namespace_use_clause,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(3074), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(2008), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [48170] = 8,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1742), 1,
      sym_name,
    STATE(2372), 1,
      sym_namespace_use_clause,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(3074), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(2008), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [48201] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3076), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48220] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1856), 1,
      anon_sym_BSLASH,
    STATE(2291), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1859), 11,
      anon_sym_COMMA,
      aux_sym_namespace_use_clause_token1,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      aux_sym_class_interface_clause_token1,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [48243] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3078), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48262] = 8,
    ACTIONS(27), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2102), 1,
      sym_name,
    STATE(2026), 1,
      sym_namespace_use_clause,
    STATE(2519), 1,
      sym_namespace_name,
    ACTIONS(3080), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
    ACTIONS(2114), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1776), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [48293] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3082), 1,
      anon_sym_LBRACE,
    ACTIONS(3086), 1,
      anon_sym_BSLASHu,
    ACTIONS(3088), 1,
      anon_sym_DOLLAR,
    STATE(1457), 1,
      sym_variable_name,
    STATE(1532), 1,
      sym__simple_string_member_access_expression,
    ACTIONS(3084), 2,
      sym_encapsed_string_chars_heredoc,
      sym_escape_sequence,
    STATE(1357), 6,
      sym__complex_string_part,
      sym__simple_string_subscript_expression,
      sym__simple_string_part,
      aux_sym__interpolated_string_body_heredoc,
      sym_dynamic_variable_name,
      sym__simple_variable,
  [48324] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3090), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48343] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1856), 1,
      anon_sym_BSLASH,
    STATE(2291), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3092), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3094), 7,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [48368] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3096), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48387] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3098), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48406] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3100), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48425] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3102), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48444] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3104), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48463] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3106), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48482] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3108), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48501] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3110), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48520] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3112), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      anon_sym_RBRACE,
      aux_sym_enum_case_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48539] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3114), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48558] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3116), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48577] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3118), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48596] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3120), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48615] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3122), 13,
      aux_sym_function_static_declaration_token1,
      aux_sym_namespace_use_declaration_token1,
      aux_sym__namespace_use_type_token1,
      aux_sym__namespace_use_type_token2,
      anon_sym_RBRACE,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
      anon_sym_POUND_LBRACK,
  [48634] = 7,
    ACTIONS(1588), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3124), 1,
      sym_name,
    ACTIONS(3128), 1,
      anon_sym_LBRACE,
    STATE(706), 1,
      sym__member_name,
    ACTIONS(3126), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(718), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [48662] = 7,
    ACTIONS(1592), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3130), 1,
      sym_name,
    ACTIONS(3134), 1,
      anon_sym_LBRACE,
    STATE(832), 1,
      sym__member_name,
    ACTIONS(3132), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(843), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [48690] = 8,
    ACTIONS(1592), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3136), 1,
      sym_name,
    ACTIONS(3140), 1,
      anon_sym_LBRACE,
    STATE(2413), 1,
      sym__member_name,
    STATE(917), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(3138), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(868), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [48720] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3147), 1,
      anon_sym_AMP,
    ACTIONS(3154), 1,
      anon_sym_PIPE,
    STATE(1464), 1,
      aux_sym_union_type_repeat1,
    STATE(1466), 1,
      aux_sym_intersection_type_repeat1,
    STATE(1473), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3149), 3,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
    ACTIONS(3142), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
  [48750] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3156), 1,
      sym_name,
    ACTIONS(3158), 1,
      anon_sym_RBRACE,
    STATE(1694), 1,
      sym_attribute_list,
    STATE(1962), 1,
      sym_final_modifier,
    STATE(2621), 1,
      sym_reference_modifier,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    STATE(1433), 2,
      sym_property_hook,
      aux_sym_property_hook_list_repeat1,
  [48786] = 8,
    ACTIONS(1588), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3160), 1,
      sym_name,
    ACTIONS(3164), 1,
      anon_sym_LBRACE,
    STATE(2245), 1,
      sym__member_name,
    STATE(743), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(3162), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(716), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [48816] = 7,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3166), 1,
      sym_name,
    ACTIONS(3170), 1,
      anon_sym_LBRACE,
    STATE(588), 1,
      sym__member_name,
    ACTIONS(3168), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(598), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [48844] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1822), 1,
      anon_sym_DOLLAR,
    ACTIONS(3172), 1,
      sym_name,
    ACTIONS(3174), 1,
      anon_sym_LBRACE,
    STATE(883), 1,
      sym__member_name,
    ACTIONS(3138), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(938), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [48872] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1822), 1,
      anon_sym_DOLLAR,
    ACTIONS(3172), 1,
      sym_name,
    ACTIONS(3174), 1,
      anon_sym_LBRACE,
    STATE(884), 1,
      sym__member_name,
    ACTIONS(3138), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(938), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [48900] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1822), 1,
      anon_sym_DOLLAR,
    ACTIONS(3172), 1,
      sym_name,
    ACTIONS(3174), 1,
      anon_sym_LBRACE,
    STATE(940), 1,
      sym__member_name,
    ACTIONS(3138), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(938), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [48928] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1822), 1,
      anon_sym_DOLLAR,
    ACTIONS(3172), 1,
      sym_name,
    ACTIONS(3174), 1,
      anon_sym_LBRACE,
    STATE(944), 1,
      sym__member_name,
    ACTIONS(3138), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(938), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [48956] = 7,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3166), 1,
      sym_name,
    ACTIONS(3170), 1,
      anon_sym_LBRACE,
    STATE(589), 1,
      sym__member_name,
    ACTIONS(3168), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(598), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [48984] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1828), 1,
      anon_sym_DOLLAR,
    ACTIONS(3160), 1,
      sym_name,
    ACTIONS(3164), 1,
      anon_sym_LBRACE,
    STATE(2255), 1,
      sym__member_name,
    STATE(743), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(3162), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(764), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [49014] = 7,
    ACTIONS(1588), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3124), 1,
      sym_name,
    ACTIONS(3128), 1,
      anon_sym_LBRACE,
    STATE(707), 1,
      sym__member_name,
    ACTIONS(3126), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(718), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [49042] = 8,
    ACTIONS(1592), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3160), 1,
      sym_name,
    ACTIONS(3164), 1,
      anon_sym_LBRACE,
    STATE(2413), 1,
      sym__member_name,
    STATE(743), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(3162), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(868), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [49072] = 8,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(3176), 1,
      anon_sym_RBRACK,
    STATE(2404), 1,
      sym_attribute,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1840), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [49102] = 7,
    ACTIONS(125), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3178), 1,
      sym_name,
    ACTIONS(3182), 1,
      anon_sym_LBRACE,
    STATE(615), 1,
      sym__member_name,
    ACTIONS(3180), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(641), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [49130] = 8,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(3184), 1,
      anon_sym_RBRACK,
    STATE(2404), 1,
      sym_attribute,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1840), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [49160] = 8,
    ACTIONS(125), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3136), 1,
      sym_name,
    ACTIONS(3140), 1,
      anon_sym_LBRACE,
    STATE(2278), 1,
      sym__member_name,
    STATE(917), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(3138), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(637), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [49190] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1828), 1,
      anon_sym_DOLLAR,
    ACTIONS(3186), 1,
      sym_name,
    ACTIONS(3188), 1,
      anon_sym_LBRACE,
    STATE(736), 1,
      sym__member_name,
    ACTIONS(3162), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(765), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [49218] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1828), 1,
      anon_sym_DOLLAR,
    ACTIONS(3186), 1,
      sym_name,
    ACTIONS(3188), 1,
      anon_sym_LBRACE,
    STATE(740), 1,
      sym__member_name,
    ACTIONS(3162), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(765), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [49246] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1828), 1,
      anon_sym_DOLLAR,
    ACTIONS(3186), 1,
      sym_name,
    ACTIONS(3188), 1,
      anon_sym_LBRACE,
    STATE(767), 1,
      sym__member_name,
    ACTIONS(3162), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(765), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [49274] = 8,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(3190), 1,
      anon_sym_RBRACK,
    STATE(2404), 1,
      sym_attribute,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1840), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [49304] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3156), 1,
      sym_name,
    ACTIONS(3192), 1,
      anon_sym_RBRACE,
    STATE(1694), 1,
      sym_attribute_list,
    STATE(1962), 1,
      sym_final_modifier,
    STATE(2621), 1,
      sym_reference_modifier,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    STATE(1411), 2,
      sym_property_hook,
      aux_sym_property_hook_list_repeat1,
  [49340] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1828), 1,
      anon_sym_DOLLAR,
    ACTIONS(3186), 1,
      sym_name,
    ACTIONS(3188), 1,
      anon_sym_LBRACE,
    STATE(768), 1,
      sym__member_name,
    ACTIONS(3162), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(765), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [49368] = 8,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1742), 1,
      sym_name,
    ACTIONS(3194), 1,
      anon_sym_RBRACK,
    STATE(2404), 1,
      sym_attribute,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1840), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [49398] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3196), 1,
      sym_name,
    ACTIONS(3199), 1,
      anon_sym_AMP,
    ACTIONS(3202), 1,
      anon_sym_RBRACE,
    ACTIONS(3204), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(3207), 1,
      anon_sym_POUND_LBRACK,
    STATE(1694), 1,
      sym_attribute_list,
    STATE(1962), 1,
      sym_final_modifier,
    STATE(2621), 1,
      sym_reference_modifier,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    STATE(1433), 2,
      sym_property_hook,
      aux_sym_property_hook_list_repeat1,
  [49434] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3212), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(3214), 1,
      aux_sym__arrow_function_header_token1,
    ACTIONS(3210), 10,
      aux_sym_function_static_declaration_token1,
      aux_sym__namespace_use_type_token2,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
  [49456] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3156), 1,
      sym_name,
    ACTIONS(3216), 1,
      anon_sym_RBRACE,
    STATE(1694), 1,
      sym_attribute_list,
    STATE(1962), 1,
      sym_final_modifier,
    STATE(2621), 1,
      sym_reference_modifier,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    STATE(1436), 2,
      sym_property_hook,
      aux_sym_property_hook_list_repeat1,
  [49492] = 11,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(39), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(115), 1,
      anon_sym_POUND_LBRACK,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3156), 1,
      sym_name,
    ACTIONS(3218), 1,
      anon_sym_RBRACE,
    STATE(1694), 1,
      sym_attribute_list,
    STATE(1962), 1,
      sym_final_modifier,
    STATE(2621), 1,
      sym_reference_modifier,
    STATE(1325), 2,
      sym_attribute_group,
      aux_sym_attribute_list_repeat1,
    STATE(1433), 2,
      sym_property_hook,
      aux_sym_property_hook_list_repeat1,
  [49528] = 7,
    ACTIONS(125), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3178), 1,
      sym_name,
    ACTIONS(3182), 1,
      anon_sym_LBRACE,
    STATE(617), 1,
      sym__member_name,
    ACTIONS(3180), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(641), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [49556] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1822), 1,
      anon_sym_DOLLAR,
    ACTIONS(3160), 1,
      sym_name,
    ACTIONS(3164), 1,
      anon_sym_LBRACE,
    STATE(2227), 1,
      sym__member_name,
    STATE(743), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(3162), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(935), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [49586] = 7,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3166), 1,
      sym_name,
    ACTIONS(3170), 1,
      anon_sym_LBRACE,
    STATE(588), 1,
      sym__member_name,
    ACTIONS(3162), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(598), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [49614] = 7,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3166), 1,
      sym_name,
    ACTIONS(3170), 1,
      anon_sym_LBRACE,
    STATE(589), 1,
      sym__member_name,
    ACTIONS(3162), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(598), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [49642] = 7,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3166), 1,
      sym_name,
    ACTIONS(3170), 1,
      anon_sym_LBRACE,
    STATE(590), 1,
      sym__member_name,
    ACTIONS(3162), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(598), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [49670] = 7,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3166), 1,
      sym_name,
    ACTIONS(3170), 1,
      anon_sym_LBRACE,
    STATE(599), 1,
      sym__member_name,
    ACTIONS(3162), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(598), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [49698] = 7,
    ACTIONS(1592), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3130), 1,
      sym_name,
    ACTIONS(3134), 1,
      anon_sym_LBRACE,
    STATE(831), 1,
      sym__member_name,
    ACTIONS(3132), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(843), 5,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
      sym__reserved_identifier,
      sym__identifier,
  [49726] = 8,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3160), 1,
      sym_name,
    ACTIONS(3164), 1,
      anon_sym_LBRACE,
    STATE(2285), 1,
      sym__member_name,
    STATE(743), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(3162), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(596), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [49756] = 7,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1742), 1,
      sym_name,
    STATE(2198), 1,
      sym_attribute,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1840), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [49783] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3220), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3222), 7,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [49802] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3224), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(3226), 1,
      aux_sym__arrow_function_header_token1,
    ACTIONS(3210), 9,
      aux_sym_function_static_declaration_token1,
      aux_sym_class_declaration_token1,
      aux_sym_final_modifier_token1,
      aux_sym_abstract_modifier_token1,
      aux_sym_readonly_modifier_token1,
      sym_var_modifier,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
  [49823] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3228), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3230), 7,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [49842] = 7,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1742), 1,
      sym_name,
    STATE(2160), 1,
      sym_attribute,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1840), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [49869] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3092), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3094), 7,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [49888] = 7,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1742), 1,
      sym_name,
    STATE(2404), 1,
      sym_attribute,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1840), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [49915] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3232), 1,
      sym_name,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1511), 1,
      sym_reference_modifier,
    STATE(1647), 1,
      sym_formal_parameters,
    STATE(2412), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [49943] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3236), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3239), 6,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [49961] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3232), 1,
      sym_name,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1497), 1,
      sym_reference_modifier,
    STATE(1752), 1,
      sym_formal_parameters,
    STATE(2412), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [49989] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3242), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3244), 6,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [50007] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3250), 1,
      anon_sym_PIPE,
    STATE(1456), 1,
      aux_sym_union_type_repeat1,
    ACTIONS(3246), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3248), 4,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
  [50029] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3255), 1,
      anon_sym_DASH_GT,
    ACTIONS(3257), 1,
      anon_sym_LBRACK,
    ACTIONS(3261), 1,
      sym_encapsed_string_chars_after_variable_heredoc,
    ACTIONS(3253), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(3259), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [50053] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3265), 1,
      anon_sym_AMP,
    STATE(1458), 1,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3263), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3268), 4,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOLLAR,
  [50075] = 6,
    ACTIONS(27), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2102), 1,
      sym_name,
    STATE(2519), 1,
      sym_namespace_name,
    ACTIONS(2114), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1846), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [50099] = 6,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1742), 1,
      sym_name,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1802), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [50123] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3274), 1,
      anon_sym_PIPE,
    STATE(1461), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3270), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3272), 4,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
  [50145] = 6,
    ACTIONS(27), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2102), 1,
      sym_name,
    STATE(2519), 1,
      sym_namespace_name,
    ACTIONS(2114), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1520), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [50169] = 6,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1742), 1,
      sym_name,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1879), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [50193] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3281), 1,
      anon_sym_PIPE,
    STATE(1456), 1,
      aux_sym_union_type_repeat1,
    ACTIONS(3277), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3279), 4,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
  [50215] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    STATE(1461), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3283), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3285), 4,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
  [50237] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3147), 1,
      anon_sym_AMP,
    STATE(1458), 1,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3289), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3291), 4,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOLLAR,
  [50259] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3246), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3248), 6,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [50277] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3270), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3272), 6,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [50295] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2156), 1,
      anon_sym_COLON,
    ACTIONS(2870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1834), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__arrow_function_header_token1,
    ACTIONS(1638), 6,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [50317] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    STATE(1465), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3293), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3295), 4,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
  [50339] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1658), 5,
      sym_encapsed_string_chars_heredoc,
      sym_encapsed_string_chars_after_variable_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
    ACTIONS(1660), 5,
      anon_sym_LBRACE,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
  [50357] = 6,
    ACTIONS(27), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2102), 1,
      sym_name,
    STATE(2517), 1,
      sym_namespace_name,
    ACTIONS(2114), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1816), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [50381] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    STATE(1461), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3297), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3299), 4,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOLLAR,
  [50403] = 6,
    ACTIONS(27), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2102), 1,
      sym_name,
    STATE(2519), 1,
      sym_namespace_name,
    ACTIONS(2114), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1816), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [50427] = 6,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1742), 1,
      sym_name,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(1979), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [50451] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3263), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3268), 6,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_RPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [50469] = 6,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1742), 1,
      sym_name,
    STATE(2646), 1,
      sym_namespace_name,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
    STATE(2051), 4,
      sym_qualified_name,
      sym__name,
      sym__reserved_identifier,
      sym__identifier,
  [50493] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    ACTIONS(3307), 1,
      anon_sym_LPAREN2,
    STATE(1536), 1,
      sym_formal_parameters,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3301), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1667), 2,
      sym__property_hook_body,
      sym__semicolon,
  [50520] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1660), 1,
      anon_sym_BSLASHu,
    ACTIONS(1658), 8,
      sym_encapsed_string_chars,
      sym_encapsed_string_chars_after_variable,
      anon_sym_LBRACE,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [50537] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    ACTIONS(3307), 1,
      anon_sym_LPAREN2,
    STATE(1544), 1,
      sym_formal_parameters,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3309), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1762), 2,
      sym__property_hook_body,
      sym__semicolon,
  [50564] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    ACTIONS(3307), 1,
      anon_sym_LPAREN2,
    STATE(1545), 1,
      sym_formal_parameters,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3311), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1763), 2,
      sym__property_hook_body,
      sym__semicolon,
  [50591] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    ACTIONS(3307), 1,
      anon_sym_LPAREN2,
    STATE(1547), 1,
      sym_formal_parameters,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3313), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1765), 2,
      sym__property_hook_body,
      sym__semicolon,
  [50618] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1660), 1,
      anon_sym_BSLASHu,
    ACTIONS(1658), 8,
      sym_execution_string_chars,
      sym_execution_string_chars_after_variable,
      anon_sym_LBRACE,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [50635] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3253), 1,
      anon_sym_BSLASHu,
    ACTIONS(3315), 1,
      anon_sym_DASH_GT,
    ACTIONS(3317), 1,
      anon_sym_LBRACK,
    ACTIONS(3319), 1,
      sym_encapsed_string_chars_after_variable,
    ACTIONS(3259), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [50658] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    ACTIONS(3307), 1,
      anon_sym_LPAREN2,
    STATE(1548), 1,
      sym_formal_parameters,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3321), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1766), 2,
      sym__property_hook_body,
      sym__semicolon,
  [50685] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    ACTIONS(3307), 1,
      anon_sym_LPAREN2,
    STATE(1557), 1,
      sym_formal_parameters,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3323), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1691), 2,
      sym__property_hook_body,
      sym__semicolon,
  [50712] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3325), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
    ACTIONS(3327), 5,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [50729] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    ACTIONS(3307), 1,
      anon_sym_LPAREN2,
    STATE(1537), 1,
      sym_formal_parameters,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3329), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1683), 2,
      sym__property_hook_body,
      sym__semicolon,
  [50756] = 8,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    ACTIONS(3307), 1,
      anon_sym_LPAREN2,
    STATE(1560), 1,
      sym_formal_parameters,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3331), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1736), 2,
      sym__property_hook_body,
      sym__semicolon,
  [50783] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1856), 1,
      anon_sym_BSLASH,
    ACTIONS(1859), 1,
      anon_sym_EQ,
    STATE(2291), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3094), 2,
      anon_sym_AMP,
      anon_sym_PIPE,
    ACTIONS(3092), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
  [50806] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3253), 1,
      anon_sym_BSLASHu,
    ACTIONS(3333), 1,
      anon_sym_DASH_GT,
    ACTIONS(3335), 1,
      anon_sym_LBRACK,
    ACTIONS(3337), 1,
      sym_execution_string_chars_after_variable,
    ACTIONS(3259), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [50829] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1834), 2,
      aux_sym__namespace_use_type_token1,
      aux_sym__arrow_function_header_token1,
    ACTIONS(1638), 6,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [50848] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(3339), 1,
      anon_sym_COMMA,
    ACTIONS(3341), 1,
      anon_sym_RPAREN2,
    STATE(1940), 1,
      aux_sym_unset_statement_repeat1,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [50870] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2156), 1,
      anon_sym_COLON,
    ACTIONS(2870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1638), 6,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [50888] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3343), 1,
      anon_sym_AMP,
    ACTIONS(3345), 1,
      anon_sym_PIPE,
    STATE(1677), 1,
      aux_sym_intersection_type_repeat1,
    STATE(1710), 1,
      aux_sym_union_type_repeat1,
    STATE(1712), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3149), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [50912] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(973), 1,
      sym_declaration_list,
    STATE(1641), 1,
      sym_arguments,
    STATE(1809), 1,
      sym_base_clause,
    STATE(2324), 1,
      sym_class_interface_clause,
  [50940] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    ACTIONS(3353), 1,
      sym_name,
    STATE(1665), 1,
      sym_formal_parameters,
    STATE(2247), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [50962] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(808), 1,
      sym_declaration_list,
    STATE(1609), 1,
      sym_arguments,
    STATE(1849), 1,
      sym_base_clause,
    STATE(2312), 1,
      sym_class_interface_clause,
  [50990] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(816), 1,
      sym_declaration_list,
    STATE(1610), 1,
      sym_arguments,
    STATE(1868), 1,
      sym_base_clause,
    STATE(2355), 1,
      sym_class_interface_clause,
  [51018] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(891), 1,
      sym_declaration_list,
    STATE(1640), 1,
      sym_arguments,
    STATE(1813), 1,
      sym_base_clause,
    STATE(2341), 1,
      sym_class_interface_clause,
  [51046] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1856), 1,
      anon_sym_BSLASH,
    ACTIONS(3357), 1,
      aux_sym_namespace_use_clause_token1,
    STATE(2291), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1859), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [51066] = 6,
    ACTIONS(1039), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3359), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(3362), 1,
      aux_sym_finally_clause_token1,
    ACTIONS(1037), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
    STATE(1502), 3,
      sym_catch_clause,
      sym_finally_clause,
      aux_sym_try_statement_repeat1,
  [51088] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(931), 1,
      sym_declaration_list,
    STATE(1621), 1,
      sym_arguments,
    STATE(1830), 1,
      sym_base_clause,
    STATE(2233), 1,
      sym_class_interface_clause,
  [51116] = 6,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(3365), 1,
      anon_sym_RPAREN2,
    STATE(2094), 1,
      aux_sym__list_destructing_repeat1,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [51138] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(801), 1,
      sym_declaration_list,
    STATE(1602), 1,
      sym_arguments,
    STATE(1797), 1,
      sym_base_clause,
    STATE(2273), 1,
      sym_class_interface_clause,
  [51166] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(3367), 1,
      anon_sym_COMMA,
    ACTIONS(3369), 1,
      anon_sym_RPAREN2,
    STATE(1913), 1,
      aux_sym_unset_statement_repeat1,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [51188] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(809), 1,
      sym_declaration_list,
    STATE(1646), 1,
      sym_arguments,
    STATE(1850), 1,
      sym_base_clause,
    STATE(2315), 1,
      sym_class_interface_clause,
  [51216] = 6,
    ACTIONS(1031), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3371), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(3373), 1,
      aux_sym_finally_clause_token1,
    ACTIONS(1029), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
    STATE(1502), 3,
      sym_catch_clause,
      sym_finally_clause,
      aux_sym_try_statement_repeat1,
  [51238] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3375), 1,
      sym_name,
    STATE(1572), 1,
      sym_reference_modifier,
    STATE(2417), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [51260] = 9,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(899), 1,
      sym_declaration_list,
    STATE(1584), 1,
      sym_arguments,
    STATE(1854), 1,
      sym_base_clause,
    STATE(2310), 1,
      sym_class_interface_clause,
  [51288] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    ACTIONS(3353), 1,
      sym_name,
    STATE(1664), 1,
      sym_formal_parameters,
    STATE(2247), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [51310] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3379), 1,
      anon_sym_COMMA,
    ACTIONS(3381), 1,
      anon_sym_LBRACE,
    STATE(1559), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(3377), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1393), 2,
      sym_property_hook_list,
      sym__semicolon,
  [51331] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1856), 1,
      anon_sym_BSLASH,
    STATE(2291), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3094), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [51348] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3339), 1,
      anon_sym_COMMA,
    ACTIONS(3341), 1,
      anon_sym_RPAREN2,
    STATE(1940), 1,
      aux_sym_unset_statement_repeat1,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [51367] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3367), 1,
      anon_sym_COMMA,
    ACTIONS(3369), 1,
      anon_sym_RPAREN2,
    STATE(1913), 1,
      aux_sym_unset_statement_repeat1,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [51386] = 5,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3365), 1,
      anon_sym_RPAREN2,
    STATE(2094), 1,
      aux_sym__list_destructing_repeat1,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [51405] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(3385), 1,
      anon_sym_DASH,
    ACTIONS(3383), 2,
      sym_integer,
      sym_name,
    STATE(2638), 3,
      sym__simple_string_subscript_unary_expression,
      sym__simple_string_array_access_argument,
      sym_variable_name,
  [51424] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3379), 1,
      anon_sym_COMMA,
    ACTIONS(3381), 1,
      anon_sym_LBRACE,
    STATE(1531), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(3387), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1400), 2,
      sym_property_hook_list,
      sym__semicolon,
  [51445] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1638), 6,
      anon_sym_COMMA,
      anon_sym_LPAREN2,
      anon_sym_RPAREN2,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [51460] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3391), 1,
      anon_sym_COMMA,
    ACTIONS(3393), 1,
      anon_sym_LBRACE,
    STATE(1524), 1,
      aux_sym_base_clause_repeat1,
    ACTIONS(3389), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1366), 2,
      sym_use_list,
      sym__semicolon,
  [51481] = 5,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3365), 1,
      anon_sym_RPAREN2,
    STATE(2094), 1,
      aux_sym__list_destructing_repeat1,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [51500] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3397), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3400), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3395), 2,
      anon_sym_RBRACE,
      aux_sym_switch_block_token1,
    STATE(1522), 3,
      sym_case_statement,
      sym_default_statement,
      aux_sym_switch_block_repeat1,
  [51519] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(3385), 1,
      anon_sym_DASH,
    ACTIONS(3403), 2,
      sym_integer,
      sym_name,
    STATE(2425), 3,
      sym__simple_string_subscript_unary_expression,
      sym__simple_string_array_access_argument,
      sym_variable_name,
  [51538] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3391), 1,
      anon_sym_COMMA,
    ACTIONS(3393), 1,
      anon_sym_LBRACE,
    STATE(1654), 1,
      aux_sym_base_clause_repeat1,
    ACTIONS(3405), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1375), 2,
      sym_use_list,
      sym__semicolon,
  [51559] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3379), 1,
      anon_sym_COMMA,
    ACTIONS(3381), 1,
      anon_sym_LBRACE,
    STATE(1745), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(3407), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1404), 2,
      sym_property_hook_list,
      sym__semicolon,
  [51580] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1678), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(1676), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [51595] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1696), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(1694), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [51610] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3379), 1,
      anon_sym_COMMA,
    ACTIONS(3381), 1,
      anon_sym_LBRACE,
    STATE(1555), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(3409), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1405), 2,
      sym_property_hook_list,
      sym__semicolon,
  [51631] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3411), 1,
      sym_name,
    STATE(1707), 1,
      sym_const_element,
    STATE(2480), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [51650] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3413), 1,
      anon_sym_AMP,
    ACTIONS(3419), 1,
      anon_sym_PIPE,
    STATE(1756), 1,
      aux_sym_union_type_repeat1,
    STATE(1757), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    STATE(1858), 1,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3149), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [51673] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3379), 1,
      anon_sym_COMMA,
    ACTIONS(3381), 1,
      anon_sym_LBRACE,
    STATE(1745), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(3421), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1406), 2,
      sym_property_hook_list,
      sym__semicolon,
  [51694] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3423), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(3425), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [51709] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2538), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1859), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [51724] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2630), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1859), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [51739] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3411), 1,
      sym_name,
    STATE(1680), 1,
      sym_const_element,
    STATE(2480), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [51758] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3427), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1682), 2,
      sym__property_hook_body,
      sym__semicolon,
  [51779] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3429), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1708), 2,
      sym__property_hook_body,
      sym__semicolon,
  [51800] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1660), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(1658), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [51815] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(3431), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [51832] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3433), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(2957), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [51847] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2624), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1859), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [51862] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3435), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(3437), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [51877] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3439), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(3441), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [51892] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3443), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1723), 2,
      sym__property_hook_body,
      sym__semicolon,
  [51913] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3445), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1738), 2,
      sym__property_hook_body,
      sym__semicolon,
  [51934] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3379), 1,
      anon_sym_COMMA,
    ACTIONS(3381), 1,
      anon_sym_LBRACE,
    STATE(1525), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(3447), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1398), 2,
      sym_property_hook_list,
      sym__semicolon,
  [51955] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3449), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1743), 2,
      sym__property_hook_body,
      sym__semicolon,
  [51976] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3451), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1675), 2,
      sym__property_hook_body,
      sym__semicolon,
  [51997] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3367), 1,
      anon_sym_COMMA,
    ACTIONS(3369), 1,
      anon_sym_RPAREN2,
    STATE(1913), 1,
      aux_sym_unset_statement_repeat1,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [52016] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(3453), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [52033] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3411), 1,
      sym_name,
    STATE(2129), 1,
      sym_const_element,
    STATE(2480), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [52052] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3411), 1,
      sym_name,
    STATE(1702), 1,
      sym_const_element,
    STATE(2480), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [52071] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(3455), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [52088] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3457), 3,
      anon_sym_LBRACE,
      anon_sym_BSLASHu,
      anon_sym_DOLLAR,
    ACTIONS(3459), 4,
      sym_encapsed_string_chars_heredoc,
      sym_heredoc_end,
      sym_escape_sequence,
      sym__new_line,
  [52103] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3379), 1,
      anon_sym_COMMA,
    ACTIONS(3381), 1,
      anon_sym_LBRACE,
    STATE(1745), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(3461), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1386), 2,
      sym_property_hook_list,
      sym__semicolon,
  [52124] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2138), 3,
      anon_sym_LPAREN2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
    ACTIONS(2136), 4,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
      sym_name,
  [52139] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3463), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1681), 2,
      sym__property_hook_body,
      sym__semicolon,
  [52160] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(3385), 1,
      anon_sym_DASH,
    ACTIONS(3465), 2,
      sym_integer,
      sym_name,
    STATE(2626), 3,
      sym__simple_string_subscript_unary_expression,
      sym__simple_string_array_access_argument,
      sym_variable_name,
  [52179] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3379), 1,
      anon_sym_COMMA,
    ACTIONS(3381), 1,
      anon_sym_LBRACE,
    STATE(1745), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(3467), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1399), 2,
      sym_property_hook_list,
      sym__semicolon,
  [52200] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3303), 1,
      anon_sym_LBRACE,
    ACTIONS(3305), 1,
      anon_sym_EQ_GT,
    STATE(1737), 1,
      sym_compound_statement,
    ACTIONS(3469), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    STATE(1761), 2,
      sym__property_hook_body,
      sym__semicolon,
  [52221] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3339), 1,
      anon_sym_COMMA,
    ACTIONS(3341), 1,
      anon_sym_RPAREN2,
    STATE(1940), 1,
      aux_sym_unset_statement_repeat1,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [52240] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3411), 1,
      sym_name,
    STATE(1672), 1,
      sym_const_element,
    STATE(2480), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [52259] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2604), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1859), 5,
      anon_sym_LPAREN2,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [52274] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(1774), 1,
      sym_base_clause,
    STATE(2090), 1,
      sym_declaration_list,
    STATE(2363), 1,
      sym_class_interface_clause,
  [52296] = 7,
    ACTIONS(1047), 1,
      aux_sym_while_statement_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3471), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3473), 1,
      aux_sym_else_clause_token1,
    STATE(1733), 1,
      aux_sym_if_statement_repeat1,
    STATE(2023), 1,
      sym_else_clause,
    STATE(2092), 1,
      sym_else_if_clause,
  [52318] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3435), 1,
      anon_sym_BSLASHu,
    ACTIONS(3437), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [52332] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3435), 1,
      anon_sym_BSLASHu,
    ACTIONS(3437), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [52346] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_BSLASHu,
    ACTIONS(3441), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [52360] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3439), 1,
      anon_sym_BSLASHu,
    ACTIONS(3441), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [52374] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3457), 1,
      anon_sym_BSLASHu,
    ACTIONS(3459), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [52388] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(564), 1,
      sym_declaration_list,
    STATE(1782), 1,
      sym_base_clause,
    STATE(2385), 1,
      sym_class_interface_clause,
  [52410] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3477), 1,
      sym_name,
    STATE(2254), 2,
      sym__reserved_identifier,
      sym__identifier,
    ACTIONS(2128), 3,
      aux_sym_function_static_declaration_token1,
      anon_sym_self,
      anon_sym_parent,
  [52426] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1817), 1,
      anon_sym_BSLASH,
    STATE(2411), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(1815), 4,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [52442] = 7,
    ACTIONS(1057), 1,
      aux_sym_while_statement_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3479), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3482), 1,
      aux_sym_else_clause_token1,
    STATE(1575), 1,
      aux_sym_if_statement_repeat1,
    STATE(2092), 1,
      sym_else_if_clause,
    STATE(2111), 1,
      sym_else_clause,
  [52464] = 7,
    ACTIONS(1047), 1,
      aux_sym_while_statement_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3485), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3488), 1,
      aux_sym_else_clause_token1,
    STATE(1733), 1,
      aux_sym_if_statement_repeat1,
    STATE(2023), 1,
      sym_else_clause,
    STATE(2092), 1,
      sym_else_if_clause,
  [52486] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3453), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [52500] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3453), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [52514] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1678), 1,
      anon_sym_BSLASHu,
    ACTIONS(1676), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [52528] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1726), 1,
      anon_sym_LPAREN2,
    STATE(632), 1,
      sym_arguments,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [52544] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1696), 1,
      anon_sym_BSLASHu,
    ACTIONS(1694), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [52558] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1678), 1,
      anon_sym_BSLASHu,
    ACTIONS(1676), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [52572] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3491), 1,
      anon_sym_SQUOTE,
    ACTIONS(3493), 1,
      aux_sym_string_token2,
    ACTIONS(3495), 1,
      aux_sym_string_content_token1,
    STATE(1866), 1,
      aux_sym_string_content_repeat1,
    STATE(1645), 2,
      sym_string_content,
      aux_sym_string_repeat1,
  [52592] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1696), 1,
      anon_sym_BSLASHu,
    ACTIONS(1694), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [52606] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(956), 1,
      sym_declaration_list,
    STATE(1803), 1,
      sym_base_clause,
    STATE(2314), 1,
      sym_class_interface_clause,
  [52628] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3431), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [52642] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1660), 1,
      anon_sym_BSLASHu,
    ACTIONS(1658), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [52656] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1660), 1,
      anon_sym_BSLASHu,
    ACTIONS(1658), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [52670] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3431), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [52684] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3495), 1,
      aux_sym_string_content_token1,
    ACTIONS(3497), 1,
      anon_sym_SQUOTE,
    ACTIONS(3499), 1,
      aux_sym_string_token2,
    STATE(1866), 1,
      aux_sym_string_content_repeat1,
    STATE(1597), 2,
      sym_string_content,
      aux_sym_string_repeat1,
  [52704] = 5,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3501), 1,
      sym_name,
    ACTIONS(3503), 1,
      anon_sym_LBRACE,
    STATE(602), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [52722] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    STATE(601), 1,
      sym_arguments,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [52738] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3457), 1,
      anon_sym_BSLASHu,
    ACTIONS(3459), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [52752] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(1873), 1,
      sym_base_clause,
    STATE(2020), 1,
      sym_declaration_list,
    STATE(2260), 1,
      sym_class_interface_clause,
  [52774] = 5,
    ACTIONS(125), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3505), 1,
      sym_name,
    ACTIONS(3507), 1,
      anon_sym_LBRACE,
    STATE(628), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [52792] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3423), 1,
      anon_sym_BSLASHu,
    ACTIONS(3425), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [52806] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(511), 1,
      sym_declaration_list,
    STATE(1848), 1,
      sym_base_clause,
    STATE(2243), 1,
      sym_class_interface_clause,
  [52828] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3495), 1,
      aux_sym_string_content_token1,
    ACTIONS(3509), 1,
      anon_sym_SQUOTE,
    ACTIONS(3511), 1,
      aux_sym_string_token2,
    STATE(1866), 1,
      aux_sym_string_content_repeat1,
    STATE(1639), 2,
      sym_string_content,
      aux_sym_string_repeat1,
  [52848] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(561), 1,
      sym_declaration_list,
    STATE(1805), 1,
      sym_base_clause,
    STATE(2261), 1,
      sym_class_interface_clause,
  [52870] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3513), 1,
      aux_sym_if_statement_token2,
    ACTIONS(3515), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3517), 1,
      aux_sym_else_clause_token1,
    STATE(1605), 1,
      aux_sym_if_statement_repeat2,
    STATE(2050), 1,
      sym_else_if_clause_2,
    STATE(2591), 1,
      sym_else_clause_2,
  [52892] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3519), 1,
      anon_sym_RBRACE,
    ACTIONS(3521), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3523), 1,
      aux_sym_match_default_expression_token1,
    STATE(1606), 3,
      sym_case_statement,
      sym_default_statement,
      aux_sym_switch_block_repeat1,
  [52910] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3521), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3523), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3525), 1,
      aux_sym_switch_block_token1,
    STATE(1607), 3,
      sym_case_statement,
      sym_default_statement,
      aux_sym_switch_block_repeat1,
  [52928] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(808), 1,
      sym_declaration_list,
    STATE(1849), 1,
      sym_base_clause,
    STATE(2312), 1,
      sym_class_interface_clause,
  [52950] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(1857), 1,
      sym_base_clause,
    STATE(2174), 1,
      sym_declaration_list,
    STATE(2336), 1,
      sym_class_interface_clause,
  [52972] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(1860), 1,
      sym_base_clause,
    STATE(2191), 1,
      sym_declaration_list,
    STATE(2340), 1,
      sym_class_interface_clause,
  [52994] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3515), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3517), 1,
      aux_sym_else_clause_token1,
    ACTIONS(3527), 1,
      aux_sym_if_statement_token2,
    STATE(1735), 1,
      aux_sym_if_statement_repeat2,
    STATE(2050), 1,
      sym_else_if_clause_2,
    STATE(2487), 1,
      sym_else_clause_2,
  [53016] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3521), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3523), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3529), 1,
      anon_sym_RBRACE,
    STATE(1522), 3,
      sym_case_statement,
      sym_default_statement,
      aux_sym_switch_block_repeat1,
  [53034] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3521), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3523), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3531), 1,
      aux_sym_switch_block_token1,
    STATE(1522), 3,
      sym_case_statement,
      sym_default_statement,
      aux_sym_switch_block_repeat1,
  [53052] = 7,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3533), 1,
      sym_name,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1795), 1,
      sym_named_type,
    STATE(2060), 1,
      sym_type_list,
    STATE(2646), 1,
      sym_namespace_name,
  [53074] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(815), 1,
      sym_declaration_list,
    STATE(1875), 1,
      sym_base_clause,
    STATE(2354), 1,
      sym_class_interface_clause,
  [53096] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(825), 1,
      sym_declaration_list,
    STATE(1780), 1,
      sym_base_clause,
    STATE(2368), 1,
      sym_class_interface_clause,
  [53118] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3515), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3517), 1,
      aux_sym_else_clause_token1,
    ACTIONS(3535), 1,
      aux_sym_if_statement_token2,
    STATE(1632), 1,
      aux_sym_if_statement_repeat2,
    STATE(2050), 1,
      sym_else_if_clause_2,
    STATE(2497), 1,
      sym_else_clause_2,
  [53140] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3521), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3523), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3537), 1,
      anon_sym_RBRACE,
    STATE(1634), 3,
      sym_case_statement,
      sym_default_statement,
      aux_sym_switch_block_repeat1,
  [53158] = 7,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3533), 1,
      sym_name,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(1795), 1,
      sym_named_type,
    STATE(2085), 1,
      sym_type_list,
    STATE(2646), 1,
      sym_namespace_name,
  [53180] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3521), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3523), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3539), 1,
      aux_sym_switch_block_token1,
    STATE(1637), 3,
      sym_case_statement,
      sym_default_statement,
      aux_sym_switch_block_repeat1,
  [53198] = 5,
    ACTIONS(1592), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3541), 1,
      sym_name,
    ACTIONS(3543), 1,
      anon_sym_LBRACE,
    STATE(861), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [53216] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2092), 1,
      anon_sym_LPAREN2,
    STATE(864), 1,
      sym_arguments,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [53232] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1822), 1,
      anon_sym_DOLLAR,
    ACTIONS(3545), 1,
      sym_name,
    ACTIONS(3547), 1,
      anon_sym_LBRACE,
    STATE(921), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [53250] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(929), 1,
      sym_arguments,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [53266] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3549), 1,
      sym_name,
    ACTIONS(3551), 1,
      anon_sym_LBRACE,
    ACTIONS(3553), 1,
      anon_sym_DOLLAR,
    STATE(1578), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [53284] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1854), 1,
      anon_sym_LPAREN2,
    STATE(720), 1,
      sym_arguments,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [53300] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(973), 1,
      sym_declaration_list,
    STATE(1809), 1,
      sym_base_clause,
    STATE(2324), 1,
      sym_class_interface_clause,
  [53322] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3555), 1,
      sym_name,
    ACTIONS(3557), 1,
      anon_sym_LBRACE,
    ACTIONS(3559), 1,
      anon_sym_DOLLAR,
    STATE(1581), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [53340] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1861), 1,
      anon_sym_LPAREN2,
    STATE(761), 1,
      sym_arguments,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [53356] = 5,
    ACTIONS(1588), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3561), 1,
      sym_name,
    ACTIONS(3563), 1,
      anon_sym_LBRACE,
    STATE(732), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [53374] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1828), 1,
      anon_sym_DOLLAR,
    ACTIONS(3565), 1,
      sym_name,
    ACTIONS(3567), 1,
      anon_sym_LBRACE,
    STATE(780), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [53392] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3551), 1,
      anon_sym_LBRACE,
    ACTIONS(3553), 1,
      anon_sym_DOLLAR,
    ACTIONS(3569), 1,
      sym_name,
    STATE(1578), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [53410] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3557), 1,
      anon_sym_LBRACE,
    ACTIONS(3559), 1,
      anon_sym_DOLLAR,
    ACTIONS(3571), 1,
      sym_name,
    STATE(1581), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [53428] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3573), 1,
      sym_name,
    ACTIONS(3575), 1,
      anon_sym_LBRACE,
    ACTIONS(3577), 1,
      anon_sym_DOLLAR,
    STATE(1526), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [53446] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3575), 1,
      anon_sym_LBRACE,
    ACTIONS(3577), 1,
      anon_sym_DOLLAR,
    ACTIONS(3579), 1,
      sym_name,
    STATE(1526), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [53464] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(512), 1,
      sym_declaration_list,
    STATE(1836), 1,
      sym_base_clause,
    STATE(2356), 1,
      sym_class_interface_clause,
  [53486] = 7,
    ACTIONS(1057), 1,
      aux_sym_while_statement_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3471), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3473), 1,
      aux_sym_else_clause_token1,
    STATE(1565), 1,
      aux_sym_if_statement_repeat1,
    STATE(2092), 1,
      sym_else_if_clause,
    STATE(2111), 1,
      sym_else_clause,
  [53508] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3515), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3517), 1,
      aux_sym_else_clause_token1,
    ACTIONS(3581), 1,
      aux_sym_if_statement_token2,
    STATE(1735), 1,
      aux_sym_if_statement_repeat2,
    STATE(2050), 1,
      sym_else_if_clause_2,
    STATE(2441), 1,
      sym_else_clause_2,
  [53530] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3149), 1,
      anon_sym_DOLLAR,
    ACTIONS(3154), 1,
      anon_sym_PIPE,
    ACTIONS(3583), 1,
      anon_sym_AMP,
    STATE(1464), 1,
      aux_sym_union_type_repeat1,
    STATE(1473), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    STATE(2089), 1,
      aux_sym_intersection_type_repeat1,
  [53552] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3521), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3523), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3589), 1,
      anon_sym_RBRACE,
    STATE(1522), 3,
      sym_case_statement,
      sym_default_statement,
      aux_sym_switch_block_repeat1,
  [53570] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3455), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [53584] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3455), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [53598] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3521), 1,
      aux_sym_enum_case_token1,
    ACTIONS(3523), 1,
      aux_sym_match_default_expression_token1,
    ACTIONS(3591), 1,
      aux_sym_switch_block_token1,
    STATE(1522), 3,
      sym_case_statement,
      sym_default_statement,
      aux_sym_switch_block_repeat1,
  [53616] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3593), 1,
      anon_sym_BSLASHu,
    ACTIONS(3031), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [53630] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3595), 1,
      anon_sym_SQUOTE,
    ACTIONS(3597), 1,
      aux_sym_string_token2,
    ACTIONS(3600), 1,
      aux_sym_string_content_token1,
    STATE(1866), 1,
      aux_sym_string_content_repeat1,
    STATE(1639), 2,
      sym_string_content,
      aux_sym_string_repeat1,
  [53650] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(899), 1,
      sym_declaration_list,
    STATE(1854), 1,
      sym_base_clause,
    STATE(2310), 1,
      sym_class_interface_clause,
  [53672] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(896), 1,
      sym_declaration_list,
    STATE(1828), 1,
      sym_base_clause,
    STATE(2306), 1,
      sym_class_interface_clause,
  [53694] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3603), 1,
      anon_sym_BSLASHu,
    ACTIONS(3045), 5,
      sym_execution_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR,
  [53708] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3357), 1,
      aux_sym_namespace_use_clause_token1,
    ACTIONS(3605), 1,
      aux_sym_use_instead_of_clause_token1,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [53724] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3423), 1,
      anon_sym_BSLASHu,
    ACTIONS(3425), 5,
      sym_encapsed_string_chars,
      anon_sym_LBRACE,
      sym_escape_sequence,
      anon_sym_DQUOTE,
      anon_sym_DOLLAR,
  [53738] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3495), 1,
      aux_sym_string_content_token1,
    ACTIONS(3511), 1,
      aux_sym_string_token2,
    ACTIONS(3607), 1,
      anon_sym_SQUOTE,
    STATE(1866), 1,
      aux_sym_string_content_repeat1,
    STATE(1639), 2,
      sym_string_content,
      aux_sym_string_repeat1,
  [53758] = 7,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(816), 1,
      sym_declaration_list,
    STATE(1868), 1,
      sym_base_clause,
    STATE(2355), 1,
      sym_class_interface_clause,
  [53780] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3609), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3611), 1,
      anon_sym_LBRACE,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    STATE(1881), 1,
      sym_anonymous_function_use_clause,
    STATE(2524), 1,
      sym__return_type,
  [53799] = 5,
    ACTIONS(29), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(475), 1,
      sym_compound_statement,
    STATE(485), 1,
      sym__semicolon,
    ACTIONS(3615), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [53816] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(515), 1,
      sym__semicolon,
    STATE(1717), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(3617), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [53833] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(1835), 1,
      aux_sym__const_declaration_repeat1,
    STATE(2181), 1,
      sym__semicolon,
    ACTIONS(3621), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [53850] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3623), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [53861] = 6,
    ACTIONS(368), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3625), 1,
      sym_name,
    ACTIONS(3627), 1,
      anon_sym_BSLASH,
    STATE(1660), 1,
      sym_namespace_name,
    STATE(2006), 1,
      sym_compound_statement,
  [53880] = 5,
    ACTIONS(29), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1379), 1,
      sym_compound_statement,
    STATE(1380), 1,
      sym__semicolon,
    ACTIONS(3629), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [53897] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3633), 1,
      anon_sym_COMMA,
    STATE(1654), 1,
      aux_sym_base_clause_repeat1,
    ACTIONS(3631), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [53912] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3636), 1,
      anon_sym_LBRACE,
    ACTIONS(3638), 1,
      anon_sym_COLON,
    STATE(510), 1,
      sym_enum_declaration_list,
    STATE(2224), 1,
      sym_class_interface_clause,
  [53931] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(3640), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1973), 1,
      sym_reference_modifier,
    STATE(1987), 1,
      sym_variable_name,
  [53950] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    ACTIONS(3644), 1,
      anon_sym_EQ,
    STATE(2416), 1,
      sym_property_hook_list,
    ACTIONS(3642), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [53967] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3648), 1,
      anon_sym_COMMA,
    STATE(1695), 1,
      aux_sym_function_static_declaration_repeat1,
    STATE(1967), 1,
      sym__semicolon,
    ACTIONS(3646), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [53984] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3652), 1,
      anon_sym_COMMA,
    STATE(1697), 1,
      aux_sym_global_declaration_repeat1,
    STATE(1971), 1,
      sym__semicolon,
    ACTIONS(3650), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54001] = 5,
    ACTIONS(368), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1977), 1,
      sym_compound_statement,
    STATE(1978), 1,
      sym__semicolon,
    ACTIONS(3654), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54018] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3658), 1,
      anon_sym_COMMA,
    STATE(1698), 1,
      aux_sym_namespace_use_declaration_repeat1,
    STATE(1982), 1,
      sym__semicolon,
    ACTIONS(3656), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54035] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(1714), 1,
      aux_sym__const_declaration_repeat1,
    STATE(2011), 1,
      sym__semicolon,
    ACTIONS(3660), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54052] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3662), 1,
      anon_sym_LBRACE,
    ACTIONS(3664), 1,
      anon_sym_COLON,
    STATE(2019), 1,
      sym_enum_declaration_list,
    STATE(2258), 1,
      sym_class_interface_clause,
  [54071] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3609), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(3666), 1,
      anon_sym_LBRACE,
    STATE(2010), 1,
      sym_anonymous_function_use_clause,
    STATE(2424), 1,
      sym__return_type,
  [54090] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3609), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(3668), 1,
      anon_sym_LBRACE,
    STATE(1932), 1,
      sym_anonymous_function_use_clause,
    STATE(2481), 1,
      sym__return_type,
  [54109] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3371), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(3373), 1,
      aux_sym_finally_clause_token1,
    STATE(1508), 3,
      sym_catch_clause,
      sym_finally_clause,
      aux_sym_try_statement_repeat1,
  [54124] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3670), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3672), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [54137] = 3,
    ACTIONS(1096), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1094), 4,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [54150] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3636), 1,
      anon_sym_LBRACE,
    ACTIONS(3674), 1,
      anon_sym_COLON,
    STATE(559), 1,
      sym_enum_declaration_list,
    STATE(2246), 1,
      sym_class_interface_clause,
  [54169] = 3,
    ACTIONS(1073), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1071), 4,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [54182] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3609), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(3676), 1,
      anon_sym_LBRACE,
    STATE(1944), 1,
      sym_anonymous_function_use_clause,
    STATE(2543), 1,
      sym__return_type,
  [54201] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(565), 1,
      sym__semicolon,
    STATE(1686), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(3678), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54218] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(566), 1,
      sym__semicolon,
    STATE(1835), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(3680), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54235] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3684), 1,
      anon_sym_BSLASH,
    STATE(1727), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3682), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [54250] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3686), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3688), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [54263] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3662), 1,
      anon_sym_LBRACE,
    ACTIONS(3690), 1,
      anon_sym_COLON,
    STATE(2158), 1,
      sym_enum_declaration_list,
    STATE(2331), 1,
      sym_class_interface_clause,
  [54282] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3343), 1,
      anon_sym_AMP,
    STATE(1768), 1,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3291), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [54297] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    ACTIONS(3694), 1,
      anon_sym_EQ,
    STATE(2400), 1,
      sym_property_hook_list,
    ACTIONS(3692), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [54314] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    ACTIONS(3698), 1,
      anon_sym_EQ,
    STATE(2403), 1,
      sym_property_hook_list,
    ACTIONS(3696), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [54331] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(476), 1,
      sym__semicolon,
    STATE(1734), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(3700), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54348] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3702), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3704), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [54361] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3706), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3708), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [54374] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3710), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3712), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [54387] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(3714), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(2003), 1,
      sym_reference_modifier,
    STATE(2005), 1,
      sym_variable_name,
  [54406] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3609), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(3716), 1,
      anon_sym_LBRACE,
    STATE(2103), 1,
      sym_anonymous_function_use_clause,
    STATE(2443), 1,
      sym__return_type,
  [54425] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(539), 1,
      sym__semicolon,
    STATE(1835), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(3718), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54442] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    ACTIONS(3722), 1,
      anon_sym_EQ,
    STATE(2234), 1,
      sym_property_hook_list,
    ACTIONS(3720), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [54459] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    ACTIONS(3726), 1,
      anon_sym_EQ,
    STATE(2334), 1,
      sym_property_hook_list,
    ACTIONS(3724), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [54476] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3658), 1,
      anon_sym_COMMA,
    STATE(486), 1,
      sym__semicolon,
    STATE(1692), 1,
      aux_sym_namespace_use_declaration_repeat1,
    ACTIONS(3728), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54493] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3230), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [54504] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3730), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3732), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [54517] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3658), 1,
      anon_sym_COMMA,
    STATE(474), 1,
      sym__semicolon,
    STATE(1791), 1,
      aux_sym_namespace_use_declaration_repeat1,
    ACTIONS(3734), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54534] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3094), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [54545] = 6,
    ACTIONS(39), 1,
      aux_sym_final_modifier_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3736), 1,
      sym_name,
    STATE(2114), 1,
      sym_final_modifier,
    STATE(2474), 1,
      sym_reference_modifier,
  [54564] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3648), 1,
      anon_sym_COMMA,
    STATE(1800), 1,
      aux_sym_function_static_declaration_repeat1,
    STATE(2141), 1,
      sym__semicolon,
    ACTIONS(3738), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54581] = 5,
    ACTIONS(29), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1368), 1,
      sym_compound_statement,
    STATE(1369), 1,
      sym__semicolon,
    ACTIONS(3740), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54598] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3652), 1,
      anon_sym_COMMA,
    STATE(1871), 1,
      aux_sym_global_declaration_repeat1,
    STATE(2143), 1,
      sym__semicolon,
    ACTIONS(3742), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54615] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3658), 1,
      anon_sym_COMMA,
    STATE(1791), 1,
      aux_sym_namespace_use_declaration_repeat1,
    STATE(2157), 1,
      sym__semicolon,
    ACTIONS(3744), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54632] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      anon_sym_EQ,
    STATE(2345), 1,
      sym_property_hook_list,
    ACTIONS(3746), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [54649] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    ACTIONS(3752), 1,
      anon_sym_EQ,
    STATE(2347), 1,
      sym_property_hook_list,
    ACTIONS(3750), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [54666] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3754), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [54677] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(1650), 1,
      aux_sym__const_declaration_repeat1,
    STATE(2091), 1,
      sym__semicolon,
    ACTIONS(3756), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54694] = 5,
    ACTIONS(29), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1370), 1,
      sym_compound_statement,
    STATE(1371), 1,
      sym__semicolon,
    ACTIONS(3758), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54711] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3760), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3762), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [54724] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3766), 1,
      anon_sym_COLON,
    STATE(2175), 1,
      sym__return_type,
    ACTIONS(3764), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [54739] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3222), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [54750] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(1764), 1,
      aux_sym__const_declaration_repeat1,
    STATE(2185), 1,
      sym__semicolon,
    ACTIONS(3768), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54767] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3770), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3772), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [54780] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(477), 1,
      sym__semicolon,
    STATE(1673), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(3774), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54797] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3776), 1,
      anon_sym_PIPE,
    STATE(1767), 1,
      aux_sym_union_type_repeat1,
    ACTIONS(3279), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [54812] = 5,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(3778), 1,
      anon_sym_RPAREN2,
    STATE(2352), 2,
      sym_variable_name,
      sym_by_ref,
  [54829] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3780), 1,
      anon_sym_PIPE,
    STATE(1770), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3299), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [54844] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3784), 1,
      anon_sym_EQ,
    ACTIONS(3782), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [54857] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(1835), 1,
      aux_sym__const_declaration_repeat1,
    STATE(2200), 1,
      sym__semicolon,
    ACTIONS(3786), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54874] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3684), 1,
      anon_sym_BSLASH,
    STATE(1727), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3788), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [54889] = 3,
    ACTIONS(1077), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1075), 4,
      aux_sym_catch_clause_token1,
      aux_sym_finally_clause_token1,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [54902] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(477), 1,
      sym__semicolon,
    STATE(1835), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(3774), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54919] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3790), 1,
      aux_sym_catch_clause_token1,
    ACTIONS(3792), 1,
      aux_sym_finally_clause_token1,
    STATE(443), 3,
      sym_catch_clause,
      sym_finally_clause,
      aux_sym_try_statement_repeat1,
  [54934] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3794), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [54945] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3780), 1,
      anon_sym_PIPE,
    STATE(1770), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3285), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [54960] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1081), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(1079), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [54973] = 5,
    ACTIONS(29), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1376), 1,
      sym_compound_statement,
    STATE(1377), 1,
      sym__semicolon,
    ACTIONS(3796), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54990] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3798), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3800), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [55003] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3802), 1,
      anon_sym_PIPE,
    STATE(1724), 1,
      aux_sym_union_type_repeat1,
    ACTIONS(3248), 3,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [55018] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3805), 1,
      anon_sym_PIPE,
    STATE(1725), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3272), 3,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [55033] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3808), 1,
      anon_sym_PIPE,
    STATE(1725), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3285), 3,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [55048] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3812), 1,
      anon_sym_BSLASH,
    STATE(1727), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3810), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [55063] = 6,
    ACTIONS(633), 1,
      anon_sym_BSLASH,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3533), 1,
      sym_name,
    STATE(1450), 1,
      sym_qualified_name,
    STATE(2147), 1,
      sym_named_type,
    STATE(2646), 1,
      sym_namespace_name,
  [55082] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1941), 1,
      anon_sym_LPAREN2,
    ACTIONS(3815), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3817), 1,
      anon_sym_DASH_GT,
    ACTIONS(3819), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(3821), 1,
      anon_sym_LBRACK,
  [55101] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1923), 1,
      anon_sym_LPAREN2,
    ACTIONS(1925), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [55114] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1686), 1,
      anon_sym_LPAREN2,
    ACTIONS(1688), 4,
      anon_sym_COLON_COLON,
      anon_sym_DASH_GT,
      anon_sym_QMARK_DASH_GT,
      anon_sym_LBRACK,
  [55127] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1092), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(1090), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [55140] = 6,
    ACTIONS(1083), 1,
      aux_sym_while_statement_token1,
    ACTIONS(1085), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3823), 1,
      aux_sym_else_if_clause_token1,
    STATE(1733), 1,
      aux_sym_if_statement_repeat1,
    STATE(2092), 1,
      sym_else_if_clause,
  [55159] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(516), 1,
      sym__semicolon,
    STATE(1835), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(3826), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [55176] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3828), 1,
      aux_sym_if_statement_token2,
    ACTIONS(3830), 1,
      aux_sym_else_if_clause_token1,
    ACTIONS(3833), 1,
      aux_sym_else_clause_token1,
    STATE(1735), 1,
      aux_sym_if_statement_repeat2,
    STATE(2050), 1,
      sym_else_if_clause_2,
  [55195] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3835), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3837), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [55208] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3839), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3841), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [55221] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3843), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3845), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [55234] = 6,
    ACTIONS(29), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3625), 1,
      sym_name,
    ACTIONS(3627), 1,
      anon_sym_BSLASH,
    STATE(459), 1,
      sym_compound_statement,
    STATE(1648), 1,
      sym_namespace_name,
  [55253] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2472), 1,
      anon_sym_LBRACE,
    ACTIONS(3849), 1,
      anon_sym_EQ,
    STATE(2406), 1,
      sym_property_hook_list,
    ACTIONS(3847), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [55270] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3648), 1,
      anon_sym_COMMA,
    STATE(472), 1,
      sym__semicolon,
    STATE(1800), 1,
      aux_sym_function_static_declaration_repeat1,
    ACTIONS(3851), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [55287] = 5,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(3853), 1,
      anon_sym_RPAREN2,
    STATE(2352), 2,
      sym_variable_name,
      sym_by_ref,
  [55304] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3855), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3857), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [55317] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3652), 1,
      anon_sym_COMMA,
    STATE(473), 1,
      sym__semicolon,
    STATE(1871), 1,
      aux_sym_global_declaration_repeat1,
    ACTIONS(3859), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [55334] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3863), 1,
      anon_sym_COMMA,
    STATE(1745), 1,
      aux_sym_property_declaration_repeat1,
    ACTIONS(3861), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [55349] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3684), 1,
      anon_sym_BSLASH,
    STATE(1715), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3682), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [55364] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(1772), 1,
      aux_sym__const_declaration_repeat1,
    STATE(2200), 1,
      sym__semicolon,
    ACTIONS(3786), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [55381] = 6,
    ACTIONS(983), 1,
      anon_sym_COLON,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1647), 1,
      sym_formal_parameters,
    STATE(2264), 1,
      sym_reference_modifier,
  [55400] = 6,
    ACTIONS(983), 1,
      anon_sym_COLON,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(2135), 1,
      sym_formal_parameters,
    STATE(2348), 1,
      sym_reference_modifier,
  [55419] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3609), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(3866), 1,
      anon_sym_LBRACE,
    STATE(2216), 1,
      sym_anonymous_function_use_clause,
    STATE(2466), 1,
      sym__return_type,
  [55438] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3648), 1,
      anon_sym_COMMA,
    STATE(457), 1,
      sym__semicolon,
    STATE(1741), 1,
      aux_sym_function_static_declaration_repeat1,
    ACTIONS(3868), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [55455] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3609), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(3870), 1,
      anon_sym_LBRACE,
    STATE(2177), 1,
      sym_anonymous_function_use_clause,
    STATE(2426), 1,
      sym__return_type,
  [55474] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3872), 5,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [55485] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3766), 1,
      anon_sym_COLON,
    STATE(1893), 1,
      sym__return_type,
    ACTIONS(3874), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [55500] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3652), 1,
      anon_sym_COMMA,
    STATE(469), 1,
      sym__semicolon,
    STATE(1744), 1,
      aux_sym_global_declaration_repeat1,
    ACTIONS(3876), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [55517] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3878), 1,
      anon_sym_PIPE,
    STATE(1724), 1,
      aux_sym_union_type_repeat1,
    ACTIONS(3279), 3,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [55532] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3808), 1,
      anon_sym_PIPE,
    STATE(1725), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3299), 3,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [55547] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3780), 1,
      anon_sym_PIPE,
    STATE(1720), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3295), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [55562] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3808), 1,
      anon_sym_PIPE,
    STATE(1726), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3295), 3,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [55577] = 6,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3609), 1,
      aux_sym_namespace_use_declaration_token1,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(3880), 1,
      anon_sym_LBRACE,
    STATE(1912), 1,
      sym_anonymous_function_use_clause,
    STATE(2588), 1,
      sym__return_type,
  [55596] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3882), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3884), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [55609] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3886), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3888), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [55622] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3890), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3892), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [55635] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(1835), 1,
      aux_sym__const_declaration_repeat1,
    STATE(1963), 1,
      sym__semicolon,
    ACTIONS(3894), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [55652] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3896), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3898), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [55665] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3900), 2,
      aux_sym_final_modifier_token1,
      sym_name,
    ACTIONS(3902), 3,
      anon_sym_AMP,
      anon_sym_RBRACE,
      anon_sym_POUND_LBRACK,
  [55678] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3904), 1,
      anon_sym_PIPE,
    STATE(1767), 1,
      aux_sym_union_type_repeat1,
    ACTIONS(3248), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [55693] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3907), 1,
      anon_sym_AMP,
    STATE(1768), 1,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3268), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [55708] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3910), 1,
      sym_name,
    STATE(2073), 1,
      sym_visibility_modifier,
    ACTIONS(3912), 3,
      aux_sym_visibility_modifier_token1,
      aux_sym_visibility_modifier_token2,
      aux_sym_visibility_modifier_token3,
  [55723] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3914), 1,
      anon_sym_PIPE,
    STATE(1770), 1,
      aux_sym_disjunctive_normal_form_type_repeat1,
    ACTIONS(3272), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [55738] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3684), 1,
      anon_sym_BSLASH,
    STATE(1674), 1,
      aux_sym_namespace_name_repeat1,
    ACTIONS(3917), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [55753] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3619), 1,
      anon_sym_COMMA,
    STATE(1835), 1,
      aux_sym__const_declaration_repeat1,
    STATE(2100), 1,
      sym__semicolon,
    ACTIONS(3919), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [55770] = 4,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(2007), 2,
      sym_variable_name,
      sym_by_ref,
  [55784] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2180), 1,
      sym_declaration_list,
    STATE(2375), 1,
      sym_class_interface_clause,
  [55800] = 3,
    ACTIONS(125), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1659), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [55812] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3923), 1,
      aux_sym_namespace_use_clause_token1,
    ACTIONS(3921), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [55824] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3925), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3927), 1,
      anon_sym_DASH_GT,
    ACTIONS(3929), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(3931), 1,
      anon_sym_LBRACK,
  [55840] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2017), 1,
      sym_declaration_list,
    STATE(2257), 1,
      sym_base_clause,
  [55856] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2578), 1,
      sym_declare_directive,
    ACTIONS(3933), 3,
      anon_sym_ticks,
      anon_sym_encoding,
      anon_sym_strict_types,
  [55868] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(749), 1,
      sym_declaration_list,
    STATE(2382), 1,
      sym_class_interface_clause,
  [55884] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(2213), 1,
      sym_formal_parameters,
    STATE(2244), 1,
      sym_reference_modifier,
  [55900] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(538), 1,
      sym_declaration_list,
    STATE(2330), 1,
      sym_class_interface_clause,
  [55916] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3248), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [55926] = 4,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(1700), 2,
      sym_variable_name,
      sym_by_ref,
  [55940] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3935), 1,
      anon_sym_AMP,
    STATE(1785), 1,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3268), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [55954] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1671), 1,
      sym_formal_parameters,
    STATE(2333), 1,
      sym_reference_modifier,
  [55970] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3938), 1,
      anon_sym_COMMA,
    STATE(1787), 1,
      aux_sym_base_clause_repeat1,
    ACTIONS(3631), 2,
      anon_sym_LBRACE,
      aux_sym_class_interface_clause_token1,
  [55984] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3810), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      anon_sym_LBRACE,
  [55994] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3943), 1,
      anon_sym_EQ,
    STATE(1401), 1,
      sym__semicolon,
    ACTIONS(3941), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56008] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3947), 1,
      anon_sym_LPAREN2,
    STATE(2041), 1,
      sym__semicolon,
    ACTIONS(3945), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56022] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3951), 1,
      anon_sym_COMMA,
    STATE(1791), 1,
      aux_sym_namespace_use_declaration_repeat1,
    ACTIONS(3949), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56036] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3925), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3954), 1,
      anon_sym_DASH_GT,
    ACTIONS(3956), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(3958), 1,
      anon_sym_LBRACK,
  [56052] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3962), 1,
      anon_sym_PIPE,
    STATE(1838), 1,
      aux_sym_type_list_repeat1,
    ACTIONS(3960), 2,
      anon_sym_RPAREN2,
      anon_sym_DOLLAR,
  [56066] = 5,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3964), 1,
      anon_sym_EQ,
    ACTIONS(3966), 1,
      anon_sym_RPAREN2,
    STATE(2132), 1,
      aux_sym__list_destructing_repeat1,
  [56082] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3962), 1,
      anon_sym_PIPE,
    STATE(1793), 1,
      aux_sym_type_list_repeat1,
    ACTIONS(3968), 2,
      anon_sym_RPAREN2,
      anon_sym_DOLLAR,
  [56096] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3925), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3970), 1,
      anon_sym_DASH_GT,
    ACTIONS(3972), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(3974), 1,
      anon_sym_LBRACK,
  [56112] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(808), 1,
      sym_declaration_list,
    STATE(2312), 1,
      sym_class_interface_clause,
  [56128] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1750), 1,
      sym_formal_parameters,
    STATE(2380), 1,
      sym_reference_modifier,
  [56144] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3244), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [56154] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3978), 1,
      anon_sym_COMMA,
    STATE(1800), 1,
      aux_sym_function_static_declaration_repeat1,
    ACTIONS(3976), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56168] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3981), 4,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
  [56178] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3983), 1,
      anon_sym_COMMA,
    STATE(1870), 1,
      aux_sym_base_clause_repeat1,
    ACTIONS(3985), 2,
      anon_sym_LBRACE,
      aux_sym_class_interface_clause_token1,
  [56192] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(925), 1,
      sym_declaration_list,
    STATE(2384), 1,
      sym_class_interface_clause,
  [56208] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3272), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [56218] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(482), 1,
      sym_declaration_list,
    STATE(2263), 1,
      sym_class_interface_clause,
  [56234] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1960), 1,
      sym_formal_parameters,
    STATE(2313), 1,
      sym_reference_modifier,
  [56250] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3989), 1,
      sym_nowdoc_string,
    STATE(1847), 1,
      aux_sym_nowdoc_body_repeat1,
    ACTIONS(3987), 2,
      sym_heredoc_end,
      sym__new_line,
  [56264] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3925), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3931), 1,
      anon_sym_LBRACK,
    ACTIONS(3991), 1,
      anon_sym_DASH_GT,
    ACTIONS(3993), 1,
      anon_sym_QMARK_DASH_GT,
  [56280] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(896), 1,
      sym_declaration_list,
    STATE(2306), 1,
      sym_class_interface_clause,
  [56296] = 3,
    ACTIONS(125), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1755), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [56308] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(2162), 1,
      sym_formal_parameters,
    STATE(2381), 1,
      sym_reference_modifier,
  [56324] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3997), 1,
      aux_sym_string_content_token1,
    STATE(1812), 1,
      aux_sym_string_content_repeat1,
    ACTIONS(3995), 2,
      anon_sym_SQUOTE,
      aux_sym_string_token2,
  [56338] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(899), 1,
      sym_declaration_list,
    STATE(2310), 1,
      sym_class_interface_clause,
  [56354] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4000), 1,
      anon_sym_DQUOTE2,
    ACTIONS(4002), 1,
      sym__new_line,
    ACTIONS(4004), 1,
      sym_heredoc_end,
    STATE(2391), 1,
      sym_heredoc_body,
  [56370] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4006), 4,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
  [56380] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4010), 1,
      aux_sym_namespace_use_clause_token1,
    ACTIONS(4008), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [56392] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1752), 1,
      sym_formal_parameters,
    STATE(2286), 1,
      sym_reference_modifier,
  [56408] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4012), 1,
      anon_sym_LPAREN,
    ACTIONS(2638), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      sym_name,
  [56420] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3872), 4,
      aux_sym_namespace_use_declaration_token1,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [56430] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3925), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4014), 1,
      anon_sym_DASH_GT,
    ACTIONS(4016), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(4018), 1,
      anon_sym_LBRACK,
  [56446] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2465), 1,
      sym_declare_directive,
    ACTIONS(3933), 3,
      anon_sym_ticks,
      anon_sym_encoding,
      anon_sym_strict_types,
  [56458] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4002), 1,
      sym__new_line,
    ACTIONS(4020), 1,
      anon_sym_DQUOTE2,
    ACTIONS(4022), 1,
      sym_heredoc_end,
    STATE(2325), 1,
      sym_heredoc_body,
  [56474] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3623), 4,
      aux_sym_namespace_use_declaration_token1,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [56484] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2578), 1,
      aux_sym_function_static_declaration_token1,
    ACTIONS(2855), 1,
      aux_sym__arrow_function_header_token1,
    ACTIONS(4024), 1,
      aux_sym__namespace_use_type_token1,
    STATE(2277), 1,
      sym_static_modifier,
  [56500] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(507), 1,
      sym_declaration_list,
    STATE(2276), 1,
      sym_base_clause,
  [56516] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3754), 4,
      aux_sym_namespace_use_declaration_token1,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [56526] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1647), 1,
      sym_formal_parameters,
    STATE(2264), 1,
      sym_reference_modifier,
  [56542] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(955), 1,
      sym_declaration_list,
    STATE(2317), 1,
      sym_class_interface_clause,
  [56558] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3662), 1,
      anon_sym_LBRACE,
    STATE(2178), 1,
      sym_enum_declaration_list,
    STATE(2373), 1,
      sym_class_interface_clause,
  [56574] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(973), 1,
      sym_declaration_list,
    STATE(2324), 1,
      sym_class_interface_clause,
  [56590] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3861), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [56600] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2153), 1,
      sym_declaration_list,
    STATE(2418), 1,
      sym_base_clause,
  [56616] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1822), 1,
      anon_sym_DOLLAR,
    STATE(974), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [56628] = 4,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(1740), 2,
      sym_variable_name,
      sym_by_ref,
  [56642] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4028), 1,
      anon_sym_COMMA,
    STATE(1835), 1,
      aux_sym__const_declaration_repeat1,
    ACTIONS(4026), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56656] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(568), 1,
      sym_declaration_list,
    STATE(2389), 1,
      sym_class_interface_clause,
  [56672] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3349), 1,
      aux_sym_base_clause_token1,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(557), 1,
      sym_declaration_list,
    STATE(2223), 1,
      sym_base_clause,
  [56688] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4033), 1,
      anon_sym_PIPE,
    STATE(1838), 1,
      aux_sym_type_list_repeat1,
    ACTIONS(4031), 2,
      anon_sym_RPAREN2,
      anon_sym_DOLLAR,
  [56702] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(2135), 1,
      sym_formal_parameters,
    STATE(2348), 1,
      sym_reference_modifier,
  [56718] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    STATE(2311), 1,
      sym_arguments,
    ACTIONS(4036), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [56732] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3662), 1,
      anon_sym_LBRACE,
    STATE(1984), 1,
      sym_enum_declaration_list,
    STATE(2342), 1,
      sym_class_interface_clause,
  [56748] = 4,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(1688), 2,
      sym_variable_name,
      sym_by_ref,
  [56762] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3925), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4038), 1,
      anon_sym_DASH_GT,
    ACTIONS(4040), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(4042), 1,
      anon_sym_LBRACK,
  [56778] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4046), 1,
      anon_sym_EQ,
    ACTIONS(4044), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [56790] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3925), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4048), 1,
      anon_sym_DASH_GT,
    ACTIONS(4050), 1,
      anon_sym_QMARK_DASH_GT,
    ACTIONS(4052), 1,
      anon_sym_LBRACK,
  [56806] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3631), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [56816] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4056), 1,
      sym_nowdoc_string,
    STATE(1847), 1,
      aux_sym_nowdoc_body_repeat1,
    ACTIONS(4054), 2,
      sym_heredoc_end,
      sym__new_line,
  [56830] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(562), 1,
      sym_declaration_list,
    STATE(2366), 1,
      sym_class_interface_clause,
  [56846] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(815), 1,
      sym_declaration_list,
    STATE(2354), 1,
      sym_class_interface_clause,
  [56862] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(816), 1,
      sym_declaration_list,
    STATE(2355), 1,
      sym_class_interface_clause,
  [56878] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1832), 4,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [56888] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3794), 4,
      aux_sym_namespace_use_declaration_token1,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_EQ_GT,
  [56898] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4002), 1,
      sym__new_line,
    ACTIONS(4059), 1,
      anon_sym_DQUOTE2,
    ACTIONS(4061), 1,
      sym_heredoc_end,
    STATE(2283), 1,
      sym_heredoc_body,
  [56914] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    STATE(956), 1,
      sym_declaration_list,
    STATE(2314), 1,
      sym_class_interface_clause,
  [56930] = 3,
    ACTIONS(675), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(595), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [56942] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3636), 1,
      anon_sym_LBRACE,
    STATE(578), 1,
      sym_enum_declaration_list,
    STATE(2396), 1,
      sym_class_interface_clause,
  [56958] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2088), 1,
      sym_declaration_list,
    STATE(2362), 1,
      sym_class_interface_clause,
  [56974] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4063), 1,
      anon_sym_AMP,
    STATE(1785), 1,
      aux_sym_intersection_type_repeat1,
    ACTIONS(3291), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOLLAR,
  [56988] = 4,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(1679), 2,
      sym_variable_name,
      sym_by_ref,
  [57002] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2101), 1,
      sym_declaration_list,
    STATE(2364), 1,
      sym_class_interface_clause,
  [57018] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3268), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACE,
  [57028] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2557), 1,
      sym_declare_directive,
    ACTIONS(3933), 3,
      anon_sym_ticks,
      anon_sym_encoding,
      anon_sym_strict_types,
  [57040] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3239), 4,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PIPE,
  [57050] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1828), 1,
      anon_sym_DOLLAR,
    STATE(763), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [57062] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3636), 1,
      anon_sym_LBRACE,
    STATE(527), 1,
      sym_enum_declaration_list,
    STATE(2326), 1,
      sym_class_interface_clause,
  [57078] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4068), 1,
      aux_sym_string_content_token1,
    STATE(1812), 1,
      aux_sym_string_content_repeat1,
    ACTIONS(4066), 2,
      anon_sym_SQUOTE,
      aux_sym_string_token2,
  [57092] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4002), 1,
      sym__new_line,
    ACTIONS(4070), 1,
      anon_sym_DQUOTE2,
    ACTIONS(4072), 1,
      sym_heredoc_end,
    STATE(2335), 1,
      sym_heredoc_body,
  [57108] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(825), 1,
      sym_declaration_list,
    STATE(2368), 1,
      sym_class_interface_clause,
  [57124] = 4,
    ACTIONS(793), 1,
      anon_sym_AMP,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(2352), 2,
      sym_variable_name,
      sym_by_ref,
  [57138] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3983), 1,
      anon_sym_COMMA,
    STATE(1787), 1,
      aux_sym_base_clause_repeat1,
    ACTIONS(4074), 2,
      anon_sym_LBRACE,
      aux_sym_class_interface_clause_token1,
  [57152] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4078), 1,
      anon_sym_COMMA,
    STATE(1871), 1,
      aux_sym_global_declaration_repeat1,
    ACTIONS(4076), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57166] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4083), 1,
      anon_sym_EQ,
    STATE(1395), 1,
      sym__semicolon,
    ACTIONS(4081), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57180] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2205), 1,
      sym_declaration_list,
    STATE(2300), 1,
      sym_class_interface_clause,
  [57196] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2584), 1,
      sym_declare_directive,
    ACTIONS(3933), 3,
      anon_sym_ticks,
      anon_sym_encoding,
      anon_sym_strict_types,
  [57208] = 5,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3351), 1,
      aux_sym_class_interface_clause_token1,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(824), 1,
      sym_declaration_list,
    STATE(2370), 1,
      sym_class_interface_clause,
  [57224] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4087), 1,
      anon_sym_LPAREN2,
    STATE(464), 1,
      sym__semicolon,
    ACTIONS(4085), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57238] = 3,
    ACTIONS(125), 1,
      anon_sym_DOLLAR,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2144), 3,
      sym_dynamic_variable_name,
      sym__simple_variable,
      sym_variable_name,
  [57250] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(502), 1,
      sym__semicolon,
    ACTIONS(4089), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57261] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3983), 1,
      anon_sym_COMMA,
    ACTIONS(4091), 1,
      anon_sym_LBRACE,
    STATE(2058), 1,
      aux_sym_base_clause_repeat1,
  [57274] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4093), 1,
      anon_sym_COMMA,
    ACTIONS(4095), 1,
      anon_sym_RPAREN2,
    STATE(1908), 1,
      aux_sym_formal_parameters_repeat1,
  [57287] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4097), 1,
      anon_sym_LBRACE,
    STATE(2571), 1,
      sym__return_type,
  [57300] = 3,
    ACTIONS(1440), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1438), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57311] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    ACTIONS(4101), 1,
      anon_sym_RPAREN2,
    STATE(1911), 1,
      aux_sym_array_creation_expression_repeat1,
  [57324] = 3,
    ACTIONS(1448), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1446), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57335] = 3,
    ACTIONS(1100), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1098), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57346] = 3,
    ACTIONS(1352), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1350), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57357] = 3,
    ACTIONS(1264), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1262), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57368] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4103), 1,
      anon_sym_LBRACE,
    ACTIONS(4105), 1,
      anon_sym_COLON,
    STATE(460), 1,
      sym_switch_block,
  [57381] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4107), 1,
      anon_sym_EQ_GT,
    STATE(2523), 1,
      sym__return_type,
  [57394] = 3,
    ACTIONS(1472), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1470), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57405] = 3,
    ACTIONS(1476), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1474), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57416] = 3,
    ACTIONS(1484), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1482), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57427] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4109), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [57436] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4111), 1,
      anon_sym_COMMA,
    ACTIONS(4114), 1,
      anon_sym_RPAREN2,
    STATE(1894), 1,
      aux_sym_arguments_repeat1,
  [57449] = 4,
    ACTIONS(875), 1,
      anon_sym_RBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4116), 1,
      anon_sym_COMMA,
    STATE(2065), 1,
      aux_sym_array_creation_expression_repeat1,
  [57462] = 3,
    ACTIONS(1308), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1306), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57473] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3184), 1,
      anon_sym_RBRACK,
    ACTIONS(4118), 1,
      anon_sym_COMMA,
    STATE(1966), 1,
      aux_sym_attribute_group_repeat1,
  [57486] = 3,
    ACTIONS(1496), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1494), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57497] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4002), 1,
      sym__new_line,
    ACTIONS(4120), 1,
      sym_heredoc_end,
    STATE(2322), 1,
      sym_heredoc_body,
  [57510] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1918), 1,
      sym__semicolon,
    ACTIONS(4122), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57521] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1919), 1,
      sym__semicolon,
    ACTIONS(4124), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57532] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4126), 1,
      anon_sym_COMMA,
    ACTIONS(4128), 1,
      anon_sym_RPAREN2,
    STATE(1925), 1,
      aux_sym_arguments_repeat1,
  [57545] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1921), 1,
      sym__semicolon,
    ACTIONS(4130), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57556] = 3,
    ACTIONS(1524), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1522), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57567] = 3,
    ACTIONS(1548), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1546), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57578] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1928), 1,
      sym__semicolon,
    ACTIONS(4132), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57589] = 3,
    ACTIONS(1572), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1570), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57600] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2096), 1,
      anon_sym_RPAREN2,
    ACTIONS(4134), 1,
      anon_sym_COMMA,
    STATE(1992), 1,
      aux_sym_formal_parameters_repeat1,
  [57613] = 3,
    ACTIONS(1104), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1102), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57624] = 4,
    ACTIONS(857), 1,
      anon_sym_RPAREN2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4136), 1,
      anon_sym_COMMA,
    STATE(2184), 1,
      aux_sym_array_creation_expression_repeat1,
  [57637] = 4,
    ACTIONS(871), 1,
      anon_sym_RPAREN2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4138), 1,
      anon_sym_COMMA,
    STATE(2184), 1,
      aux_sym_array_creation_expression_repeat1,
  [57650] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4140), 1,
      anon_sym_LBRACE,
    STATE(2437), 1,
      sym__return_type,
  [57663] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1811), 1,
      anon_sym_RPAREN2,
    ACTIONS(4142), 1,
      anon_sym_COMMA,
    STATE(2071), 1,
      aux_sym_unset_statement_repeat1,
  [57676] = 3,
    ACTIONS(1112), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1110), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57687] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4144), 1,
      anon_sym_COMMA,
    ACTIONS(4146), 1,
      anon_sym_RBRACE,
    STATE(1934), 1,
      aux_sym_match_block_repeat1,
  [57700] = 3,
    ACTIONS(1124), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1122), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57711] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1937), 1,
      sym__semicolon,
    ACTIONS(4148), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57722] = 3,
    ACTIONS(1128), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1126), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57733] = 3,
    ACTIONS(1144), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1142), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57744] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1938), 1,
      sym__semicolon,
    ACTIONS(4150), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57755] = 3,
    ACTIONS(1148), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1146), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57766] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1939), 1,
      sym__semicolon,
    ACTIONS(4152), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57777] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4002), 1,
      sym__new_line,
    ACTIONS(4154), 1,
      sym_heredoc_end,
    STATE(2359), 1,
      sym_heredoc_body,
  [57790] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4156), 1,
      sym__new_line,
    ACTIONS(4158), 1,
      sym_heredoc_end,
    STATE(2360), 1,
      sym_nowdoc_body,
  [57803] = 4,
    ACTIONS(819), 1,
      anon_sym_RPAREN2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4160), 1,
      anon_sym_COMMA,
    STATE(1894), 1,
      aux_sym_arguments_repeat1,
  [57816] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1941), 1,
      sym__semicolon,
    ACTIONS(4162), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57827] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(525), 1,
      sym__semicolon,
    ACTIONS(4164), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57838] = 3,
    ACTIONS(1156), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1154), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57849] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1943), 1,
      sym__semicolon,
    ACTIONS(4166), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57860] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1946), 1,
      sym__semicolon,
    ACTIONS(4168), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57871] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(526), 1,
      sym__semicolon,
    ACTIONS(4170), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57882] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4172), 1,
      anon_sym_LBRACE,
    STATE(2438), 1,
      sym__return_type,
  [57895] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1948), 1,
      sym__semicolon,
    ACTIONS(4174), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [57906] = 4,
    ACTIONS(901), 1,
      anon_sym_RBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4176), 1,
      anon_sym_COMMA,
    STATE(2081), 1,
      aux_sym_match_block_repeat1,
  [57919] = 3,
    ACTIONS(1224), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1222), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57930] = 3,
    ACTIONS(1228), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1226), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57941] = 3,
    ACTIONS(1232), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1230), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57952] = 3,
    ACTIONS(1236), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1234), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57963] = 3,
    ACTIONS(1240), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1238), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57974] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1809), 1,
      anon_sym_RPAREN2,
    ACTIONS(4178), 1,
      anon_sym_COMMA,
    STATE(2071), 1,
      aux_sym_unset_statement_repeat1,
  [57987] = 3,
    ACTIONS(1244), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1242), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [57998] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1952), 1,
      sym__semicolon,
    ACTIONS(4180), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58009] = 3,
    ACTIONS(1248), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1246), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58020] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4182), 1,
      anon_sym_LBRACE,
    STATE(2444), 1,
      sym__return_type,
  [58033] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4184), 1,
      anon_sym_EQ_GT,
    STATE(2445), 1,
      sym__return_type,
  [58046] = 3,
    ACTIONS(1260), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1258), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58057] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1953), 1,
      sym__semicolon,
    ACTIONS(4186), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58068] = 3,
    ACTIONS(1268), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1266), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58079] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4188), 1,
      anon_sym_COMMA,
    ACTIONS(4190), 1,
      anon_sym_RBRACE,
    STATE(2159), 1,
      aux_sym_namespace_use_declaration_repeat1,
  [58092] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1955), 1,
      sym__semicolon,
    ACTIONS(4192), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58103] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1956), 1,
      sym__semicolon,
    ACTIONS(4194), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58114] = 3,
    ACTIONS(1280), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1278), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58125] = 3,
    ACTIONS(1284), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1282), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58136] = 3,
    ACTIONS(1512), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1510), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58147] = 3,
    ACTIONS(1288), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1286), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58158] = 3,
    ACTIONS(1292), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1290), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58169] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1958), 1,
      sym__semicolon,
    ACTIONS(4196), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58180] = 3,
    ACTIONS(1296), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1294), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58191] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2035), 1,
      sym__semicolon,
    ACTIONS(3529), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58202] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4198), 1,
      anon_sym_EQ_GT,
    STATE(2653), 1,
      sym__return_type,
  [58215] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(501), 1,
      sym__semicolon,
    ACTIONS(4200), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58226] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(4202), 1,
      sym_name,
    STATE(2469), 1,
      sym_reference_modifier,
  [58239] = 3,
    ACTIONS(1340), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1338), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58250] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4204), 1,
      anon_sym_COMMA,
    ACTIONS(4206), 1,
      anon_sym_RPAREN2,
    STATE(1969), 1,
      aux_sym_arguments_repeat1,
  [58263] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(3764), 1,
      anon_sym_LBRACE,
    STATE(2642), 1,
      sym__return_type,
  [58276] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4208), 1,
      anon_sym_COMMA,
    ACTIONS(4211), 1,
      anon_sym_RBRACK,
    STATE(1966), 1,
      aux_sym_attribute_group_repeat1,
  [58289] = 3,
    ACTIONS(1108), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1106), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58300] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1997), 1,
      sym__semicolon,
    ACTIONS(4213), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58311] = 4,
    ACTIONS(823), 1,
      anon_sym_RPAREN2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4215), 1,
      anon_sym_COMMA,
    STATE(1894), 1,
      aux_sym_arguments_repeat1,
  [58324] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4156), 1,
      sym__new_line,
    ACTIONS(4217), 1,
      sym_heredoc_end,
    STATE(2252), 1,
      sym_nowdoc_body,
  [58337] = 3,
    ACTIONS(1152), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1150), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58348] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3147), 1,
      anon_sym_AMP,
    ACTIONS(4219), 1,
      anon_sym_RPAREN2,
    STATE(1466), 1,
      aux_sym_intersection_type_repeat1,
  [58361] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(4221), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1993), 1,
      sym_variable_name,
  [58374] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4223), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [58383] = 3,
    ACTIONS(1276), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1274), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58394] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4188), 1,
      anon_sym_COMMA,
    ACTIONS(4225), 1,
      anon_sym_RBRACE,
    STATE(1949), 1,
      aux_sym_namespace_use_declaration_repeat1,
  [58407] = 3,
    ACTIONS(1176), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1174), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58418] = 3,
    ACTIONS(1216), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1214), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58429] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4227), 1,
      aux_sym_namespace_use_clause_token1,
    ACTIONS(4008), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [58440] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(503), 1,
      sym__semicolon,
    ACTIONS(4229), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58451] = 3,
    ACTIONS(1348), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1346), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58462] = 3,
    ACTIONS(1220), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1218), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58473] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4231), 1,
      anon_sym_COMMA,
    ACTIONS(4233), 1,
      anon_sym_RPAREN2,
    STATE(1988), 1,
      aux_sym_arguments_repeat1,
  [58486] = 3,
    ACTIONS(1584), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1582), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58497] = 3,
    ACTIONS(1272), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1270), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58508] = 3,
    ACTIONS(1356), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1354), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58519] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4237), 1,
      anon_sym_EQ,
    ACTIONS(4235), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [58530] = 4,
    ACTIONS(827), 1,
      anon_sym_RPAREN2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4239), 1,
      anon_sym_COMMA,
    STATE(1894), 1,
      aux_sym_arguments_repeat1,
  [58543] = 3,
    ACTIONS(1376), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1374), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58554] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(504), 1,
      sym__semicolon,
    ACTIONS(4241), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58565] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2121), 1,
      sym__semicolon,
    ACTIONS(4243), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58576] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4245), 1,
      anon_sym_COMMA,
    ACTIONS(4248), 1,
      anon_sym_RPAREN2,
    STATE(1992), 1,
      aux_sym_formal_parameters_repeat1,
  [58589] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4252), 1,
      anon_sym_EQ,
    ACTIONS(4250), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [58600] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4256), 1,
      anon_sym_EQ,
    ACTIONS(4254), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [58611] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(4258), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1998), 1,
      sym_variable_name,
  [58624] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(499), 1,
      sym__semicolon,
    ACTIONS(4260), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58635] = 3,
    ACTIONS(1380), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1378), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58646] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4264), 1,
      anon_sym_EQ,
    ACTIONS(4262), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [58657] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4266), 1,
      anon_sym_COMMA,
    ACTIONS(4268), 1,
      anon_sym_RPAREN2,
    STATE(2004), 1,
      aux_sym_arguments_repeat1,
  [58670] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4272), 1,
      anon_sym_EQ,
    ACTIONS(4270), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [58681] = 3,
    ACTIONS(1404), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1402), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58692] = 3,
    ACTIONS(1416), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1414), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58703] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(4274), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(1994), 1,
      sym_variable_name,
  [58716] = 4,
    ACTIONS(829), 1,
      anon_sym_RPAREN2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4276), 1,
      anon_sym_COMMA,
    STATE(1894), 1,
      aux_sym_arguments_repeat1,
  [58729] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4280), 1,
      anon_sym_EQ,
    ACTIONS(4278), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [58740] = 3,
    ACTIONS(1116), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1114), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58751] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4282), 1,
      anon_sym_COMMA,
    ACTIONS(4284), 1,
      anon_sym_RPAREN2,
    STATE(2009), 1,
      aux_sym_anonymous_function_use_clause_repeat1,
  [58764] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4286), 1,
      aux_sym_namespace_use_clause_token1,
    ACTIONS(3921), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [58775] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3778), 1,
      anon_sym_RPAREN2,
    ACTIONS(4288), 1,
      anon_sym_COMMA,
    STATE(2125), 1,
      aux_sym_anonymous_function_use_clause_repeat1,
  [58788] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4290), 1,
      anon_sym_LBRACE,
    STATE(2500), 1,
      sym__return_type,
  [58801] = 3,
    ACTIONS(1336), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1334), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58812] = 3,
    ACTIONS(1420), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1418), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58823] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4292), 1,
      sym_name,
    ACTIONS(4294), 1,
      anon_sym_LBRACE,
    STATE(2229), 1,
      sym_namespace_use_group,
  [58836] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4296), 1,
      anon_sym_COMMA,
    ACTIONS(4298), 1,
      anon_sym_RPAREN2,
    STATE(2018), 1,
      aux_sym_arguments_repeat1,
  [58849] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(536), 1,
      sym__semicolon,
    ACTIONS(4300), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58860] = 3,
    ACTIONS(1492), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1490), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58871] = 3,
    ACTIONS(1500), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1498), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58882] = 4,
    ACTIONS(835), 1,
      anon_sym_RPAREN2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4302), 1,
      anon_sym_COMMA,
    STATE(1894), 1,
      aux_sym_arguments_repeat1,
  [58895] = 3,
    ACTIONS(1508), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1506), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58906] = 3,
    ACTIONS(1516), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1514), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58917] = 3,
    ACTIONS(1432), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1430), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58928] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(537), 1,
      sym__semicolon,
    ACTIONS(4304), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [58939] = 3,
    ACTIONS(1436), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1434), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58950] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3327), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [58959] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(3874), 1,
      anon_sym_LBRACE,
    STATE(2507), 1,
      sym__return_type,
  [58972] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3949), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [58981] = 3,
    ACTIONS(1444), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1442), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [58992] = 3,
    ACTIONS(1540), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1538), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59003] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2163), 1,
      sym__semicolon,
    ACTIONS(4306), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59014] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2045), 1,
      sym__semicolon,
    ACTIONS(4308), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59025] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2169), 1,
      sym__semicolon,
    ACTIONS(4310), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59036] = 3,
    ACTIONS(1552), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1550), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59047] = 3,
    ACTIONS(1556), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1554), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59058] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4312), 1,
      anon_sym_DQUOTE,
    ACTIONS(4314), 1,
      anon_sym_SQUOTE,
    ACTIONS(4316), 1,
      sym_heredoc_start,
  [59071] = 3,
    ACTIONS(1464), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1462), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59082] = 3,
    ACTIONS(1560), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1558), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59093] = 3,
    ACTIONS(1564), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1562), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59104] = 3,
    ACTIONS(1568), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1566), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59115] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1852), 1,
      anon_sym_DOLLAR,
    STATE(1518), 1,
      sym_property_element,
    STATE(1713), 1,
      sym_variable_name,
  [59128] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2172), 1,
      sym__semicolon,
    ACTIONS(4318), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59139] = 3,
    ACTIONS(1132), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1130), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59150] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1852), 1,
      anon_sym_DOLLAR,
    STATE(1528), 1,
      sym_property_element,
    STATE(1713), 1,
      sym_variable_name,
  [59163] = 3,
    ACTIONS(1120), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1118), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59174] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1852), 1,
      anon_sym_DOLLAR,
    STATE(1713), 1,
      sym_variable_name,
    STATE(1831), 1,
      sym_property_element,
  [59187] = 3,
    ACTIONS(1212), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1210), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59198] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3976), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [59207] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2032), 1,
      sym__semicolon,
    ACTIONS(4320), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59218] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(543), 1,
      sym__semicolon,
    ACTIONS(4322), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59229] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(505), 1,
      sym__semicolon,
    ACTIONS(4324), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59240] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4328), 1,
      aux_sym_else_clause_token1,
    ACTIONS(4326), 2,
      aux_sym_if_statement_token2,
      aux_sym_else_if_clause_token1,
  [59251] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3631), 3,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      aux_sym_class_interface_clause_token1,
  [59260] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2002), 1,
      sym__semicolon,
    ACTIONS(4330), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59271] = 4,
    ACTIONS(837), 1,
      anon_sym_RPAREN2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4332), 1,
      anon_sym_COMMA,
    STATE(1894), 1,
      aux_sym_arguments_repeat1,
  [59284] = 3,
    ACTIONS(1400), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1398), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59295] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2012), 1,
      sym__semicolon,
    ACTIONS(4334), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59306] = 3,
    ACTIONS(1136), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1134), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59317] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(4336), 1,
      anon_sym_DOT_DOT_DOT,
    STATE(2077), 1,
      sym_variable_name,
  [59330] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3983), 1,
      anon_sym_COMMA,
    ACTIONS(4338), 1,
      anon_sym_LBRACE,
    STATE(1787), 1,
      aux_sym_base_clause_repeat1,
  [59343] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(570), 1,
      sym__semicolon,
    ACTIONS(4340), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59354] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(4342), 1,
      anon_sym_RPAREN2,
    STATE(2532), 1,
      sym_variable_name,
  [59367] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(577), 1,
      sym__semicolon,
    ACTIONS(4344), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59378] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4346), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [59387] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(509), 1,
      sym__semicolon,
    ACTIONS(4348), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59398] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(513), 1,
      sym__semicolon,
    ACTIONS(4350), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59409] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4352), 1,
      anon_sym_COMMA,
    ACTIONS(4355), 1,
      anon_sym_RBRACK,
    STATE(2065), 1,
      aux_sym_array_creation_expression_repeat1,
  [59422] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4355), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
      anon_sym_RBRACK,
  [59431] = 3,
    ACTIONS(1140), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1138), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59442] = 4,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1789), 1,
      anon_sym_RPAREN2,
    STATE(2130), 1,
      aux_sym__list_destructing_repeat1,
  [59455] = 4,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1789), 1,
      anon_sym_RPAREN2,
    STATE(2133), 1,
      aux_sym__list_destructing_repeat1,
  [59468] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4357), 1,
      anon_sym_COMMA,
    ACTIONS(4359), 1,
      anon_sym_RBRACE,
    STATE(2107), 1,
      aux_sym_match_block_repeat1,
  [59481] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3453), 1,
      anon_sym_RPAREN2,
    ACTIONS(4361), 1,
      anon_sym_COMMA,
    STATE(2071), 1,
      aux_sym_unset_statement_repeat1,
  [59494] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4364), 1,
      anon_sym_EQ_GT,
    STATE(2422), 1,
      sym__return_type,
  [59507] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4366), 1,
      sym_name,
    ACTIONS(4368), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59518] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4372), 1,
      aux_sym_else_clause_token1,
    ACTIONS(4370), 2,
      aux_sym_if_statement_token2,
      aux_sym_else_if_clause_token1,
  [59529] = 3,
    ACTIONS(1208), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1206), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59540] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(455), 1,
      sym__semicolon,
    ACTIONS(4374), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59551] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4378), 1,
      anon_sym_EQ,
    ACTIONS(4376), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [59562] = 3,
    ACTIONS(1488), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1486), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59573] = 3,
    ACTIONS(1504), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1502), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59584] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4380), 1,
      anon_sym_COMMA,
    ACTIONS(4382), 1,
      anon_sym_RBRACK,
    STATE(2084), 1,
      aux_sym__array_destructing_repeat1,
  [59597] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4384), 1,
      anon_sym_COMMA,
    ACTIONS(4387), 1,
      anon_sym_RBRACE,
    STATE(2081), 1,
      aux_sym_match_block_repeat1,
  [59610] = 4,
    ACTIONS(867), 1,
      anon_sym_RBRACK,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4389), 1,
      anon_sym_COMMA,
    STATE(2065), 1,
      aux_sym_array_creation_expression_repeat1,
  [59623] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2576), 1,
      anon_sym_EQ_GT,
    ACTIONS(4391), 1,
      anon_sym_COMMA,
    STATE(2083), 1,
      aux_sym_match_condition_list_repeat1,
  [59636] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4394), 1,
      anon_sym_COMMA,
    ACTIONS(4397), 1,
      anon_sym_RBRACK,
    STATE(2084), 1,
      aux_sym__array_destructing_repeat1,
  [59649] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    ACTIONS(4399), 1,
      anon_sym_RPAREN2,
    STATE(2549), 1,
      sym_variable_name,
  [59662] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(548), 1,
      sym__semicolon,
    ACTIONS(3589), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59673] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3194), 1,
      anon_sym_RBRACK,
    ACTIONS(4401), 1,
      anon_sym_COMMA,
    STATE(1966), 1,
      aux_sym_attribute_group_repeat1,
  [59686] = 3,
    ACTIONS(1520), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1518), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59697] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3291), 1,
      anon_sym_DOLLAR,
    ACTIONS(4403), 1,
      anon_sym_AMP,
    STATE(1458), 1,
      aux_sym_intersection_type_repeat1,
  [59710] = 3,
    ACTIONS(1528), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1526), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59721] = 3,
    ACTIONS(1532), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1530), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59732] = 3,
    ACTIONS(1252), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1250), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59743] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4406), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [59752] = 4,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4408), 1,
      anon_sym_RPAREN2,
    STATE(2133), 1,
      aux_sym__list_destructing_repeat1,
  [59765] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4410), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [59774] = 4,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4412), 1,
      anon_sym_RPAREN2,
    STATE(2133), 1,
      aux_sym__list_destructing_repeat1,
  [59787] = 3,
    ACTIONS(1452), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1450), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59798] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(547), 1,
      sym__semicolon,
    ACTIONS(4414), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59809] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(549), 1,
      sym__semicolon,
    ACTIONS(4416), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59820] = 3,
    ACTIONS(1536), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1534), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59831] = 3,
    ACTIONS(1544), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1542), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59842] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4418), 1,
      anon_sym_EQ_GT,
    STATE(2529), 1,
      sym__return_type,
  [59855] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4420), 1,
      anon_sym_LBRACE,
    STATE(2572), 1,
      sym__return_type,
  [59868] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(470), 1,
      sym__semicolon,
    ACTIONS(4422), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59879] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4002), 1,
      sym__new_line,
    ACTIONS(4424), 1,
      sym_heredoc_end,
    STATE(2346), 1,
      sym_heredoc_body,
  [59892] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4426), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [59901] = 4,
    ACTIONS(887), 1,
      anon_sym_RBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4428), 1,
      anon_sym_COMMA,
    STATE(2081), 1,
      aux_sym_match_block_repeat1,
  [59914] = 4,
    ACTIONS(1021), 1,
      anon_sym_EQ_GT,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4430), 1,
      anon_sym_COMMA,
    STATE(2083), 1,
      aux_sym_match_condition_list_repeat1,
  [59927] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1852), 1,
      anon_sym_DOLLAR,
    STATE(1844), 1,
      sym_variable_name,
    STATE(2046), 1,
      sym_static_variable_declaration,
  [59940] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4432), 1,
      anon_sym_COMMA,
    ACTIONS(4434), 1,
      anon_sym_RPAREN2,
    STATE(1910), 1,
      aux_sym_array_creation_expression_repeat1,
  [59953] = 3,
    ACTIONS(1256), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1254), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [59964] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(514), 1,
      sym__semicolon,
    ACTIONS(4436), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59975] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(484), 1,
      sym__semicolon,
    ACTIONS(4438), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [59986] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1899), 1,
      anon_sym_AMP,
    ACTIONS(4440), 1,
      sym_name,
    STATE(2606), 1,
      sym_reference_modifier,
  [59999] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(489), 1,
      sym__semicolon,
    ACTIONS(4442), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60010] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1989), 1,
      sym__semicolon,
    ACTIONS(4444), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60021] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(490), 1,
      sym__semicolon,
    ACTIONS(4446), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60032] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4002), 1,
      sym__new_line,
    ACTIONS(4448), 1,
      sym_heredoc_end,
    STATE(2236), 1,
      sym_heredoc_body,
  [60045] = 3,
    ACTIONS(1576), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1574), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60056] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4450), 1,
      anon_sym_COMMA,
    ACTIONS(4452), 1,
      anon_sym_RPAREN2,
    STATE(2208), 1,
      aux_sym_formal_parameters_repeat1,
  [60069] = 3,
    ACTIONS(1580), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1578), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60080] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3964), 1,
      anon_sym_EQ,
    ACTIONS(4454), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [60091] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(491), 1,
      sym__semicolon,
    ACTIONS(4456), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60102] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(492), 1,
      sym__semicolon,
    ACTIONS(4458), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60113] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4460), 1,
      anon_sym_COMMA,
    ACTIONS(4463), 1,
      anon_sym_RPAREN2,
    STATE(2125), 1,
      aux_sym_anonymous_function_use_clause_repeat1,
  [60126] = 4,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4465), 1,
      anon_sym_RPAREN2,
    STATE(2096), 1,
      aux_sym__list_destructing_repeat1,
  [60139] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4467), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [60148] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1351), 1,
      sym__semicolon,
    ACTIONS(4469), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60159] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4026), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [60168] = 4,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4471), 1,
      anon_sym_RPAREN2,
    STATE(2133), 1,
      aux_sym__list_destructing_repeat1,
  [60181] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4473), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [60190] = 4,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4475), 1,
      anon_sym_RPAREN2,
    STATE(2133), 1,
      aux_sym__list_destructing_repeat1,
  [60203] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1800), 1,
      anon_sym_RPAREN2,
    ACTIONS(4477), 1,
      anon_sym_COMMA,
    STATE(2133), 1,
      aux_sym__list_destructing_repeat1,
  [60216] = 3,
    ACTIONS(1160), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1158), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60227] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4480), 1,
      anon_sym_EQ_GT,
    STATE(2476), 1,
      sym__return_type,
  [60240] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2188), 1,
      sym__semicolon,
    ACTIONS(4482), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60251] = 4,
    ACTIONS(889), 1,
      anon_sym_COMMA,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3365), 1,
      anon_sym_RPAREN2,
    STATE(2094), 1,
      aux_sym__list_destructing_repeat1,
  [60264] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4484), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [60273] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4486), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [60282] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2189), 1,
      sym__semicolon,
    ACTIONS(4488), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60293] = 3,
    ACTIONS(1164), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1162), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60304] = 3,
    ACTIONS(1368), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1366), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60315] = 3,
    ACTIONS(1168), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1166), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60326] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4076), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [60335] = 3,
    ACTIONS(1480), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1478), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60346] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(553), 1,
      sym__semicolon,
    ACTIONS(4490), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60357] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4031), 3,
      anon_sym_RPAREN2,
      anon_sym_PIPE,
      anon_sym_DOLLAR,
  [60366] = 3,
    ACTIONS(1300), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1298), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60377] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3964), 1,
      anon_sym_EQ,
    ACTIONS(4492), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [60388] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(556), 1,
      sym__semicolon,
    ACTIONS(4494), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60399] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4380), 1,
      anon_sym_COMMA,
    ACTIONS(4496), 1,
      anon_sym_RBRACK,
    STATE(2080), 1,
      aux_sym__array_destructing_repeat1,
  [60412] = 4,
    ACTIONS(983), 1,
      anon_sym_COLON,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2134), 1,
      anon_sym_LPAREN2,
    STATE(2230), 1,
      sym_parenthesized_expression,
  [60425] = 3,
    ACTIONS(1304), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1302), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60436] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4498), 1,
      anon_sym_COMMA,
    ACTIONS(4500), 1,
      anon_sym_RBRACK,
    STATE(2082), 1,
      aux_sym_array_creation_expression_repeat1,
  [60449] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4502), 1,
      anon_sym_COMMA,
    ACTIONS(4504), 1,
      anon_sym_RPAREN2,
    STATE(2053), 1,
      aux_sym_arguments_repeat1,
  [60462] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4380), 1,
      anon_sym_COMMA,
    ACTIONS(4496), 1,
      anon_sym_RBRACK,
    STATE(2084), 1,
      aux_sym__array_destructing_repeat1,
  [60475] = 3,
    ACTIONS(1172), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1170), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60486] = 3,
    ACTIONS(1316), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1314), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60497] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3949), 1,
      anon_sym_RBRACE,
    ACTIONS(4506), 1,
      anon_sym_COMMA,
    STATE(2159), 1,
      aux_sym_namespace_use_declaration_repeat1,
  [60510] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4509), 1,
      anon_sym_COMMA,
    ACTIONS(4511), 1,
      anon_sym_RBRACK,
    STATE(2087), 1,
      aux_sym_attribute_group_repeat1,
  [60523] = 3,
    ACTIONS(1360), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1358), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60534] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4513), 1,
      anon_sym_EQ_GT,
    STATE(2513), 1,
      sym__return_type,
  [60547] = 3,
    ACTIONS(1460), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1458), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60558] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4515), 1,
      anon_sym_DQUOTE,
    ACTIONS(4517), 1,
      anon_sym_SQUOTE,
    ACTIONS(4519), 1,
      sym_heredoc_start,
  [60571] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(493), 1,
      sym__semicolon,
    ACTIONS(4521), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60582] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4523), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [60591] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4525), 3,
      anon_sym_COMMA,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [60600] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(496), 1,
      sym__semicolon,
    ACTIONS(4527), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60611] = 3,
    ACTIONS(1468), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1466), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60622] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1885), 1,
      sym__semicolon,
    ACTIONS(4529), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60633] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4533), 1,
      anon_sym_EQ,
    ACTIONS(4531), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [60644] = 3,
    ACTIONS(1332), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1330), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60655] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4535), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [60664] = 3,
    ACTIONS(1320), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1318), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60675] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4537), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [60684] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2027), 1,
      sym__semicolon,
    ACTIONS(4539), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60695] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4541), 1,
      anon_sym_LBRACE,
    STATE(2483), 1,
      sym__return_type,
  [60708] = 3,
    ACTIONS(1384), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1382), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60719] = 3,
    ACTIONS(1396), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1394), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60730] = 3,
    ACTIONS(1424), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1422), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60741] = 3,
    ACTIONS(1428), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1426), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60752] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1985), 1,
      sym__semicolon,
    ACTIONS(4543), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60763] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(498), 1,
      sym__semicolon,
    ACTIONS(4545), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60774] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4355), 1,
      anon_sym_RPAREN2,
    ACTIONS(4547), 1,
      anon_sym_COMMA,
    STATE(2184), 1,
      aux_sym_array_creation_expression_repeat1,
  [60787] = 3,
    ACTIONS(1180), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1178), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60798] = 3,
    ACTIONS(1456), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1454), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60809] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4292), 1,
      sym_name,
    ACTIONS(4294), 1,
      anon_sym_LBRACE,
    STATE(2361), 1,
      sym_namespace_use_group,
  [60822] = 3,
    ACTIONS(1312), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1310), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60833] = 3,
    ACTIONS(1328), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1326), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60844] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1892), 1,
      sym__semicolon,
    ACTIONS(4550), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60855] = 3,
    ACTIONS(1324), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1322), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60866] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2028), 1,
      sym__semicolon,
    ACTIONS(4552), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60877] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1898), 1,
      sym__semicolon,
    ACTIONS(4554), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [60888] = 3,
    ACTIONS(1364), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1362), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60899] = 3,
    ACTIONS(1372), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1370), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60910] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4556), 1,
      anon_sym_LBRACE,
    ACTIONS(4558), 1,
      anon_sym_COLON,
    STATE(2043), 1,
      sym_switch_block,
  [60923] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4560), 1,
      anon_sym_COMMA,
    ACTIONS(4562), 1,
      anon_sym_RBRACK,
    STATE(1895), 1,
      aux_sym_array_creation_expression_repeat1,
  [60936] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4564), 1,
      anon_sym_COMMA,
    ACTIONS(4566), 1,
      anon_sym_RBRACK,
    STATE(1897), 1,
      aux_sym_attribute_group_repeat1,
  [60949] = 3,
    ACTIONS(1388), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1386), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60960] = 3,
    ACTIONS(1184), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1182), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60971] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3964), 1,
      anon_sym_EQ,
    ACTIONS(4568), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [60982] = 3,
    ACTIONS(1192), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1190), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [60993] = 3,
    ACTIONS(1196), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1194), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [61004] = 3,
    ACTIONS(1200), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1198), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [61015] = 3,
    ACTIONS(1204), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1202), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [61026] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2754), 3,
      sym__automatic_semicolon,
      anon_sym_SEMI,
      sym_name,
  [61035] = 3,
    ACTIONS(1392), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1390), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [61046] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2100), 1,
      anon_sym_RPAREN2,
    ACTIONS(4570), 1,
      anon_sym_COMMA,
    STATE(1992), 1,
      aux_sym_formal_parameters_repeat1,
  [61059] = 3,
    ACTIONS(1408), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1406), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [61070] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(463), 1,
      sym__semicolon,
    ACTIONS(4572), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [61081] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(467), 1,
      sym__semicolon,
    ACTIONS(4574), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [61092] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(468), 1,
      sym__semicolon,
    ACTIONS(4576), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [61103] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4578), 1,
      anon_sym_EQ_GT,
    STATE(2433), 1,
      sym__return_type,
  [61116] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(567), 1,
      sym__semicolon,
    ACTIONS(4580), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [61127] = 3,
    ACTIONS(1412), 1,
      aux_sym_else_clause_token1,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1410), 2,
      aux_sym_while_statement_token1,
      aux_sym_else_if_clause_token1,
  [61138] = 4,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3613), 1,
      anon_sym_COLON,
    ACTIONS(4582), 1,
      anon_sym_LBRACE,
    STATE(2614), 1,
      sym__return_type,
  [61151] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4584), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [61159] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4586), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [61167] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(929), 1,
      sym_arguments,
  [61177] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2016), 1,
      sym_declaration_list,
  [61187] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4248), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [61195] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4588), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [61203] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(479), 1,
      sym_declaration_list,
  [61213] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3636), 1,
      anon_sym_LBRACE,
    STATE(558), 1,
      sym_enum_declaration_list,
  [61223] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(555), 1,
      sym_declaration_list,
  [61233] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4590), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [61241] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2144), 1,
      anon_sym_LPAREN2,
    STATE(949), 1,
      sym_arguments,
  [61251] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(2218), 1,
      sym_variable_name,
  [61261] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4592), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [61269] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4594), 1,
      anon_sym_LBRACE,
    STATE(993), 1,
      sym_match_block,
  [61279] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4596), 1,
      anon_sym_LBRACE,
    STATE(450), 1,
      sym_compound_statement,
  [61289] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4598), 1,
      anon_sym_LBRACE,
    STATE(1090), 1,
      sym_compound_statement,
  [61299] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    STATE(973), 1,
      sym_declaration_list,
  [61309] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4600), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [61317] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4602), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [61325] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4604), 1,
      sym__new_line,
    ACTIONS(4606), 1,
      sym_heredoc_end,
  [61335] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1854), 1,
      anon_sym_LPAREN2,
    STATE(720), 1,
      sym_arguments,
  [61345] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(2349), 1,
      sym_variable_name,
  [61355] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4608), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [61363] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(2222), 1,
      sym_variable_name,
  [61373] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4610), 2,
      anon_sym_string,
      anon_sym_int,
  [61381] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4612), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [61389] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(562), 1,
      sym_declaration_list,
  [61399] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1889), 1,
      sym_formal_parameters,
  [61409] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1854), 1,
      anon_sym_LPAREN2,
    STATE(726), 1,
      sym_arguments,
  [61419] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3636), 1,
      anon_sym_LBRACE,
    STATE(481), 1,
      sym_enum_declaration_list,
  [61429] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(2025), 1,
      sym_formal_parameters,
  [61439] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4614), 2,
      anon_sym_LBRACE,
      anon_sym_COLON,
  [61447] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4616), 1,
      sym_nowdoc_string,
    STATE(1807), 1,
      aux_sym_nowdoc_body_repeat1,
  [61457] = 3,
    ACTIONS(368), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2134), 1,
      sym_compound_statement,
  [61467] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1861), 1,
      anon_sym_LPAREN2,
    STATE(761), 1,
      sym_arguments,
  [61477] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4618), 1,
      sym__new_line,
    ACTIONS(4620), 1,
      sym_heredoc_end,
  [61487] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4622), 1,
      anon_sym_LBRACE,
    STATE(1103), 1,
      sym_match_block,
  [61497] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3307), 1,
      anon_sym_LPAREN2,
    STATE(1754), 1,
      sym_formal_parameters,
  [61507] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1861), 1,
      anon_sym_LPAREN2,
    STATE(771), 1,
      sym_arguments,
  [61517] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4114), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [61525] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2202), 1,
      sym_declaration_list,
  [61535] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3662), 1,
      anon_sym_LBRACE,
    STATE(2204), 1,
      sym_enum_declaration_list,
  [61545] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2536), 1,
      anon_sym_RPAREN2,
    ACTIONS(4624), 1,
      anon_sym_EQ,
  [61555] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2205), 1,
      sym_declaration_list,
  [61565] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(482), 1,
      sym_declaration_list,
  [61575] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2630), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [61583] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(520), 1,
      sym_declaration_list,
  [61593] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1664), 1,
      sym_formal_parameters,
  [61603] = 3,
    ACTIONS(368), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1716), 1,
      sym_compound_statement,
  [61613] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4626), 1,
      anon_sym_LPAREN2,
    STATE(2055), 1,
      sym_parenthesized_expression,
  [61623] = 3,
    ACTIONS(368), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1666), 1,
      sym_compound_statement,
  [61633] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4628), 1,
      anon_sym_LPAREN2,
    STATE(79), 1,
      sym_parenthesized_expression,
  [61643] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2134), 1,
      anon_sym_LPAREN2,
    STATE(2230), 1,
      sym_parenthesized_expression,
  [61653] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2134), 1,
      anon_sym_LPAREN2,
    STATE(2196), 1,
      sym_parenthesized_expression,
  [61663] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4223), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [61671] = 3,
    ACTIONS(537), 1,
      anon_sym_COLON,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2074), 1,
      sym_colon_block,
  [61681] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(808), 1,
      sym_declaration_list,
  [61691] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(2235), 1,
      sym_variable_name,
  [61701] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4628), 1,
      anon_sym_LPAREN2,
    STATE(115), 1,
      sym_parenthesized_expression,
  [61711] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(554), 1,
      sym_declaration_list,
  [61721] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3224), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(3226), 1,
      aux_sym__arrow_function_header_token1,
  [61731] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1726), 1,
      anon_sym_LPAREN2,
    STATE(619), 1,
      sym_arguments,
  [61741] = 3,
    ACTIONS(29), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(471), 1,
      sym_compound_statement,
  [61751] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4630), 2,
      anon_sym_string,
      anon_sym_int,
  [61759] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2134), 1,
      anon_sym_LPAREN2,
    STATE(2272), 1,
      sym_parenthesized_expression,
  [61769] = 3,
    ACTIONS(352), 1,
      anon_sym_COLON,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(2421), 1,
      sym_colon_block,
  [61779] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4632), 1,
      sym__new_line,
    ACTIONS(4634), 1,
      sym_heredoc_end,
  [61789] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3212), 1,
      aux_sym__namespace_use_type_token1,
    ACTIONS(3214), 1,
      aux_sym__arrow_function_header_token1,
  [61799] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    STATE(606), 1,
      sym_arguments,
  [61809] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1665), 1,
      sym_formal_parameters,
  [61819] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2148), 1,
      sym_declaration_list,
  [61829] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    STATE(802), 1,
      sym_arguments,
  [61839] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4636), 2,
      anon_sym_string,
      anon_sym_int,
  [61847] = 3,
    ACTIONS(29), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(508), 1,
      sym_compound_statement,
  [61857] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1817), 1,
      anon_sym_BSLASH,
    STATE(2358), 1,
      aux_sym_namespace_name_repeat1,
  [61867] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4638), 1,
      anon_sym_LPAREN2,
    ACTIONS(4640), 1,
      anon_sym_RPAREN2,
  [61877] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(2304), 1,
      sym_variable_name,
  [61887] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4628), 1,
      anon_sym_LPAREN2,
    STATE(55), 1,
      sym_parenthesized_expression,
  [61897] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4596), 1,
      anon_sym_LBRACE,
    STATE(454), 1,
      sym_compound_statement,
  [61907] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4642), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [61915] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3964), 2,
      anon_sym_EQ,
      anon_sym_RPAREN2,
  [61923] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1726), 1,
      anon_sym_LPAREN2,
    STATE(632), 1,
      sym_arguments,
  [61933] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2538), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [61941] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(1986), 1,
      sym_declaration_list,
  [61951] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1941), 1,
      anon_sym_LPAREN2,
    ACTIONS(3815), 1,
      anon_sym_COLON_COLON,
  [61961] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2494), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN2,
  [61969] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4397), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [61977] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4644), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [61985] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4628), 1,
      anon_sym_LPAREN2,
    STATE(109), 1,
      sym_parenthesized_expression,
  [61995] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    STATE(955), 1,
      sym_declaration_list,
  [62005] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4387), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [62013] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4646), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62021] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4648), 2,
      anon_sym_SEMI,
      anon_sym_COLON,
  [62029] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    STATE(956), 1,
      sym_declaration_list,
  [62039] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4650), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [62047] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(815), 1,
      sym_declaration_list,
  [62057] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1945), 1,
      sym_formal_parameters,
  [62067] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    STATE(925), 1,
      sym_declaration_list,
  [62077] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(816), 1,
      sym_declaration_list,
  [62087] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4368), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [62095] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    STATE(930), 1,
      sym_declaration_list,
  [62105] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2494), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [62113] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4652), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62121] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2624), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62129] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3455), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62137] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4654), 1,
      sym__new_line,
    ACTIONS(4656), 1,
      sym_heredoc_end,
  [62147] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(2386), 1,
      sym_variable_name,
  [62157] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    STATE(896), 1,
      sym_declaration_list,
  [62167] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4658), 1,
      sym__new_line,
    ACTIONS(4660), 1,
      sym_heredoc_end,
  [62177] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3636), 1,
      anon_sym_LBRACE,
    STATE(550), 1,
      sym_enum_declaration_list,
  [62187] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(2226), 1,
      sym_variable_name,
  [62197] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4535), 2,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
  [62205] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4662), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [62213] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(551), 1,
      sym_declaration_list,
  [62223] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3662), 1,
      anon_sym_LBRACE,
    STATE(2079), 1,
      sym_enum_declaration_list,
  [62233] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4665), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [62241] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1685), 1,
      sym_formal_parameters,
  [62251] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4667), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62259] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4669), 1,
      sym__new_line,
    ACTIONS(4671), 1,
      sym_heredoc_end,
  [62269] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2088), 1,
      sym_declaration_list,
  [62279] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4673), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62287] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4568), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [62295] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4628), 1,
      anon_sym_LPAREN2,
    STATE(107), 1,
      sym_parenthesized_expression,
  [62305] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2101), 1,
      sym_declaration_list,
  [62315] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    STATE(899), 1,
      sym_declaration_list,
  [62325] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3662), 1,
      anon_sym_LBRACE,
    STATE(2119), 1,
      sym_enum_declaration_list,
  [62335] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3098), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62343] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4626), 1,
      anon_sym_LPAREN2,
    STATE(2022), 1,
      sym_parenthesized_expression,
  [62353] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4675), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62361] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4677), 1,
      sym__new_line,
    ACTIONS(4679), 1,
      sym_heredoc_end,
  [62371] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4681), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62379] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(2102), 1,
      sym_formal_parameters,
  [62389] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4683), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62397] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1800), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62405] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4685), 2,
      anon_sym_LBRACE,
      anon_sym_COLON,
  [62413] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4463), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62421] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(506), 1,
      sym_declaration_list,
  [62431] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(824), 1,
      sym_declaration_list,
  [62441] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(825), 1,
      sym_declaration_list,
  [62451] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(568), 1,
      sym_declaration_list,
  [62461] = 3,
    ACTIONS(368), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1718), 1,
      sym_compound_statement,
  [62471] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4687), 1,
      anon_sym_BSLASH,
    STATE(2358), 1,
      aux_sym_namespace_name_repeat1,
  [62481] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4690), 1,
      sym__new_line,
    ACTIONS(4692), 1,
      sym_heredoc_end,
  [62491] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4694), 1,
      sym__new_line,
    ACTIONS(4696), 1,
      sym_heredoc_end,
  [62501] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4698), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [62509] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2179), 1,
      sym_declaration_list,
  [62519] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2180), 1,
      sym_declaration_list,
  [62529] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2186), 1,
      sym_declaration_list,
  [62539] = 3,
    ACTIONS(368), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1668), 1,
      sym_compound_statement,
  [62549] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(531), 1,
      sym_declaration_list,
  [62559] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3431), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [62567] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(749), 1,
      sym_declaration_list,
  [62577] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4596), 1,
      anon_sym_LBRACE,
    STATE(449), 1,
      sym_compound_statement,
  [62587] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(750), 1,
      sym_declaration_list,
  [62597] = 3,
    ACTIONS(983), 1,
      anon_sym_COLON,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4638), 1,
      anon_sym_LPAREN2,
  [62607] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3949), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [62615] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3662), 1,
      anon_sym_LBRACE,
    STATE(1890), 1,
      sym_enum_declaration_list,
  [62625] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4700), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [62633] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(1891), 1,
      sym_declaration_list,
  [62643] = 3,
    ACTIONS(368), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1670), 1,
      sym_compound_statement,
  [62653] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4628), 1,
      anon_sym_LPAREN2,
    STATE(75), 1,
      sym_parenthesized_expression,
  [62663] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2604), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62671] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4628), 1,
      anon_sym_LPAREN2,
    STATE(76), 1,
      sym_parenthesized_expression,
  [62681] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1760), 1,
      sym_formal_parameters,
  [62691] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(2072), 1,
      sym_formal_parameters,
  [62701] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(753), 1,
      sym_declaration_list,
  [62711] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4426), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [62719] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3347), 1,
      anon_sym_LBRACE,
    STATE(915), 1,
      sym_declaration_list,
  [62729] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(538), 1,
      sym_declaration_list,
  [62739] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4703), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62747] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4628), 1,
      anon_sym_LPAREN2,
    STATE(62), 1,
      sym_parenthesized_expression,
  [62757] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2134), 1,
      anon_sym_LPAREN2,
    STATE(2253), 1,
      sym_parenthesized_expression,
  [62767] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3475), 1,
      anon_sym_LBRACE,
    STATE(546), 1,
      sym_declaration_list,
  [62777] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4705), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [62785] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4707), 1,
      sym__new_line,
    ACTIONS(4709), 1,
      sym_heredoc_end,
  [62795] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4628), 1,
      anon_sym_LPAREN2,
    STATE(106), 1,
      sym_parenthesized_expression,
  [62805] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4711), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62813] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4713), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62821] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4715), 2,
      anon_sym_string,
      anon_sym_int,
  [62829] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3636), 1,
      anon_sym_LBRACE,
    STATE(576), 1,
      sym_enum_declaration_list,
  [62839] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4717), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [62847] = 3,
    ACTIONS(368), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(1896), 1,
      sym_compound_statement,
  [62857] = 3,
    ACTIONS(368), 1,
      anon_sym_LBRACE,
    ACTIONS(1602), 1,
      sym_comment,
    STATE(980), 1,
      sym_compound_statement,
  [62867] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4719), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62875] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2134), 1,
      anon_sym_LPAREN2,
    STATE(1888), 1,
      sym_parenthesized_expression,
  [62885] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4628), 1,
      anon_sym_LPAREN2,
    STATE(98), 1,
      sym_parenthesized_expression,
  [62895] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4721), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62903] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4211), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [62911] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2092), 1,
      anon_sym_LPAREN2,
    STATE(864), 1,
      sym_arguments,
  [62921] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4723), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62929] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3070), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62937] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4725), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [62945] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4727), 2,
      anon_sym_LBRACE,
      anon_sym_COLON,
  [62953] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1620), 1,
      anon_sym_LPAREN2,
    STATE(601), 1,
      sym_arguments,
  [62963] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4729), 1,
      anon_sym_BSLASH,
    STATE(2358), 1,
      aux_sym_namespace_name_repeat1,
  [62973] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3234), 1,
      anon_sym_LPAREN2,
    STATE(1965), 1,
      sym_formal_parameters,
  [62983] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(2092), 1,
      anon_sym_LPAREN2,
    STATE(847), 1,
      sym_arguments,
  [62993] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4732), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [63001] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1915), 1,
      anon_sym_DOLLAR,
    STATE(2239), 1,
      sym_variable_name,
  [63011] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4734), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN2,
  [63019] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3307), 1,
      anon_sym_LPAREN2,
    STATE(1705), 1,
      sym_formal_parameters,
  [63029] = 3,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3355), 1,
      anon_sym_LBRACE,
    STATE(2078), 1,
      sym_declaration_list,
  [63039] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4736), 1,
      anon_sym_COLON_COLON,
  [63046] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4738), 1,
      anon_sym_SEMI,
  [63053] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4740), 1,
      aux_sym_if_statement_token2,
  [63060] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4742), 1,
      anon_sym_EQ_GT,
  [63067] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4562), 1,
      anon_sym_RBRACK,
  [63074] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4744), 1,
      anon_sym_LBRACE,
  [63081] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4746), 1,
      anon_sym_RBRACK,
  [63088] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4748), 1,
      anon_sym_LBRACE,
  [63095] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4750), 1,
      sym_heredoc_end,
  [63102] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4752), 1,
      sym_heredoc_end,
  [63109] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4754), 1,
      ts_builtin_sym_end,
  [63116] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4756), 1,
      sym_heredoc_end,
  [63123] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4758), 1,
      sym_name,
  [63130] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4760), 1,
      sym_name,
  [63137] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4762), 1,
      anon_sym_EQ_GT,
  [63144] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4764), 1,
      sym_heredoc_end,
  [63151] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4766), 1,
      sym_name,
  [63158] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4768), 1,
      sym_name,
  [63165] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4770), 1,
      anon_sym_LBRACE,
  [63172] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4772), 1,
      anon_sym_LBRACE,
  [63179] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4774), 1,
      anon_sym_LPAREN2,
  [63186] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4776), 1,
      anon_sym_LPAREN2,
  [63193] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4778), 1,
      aux_sym_if_statement_token2,
  [63200] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4780), 1,
      sym_name,
  [63207] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4782), 1,
      anon_sym_LBRACE,
  [63214] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4784), 1,
      anon_sym_LBRACE,
  [63221] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4786), 1,
      anon_sym_EQ_GT,
  [63228] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4788), 1,
      aux_sym_while_statement_token1,
  [63235] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4233), 1,
      anon_sym_RPAREN2,
  [63242] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4790), 1,
      anon_sym_LPAREN2,
  [63249] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4792), 1,
      anon_sym_EQ_GT,
  [63256] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4794), 1,
      anon_sym_COLON_COLON,
  [63263] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4796), 1,
      sym_name,
  [63270] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4798), 1,
      sym_heredoc_start,
  [63277] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4800), 1,
      anon_sym_COLON_COLON,
  [63284] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4802), 1,
      sym_name,
  [63291] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4804), 1,
      sym_name,
  [63298] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4806), 1,
      anon_sym_SQUOTE2,
  [63305] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4808), 1,
      sym_name,
  [63312] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4810), 1,
      sym_name,
  [63319] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4812), 1,
      sym_name,
  [63326] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4814), 1,
      anon_sym_EQ_GT,
  [63333] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4816), 1,
      anon_sym_EQ,
  [63340] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4818), 1,
      anon_sym_RPAREN2,
  [63347] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4820), 1,
      sym_name,
  [63354] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4822), 1,
      sym_name,
  [63361] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4824), 1,
      anon_sym_RPAREN2,
  [63368] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4826), 1,
      anon_sym_LBRACE,
  [63375] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4828), 1,
      aux_sym_while_statement_token2,
  [63382] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4830), 1,
      anon_sym_RPAREN2,
  [63389] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4832), 1,
      sym_name,
  [63396] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4834), 1,
      anon_sym_RPAREN2,
  [63403] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4836), 1,
      anon_sym_RPAREN2,
  [63410] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4838), 1,
      anon_sym_RPAREN2,
  [63417] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4840), 1,
      anon_sym_RPAREN2,
  [63424] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4842), 1,
      sym_name,
  [63431] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4844), 1,
      anon_sym_RPAREN2,
  [63438] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4846), 1,
      anon_sym_EQ_GT,
  [63445] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4848), 1,
      sym_name,
  [63452] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4850), 1,
      anon_sym_BSLASH,
  [63459] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4095), 1,
      anon_sym_RPAREN2,
  [63466] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4852), 1,
      anon_sym_EQ,
  [63473] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4854), 1,
      anon_sym_LBRACE,
  [63480] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4856), 1,
      anon_sym_RPAREN2,
  [63487] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4858), 1,
      anon_sym_LBRACE,
  [63494] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4860), 1,
      anon_sym_LPAREN2,
  [63501] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4862), 1,
      anon_sym_RPAREN2,
  [63508] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4864), 1,
      anon_sym_RPAREN2,
  [63515] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4866), 1,
      aux_sym_if_statement_token2,
  [63522] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4868), 1,
      anon_sym_RPAREN2,
  [63529] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4870), 1,
      anon_sym_SEMI,
  [63536] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4298), 1,
      anon_sym_RPAREN2,
  [63543] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4872), 1,
      anon_sym_RPAREN2,
  [63550] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4874), 1,
      anon_sym_SEMI,
  [63557] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4876), 1,
      sym_name,
  [63564] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4878), 1,
      anon_sym_COLON_COLON,
  [63571] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4880), 1,
      anon_sym_SQUOTE2,
  [63578] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3810), 1,
      anon_sym_BSLASH,
  [63585] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4882), 1,
      aux_sym_if_statement_token2,
  [63592] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4884), 1,
      anon_sym_LPAREN2,
  [63599] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4886), 1,
      sym_name,
  [63606] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4888), 1,
      anon_sym_LBRACE,
  [63613] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4359), 1,
      anon_sym_RBRACE,
  [63620] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4128), 1,
      anon_sym_RPAREN2,
  [63627] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4434), 1,
      anon_sym_RPAREN2,
  [63634] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4890), 1,
      sym_name,
  [63641] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4892), 1,
      sym_name,
  [63648] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4894), 1,
      sym_heredoc_end,
  [63655] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4109), 1,
      anon_sym_LBRACE,
  [63662] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4896), 1,
      sym_heredoc_end,
  [63669] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4898), 1,
      anon_sym_RPAREN2,
  [63676] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4900), 1,
      anon_sym_SEMI,
  [63683] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4902), 1,
      anon_sym_EQ_GT,
  [63690] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4904), 1,
      anon_sym_SEMI,
  [63697] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4906), 1,
      anon_sym_EQ_GT,
  [63704] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4908), 1,
      sym_name,
  [63711] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4910), 1,
      sym_name,
  [63718] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4912), 1,
      sym_name,
  [63725] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4914), 1,
      anon_sym_BSLASH,
  [63732] = 2,
    ACTIONS(983), 1,
      anon_sym_COLON,
    ACTIONS(1602), 1,
      sym_comment,
  [63739] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4916), 1,
      anon_sym_BSLASH,
  [63746] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4918), 1,
      anon_sym_RPAREN2,
  [63753] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4920), 1,
      anon_sym_RPAREN2,
  [63760] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4922), 1,
      anon_sym_RPAREN2,
  [63767] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4924), 1,
      anon_sym_EQ_GT,
  [63774] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4926), 1,
      anon_sym_LBRACE,
  [63781] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4292), 1,
      sym_name,
  [63788] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4928), 1,
      anon_sym_EQ,
  [63795] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4930), 1,
      sym_name,
  [63802] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4932), 1,
      anon_sym_SEMI,
  [63809] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4934), 1,
      anon_sym_EQ_GT,
  [63816] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4936), 1,
      sym_heredoc_end,
  [63823] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4938), 1,
      sym_heredoc_end,
  [63830] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4940), 1,
      anon_sym_RPAREN2,
  [63837] = 2,
    ACTIONS(517), 1,
      ts_builtin_sym_end,
    ACTIONS(1602), 1,
      sym_comment,
  [63844] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4942), 1,
      anon_sym_RPAREN,
  [63851] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3505), 1,
      sym_name,
  [63858] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4944), 1,
      anon_sym_EQ_GT,
  [63865] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4101), 1,
      anon_sym_RPAREN2,
  [63872] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4504), 1,
      anon_sym_RPAREN2,
  [63879] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4946), 1,
      anon_sym_EQ,
  [63886] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4948), 1,
      sym_name,
  [63893] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4950), 1,
      sym_name,
  [63900] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4952), 1,
      anon_sym_RPAREN2,
  [63907] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4954), 1,
      anon_sym_LBRACE,
  [63914] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4956), 1,
      aux_sym_foreach_statement_token2,
  [63921] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4958), 1,
      anon_sym_RPAREN2,
  [63928] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4960), 1,
      anon_sym_RPAREN2,
  [63935] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4962), 1,
      sym_name,
  [63942] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4964), 1,
      anon_sym_RPAREN2,
  [63949] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4966), 1,
      anon_sym_RPAREN2,
  [63956] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4968), 1,
      aux_sym_foreach_statement_token2,
  [63963] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4970), 1,
      anon_sym_SEMI,
  [63970] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4972), 1,
      anon_sym_EQ_GT,
  [63977] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4974), 1,
      anon_sym_EQ,
  [63984] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4976), 1,
      anon_sym_RPAREN2,
  [63991] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4268), 1,
      anon_sym_RPAREN2,
  [63998] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4978), 1,
      sym_name,
  [64005] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4980), 1,
      anon_sym_RPAREN2,
  [64012] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4982), 1,
      anon_sym_BSLASH,
  [64019] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4984), 1,
      anon_sym_LPAREN2,
  [64026] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4986), 1,
      anon_sym_RPAREN,
  [64033] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4640), 1,
      anon_sym_RPAREN2,
  [64040] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4988), 1,
      anon_sym_RPAREN2,
  [64047] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4990), 1,
      sym_name,
  [64054] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4992), 1,
      anon_sym_RPAREN2,
  [64061] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4994), 1,
      anon_sym_BSLASH,
  [64068] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4996), 1,
      aux_sym_while_statement_token1,
  [64075] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4998), 1,
      anon_sym_RPAREN2,
  [64082] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5000), 1,
      anon_sym_RPAREN2,
  [64089] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5002), 1,
      anon_sym_COLON_COLON,
  [64096] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5004), 1,
      anon_sym_RPAREN2,
  [64103] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5006), 1,
      anon_sym_LBRACE,
  [64110] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5008), 1,
      anon_sym_LBRACE,
  [64117] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5010), 1,
      anon_sym_RPAREN2,
  [64124] = 2,
    ACTIONS(485), 1,
      ts_builtin_sym_end,
    ACTIONS(1602), 1,
      sym_comment,
  [64131] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5012), 1,
      anon_sym_LPAREN2,
  [64138] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5014), 1,
      anon_sym_LPAREN2,
  [64145] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5016), 1,
      sym_heredoc_end,
  [64152] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5018), 1,
      anon_sym_RPAREN2,
  [64159] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5020), 1,
      anon_sym_RPAREN2,
  [64166] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5022), 1,
      sym_heredoc_start,
  [64173] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5024), 1,
      sym_name,
  [64180] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5026), 1,
      anon_sym_RBRACK,
  [64187] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5028), 1,
      sym_name,
  [64194] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5030), 1,
      anon_sym_RPAREN2,
  [64201] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5032), 1,
      anon_sym_SEMI,
  [64208] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5034), 1,
      sym_heredoc_end,
  [64215] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5036), 1,
      sym_name,
  [64222] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5038), 1,
      anon_sym_LBRACE,
  [64229] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5040), 1,
      sym_name,
  [64236] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5042), 1,
      sym_name,
  [64243] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5044), 1,
      aux_sym_if_statement_token2,
  [64250] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5046), 1,
      sym_name,
  [64257] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5048), 1,
      anon_sym_RPAREN2,
  [64264] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5050), 1,
      sym_name,
  [64271] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5052), 1,
      anon_sym_EQ_GT,
  [64278] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5054), 1,
      anon_sym_RPAREN2,
  [64285] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5056), 1,
      anon_sym_RPAREN2,
  [64292] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5058), 1,
      anon_sym_RPAREN2,
  [64299] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4146), 1,
      anon_sym_RBRACE,
  [64306] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5060), 1,
      anon_sym_RPAREN2,
  [64313] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5062), 1,
      aux_sym_while_statement_token2,
  [64320] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5064), 1,
      anon_sym_COLON_COLON,
  [64327] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4624), 1,
      anon_sym_EQ,
  [64334] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5066), 1,
      anon_sym_RPAREN2,
  [64341] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5068), 1,
      anon_sym_RPAREN2,
  [64348] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5070), 1,
      sym_name,
  [64355] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5072), 1,
      sym_name,
  [64362] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4638), 1,
      anon_sym_LPAREN2,
  [64369] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5074), 1,
      anon_sym_RPAREN2,
  [64376] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5076), 1,
      sym_heredoc_start,
  [64383] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5078), 1,
      sym_integer,
  [64390] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5080), 1,
      sym_heredoc_start,
  [64397] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5082), 1,
      anon_sym_LPAREN2,
  [64404] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5084), 1,
      anon_sym_LBRACE,
  [64411] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5086), 1,
      anon_sym_SEMI,
  [64418] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5088), 1,
      sym_name,
  [64425] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5090), 1,
      sym_name,
  [64432] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5092), 1,
      sym_name,
  [64439] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4452), 1,
      anon_sym_RPAREN2,
  [64446] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5094), 1,
      sym_name,
  [64453] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5096), 1,
      sym_name,
  [64460] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5098), 1,
      sym_name,
  [64467] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5100), 1,
      sym_name,
  [64474] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(3501), 1,
      sym_name,
  [64481] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5102), 1,
      sym_name,
  [64488] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5104), 1,
      anon_sym_RBRACK,
  [64495] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4206), 1,
      anon_sym_RPAREN2,
  [64502] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5106), 1,
      anon_sym_BSLASH,
  [64509] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5108), 1,
      sym_name,
  [64516] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5110), 1,
      ts_builtin_sym_end,
  [64523] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5112), 1,
      anon_sym_LPAREN2,
  [64530] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5114), 1,
      sym_name,
  [64537] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5116), 1,
      sym_name,
  [64544] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5118), 1,
      anon_sym_RPAREN2,
  [64551] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5120), 1,
      anon_sym_LPAREN2,
  [64558] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5122), 1,
      sym_name,
  [64565] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5124), 1,
      anon_sym_SEMI,
  [64572] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5126), 1,
      anon_sym_RBRACK,
  [64579] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5128), 1,
      anon_sym_LPAREN2,
  [64586] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5130), 1,
      anon_sym_SEMI,
  [64593] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5132), 1,
      anon_sym_LPAREN2,
  [64600] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4537), 1,
      anon_sym_LBRACE,
  [64607] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5134), 1,
      anon_sym_LPAREN2,
  [64614] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5136), 1,
      anon_sym_LPAREN2,
  [64621] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5138), 1,
      anon_sym_LPAREN2,
  [64628] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5140), 1,
      anon_sym_BSLASH,
  [64635] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5142), 1,
      anon_sym_SEMI,
  [64642] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5144), 1,
      anon_sym_COLON_COLON,
  [64649] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5146), 1,
      anon_sym_LPAREN2,
  [64656] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5148), 1,
      anon_sym_LPAREN2,
  [64663] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(4500), 1,
      anon_sym_RBRACK,
  [64670] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5150), 1,
      anon_sym_LPAREN2,
  [64677] = 2,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(5152), 1,
      anon_sym_EQ_GT,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(584)] = 0,
  [SMALL_STATE(585)] = 70,
  [SMALL_STATE(586)] = 140,
  [SMALL_STATE(587)] = 210,
  [SMALL_STATE(588)] = 280,
  [SMALL_STATE(589)] = 353,
  [SMALL_STATE(590)] = 426,
  [SMALL_STATE(591)] = 494,
  [SMALL_STATE(592)] = 562,
  [SMALL_STATE(593)] = 630,
  [SMALL_STATE(594)] = 698,
  [SMALL_STATE(595)] = 768,
  [SMALL_STATE(596)] = 836,
  [SMALL_STATE(597)] = 906,
  [SMALL_STATE(598)] = 974,
  [SMALL_STATE(599)] = 1042,
  [SMALL_STATE(600)] = 1110,
  [SMALL_STATE(601)] = 1178,
  [SMALL_STATE(602)] = 1246,
  [SMALL_STATE(603)] = 1314,
  [SMALL_STATE(604)] = 1392,
  [SMALL_STATE(605)] = 1460,
  [SMALL_STATE(606)] = 1528,
  [SMALL_STATE(607)] = 1596,
  [SMALL_STATE(608)] = 1664,
  [SMALL_STATE(609)] = 1732,
  [SMALL_STATE(610)] = 1800,
  [SMALL_STATE(611)] = 1868,
  [SMALL_STATE(612)] = 1936,
  [SMALL_STATE(613)] = 2004,
  [SMALL_STATE(614)] = 2079,
  [SMALL_STATE(615)] = 2154,
  [SMALL_STATE(616)] = 2223,
  [SMALL_STATE(617)] = 2298,
  [SMALL_STATE(618)] = 2367,
  [SMALL_STATE(619)] = 2431,
  [SMALL_STATE(620)] = 2495,
  [SMALL_STATE(621)] = 2559,
  [SMALL_STATE(622)] = 2631,
  [SMALL_STATE(623)] = 2695,
  [SMALL_STATE(624)] = 2759,
  [SMALL_STATE(625)] = 2823,
  [SMALL_STATE(626)] = 2935,
  [SMALL_STATE(627)] = 3009,
  [SMALL_STATE(628)] = 3127,
  [SMALL_STATE(629)] = 3191,
  [SMALL_STATE(630)] = 3255,
  [SMALL_STATE(631)] = 3367,
  [SMALL_STATE(632)] = 3431,
  [SMALL_STATE(633)] = 3495,
  [SMALL_STATE(634)] = 3559,
  [SMALL_STATE(635)] = 3623,
  [SMALL_STATE(636)] = 3687,
  [SMALL_STATE(637)] = 3751,
  [SMALL_STATE(638)] = 3817,
  [SMALL_STATE(639)] = 3891,
  [SMALL_STATE(640)] = 4009,
  [SMALL_STATE(641)] = 4075,
  [SMALL_STATE(642)] = 4139,
  [SMALL_STATE(643)] = 4251,
  [SMALL_STATE(644)] = 4323,
  [SMALL_STATE(645)] = 4397,
  [SMALL_STATE(646)] = 4461,
  [SMALL_STATE(647)] = 4532,
  [SMALL_STATE(648)] = 4611,
  [SMALL_STATE(649)] = 4674,
  [SMALL_STATE(650)] = 4737,
  [SMALL_STATE(651)] = 4800,
  [SMALL_STATE(652)] = 4863,
  [SMALL_STATE(653)] = 4926,
  [SMALL_STATE(654)] = 4989,
  [SMALL_STATE(655)] = 5060,
  [SMALL_STATE(656)] = 5131,
  [SMALL_STATE(657)] = 5246,
  [SMALL_STATE(658)] = 5309,
  [SMALL_STATE(659)] = 5382,
  [SMALL_STATE(660)] = 5453,
  [SMALL_STATE(661)] = 5526,
  [SMALL_STATE(662)] = 5589,
  [SMALL_STATE(663)] = 5660,
  [SMALL_STATE(664)] = 5731,
  [SMALL_STATE(665)] = 5807,
  [SMALL_STATE(666)] = 5881,
  [SMALL_STATE(667)] = 5951,
  [SMALL_STATE(668)] = 6021,
  [SMALL_STATE(669)] = 6095,
  [SMALL_STATE(670)] = 6165,
  [SMALL_STATE(671)] = 6239,
  [SMALL_STATE(672)] = 6315,
  [SMALL_STATE(673)] = 6385,
  [SMALL_STATE(674)] = 6456,
  [SMALL_STATE(675)] = 6565,
  [SMALL_STATE(676)] = 6674,
  [SMALL_STATE(677)] = 6745,
  [SMALL_STATE(678)] = 6816,
  [SMALL_STATE(679)] = 6887,
  [SMALL_STATE(680)] = 6958,
  [SMALL_STATE(681)] = 7031,
  [SMALL_STATE(682)] = 7140,
  [SMALL_STATE(683)] = 7211,
  [SMALL_STATE(684)] = 7320,
  [SMALL_STATE(685)] = 7384,
  [SMALL_STATE(686)] = 7490,
  [SMALL_STATE(687)] = 7596,
  [SMALL_STATE(688)] = 7702,
  [SMALL_STATE(689)] = 7808,
  [SMALL_STATE(690)] = 7914,
  [SMALL_STATE(691)] = 8020,
  [SMALL_STATE(692)] = 8126,
  [SMALL_STATE(693)] = 8232,
  [SMALL_STATE(694)] = 8302,
  [SMALL_STATE(695)] = 8408,
  [SMALL_STATE(696)] = 8478,
  [SMALL_STATE(697)] = 8584,
  [SMALL_STATE(698)] = 8690,
  [SMALL_STATE(699)] = 8796,
  [SMALL_STATE(700)] = 8902,
  [SMALL_STATE(701)] = 8960,
  [SMALL_STATE(702)] = 9020,
  [SMALL_STATE(703)] = 9117,
  [SMALL_STATE(704)] = 9176,
  [SMALL_STATE(705)] = 9273,
  [SMALL_STATE(706)] = 9328,
  [SMALL_STATE(707)] = 9387,
  [SMALL_STATE(708)] = 9446,
  [SMALL_STATE(709)] = 9501,
  [SMALL_STATE(710)] = 9555,
  [SMALL_STATE(711)] = 9609,
  [SMALL_STATE(712)] = 9663,
  [SMALL_STATE(713)] = 9723,
  [SMALL_STATE(714)] = 9779,
  [SMALL_STATE(715)] = 9835,
  [SMALL_STATE(716)] = 9889,
  [SMALL_STATE(717)] = 9945,
  [SMALL_STATE(718)] = 9999,
  [SMALL_STATE(719)] = 10053,
  [SMALL_STATE(720)] = 10107,
  [SMALL_STATE(721)] = 10161,
  [SMALL_STATE(722)] = 10215,
  [SMALL_STATE(723)] = 10269,
  [SMALL_STATE(724)] = 10323,
  [SMALL_STATE(725)] = 10377,
  [SMALL_STATE(726)] = 10433,
  [SMALL_STATE(727)] = 10487,
  [SMALL_STATE(728)] = 10541,
  [SMALL_STATE(729)] = 10597,
  [SMALL_STATE(730)] = 10651,
  [SMALL_STATE(731)] = 10711,
  [SMALL_STATE(732)] = 10765,
  [SMALL_STATE(733)] = 10819,
  [SMALL_STATE(734)] = 10876,
  [SMALL_STATE(735)] = 10929,
  [SMALL_STATE(736)] = 10982,
  [SMALL_STATE(737)] = 11039,
  [SMALL_STATE(738)] = 11094,
  [SMALL_STATE(739)] = 11147,
  [SMALL_STATE(740)] = 11200,
  [SMALL_STATE(741)] = 11257,
  [SMALL_STATE(742)] = 11310,
  [SMALL_STATE(743)] = 11363,
  [SMALL_STATE(744)] = 11418,
  [SMALL_STATE(745)] = 11471,
  [SMALL_STATE(746)] = 11524,
  [SMALL_STATE(747)] = 11581,
  [SMALL_STATE(748)] = 11638,
  [SMALL_STATE(749)] = 11690,
  [SMALL_STATE(750)] = 11742,
  [SMALL_STATE(751)] = 11794,
  [SMALL_STATE(752)] = 11846,
  [SMALL_STATE(753)] = 11898,
  [SMALL_STATE(754)] = 11950,
  [SMALL_STATE(755)] = 12006,
  [SMALL_STATE(756)] = 12060,
  [SMALL_STATE(757)] = 12114,
  [SMALL_STATE(758)] = 12166,
  [SMALL_STATE(759)] = 12218,
  [SMALL_STATE(760)] = 12270,
  [SMALL_STATE(761)] = 12324,
  [SMALL_STATE(762)] = 12376,
  [SMALL_STATE(763)] = 12428,
  [SMALL_STATE(764)] = 12480,
  [SMALL_STATE(765)] = 12534,
  [SMALL_STATE(766)] = 12586,
  [SMALL_STATE(767)] = 12638,
  [SMALL_STATE(768)] = 12690,
  [SMALL_STATE(769)] = 12742,
  [SMALL_STATE(770)] = 12794,
  [SMALL_STATE(771)] = 12846,
  [SMALL_STATE(772)] = 12898,
  [SMALL_STATE(773)] = 12950,
  [SMALL_STATE(774)] = 13002,
  [SMALL_STATE(775)] = 13054,
  [SMALL_STATE(776)] = 13106,
  [SMALL_STATE(777)] = 13158,
  [SMALL_STATE(778)] = 13210,
  [SMALL_STATE(779)] = 13262,
  [SMALL_STATE(780)] = 13314,
  [SMALL_STATE(781)] = 13366,
  [SMALL_STATE(782)] = 13418,
  [SMALL_STATE(783)] = 13512,
  [SMALL_STATE(784)] = 13566,
  [SMALL_STATE(785)] = 13622,
  [SMALL_STATE(786)] = 13678,
  [SMALL_STATE(787)] = 13732,
  [SMALL_STATE(788)] = 13786,
  [SMALL_STATE(789)] = 13838,
  [SMALL_STATE(790)] = 13890,
  [SMALL_STATE(791)] = 13950,
  [SMALL_STATE(792)] = 14002,
  [SMALL_STATE(793)] = 14058,
  [SMALL_STATE(794)] = 14114,
  [SMALL_STATE(795)] = 14166,
  [SMALL_STATE(796)] = 14218,
  [SMALL_STATE(797)] = 14312,
  [SMALL_STATE(798)] = 14364,
  [SMALL_STATE(799)] = 14416,
  [SMALL_STATE(800)] = 14468,
  [SMALL_STATE(801)] = 14520,
  [SMALL_STATE(802)] = 14572,
  [SMALL_STATE(803)] = 14624,
  [SMALL_STATE(804)] = 14676,
  [SMALL_STATE(805)] = 14728,
  [SMALL_STATE(806)] = 14780,
  [SMALL_STATE(807)] = 14832,
  [SMALL_STATE(808)] = 14884,
  [SMALL_STATE(809)] = 14936,
  [SMALL_STATE(810)] = 14988,
  [SMALL_STATE(811)] = 15040,
  [SMALL_STATE(812)] = 15092,
  [SMALL_STATE(813)] = 15144,
  [SMALL_STATE(814)] = 15196,
  [SMALL_STATE(815)] = 15248,
  [SMALL_STATE(816)] = 15300,
  [SMALL_STATE(817)] = 15352,
  [SMALL_STATE(818)] = 15404,
  [SMALL_STATE(819)] = 15456,
  [SMALL_STATE(820)] = 15508,
  [SMALL_STATE(821)] = 15560,
  [SMALL_STATE(822)] = 15612,
  [SMALL_STATE(823)] = 15664,
  [SMALL_STATE(824)] = 15716,
  [SMALL_STATE(825)] = 15768,
  [SMALL_STATE(826)] = 15820,
  [SMALL_STATE(827)] = 15872,
  [SMALL_STATE(828)] = 15924,
  [SMALL_STATE(829)] = 15977,
  [SMALL_STATE(830)] = 16030,
  [SMALL_STATE(831)] = 16085,
  [SMALL_STATE(832)] = 16140,
  [SMALL_STATE(833)] = 16195,
  [SMALL_STATE(834)] = 16286,
  [SMALL_STATE(835)] = 16377,
  [SMALL_STATE(836)] = 16468,
  [SMALL_STATE(837)] = 16521,
  [SMALL_STATE(838)] = 16572,
  [SMALL_STATE(839)] = 16625,
  [SMALL_STATE(840)] = 16716,
  [SMALL_STATE(841)] = 16771,
  [SMALL_STATE(842)] = 16824,
  [SMALL_STATE(843)] = 16879,
  [SMALL_STATE(844)] = 16929,
  [SMALL_STATE(845)] = 16979,
  [SMALL_STATE(846)] = 17029,
  [SMALL_STATE(847)] = 17079,
  [SMALL_STATE(848)] = 17129,
  [SMALL_STATE(849)] = 17179,
  [SMALL_STATE(850)] = 17229,
  [SMALL_STATE(851)] = 17279,
  [SMALL_STATE(852)] = 17373,
  [SMALL_STATE(853)] = 17467,
  [SMALL_STATE(854)] = 17519,
  [SMALL_STATE(855)] = 17569,
  [SMALL_STATE(856)] = 17623,
  [SMALL_STATE(857)] = 17673,
  [SMALL_STATE(858)] = 17723,
  [SMALL_STATE(859)] = 17777,
  [SMALL_STATE(860)] = 17827,
  [SMALL_STATE(861)] = 17877,
  [SMALL_STATE(862)] = 17927,
  [SMALL_STATE(863)] = 17983,
  [SMALL_STATE(864)] = 18039,
  [SMALL_STATE(865)] = 18089,
  [SMALL_STATE(866)] = 18183,
  [SMALL_STATE(867)] = 18271,
  [SMALL_STATE(868)] = 18321,
  [SMALL_STATE(869)] = 18373,
  [SMALL_STATE(870)] = 18423,
  [SMALL_STATE(871)] = 18476,
  [SMALL_STATE(872)] = 18525,
  [SMALL_STATE(873)] = 18574,
  [SMALL_STATE(874)] = 18623,
  [SMALL_STATE(875)] = 18672,
  [SMALL_STATE(876)] = 18725,
  [SMALL_STATE(877)] = 18776,
  [SMALL_STATE(878)] = 18829,
  [SMALL_STATE(879)] = 18878,
  [SMALL_STATE(880)] = 18927,
  [SMALL_STATE(881)] = 18978,
  [SMALL_STATE(882)] = 19029,
  [SMALL_STATE(883)] = 19078,
  [SMALL_STATE(884)] = 19131,
  [SMALL_STATE(885)] = 19184,
  [SMALL_STATE(886)] = 19235,
  [SMALL_STATE(887)] = 19284,
  [SMALL_STATE(888)] = 19333,
  [SMALL_STATE(889)] = 19382,
  [SMALL_STATE(890)] = 19431,
  [SMALL_STATE(891)] = 19479,
  [SMALL_STATE(892)] = 19527,
  [SMALL_STATE(893)] = 19577,
  [SMALL_STATE(894)] = 19625,
  [SMALL_STATE(895)] = 19675,
  [SMALL_STATE(896)] = 19723,
  [SMALL_STATE(897)] = 19771,
  [SMALL_STATE(898)] = 19821,
  [SMALL_STATE(899)] = 19869,
  [SMALL_STATE(900)] = 19917,
  [SMALL_STATE(901)] = 19965,
  [SMALL_STATE(902)] = 20013,
  [SMALL_STATE(903)] = 20061,
  [SMALL_STATE(904)] = 20109,
  [SMALL_STATE(905)] = 20165,
  [SMALL_STATE(906)] = 20213,
  [SMALL_STATE(907)] = 20261,
  [SMALL_STATE(908)] = 20309,
  [SMALL_STATE(909)] = 20357,
  [SMALL_STATE(910)] = 20405,
  [SMALL_STATE(911)] = 20453,
  [SMALL_STATE(912)] = 20501,
  [SMALL_STATE(913)] = 20549,
  [SMALL_STATE(914)] = 20597,
  [SMALL_STATE(915)] = 20647,
  [SMALL_STATE(916)] = 20695,
  [SMALL_STATE(917)] = 20743,
  [SMALL_STATE(918)] = 20793,
  [SMALL_STATE(919)] = 20843,
  [SMALL_STATE(920)] = 20893,
  [SMALL_STATE(921)] = 20941,
  [SMALL_STATE(922)] = 20989,
  [SMALL_STATE(923)] = 21037,
  [SMALL_STATE(924)] = 21093,
  [SMALL_STATE(925)] = 21149,
  [SMALL_STATE(926)] = 21197,
  [SMALL_STATE(927)] = 21249,
  [SMALL_STATE(928)] = 21297,
  [SMALL_STATE(929)] = 21349,
  [SMALL_STATE(930)] = 21397,
  [SMALL_STATE(931)] = 21445,
  [SMALL_STATE(932)] = 21493,
  [SMALL_STATE(933)] = 21541,
  [SMALL_STATE(934)] = 21589,
  [SMALL_STATE(935)] = 21637,
  [SMALL_STATE(936)] = 21687,
  [SMALL_STATE(937)] = 21735,
  [SMALL_STATE(938)] = 21787,
  [SMALL_STATE(939)] = 21835,
  [SMALL_STATE(940)] = 21883,
  [SMALL_STATE(941)] = 21931,
  [SMALL_STATE(942)] = 21983,
  [SMALL_STATE(943)] = 22031,
  [SMALL_STATE(944)] = 22079,
  [SMALL_STATE(945)] = 22127,
  [SMALL_STATE(946)] = 22175,
  [SMALL_STATE(947)] = 22223,
  [SMALL_STATE(948)] = 22271,
  [SMALL_STATE(949)] = 22319,
  [SMALL_STATE(950)] = 22367,
  [SMALL_STATE(951)] = 22415,
  [SMALL_STATE(952)] = 22463,
  [SMALL_STATE(953)] = 22511,
  [SMALL_STATE(954)] = 22559,
  [SMALL_STATE(955)] = 22607,
  [SMALL_STATE(956)] = 22655,
  [SMALL_STATE(957)] = 22703,
  [SMALL_STATE(958)] = 22751,
  [SMALL_STATE(959)] = 22799,
  [SMALL_STATE(960)] = 22847,
  [SMALL_STATE(961)] = 22895,
  [SMALL_STATE(962)] = 22943,
  [SMALL_STATE(963)] = 22991,
  [SMALL_STATE(964)] = 23039,
  [SMALL_STATE(965)] = 23091,
  [SMALL_STATE(966)] = 23139,
  [SMALL_STATE(967)] = 23187,
  [SMALL_STATE(968)] = 23235,
  [SMALL_STATE(969)] = 23283,
  [SMALL_STATE(970)] = 23331,
  [SMALL_STATE(971)] = 23379,
  [SMALL_STATE(972)] = 23431,
  [SMALL_STATE(973)] = 23487,
  [SMALL_STATE(974)] = 23535,
  [SMALL_STATE(975)] = 23583,
  [SMALL_STATE(976)] = 23632,
  [SMALL_STATE(977)] = 23679,
  [SMALL_STATE(978)] = 23726,
  [SMALL_STATE(979)] = 23773,
  [SMALL_STATE(980)] = 23822,
  [SMALL_STATE(981)] = 23869,
  [SMALL_STATE(982)] = 23916,
  [SMALL_STATE(983)] = 23963,
  [SMALL_STATE(984)] = 24010,
  [SMALL_STATE(985)] = 24059,
  [SMALL_STATE(986)] = 24106,
  [SMALL_STATE(987)] = 24157,
  [SMALL_STATE(988)] = 24206,
  [SMALL_STATE(989)] = 24255,
  [SMALL_STATE(990)] = 24302,
  [SMALL_STATE(991)] = 24353,
  [SMALL_STATE(992)] = 24400,
  [SMALL_STATE(993)] = 24447,
  [SMALL_STATE(994)] = 24494,
  [SMALL_STATE(995)] = 24541,
  [SMALL_STATE(996)] = 24590,
  [SMALL_STATE(997)] = 24637,
  [SMALL_STATE(998)] = 24686,
  [SMALL_STATE(999)] = 24735,
  [SMALL_STATE(1000)] = 24782,
  [SMALL_STATE(1001)] = 24829,
  [SMALL_STATE(1002)] = 24876,
  [SMALL_STATE(1003)] = 24923,
  [SMALL_STATE(1004)] = 24970,
  [SMALL_STATE(1005)] = 25017,
  [SMALL_STATE(1006)] = 25064,
  [SMALL_STATE(1007)] = 25111,
  [SMALL_STATE(1008)] = 25158,
  [SMALL_STATE(1009)] = 25204,
  [SMALL_STATE(1010)] = 25251,
  [SMALL_STATE(1011)] = 25296,
  [SMALL_STATE(1012)] = 25345,
  [SMALL_STATE(1013)] = 25427,
  [SMALL_STATE(1014)] = 25477,
  [SMALL_STATE(1015)] = 25553,
  [SMALL_STATE(1016)] = 25633,
  [SMALL_STATE(1017)] = 25677,
  [SMALL_STATE(1018)] = 25755,
  [SMALL_STATE(1019)] = 25827,
  [SMALL_STATE(1020)] = 25897,
  [SMALL_STATE(1021)] = 25981,
  [SMALL_STATE(1022)] = 26047,
  [SMALL_STATE(1023)] = 26107,
  [SMALL_STATE(1024)] = 26163,
  [SMALL_STATE(1025)] = 26215,
  [SMALL_STATE(1026)] = 26269,
  [SMALL_STATE(1027)] = 26351,
  [SMALL_STATE(1028)] = 26433,
  [SMALL_STATE(1029)] = 26515,
  [SMALL_STATE(1030)] = 26597,
  [SMALL_STATE(1031)] = 26679,
  [SMALL_STATE(1032)] = 26761,
  [SMALL_STATE(1033)] = 26807,
  [SMALL_STATE(1034)] = 26889,
  [SMALL_STATE(1035)] = 26971,
  [SMALL_STATE(1036)] = 27047,
  [SMALL_STATE(1037)] = 27129,
  [SMALL_STATE(1038)] = 27205,
  [SMALL_STATE(1039)] = 27281,
  [SMALL_STATE(1040)] = 27357,
  [SMALL_STATE(1041)] = 27403,
  [SMALL_STATE(1042)] = 27467,
  [SMALL_STATE(1043)] = 27543,
  [SMALL_STATE(1044)] = 27619,
  [SMALL_STATE(1045)] = 27687,
  [SMALL_STATE(1046)] = 27744,
  [SMALL_STATE(1047)] = 27803,
  [SMALL_STATE(1048)] = 27858,
  [SMALL_STATE(1049)] = 27901,
  [SMALL_STATE(1050)] = 27944,
  [SMALL_STATE(1051)] = 27997,
  [SMALL_STATE(1052)] = 28042,
  [SMALL_STATE(1053)] = 28085,
  [SMALL_STATE(1054)] = 28166,
  [SMALL_STATE(1055)] = 28241,
  [SMALL_STATE(1056)] = 28322,
  [SMALL_STATE(1057)] = 28403,
  [SMALL_STATE(1058)] = 28446,
  [SMALL_STATE(1059)] = 28527,
  [SMALL_STATE(1060)] = 28590,
  [SMALL_STATE(1061)] = 28657,
  [SMALL_STATE(1062)] = 28706,
  [SMALL_STATE(1063)] = 28781,
  [SMALL_STATE(1064)] = 28826,
  [SMALL_STATE(1065)] = 28901,
  [SMALL_STATE(1066)] = 28980,
  [SMALL_STATE(1067)] = 29057,
  [SMALL_STATE(1068)] = 29128,
  [SMALL_STATE(1069)] = 29197,
  [SMALL_STATE(1070)] = 29262,
  [SMALL_STATE(1071)] = 29321,
  [SMALL_STATE(1072)] = 29376,
  [SMALL_STATE(1073)] = 29427,
  [SMALL_STATE(1074)] = 29480,
  [SMALL_STATE(1075)] = 29525,
  [SMALL_STATE(1076)] = 29606,
  [SMALL_STATE(1077)] = 29681,
  [SMALL_STATE(1078)] = 29756,
  [SMALL_STATE(1079)] = 29831,
  [SMALL_STATE(1080)] = 29906,
  [SMALL_STATE(1081)] = 29981,
  [SMALL_STATE(1082)] = 30056,
  [SMALL_STATE(1083)] = 30137,
  [SMALL_STATE(1084)] = 30182,
  [SMALL_STATE(1085)] = 30255,
  [SMALL_STATE(1086)] = 30330,
  [SMALL_STATE(1087)] = 30411,
  [SMALL_STATE(1088)] = 30486,
  [SMALL_STATE(1089)] = 30529,
  [SMALL_STATE(1090)] = 30572,
  [SMALL_STATE(1091)] = 30615,
  [SMALL_STATE(1092)] = 30696,
  [SMALL_STATE(1093)] = 30739,
  [SMALL_STATE(1094)] = 30820,
  [SMALL_STATE(1095)] = 30863,
  [SMALL_STATE(1096)] = 30906,
  [SMALL_STATE(1097)] = 30949,
  [SMALL_STATE(1098)] = 30992,
  [SMALL_STATE(1099)] = 31035,
  [SMALL_STATE(1100)] = 31078,
  [SMALL_STATE(1101)] = 31121,
  [SMALL_STATE(1102)] = 31164,
  [SMALL_STATE(1103)] = 31245,
  [SMALL_STATE(1104)] = 31288,
  [SMALL_STATE(1105)] = 31369,
  [SMALL_STATE(1106)] = 31412,
  [SMALL_STATE(1107)] = 31455,
  [SMALL_STATE(1108)] = 31530,
  [SMALL_STATE(1109)] = 31573,
  [SMALL_STATE(1110)] = 31654,
  [SMALL_STATE(1111)] = 31697,
  [SMALL_STATE(1112)] = 31778,
  [SMALL_STATE(1113)] = 31821,
  [SMALL_STATE(1114)] = 31884,
  [SMALL_STATE(1115)] = 31951,
  [SMALL_STATE(1116)] = 32000,
  [SMALL_STATE(1117)] = 32075,
  [SMALL_STATE(1118)] = 32120,
  [SMALL_STATE(1119)] = 32195,
  [SMALL_STATE(1120)] = 32238,
  [SMALL_STATE(1121)] = 32317,
  [SMALL_STATE(1122)] = 32394,
  [SMALL_STATE(1123)] = 32475,
  [SMALL_STATE(1124)] = 32556,
  [SMALL_STATE(1125)] = 32639,
  [SMALL_STATE(1126)] = 32720,
  [SMALL_STATE(1127)] = 32801,
  [SMALL_STATE(1128)] = 32882,
  [SMALL_STATE(1129)] = 32963,
  [SMALL_STATE(1130)] = 33006,
  [SMALL_STATE(1131)] = 33077,
  [SMALL_STATE(1132)] = 33160,
  [SMALL_STATE(1133)] = 33203,
  [SMALL_STATE(1134)] = 33284,
  [SMALL_STATE(1135)] = 33353,
  [SMALL_STATE(1136)] = 33418,
  [SMALL_STATE(1137)] = 33469,
  [SMALL_STATE(1138)] = 33543,
  [SMALL_STATE(1139)] = 33621,
  [SMALL_STATE(1140)] = 33697,
  [SMALL_STATE(1141)] = 33767,
  [SMALL_STATE(1142)] = 33835,
  [SMALL_STATE(1143)] = 33899,
  [SMALL_STATE(1144)] = 33957,
  [SMALL_STATE(1145)] = 34011,
  [SMALL_STATE(1146)] = 34061,
  [SMALL_STATE(1147)] = 34113,
  [SMALL_STATE(1148)] = 34157,
  [SMALL_STATE(1149)] = 34231,
  [SMALL_STATE(1150)] = 34305,
  [SMALL_STATE(1151)] = 34379,
  [SMALL_STATE(1152)] = 34459,
  [SMALL_STATE(1153)] = 34533,
  [SMALL_STATE(1154)] = 34613,
  [SMALL_STATE(1155)] = 34693,
  [SMALL_STATE(1156)] = 34767,
  [SMALL_STATE(1157)] = 34847,
  [SMALL_STATE(1158)] = 34927,
  [SMALL_STATE(1159)] = 35007,
  [SMALL_STATE(1160)] = 35089,
  [SMALL_STATE(1161)] = 35169,
  [SMALL_STATE(1162)] = 35249,
  [SMALL_STATE(1163)] = 35329,
  [SMALL_STATE(1164)] = 35409,
  [SMALL_STATE(1165)] = 35471,
  [SMALL_STATE(1166)] = 35537,
  [SMALL_STATE(1167)] = 35585,
  [SMALL_STATE(1168)] = 35659,
  [SMALL_STATE(1169)] = 35703,
  [SMALL_STATE(1170)] = 35786,
  [SMALL_STATE(1171)] = 35869,
  [SMALL_STATE(1172)] = 35948,
  [SMALL_STATE(1173)] = 35991,
  [SMALL_STATE(1174)] = 36074,
  [SMALL_STATE(1175)] = 36157,
  [SMALL_STATE(1176)] = 36240,
  [SMALL_STATE(1177)] = 36283,
  [SMALL_STATE(1178)] = 36366,
  [SMALL_STATE(1179)] = 36449,
  [SMALL_STATE(1180)] = 36532,
  [SMALL_STATE(1181)] = 36612,
  [SMALL_STATE(1182)] = 36678,
  [SMALL_STATE(1183)] = 36740,
  [SMALL_STATE(1184)] = 36820,
  [SMALL_STATE(1185)] = 36882,
  [SMALL_STATE(1186)] = 36962,
  [SMALL_STATE(1187)] = 37042,
  [SMALL_STATE(1188)] = 37104,
  [SMALL_STATE(1189)] = 37184,
  [SMALL_STATE(1190)] = 37264,
  [SMALL_STATE(1191)] = 37326,
  [SMALL_STATE(1192)] = 37406,
  [SMALL_STATE(1193)] = 37472,
  [SMALL_STATE(1194)] = 37552,
  [SMALL_STATE(1195)] = 37632,
  [SMALL_STATE(1196)] = 37712,
  [SMALL_STATE(1197)] = 37792,
  [SMALL_STATE(1198)] = 37872,
  [SMALL_STATE(1199)] = 37952,
  [SMALL_STATE(1200)] = 38032,
  [SMALL_STATE(1201)] = 38112,
  [SMALL_STATE(1202)] = 38194,
  [SMALL_STATE(1203)] = 38272,
  [SMALL_STATE(1204)] = 38350,
  [SMALL_STATE(1205)] = 38430,
  [SMALL_STATE(1206)] = 38509,
  [SMALL_STATE(1207)] = 38586,
  [SMALL_STATE(1208)] = 38663,
  [SMALL_STATE(1209)] = 38734,
  [SMALL_STATE(1210)] = 38811,
  [SMALL_STATE(1211)] = 38888,
  [SMALL_STATE(1212)] = 38959,
  [SMALL_STATE(1213)] = 39036,
  [SMALL_STATE(1214)] = 39113,
  [SMALL_STATE(1215)] = 39190,
  [SMALL_STATE(1216)] = 39267,
  [SMALL_STATE(1217)] = 39338,
  [SMALL_STATE(1218)] = 39415,
  [SMALL_STATE(1219)] = 39492,
  [SMALL_STATE(1220)] = 39571,
  [SMALL_STATE(1221)] = 39650,
  [SMALL_STATE(1222)] = 39727,
  [SMALL_STATE(1223)] = 39804,
  [SMALL_STATE(1224)] = 39875,
  [SMALL_STATE(1225)] = 39952,
  [SMALL_STATE(1226)] = 40031,
  [SMALL_STATE(1227)] = 40108,
  [SMALL_STATE(1228)] = 40187,
  [SMALL_STATE(1229)] = 40258,
  [SMALL_STATE(1230)] = 40335,
  [SMALL_STATE(1231)] = 40412,
  [SMALL_STATE(1232)] = 40483,
  [SMALL_STATE(1233)] = 40554,
  [SMALL_STATE(1234)] = 40594,
  [SMALL_STATE(1235)] = 40670,
  [SMALL_STATE(1236)] = 40746,
  [SMALL_STATE(1237)] = 40806,
  [SMALL_STATE(1238)] = 40882,
  [SMALL_STATE(1239)] = 40958,
  [SMALL_STATE(1240)] = 41034,
  [SMALL_STATE(1241)] = 41110,
  [SMALL_STATE(1242)] = 41170,
  [SMALL_STATE(1243)] = 41246,
  [SMALL_STATE(1244)] = 41322,
  [SMALL_STATE(1245)] = 41398,
  [SMALL_STATE(1246)] = 41474,
  [SMALL_STATE(1247)] = 41550,
  [SMALL_STATE(1248)] = 41626,
  [SMALL_STATE(1249)] = 41702,
  [SMALL_STATE(1250)] = 41778,
  [SMALL_STATE(1251)] = 41854,
  [SMALL_STATE(1252)] = 41930,
  [SMALL_STATE(1253)] = 42006,
  [SMALL_STATE(1254)] = 42082,
  [SMALL_STATE(1255)] = 42158,
  [SMALL_STATE(1256)] = 42234,
  [SMALL_STATE(1257)] = 42310,
  [SMALL_STATE(1258)] = 42386,
  [SMALL_STATE(1259)] = 42462,
  [SMALL_STATE(1260)] = 42538,
  [SMALL_STATE(1261)] = 42614,
  [SMALL_STATE(1262)] = 42690,
  [SMALL_STATE(1263)] = 42766,
  [SMALL_STATE(1264)] = 42842,
  [SMALL_STATE(1265)] = 42918,
  [SMALL_STATE(1266)] = 42994,
  [SMALL_STATE(1267)] = 43070,
  [SMALL_STATE(1268)] = 43146,
  [SMALL_STATE(1269)] = 43222,
  [SMALL_STATE(1270)] = 43298,
  [SMALL_STATE(1271)] = 43374,
  [SMALL_STATE(1272)] = 43450,
  [SMALL_STATE(1273)] = 43526,
  [SMALL_STATE(1274)] = 43602,
  [SMALL_STATE(1275)] = 43678,
  [SMALL_STATE(1276)] = 43754,
  [SMALL_STATE(1277)] = 43830,
  [SMALL_STATE(1278)] = 43906,
  [SMALL_STATE(1279)] = 43982,
  [SMALL_STATE(1280)] = 44058,
  [SMALL_STATE(1281)] = 44134,
  [SMALL_STATE(1282)] = 44210,
  [SMALL_STATE(1283)] = 44286,
  [SMALL_STATE(1284)] = 44362,
  [SMALL_STATE(1285)] = 44438,
  [SMALL_STATE(1286)] = 44514,
  [SMALL_STATE(1287)] = 44590,
  [SMALL_STATE(1288)] = 44627,
  [SMALL_STATE(1289)] = 44692,
  [SMALL_STATE(1290)] = 44729,
  [SMALL_STATE(1291)] = 44794,
  [SMALL_STATE(1292)] = 44831,
  [SMALL_STATE(1293)] = 44868,
  [SMALL_STATE(1294)] = 44933,
  [SMALL_STATE(1295)] = 44998,
  [SMALL_STATE(1296)] = 45063,
  [SMALL_STATE(1297)] = 45099,
  [SMALL_STATE(1298)] = 45152,
  [SMALL_STATE(1299)] = 45205,
  [SMALL_STATE(1300)] = 45242,
  [SMALL_STATE(1301)] = 45279,
  [SMALL_STATE(1302)] = 45310,
  [SMALL_STATE(1303)] = 45341,
  [SMALL_STATE(1304)] = 45372,
  [SMALL_STATE(1305)] = 45414,
  [SMALL_STATE(1306)] = 45456,
  [SMALL_STATE(1307)] = 45498,
  [SMALL_STATE(1308)] = 45548,
  [SMALL_STATE(1309)] = 45598,
  [SMALL_STATE(1310)] = 45640,
  [SMALL_STATE(1311)] = 45682,
  [SMALL_STATE(1312)] = 45738,
  [SMALL_STATE(1313)] = 45780,
  [SMALL_STATE(1314)] = 45822,
  [SMALL_STATE(1315)] = 45864,
  [SMALL_STATE(1316)] = 45906,
  [SMALL_STATE(1317)] = 45962,
  [SMALL_STATE(1318)] = 46004,
  [SMALL_STATE(1319)] = 46046,
  [SMALL_STATE(1320)] = 46085,
  [SMALL_STATE(1321)] = 46124,
  [SMALL_STATE(1322)] = 46163,
  [SMALL_STATE(1323)] = 46196,
  [SMALL_STATE(1324)] = 46235,
  [SMALL_STATE(1325)] = 46274,
  [SMALL_STATE(1326)] = 46307,
  [SMALL_STATE(1327)] = 46346,
  [SMALL_STATE(1328)] = 46382,
  [SMALL_STATE(1329)] = 46418,
  [SMALL_STATE(1330)] = 46445,
  [SMALL_STATE(1331)] = 46472,
  [SMALL_STATE(1332)] = 46517,
  [SMALL_STATE(1333)] = 46544,
  [SMALL_STATE(1334)] = 46578,
  [SMALL_STATE(1335)] = 46620,
  [SMALL_STATE(1336)] = 46654,
  [SMALL_STATE(1337)] = 46688,
  [SMALL_STATE(1338)] = 46727,
  [SMALL_STATE(1339)] = 46766,
  [SMALL_STATE(1340)] = 46805,
  [SMALL_STATE(1341)] = 46844,
  [SMALL_STATE(1342)] = 46883,
  [SMALL_STATE(1343)] = 46919,
  [SMALL_STATE(1344)] = 46955,
  [SMALL_STATE(1345)] = 46991,
  [SMALL_STATE(1346)] = 47027,
  [SMALL_STATE(1347)] = 47067,
  [SMALL_STATE(1348)] = 47103,
  [SMALL_STATE(1349)] = 47139,
  [SMALL_STATE(1350)] = 47175,
  [SMALL_STATE(1351)] = 47211,
  [SMALL_STATE(1352)] = 47235,
  [SMALL_STATE(1353)] = 47275,
  [SMALL_STATE(1354)] = 47311,
  [SMALL_STATE(1355)] = 47348,
  [SMALL_STATE(1356)] = 47383,
  [SMALL_STATE(1357)] = 47418,
  [SMALL_STATE(1358)] = 47453,
  [SMALL_STATE(1359)] = 47487,
  [SMALL_STATE(1360)] = 47521,
  [SMALL_STATE(1361)] = 47555,
  [SMALL_STATE(1362)] = 47589,
  [SMALL_STATE(1363)] = 47623,
  [SMALL_STATE(1364)] = 47657,
  [SMALL_STATE(1365)] = 47691,
  [SMALL_STATE(1366)] = 47725,
  [SMALL_STATE(1367)] = 47745,
  [SMALL_STATE(1368)] = 47779,
  [SMALL_STATE(1369)] = 47799,
  [SMALL_STATE(1370)] = 47819,
  [SMALL_STATE(1371)] = 47839,
  [SMALL_STATE(1372)] = 47859,
  [SMALL_STATE(1373)] = 47893,
  [SMALL_STATE(1374)] = 47927,
  [SMALL_STATE(1375)] = 47947,
  [SMALL_STATE(1376)] = 47967,
  [SMALL_STATE(1377)] = 47987,
  [SMALL_STATE(1378)] = 48007,
  [SMALL_STATE(1379)] = 48041,
  [SMALL_STATE(1380)] = 48061,
  [SMALL_STATE(1381)] = 48081,
  [SMALL_STATE(1382)] = 48101,
  [SMALL_STATE(1383)] = 48120,
  [SMALL_STATE(1384)] = 48139,
  [SMALL_STATE(1385)] = 48170,
  [SMALL_STATE(1386)] = 48201,
  [SMALL_STATE(1387)] = 48220,
  [SMALL_STATE(1388)] = 48243,
  [SMALL_STATE(1389)] = 48262,
  [SMALL_STATE(1390)] = 48293,
  [SMALL_STATE(1391)] = 48324,
  [SMALL_STATE(1392)] = 48343,
  [SMALL_STATE(1393)] = 48368,
  [SMALL_STATE(1394)] = 48387,
  [SMALL_STATE(1395)] = 48406,
  [SMALL_STATE(1396)] = 48425,
  [SMALL_STATE(1397)] = 48444,
  [SMALL_STATE(1398)] = 48463,
  [SMALL_STATE(1399)] = 48482,
  [SMALL_STATE(1400)] = 48501,
  [SMALL_STATE(1401)] = 48520,
  [SMALL_STATE(1402)] = 48539,
  [SMALL_STATE(1403)] = 48558,
  [SMALL_STATE(1404)] = 48577,
  [SMALL_STATE(1405)] = 48596,
  [SMALL_STATE(1406)] = 48615,
  [SMALL_STATE(1407)] = 48634,
  [SMALL_STATE(1408)] = 48662,
  [SMALL_STATE(1409)] = 48690,
  [SMALL_STATE(1410)] = 48720,
  [SMALL_STATE(1411)] = 48750,
  [SMALL_STATE(1412)] = 48786,
  [SMALL_STATE(1413)] = 48816,
  [SMALL_STATE(1414)] = 48844,
  [SMALL_STATE(1415)] = 48872,
  [SMALL_STATE(1416)] = 48900,
  [SMALL_STATE(1417)] = 48928,
  [SMALL_STATE(1418)] = 48956,
  [SMALL_STATE(1419)] = 48984,
  [SMALL_STATE(1420)] = 49014,
  [SMALL_STATE(1421)] = 49042,
  [SMALL_STATE(1422)] = 49072,
  [SMALL_STATE(1423)] = 49102,
  [SMALL_STATE(1424)] = 49130,
  [SMALL_STATE(1425)] = 49160,
  [SMALL_STATE(1426)] = 49190,
  [SMALL_STATE(1427)] = 49218,
  [SMALL_STATE(1428)] = 49246,
  [SMALL_STATE(1429)] = 49274,
  [SMALL_STATE(1430)] = 49304,
  [SMALL_STATE(1431)] = 49340,
  [SMALL_STATE(1432)] = 49368,
  [SMALL_STATE(1433)] = 49398,
  [SMALL_STATE(1434)] = 49434,
  [SMALL_STATE(1435)] = 49456,
  [SMALL_STATE(1436)] = 49492,
  [SMALL_STATE(1437)] = 49528,
  [SMALL_STATE(1438)] = 49556,
  [SMALL_STATE(1439)] = 49586,
  [SMALL_STATE(1440)] = 49614,
  [SMALL_STATE(1441)] = 49642,
  [SMALL_STATE(1442)] = 49670,
  [SMALL_STATE(1443)] = 49698,
  [SMALL_STATE(1444)] = 49726,
  [SMALL_STATE(1445)] = 49756,
  [SMALL_STATE(1446)] = 49783,
  [SMALL_STATE(1447)] = 49802,
  [SMALL_STATE(1448)] = 49823,
  [SMALL_STATE(1449)] = 49842,
  [SMALL_STATE(1450)] = 49869,
  [SMALL_STATE(1451)] = 49888,
  [SMALL_STATE(1452)] = 49915,
  [SMALL_STATE(1453)] = 49943,
  [SMALL_STATE(1454)] = 49961,
  [SMALL_STATE(1455)] = 49989,
  [SMALL_STATE(1456)] = 50007,
  [SMALL_STATE(1457)] = 50029,
  [SMALL_STATE(1458)] = 50053,
  [SMALL_STATE(1459)] = 50075,
  [SMALL_STATE(1460)] = 50099,
  [SMALL_STATE(1461)] = 50123,
  [SMALL_STATE(1462)] = 50145,
  [SMALL_STATE(1463)] = 50169,
  [SMALL_STATE(1464)] = 50193,
  [SMALL_STATE(1465)] = 50215,
  [SMALL_STATE(1466)] = 50237,
  [SMALL_STATE(1467)] = 50259,
  [SMALL_STATE(1468)] = 50277,
  [SMALL_STATE(1469)] = 50295,
  [SMALL_STATE(1470)] = 50317,
  [SMALL_STATE(1471)] = 50339,
  [SMALL_STATE(1472)] = 50357,
  [SMALL_STATE(1473)] = 50381,
  [SMALL_STATE(1474)] = 50403,
  [SMALL_STATE(1475)] = 50427,
  [SMALL_STATE(1476)] = 50451,
  [SMALL_STATE(1477)] = 50469,
  [SMALL_STATE(1478)] = 50493,
  [SMALL_STATE(1479)] = 50520,
  [SMALL_STATE(1480)] = 50537,
  [SMALL_STATE(1481)] = 50564,
  [SMALL_STATE(1482)] = 50591,
  [SMALL_STATE(1483)] = 50618,
  [SMALL_STATE(1484)] = 50635,
  [SMALL_STATE(1485)] = 50658,
  [SMALL_STATE(1486)] = 50685,
  [SMALL_STATE(1487)] = 50712,
  [SMALL_STATE(1488)] = 50729,
  [SMALL_STATE(1489)] = 50756,
  [SMALL_STATE(1490)] = 50783,
  [SMALL_STATE(1491)] = 50806,
  [SMALL_STATE(1492)] = 50829,
  [SMALL_STATE(1493)] = 50848,
  [SMALL_STATE(1494)] = 50870,
  [SMALL_STATE(1495)] = 50888,
  [SMALL_STATE(1496)] = 50912,
  [SMALL_STATE(1497)] = 50940,
  [SMALL_STATE(1498)] = 50962,
  [SMALL_STATE(1499)] = 50990,
  [SMALL_STATE(1500)] = 51018,
  [SMALL_STATE(1501)] = 51046,
  [SMALL_STATE(1502)] = 51066,
  [SMALL_STATE(1503)] = 51088,
  [SMALL_STATE(1504)] = 51116,
  [SMALL_STATE(1505)] = 51138,
  [SMALL_STATE(1506)] = 51166,
  [SMALL_STATE(1507)] = 51188,
  [SMALL_STATE(1508)] = 51216,
  [SMALL_STATE(1509)] = 51238,
  [SMALL_STATE(1510)] = 51260,
  [SMALL_STATE(1511)] = 51288,
  [SMALL_STATE(1512)] = 51310,
  [SMALL_STATE(1513)] = 51331,
  [SMALL_STATE(1514)] = 51348,
  [SMALL_STATE(1515)] = 51367,
  [SMALL_STATE(1516)] = 51386,
  [SMALL_STATE(1517)] = 51405,
  [SMALL_STATE(1518)] = 51424,
  [SMALL_STATE(1519)] = 51445,
  [SMALL_STATE(1520)] = 51460,
  [SMALL_STATE(1521)] = 51481,
  [SMALL_STATE(1522)] = 51500,
  [SMALL_STATE(1523)] = 51519,
  [SMALL_STATE(1524)] = 51538,
  [SMALL_STATE(1525)] = 51559,
  [SMALL_STATE(1526)] = 51580,
  [SMALL_STATE(1527)] = 51595,
  [SMALL_STATE(1528)] = 51610,
  [SMALL_STATE(1529)] = 51631,
  [SMALL_STATE(1530)] = 51650,
  [SMALL_STATE(1531)] = 51673,
  [SMALL_STATE(1532)] = 51694,
  [SMALL_STATE(1533)] = 51709,
  [SMALL_STATE(1534)] = 51724,
  [SMALL_STATE(1535)] = 51739,
  [SMALL_STATE(1536)] = 51758,
  [SMALL_STATE(1537)] = 51779,
  [SMALL_STATE(1538)] = 51800,
  [SMALL_STATE(1539)] = 51815,
  [SMALL_STATE(1540)] = 51832,
  [SMALL_STATE(1541)] = 51847,
  [SMALL_STATE(1542)] = 51862,
  [SMALL_STATE(1543)] = 51877,
  [SMALL_STATE(1544)] = 51892,
  [SMALL_STATE(1545)] = 51913,
  [SMALL_STATE(1546)] = 51934,
  [SMALL_STATE(1547)] = 51955,
  [SMALL_STATE(1548)] = 51976,
  [SMALL_STATE(1549)] = 51997,
  [SMALL_STATE(1550)] = 52016,
  [SMALL_STATE(1551)] = 52033,
  [SMALL_STATE(1552)] = 52052,
  [SMALL_STATE(1553)] = 52071,
  [SMALL_STATE(1554)] = 52088,
  [SMALL_STATE(1555)] = 52103,
  [SMALL_STATE(1556)] = 52124,
  [SMALL_STATE(1557)] = 52139,
  [SMALL_STATE(1558)] = 52160,
  [SMALL_STATE(1559)] = 52179,
  [SMALL_STATE(1560)] = 52200,
  [SMALL_STATE(1561)] = 52221,
  [SMALL_STATE(1562)] = 52240,
  [SMALL_STATE(1563)] = 52259,
  [SMALL_STATE(1564)] = 52274,
  [SMALL_STATE(1565)] = 52296,
  [SMALL_STATE(1566)] = 52318,
  [SMALL_STATE(1567)] = 52332,
  [SMALL_STATE(1568)] = 52346,
  [SMALL_STATE(1569)] = 52360,
  [SMALL_STATE(1570)] = 52374,
  [SMALL_STATE(1571)] = 52388,
  [SMALL_STATE(1572)] = 52410,
  [SMALL_STATE(1573)] = 52426,
  [SMALL_STATE(1574)] = 52442,
  [SMALL_STATE(1575)] = 52464,
  [SMALL_STATE(1576)] = 52486,
  [SMALL_STATE(1577)] = 52500,
  [SMALL_STATE(1578)] = 52514,
  [SMALL_STATE(1579)] = 52528,
  [SMALL_STATE(1580)] = 52544,
  [SMALL_STATE(1581)] = 52558,
  [SMALL_STATE(1582)] = 52572,
  [SMALL_STATE(1583)] = 52592,
  [SMALL_STATE(1584)] = 52606,
  [SMALL_STATE(1585)] = 52628,
  [SMALL_STATE(1586)] = 52642,
  [SMALL_STATE(1587)] = 52656,
  [SMALL_STATE(1588)] = 52670,
  [SMALL_STATE(1589)] = 52684,
  [SMALL_STATE(1590)] = 52704,
  [SMALL_STATE(1591)] = 52722,
  [SMALL_STATE(1592)] = 52738,
  [SMALL_STATE(1593)] = 52752,
  [SMALL_STATE(1594)] = 52774,
  [SMALL_STATE(1595)] = 52792,
  [SMALL_STATE(1596)] = 52806,
  [SMALL_STATE(1597)] = 52828,
  [SMALL_STATE(1598)] = 52848,
  [SMALL_STATE(1599)] = 52870,
  [SMALL_STATE(1600)] = 52892,
  [SMALL_STATE(1601)] = 52910,
  [SMALL_STATE(1602)] = 52928,
  [SMALL_STATE(1603)] = 52950,
  [SMALL_STATE(1604)] = 52972,
  [SMALL_STATE(1605)] = 52994,
  [SMALL_STATE(1606)] = 53016,
  [SMALL_STATE(1607)] = 53034,
  [SMALL_STATE(1608)] = 53052,
  [SMALL_STATE(1609)] = 53074,
  [SMALL_STATE(1610)] = 53096,
  [SMALL_STATE(1611)] = 53118,
  [SMALL_STATE(1612)] = 53140,
  [SMALL_STATE(1613)] = 53158,
  [SMALL_STATE(1614)] = 53180,
  [SMALL_STATE(1615)] = 53198,
  [SMALL_STATE(1616)] = 53216,
  [SMALL_STATE(1617)] = 53232,
  [SMALL_STATE(1618)] = 53250,
  [SMALL_STATE(1619)] = 53266,
  [SMALL_STATE(1620)] = 53284,
  [SMALL_STATE(1621)] = 53300,
  [SMALL_STATE(1622)] = 53322,
  [SMALL_STATE(1623)] = 53340,
  [SMALL_STATE(1624)] = 53356,
  [SMALL_STATE(1625)] = 53374,
  [SMALL_STATE(1626)] = 53392,
  [SMALL_STATE(1627)] = 53410,
  [SMALL_STATE(1628)] = 53428,
  [SMALL_STATE(1629)] = 53446,
  [SMALL_STATE(1630)] = 53464,
  [SMALL_STATE(1631)] = 53486,
  [SMALL_STATE(1632)] = 53508,
  [SMALL_STATE(1633)] = 53530,
  [SMALL_STATE(1634)] = 53552,
  [SMALL_STATE(1635)] = 53570,
  [SMALL_STATE(1636)] = 53584,
  [SMALL_STATE(1637)] = 53598,
  [SMALL_STATE(1638)] = 53616,
  [SMALL_STATE(1639)] = 53630,
  [SMALL_STATE(1640)] = 53650,
  [SMALL_STATE(1641)] = 53672,
  [SMALL_STATE(1642)] = 53694,
  [SMALL_STATE(1643)] = 53708,
  [SMALL_STATE(1644)] = 53724,
  [SMALL_STATE(1645)] = 53738,
  [SMALL_STATE(1646)] = 53758,
  [SMALL_STATE(1647)] = 53780,
  [SMALL_STATE(1648)] = 53799,
  [SMALL_STATE(1649)] = 53816,
  [SMALL_STATE(1650)] = 53833,
  [SMALL_STATE(1651)] = 53850,
  [SMALL_STATE(1652)] = 53861,
  [SMALL_STATE(1653)] = 53880,
  [SMALL_STATE(1654)] = 53897,
  [SMALL_STATE(1655)] = 53912,
  [SMALL_STATE(1656)] = 53931,
  [SMALL_STATE(1657)] = 53950,
  [SMALL_STATE(1658)] = 53967,
  [SMALL_STATE(1659)] = 53984,
  [SMALL_STATE(1660)] = 54001,
  [SMALL_STATE(1661)] = 54018,
  [SMALL_STATE(1662)] = 54035,
  [SMALL_STATE(1663)] = 54052,
  [SMALL_STATE(1664)] = 54071,
  [SMALL_STATE(1665)] = 54090,
  [SMALL_STATE(1666)] = 54109,
  [SMALL_STATE(1667)] = 54124,
  [SMALL_STATE(1668)] = 54137,
  [SMALL_STATE(1669)] = 54150,
  [SMALL_STATE(1670)] = 54169,
  [SMALL_STATE(1671)] = 54182,
  [SMALL_STATE(1672)] = 54201,
  [SMALL_STATE(1673)] = 54218,
  [SMALL_STATE(1674)] = 54235,
  [SMALL_STATE(1675)] = 54250,
  [SMALL_STATE(1676)] = 54263,
  [SMALL_STATE(1677)] = 54282,
  [SMALL_STATE(1678)] = 54297,
  [SMALL_STATE(1679)] = 54314,
  [SMALL_STATE(1680)] = 54331,
  [SMALL_STATE(1681)] = 54348,
  [SMALL_STATE(1682)] = 54361,
  [SMALL_STATE(1683)] = 54374,
  [SMALL_STATE(1684)] = 54387,
  [SMALL_STATE(1685)] = 54406,
  [SMALL_STATE(1686)] = 54425,
  [SMALL_STATE(1687)] = 54442,
  [SMALL_STATE(1688)] = 54459,
  [SMALL_STATE(1689)] = 54476,
  [SMALL_STATE(1690)] = 54493,
  [SMALL_STATE(1691)] = 54504,
  [SMALL_STATE(1692)] = 54517,
  [SMALL_STATE(1693)] = 54534,
  [SMALL_STATE(1694)] = 54545,
  [SMALL_STATE(1695)] = 54564,
  [SMALL_STATE(1696)] = 54581,
  [SMALL_STATE(1697)] = 54598,
  [SMALL_STATE(1698)] = 54615,
  [SMALL_STATE(1699)] = 54632,
  [SMALL_STATE(1700)] = 54649,
  [SMALL_STATE(1701)] = 54666,
  [SMALL_STATE(1702)] = 54677,
  [SMALL_STATE(1703)] = 54694,
  [SMALL_STATE(1704)] = 54711,
  [SMALL_STATE(1705)] = 54724,
  [SMALL_STATE(1706)] = 54739,
  [SMALL_STATE(1707)] = 54750,
  [SMALL_STATE(1708)] = 54767,
  [SMALL_STATE(1709)] = 54780,
  [SMALL_STATE(1710)] = 54797,
  [SMALL_STATE(1711)] = 54812,
  [SMALL_STATE(1712)] = 54829,
  [SMALL_STATE(1713)] = 54844,
  [SMALL_STATE(1714)] = 54857,
  [SMALL_STATE(1715)] = 54874,
  [SMALL_STATE(1716)] = 54889,
  [SMALL_STATE(1717)] = 54902,
  [SMALL_STATE(1718)] = 54919,
  [SMALL_STATE(1719)] = 54934,
  [SMALL_STATE(1720)] = 54945,
  [SMALL_STATE(1721)] = 54960,
  [SMALL_STATE(1722)] = 54973,
  [SMALL_STATE(1723)] = 54990,
  [SMALL_STATE(1724)] = 55003,
  [SMALL_STATE(1725)] = 55018,
  [SMALL_STATE(1726)] = 55033,
  [SMALL_STATE(1727)] = 55048,
  [SMALL_STATE(1728)] = 55063,
  [SMALL_STATE(1729)] = 55082,
  [SMALL_STATE(1730)] = 55101,
  [SMALL_STATE(1731)] = 55114,
  [SMALL_STATE(1732)] = 55127,
  [SMALL_STATE(1733)] = 55140,
  [SMALL_STATE(1734)] = 55159,
  [SMALL_STATE(1735)] = 55176,
  [SMALL_STATE(1736)] = 55195,
  [SMALL_STATE(1737)] = 55208,
  [SMALL_STATE(1738)] = 55221,
  [SMALL_STATE(1739)] = 55234,
  [SMALL_STATE(1740)] = 55253,
  [SMALL_STATE(1741)] = 55270,
  [SMALL_STATE(1742)] = 55287,
  [SMALL_STATE(1743)] = 55304,
  [SMALL_STATE(1744)] = 55317,
  [SMALL_STATE(1745)] = 55334,
  [SMALL_STATE(1746)] = 55349,
  [SMALL_STATE(1747)] = 55364,
  [SMALL_STATE(1748)] = 55381,
  [SMALL_STATE(1749)] = 55400,
  [SMALL_STATE(1750)] = 55419,
  [SMALL_STATE(1751)] = 55438,
  [SMALL_STATE(1752)] = 55455,
  [SMALL_STATE(1753)] = 55474,
  [SMALL_STATE(1754)] = 55485,
  [SMALL_STATE(1755)] = 55500,
  [SMALL_STATE(1756)] = 55517,
  [SMALL_STATE(1757)] = 55532,
  [SMALL_STATE(1758)] = 55547,
  [SMALL_STATE(1759)] = 55562,
  [SMALL_STATE(1760)] = 55577,
  [SMALL_STATE(1761)] = 55596,
  [SMALL_STATE(1762)] = 55609,
  [SMALL_STATE(1763)] = 55622,
  [SMALL_STATE(1764)] = 55635,
  [SMALL_STATE(1765)] = 55652,
  [SMALL_STATE(1766)] = 55665,
  [SMALL_STATE(1767)] = 55678,
  [SMALL_STATE(1768)] = 55693,
  [SMALL_STATE(1769)] = 55708,
  [SMALL_STATE(1770)] = 55723,
  [SMALL_STATE(1771)] = 55738,
  [SMALL_STATE(1772)] = 55753,
  [SMALL_STATE(1773)] = 55770,
  [SMALL_STATE(1774)] = 55784,
  [SMALL_STATE(1775)] = 55800,
  [SMALL_STATE(1776)] = 55812,
  [SMALL_STATE(1777)] = 55824,
  [SMALL_STATE(1778)] = 55840,
  [SMALL_STATE(1779)] = 55856,
  [SMALL_STATE(1780)] = 55868,
  [SMALL_STATE(1781)] = 55884,
  [SMALL_STATE(1782)] = 55900,
  [SMALL_STATE(1783)] = 55916,
  [SMALL_STATE(1784)] = 55926,
  [SMALL_STATE(1785)] = 55940,
  [SMALL_STATE(1786)] = 55954,
  [SMALL_STATE(1787)] = 55970,
  [SMALL_STATE(1788)] = 55984,
  [SMALL_STATE(1789)] = 55994,
  [SMALL_STATE(1790)] = 56008,
  [SMALL_STATE(1791)] = 56022,
  [SMALL_STATE(1792)] = 56036,
  [SMALL_STATE(1793)] = 56052,
  [SMALL_STATE(1794)] = 56066,
  [SMALL_STATE(1795)] = 56082,
  [SMALL_STATE(1796)] = 56096,
  [SMALL_STATE(1797)] = 56112,
  [SMALL_STATE(1798)] = 56128,
  [SMALL_STATE(1799)] = 56144,
  [SMALL_STATE(1800)] = 56154,
  [SMALL_STATE(1801)] = 56168,
  [SMALL_STATE(1802)] = 56178,
  [SMALL_STATE(1803)] = 56192,
  [SMALL_STATE(1804)] = 56208,
  [SMALL_STATE(1805)] = 56218,
  [SMALL_STATE(1806)] = 56234,
  [SMALL_STATE(1807)] = 56250,
  [SMALL_STATE(1808)] = 56264,
  [SMALL_STATE(1809)] = 56280,
  [SMALL_STATE(1810)] = 56296,
  [SMALL_STATE(1811)] = 56308,
  [SMALL_STATE(1812)] = 56324,
  [SMALL_STATE(1813)] = 56338,
  [SMALL_STATE(1814)] = 56354,
  [SMALL_STATE(1815)] = 56370,
  [SMALL_STATE(1816)] = 56380,
  [SMALL_STATE(1817)] = 56392,
  [SMALL_STATE(1818)] = 56408,
  [SMALL_STATE(1819)] = 56420,
  [SMALL_STATE(1820)] = 56430,
  [SMALL_STATE(1821)] = 56446,
  [SMALL_STATE(1822)] = 56458,
  [SMALL_STATE(1823)] = 56474,
  [SMALL_STATE(1824)] = 56484,
  [SMALL_STATE(1825)] = 56500,
  [SMALL_STATE(1826)] = 56516,
  [SMALL_STATE(1827)] = 56526,
  [SMALL_STATE(1828)] = 56542,
  [SMALL_STATE(1829)] = 56558,
  [SMALL_STATE(1830)] = 56574,
  [SMALL_STATE(1831)] = 56590,
  [SMALL_STATE(1832)] = 56600,
  [SMALL_STATE(1833)] = 56616,
  [SMALL_STATE(1834)] = 56628,
  [SMALL_STATE(1835)] = 56642,
  [SMALL_STATE(1836)] = 56656,
  [SMALL_STATE(1837)] = 56672,
  [SMALL_STATE(1838)] = 56688,
  [SMALL_STATE(1839)] = 56702,
  [SMALL_STATE(1840)] = 56718,
  [SMALL_STATE(1841)] = 56732,
  [SMALL_STATE(1842)] = 56748,
  [SMALL_STATE(1843)] = 56762,
  [SMALL_STATE(1844)] = 56778,
  [SMALL_STATE(1845)] = 56790,
  [SMALL_STATE(1846)] = 56806,
  [SMALL_STATE(1847)] = 56816,
  [SMALL_STATE(1848)] = 56830,
  [SMALL_STATE(1849)] = 56846,
  [SMALL_STATE(1850)] = 56862,
  [SMALL_STATE(1851)] = 56878,
  [SMALL_STATE(1852)] = 56888,
  [SMALL_STATE(1853)] = 56898,
  [SMALL_STATE(1854)] = 56914,
  [SMALL_STATE(1855)] = 56930,
  [SMALL_STATE(1856)] = 56942,
  [SMALL_STATE(1857)] = 56958,
  [SMALL_STATE(1858)] = 56974,
  [SMALL_STATE(1859)] = 56988,
  [SMALL_STATE(1860)] = 57002,
  [SMALL_STATE(1861)] = 57018,
  [SMALL_STATE(1862)] = 57028,
  [SMALL_STATE(1863)] = 57040,
  [SMALL_STATE(1864)] = 57050,
  [SMALL_STATE(1865)] = 57062,
  [SMALL_STATE(1866)] = 57078,
  [SMALL_STATE(1867)] = 57092,
  [SMALL_STATE(1868)] = 57108,
  [SMALL_STATE(1869)] = 57124,
  [SMALL_STATE(1870)] = 57138,
  [SMALL_STATE(1871)] = 57152,
  [SMALL_STATE(1872)] = 57166,
  [SMALL_STATE(1873)] = 57180,
  [SMALL_STATE(1874)] = 57196,
  [SMALL_STATE(1875)] = 57208,
  [SMALL_STATE(1876)] = 57224,
  [SMALL_STATE(1877)] = 57238,
  [SMALL_STATE(1878)] = 57250,
  [SMALL_STATE(1879)] = 57261,
  [SMALL_STATE(1880)] = 57274,
  [SMALL_STATE(1881)] = 57287,
  [SMALL_STATE(1882)] = 57300,
  [SMALL_STATE(1883)] = 57311,
  [SMALL_STATE(1884)] = 57324,
  [SMALL_STATE(1885)] = 57335,
  [SMALL_STATE(1886)] = 57346,
  [SMALL_STATE(1887)] = 57357,
  [SMALL_STATE(1888)] = 57368,
  [SMALL_STATE(1889)] = 57381,
  [SMALL_STATE(1890)] = 57394,
  [SMALL_STATE(1891)] = 57405,
  [SMALL_STATE(1892)] = 57416,
  [SMALL_STATE(1893)] = 57427,
  [SMALL_STATE(1894)] = 57436,
  [SMALL_STATE(1895)] = 57449,
  [SMALL_STATE(1896)] = 57462,
  [SMALL_STATE(1897)] = 57473,
  [SMALL_STATE(1898)] = 57486,
  [SMALL_STATE(1899)] = 57497,
  [SMALL_STATE(1900)] = 57510,
  [SMALL_STATE(1901)] = 57521,
  [SMALL_STATE(1902)] = 57532,
  [SMALL_STATE(1903)] = 57545,
  [SMALL_STATE(1904)] = 57556,
  [SMALL_STATE(1905)] = 57567,
  [SMALL_STATE(1906)] = 57578,
  [SMALL_STATE(1907)] = 57589,
  [SMALL_STATE(1908)] = 57600,
  [SMALL_STATE(1909)] = 57613,
  [SMALL_STATE(1910)] = 57624,
  [SMALL_STATE(1911)] = 57637,
  [SMALL_STATE(1912)] = 57650,
  [SMALL_STATE(1913)] = 57663,
  [SMALL_STATE(1914)] = 57676,
  [SMALL_STATE(1915)] = 57687,
  [SMALL_STATE(1916)] = 57700,
  [SMALL_STATE(1917)] = 57711,
  [SMALL_STATE(1918)] = 57722,
  [SMALL_STATE(1919)] = 57733,
  [SMALL_STATE(1920)] = 57744,
  [SMALL_STATE(1921)] = 57755,
  [SMALL_STATE(1922)] = 57766,
  [SMALL_STATE(1923)] = 57777,
  [SMALL_STATE(1924)] = 57790,
  [SMALL_STATE(1925)] = 57803,
  [SMALL_STATE(1926)] = 57816,
  [SMALL_STATE(1927)] = 57827,
  [SMALL_STATE(1928)] = 57838,
  [SMALL_STATE(1929)] = 57849,
  [SMALL_STATE(1930)] = 57860,
  [SMALL_STATE(1931)] = 57871,
  [SMALL_STATE(1932)] = 57882,
  [SMALL_STATE(1933)] = 57895,
  [SMALL_STATE(1934)] = 57906,
  [SMALL_STATE(1935)] = 57919,
  [SMALL_STATE(1936)] = 57930,
  [SMALL_STATE(1937)] = 57941,
  [SMALL_STATE(1938)] = 57952,
  [SMALL_STATE(1939)] = 57963,
  [SMALL_STATE(1940)] = 57974,
  [SMALL_STATE(1941)] = 57987,
  [SMALL_STATE(1942)] = 57998,
  [SMALL_STATE(1943)] = 58009,
  [SMALL_STATE(1944)] = 58020,
  [SMALL_STATE(1945)] = 58033,
  [SMALL_STATE(1946)] = 58046,
  [SMALL_STATE(1947)] = 58057,
  [SMALL_STATE(1948)] = 58068,
  [SMALL_STATE(1949)] = 58079,
  [SMALL_STATE(1950)] = 58092,
  [SMALL_STATE(1951)] = 58103,
  [SMALL_STATE(1952)] = 58114,
  [SMALL_STATE(1953)] = 58125,
  [SMALL_STATE(1954)] = 58136,
  [SMALL_STATE(1955)] = 58147,
  [SMALL_STATE(1956)] = 58158,
  [SMALL_STATE(1957)] = 58169,
  [SMALL_STATE(1958)] = 58180,
  [SMALL_STATE(1959)] = 58191,
  [SMALL_STATE(1960)] = 58202,
  [SMALL_STATE(1961)] = 58215,
  [SMALL_STATE(1962)] = 58226,
  [SMALL_STATE(1963)] = 58239,
  [SMALL_STATE(1964)] = 58250,
  [SMALL_STATE(1965)] = 58263,
  [SMALL_STATE(1966)] = 58276,
  [SMALL_STATE(1967)] = 58289,
  [SMALL_STATE(1968)] = 58300,
  [SMALL_STATE(1969)] = 58311,
  [SMALL_STATE(1970)] = 58324,
  [SMALL_STATE(1971)] = 58337,
  [SMALL_STATE(1972)] = 58348,
  [SMALL_STATE(1973)] = 58361,
  [SMALL_STATE(1974)] = 58374,
  [SMALL_STATE(1975)] = 58383,
  [SMALL_STATE(1976)] = 58394,
  [SMALL_STATE(1977)] = 58407,
  [SMALL_STATE(1978)] = 58418,
  [SMALL_STATE(1979)] = 58429,
  [SMALL_STATE(1980)] = 58440,
  [SMALL_STATE(1981)] = 58451,
  [SMALL_STATE(1982)] = 58462,
  [SMALL_STATE(1983)] = 58473,
  [SMALL_STATE(1984)] = 58486,
  [SMALL_STATE(1985)] = 58497,
  [SMALL_STATE(1986)] = 58508,
  [SMALL_STATE(1987)] = 58519,
  [SMALL_STATE(1988)] = 58530,
  [SMALL_STATE(1989)] = 58543,
  [SMALL_STATE(1990)] = 58554,
  [SMALL_STATE(1991)] = 58565,
  [SMALL_STATE(1992)] = 58576,
  [SMALL_STATE(1993)] = 58589,
  [SMALL_STATE(1994)] = 58600,
  [SMALL_STATE(1995)] = 58611,
  [SMALL_STATE(1996)] = 58624,
  [SMALL_STATE(1997)] = 58635,
  [SMALL_STATE(1998)] = 58646,
  [SMALL_STATE(1999)] = 58657,
  [SMALL_STATE(2000)] = 58670,
  [SMALL_STATE(2001)] = 58681,
  [SMALL_STATE(2002)] = 58692,
  [SMALL_STATE(2003)] = 58703,
  [SMALL_STATE(2004)] = 58716,
  [SMALL_STATE(2005)] = 58729,
  [SMALL_STATE(2006)] = 58740,
  [SMALL_STATE(2007)] = 58751,
  [SMALL_STATE(2008)] = 58764,
  [SMALL_STATE(2009)] = 58775,
  [SMALL_STATE(2010)] = 58788,
  [SMALL_STATE(2011)] = 58801,
  [SMALL_STATE(2012)] = 58812,
  [SMALL_STATE(2013)] = 58823,
  [SMALL_STATE(2014)] = 58836,
  [SMALL_STATE(2015)] = 58849,
  [SMALL_STATE(2016)] = 58860,
  [SMALL_STATE(2017)] = 58871,
  [SMALL_STATE(2018)] = 58882,
  [SMALL_STATE(2019)] = 58895,
  [SMALL_STATE(2020)] = 58906,
  [SMALL_STATE(2021)] = 58917,
  [SMALL_STATE(2022)] = 58928,
  [SMALL_STATE(2023)] = 58939,
  [SMALL_STATE(2024)] = 58950,
  [SMALL_STATE(2025)] = 58959,
  [SMALL_STATE(2026)] = 58972,
  [SMALL_STATE(2027)] = 58981,
  [SMALL_STATE(2028)] = 58992,
  [SMALL_STATE(2029)] = 59003,
  [SMALL_STATE(2030)] = 59014,
  [SMALL_STATE(2031)] = 59025,
  [SMALL_STATE(2032)] = 59036,
  [SMALL_STATE(2033)] = 59047,
  [SMALL_STATE(2034)] = 59058,
  [SMALL_STATE(2035)] = 59071,
  [SMALL_STATE(2036)] = 59082,
  [SMALL_STATE(2037)] = 59093,
  [SMALL_STATE(2038)] = 59104,
  [SMALL_STATE(2039)] = 59115,
  [SMALL_STATE(2040)] = 59128,
  [SMALL_STATE(2041)] = 59139,
  [SMALL_STATE(2042)] = 59150,
  [SMALL_STATE(2043)] = 59163,
  [SMALL_STATE(2044)] = 59174,
  [SMALL_STATE(2045)] = 59187,
  [SMALL_STATE(2046)] = 59198,
  [SMALL_STATE(2047)] = 59207,
  [SMALL_STATE(2048)] = 59218,
  [SMALL_STATE(2049)] = 59229,
  [SMALL_STATE(2050)] = 59240,
  [SMALL_STATE(2051)] = 59251,
  [SMALL_STATE(2052)] = 59260,
  [SMALL_STATE(2053)] = 59271,
  [SMALL_STATE(2054)] = 59284,
  [SMALL_STATE(2055)] = 59295,
  [SMALL_STATE(2056)] = 59306,
  [SMALL_STATE(2057)] = 59317,
  [SMALL_STATE(2058)] = 59330,
  [SMALL_STATE(2059)] = 59343,
  [SMALL_STATE(2060)] = 59354,
  [SMALL_STATE(2061)] = 59367,
  [SMALL_STATE(2062)] = 59378,
  [SMALL_STATE(2063)] = 59387,
  [SMALL_STATE(2064)] = 59398,
  [SMALL_STATE(2065)] = 59409,
  [SMALL_STATE(2066)] = 59422,
  [SMALL_STATE(2067)] = 59431,
  [SMALL_STATE(2068)] = 59442,
  [SMALL_STATE(2069)] = 59455,
  [SMALL_STATE(2070)] = 59468,
  [SMALL_STATE(2071)] = 59481,
  [SMALL_STATE(2072)] = 59494,
  [SMALL_STATE(2073)] = 59507,
  [SMALL_STATE(2074)] = 59518,
  [SMALL_STATE(2075)] = 59529,
  [SMALL_STATE(2076)] = 59540,
  [SMALL_STATE(2077)] = 59551,
  [SMALL_STATE(2078)] = 59562,
  [SMALL_STATE(2079)] = 59573,
  [SMALL_STATE(2080)] = 59584,
  [SMALL_STATE(2081)] = 59597,
  [SMALL_STATE(2082)] = 59610,
  [SMALL_STATE(2083)] = 59623,
  [SMALL_STATE(2084)] = 59636,
  [SMALL_STATE(2085)] = 59649,
  [SMALL_STATE(2086)] = 59662,
  [SMALL_STATE(2087)] = 59673,
  [SMALL_STATE(2088)] = 59686,
  [SMALL_STATE(2089)] = 59697,
  [SMALL_STATE(2090)] = 59710,
  [SMALL_STATE(2091)] = 59721,
  [SMALL_STATE(2092)] = 59732,
  [SMALL_STATE(2093)] = 59743,
  [SMALL_STATE(2094)] = 59752,
  [SMALL_STATE(2095)] = 59765,
  [SMALL_STATE(2096)] = 59774,
  [SMALL_STATE(2097)] = 59787,
  [SMALL_STATE(2098)] = 59798,
  [SMALL_STATE(2099)] = 59809,
  [SMALL_STATE(2100)] = 59820,
  [SMALL_STATE(2101)] = 59831,
  [SMALL_STATE(2102)] = 59842,
  [SMALL_STATE(2103)] = 59855,
  [SMALL_STATE(2104)] = 59868,
  [SMALL_STATE(2105)] = 59879,
  [SMALL_STATE(2106)] = 59892,
  [SMALL_STATE(2107)] = 59901,
  [SMALL_STATE(2108)] = 59914,
  [SMALL_STATE(2109)] = 59927,
  [SMALL_STATE(2110)] = 59940,
  [SMALL_STATE(2111)] = 59953,
  [SMALL_STATE(2112)] = 59964,
  [SMALL_STATE(2113)] = 59975,
  [SMALL_STATE(2114)] = 59986,
  [SMALL_STATE(2115)] = 59999,
  [SMALL_STATE(2116)] = 60010,
  [SMALL_STATE(2117)] = 60021,
  [SMALL_STATE(2118)] = 60032,
  [SMALL_STATE(2119)] = 60045,
  [SMALL_STATE(2120)] = 60056,
  [SMALL_STATE(2121)] = 60069,
  [SMALL_STATE(2122)] = 60080,
  [SMALL_STATE(2123)] = 60091,
  [SMALL_STATE(2124)] = 60102,
  [SMALL_STATE(2125)] = 60113,
  [SMALL_STATE(2126)] = 60126,
  [SMALL_STATE(2127)] = 60139,
  [SMALL_STATE(2128)] = 60148,
  [SMALL_STATE(2129)] = 60159,
  [SMALL_STATE(2130)] = 60168,
  [SMALL_STATE(2131)] = 60181,
  [SMALL_STATE(2132)] = 60190,
  [SMALL_STATE(2133)] = 60203,
  [SMALL_STATE(2134)] = 60216,
  [SMALL_STATE(2135)] = 60227,
  [SMALL_STATE(2136)] = 60240,
  [SMALL_STATE(2137)] = 60251,
  [SMALL_STATE(2138)] = 60264,
  [SMALL_STATE(2139)] = 60273,
  [SMALL_STATE(2140)] = 60282,
  [SMALL_STATE(2141)] = 60293,
  [SMALL_STATE(2142)] = 60304,
  [SMALL_STATE(2143)] = 60315,
  [SMALL_STATE(2144)] = 60326,
  [SMALL_STATE(2145)] = 60335,
  [SMALL_STATE(2146)] = 60346,
  [SMALL_STATE(2147)] = 60357,
  [SMALL_STATE(2148)] = 60366,
  [SMALL_STATE(2149)] = 60377,
  [SMALL_STATE(2150)] = 60388,
  [SMALL_STATE(2151)] = 60399,
  [SMALL_STATE(2152)] = 60412,
  [SMALL_STATE(2153)] = 60425,
  [SMALL_STATE(2154)] = 60436,
  [SMALL_STATE(2155)] = 60449,
  [SMALL_STATE(2156)] = 60462,
  [SMALL_STATE(2157)] = 60475,
  [SMALL_STATE(2158)] = 60486,
  [SMALL_STATE(2159)] = 60497,
  [SMALL_STATE(2160)] = 60510,
  [SMALL_STATE(2161)] = 60523,
  [SMALL_STATE(2162)] = 60534,
  [SMALL_STATE(2163)] = 60547,
  [SMALL_STATE(2164)] = 60558,
  [SMALL_STATE(2165)] = 60571,
  [SMALL_STATE(2166)] = 60582,
  [SMALL_STATE(2167)] = 60591,
  [SMALL_STATE(2168)] = 60600,
  [SMALL_STATE(2169)] = 60611,
  [SMALL_STATE(2170)] = 60622,
  [SMALL_STATE(2171)] = 60633,
  [SMALL_STATE(2172)] = 60644,
  [SMALL_STATE(2173)] = 60655,
  [SMALL_STATE(2174)] = 60664,
  [SMALL_STATE(2175)] = 60675,
  [SMALL_STATE(2176)] = 60684,
  [SMALL_STATE(2177)] = 60695,
  [SMALL_STATE(2178)] = 60708,
  [SMALL_STATE(2179)] = 60719,
  [SMALL_STATE(2180)] = 60730,
  [SMALL_STATE(2181)] = 60741,
  [SMALL_STATE(2182)] = 60752,
  [SMALL_STATE(2183)] = 60763,
  [SMALL_STATE(2184)] = 60774,
  [SMALL_STATE(2185)] = 60787,
  [SMALL_STATE(2186)] = 60798,
  [SMALL_STATE(2187)] = 60809,
  [SMALL_STATE(2188)] = 60822,
  [SMALL_STATE(2189)] = 60833,
  [SMALL_STATE(2190)] = 60844,
  [SMALL_STATE(2191)] = 60855,
  [SMALL_STATE(2192)] = 60866,
  [SMALL_STATE(2193)] = 60877,
  [SMALL_STATE(2194)] = 60888,
  [SMALL_STATE(2195)] = 60899,
  [SMALL_STATE(2196)] = 60910,
  [SMALL_STATE(2197)] = 60923,
  [SMALL_STATE(2198)] = 60936,
  [SMALL_STATE(2199)] = 60949,
  [SMALL_STATE(2200)] = 60960,
  [SMALL_STATE(2201)] = 60971,
  [SMALL_STATE(2202)] = 60982,
  [SMALL_STATE(2203)] = 60993,
  [SMALL_STATE(2204)] = 61004,
  [SMALL_STATE(2205)] = 61015,
  [SMALL_STATE(2206)] = 61026,
  [SMALL_STATE(2207)] = 61035,
  [SMALL_STATE(2208)] = 61046,
  [SMALL_STATE(2209)] = 61059,
  [SMALL_STATE(2210)] = 61070,
  [SMALL_STATE(2211)] = 61081,
  [SMALL_STATE(2212)] = 61092,
  [SMALL_STATE(2213)] = 61103,
  [SMALL_STATE(2214)] = 61116,
  [SMALL_STATE(2215)] = 61127,
  [SMALL_STATE(2216)] = 61138,
  [SMALL_STATE(2217)] = 61151,
  [SMALL_STATE(2218)] = 61159,
  [SMALL_STATE(2219)] = 61167,
  [SMALL_STATE(2220)] = 61177,
  [SMALL_STATE(2221)] = 61187,
  [SMALL_STATE(2222)] = 61195,
  [SMALL_STATE(2223)] = 61203,
  [SMALL_STATE(2224)] = 61213,
  [SMALL_STATE(2225)] = 61223,
  [SMALL_STATE(2226)] = 61233,
  [SMALL_STATE(2227)] = 61241,
  [SMALL_STATE(2228)] = 61251,
  [SMALL_STATE(2229)] = 61261,
  [SMALL_STATE(2230)] = 61269,
  [SMALL_STATE(2231)] = 61279,
  [SMALL_STATE(2232)] = 61289,
  [SMALL_STATE(2233)] = 61299,
  [SMALL_STATE(2234)] = 61309,
  [SMALL_STATE(2235)] = 61317,
  [SMALL_STATE(2236)] = 61325,
  [SMALL_STATE(2237)] = 61335,
  [SMALL_STATE(2238)] = 61345,
  [SMALL_STATE(2239)] = 61355,
  [SMALL_STATE(2240)] = 61363,
  [SMALL_STATE(2241)] = 61373,
  [SMALL_STATE(2242)] = 61381,
  [SMALL_STATE(2243)] = 61389,
  [SMALL_STATE(2244)] = 61399,
  [SMALL_STATE(2245)] = 61409,
  [SMALL_STATE(2246)] = 61419,
  [SMALL_STATE(2247)] = 61429,
  [SMALL_STATE(2248)] = 61439,
  [SMALL_STATE(2249)] = 61447,
  [SMALL_STATE(2250)] = 61457,
  [SMALL_STATE(2251)] = 61467,
  [SMALL_STATE(2252)] = 61477,
  [SMALL_STATE(2253)] = 61487,
  [SMALL_STATE(2254)] = 61497,
  [SMALL_STATE(2255)] = 61507,
  [SMALL_STATE(2256)] = 61517,
  [SMALL_STATE(2257)] = 61525,
  [SMALL_STATE(2258)] = 61535,
  [SMALL_STATE(2259)] = 61545,
  [SMALL_STATE(2260)] = 61555,
  [SMALL_STATE(2261)] = 61565,
  [SMALL_STATE(2262)] = 61575,
  [SMALL_STATE(2263)] = 61583,
  [SMALL_STATE(2264)] = 61593,
  [SMALL_STATE(2265)] = 61603,
  [SMALL_STATE(2266)] = 61613,
  [SMALL_STATE(2267)] = 61623,
  [SMALL_STATE(2268)] = 61633,
  [SMALL_STATE(2269)] = 61643,
  [SMALL_STATE(2270)] = 61653,
  [SMALL_STATE(2271)] = 61663,
  [SMALL_STATE(2272)] = 61671,
  [SMALL_STATE(2273)] = 61681,
  [SMALL_STATE(2274)] = 61691,
  [SMALL_STATE(2275)] = 61701,
  [SMALL_STATE(2276)] = 61711,
  [SMALL_STATE(2277)] = 61721,
  [SMALL_STATE(2278)] = 61731,
  [SMALL_STATE(2279)] = 61741,
  [SMALL_STATE(2280)] = 61751,
  [SMALL_STATE(2281)] = 61759,
  [SMALL_STATE(2282)] = 61769,
  [SMALL_STATE(2283)] = 61779,
  [SMALL_STATE(2284)] = 61789,
  [SMALL_STATE(2285)] = 61799,
  [SMALL_STATE(2286)] = 61809,
  [SMALL_STATE(2287)] = 61819,
  [SMALL_STATE(2288)] = 61829,
  [SMALL_STATE(2289)] = 61839,
  [SMALL_STATE(2290)] = 61847,
  [SMALL_STATE(2291)] = 61857,
  [SMALL_STATE(2292)] = 61867,
  [SMALL_STATE(2293)] = 61877,
  [SMALL_STATE(2294)] = 61887,
  [SMALL_STATE(2295)] = 61897,
  [SMALL_STATE(2296)] = 61907,
  [SMALL_STATE(2297)] = 61915,
  [SMALL_STATE(2298)] = 61923,
  [SMALL_STATE(2299)] = 61933,
  [SMALL_STATE(2300)] = 61941,
  [SMALL_STATE(2301)] = 61951,
  [SMALL_STATE(2302)] = 61961,
  [SMALL_STATE(2303)] = 61969,
  [SMALL_STATE(2304)] = 61977,
  [SMALL_STATE(2305)] = 61985,
  [SMALL_STATE(2306)] = 61995,
  [SMALL_STATE(2307)] = 62005,
  [SMALL_STATE(2308)] = 62013,
  [SMALL_STATE(2309)] = 62021,
  [SMALL_STATE(2310)] = 62029,
  [SMALL_STATE(2311)] = 62039,
  [SMALL_STATE(2312)] = 62047,
  [SMALL_STATE(2313)] = 62057,
  [SMALL_STATE(2314)] = 62067,
  [SMALL_STATE(2315)] = 62077,
  [SMALL_STATE(2316)] = 62087,
  [SMALL_STATE(2317)] = 62095,
  [SMALL_STATE(2318)] = 62105,
  [SMALL_STATE(2319)] = 62113,
  [SMALL_STATE(2320)] = 62121,
  [SMALL_STATE(2321)] = 62129,
  [SMALL_STATE(2322)] = 62137,
  [SMALL_STATE(2323)] = 62147,
  [SMALL_STATE(2324)] = 62157,
  [SMALL_STATE(2325)] = 62167,
  [SMALL_STATE(2326)] = 62177,
  [SMALL_STATE(2327)] = 62187,
  [SMALL_STATE(2328)] = 62197,
  [SMALL_STATE(2329)] = 62205,
  [SMALL_STATE(2330)] = 62213,
  [SMALL_STATE(2331)] = 62223,
  [SMALL_STATE(2332)] = 62233,
  [SMALL_STATE(2333)] = 62241,
  [SMALL_STATE(2334)] = 62251,
  [SMALL_STATE(2335)] = 62259,
  [SMALL_STATE(2336)] = 62269,
  [SMALL_STATE(2337)] = 62279,
  [SMALL_STATE(2338)] = 62287,
  [SMALL_STATE(2339)] = 62295,
  [SMALL_STATE(2340)] = 62305,
  [SMALL_STATE(2341)] = 62315,
  [SMALL_STATE(2342)] = 62325,
  [SMALL_STATE(2343)] = 62335,
  [SMALL_STATE(2344)] = 62343,
  [SMALL_STATE(2345)] = 62353,
  [SMALL_STATE(2346)] = 62361,
  [SMALL_STATE(2347)] = 62371,
  [SMALL_STATE(2348)] = 62379,
  [SMALL_STATE(2349)] = 62389,
  [SMALL_STATE(2350)] = 62397,
  [SMALL_STATE(2351)] = 62405,
  [SMALL_STATE(2352)] = 62413,
  [SMALL_STATE(2353)] = 62421,
  [SMALL_STATE(2354)] = 62431,
  [SMALL_STATE(2355)] = 62441,
  [SMALL_STATE(2356)] = 62451,
  [SMALL_STATE(2357)] = 62461,
  [SMALL_STATE(2358)] = 62471,
  [SMALL_STATE(2359)] = 62481,
  [SMALL_STATE(2360)] = 62491,
  [SMALL_STATE(2361)] = 62501,
  [SMALL_STATE(2362)] = 62509,
  [SMALL_STATE(2363)] = 62519,
  [SMALL_STATE(2364)] = 62529,
  [SMALL_STATE(2365)] = 62539,
  [SMALL_STATE(2366)] = 62549,
  [SMALL_STATE(2367)] = 62559,
  [SMALL_STATE(2368)] = 62567,
  [SMALL_STATE(2369)] = 62577,
  [SMALL_STATE(2370)] = 62587,
  [SMALL_STATE(2371)] = 62597,
  [SMALL_STATE(2372)] = 62607,
  [SMALL_STATE(2373)] = 62615,
  [SMALL_STATE(2374)] = 62625,
  [SMALL_STATE(2375)] = 62633,
  [SMALL_STATE(2376)] = 62643,
  [SMALL_STATE(2377)] = 62653,
  [SMALL_STATE(2378)] = 62663,
  [SMALL_STATE(2379)] = 62671,
  [SMALL_STATE(2380)] = 62681,
  [SMALL_STATE(2381)] = 62691,
  [SMALL_STATE(2382)] = 62701,
  [SMALL_STATE(2383)] = 62711,
  [SMALL_STATE(2384)] = 62719,
  [SMALL_STATE(2385)] = 62729,
  [SMALL_STATE(2386)] = 62739,
  [SMALL_STATE(2387)] = 62747,
  [SMALL_STATE(2388)] = 62757,
  [SMALL_STATE(2389)] = 62767,
  [SMALL_STATE(2390)] = 62777,
  [SMALL_STATE(2391)] = 62785,
  [SMALL_STATE(2392)] = 62795,
  [SMALL_STATE(2393)] = 62805,
  [SMALL_STATE(2394)] = 62813,
  [SMALL_STATE(2395)] = 62821,
  [SMALL_STATE(2396)] = 62829,
  [SMALL_STATE(2397)] = 62839,
  [SMALL_STATE(2398)] = 62847,
  [SMALL_STATE(2399)] = 62857,
  [SMALL_STATE(2400)] = 62867,
  [SMALL_STATE(2401)] = 62875,
  [SMALL_STATE(2402)] = 62885,
  [SMALL_STATE(2403)] = 62895,
  [SMALL_STATE(2404)] = 62903,
  [SMALL_STATE(2405)] = 62911,
  [SMALL_STATE(2406)] = 62921,
  [SMALL_STATE(2407)] = 62929,
  [SMALL_STATE(2408)] = 62937,
  [SMALL_STATE(2409)] = 62945,
  [SMALL_STATE(2410)] = 62953,
  [SMALL_STATE(2411)] = 62963,
  [SMALL_STATE(2412)] = 62973,
  [SMALL_STATE(2413)] = 62983,
  [SMALL_STATE(2414)] = 62993,
  [SMALL_STATE(2415)] = 63001,
  [SMALL_STATE(2416)] = 63011,
  [SMALL_STATE(2417)] = 63019,
  [SMALL_STATE(2418)] = 63029,
  [SMALL_STATE(2419)] = 63039,
  [SMALL_STATE(2420)] = 63046,
  [SMALL_STATE(2421)] = 63053,
  [SMALL_STATE(2422)] = 63060,
  [SMALL_STATE(2423)] = 63067,
  [SMALL_STATE(2424)] = 63074,
  [SMALL_STATE(2425)] = 63081,
  [SMALL_STATE(2426)] = 63088,
  [SMALL_STATE(2427)] = 63095,
  [SMALL_STATE(2428)] = 63102,
  [SMALL_STATE(2429)] = 63109,
  [SMALL_STATE(2430)] = 63116,
  [SMALL_STATE(2431)] = 63123,
  [SMALL_STATE(2432)] = 63130,
  [SMALL_STATE(2433)] = 63137,
  [SMALL_STATE(2434)] = 63144,
  [SMALL_STATE(2435)] = 63151,
  [SMALL_STATE(2436)] = 63158,
  [SMALL_STATE(2437)] = 63165,
  [SMALL_STATE(2438)] = 63172,
  [SMALL_STATE(2439)] = 63179,
  [SMALL_STATE(2440)] = 63186,
  [SMALL_STATE(2441)] = 63193,
  [SMALL_STATE(2442)] = 63200,
  [SMALL_STATE(2443)] = 63207,
  [SMALL_STATE(2444)] = 63214,
  [SMALL_STATE(2445)] = 63221,
  [SMALL_STATE(2446)] = 63228,
  [SMALL_STATE(2447)] = 63235,
  [SMALL_STATE(2448)] = 63242,
  [SMALL_STATE(2449)] = 63249,
  [SMALL_STATE(2450)] = 63256,
  [SMALL_STATE(2451)] = 63263,
  [SMALL_STATE(2452)] = 63270,
  [SMALL_STATE(2453)] = 63277,
  [SMALL_STATE(2454)] = 63284,
  [SMALL_STATE(2455)] = 63291,
  [SMALL_STATE(2456)] = 63298,
  [SMALL_STATE(2457)] = 63305,
  [SMALL_STATE(2458)] = 63312,
  [SMALL_STATE(2459)] = 63319,
  [SMALL_STATE(2460)] = 63326,
  [SMALL_STATE(2461)] = 63333,
  [SMALL_STATE(2462)] = 63340,
  [SMALL_STATE(2463)] = 63347,
  [SMALL_STATE(2464)] = 63354,
  [SMALL_STATE(2465)] = 63361,
  [SMALL_STATE(2466)] = 63368,
  [SMALL_STATE(2467)] = 63375,
  [SMALL_STATE(2468)] = 63382,
  [SMALL_STATE(2469)] = 63389,
  [SMALL_STATE(2470)] = 63396,
  [SMALL_STATE(2471)] = 63403,
  [SMALL_STATE(2472)] = 63410,
  [SMALL_STATE(2473)] = 63417,
  [SMALL_STATE(2474)] = 63424,
  [SMALL_STATE(2475)] = 63431,
  [SMALL_STATE(2476)] = 63438,
  [SMALL_STATE(2477)] = 63445,
  [SMALL_STATE(2478)] = 63452,
  [SMALL_STATE(2479)] = 63459,
  [SMALL_STATE(2480)] = 63466,
  [SMALL_STATE(2481)] = 63473,
  [SMALL_STATE(2482)] = 63480,
  [SMALL_STATE(2483)] = 63487,
  [SMALL_STATE(2484)] = 63494,
  [SMALL_STATE(2485)] = 63501,
  [SMALL_STATE(2486)] = 63508,
  [SMALL_STATE(2487)] = 63515,
  [SMALL_STATE(2488)] = 63522,
  [SMALL_STATE(2489)] = 63529,
  [SMALL_STATE(2490)] = 63536,
  [SMALL_STATE(2491)] = 63543,
  [SMALL_STATE(2492)] = 63550,
  [SMALL_STATE(2493)] = 63557,
  [SMALL_STATE(2494)] = 63564,
  [SMALL_STATE(2495)] = 63571,
  [SMALL_STATE(2496)] = 63578,
  [SMALL_STATE(2497)] = 63585,
  [SMALL_STATE(2498)] = 63592,
  [SMALL_STATE(2499)] = 63599,
  [SMALL_STATE(2500)] = 63606,
  [SMALL_STATE(2501)] = 63613,
  [SMALL_STATE(2502)] = 63620,
  [SMALL_STATE(2503)] = 63627,
  [SMALL_STATE(2504)] = 63634,
  [SMALL_STATE(2505)] = 63641,
  [SMALL_STATE(2506)] = 63648,
  [SMALL_STATE(2507)] = 63655,
  [SMALL_STATE(2508)] = 63662,
  [SMALL_STATE(2509)] = 63669,
  [SMALL_STATE(2510)] = 63676,
  [SMALL_STATE(2511)] = 63683,
  [SMALL_STATE(2512)] = 63690,
  [SMALL_STATE(2513)] = 63697,
  [SMALL_STATE(2514)] = 63704,
  [SMALL_STATE(2515)] = 63711,
  [SMALL_STATE(2516)] = 63718,
  [SMALL_STATE(2517)] = 63725,
  [SMALL_STATE(2518)] = 63732,
  [SMALL_STATE(2519)] = 63739,
  [SMALL_STATE(2520)] = 63746,
  [SMALL_STATE(2521)] = 63753,
  [SMALL_STATE(2522)] = 63760,
  [SMALL_STATE(2523)] = 63767,
  [SMALL_STATE(2524)] = 63774,
  [SMALL_STATE(2525)] = 63781,
  [SMALL_STATE(2526)] = 63788,
  [SMALL_STATE(2527)] = 63795,
  [SMALL_STATE(2528)] = 63802,
  [SMALL_STATE(2529)] = 63809,
  [SMALL_STATE(2530)] = 63816,
  [SMALL_STATE(2531)] = 63823,
  [SMALL_STATE(2532)] = 63830,
  [SMALL_STATE(2533)] = 63837,
  [SMALL_STATE(2534)] = 63844,
  [SMALL_STATE(2535)] = 63851,
  [SMALL_STATE(2536)] = 63858,
  [SMALL_STATE(2537)] = 63865,
  [SMALL_STATE(2538)] = 63872,
  [SMALL_STATE(2539)] = 63879,
  [SMALL_STATE(2540)] = 63886,
  [SMALL_STATE(2541)] = 63893,
  [SMALL_STATE(2542)] = 63900,
  [SMALL_STATE(2543)] = 63907,
  [SMALL_STATE(2544)] = 63914,
  [SMALL_STATE(2545)] = 63921,
  [SMALL_STATE(2546)] = 63928,
  [SMALL_STATE(2547)] = 63935,
  [SMALL_STATE(2548)] = 63942,
  [SMALL_STATE(2549)] = 63949,
  [SMALL_STATE(2550)] = 63956,
  [SMALL_STATE(2551)] = 63963,
  [SMALL_STATE(2552)] = 63970,
  [SMALL_STATE(2553)] = 63977,
  [SMALL_STATE(2554)] = 63984,
  [SMALL_STATE(2555)] = 63991,
  [SMALL_STATE(2556)] = 63998,
  [SMALL_STATE(2557)] = 64005,
  [SMALL_STATE(2558)] = 64012,
  [SMALL_STATE(2559)] = 64019,
  [SMALL_STATE(2560)] = 64026,
  [SMALL_STATE(2561)] = 64033,
  [SMALL_STATE(2562)] = 64040,
  [SMALL_STATE(2563)] = 64047,
  [SMALL_STATE(2564)] = 64054,
  [SMALL_STATE(2565)] = 64061,
  [SMALL_STATE(2566)] = 64068,
  [SMALL_STATE(2567)] = 64075,
  [SMALL_STATE(2568)] = 64082,
  [SMALL_STATE(2569)] = 64089,
  [SMALL_STATE(2570)] = 64096,
  [SMALL_STATE(2571)] = 64103,
  [SMALL_STATE(2572)] = 64110,
  [SMALL_STATE(2573)] = 64117,
  [SMALL_STATE(2574)] = 64124,
  [SMALL_STATE(2575)] = 64131,
  [SMALL_STATE(2576)] = 64138,
  [SMALL_STATE(2577)] = 64145,
  [SMALL_STATE(2578)] = 64152,
  [SMALL_STATE(2579)] = 64159,
  [SMALL_STATE(2580)] = 64166,
  [SMALL_STATE(2581)] = 64173,
  [SMALL_STATE(2582)] = 64180,
  [SMALL_STATE(2583)] = 64187,
  [SMALL_STATE(2584)] = 64194,
  [SMALL_STATE(2585)] = 64201,
  [SMALL_STATE(2586)] = 64208,
  [SMALL_STATE(2587)] = 64215,
  [SMALL_STATE(2588)] = 64222,
  [SMALL_STATE(2589)] = 64229,
  [SMALL_STATE(2590)] = 64236,
  [SMALL_STATE(2591)] = 64243,
  [SMALL_STATE(2592)] = 64250,
  [SMALL_STATE(2593)] = 64257,
  [SMALL_STATE(2594)] = 64264,
  [SMALL_STATE(2595)] = 64271,
  [SMALL_STATE(2596)] = 64278,
  [SMALL_STATE(2597)] = 64285,
  [SMALL_STATE(2598)] = 64292,
  [SMALL_STATE(2599)] = 64299,
  [SMALL_STATE(2600)] = 64306,
  [SMALL_STATE(2601)] = 64313,
  [SMALL_STATE(2602)] = 64320,
  [SMALL_STATE(2603)] = 64327,
  [SMALL_STATE(2604)] = 64334,
  [SMALL_STATE(2605)] = 64341,
  [SMALL_STATE(2606)] = 64348,
  [SMALL_STATE(2607)] = 64355,
  [SMALL_STATE(2608)] = 64362,
  [SMALL_STATE(2609)] = 64369,
  [SMALL_STATE(2610)] = 64376,
  [SMALL_STATE(2611)] = 64383,
  [SMALL_STATE(2612)] = 64390,
  [SMALL_STATE(2613)] = 64397,
  [SMALL_STATE(2614)] = 64404,
  [SMALL_STATE(2615)] = 64411,
  [SMALL_STATE(2616)] = 64418,
  [SMALL_STATE(2617)] = 64425,
  [SMALL_STATE(2618)] = 64432,
  [SMALL_STATE(2619)] = 64439,
  [SMALL_STATE(2620)] = 64446,
  [SMALL_STATE(2621)] = 64453,
  [SMALL_STATE(2622)] = 64460,
  [SMALL_STATE(2623)] = 64467,
  [SMALL_STATE(2624)] = 64474,
  [SMALL_STATE(2625)] = 64481,
  [SMALL_STATE(2626)] = 64488,
  [SMALL_STATE(2627)] = 64495,
  [SMALL_STATE(2628)] = 64502,
  [SMALL_STATE(2629)] = 64509,
  [SMALL_STATE(2630)] = 64516,
  [SMALL_STATE(2631)] = 64523,
  [SMALL_STATE(2632)] = 64530,
  [SMALL_STATE(2633)] = 64537,
  [SMALL_STATE(2634)] = 64544,
  [SMALL_STATE(2635)] = 64551,
  [SMALL_STATE(2636)] = 64558,
  [SMALL_STATE(2637)] = 64565,
  [SMALL_STATE(2638)] = 64572,
  [SMALL_STATE(2639)] = 64579,
  [SMALL_STATE(2640)] = 64586,
  [SMALL_STATE(2641)] = 64593,
  [SMALL_STATE(2642)] = 64600,
  [SMALL_STATE(2643)] = 64607,
  [SMALL_STATE(2644)] = 64614,
  [SMALL_STATE(2645)] = 64621,
  [SMALL_STATE(2646)] = 64628,
  [SMALL_STATE(2647)] = 64635,
  [SMALL_STATE(2648)] = 64642,
  [SMALL_STATE(2649)] = 64649,
  [SMALL_STATE(2650)] = 64656,
  [SMALL_STATE(2651)] = 64663,
  [SMALL_STATE(2652)] = 64670,
  [SMALL_STATE(2653)] = 64677,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 0, 0, 0),
  [7] = {.entry = {.count = 1, .reusable = false}}, SHIFT(972),
  [9] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2574),
  [11] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [13] = {.entry = {.count = 1, .reusable = true}}, SHIFT(545),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1328),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1810),
  [19] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1739),
  [21] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1364),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1452),
  [25] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1190),
  [27] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2636),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2623),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2436),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2451),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2527),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1292),
  [41] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1296),
  [43] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1287),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1341),
  [47] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1233),
  [49] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1839),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2639),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(225),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1876),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2498),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2439),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1095),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2357),
  [67] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2633),
  [69] = {.entry = {.count = 1, .reusable = false}}, SHIFT(190),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(194),
  [73] = {.entry = {.count = 1, .reusable = false}}, SHIFT(202),
  [75] = {.entry = {.count = 1, .reusable = false}}, SHIFT(321),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2294),
  [79] = {.entry = {.count = 1, .reusable = false}}, SHIFT(110),
  [81] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2484),
  [83] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2559),
  [85] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2387),
  [87] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2388),
  [89] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2401),
  [91] = {.entry = {.count = 1, .reusable = false}}, SHIFT(328),
  [93] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
  [95] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [97] = {.entry = {.count = 1, .reusable = false}}, SHIFT(582),
  [99] = {.entry = {.count = 1, .reusable = false}}, SHIFT(319),
  [101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(851),
  [103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(699),
  [105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2448),
  [107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1519),
  [111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1129),
  [113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1099),
  [115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1449),
  [117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1365),
  [119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1582),
  [121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2164),
  [123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1378),
  [125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1594),
  [127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(119),
  [129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(329),
  [131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(331),
  [133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(332),
  [135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(334),
  [137] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0),
  [139] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(972),
  [142] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(545),
  [145] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1328),
  [148] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1810),
  [151] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1739),
  [154] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1364),
  [157] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1452),
  [160] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1190),
  [163] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2636),
  [166] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(101),
  [169] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2623),
  [172] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2436),
  [175] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2451),
  [178] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0),
  [180] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2527),
  [183] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1292),
  [186] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1296),
  [189] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1287),
  [192] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1341),
  [195] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1233),
  [198] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(121),
  [201] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1839),
  [204] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2639),
  [207] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(225),
  [210] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1876),
  [213] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2498),
  [216] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2439),
  [219] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1095),
  [222] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2357),
  [225] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2633),
  [228] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(190),
  [231] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(194),
  [234] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(202),
  [237] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(321),
  [240] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2294),
  [243] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(110),
  [246] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2484),
  [249] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2559),
  [252] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2387),
  [255] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2388),
  [258] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2401),
  [261] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(328),
  [264] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(328),
  [267] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(318),
  [270] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(582),
  [273] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(319),
  [276] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(851),
  [279] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(699),
  [282] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2448),
  [285] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(140),
  [288] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1519),
  [291] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1129),
  [294] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1099),
  [297] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1449),
  [300] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1365),
  [303] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1582),
  [306] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2164),
  [309] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1378),
  [312] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1594),
  [315] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(119),
  [318] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(329),
  [321] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(331),
  [324] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(332),
  [327] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(334),
  [330] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 3, 0, 135),
  [332] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 3, 0, 135),
  [334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_default_statement, 3, 0, 0),
  [336] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_default_statement, 3, 0, 0),
  [338] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 4, 0, 135),
  [340] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_statement, 4, 0, 135),
  [342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_default_statement, 2, 0, 0),
  [344] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_default_statement, 2, 0, 0),
  [346] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_colon_block, 1, 0, 0),
  [348] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_colon_block, 2, 0, 0),
  [350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(544),
  [352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7),
  [354] = {.entry = {.count = 1, .reusable = false}}, SHIFT(904),
  [356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1884),
  [358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1327),
  [360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1775),
  [362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1652),
  [364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1367),
  [366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1182),
  [368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [370] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2587),
  [372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2589),
  [374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2592),
  [376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2594),
  [378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1340),
  [380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(231),
  [382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1790),
  [384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2575),
  [386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2576),
  [388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2267),
  [390] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2516),
  [392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(195),
  [394] = {.entry = {.count = 1, .reusable = false}}, SHIFT(198),
  [396] = {.entry = {.count = 1, .reusable = false}}, SHIFT(199),
  [398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2377),
  [400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(116),
  [402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2631),
  [404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2644),
  [406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2379),
  [408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2270),
  [410] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2643),
  [413] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2392),
  [416] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2650),
  [419] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2652),
  [422] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2268),
  [425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2643),
  [427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2392),
  [429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2650),
  [431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2652),
  [433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2268),
  [435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2613),
  [437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2339),
  [439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2641),
  [441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2649),
  [443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2402),
  [445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2195),
  [447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(456),
  [451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(462),
  [455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(552),
  [459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(488),
  [463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2001),
  [467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(80),
  [469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2145),
  [471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1936),
  [475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(96),
  [477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2207),
  [479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2215),
  [483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [485] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 1, 0, 0),
  [487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2533),
  [489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1905),
  [495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1909),
  [499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1916),
  [503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(569),
  [507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(535),
  [511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(533),
  [515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [517] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 2, 0, 0),
  [519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2429),
  [521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(530),
  [523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2064),
  [527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2146),
  [529] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2150),
  [531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2210),
  [533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2211),
  [535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2212),
  [537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(12),
  [539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2104),
  [541] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2117),
  [543] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2123),
  [545] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2124),
  [547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2165),
  [549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2168),
  [551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2183),
  [553] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1961),
  [555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1878),
  [557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1980),
  [559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1990),
  [561] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2049),
  [563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2097),
  [565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(827),
  [567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(797),
  [569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2140),
  [571] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2190),
  [573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2193),
  [575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1900),
  [577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1901),
  [579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1903),
  [581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1906),
  [583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1920),
  [585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1922),
  [587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1926),
  [589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1929),
  [591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1930),
  [593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1933),
  [595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1942),
  [597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1947),
  [599] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1950),
  [601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1951),
  [603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1957),
  [605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1088),
  [607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1057),
  [609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(451),
  [613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(453),
  [615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1721),
  [617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1732),
  [619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(528),
  [621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(730),
  [623] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield_expression, 1, 0, 0),
  [625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(694),
  [627] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1492),
  [629] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_yield_expression, 1, 0, 0),
  [631] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1827),
  [633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2504),
  [635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2608),
  [641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1003),
  [643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(286),
  [645] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2269),
  [647] = {.entry = {.count = 1, .reusable = false}}, SHIFT(287),
  [649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [653] = {.entry = {.count = 1, .reusable = false}}, SHIFT(580),
  [655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(289),
  [657] = {.entry = {.count = 1, .reusable = false}}, SHIFT(852),
  [659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(692),
  [661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(999),
  [665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1000),
  [667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1360),
  [669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1589),
  [671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2034),
  [673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1361),
  [675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1590),
  [677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(117),
  [679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(296),
  [681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(290),
  [683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(291),
  [685] = {.entry = {.count = 1, .reusable = false}}, SHIFT(292),
  [687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(293),
  [689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(376),
  [693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(346),
  [695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(347),
  [697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(579),
  [703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(349),
  [705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(118),
  [709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(354),
  [711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(350),
  [713] = {.entry = {.count = 1, .reusable = false}}, SHIFT(351),
  [715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(352),
  [717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(437),
  [719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(863),
  [721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(687),
  [723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(385),
  [727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(387),
  [733] = {.entry = {.count = 1, .reusable = false}}, SHIFT(388),
  [735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(581),
  [741] = {.entry = {.count = 1, .reusable = false}}, SHIFT(390),
  [743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [745] = {.entry = {.count = 1, .reusable = false}}, SHIFT(120),
  [747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(395),
  [749] = {.entry = {.count = 1, .reusable = false}}, SHIFT(391),
  [751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(392),
  [753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(393),
  [755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(394),
  [757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2292),
  [759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2561),
  [761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(924),
  [763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(873),
  [765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1469),
  [767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2518),
  [769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1748),
  [771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(867),
  [773] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1749),
  [775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2371),
  [779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(237),
  [781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2152),
  [783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1494),
  [785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1172),
  [787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1176),
  [789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(586),
  [791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(634),
  [793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(688),
  [795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(918),
  [799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(855),
  [801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(933),
  [803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(729),
  [805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(762),
  [807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(885),
  [811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(793),
  [813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(635),
  [817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(723),
  [819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(587),
  [821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(969),
  [823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(857),
  [825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(860),
  [827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(967),
  [829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(721),
  [831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(585),
  [833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(778),
  [835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(776),
  [837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(629),
  [839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2651),
  [841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(916),
  [843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2503),
  [845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(954),
  [847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2537),
  [849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(799),
  [851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2423),
  [853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(932),
  [857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(927),
  [859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(823),
  [861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2501),
  [863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1089),
  [865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2511),
  [867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(934),
  [869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(817),
  [871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(814),
  [873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(900),
  [875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(810),
  [877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2599),
  [879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(985),
  [881] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__array_destructing_repeat1, 1, 0, 0),
  [883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1101),
  [889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2062),
  [893] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__array_destructing_repeat1, 1, 0, 0), SHIFT(936),
  [896] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__array_destructing_repeat1, 1, 0, 0), SHIFT(803),
  [899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1049),
  [901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1005),
  [903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1001),
  [905] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_destructing_repeat1, 1, 0, 0),
  [907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(24),
  [913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2056),
  [925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(19),
  [927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2067),
  [931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2075),
  [933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(25),
  [937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(483),
  [939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(33),
  [943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(48),
  [947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(196),
  [955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(39),
  [965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(18),
  [975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(687),
  [979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_condition_list, 3, 0, 0),
  [983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(854),
  [985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(769),
  [987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(624),
  [989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2113),
  [991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(373),
  [995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(414),
  [997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(724),
  [999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [1001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [1003] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_placeholder, 1, 0, 0),
  [1005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [1007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2030),
  [1009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [1011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(600),
  [1013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(939),
  [1015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(591),
  [1017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(766),
  [1019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(946),
  [1021] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_condition_list, 2, 0, 0),
  [1023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [1025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [1027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [1029] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 3, 0, 6),
  [1031] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 3, 0, 6),
  [1033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2645),
  [1035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2231),
  [1037] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0),
  [1039] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0),
  [1041] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2645),
  [1044] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2231),
  [1047] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 4, 0, 53),
  [1049] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 53),
  [1051] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 53), SHIFT(2275),
  [1054] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 53), SHIFT(112),
  [1057] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 3, 0, 23),
  [1059] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 3, 0, 23),
  [1061] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_statement, 3, 0, 23), SHIFT(2275),
  [1064] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_statement, 3, 0, 23), SHIFT(112),
  [1067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2275),
  [1069] = {.entry = {.count = 1, .reusable = false}}, SHIFT(114),
  [1071] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_catch_clause, 6, 0, 187),
  [1073] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_clause, 6, 0, 187),
  [1075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_finally_clause, 2, 0, 6),
  [1077] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_finally_clause, 2, 0, 6),
  [1079] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_statement, 2, 0, 0),
  [1081] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_statement, 2, 0, 0),
  [1083] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat1, 2, 0, 96),
  [1085] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat1, 2, 0, 96),
  [1087] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat1, 2, 0, 96), SHIFT_REPEAT(2275),
  [1090] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_statement, 3, 0, 0),
  [1092] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_statement, 3, 0, 0),
  [1094] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_catch_clause, 5, 0, 164),
  [1096] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_clause, 5, 0, 164),
  [1098] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 7, 0, 95),
  [1100] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 7, 0, 95),
  [1102] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 191),
  [1104] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 191),
  [1106] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_static_declaration, 3, 0, 0),
  [1108] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_static_declaration, 3, 0, 0),
  [1110] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 192),
  [1112] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 192),
  [1114] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_definition, 2, 0, 6),
  [1116] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_definition, 2, 0, 6),
  [1118] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_statement, 3, 0, 23),
  [1120] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_statement, 3, 0, 23),
  [1122] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 193),
  [1124] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 193),
  [1126] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 171),
  [1128] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 171),
  [1130] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exit_statement, 2, 0, 0),
  [1132] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exit_statement, 2, 0, 0),
  [1134] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_continue_statement, 2, 0, 0),
  [1136] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_continue_statement, 2, 0, 0),
  [1138] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_break_statement, 2, 0, 0),
  [1140] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_break_statement, 2, 0, 0),
  [1142] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 166),
  [1144] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 166),
  [1146] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 168),
  [1148] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 168),
  [1150] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_global_declaration, 3, 0, 0),
  [1152] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_global_declaration, 3, 0, 0),
  [1154] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 170),
  [1156] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 170),
  [1158] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 3, 0, 33),
  [1160] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 3, 0, 33),
  [1162] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_static_declaration, 4, 0, 0),
  [1164] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_static_declaration, 4, 0, 0),
  [1166] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_global_declaration, 4, 0, 0),
  [1168] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_global_declaration, 4, 0, 0),
  [1170] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_declaration, 4, 0, 0),
  [1172] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_use_declaration, 4, 0, 0),
  [1174] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_definition, 3, 0, 15),
  [1176] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_definition, 3, 0, 15),
  [1178] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 4, 0, 46),
  [1180] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 4, 0, 46),
  [1182] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 4, 0, 0),
  [1184] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 4, 0, 0),
  [1186] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration_list, 2, 0, 0),
  [1188] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration_list, 2, 0, 0),
  [1190] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_declaration, 4, 0, 48),
  [1192] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interface_declaration, 4, 0, 48),
  [1194] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration_list, 2, 0, 0),
  [1196] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration_list, 2, 0, 0),
  [1198] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 4, 0, 48),
  [1200] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 4, 0, 48),
  [1202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 4, 0, 48),
  [1204] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 4, 0, 48),
  [1206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 2, 0, 0),
  [1208] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 2, 0, 0),
  [1210] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exit_statement, 4, 0, 0),
  [1212] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exit_statement, 4, 0, 0),
  [1214] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_definition, 3, 0, 16),
  [1216] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_definition, 3, 0, 16),
  [1218] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_declaration, 3, 0, 0),
  [1220] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_use_declaration, 3, 0, 0),
  [1222] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 203),
  [1224] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 203),
  [1226] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 204),
  [1228] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 204),
  [1230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreach_statement, 9, 0, 171),
  [1232] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_foreach_statement, 9, 0, 171),
  [1234] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 209),
  [1236] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 209),
  [1238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 210),
  [1240] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 210),
  [1242] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 189),
  [1244] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 189),
  [1246] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 211),
  [1248] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 211),
  [1250] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat1, 1, 0, 51),
  [1252] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat1, 1, 0, 51),
  [1254] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 4, 0, 52),
  [1256] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 52),
  [1258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 191),
  [1260] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 191),
  [1262] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_block, 2, 0, 0),
  [1264] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_block, 2, 0, 0),
  [1266] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 10, 0, 193),
  [1268] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 10, 0, 193),
  [1270] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_declaration, 3, 0, 18),
  [1272] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_namespace_use_declaration, 3, 0, 18),
  [1274] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement, 1, 0, 0),
  [1276] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement, 1, 0, 0),
  [1278] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 11, 0, 213),
  [1280] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 11, 0, 213),
  [1282] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 11, 0, 214),
  [1284] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 11, 0, 214),
  [1286] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 11, 0, 215),
  [1288] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 11, 0, 215),
  [1290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 11, 0, 204),
  [1292] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 11, 0, 204),
  [1294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 12, 0, 216),
  [1296] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 12, 0, 216),
  [1298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_declaration, 4, 0, 72),
  [1300] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_declaration, 4, 0, 72),
  [1302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_declaration, 4, 0, 72),
  [1304] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interface_declaration, 4, 0, 72),
  [1306] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 2, 0, 12),
  [1308] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 2, 0, 12),
  [1310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unset_statement, 7, 0, 0),
  [1312] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unset_statement, 7, 0, 0),
  [1314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 4, 0, 72),
  [1316] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 4, 0, 72),
  [1318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 4, 0, 72),
  [1320] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 4, 0, 72),
  [1322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 4, 0, 74),
  [1324] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 4, 0, 74),
  [1326] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declare_statement, 7, 0, 0),
  [1328] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declare_statement, 7, 0, 0),
  [1330] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_block, 4, 0, 0),
  [1332] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_block, 4, 0, 0),
  [1334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 3, 0, 0),
  [1336] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 3, 0, 0),
  [1338] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 5, 0, 46),
  [1340] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 5, 0, 46),
  [1342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration_list, 3, 0, 0),
  [1344] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration_list, 3, 0, 0),
  [1346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration_list, 3, 0, 0),
  [1348] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration_list, 3, 0, 0),
  [1350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_named_label_statement, 2, 0, 0),
  [1352] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_named_label_statement, 2, 0, 0),
  [1354] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 5, 0, 93),
  [1356] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 5, 0, 93),
  [1358] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_if_clause, 3, 0, 23),
  [1360] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_if_clause, 3, 0, 23),
  [1362] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 165),
  [1364] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 165),
  [1366] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 6, 0, 132),
  [1368] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 6, 0, 132),
  [1370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 166),
  [1372] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 166),
  [1374] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exit_statement, 5, 0, 0),
  [1376] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exit_statement, 5, 0, 0),
  [1378] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unset_statement, 5, 0, 0),
  [1380] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unset_statement, 5, 0, 0),
  [1382] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 6, 0, 145),
  [1384] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 6, 0, 145),
  [1386] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 167),
  [1388] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 167),
  [1390] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 168),
  [1392] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 168),
  [1394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 6, 0, 146),
  [1396] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 6, 0, 146),
  [1398] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_declaration, 1, 0, 1),
  [1400] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_declaration, 1, 0, 1),
  [1402] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declare_statement, 5, 0, 0),
  [1404] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declare_statement, 5, 0, 0),
  [1406] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 169),
  [1408] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 169),
  [1410] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 170),
  [1412] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 170),
  [1414] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_statement, 5, 0, 23),
  [1416] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_statement, 5, 0, 23),
  [1418] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_statement, 5, 0, 94),
  [1420] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_statement, 5, 0, 94),
  [1422] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 6, 0, 149),
  [1424] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 6, 0, 149),
  [1426] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 6, 0, 114),
  [1428] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 6, 0, 114),
  [1430] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause, 2, 0, 6),
  [1432] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 2, 0, 6),
  [1434] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 5, 0, 95),
  [1436] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 5, 0, 95),
  [1438] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreach_statement, 7, 0, 171),
  [1440] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_foreach_statement, 7, 0, 171),
  [1442] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 5, 0, 23),
  [1444] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 5, 0, 23),
  [1446] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreach_statement, 7, 0, 0),
  [1448] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_foreach_statement, 7, 0, 0),
  [1450] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_empty_statement, 1, 0, 0),
  [1452] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_empty_statement, 1, 0, 0),
  [1454] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 6, 0, 150),
  [1456] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 6, 0, 150),
  [1458] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 6, 0, 52),
  [1460] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 6, 0, 52),
  [1462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_block, 3, 0, 0),
  [1464] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_block, 3, 0, 0),
  [1466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 6, 0, 53),
  [1468] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 6, 0, 53),
  [1470] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 7, 0, 174),
  [1472] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 7, 0, 174),
  [1474] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 7, 0, 176),
  [1476] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 7, 0, 176),
  [1478] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 6, 0, 0),
  [1480] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 6, 0, 0),
  [1482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declare_statement, 8, 0, 0),
  [1484] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declare_statement, 8, 0, 0),
  [1486] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_declaration, 5, 0, 110),
  [1488] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interface_declaration, 5, 0, 110),
  [1490] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_declaration, 3, 0, 15),
  [1492] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_declaration, 3, 0, 15),
  [1494] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 0),
  [1496] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 0),
  [1498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_declaration, 3, 0, 15),
  [1500] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interface_declaration, 3, 0, 15),
  [1502] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 5, 0, 110),
  [1504] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 5, 0, 110),
  [1506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 3, 0, 15),
  [1508] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 3, 0, 15),
  [1510] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_statement, 2, 0, 0),
  [1512] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_statement, 2, 0, 0),
  [1514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 3, 0, 15),
  [1516] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 3, 0, 15),
  [1518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 5, 0, 110),
  [1520] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 5, 0, 110),
  [1522] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 188),
  [1524] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 188),
  [1526] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 5, 0, 113),
  [1528] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 5, 0, 113),
  [1530] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 5, 0, 114),
  [1532] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 5, 0, 114),
  [1534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__const_declaration, 5, 0, 0),
  [1536] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__const_declaration, 5, 0, 0),
  [1538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_echo_statement, 3, 0, 0),
  [1540] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_echo_statement, 3, 0, 0),
  [1542] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 5, 0, 115),
  [1544] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 5, 0, 115),
  [1546] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 189),
  [1548] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 189),
  [1550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_goto_statement, 3, 0, 0),
  [1552] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_goto_statement, 3, 0, 0),
  [1554] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_continue_statement, 3, 0, 0),
  [1556] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_continue_statement, 3, 0, 0),
  [1558] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_break_statement, 3, 0, 0),
  [1560] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_break_statement, 3, 0, 0),
  [1562] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 3, 0, 0),
  [1564] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 3, 0, 0),
  [1566] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_statement, 3, 0, 23),
  [1568] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_statement, 3, 0, 23),
  [1570] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 190),
  [1572] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 190),
  [1574] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 6, 0, 131),
  [1576] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 6, 0, 131),
  [1578] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unset_statement, 6, 0, 0),
  [1580] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unset_statement, 6, 0, 0),
  [1582] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_declaration, 5, 0, 92),
  [1584] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_declaration, 5, 0, 92),
  [1586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [1588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1624),
  [1590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [1592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1615),
  [1594] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [1596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [1598] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 3, 0, 0),
  [1600] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 3, 0, 0),
  [1602] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [1604] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 5, 0, 0),
  [1606] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 5, 0, 0),
  [1608] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 2, 0, 0),
  [1610] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 2, 0, 0),
  [1612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 4, 0, 0),
  [1614] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 4, 0, 0),
  [1616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_access_expression, 3, 0, 31),
  [1618] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_access_expression, 3, 0, 31),
  [1620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [1622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nullsafe_member_access_expression, 3, 0, 31),
  [1624] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_nullsafe_member_access_expression, 3, 0, 31),
  [1626] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__variable_member_access_expression, 3, 0, 31),
  [1628] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__variable_member_access_expression, 3, 0, 31),
  [1630] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__variable_subscript_expression, 3, 0, 0),
  [1632] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__variable_subscript_expression, 3, 0, 0),
  [1634] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__new_variable, 1, 0, 4),
  [1636] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__new_variable, 1, 0, 4),
  [1638] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__reserved_identifier, 1, 0, 0),
  [1640] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__reserved_identifier, 1, 0, 0),
  [1642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__new_variable, 1, 0, 0),
  [1644] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__new_variable, 1, 0, 0),
  [1646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__callable_variable, 1, 0, 0),
  [1648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__variable_scoped_property_access_expression, 3, 0, 25),
  [1650] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__variable_scoped_property_access_expression, 3, 0, 25),
  [1652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_property_access_expression, 3, 0, 25),
  [1654] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scoped_property_access_expression, 3, 0, 25),
  [1656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__member_name, 1, 0, 5),
  [1658] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_name, 2, 0, 0),
  [1660] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable_name, 2, 0, 0),
  [1662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__member_name, 1, 0, 5),
  [1664] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__variable_nullsafe_member_access_expression, 3, 0, 31),
  [1666] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__variable_nullsafe_member_access_expression, 3, 0, 31),
  [1668] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__dereferencable_subscript_expression, 3, 0, 0),
  [1670] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__dereferencable_subscript_expression, 3, 0, 0),
  [1672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call_expression, 2, 0, 11),
  [1674] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call_expression, 2, 0, 11),
  [1676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dynamic_variable_name, 2, 0, 0),
  [1678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dynamic_variable_name, 2, 0, 0),
  [1680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 0),
  [1682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 0),
  [1684] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cast_variable, 4, 0, 50),
  [1686] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0),
  [1688] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__dereferencable_expression, 1, 0, 0),
  [1690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(989),
  [1692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cast_variable, 4, 0, 50),
  [1694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dynamic_variable_name, 4, 0, 0),
  [1696] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dynamic_variable_name, 4, 0, 0),
  [1698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__new_variable, 1, 0, 3),
  [1700] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__new_variable, 1, 0, 3),
  [1702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_call_expression, 4, 0, 68),
  [1704] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scoped_call_expression, 4, 0, 68),
  [1706] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_call_expression, 4, 0, 69),
  [1708] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_call_expression, 4, 0, 69),
  [1710] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nullsafe_member_call_expression, 4, 0, 69),
  [1712] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_nullsafe_member_call_expression, 4, 0, 69),
  [1714] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__dereferencable_subscript_expression, 4, 0, 0),
  [1716] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__dereferencable_subscript_expression, 4, 0, 0),
  [1718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__variable_subscript_expression, 4, 0, 0),
  [1720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__variable_subscript_expression, 4, 0, 0),
  [1722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__member_name, 3, 0, 16),
  [1724] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__member_name, 3, 0, 16),
  [1726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(134),
  [1728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(245),
  [1730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [1732] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1501),
  [1734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1381),
  [1736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [1738] = {.entry = {.count = 1, .reusable = false}}, SHIFT(865),
  [1740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1119),
  [1742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1387),
  [1744] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1501),
  [1747] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1519),
  [1750] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2504),
  [1753] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0),
  [1755] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(131),
  [1758] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2608),
  [1761] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(865),
  [1764] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(159),
  [1767] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1360),
  [1770] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1589),
  [1773] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2034),
  [1776] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1590),
  [1779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(234),
  [1781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [1783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1374),
  [1785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(248),
  [1787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [1789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2127),
  [1791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(241),
  [1793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [1795] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 0), REDUCE(sym__array_destructing_element, 1, 0, 0),
  [1798] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing_element, 1, 0, 0),
  [1800] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_destructing_repeat1, 2, 0, 0),
  [1802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2136),
  [1804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2063),
  [1806] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 0), REDUCE(sym__array_destructing_element, 3, 0, 0),
  [1809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2061),
  [1811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1991),
  [1813] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_qualified_name, 2, 0, 8),
  [1815] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_qualified_name, 2, 0, 8),
  [1817] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_namespace_name, 2, 0, 0), SHIFT(2493),
  [1820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [1822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1617),
  [1824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(125),
  [1826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [1828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1625),
  [1830] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_qualified_name, 3, 0, 26),
  [1832] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_qualified_name, 3, 0, 26),
  [1834] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_modifier, 1, 0, 0),
  [1836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1392),
  [1838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1291),
  [1840] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1509),
  [1842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1184),
  [1844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1448),
  [1846] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1045),
  [1848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1305),
  [1850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1336),
  [1852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2535),
  [1854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [1856] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_namespace_name, 1, 0, 0), SHIFT(2493),
  [1859] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__identifier, 1, 0, 0),
  [1861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [1863] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_constant_access_expression, 5, 0, 107),
  [1865] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_constant_access_expression, 5, 0, 107),
  [1867] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_constant_access_expression, 3, 0, 0),
  [1869] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_constant_access_expression, 3, 0, 0),
  [1871] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_by_ref, 2, 0, 0),
  [1873] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_by_ref, 2, 0, 0),
  [1875] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nowdoc, 6, 0, 141),
  [1877] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_nowdoc, 6, 0, 141),
  [1879] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 6, 0, 172),
  [1881] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 6, 0, 172),
  [1883] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 6, 0, 132),
  [1885] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 6, 0, 132),
  [1887] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 7, 0, 173),
  [1889] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 7, 0, 173),
  [1891] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nowdoc, 7, 0, 173),
  [1893] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_nowdoc, 7, 0, 173),
  [1895] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 7, 0, 194),
  [1897] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 7, 0, 194),
  [1899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1556),
  [1901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2479),
  [1903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2547),
  [1905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1318),
  [1907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1701),
  [1909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2293),
  [1911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1333),
  [1913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1445),
  [1915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2624),
  [1917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__name, 1, 0, 0),
  [1919] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_creation_expression, 1, 0, 0),
  [1921] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_creation_expression, 1, 0, 0),
  [1923] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 2),
  [1925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__dereferencable_expression, 1, 0, 2),
  [1927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__dereferencable_scalar, 1, 0, 0),
  [1929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_literal, 1, 0, 0),
  [1931] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_literal, 1, 0, 0),
  [1933] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 6, 0, 141),
  [1935] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 6, 0, 141),
  [1937] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__new_dereferencable_expression, 2, 0, 0),
  [1939] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__new_dereferencable_expression, 2, 0, 0),
  [1941] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_name_reference, 1, 0, 0),
  [1943] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_name_reference, 1, 0, 0),
  [1945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1864),
  [1947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1428),
  [1949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1431),
  [1951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [1953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 6, 0, 142),
  [1955] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 6, 0, 142),
  [1957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_update_expression, 2, 0, 9),
  [1959] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_update_expression, 2, 0, 9),
  [1961] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 2, 0, 0),
  [1963] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 2, 0, 0),
  [1965] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__array_destructing, 2, 0, 0), REDUCE(sym_array_creation_expression, 2, 0, 0),
  [1968] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__array_destructing, 2, 0, 0),
  [1970] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_encapsed_string, 2, 0, 0),
  [1972] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_encapsed_string, 2, 0, 0),
  [1974] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 2, 0, 0),
  [1976] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 2, 0, 0),
  [1978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2619),
  [1980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1826),
  [1982] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 3, 0, 22),
  [1984] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 3, 0, 22),
  [1986] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 2, 0, 6),
  [1988] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 2, 0, 6),
  [1990] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__new_dereferencable_expression, 3, 0, 0),
  [1992] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__new_dereferencable_expression, 3, 0, 0),
  [1994] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 3, 0, 0),
  [1996] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 3, 0, 0),
  [1998] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_encapsed_string, 3, 0, 0),
  [2000] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_encapsed_string, 3, 0, 0),
  [2002] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 3, 0, 0),
  [2004] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 3, 0, 0),
  [2006] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 3, 0, 24),
  [2008] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 3, 0, 24),
  [2010] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 4, 0, 22),
  [2012] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 4, 0, 22),
  [2014] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 3, 0, 36),
  [2016] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 3, 0, 36),
  [2018] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 3, 0, 54),
  [2020] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 3, 0, 54),
  [2022] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 4, 0, 0),
  [2024] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 4, 0, 0),
  [2026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 4, 0, 58),
  [2028] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 4, 0, 58),
  [2030] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 4, 0, 59),
  [2032] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 4, 0, 59),
  [2034] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 4, 0, 60),
  [2036] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 4, 0, 60),
  [2038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 5, 0, 22),
  [2040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 5, 0, 22),
  [2042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 4, 0, 97),
  [2044] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 4, 0, 97),
  [2046] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 4, 0, 98),
  [2048] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 4, 0, 98),
  [2050] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 5, 0, 0),
  [2052] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 5, 0, 0),
  [2054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 5, 0, 99),
  [2056] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 5, 0, 99),
  [2058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 5, 0, 100),
  [2060] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 5, 0, 100),
  [2062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 5, 0, 101),
  [2064] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 5, 0, 101),
  [2066] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 5, 0, 102),
  [2068] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 5, 0, 102),
  [2070] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nowdoc, 5, 0, 101),
  [2072] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_nowdoc, 5, 0, 101),
  [2074] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_creation_expression, 6, 0, 22),
  [2076] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_creation_expression, 6, 0, 22),
  [2078] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 5, 0, 136),
  [2080] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 5, 0, 136),
  [2082] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_class, 5, 0, 137),
  [2084] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_class, 5, 0, 137),
  [2086] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc, 6, 0, 140),
  [2088] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_heredoc, 6, 0, 140),
  [2090] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__identifier, 1, 0, 0),
  [2092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [2094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1823),
  [2096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1719),
  [2098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1651),
  [2100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1852),
  [2102] = {.entry = {.count = 1, .reusable = false}}, SHIFT(928),
  [2104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(713),
  [2106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2590),
  [2108] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1503),
  [2110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1353),
  [2112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [2114] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1008),
  [2116] = {.entry = {.count = 1, .reusable = false}}, SHIFT(830),
  [2118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(701),
  [2120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2632),
  [2122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1505),
  [2124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1348),
  [2126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [2128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(837),
  [2130] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__argument_name, 2, 0, 61),
  [2132] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__argument_name, 2, 0, 61),
  [2134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [2136] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_reference_modifier, 1, 0, 0),
  [2138] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_reference_modifier, 1, 0, 0),
  [2140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__new_non_dereferencable_expression, 2, 0, 0),
  [2142] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__new_non_dereferencable_expression, 2, 0, 0),
  [2144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [2146] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1886),
  [2148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1833),
  [2150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1416),
  [2152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1417),
  [2154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [2156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(854),
  [2158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(519),
  [2160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_expression, 3, 0, 29),
  [2162] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_expression, 3, 0, 29),
  [2164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_error_suppression_expression, 2, 0, 0),
  [2166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_error_suppression_expression, 2, 0, 0),
  [2168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_element_initializer, 1, 0, 0),
  [2170] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_element_initializer, 1, 0, 0),
  [2172] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 2, 0, 13),
  [2174] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 2, 0, 13),
  [2176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_clone_expression, 2, 0, 0),
  [2178] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_clone_expression, 2, 0, 0),
  [2180] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cast_expression, 4, 0, 50),
  [2182] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cast_expression, 4, 0, 50),
  [2184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_block, 2, 0, 0),
  [2186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_block, 2, 0, 0),
  [2188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression, 1, 0, 0),
  [2190] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression, 1, 0, 0),
  [2192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1307),
  [2194] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_update_expression, 2, 0, 14),
  [2196] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_update_expression, 2, 0, 14),
  [2198] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing, 2, 0, 0),
  [2200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_op_expression, 2, 0, 9),
  [2202] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_op_expression, 2, 0, 9),
  [2204] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shell_command_expression, 2, 0, 0),
  [2206] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_shell_command_expression, 2, 0, 0),
  [2208] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_expression, 3, 0, 23),
  [2210] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_expression, 3, 0, 23),
  [2212] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield_expression, 2, 0, 0),
  [2214] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_yield_expression, 2, 0, 0),
  [2216] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_block, 3, 0, 0),
  [2218] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_block, 3, 0, 0),
  [2220] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_null, 1, 0, 0),
  [2222] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_null, 1, 0, 0),
  [2224] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean, 1, 0, 0),
  [2226] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean, 1, 0, 0),
  [2228] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_block, 5, 0, 0),
  [2230] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_block, 5, 0, 0),
  [2232] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shell_command_expression, 3, 0, 0),
  [2234] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_shell_command_expression, 3, 0, 0),
  [2236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_block, 4, 0, 0),
  [2238] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_block, 4, 0, 0),
  [2240] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_element_initializer, 3, 0, 0),
  [2242] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_element_initializer, 3, 0, 0),
  [2244] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arrow_function, 3, 0, 28),
  [2246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(298),
  [2248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(246),
  [2250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(299),
  [2252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [2254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [2256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [2258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [2260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [2262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [2264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [2266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [2268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [2270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(309),
  [2272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [2274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(310),
  [2276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [2278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [2280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [2282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(313),
  [2284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [2286] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_reference_assignment_expression, 4, 0, 67),
  [2288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [2290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_expression, 2, 0, 0),
  [2292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_once_expression, 2, 0, 0),
  [2294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_throw_expression, 2, 0, 0),
  [2296] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_unpacking, 2, 0, 0),
  [2298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_print_intrinsic, 2, 0, 0),
  [2300] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_require_expression, 2, 0, 0),
  [2302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_require_once_expression, 2, 0, 0),
  [2304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield_expression, 3, 0, 0),
  [2306] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_expression, 3, 0, 30),
  [2308] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_expression, 5, 0, 106),
  [2310] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_conditional_expression, 5, 0, 106),
  [2312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_augmented_assignment_expression, 3, 0, 29),
  [2314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_expression, 4, 0, 66),
  [2316] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_conditional_expression, 4, 0, 66),
  [2318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0),
  [2320] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1291),
  [2323] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0),
  [2325] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1292),
  [2328] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1296),
  [2331] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1287),
  [2334] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1045),
  [2337] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1233),
  [2340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [2342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [2344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(278),
  [2346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [2348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [2350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [2352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(282),
  [2354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [2356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(261),
  [2358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(258),
  [2360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(262),
  [2362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [2364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [2366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [2368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [2370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [2372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [2374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [2376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(259),
  [2378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [2380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(356),
  [2382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(242),
  [2384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(357),
  [2386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [2388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(359),
  [2390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(360),
  [2392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [2394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [2396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [2398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(364),
  [2400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [2402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(366),
  [2404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(367),
  [2406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(367),
  [2408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(368),
  [2410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [2412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [2414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [2416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(371),
  [2418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [2420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1308),
  [2422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2415),
  [2424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [2426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [2428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(397),
  [2430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(243),
  [2432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(398),
  [2434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [2436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(400),
  [2438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(401),
  [2440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [2442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [2444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(407),
  [2446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(408),
  [2448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [2450] = {.entry = {.count = 1, .reusable = false}}, SHIFT(409),
  [2452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [2454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [2456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [2458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(412),
  [2460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [2462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(402),
  [2464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [2466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [2468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [2470] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 152),
  [2472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1430),
  [2474] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 118),
  [2476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_element, 3, 0, 86),
  [2478] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 177),
  [2480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 183),
  [2482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 7, 0, 201),
  [2484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 151),
  [2486] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 182),
  [2488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 158),
  [2490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2037),
  [2492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1490),
  [2494] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_expression, 3, 0, 0),
  [2496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [2498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [2500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(560),
  [2502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(573),
  [2504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expressions, 1, 0, 0),
  [2506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1704),
  [2508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [2510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(572),
  [2512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1954),
  [2514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2033),
  [2516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2036),
  [2518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1383),
  [2520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [2522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1388),
  [2524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [2526] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_condition_list, 1, 0, 0),
  [2528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_element, 3, 0, 0),
  [2530] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_variable_declaration, 3, 0, 34),
  [2532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [2534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [2536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__foreach_value, 1, 0, 0),
  [2538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument, 3, 0, 103),
  [2540] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 6, 0, 184),
  [2542] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1291),
  [2545] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1462),
  [2548] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1509),
  [2551] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1190),
  [2554] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0),
  [2556] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1292),
  [2559] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1296),
  [2562] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1287),
  [2565] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(704),
  [2568] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1233),
  [2571] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1449),
  [2574] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 5, 0, 155),
  [2576] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_match_condition_list_repeat1, 2, 0, 0),
  [2578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1291),
  [2580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1462),
  [2582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1509),
  [2584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1190),
  [2586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(960),
  [2588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1292),
  [2590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1296),
  [2592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1287),
  [2594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(704),
  [2596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1233),
  [2598] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 3, 0, 86),
  [2600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 5, 0, 160),
  [2602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 5, 0, 156),
  [2604] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument, 1, 0, 0),
  [2606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(517),
  [2608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3),
  [2610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_conditional_expression, 3, 0, 134),
  [2612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [2614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [2616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 4, 0, 126),
  [2618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 4, 0, 116),
  [2620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(705),
  [2622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_default_expression, 3, 0, 133),
  [2624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument, 2, 0, 63),
  [2626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(943),
  [2628] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 4, 0, 120),
  [2630] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument, 2, 0, 62),
  [2632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [2634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(478),
  [2636] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_visibility_modifier, 1, 0, 0),
  [2638] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_visibility_modifier, 1, 0, 0),
  [2640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2442),
  [2642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(633),
  [2644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(651),
  [2646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(953),
  [2648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(627),
  [2650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(919),
  [2652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(631),
  [2654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1927),
  [2656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(798),
  [2658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2116),
  [2660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1567),
  [2662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [2664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [2666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(609),
  [2668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(610),
  [2670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(737),
  [2672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(611),
  [2674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [2676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [2678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1542),
  [2680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(845),
  [2682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(850),
  [2684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(879),
  [2686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [2688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1016),
  [2690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(947),
  [2692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(961),
  [2694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(965),
  [2696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(968),
  [2698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(963),
  [2700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1580),
  [2702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(715),
  [2704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(744),
  [2706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(722),
  [2708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1583),
  [2710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(774),
  [2712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(775),
  [2714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(777),
  [2716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(717),
  [2718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(781),
  [2720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1527),
  [2722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(639),
  [2724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(656),
  [2726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1566),
  [2728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [2730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(645),
  [2732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [2734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(335),
  [2736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [2738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [2740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(583),
  [2742] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_readonly_modifier, 1, 0, 0),
  [2744] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_readonly_modifier, 1, 0, 0),
  [2746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(518),
  [2748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2629),
  [2750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1339),
  [2752] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_visibility_modifier, 4, 0, 49),
  [2754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_visibility_modifier, 4, 0, 49),
  [2756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1981),
  [2758] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_static_modifier, 1, 0, 0),
  [2760] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_final_modifier, 1, 0, 0),
  [2762] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_final_modifier, 1, 0, 0),
  [2764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2203),
  [2766] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1291),
  [2769] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1462),
  [2772] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1509),
  [2775] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0),
  [2777] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2629),
  [2780] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1292),
  [2783] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1296),
  [2786] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1287),
  [2789] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1339),
  [2792] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1233),
  [2795] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_declaration_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1449),
  [2798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(480),
  [2800] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_modifier, 1, 0, 0),
  [2802] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_modifier, 1, 0, 0),
  [2804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2328),
  [2806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1513),
  [2808] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1690),
  [2810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1312),
  [2812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1335),
  [2814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2173),
  [2816] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_attribute_list_repeat1, 2, 0, 0),
  [2818] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_list_repeat1, 2, 0, 0),
  [2820] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1445),
  [2823] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_list, 1, 0, 0),
  [2825] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_list, 1, 0, 0),
  [2827] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_group, 3, 0, 0),
  [2829] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_group, 3, 0, 0),
  [2831] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_group, 4, 0, 0),
  [2833] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_group, 4, 0, 0),
  [2835] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_group, 5, 0, 0),
  [2837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_group, 5, 0, 0),
  [2839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1314),
  [2841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1304),
  [2843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1454),
  [2845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2454),
  [2847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2455),
  [2849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2457),
  [2851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2459),
  [2853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1349),
  [2855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1781),
  [2857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2616),
  [2859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2541),
  [2861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2583),
  [2863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2432),
  [2865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1342),
  [2867] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1449),
  [2870] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_relative_scope, 1, 0, 0),
  [2872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(702),
  [2874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2563),
  [2876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1337),
  [2878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1045),
  [2880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1003),
  [2882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(999),
  [2884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1000),
  [2886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1187),
  [2888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2464),
  [2890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1184),
  [2892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2581),
  [2894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2607),
  [2896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1499),
  [2898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1510),
  [2900] = {.entry = {.count = 1, .reusable = false}}, SHIFT(383),
  [2902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1356),
  [2904] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1356),
  [2906] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_heredoc_body, 2, 0, 0), SHIFT(1390),
  [2909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1628),
  [2911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_heredoc_body, 2, 0, 0),
  [2913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1507),
  [2915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1344),
  [2917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1498),
  [2919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2499),
  [2921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1500),
  [2923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1345),
  [2925] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_use_list_repeat1, 2, 0, 0),
  [2927] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_heredoc_body_repeat1, 2, 0, 0), SHIFT_REPEAT(383),
  [2930] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1356),
  [2933] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_heredoc_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1356),
  [2936] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1390),
  [2939] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_heredoc_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1628),
  [2942] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 2, 0, 0),
  [2944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1496),
  [2946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1390),
  [2948] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 2, 0, 0), SHIFT_REPEAT(383),
  [2951] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 2, 0, 0), SHIFT_REPEAT(1355),
  [2954] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 2, 0, 0), SHIFT_REPEAT(1355),
  [2957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 2, 0, 0),
  [2959] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 2, 0, 0), SHIFT_REPEAT(1628),
  [2962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1355),
  [2964] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1355),
  [2966] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_heredoc_body_repeat1, 1, 0, 0),
  [2968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [2970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1372),
  [2972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1372),
  [2974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(898),
  [2976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1619),
  [2978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [2980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1373),
  [2982] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1373),
  [2984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1092),
  [2986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1622),
  [2988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1362),
  [2990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1362),
  [2992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(794),
  [2994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1363),
  [2996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1363),
  [2998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(992),
  [3000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [3002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1002),
  [3004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1472),
  [3006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1358),
  [3008] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1358),
  [3010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(945),
  [3012] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_declaration, 3, 0, 0),
  [3014] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 3, 0, 33),
  [3016] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 3, 0, 128),
  [3018] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 3, 0, 129),
  [3020] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 3, 0, 130),
  [3022] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body, 2, 0, 0), SHIFT_REPEAT(379),
  [3025] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body, 2, 0, 0), SHIFT_REPEAT(1372),
  [3028] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body, 2, 0, 0), SHIFT_REPEAT(1372),
  [3031] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body, 2, 0, 0),
  [3033] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_string_body, 2, 0, 0), SHIFT_REPEAT(1619),
  [3036] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_execution_operator_body, 2, 0, 0), SHIFT_REPEAT(338),
  [3039] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_execution_operator_body, 2, 0, 0), SHIFT_REPEAT(1373),
  [3042] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolated_execution_operator_body, 2, 0, 0), SHIFT_REPEAT(1373),
  [3045] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolated_execution_operator_body, 2, 0, 0),
  [3047] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolated_execution_operator_body, 2, 0, 0), SHIFT_REPEAT(1622),
  [3050] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_list, 2, 0, 0),
  [3052] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_declaration, 4, 0, 0),
  [3054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 4, 0, 162),
  [3056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 4, 0, 163),
  [3058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1359),
  [3060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1359),
  [3062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1112),
  [3064] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 2, 0, 12),
  [3066] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_declaration, 2, 0, 90),
  [3068] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_list, 3, 0, 0),
  [3070] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook_list, 2, 0, 0),
  [3072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_case, 5, 0, 186),
  [3074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1475),
  [3076] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 6, 0, 185),
  [3078] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_case, 6, 0, 202),
  [3080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1474),
  [3082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [3084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1357),
  [3086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1357),
  [3088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1628),
  [3090] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_const_declaration, 3, 0, 127),
  [3092] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_named_type, 1, 0, 0),
  [3094] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_named_type, 1, 0, 0),
  [3096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 3, 0, 0),
  [3098] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook_list, 3, 0, 0),
  [3100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_case, 3, 0, 16),
  [3102] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__member_declaration, 1, 0, 47),
  [3104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_const_declaration, 1, 0, 1),
  [3106] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 4, 0, 161),
  [3108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 4, 0, 0),
  [3110] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 4, 0, 46),
  [3112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_case, 4, 0, 82),
  [3114] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_const_declaration, 2, 0, 89),
  [3116] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_const_declaration, 2, 0, 91),
  [3118] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 5, 0, 161),
  [3120] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 5, 0, 185),
  [3122] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_declaration, 5, 0, 46),
  [3124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(718),
  [3126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(727),
  [3128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [3130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(843),
  [3132] = {.entry = {.count = 1, .reusable = false}}, SHIFT(869),
  [3134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [3136] = {.entry = {.count = 1, .reusable = false}}, SHIFT(917),
  [3138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(893),
  [3140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [3142] = {.entry = {.count = 4, .reusable = false}}, REDUCE(sym_type, 1, 0, 0), REDUCE(sym_union_type, 1, 0, 0), REDUCE(sym_intersection_type, 1, 0, 0), REDUCE(sym_disjunctive_normal_form_type, 1, -1, 0),
  [3147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1326),
  [3149] = {.entry = {.count = 4, .reusable = true}}, REDUCE(sym_type, 1, 0, 0), REDUCE(sym_union_type, 1, 0, 0), REDUCE(sym_intersection_type, 1, 0, 0), REDUCE(sym_disjunctive_normal_form_type, 1, -1, 0),
  [3154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1310),
  [3156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1489),
  [3158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2343),
  [3160] = {.entry = {.count = 1, .reusable = false}}, SHIFT(743),
  [3162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(709),
  [3164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [3166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(598),
  [3168] = {.entry = {.count = 1, .reusable = false}}, SHIFT(593),
  [3170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [3172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(938),
  [3174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [3176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1332),
  [3178] = {.entry = {.count = 1, .reusable = false}}, SHIFT(641),
  [3180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(636),
  [3182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [3184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1302),
  [3186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(765),
  [3188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [3190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1303),
  [3192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2407),
  [3194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1329),
  [3196] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_property_hook_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1489),
  [3199] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_property_hook_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1556),
  [3202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_property_hook_list_repeat1, 2, 0, 0),
  [3204] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_property_hook_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1292),
  [3207] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_property_hook_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1449),
  [3210] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__modifier, 1, 0, 0),
  [3212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1798),
  [3214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1811),
  [3216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1382),
  [3218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1394),
  [3220] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_optional_type, 2, 0, 0),
  [3222] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_optional_type, 2, 0, 0),
  [3224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1786),
  [3226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1806),
  [3228] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_primitive_type, 1, 0, 0),
  [3230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_primitive_type, 1, 0, 0),
  [3232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2412),
  [3234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(796),
  [3236] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0), REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0),
  [3239] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0), REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0),
  [3242] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 4, 0, 0),
  [3244] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 4, 0, 0),
  [3246] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0),
  [3248] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0),
  [3250] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1324),
  [3253] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_variable, 1, 0, 0),
  [3255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2625),
  [3257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1517),
  [3259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_variable, 1, 0, 0),
  [3261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1540),
  [3263] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_intersection_type_repeat1, 2, 0, 0),
  [3265] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_intersection_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1326),
  [3268] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_intersection_type_repeat1, 2, 0, 0),
  [3270] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0),
  [3272] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0),
  [3274] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1317),
  [3277] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_type, 2, 0, 0),
  [3279] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_type, 2, 0, 0),
  [3281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1324),
  [3283] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_disjunctive_normal_form_type, 4, -1, 0),
  [3285] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_disjunctive_normal_form_type, 4, -1, 0),
  [3287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1317),
  [3289] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_intersection_type, 2, 0, 0),
  [3291] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_intersection_type, 2, 0, 0),
  [3293] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_disjunctive_normal_form_type, 3, -1, 0),
  [3295] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_disjunctive_normal_form_type, 3, -1, 0),
  [3297] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_disjunctive_normal_form_type, 2, -1, 0),
  [3299] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_disjunctive_normal_form_type, 2, -1, 0),
  [3301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1667),
  [3303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(104),
  [3305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [3307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [3309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1762),
  [3311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1763),
  [3313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1765),
  [3315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2435),
  [3317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1558),
  [3319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1638),
  [3321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1766),
  [3323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1691),
  [3325] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type, 1, 0, 0),
  [3327] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type, 1, 0, 0),
  [3329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1683),
  [3331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1736),
  [3333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2477),
  [3335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1523),
  [3337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1642),
  [3339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(681),
  [3341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1931),
  [3343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1323),
  [3345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1309),
  [3347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1228),
  [3349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1460),
  [3351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1463),
  [3353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2247),
  [3355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1231),
  [3357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1769),
  [3359] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2635),
  [3362] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(2265),
  [3365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2093),
  [3367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(683),
  [3369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1968),
  [3371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2635),
  [3373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2265),
  [3375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2417),
  [3377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1393),
  [3379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2044),
  [3381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1435),
  [3383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2638),
  [3385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2611),
  [3387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1400),
  [3389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1366),
  [3391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1459),
  [3393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(642),
  [3395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_switch_block_repeat1, 2, 0, 0),
  [3397] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_switch_block_repeat1, 2, 0, 0), SHIFT_REPEAT(333),
  [3400] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_switch_block_repeat1, 2, 0, 0), SHIFT_REPEAT(2309),
  [3403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2425),
  [3405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1375),
  [3407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1404),
  [3409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1405),
  [3411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2480),
  [3413] = {.entry = {.count = 5, .reusable = true}}, REDUCE(sym_type, 1, 0, 0), REDUCE(sym_union_type, 1, 0, 0), REDUCE(sym_intersection_type, 1, 0, 0), REDUCE(sym_disjunctive_normal_form_type, 1, -1, 0), SHIFT(1319),
  [3419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1313),
  [3421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1406),
  [3423] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_string_part, 1, 0, 3),
  [3425] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_string_part, 1, 0, 3),
  [3427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1682),
  [3429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1708),
  [3431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing_element, 3, 0, 0),
  [3433] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body_heredoc, 2, 0, 0),
  [3435] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__complex_string_part, 3, 0, 0),
  [3437] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__complex_string_part, 3, 0, 0),
  [3439] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_string_member_access_expression, 3, 0, 57),
  [3441] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_string_member_access_expression, 3, 0, 57),
  [3443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1723),
  [3445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1738),
  [3447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1398),
  [3449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1743),
  [3451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1675),
  [3453] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_unset_statement_repeat1, 2, 0, 0),
  [3455] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_destructing_repeat1, 4, 0, 0),
  [3457] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_string_subscript_expression, 4, 0, 0),
  [3459] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_string_subscript_expression, 4, 0, 0),
  [3461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1386),
  [3463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1681),
  [3465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2626),
  [3467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1399),
  [3469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1761),
  [3471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2305),
  [3473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(111),
  [3475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1232),
  [3477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2254),
  [3479] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_if_statement, 3, 0, 23), SHIFT(2305),
  [3482] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_statement, 3, 0, 23), SHIFT(113),
  [3485] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_if_statement, 4, 0, 53), SHIFT(2305),
  [3488] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 53), SHIFT(113),
  [3491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(952),
  [3493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1645),
  [3495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1866),
  [3497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(795),
  [3499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1597),
  [3501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(597),
  [3503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [3505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(618),
  [3507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [3509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(805),
  [3511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1639),
  [3513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2176),
  [3515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2281),
  [3517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2282),
  [3519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1887),
  [3521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [3523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2309),
  [3525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1959),
  [3527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2031),
  [3529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2035),
  [3531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2040),
  [3533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1392),
  [3535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2048),
  [3537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(497),
  [3539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2086),
  [3541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(856),
  [3543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [3545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(920),
  [3547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [3549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1479),
  [3551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(420),
  [3553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1626),
  [3555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1483),
  [3557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(422),
  [3559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1627),
  [3561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(731),
  [3563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(424),
  [3565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(779),
  [3567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [3569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1586),
  [3571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1587),
  [3573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1471),
  [3575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [3577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1629),
  [3579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1538),
  [3581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2099),
  [3583] = {.entry = {.count = 5, .reusable = true}}, REDUCE(sym_type, 1, 0, 0), REDUCE(sym_union_type, 1, 0, 0), REDUCE(sym_intersection_type, 1, 0, 0), REDUCE(sym_disjunctive_normal_form_type, 1, -1, 0), SHIFT(1326),
  [3589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(548),
  [3591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2112),
  [3593] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolated_string_body, 2, 0, 0),
  [3595] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2, 0, 0),
  [3597] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1639),
  [3600] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1866),
  [3603] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolated_execution_operator_body, 2, 0, 0),
  [3605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2620),
  [3607] = {.entry = {.count = 1, .reusable = false}}, SHIFT(901),
  [3609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2440),
  [3611] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 2, 0, 7),
  [3613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1297),
  [3615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [3617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(515),
  [3619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1551),
  [3621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2181),
  [3623] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_formal_parameters, 5, 0, 0),
  [3625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1771),
  [3627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2622),
  [3629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1380),
  [3631] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_base_clause_repeat1, 2, 0, 0),
  [3633] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_base_clause_repeat1, 2, 0, 0), SHIFT_REPEAT(1459),
  [3636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1295),
  [3638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2395),
  [3640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2327),
  [3642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 2, 0, 38),
  [3644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [3646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1967),
  [3648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2109),
  [3650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1971),
  [3652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1877),
  [3654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1978),
  [3656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1982),
  [3658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1389),
  [3660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2011),
  [3662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1293),
  [3664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2280),
  [3666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 3, 0, 19),
  [3668] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 70),
  [3670] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 4, 0, 200),
  [3672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 4, 0, 200),
  [3674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2241),
  [3676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 73),
  [3678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(565),
  [3680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [3682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_name, 2, 0, 0),
  [3684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2556),
  [3686] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 5, 0, 205),
  [3688] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 5, 0, 205),
  [3690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2289),
  [3692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 3, 0, 78),
  [3694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(427),
  [3696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 3, 0, 79),
  [3698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(428),
  [3700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(476),
  [3702] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 5, 0, 206),
  [3704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 5, 0, 206),
  [3706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 5, 0, 207),
  [3708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 5, 0, 207),
  [3710] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 5, 0, 208),
  [3712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 5, 0, 208),
  [3714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2274),
  [3716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 111),
  [3718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(539),
  [3720] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 3, 0, 84),
  [3722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [3724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 117),
  [3726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [3728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(486),
  [3730] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 4, 0, 199),
  [3732] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 4, 0, 199),
  [3734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [3736] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1482),
  [3738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2141),
  [3740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1369),
  [3742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2143),
  [3744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2157),
  [3746] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 122),
  [3748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [3750] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 123),
  [3752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [3754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_formal_parameters, 2, 0, 0),
  [3756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2091),
  [3758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1371),
  [3760] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__property_hook_body, 3, 0, 6),
  [3762] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__property_hook_body, 3, 0, 6),
  [3764] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__function_definition_header, 3, 0, 21),
  [3766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1298),
  [3768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2185),
  [3770] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 6, 0, 212),
  [3772] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 6, 0, 212),
  [3774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(477),
  [3776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1320),
  [3778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2351),
  [3780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1306),
  [3782] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_element, 1, 0, 5),
  [3784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [3786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2200),
  [3788] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_name, 3, 0, 0),
  [3790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2645),
  [3792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2231),
  [3794] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_formal_parameters, 4, 0, 0),
  [3796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1377),
  [3798] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 4, 0, 195),
  [3800] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 4, 0, 195),
  [3802] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1321),
  [3805] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1315),
  [3808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1315),
  [3810] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_namespace_name_repeat1, 2, 0, 0),
  [3812] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_namespace_name_repeat1, 2, 0, 0), SHIFT_REPEAT(2556),
  [3815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1855),
  [3817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1441),
  [3819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1442),
  [3821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [3823] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat1, 2, 0, 96), SHIFT_REPEAT(2305),
  [3826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(516),
  [3828] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat2, 2, 0, 96),
  [3830] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat2, 2, 0, 96), SHIFT_REPEAT(2281),
  [3833] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat2, 2, 0, 96),
  [3835] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 2, 0, 153),
  [3837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 2, 0, 153),
  [3839] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__property_hook_body, 1, 0, 154),
  [3841] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__property_hook_body, 1, 0, 154),
  [3843] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 4, 0, 196),
  [3845] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 4, 0, 196),
  [3847] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 157),
  [3849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [3851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [3853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2409),
  [3855] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 4, 0, 198),
  [3857] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 4, 0, 198),
  [3859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(473),
  [3861] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_property_declaration_repeat1, 2, 0, 0),
  [3863] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_property_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(2044),
  [3866] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 3, 0, 27),
  [3868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(457),
  [3870] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 3, 0, 32),
  [3872] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_formal_parameters, 3, 0, 0),
  [3874] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__function_definition_header, 4, 0, 42),
  [3876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(469),
  [3878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1321),
  [3880] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 64),
  [3882] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 3, 0, 178),
  [3884] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 3, 0, 178),
  [3886] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 3, 0, 179),
  [3888] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 3, 0, 179),
  [3890] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 3, 0, 180),
  [3892] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 3, 0, 180),
  [3894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1963),
  [3896] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 3, 0, 181),
  [3898] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 3, 0, 181),
  [3900] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_hook, 4, 0, 197),
  [3902] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_hook, 4, 0, 197),
  [3904] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_union_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1320),
  [3907] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_intersection_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1323),
  [3910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2316),
  [3912] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1818),
  [3914] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_disjunctive_normal_form_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1306),
  [3917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_name, 1, 0, 0),
  [3919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2100),
  [3921] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_clause, 1, 0, 0),
  [3923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2514),
  [3925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__scope_resolution_qualifier, 1, 0, 0),
  [3927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1413),
  [3929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1418),
  [3931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [3933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2526),
  [3935] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_intersection_type_repeat1, 2, 0, 0), SHIFT_REPEAT(1319),
  [3938] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_base_clause_repeat1, 2, 0, 0), SHIFT_REPEAT(1477),
  [3941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1401),
  [3943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [3945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2041),
  [3947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [3949] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_namespace_use_declaration_repeat1, 2, 0, 0),
  [3951] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_namespace_use_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1389),
  [3954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1426),
  [3956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1427),
  [3958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [3960] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_list, 2, 0, 0),
  [3962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1728),
  [3964] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_literal, 1, 0, 0),
  [3966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2131),
  [3968] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_list, 1, 0, 0),
  [3970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1423),
  [3972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1437),
  [3974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [3976] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_static_declaration_repeat1, 2, 0, 0),
  [3978] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_static_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(2109),
  [3981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing, 3, 0, 0),
  [3983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1477),
  [3985] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_base_clause, 2, 0, 0),
  [3987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nowdoc_body, 2, 0, 0),
  [3989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1847),
  [3991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1439),
  [3993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1440),
  [3995] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0),
  [3997] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(1812),
  [4000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2118),
  [4002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1354),
  [4004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(911),
  [4006] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing, 4, 0, 0),
  [4008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_clause, 2, 0, 17),
  [4010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2458),
  [4012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2431),
  [4014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1443),
  [4016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1408),
  [4018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [4020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1923),
  [4022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(813),
  [4024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1817),
  [4026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__const_declaration_repeat1, 2, 0, 0),
  [4028] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__const_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1551),
  [4031] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_list_repeat1, 2, 0, 0),
  [4033] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1728),
  [4036] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 1, 0, 0),
  [4038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1414),
  [4040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1415),
  [4042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [4044] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_variable_declaration, 1, 0, 5),
  [4046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [4048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1407),
  [4050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1420),
  [4052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [4054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_nowdoc_body_repeat1, 2, 0, 0),
  [4056] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_nowdoc_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1847),
  [4059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1899),
  [4061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(806),
  [4063] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_intersection_type, 2, 0, 0), SHIFT(1319),
  [4066] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_content, 1, 0, 0),
  [4068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1812),
  [4070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2105),
  [4072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(902),
  [4074] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_base_clause, 3, 0, 0),
  [4076] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_global_declaration_repeat1, 2, 0, 0),
  [4078] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_global_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1877),
  [4081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1395),
  [4083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [4085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [4087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(240),
  [4089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(502),
  [4091] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface_clause, 2, 0, 0),
  [4093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(834),
  [4095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1753),
  [4097] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 3, 0, 7),
  [4099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [4101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(807),
  [4103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1612),
  [4105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1614),
  [4107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 4, 0, 70),
  [4109] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__function_definition_header, 5, 0, 88),
  [4111] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(157),
  [4114] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_arguments_repeat1, 2, 0, 0),
  [4116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [4118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1429),
  [4120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(811),
  [4122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1918),
  [4124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1919),
  [4126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [4128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(584),
  [4130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1921),
  [4132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1928),
  [4134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(835),
  [4136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [4138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [4140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 64),
  [4142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(674),
  [4144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [4146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(996),
  [4148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1937),
  [4150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1938),
  [4152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1939),
  [4154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [4156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2249),
  [4158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(822),
  [4160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [4162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1941),
  [4164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(525),
  [4166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1943),
  [4168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1946),
  [4170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(526),
  [4172] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 70),
  [4174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1948),
  [4176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(179),
  [4178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(675),
  [4180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1952),
  [4182] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 73),
  [4184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 5, 0, 111),
  [4186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1953),
  [4188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1385),
  [4190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2332),
  [4192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1955),
  [4194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1956),
  [4196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1958),
  [4198] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 4, 0, 73),
  [4200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(501),
  [4202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1481),
  [4204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [4206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(846),
  [4208] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_group_repeat1, 2, 0, 0), SHIFT_REPEAT(1451),
  [4211] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_group_repeat1, 2, 0, 0),
  [4213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1997),
  [4215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [4217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(909),
  [4219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_intersection_type, 1, 0, 0),
  [4221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2228),
  [4223] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_clause, 4, 0, 75),
  [4225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2390),
  [4227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2540),
  [4229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(503),
  [4231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [4233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(948),
  [4235] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 2, 0, 39),
  [4237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [4239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [4241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(504),
  [4243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2121),
  [4245] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_formal_parameters_repeat1, 2, 0, 0), SHIFT_REPEAT(866),
  [4248] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_formal_parameters_repeat1, 2, 0, 0),
  [4250] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 3, 0, 81),
  [4252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [4254] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 4, 0, 125),
  [4256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [4258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2240),
  [4260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(499),
  [4262] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 3, 0, 83),
  [4264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [4266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [4268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(719),
  [4270] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 2, 0, 40),
  [4272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [4274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2238),
  [4276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [4278] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 3, 0, 85),
  [4280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(432),
  [4282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1711),
  [4284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2248),
  [4286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2463),
  [4288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1742),
  [4290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 19),
  [4292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [4294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1384),
  [4296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [4298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(770),
  [4300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(536),
  [4302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [4304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(537),
  [4306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2163),
  [4308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2045),
  [4310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2169),
  [4312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2452),
  [4314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2580),
  [4316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1853),
  [4318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2172),
  [4320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2032),
  [4322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(543),
  [4324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(505),
  [4326] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat2, 1, 0, 51),
  [4328] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat2, 1, 0, 51),
  [4330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2002),
  [4332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [4334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2012),
  [4336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2323),
  [4338] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_interface_clause, 3, 0, 0),
  [4340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [4342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2365),
  [4344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(577),
  [4346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 3, 0, 0),
  [4348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(509),
  [4350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(513),
  [4352] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_creation_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(172),
  [4355] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_array_creation_expression_repeat1, 2, 0, 0),
  [4357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [4359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1048),
  [4361] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unset_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(697),
  [4364] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 4, 0, 64),
  [4366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2397),
  [4368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_as_clause, 3, 0, 0),
  [4370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_if_clause_2, 3, 0, 23),
  [4372] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_if_clause_2, 3, 0, 23),
  [4374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(455),
  [4376] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 2, 0, 37),
  [4378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [4380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [4382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1815),
  [4384] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2, 0, 0), SHIFT_REPEAT(185),
  [4387] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2, 0, 0),
  [4389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [4391] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_condition_list_repeat1, 2, 0, 0), SHIFT_REPEAT(281),
  [4394] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__array_destructing_repeat1, 2, 0, 0), SHIFT_REPEAT(170),
  [4397] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__array_destructing_repeat1, 2, 0, 0),
  [4399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2295),
  [4401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1422),
  [4403] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_intersection_type, 2, 0, 0), SHIFT(1326),
  [4406] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 6, 0, 0),
  [4408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2166),
  [4410] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 6, 0, 139),
  [4412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2167),
  [4414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(547),
  [4416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(549),
  [4418] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 3, 0, 19),
  [4420] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 6, 0, 111),
  [4422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [4424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(895),
  [4426] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_clause, 3, 0, 35),
  [4428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [4430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [4432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [4434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(942),
  [4436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [4438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(484),
  [4440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1478),
  [4442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(489),
  [4444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1989),
  [4446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(490),
  [4448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(907),
  [4450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [4452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1819),
  [4454] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_destructing_repeat1, 2, 0, 55),
  [4456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(491),
  [4458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(492),
  [4460] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_anonymous_function_use_clause_repeat1, 2, 0, 0), SHIFT_REPEAT(1869),
  [4463] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_anonymous_function_use_clause_repeat1, 2, 0, 0),
  [4465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2095),
  [4467] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 4, 0, 0),
  [4469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1351),
  [4471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2138),
  [4473] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 4, 0, 56),
  [4475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2139),
  [4477] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__list_destructing_repeat1, 2, 0, 0), SHIFT_REPEAT(182),
  [4480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 2, 0, 7),
  [4482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2188),
  [4484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 5, 0, 0),
  [4486] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 5, 0, 56),
  [4488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2189),
  [4490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(553),
  [4492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing_element, 1, 0, 10),
  [4494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(556),
  [4496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1801),
  [4498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [4500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(936),
  [4502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [4504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(623),
  [4506] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_namespace_use_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1385),
  [4509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1432),
  [4511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1330),
  [4513] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 3, 0, 27),
  [4515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2610),
  [4517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2612),
  [4519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1867),
  [4521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(493),
  [4523] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 7, 0, 0),
  [4525] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__list_destructing, 7, 0, 139),
  [4527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [4529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1885),
  [4531] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_parameter, 1, 0, 5),
  [4533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [4535] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__return_type, 2, 0, 43),
  [4537] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__function_definition_header, 4, 0, 45),
  [4539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2027),
  [4541] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 32),
  [4543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1985),
  [4545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(498),
  [4547] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_creation_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(173),
  [4550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1892),
  [4552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2028),
  [4554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1898),
  [4556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1600),
  [4558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1601),
  [4560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [4562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(803),
  [4564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1424),
  [4566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1301),
  [4568] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_destructing_element, 3, 0, 56),
  [4570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(833),
  [4572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [4574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [4576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(468),
  [4578] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 3, 0, 32),
  [4580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(567),
  [4582] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 27),
  [4584] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 8, 0, 201),
  [4586] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 4, 0, 119),
  [4588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 4, 0, 121),
  [4590] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 3, 0, 80),
  [4592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__namespace_use_group, 3, 0, 36),
  [4594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [4596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(102),
  [4598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [4600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 84),
  [4602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 4, 0, 124),
  [4604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2430),
  [4606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(958),
  [4608] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 3, 0, 82),
  [4610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1856),
  [4612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_instead_of_clause, 3, 0, 0),
  [4614] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function_use_clause, 4, 0, 0),
  [4616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1807),
  [4618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2434),
  [4620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(962),
  [4622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [4624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [4626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [4628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [4630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1841),
  [4632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2427),
  [4634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(812),
  [4636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1829),
  [4638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [4640] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cast_type, 1, 0, 0),
  [4642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 7, 0, 182),
  [4644] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 2, 0, 16),
  [4646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 7, 0, 183),
  [4648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6),
  [4650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 2, 0, 7),
  [4652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__list_destructing_repeat1, 4, 0, 138),
  [4654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2506),
  [4656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(818),
  [4658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2508),
  [4660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(821),
  [4662] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__array_destructing_element, 3, 0, 0), REDUCE(sym_array_element_initializer, 3, 0, 0),
  [4665] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_group, 4, 0, 0),
  [4667] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 117),
  [4669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2577),
  [4671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(910),
  [4673] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 118),
  [4675] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 122),
  [4677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2428),
  [4679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(905),
  [4681] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 5, 0, 123),
  [4683] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 5, 0, 159),
  [4685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function_use_clause, 5, 0, 0),
  [4687] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_namespace_name_repeat1, 2, 0, 0), SHIFT_REPEAT(2493),
  [4690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2530),
  [4692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [4694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2531),
  [4696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(748),
  [4698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__namespace_use_group, 4, 0, 76),
  [4700] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__array_destructing_element, 1, 0, 0), REDUCE(sym_array_element_initializer, 1, 0, 0),
  [4703] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 3, 0, 77),
  [4705] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_use_group, 3, 0, 0),
  [4707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2586),
  [4709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(908),
  [4711] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 151),
  [4713] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 152),
  [4715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1865),
  [4717] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_as_clause, 4, 0, 0),
  [4719] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 78),
  [4721] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 4, 0, 79),
  [4723] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 157),
  [4725] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 6, 0, 158),
  [4727] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function_use_clause, 6, 0, 0),
  [4729] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_namespace_name, 3, 0, 0), SHIFT(2493),
  [4732] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 7, 0, 177),
  [4734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_promotion_parameter, 3, 0, 38),
  [4736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1444),
  [4738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [4740] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause_2, 2, 0, 6),
  [4742] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 5, 0, 104),
  [4744] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 41),
  [4746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1570),
  [4748] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 71),
  [4750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(819),
  [4752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(957),
  [4754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 3, 0, 0),
  [4756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(890),
  [4758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2534),
  [4760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1596),
  [4762] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 4, 0, 71),
  [4764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(966),
  [4766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1569),
  [4768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1837),
  [4770] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 6, 0, 143),
  [4772] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 6, 0, 144),
  [4774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1862),
  [4776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1773),
  [4778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2076),
  [4780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2560),
  [4782] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 6, 0, 147),
  [4784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 6, 0, 148),
  [4786] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 6, 0, 147),
  [4788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2266),
  [4790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [4792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [4794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1421),
  [4796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1669),
  [4798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1822),
  [4800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1419),
  [4802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2287),
  [4804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1832),
  [4806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1970),
  [4808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1676),
  [4810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1974),
  [4812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1603),
  [4814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [4816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [4818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [4820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2383),
  [4822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1604),
  [4824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [4826] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 65),
  [4828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2015),
  [4830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [4832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1485),
  [4834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1799),
  [4836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [4838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [4840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [4842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1486),
  [4844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [4846] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 3, 0, 20),
  [4848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1568),
  [4850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2617),
  [4852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [4854] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 108),
  [4856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(16),
  [4858] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 109),
  [4860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [4862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(10),
  [4864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [4866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2170),
  [4868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(9),
  [4870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [4872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(686),
  [4874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [4876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2496),
  [4878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1412),
  [4880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1924),
  [4882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2098),
  [4884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(685),
  [4886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1564),
  [4888] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 87),
  [4890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(684),
  [4892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1851),
  [4894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [4896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(791),
  [4898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1758),
  [4900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [4902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [4904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [4906] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 4, 0, 65),
  [4908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2106),
  [4910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(874),
  [4912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2047),
  [4914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2187),
  [4916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2525),
  [4918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(441),
  [4920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(28),
  [4922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(29),
  [4924] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 5, 0, 108),
  [4926] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 3, 0, 20),
  [4928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1338),
  [4930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1598),
  [4932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [4934] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 4, 0, 41),
  [4936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(751),
  [4938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(752),
  [4940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2376),
  [4942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2206),
  [4944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [4946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [4948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2271),
  [4950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1825),
  [4952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(20),
  [4954] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 112),
  [4956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2115),
  [4958] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declare_directive, 3, 0, 0),
  [4960] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreach_pair, 3, 0, 0),
  [4962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1573),
  [4964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(31),
  [4966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2369),
  [4968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1917),
  [4970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [4972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [4974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [4976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [4978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1788),
  [4980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(40),
  [4982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2505),
  [4984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [4986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1289),
  [4988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(690),
  [4990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1789),
  [4992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(32),
  [4994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2515),
  [4996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2344),
  [4998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(17),
  [5000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(696),
  [5002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1425),
  [5004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [5006] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 4, 0, 44),
  [5008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 7, 0, 175),
  [5010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1470),
  [5012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(691),
  [5014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1779),
  [5016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(906),
  [5018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(21),
  [5020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1455),
  [5022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2495),
  [5024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1630),
  [5026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_string_subscript_unary_expression, 2, 0, 0),
  [5028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1655),
  [5030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(34),
  [5032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [5034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(959),
  [5036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2220),
  [5038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 104),
  [5040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1778),
  [5042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [5044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2029),
  [5046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1663),
  [5048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(37),
  [5050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1593),
  [5052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [5054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(14),
  [5056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [5058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(42),
  [5060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [5062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2052),
  [5064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1438),
  [5066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(689),
  [5068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(439),
  [5070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1488),
  [5072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1571),
  [5074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [5076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1814),
  [5078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2582),
  [5080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2456),
  [5082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1821),
  [5084] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anonymous_function_header, 5, 0, 105),
  [5086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [5088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2353),
  [5090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1010),
  [5092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(700),
  [5094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2242),
  [5096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1480),
  [5098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1746),
  [5100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2225),
  [5102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1543),
  [5104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1592),
  [5106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2013),
  [5108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1872),
  [5110] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [5112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [5114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [5116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2059),
  [5118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1759),
  [5120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1608),
  [5122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(840),
  [5124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [5126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1554),
  [5128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [5130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [5132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [5134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1874),
  [5136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [5138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1613),
  [5140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2618),
  [5142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [5144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1409),
  [5146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [5148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [5150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [5152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__arrow_function_header, 5, 0, 112),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token__automatic_semicolon = 0,
  ts_external_token_encapsed_string_chars = 1,
  ts_external_token_encapsed_string_chars_after_variable = 2,
  ts_external_token_execution_string_chars = 3,
  ts_external_token_execution_string_chars_after_variable = 4,
  ts_external_token_encapsed_string_chars_heredoc = 5,
  ts_external_token_encapsed_string_chars_after_variable_heredoc = 6,
  ts_external_token__eof = 7,
  ts_external_token_heredoc_start = 8,
  ts_external_token_heredoc_end = 9,
  ts_external_token_nowdoc_string = 10,
  ts_external_token_sentinel_error = 11,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__automatic_semicolon] = sym__automatic_semicolon,
  [ts_external_token_encapsed_string_chars] = sym_encapsed_string_chars,
  [ts_external_token_encapsed_string_chars_after_variable] = sym_encapsed_string_chars_after_variable,
  [ts_external_token_execution_string_chars] = sym_execution_string_chars,
  [ts_external_token_execution_string_chars_after_variable] = sym_execution_string_chars_after_variable,
  [ts_external_token_encapsed_string_chars_heredoc] = sym_encapsed_string_chars_heredoc,
  [ts_external_token_encapsed_string_chars_after_variable_heredoc] = sym_encapsed_string_chars_after_variable_heredoc,
  [ts_external_token__eof] = sym__eof,
  [ts_external_token_heredoc_start] = sym_heredoc_start,
  [ts_external_token_heredoc_end] = sym_heredoc_end,
  [ts_external_token_nowdoc_string] = sym_nowdoc_string,
  [ts_external_token_sentinel_error] = sym_sentinel_error,
};

static const bool ts_external_scanner_states[14][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__automatic_semicolon] = true,
    [ts_external_token_encapsed_string_chars] = true,
    [ts_external_token_encapsed_string_chars_after_variable] = true,
    [ts_external_token_execution_string_chars] = true,
    [ts_external_token_execution_string_chars_after_variable] = true,
    [ts_external_token_encapsed_string_chars_heredoc] = true,
    [ts_external_token_encapsed_string_chars_after_variable_heredoc] = true,
    [ts_external_token__eof] = true,
    [ts_external_token_heredoc_start] = true,
    [ts_external_token_heredoc_end] = true,
    [ts_external_token_nowdoc_string] = true,
    [ts_external_token_sentinel_error] = true,
  },
  [2] = {
    [ts_external_token__automatic_semicolon] = true,
  },
  [3] = {
    [ts_external_token_encapsed_string_chars_heredoc] = true,
    [ts_external_token_heredoc_end] = true,
  },
  [4] = {
    [ts_external_token_encapsed_string_chars_heredoc] = true,
  },
  [5] = {
    [ts_external_token_encapsed_string_chars] = true,
  },
  [6] = {
    [ts_external_token_execution_string_chars] = true,
  },
  [7] = {
    [ts_external_token_encapsed_string_chars_heredoc] = true,
    [ts_external_token_encapsed_string_chars_after_variable_heredoc] = true,
    [ts_external_token_heredoc_end] = true,
  },
  [8] = {
    [ts_external_token_encapsed_string_chars] = true,
    [ts_external_token_encapsed_string_chars_after_variable] = true,
  },
  [9] = {
    [ts_external_token_execution_string_chars] = true,
    [ts_external_token_execution_string_chars_after_variable] = true,
  },
  [10] = {
    [ts_external_token_heredoc_end] = true,
    [ts_external_token_nowdoc_string] = true,
  },
  [11] = {
    [ts_external_token_heredoc_end] = true,
  },
  [12] = {
    [ts_external_token_heredoc_start] = true,
  },
  [13] = {
    [ts_external_token_nowdoc_string] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_php_only_external_scanner_create(void);
void tree_sitter_php_only_external_scanner_destroy(void *);
bool tree_sitter_php_only_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_php_only_external_scanner_serialize(void *, char *);
void tree_sitter_php_only_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_php_only(void) {
  static const TSLanguage language = {
    .version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym_name,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_php_only_external_scanner_create,
      tree_sitter_php_only_external_scanner_destroy,
      tree_sitter_php_only_external_scanner_scan,
      tree_sitter_php_only_external_scanner_serialize,
      tree_sitter_php_only_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
